#!/usr/bin/perl
#
# get_cuda_toolkit_version.pl
#
use File::Basename;
use FileHandle;
use Getopt::Long;
use English;

use strict 'vars';

$::g_prog_name = basename( $0 ); 
my $cuda_include_dir ;
my $cuda_h;
my $fh;
my $line;
my $cuda_toolkit_version = '0.0';
my $cuda_version_tag;
my $remaining_char;
my $preproc;
my $bCudaToolkitInternalVersion = 0;

my %cuda_version_map = (
  '2020'     => '2.2',
  '2030b'    => '2.3b',
  '2030'     => '2.3'
);

my %optctl = (
  'internal-version'           => \$bCudaToolkitInternalVersion
);


# parse the command line 
if( GetOptions( \%optctl, 'internal-version!' ) != 1 ) {
  fatal_error( __LINE__, "command line not understood, check the syntax\n" );
}

if( ! exists( $ENV{'CUDA_INC_PATH'} ) ) {
  print STDERR "$::g_prog_name: error: CUDA_INC_PATH is missing in your environment.\n"; 
  die;
}

if( ! exists( $ENV{'CUDA_BIN_PATH'} ) ) {
  print STDERR "$::g_prog_name: error: CUDA_BIN_PATH is missing in your environment.\n"; 
  die;
}

$cuda_h = File::Spec->catfile( $ENV{'CUDA_INC_PATH'} , 'cuda.h' );

$fh = new FileHandle $cuda_h, "r";
if ( ! defined $fh ) {
  print STDERR "$::g_prog_name: error: file \"$cuda_h\" not found\n";
  die;
}

while( defined( $fh ) && ( ! $fh->eof() ) ) {
  $line = $fh->getline();
  chomp;
  
  #    #define CUDA_VERSION 2020 /* 2.2 */
  if ( $line =~ /^#define[ \t]+CUDA_VERSION/ ) {
    ( $preproc, $cuda_version_tag, $cuda_toolkit_version, $remaining_char ) = split( '[ \t]+', $line, 4 );
    undef $fh;

  }
}

if( defined( $fh ) ) {
  $fh->close();
}

if( $cuda_toolkit_version eq '2030' ) {
  #either cuda 2.3 or 2.3b
  my $nvcc_version_resp;
  my @lines;
  my $line;
  my $release_string;
  my $nvcc_release;
  my $cuda_tool_string;
  my $cuda_tool_internal_version;
  my $nvcc = File::Spec->catfile( $ENV{'CUDA_BIN_PATH'} , 'nvcc' );
  
  if( $OSNAME eq  'MSWin32' ) {
    $nvcc .= '.exe';
  }
  
  $nvcc_version_resp = `$nvcc --version`;
  if( $? != 0 ) {
    print STDERR "$::g_prog_name: internal error: nvcc does not seem to understand --version anymore\n";
    exit 1;
  }
    
  @lines = split /\n/, $nvcc_version_resp;
  foreach $line (@lines) {
    if( $line =~ /^Cuda compilation tools/ ) {
      ($cuda_tool_string, $release_string, $cuda_tool_internal_version ) = split( ',', $line, 3 );
      if( $release_string =~ /release\s+([0-9.]+)/ ) {
        $nvcc_release = $1;
        if( $nvcc_release ne '2.3' ) {
          $cuda_toolkit_version = '2030b';
        }
      }
    }
  }
}

if( $bCudaToolkitInternalVersion ) {
  print "$cuda_toolkit_version";
}
else {
  if( exists( $cuda_version_map{ $cuda_toolkit_version } ) ) {
    print "$cuda_version_map{ $cuda_toolkit_version }"; 
  }
  else {
    print STDERR "$::g_prog_name: error: the internal CUDA version ID $cuda_toolkit_version is not understood.\n";
    die;
  }
}
