#!/usr/bin/perl
#
# parse_pkg_filename.pl
#
use File::Basename;
use Getopt::Long;
use English;

use strict 'vars';

$::g_script_error = "parse_pkg_filename.pl: error:";
my $b_help = 0;
my $b_print_arch = 0;
my $b_print_version = 0;
my $b_print_pkg_name = 0;
my $pkg_filename;
my %pkg_info_tokens;
my $pkg_filename_tmp;
my @pkg_token_list;
my $pkg_token_cnt;
my $pkg_token;

my %optctl = (
  'help'               => \$b_help,
  'arch'               => \$b_print_arch,
  'version'            => \$b_print_version,
  'name'               => \$b_print_pkg_name
);


# parse the command line 
if( GetOptions( \%optctl, 'help!', 'arch!', 'version!', 'name!' ) != 1 ) {
  die "$::g_script_error command line not understood, check the syntax\n";
}

if( $b_help ) {
  usage();
  exit 0;
}

if( scalar( @ARGV ) != 1 ) {
  usage();
  exit 1;
}

if( ( $b_print_arch + $b_print_version +  $b_print_pkg_name ) == 0 ) {
  die "$::g_script_error no package attribute selected; you may use --help to get the list of supported package attributes."; 
}

if( ! -f $ARGV[ 0 ] ) {
  die "$::g_script_error the package file \"$ARGV[ 0 ]\" does not exist.";
}
else {
  $pkg_filename = basename( $ARGV[ 0 ] );
}

$pkg_filename_tmp = $pkg_filename;
$pkg_filename_tmp =~ s/\.deb$//;

#print "pkg: $pkg_filename_tmp\n";

# Two variants of package filenames are supported:
#
#   package_version_arch.deb
#   package_version_codename_arch.deb
#   

@pkg_token_list = split( '_', $pkg_filename_tmp );

#print "tokens: @pkg_token_list \n";

$pkg_token_cnt = scalar( @pkg_token_list );

if( ( $pkg_token_cnt != 3 ) && ( $pkg_token_cnt != 4 ) ) {
  die "$::g_script_error the package filename \"$pkg_filename\" is not understood.";
}

# add the tokens into the hash
$pkg_token = shift @pkg_token_list;
$pkg_info_tokens{ 'name' } = $pkg_token;

$pkg_token = shift @pkg_token_list;
$pkg_info_tokens{ 'version' } = $pkg_token;

if( scalar( @pkg_token_list ) == 2 ) {
  # ignore the codename 
  $pkg_token = shift @pkg_token_list;
}

$pkg_token = shift @pkg_token_list;
$pkg_info_tokens{ 'arch' } = $pkg_token;


print "$pkg_info_tokens{ 'name' }\n" if( $b_print_pkg_name );
print "$pkg_info_tokens{ 'version' }\n" if( $b_print_version );
print "$pkg_info_tokens{ 'arch' }\n" if( $b_print_arch );



sub usage {
  my $how_to_call = <<"__END_OF_USAGE__";

Usage: parse_pkg_filename.pl [options] pkg_file

  The script parse_pkg_filename.pl analyzes a package filename to extract 
  the package name, version and architecture. 

  options:=
     --help              show usage information
     
     --arch              print the package architecture [default: off]
     
     --version           print the package version [default: off]
     
     --name              print the package name [default: off]


__END_OF_USAGE__

  print $how_to_call;
}