#!/usr/bin/perl
#
# find_latest_msvc_version.pl
#
# ABSTRACT
# ==========
# The script returns the latest msvc version and is just a helper script not intended to be used
#  by the end user.  It needs to be extended for every new msvc version supported by boost.
#
# example:
#  Visual Studio 2005 => 8.0
#  Visual Studio 2008 => 9.0
#
# NOTES
# =======
# The script has not been tested for the corresponding MSVC express versions.
# 
use Win32::TieRegistry;
use File::Spec;

use strict 'vars';

my $msvc_version = '0.0';

# Known toolset versions, in order of preference.
my @known_msvc_versions = ( '9.0', '9.0express', '8.0',  '8.0express' );
my @registry_sections = ( 'Microsoft', 'Wow6432Node\Microsoft' );

# Names of registry keys containing the Visual C++ installation path (relative
# to "HKEY_LOCAL_MACHINE\SOFTWARE\\Microsoft").
my %msvc_registry_keys = (
  '9.0' => 'VisualStudio\9.0\Setup\VC',
  '9.0express' => 'VCExpress\9.0\Setup\VC',
  '8.0' => 'VisualStudio\8.0\Setup\VC',
  '8.0express' => 'VCExpress\8.0\Setup\VC' 
  );

my $reg_key;
my $vc_product_dir;
my $vc_product_bin_dir;
my $reg_section;
my $msvc_version;
my $version = 0.0;
my $h_reg_key_software;
my $h_reg_key;

$h_reg_key_software = new Win32::TieRegistry "LMachine\\Software", { Access=>KEY_READ, Delimiter=>"\\" } ;
if( ! defined( $h_reg_key_software ) ) {
  die "find_latest_msvc_version: cannot read windows registry HKLM\\Software\n";
}

foreach $msvc_version (@known_msvc_versions) {
  foreach $reg_section (@registry_sections) {
    $reg_key = $reg_section . "\\";
    $reg_key .= $msvc_registry_keys{ $msvc_version };
    #$reg_key .= $msvc_registry_keys{ '10' };
    #print "$reg_key\n";
    
    $h_reg_key = $h_reg_key_software->Open( $reg_key, { Access=>KEY_READ, Delimiter=>"\\" } );
    if( defined( $h_reg_key ) ) {
      #print "accessing registry $reg_key \n";
      $vc_product_dir = $h_reg_key->GetValue( 'ProductDir' );
      #print "product dir: $vc_product_dir \n";
      
      $vc_product_bin_dir = File::Spec->catfile( $vc_product_dir, 'bin' );
      if( -d $vc_product_bin_dir ) {
        $version = $msvc_version;
        
        #just test stuff....
        #$version = '6.0express';
        
        $version =~ s/([0-9.]+).*/\1/;
        #closing registry keys
        undef $h_reg_key ;
        undef $h_reg_key_software;

        #print the VC version 9.0, 8.0 , ...
        print $version;
        exit 0;
      }
      
      undef $h_reg_key
    }
  }
}

#close registry key software
undef $h_reg_key_software;

print $msvc_version;
