#!/usr/bin/perl
#
# get_msvc_includes_amd64.pl
#
use File::Basename;
use File::Temp;
use File::Spec;
use File::Path;

use FileHandle;
use Getopt::Long;
use English;

use strict 'vars';

$::g_prog_name = basename( $0 ); 

my $vc_bin_dir = 'C:\Program Files (x86)\Microsoft Visual Studio 9.0\VC\bin' ;
#my $vc_bin_dir = '' ;
my $vc_include_path = '';
#$vc_include_path = '-I"C:\Program Files (x86)\Microsoft Visual Studio 9.0\VC\\include"';
#$vc_include_path .= ' -I"C:\Program Files\Microsoft SDKs\Windows\v6.1\include"';

my $vc_setup_script;
my $vc_setup_inspect_script;
my $vc_environment;
my @vc_include_dir_list;
my $include_dir;
my $include_dir_normalized;
my $cuda_toolkit_version;

if( scalar(@ARGV) != 1 ) {
  print STDERR "$::g_prog_name: error: the cuda toolkit version must be specified; e.g. 2.3\n";
  exit 1;
}

$cuda_toolkit_version = $ARGV[ 0 ];

$vc_setup_inspect_script = create_vc_include_inspector_amd64( $vc_bin_dir );
#print $vc_setup_inspect_script;

#$vc_setup_inspect_script = File::Spec->catfile( 'bin dir', 'vc_amd64_env_test.bat' );

$vc_environment = `"$vc_setup_inspect_script"`;
if( $? != 0 ) {
  print STDERR "$::g_prog_name: error: $vc_setup_inspect_script has failed\n";
  die;
}

rmtree( dirname( $vc_setup_inspect_script ), { verbose => 0 } );


#print "-----------------\n";
#print "$vc_environment";
#print "-----------------\n\n";

@vc_include_dir_list = split /;/, $vc_environment;

if( $cuda_toolkit_version eq '2.3' ) {
  #filter the include directories
  foreach $include_dir (@vc_include_dir_list) {
    $include_dir_normalized = File::Spec->canonpath( $include_dir ); 
    
    if( -d "$include_dir_normalized" ) {
      if( $vc_include_path ne '' ) {
        $vc_include_path .= ',';
      }
      $vc_include_path .= "\"$include_dir_normalized\"";
    }
  }

  print "--system-include $vc_include_path";
}
else {
  #filter the include directories
  foreach $include_dir (@vc_include_dir_list) {
    $include_dir_normalized = File::Spec->canonpath( $include_dir ); 
    
    if( -d "$include_dir_normalized" ) {
      if( $vc_include_path ne '' ) {
        $vc_include_path .= ' ';
      }
      $vc_include_path .= "-I\"$include_dir_normalized\"";
    }
  }

  print "$vc_include_path";
}
sub create_vc_include_inspector_amd64 {
  my ($vc_bin_dir) = @_;
  my $fh;
  my $vc_setup_inspect_script;
  my $vc_setup_script;
  my $tempdir;
  
  $vc_setup_script = File::Spec->catfile( $vc_bin_dir, 'amd64', 'vcvarsamd64.bat' );
  if( ! -f "$vc_setup_script" ) {
    print STDERR "$::g_prog_name: error: \"$vc_setup_script\" does not exist\n";
    die;
  }
  
  $tempdir = File::Temp->tempdir( 'get_msvc_includes_amd64_XXXX' );
  
  $vc_setup_inspect_script = File::Spec->catfile( $tempdir, 'vc_setup_inspect_script.bat' );
  
  $fh = new FileHandle $vc_setup_inspect_script, "w";
  if( ! defined( $fh ) ) {
    print STDERR "$::g_prog_name: error: cannot open file $vc_setup_inspect_script\n";
    die
  }
  
  print $fh "\@echo off\n";
  print $fh "call \"$vc_setup_script\" >nul\n";
  print $fh "echo %INCLUDE%\n";

  undef $fh;
  
  return $vc_setup_inspect_script;
}
