#
# get_system_info_win.pl
#

use English;
use strict 'vars';

use Win32::TieRegistry;

my $system_info;

die if( $OSNAME ne 'MSWin32' ); 


$system_info = get_system_info();

print "$system_info\n";

sub get_system_info {
  my $system_info;
  my $num_processors;
  
  $system_info = 'windows';
  
  # architecture detection
  if( is_windows_64bit() ) {
    $system_info .= ':x86_64';
  }
  else {
    $system_info .= ':x86';
  }
  
  # detection of the number of installed/enabled processors
  $num_processors = get_num_processors();
  #print "Number of processors: $num_processors \n";
  $system_info .= ":$num_processors";
  
  # distro and codename attributes not yet supported
  $system_info .= ":unknown:unknown";
   
  return $system_info;
}


sub is_windows_64bit {
  my $windows_64bit = 0;
  my $proc_arch;
  

  if( ! exists( $ENV{'PROCESSOR_ARCHITECTURE'} ) ) {
    die "is_windows_64bit(): the environment variable PROCESSOR_ARCHITECTURE does not exist\n";
  }
  
  $proc_arch = $ENV{'PROCESSOR_ARCHITECTURE'};
  if( $proc_arch =~ /AMD64/ ) {
    $windows_64bit = 1;
  }
  
  return $windows_64bit;
}

sub get_num_processors {
  my $h_reg_key_hardware;
  my $h_reg_key;
  my $num_procs = 0;
  my $proc_id = 0;
  my $proc_key;
  my @proc_value_names;
  my $proc_value_name;
  
  my $reg_section = "LMachine\\Hardware\\Description\\System\\CentralProcessor";
  
  $h_reg_key_hardware = new Win32::TieRegistry $reg_section, { Access=>KEY_READ, Delimiter=>"\\" } ;
  if( ! defined( $h_reg_key_hardware ) ) {
    die "get_num_procs_win.pl: cannot read windows registry $reg_section \n";
  }

  while( defined( $h_reg_key_hardware ) ) {
    $h_reg_key = $h_reg_key_hardware->Open( $proc_id, { Access=>KEY_READ, Delimiter=>"\\" } );
    if( defined( $h_reg_key ) ) {
      # make sure the processor key contains a value indentified by 'Identifier'.
      @proc_value_names = $h_reg_key->ValueNames;
      foreach $proc_value_name (@proc_value_names) {
        if( $proc_value_name eq 'Identifier' ) {
          $num_procs++;
        }
      }
      $proc_id++;
      undef $h_reg_key;
    }
    else {
      undef $h_reg_key_hardware;
    } 
  }
  
  return $num_procs;
}
 