/**
  \ingroup CudaTestLibSample
  \file    CudaTestLibSample.cpp
  \brief   This CudaTestLibSample.cpp file contains the main entry point of a Sample application
  \author  waizenegger@hhi.fraunhofer.de
  \date    6/26/2009

  Copyright:
  2009 Fraunhofer Institute for Telecommunications, Heinrich-Hertz-Institut (HHI)
  The copyright of this software source code is the property of HHI.
  This software may be used and/or copied only with the written permission
  of HHI and in accordance with the terms and conditions stipulated
  in the agreement/contract under which the software has been supplied.
  The software distributed under this license is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either expressed or implied.
*/

#include "CudaTestPch.h"

#include <iostream>
#include <iomanip>

#include <boost/assert.hpp>


#include "SquareMeGold.h"
#include "SquareMeCuda.h"




int testSquareMe( int argc, char* argv[] )
{
  std::cout << "\nSquaring some floats ..." << std::endl;
  SquareMeGold cSmGold;

  unsigned int uiLength= 4;
  float* pfIn= new float[uiLength];

  for (unsigned int i=0; i < uiLength; i++)
  {
    pfIn[i]= static_cast< float >( i );
  }

  float* pfOut= new float[uiLength];  

  std::cout << "\nExecuting C++ gold implementation" << std::endl;
  cSmGold.squareMe(pfIn, pfOut, uiLength);

  std::cout << "Result:" << std::endl;
  for (unsigned int i=0; i < uiLength; i++)
  {
    std::cout << std::setw(5) << pfOut[i];
  }

  std::cout << "\nExecuting CUDA implementation" << std::endl;
  SquareMeCuda cSmCuda;
  // init device
  SquareMeCudaInitPar cPar;
  cPar.setDeviceId( 0 );
  cPar.setArraySize( uiLength );
  BOOST_VERIFY( 0 == cSmCuda.initGpu( cPar ) );
  cSmCuda.squareMe( pfIn, pfOut, uiLength );	
  
  // uninit device 
  BOOST_VERIFY( 0 == cSmCuda.uninitGpu() );

  std::cout << "Result:" << std::endl;
  for ( unsigned int i=0; i < uiLength; i++ )
  {
    std::cout << std::setw(5) << pfOut[i];
  }
  std::cout << std::endl;

  delete[] pfIn;
  delete[] pfOut;

  return 0;
}


int main( int argc, char* argv[] )
{  

  BOOST_VERIFY( 0 == testSquareMe( argc, argv ) );

  std::cout << "press any key to continue" << std::endl;
  ::getchar();

  return 0;
}
