/**
\ingroup 
\file    SquareMe.cu
\brief   
\author  wolfgang.waizenegger@hhi.fraunhofer.de
\date    06/26/2009

Copyright:
2008 Fraunhofer Institute for Telecommunications, Heinrich-Hertz-Institut (HHI)
The copyright of this software source code is the property of HHI.
This software may be used and/or copied only with the written permission
of HHI and in accordance with the terms and conditions stipulated
in the agreement/contract under which the software has been supplied.
The software distributed under this license is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied.
*/

#include "CudaTestPch.h"

#include "cutil_inline.h"

// cuda header
#include "SquareMe.cuh"

// includes, kernels
#include "SquareMe_kernel.cu"


extern "C" 
{  

int initGpu_SquareMeCuda( SquareMeCudaInitPar* pcPar, SquareMeCudaDevMem* pcDevMem )
{
  // set device
  cutilSafeCall( cudaSetDevice( pcPar->getDeviceId() ) );  
  unsigned int uiArraySize= pcPar->getArraySize();

  pcDevMem->setArraySize( uiArraySize );

  float* pfIn_gpu;
  float* pfOut_gpu;

  /* allocate an input and output data array on the GPU*/
  cutilSafeCall( cudaMalloc( (void **) &pfIn_gpu,  sizeof(float) * uiArraySize ) );  
  cutilSafeCall( cudaMalloc( (void **) &pfOut_gpu, sizeof(float) * uiArraySize ) );

  pcDevMem->setInputArray( pfIn_gpu );
  pcDevMem->setOutputArray( pfOut_gpu );

  return 0;
}

int uninitGpu_SquareMeCuda( SquareMeCudaDevMem* pcDevMem )
{
  /* free device memory */
  cutilSafeCall( cudaFree( pcDevMem->getInputArray() ) );
  cutilSafeCall( cudaFree( pcDevMem->getOutputArray() ) ); 

  return 0;
}

void SquareMe_CUDA(float* pfIn, float* pfOut, SquareMeCudaDevMem* pcDevMem ) 
{
  unsigned int uiLength = pcDevMem->getArraySize();

	float* pfIn_gpu  = pcDevMem->getInputArray();
  float* pfOut_gpu = pcDevMem->getOutputArray();

	/* copy data to GPU */
	cutilSafeCall( cudaMemcpy( pfIn_gpu, pfIn, sizeof(float) * uiLength, cudaMemcpyHostToDevice ) ); 

	/* Compute execution configuration using 128 threads per block */
	int threadNum= 240;
	dim3 dimBlock(threadNum);
	dim3 dimGrid(uiLength/dimBlock.x);
	if ( uiLength % threadNum !=0 ) dimGrid.x += 1;

	/* Call function on GPU */
	squareMe_kernel<<<dimGrid,dimBlock>>>(pfIn_gpu, pfOut_gpu, uiLength);
  
  /* check for errors */
  CUT_CHECK_ERROR( "SquareMe_CUDA::squareMe_kernel" );
  
	/* Copy result back to host */
	cutilSafeCall( cudaMemcpy( pfOut, pfOut_gpu, sizeof(float) * uiLength, cudaMemcpyDeviceToHost ) );
}

} // extern "C"