/**
\ingroup 
\file    SquareMe.cuh
\brief   
\author  wolfgang.waizenegger@hhi.fraunhofer.de
\date    06/26/2009

Copyright:
2008 Fraunhofer Institute for Telecommunications, Heinrich-Hertz-Institut (HHI)
The copyright of this software source code is the property of HHI.
This software may be used and/or copied only with the written permission
of HHI and in accordance with the terms and conditions stipulated
in the agreement/contract under which the software has been supplied.
The software distributed under this license is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied.
*/

#ifndef SQUARE_ME_CUH_062620091400
#define SQUARE_ME_CUH_062620091400

// parameter class(es)
class SquareMeCudaInitPar
{
public:
  SquareMeCudaInitPar(void)
    : m_uiDeviceId   ( 0 )
    , m_uiArraySize  ( 0 )
  {}

  ~SquareMeCudaInitPar(void) {}

  void setDeviceId( unsigned int uiDeviceId ) { m_uiDeviceId = uiDeviceId; }
  unsigned int getDeviceId() { return m_uiDeviceId; }

  void setArraySize( unsigned int uiArraySize ) { m_uiArraySize = uiArraySize; }
  unsigned int getArraySize() { return m_uiArraySize; }

protected:
  unsigned int m_uiDeviceId;
  unsigned int m_uiArraySize;
};

class SquareMeCudaDevMem
{
public:
  SquareMeCudaDevMem(void)
    : m_uiArraySize   ( 0 )
    , m_pfInput_gpu   ( 0 )
    , m_pfOutput_gpu  ( 0 )
  {}

  ~SquareMeCudaDevMem(void) {}

  void setArraySize( unsigned int uiArraySize ) { m_uiArraySize = uiArraySize; }
  unsigned int getArraySize() { return m_uiArraySize; }

  void setInputArray( float* pfInputArray ) { m_pfInput_gpu = pfInputArray; }
  float* getInputArray() { return m_pfInput_gpu; }

  void setOutputArray( float* pfOutputArray ) { m_pfOutput_gpu = pfOutputArray; }
  float* getOutputArray() { return m_pfOutput_gpu; }


protected:
  unsigned int   m_uiArraySize;
  float* m_pfInput_gpu;
  float* m_pfOutput_gpu;  
};


extern "C"
{
  int initGpu_SquareMeCuda( SquareMeCudaInitPar* pcPar, SquareMeCudaDevMem* pcDevMem );
  int uninitGpu_SquareMeCuda( SquareMeCudaDevMem* pcDevMem );

  void SquareMe_CUDA(float* pfIn, float* pfOut, SquareMeCudaDevMem* pcDevMem );
}; // extern "C"


#endif // #ifndef SQUARE_ME_CUH_062620091400
