;; X86 ABI (windows, cdecl)
;; ========================
;; All arguments are widened to 32 bits when they are passed. Return values are also widened 
;; to 32 bits and returned in the EAX register, except for 8-byte structures, 
;; which are returned in the EDX:EAX register pair. 
;; Larger structures are returned in the EAX register as pointers to hidden return structures. 
;; Parameters are pushed onto the stack from right to left. 
;; Structures that are not PODs will not be returned in registers.
;;
;; The compiler generates prolog and epilog code to save and restore the ESI, EDI, EBX, and EBP 
;; registers, if they are used in the function.
;;
;;
;;
;; Notes on X86_64 ABI(s)
;; =========================
;; WIN64 calling convention:
;;
;; Register   Purpose                                     Saved across calls
;; ========================================================================
;; RAX        temp register; return value                 No
;; RBX        callee-saved                                Yes
;; RCX        used to pass 1st argument to functions      No
;; RDX        used to pass 2nd argument to functions      No
;; RSP        stack pointer                               Yes
;; RBP        callee-saved; base pointer                  Yes
;; RSI        callee-saved registers                      Yes
;; RDI        callee-saved registers                      Yes
;; R8         used to pass 3th argument to functions      No
;; R9         used to pass 4th argument to functions      No
;; R10-R11    temporary                                   No
;; R12-R15    callee-saved registers                      Yes
;;
;; XMM0       used to pass 1th FP argument to functions   No 
;; XMM1       used to pass 2th FP argument to functions   No 
;; XMM2       used to pass 3th FP argument to functions   No
;; XMM3       used to pass 4th FP argument to functions   No
;; XMM4:XMM5  Volatile Must be preserved as required by 
;;            caller. 
;; XMM6:XMM15 callee-saved registers
;;
;;
;;
;; LINUX: X86_64 ABI:
;;
;; Register    Purpose                                  Saved across calls
;; ========================================================================
;; %rax        temp register; return value              No
;; %rbx        callee-saved                             Yes
;; %rcx        used to pass 4th argument to functions   No
;; %rdx        used to pass 3rd argument to functions   No
;; %rsp        stack pointer                            Yes
;; %rbp        callee-saved; base pointer               Yes
;; %rsi        used to pass 2nd argument to functions   No
;; %rdi        used to pass 1st argument to functions   No
;; %r8         used to pass 5th argument to functions   No
;; %r9         used to pass 6th argument to functions   No
;; %r10-%r11   temporary                                No
;; %r12-%r15   callee-saved registers                   Yes
;;
;; 
;; %xmm0-%xmm1 used to pass FP argument to functions    No
;; %xmm2-%xmm7 used to pass floating point arguments    No
;; %xmm8-%xmm15 temporary registers                     No
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			CODE SEGMENT
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

segment .text

;; this defintion solves the problem with the different link names 
;; under Win32 and Linux

%ifdef MSYS_WIN32
  %macro cglobal 1
      global _%1
  %define %1 _%1
  %endmacro
%elifdef MSYS_WIN64 MSYS_LINUX64
  %macro cglobal 1
      global %1
  %endmacro
%elifdef MSYS_LINUX
  %macro cglobal 1
      global %1
  %endmacro
%endif

;---------------------- global variables ------------------------------------
  cglobal  add_int_values_asm

;extern

;;
;; int add_int_values_asm( int iValue1, int iValue2 );
;;
%ifndef MSYS_WIN64 MSYS_LINUX64
;; 32 bit implementation
add_int_values_asm:
  push      ebp                       ; save frame pointer
  mov       ebp, esp                  ; setup access to parameters via ebp
  push      ebp                       ; saves esp on the stack

  mov       eax, dword [ebp + 8]      ; 1st parameter 
  add       eax, dword [ebp + 12]     ; 2nd parameter 

  pop       ebp
  mov       esp, ebp                  ; recover esp
  pop       ebp                       ; recover frame pointer
  ret
%endif

%ifdef MSYS_WIN64 MSYS_LINUX64
;; 64 bit implementation
;;
;; WINDOWS ABI:
;; rcx        used to pass 1st argument to functions
;; rdx        used to pass 2nd argument to functions

;; LINUX ABI:  
;; %rsi        used to pass 2nd argument to functions
;; %rdi        used to pass 1st argument to functions

add_int_values_asm:

  push  rbx                            ; save frame pointer

%ifdef MSYS_WIN64
  ; transform windows ABI to LINUX ABI
  push  rdi
  push  rsi
  mov   rdi, rcx
  mov   rsi, rdx
%endif

  mov   rax, rdi
  add   rax, rsi

%ifdef MSYS_WIN64
  pop rsi                       ; recover callee saved registers on windows
  pop rdi
%endif

  pop rbx                        ; restore frame pointer
  ret
%endif
