/**
  \ingroup PurifyTest
  \file    PurifyTest.cpp
  \brief   This PurifyTest.cpp file contains the main entry point of a test application to
           run some basic checks to see whether purify is able to catch memory leaks.
  \author  rauth@hhi.fraunhofer.de
  \date    7/15/2009

  Copyright:
  2009 Fraunhofer Institute for Telecommunications, Heinrich-Hertz-Institut (HHI)
  The copyright of this software source code is the property of HHI.
  This software may be used and/or copied only with the written permission
  of HHI and in accordance with the terms and conditions stipulated
  in the agreement/contract under which the software has been supplied.
  The software distributed under this license is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either expressed or implied.
*/

#include "PurifyTestPch.h"
#include <iostream>
#include <string>
#include <memory.h>

void usage()
{
  std::clog << "usage: PurifyTest leak | no-leak" << std::endl;
}

char* allocBytes( unsigned int uiSize )
{
  char* pchByteVector = new char[ uiSize ];

  return pchByteVector;
}

int main( int argc, char* argv[] )
{
  std::string cTestOption;
  bool bLeakTest = true;
  unsigned int uiNumBytes = 10;

  std::cout << "PurifyTest: starting ... " << std::endl;
 
  if( argc != 2 )
  {
    usage();
    return 1;
  }

  cTestOption = argv[ 1 ];
  if( ( cTestOption != "leak" ) && ( cTestOption != "no-leak" ) ) 
  {
    usage();
    return 1;
  }

  bLeakTest = ( cTestOption == "no-leak" ) ? false : true;

  {
    std::cout << "Allocating " << uiNumBytes << " number of bytes\n";

    char* pchByteVector = allocBytes( uiNumBytes );

    memset( pchByteVector, 0, uiNumBytes );

    if( bLeakTest )
    {
      std::cout << "You've choosen to activate the leak test, the memory checker should find a leak of 10 bytes." << std::endl;
    }
    else 
    {
      delete [] pchByteVector;
      pchByteVector = NULL;
    }
  }

  std::cout << "PurifyTest: exiting ... " << std::endl;

  return 0;
}
