/**
  \ingroup UTFexampleLibTest
  \file    DummyTester.cpp
  \brief   This DummyTester.cpp file implements test functionality in native C++ without boost specific functions. Test functionality always in here and NOT in the lib
  \author  selinger@hhi.fraunhofer.de
  \date    3/3/2009

  Copyright:
  2005 Fraunhofer Institute for Telecommunications, Heinrich-Hertz-Institut (HHI)
  The copyright of this software source code is the property of HHI.
  This software may be used and/or copied only with the written permission
  of HHI and in accordance with the terms and conditions stipulated
  in the agreement/contract under which the software has been supplied.
  The software distributed under this license is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either expressed or implied.
*/

#include "UTFexampleLibTestPch.h"
#include "DummyTester.h"
#include "UTFexampleLib/Dummy.h"
#include <fstream>

bool DummyTester::testAddNumbers( const std::string& rcFileNameRef, const std::string& rcFileNameTest )
{
/////////////// Run the algorithm under test and write results to a file
  std::ofstream cOfStreamTest;
  cOfStreamTest.open( rcFileNameTest.c_str(), std::ios::out );

  if( false == cOfStreamTest.is_open() )
  {
    throw( "\nCould not open reference data file: " + rcFileNameTest );
  }

  // class under Tests
  UTFexampleLib::Dummy cDummy;

  for( unsigned int ui = 0; ui <= 20; ui++ )
  {
    // run the tests
    int iResult = cDummy.addNumbers( ui, ui );

    // write results to file
    cOfStreamTest << iResult << std::endl;
  }

  cOfStreamTest.close();

/////////////// Open Ref and Test file
  
  std::ifstream cIfStreamRef;
  cIfStreamRef.open( rcFileNameRef.c_str(), std::ios::in );

  if( false == cIfStreamRef.is_open() )
  {
    throw( "\nCould not open reference data file: " + rcFileNameRef );
  }

  std::ifstream cIfStreamTest;
  cIfStreamTest.open( rcFileNameTest.c_str(), std::ios::in );

  if( false == cIfStreamTest.is_open() )
  {
    throw( "\nCould not open test data file for reading: " + rcFileNameTest );
  }

/////////////// Compare the content of the files

  std::string cStringRef;
  std::string cStringTest;
  bool        bResult = false;

  while( false == cIfStreamRef.eof() )
  { 
    bResult = true;

    // read the ref data from file
    cIfStreamRef  >> cStringRef;
 
    // read the generated test data from file
    cIfStreamTest >> cStringTest;

    //std::cerr << cStringTest << std::endl;

    if( cStringTest != cStringRef )
    {
      bResult = false;
      break;
    }
  }

  // verify that files have been read up to the end
  if( cIfStreamRef.eof() != cIfStreamTest.eof() )
  {
    bResult = false;
  }

  cIfStreamRef.close();
  cIfStreamTest.close();

  return bResult;
}
