/**
  \ingroup UTFexampleLibTest
  \file    TestSuiteDefault.cpp
  \brief   This TestSuiteDefault.cpp file contains the default test suite and should always be present in a libraries test project.
  \author  selinger@hhi.fraunhofer.de
  \date    3/3/2009

  This file is 1 of N test suites for the UTFexampleLibTest lib. It should always be present and contain some or all of the test cases of your lib.
  Group related test cases into different test suites and optionally into separate files for ease of use when you have big libs with different components.
*/

/// 1. A must! As always I first need to include the precompiled header file
#include "UTFexampleLibTestPch.h"

/// 2. A must! This boost UTF header file must always be included. We use the static library variant of the UTF. It is not recommended to use the single header library variant
#include <boost/test/unit_test.hpp>

/// Include the header file(s) of the class(es) you want to test or as a level of indirection use the classes that will test the lib 
#include "DummyTester.h"
#include "UTFexampleLib/Dummy.h"


// Create and name the test suite
BOOST_AUTO_TEST_SUITE( ts1_default ) 

// Test case demonstrating direct testing of a class of the lib
BOOST_AUTO_TEST_CASE( tc1_simple ) { UTFexampleLib::Dummy cDummy; BOOST_CHECK_EQUAL( cDummy.addNumbers( 7, 8 ), 15 ); }

// Test case demonstrating complex testing with test data files
BOOST_AUTO_TEST_CASE( tc2_complex ) { UTFexampleLib::Dummy cDummy; BOOST_CHECK_EQUAL( DummyTester::testAddNumbers( "Ref_Numbers.txt", "Test_Numbers.txt" ), true ); }

// End of test suite
BOOST_AUTO_TEST_SUITE_END()


