/**
  \ingroup UTFexampleLibTest
  \file    UTFexampleLibTest.cpp
  \brief   This UTFexampleLibTest.cpp file is the root for complex library regression testing. It contains only the main test entry point, the master suite definition and the global fixture
  \author  selinger@hhi.fraunhofer.de
  \date    3/3/2009

  Complete documentation at: http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/index.html
  Runtime reference:         http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/runtime-config/reference.html
  Test tool reference:       http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/testing-tools/reference.html

  The <libname>Test project (UTFexampleLibTest in this case) shall always serve as a complete regression test project for the library.
  This project demonstrates the intended usage of the svn_<libname>/TestData/<libname> folder that must contain (by convention) the self-test data of the lib under test (where appropriate).
  Although all tests can be done within on file (UTFexampleLibTest.cpp) a reasonable test separates tests into 3 kinds of files:
  1. UTFexampleLibTest.cpp: 1 file that contains only the global fixture and the main entry point with master test suite definition.
  2. TestSuiteXXX.cpp     1..N files that contain only the test suite definition with all test cases. TestSuiteDefault.cpp should always be present.
  3. XXXTester.cpp        1..N files that contain the implementation of test functionality keeping the library free of any test code.
                          XXXTester.cpp like files should only use native C++ without boost UTF stuff.

  NORMALLY YOU CAN LEAVE THIS FILE AS IS. NOTHING TODO. Edit TestSuiteDefault.cpp in order to add tests.
*/

/// 1. A must! As always I first need to include the precompiled header file
#include "UTFexampleLibTestPch.h"

/// 2. A must! This boost UTF macro  BOOST_TEST_MODULE  generates the main entry point for this test project and names the master test suite. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/master-test-suite.html
#define BOOST_TEST_MODULE ts_master master_test_suite

/// 3. A must! This boost UTF header file must always be included. We use the static/dynamic library variant of the UTF.
#include <boost/test/unit_test.hpp>

#include <boost/filesystem.hpp>

#if defined( _WIN32 )
  #include <direct.h>
#elif defined( linux ) || defined( sun )
  #include <unistd.h>
#else
  #error currently unsupported platform for chdir
#endif


// Global test suite fixture (test initialization). You may remove it when you do not need global initialization or no test data.
struct GlobalConfigFixture
{
  GlobalConfigFixture()
  {
    BOOST_TEST_MESSAGE( "Global fixture setup. Changing the current directory to the TestData/... directory" );

    // Get the current dir like getCWD()
    boost::filesystem::path cPathTestData  = boost::filesystem::current_path();
    boost::filesystem::path cPathExtension = "TestData/UTFexampleLib";  // <------ edit this line for an additional test project in the repository

    //SR added the if construct to allow to be in the jamroot directory to invoke the test. We used to have just the while, which requires
    //cd to bin or below.
    if( true == boost::filesystem::is_directory( cPathTestData / cPathExtension ) )
    {
       // Found it. Add the path extension
       cPathTestData /= cPathExtension;
    }
    else
    {
      // Check for parent path
      while( true == cPathTestData.has_parent_path() )
      {
        // One level up path
        cPathTestData = cPathTestData.parent_path();

        // Check whether at the current path we find TestData/...
        if( true == boost::filesystem::is_directory( cPathTestData / cPathExtension ) )
        {
          // Found it. Add the path extension
          cPathTestData /= cPathExtension;
          break;
        }
      }
    }

    // Check that we did not end up with nothing found.
    if( false == cPathTestData.has_parent_path() )
    {
      // Throw a message indicating that dir was not found.
      throw( "GlobalConfigFixture:\nFailed changing to folder " + cPathExtension.string() + "\nPlease ensure that the directory exists." );
    }
    else
    { // Go into that dir as current working directory. xxx to be replaced with later version of boost
      int iRetv = 0;

      // Change the current working directory to the TestData/... path
      #if defined( _WIN32 )
         iRetv = _chdir( cPathTestData.string().c_str() );
      #elif defined( linux ) || defined( sun )
         iRetv = chdir( cPathTestData.string().c_str() );
      #else
         #error changeDirectory is not yet implemented on this platform
      #endif

      if( 0 != iRetv )
      {
        throw( "GlobalConfigFixture:\nCould not change current working directory to folder: " + cPathTestData.string() );
      }
    } // end readability scope
  } // end constructor

  ~GlobalConfigFixture()
  {
    // Just an informative message
    BOOST_TEST_MESSAGE( "global fixture teardown" );
  }
}; // end struct


// Apply the global fixture
BOOST_GLOBAL_FIXTURE( GlobalConfigFixture );

