/**
  \ingroup UTFexampleLibTest1
  \file    UTFexampleLibTest1.cpp
  \brief   This UTFexampleLibTest1.cpp file contains a simple example for using the boost unit test framework (UTF). It contains all you need for testing.
  \author  selinger@hhi.fraunhofer.de
  \date    3/3/2009

  This example is designed to show how simple it is writing a test with the boost unit test framework (UTF).
  All you need are 3 macros and 1 include.
  You do not write a main() function. Its contained in the BOOST_TEST_MODULE macro.

  What you get is:
  -- a unified test environment with unified test code and unified test output.
  -- zero-effort to control (run selective) tests from the command-line
  -- zero-effort to control the test output from the command-line
  -- support of test automation especially with boost-build
  -- basic check for memory leaks out-of-the-box
  -- boost UTF catches system errors (exceptions/asserts) AND continues the tests.
  -- boost UTF allows to check for defines, compile success, compile failures, defines, expected exceptions, expected std::out, ... 

  // Find the complete documentation at:   http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/index.html
  // Find the runtime parameter reference: http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/runtime-config/reference.html
  // Find the test tool reference:         http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/testing-tools/reference.html
*/

/////////////////////////////////////////////// 
// Required includes and defines 
///////////////////////////////////////////////

/// 1. A must! As always I first need to include the precompiled header file
#include "UTFexampleLibTest1Pch.h"

/// 2. A must! This boost UTF macro  BOOST_TEST_MODULE  generates the main entry point for this test project and names the master test suite. The name is up to you. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/master-test-suite.html
#define BOOST_TEST_MODULE ts_master master_test_suite

/// 3. A must! This boost UTF header file must always be included. We use the static library variant of the UTF. It is not recommended to use the single header library variant
#include <boost/test/unit_test.hpp>

/// Include the header file(s) of the class(es) you want to test
#include "UTFexampleLib/UTFsample1.h"


///////////////////////////////////////////////
// Test cases (named tc1-tc5) are defined by macros below. 
///////////////////////////////////////////////

// This boost UTF macro  BOOST_AUTO_TEST_CASE  adds one test case to a test suite. E.g. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/auto-nullary-test-case.html
BOOST_AUTO_TEST_CASE( tc1 ) 
{ 
  // Make an instance of the class under test
  utfexamplelib::UTFsample1 cUTFsample1;

  // Use boost UTF macro  BOOST_CHECK_EQUAL to check whether the method under test returns the correct result
  BOOST_CHECK_EQUAL( cUTFsample1.addNumbers( 4, 5 ), 9 ); 
}


// Writing more tests as single lines for ease of reading/writing. 
BOOST_AUTO_TEST_CASE( tc2 ) { utfexamplelib::UTFsample1 cUTFsample1; BOOST_CHECK_EQUAL( cUTFsample1.addNumbers( 4, 1 ), 5 ); }
BOOST_AUTO_TEST_CASE( tc3 ) { utfexamplelib::UTFsample1 cUTFsample1; BOOST_CHECK_EQUAL( cUTFsample1.addNumbers( 4, 2 ), 6 ); }
BOOST_AUTO_TEST_CASE( tc4 ) { utfexamplelib::UTFsample1 cUTFsample1; BOOST_CHECK_EQUAL( cUTFsample1.addNumbers( 4, 3 ), 7 ); }
BOOST_AUTO_TEST_CASE( tc5 ) { utfexamplelib::UTFsample1 cUTFsample1; BOOST_CHECK_EQUAL( cUTFsample1.addNumbers( 4, 4 ), 8 ); } 

// You are done. Just compile&run the test project and look at the output.
