/**
  \ingroup UTFexampleLibTest3
  \file    UTFexampleLibTest3.cpp
  \brief   This UTFexampleLibTest3.cpp file shows some features of boost UTF
  \author  selinger@hhi.fraunhofer.de
  \date    3/3/2009

  This example is designed to demonstrate failures and how boost UTF reports them.
  failures are:   system C/C++ assert;  system exception;  checking for user exception;  checking for defines
  
  What you get is:  4 failures detected 

  // Find the complete documentation at:   http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/index.html
  // Find the runtime parameter reference: http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/runtime-config/reference.html
  // Find the test tool reference:         http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/testing-tools/reference.html
*/

///////////////////////////////////////////////
// Required includes and defines 
///////////////////////////////////////////////

/// 1. A must! As always I first need to include the precompiled header file
#include "UTFexampleLibTest3Pch.h"

/// 2. A must! This boost UTF macro  BOOST_TEST_MODULE  (use once and only once) generates the main entry point for this test project and names the master test suite. The name is up to you. http://www.boost.org/doc/libs/1_38_0/libs/test/doc/html/utf/user-guide/test-organization/master-test-suite.html
#define BOOST_TEST_MODULE ts_master master_test_stuite

/// 3. A must! This boost UTF header file must always be included. We use the static library variant of the UTF. It is not recommended to use the single header library variant
#include <boost/test/unit_test.hpp>

/// Include the header file(s) of the class(es) you want to test
#include "UTFexampleLib/UTFsample2.h"
#include <iostream>


// this is not a test but some work around to display some initial message
BOOST_AUTO_TEST_CASE( tc0_dummy ) 
{
  std::cerr << "\n\n--- These tests fail on purpose! It shall demonstrate how test errors (4) are reported.\n";
  std::cerr << "--- Play with the command-line parameters that boost UTF provides.\n\n\n";
}


// We add a new test and name it.
BOOST_AUTO_TEST_CASE( tc1_assert ) 
{ 
  BOOST_TEST_MESSAGE( "\n---------------------------------- This test case is expected to fail once " );

  // Instance of the class under test.
  utfexamplelib::UTFsample2 cUTFsample2;

  // This call will not complete normally but run into an assert. UTF will catch and report this assert with test case and line numbers.
  BOOST_CHECK_EQUAL( cUTFsample2.doAssert(), 0 ); 
}

BOOST_AUTO_TEST_CASE( tc2_throw_system_exception )
{
  BOOST_TEST_MESSAGE( "\n---------------------------------- This test case is expected to fail once " );

  // Instance of the class under test
  utfexamplelib::UTFsample2 cUTFsample2;

  // Check for: not throwing an exception.
  BOOST_CHECK_NO_THROW( cUTFsample2.divideNumbersMayThrowException( 5, 1 ) );
  
  // Check for: throw divide by zero caught by UTF but tests continues.
  BOOST_CHECK_NO_THROW( cUTFsample2.divideNumbersMayThrowException( 5, 0 ) );
}

BOOST_AUTO_TEST_CASE( tc3_exception )
{
  BOOST_TEST_MESSAGE( "\n---------------------------------- This test case is expected to fail once " );

  // Instance of the class under test.
  utfexamplelib::UTFsample2 cUTFsample2;

  // Check for a specific exception to be thrown. No boost test error will occur.
  BOOST_CHECK_THROW( cUTFsample2.doThrowSample2Exception(), utfexamplelib::Sample2Exception );

  // Check for a specific exception but some other exception will be thrown resulting in an boost test error.
  BOOST_CHECK_THROW( cUTFsample2.doThrowOtherThanSample2Exception(), utfexamplelib::Sample2Exception );
}

BOOST_AUTO_TEST_CASE( tc4_define ) 
{ 
  BOOST_TEST_MESSAGE( "\n---------------------------------- This test case is expected to fail once " );

  #define TEST_CODE_ON
  // Trying to verify that  some #define TEST_CODE_ON is not set!
  BOOST_CHECK( !BOOST_IS_DEFINED( TEST_CODE_ON ) ); 
}
