/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef VIDEOCODECTOOLSLIB_BITSTREAMSINK_H
#define VIDEOCODECTOOLSLIB_BITSTREAMSINK_H

#include "VideoCodecToolsLib/MSysLargeFile.h"

class BitstreamSink
  : public VideoCodecSinkIf<std::list<std::vector<UChar> >*>
{
public:
  BitstreamSink( LargeFile &rcBitstreamFile ) : m_rcBitstreamFile( rcBitstreamFile ), m_uiTotalBytes( 0 ) {}

  ErrVal receive( std::list<std::vector<UChar> > *pcPacket )
  {
    UInt uiSize = 0;
    for( std::list<std::vector<UChar> >::iterator cIter = pcPacket->begin(); cIter != pcPacket->end(); ++cIter )
    {
      uiSize += UInt( (*cIter).size() );
    }
    CAOF( sizeof( UInt ) == 4 );
    for( UInt ui = 0; ui < 4; ui++ )
    {
      const UChar ucByte = uiSize >> ui*8;
      RNOK( m_rcBitstreamFile.write( &ucByte, 1 ) );
    }
    m_uiTotalBytes += 4;

    for( std::list<std::vector<UChar> >::const_iterator cIter = pcPacket->begin(); cIter != pcPacket->end(); ++cIter )
    {
      RNOK( m_rcBitstreamFile.write( &(*cIter)[0], UInt( (*cIter).size() ) ) );
      m_uiTotalBytes += (*cIter).size();
    }
    delete pcPacket;
    return Err::m_nOK;
  }
  
  size_t getTotalNumBytes() const { return m_uiTotalBytes; }

private:
  LargeFile &m_rcBitstreamFile;
  size_t m_uiTotalBytes;
};

class BitstreamSinkStartCode
  : public VideoCodecSinkIf<std::list<std::vector<UChar> >*>
{
public:
  BitstreamSinkStartCode( LargeFile &rcBitstreamFile ) : m_rcBitstreamFile( rcBitstreamFile ), m_uiTotalBytes( 0 ) {}

  ErrVal receive( std::list<std::vector<UChar> > *pcPacket )
  {
    if( 0 != pcPacket->size())
    {
      UChar ucByte = 0;
      RNOK( m_rcBitstreamFile.write( &ucByte, 1 ) );
      RNOK( m_rcBitstreamFile.write( &ucByte, 1 ) );
      RNOK( m_rcBitstreamFile.write( &ucByte, 1 ) );
      ucByte = 1;
      RNOK( m_rcBitstreamFile.write( &ucByte, 1 ) );
      m_uiTotalBytes += 4;

//      RNOK( m_rcBitstreamFile.write( &rcPacket[0], UInt( rcPacket.size() ) ) );
      for( std::list<std::vector<UChar> >::const_iterator cIter = pcPacket->begin(); cIter != pcPacket->end(); ++cIter )
      {
        RNOK( m_rcBitstreamFile.write( &(*cIter)[0], UInt( (*cIter).size() ) ) );
        m_uiTotalBytes += (*cIter).size();
      }
    }
    delete pcPacket;
    return Err::m_nOK;
  }

  size_t getTotalNumBytes() const { return m_uiTotalBytes; }

private:
  LargeFile &m_rcBitstreamFile;
  size_t m_uiTotalBytes;
};

#endif // VIDEOCODECTOOLSLIB_BITSTREAMSINK_H
