/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef VIDEOCODECTOOLSLIB_MSYSDEFINES_H
#define VIDEOCODECTOOLSLIB_MSYSDEFINES_H

#include <assert.h>

#if defined( _WIN32 )
# if !defined( WIN32 )
#   define WIN32
# endif
# if !defined( MSYS_WIN32 )
#   define MSYS_WIN32
# endif
# if defined( _WIN64 )
#   if !defined( WIN64 )
#     define WIN64
#   endif
# endif
#elif defined( linux )
# if !defined( MSYS_LINUX )
#   define MSYS_LINUX
# endif
# if defined( __x86_64 )
#   if !defined( LINUX64 )
#     define LINUX64
#   endif
# endif
#endif

#if defined( MSYS_WIN32 )
# if !defined( _MT )
#   error Please add multithreading support
# endif
#endif

#if defined( MSYS_LINUX )
# include <pthread.h>
# include <sys/time.h>
# include <sys/ioctl.h>
# include <sys/types.h>
# include <netdb.h>
# include <sys/socket.h>
# include <sys/stat.h>
# include <netinet/in.h>
# include <netinet/tcp.h>
# include <arpa/inet.h>
# include <unistd.h>
# include <fcntl.h>
# include <dlfcn.h>
# include <cerrno>
# include <cstring>
#endif

typedef int            ErrVal;

typedef unsigned char  UChar;
typedef unsigned short UShort;
typedef unsigned int   UInt;

#if defined( MSYS_WIN32 )

  typedef __int64                Int64;
  typedef unsigned __int64       UInt64;

#elif defined( MSYS_LINUX )

  typedef long long int          Int64;
  typedef unsigned long long int UInt64;

#endif

template<class T> class TypeLimits;

template<> class TypeLimits<char>
{
public:
  static const char getMin() { return -128; }
  static const char getMax() { return 127; }
  typedef short DType; // difference type
  static const char m_iMin;
  static const char m_iMax;
};

template<> class TypeLimits<UChar>
{
public:
  typedef short DType; // difference type
  static const UChar getMin() { return 0; }
  static const UChar getMax() { return 255; }
  static const UChar m_iMin;
  static const UChar m_iMax;
};

template<> class TypeLimits<short>
{
public:
  typedef int DType; // difference type
  static const short getMin() { return -32768; }
  static const short getMax() { return 32767; }
  static const short m_iMin;
  static const short m_iMax;
};

template<> class TypeLimits<UShort>
{
public:
  typedef int DType; // difference type
  static const UShort getMin() { return 0; }
  static const UShort getMax() { return 0xFFFF; }
  static const UShort m_iMin;
  static const UShort m_iMax;
};

template<> class TypeLimits<int>
{
public:
  typedef Int64 DType; // difference type
  static const int getMin() { return (-2147483647 - 1); }
  static const int getMax() { return 2147483647; }
  static const int m_iMin;
  static const int m_iMax;
};

template<> class TypeLimits<UInt>
{
public:
  typedef Int64 DType; // difference type
  static const UInt getMin() { return 0; }
  static const UInt getMax() { return 0xFFFFFFFFU; }
  static const UInt m_iMin;
  static const UInt m_iMax;
};

template<> class TypeLimits<Int64>
{
public:
  static const Int64 m_iMin;
  static const Int64 m_iMax;
};

template<> class TypeLimits<UInt64>
{
public:
  static const UInt64 m_iMin;
  static const UInt64 m_iMax;
};


template<bool> class CompileTimeAOT;
template<>     class CompileTimeAOT<false> {};

#define CAOT(x) { CompileTimeAOT< (x)>(); }
#define CAOF(x) { CompileTimeAOT<!(x)>(); }


class Err
{
public:
  enum
  {
    m_nOK               =  0,
    m_nERR              = -1,
    m_nEndOfFile        = -2,
    m_nEndOfBuffer      = -3,
    m_nInvalidParameter = -4,
    m_nDataNotAvailable = -5,
    m_nSizeNotAvailable = -5,
  };
};


#define ROF( exp )            \
{                             \
  if( !( exp ) )              \
  {                           \
    assert( 0 );              \
    return -1;                \
  }                           \
}

#define ROT( exp )            \
{                             \
  if( ( exp ) )               \
  {                           \
    assert( 0 );              \
    return -1;                \
  }                           \
}

#define ROFS( exp )           \
{                             \
  if( !( exp ) )              \
  {                           \
    return -1;                \
  }                           \
}

#define ROTS( exp )           \
{                             \
  if( ( exp ) )               \
  {                           \
    return -1;                \
  }                           \
}

#define ROFR( exp, retVal )   \
{                             \
  if( !( exp ) )              \
  {                           \
    assert( 0 );              \
    return retVal;            \
  }                           \
}

#define ROTR( exp, retVal )   \
{                             \
  if( ( exp ) )               \
  {                           \
    assert( 0 );              \
    return retVal;            \
  }                           \
}

#define ROFRS( exp, retVal )  \
{                             \
  if( !( exp ) )              \
  {                           \
    return retVal;            \
  }                           \
}

#define ROTRS( exp, retVal )  \
{                             \
  if( ( exp ) )               \
  {                           \
    return retVal;            \
  }                           \
}

#define ROFV( exp )           \
{                             \
  if( !( exp ) )              \
  {                           \
    assert( 0 );              \
    return;                   \
  }                           \
}

#define ROTV( exp )           \
{                             \
  if( ( exp ) )               \
  {                           \
    assert( 0 );              \
    return;                   \
  }                           \
}

#define ROFVS( exp )          \
{                             \
  if( !( exp ) )              \
  {                           \
    return;                   \
  }                           \
}

#define ROTVS( exp )          \
{                             \
  if( ( exp ) )               \
  {                           \
    return;                   \
  }                           \
}

#define RNOK( exp )                   \
{                                     \
  const ErrVal nMSysRetVal = ( exp ); \
  if( 0 != nMSysRetVal )              \
  {                                   \
    assert( 0 );                      \
    return nMSysRetVal;               \
  }                                   \
}

#define RNOKR( exp, retVal )        \
{                                   \
  if( 0 != ( exp ) )                \
  {                                 \
    assert( 0 );                    \
    return retVal;                  \
  }                                 \
}

#define RNOKS( exp )                  \
{                                     \
  const ErrVal nMSysRetVal = ( exp ); \
  if( 0 != nMSysRetVal )              \
  {                                   \
    return nMSysRetVal;               \
  }                                   \
}

#define RNOKRS( exp, retVal )       \
{                                   \
  if( 0 != ( exp ) )                \
  {                                 \
    return retVal;                  \
  }                                 \
}

#define RNOKV( exp )                \
{                                   \
  if( 0 != ( exp ) )                \
  {                                 \
    assert( 0 );                    \
    return;                         \
  }                                 \
}

#define RNOKVS( exp )               \
{                                   \
  if( 0 != ( exp ) )                \
  {                                 \
    return;                         \
  }                                 \
}

#define ANOK( exp )                 \
{                                   \
  if( 0 != ( exp ) )                \
  {                                 \
    assert( 0 );                    \
  }                                 \
}

#define AOF( exp )                  \
{                                   \
  if( !( exp ) )                    \
  {                                 \
    assert( 0 );                    \
  }                                 \
}

#define AOT( exp )            \
{                             \
  if( ( exp ) )               \
  {                           \
    assert( 0 );              \
  }                           \
}

#if defined( _DEBUG )

  #define AOT_DBG( exp )    AOT( exp )
  #define AOF_DBG( exp )    AOF( exp )

#else

  #define AOT_DBG( exp )    ((void)0)
  #define AOF_DBG( exp )    ((void)0)

#endif


#endif // VIDEOCODECTOOLSLIB_MSYSDEFINES_H
