/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef VIDEOCODECTOOLSLIB_MSYSTIMESTAMP_H
#define VIDEOCODECTOOLSLIB_MSYSTIMESTAMP_H

#if defined( MSYS_WIN32 )
# define NOMINMAX
# include <windows.h>

#ifdef min
#error
#endif

#ifdef max
#error
#endif

#elif defined( MSYS_LINUX )
# include <sys/time.h>
#endif

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4251 )
#endif

class TimeStamp
{
public:
  TimeStamp( UInt64 uiTimeStamp=0 ) : m_uiTimeStamp( uiTimeStamp ) { if( 0 == m_uiFreq ) { UInt64 uiFreq; TimeStamp::calcFreq( uiFreq ); } }

  TimeStamp( const TimeStamp& rcTimeStamp ) : m_uiTimeStamp( rcTimeStamp.m_uiTimeStamp ) {}

  virtual ~TimeStamp() {}

public:
  void stamp() { getTs(); }
  UInt64 getTimeStamp()  const { return m_uiTimeStamp; }
  static UInt64 getFreq() { return m_uiFreq; }
  static void setFreq( UInt64 uiFreq ) { m_uiFreq = uiFreq; }

public:
  static ErrVal calcFreq( UInt64& ruiFreq );
  void reset() { m_uiTimeStamp = 0; }

public:
  TimeStamp& operator=( const TimeStamp& rcTimeStamp ) { m_uiTimeStamp = rcTimeStamp.m_uiTimeStamp; return *this; }
  operator UInt64() { return m_uiTimeStamp; }
  bool operator==( const TimeStamp& rcTimeStamp ) { return (m_uiTimeStamp == rcTimeStamp.m_uiTimeStamp); }

  TimeStamp operator+( const TimeStamp& rcTimeStamp ) { return TimeStamp( m_uiTimeStamp + rcTimeStamp.m_uiTimeStamp ); }
  TimeStamp operator-( const TimeStamp& rcTimeStamp ) { return TimeStamp( m_uiTimeStamp - rcTimeStamp.m_uiTimeStamp ); }
  TimeStamp operator*( const TimeStamp& rcTimeStamp ) { return TimeStamp( m_uiTimeStamp * rcTimeStamp.m_uiTimeStamp ); }
  TimeStamp operator/( const TimeStamp& rcTimeStamp ) { return TimeStamp( m_uiTimeStamp / rcTimeStamp.m_uiTimeStamp ); }

  TimeStamp& operator+=( const TimeStamp& rcTimeStamp ) { m_uiTimeStamp += rcTimeStamp.m_uiTimeStamp; return *this; }
  TimeStamp& operator-=( const TimeStamp& rcTimeStamp ) { m_uiTimeStamp -= rcTimeStamp.m_uiTimeStamp; return *this; }
  TimeStamp& operator*=( const TimeStamp& rcTimeStamp ) { m_uiTimeStamp *= rcTimeStamp.m_uiTimeStamp; return *this; }
  TimeStamp& operator/=( const TimeStamp& rcTimeStamp ) { m_uiTimeStamp /= rcTimeStamp.m_uiTimeStamp; return *this; }

public:
  UInt64 getSec() const;
  UInt64 getMSec() const;
  UInt64 getUSec() const;
  UInt64 getNSec() const;

  std::string getSecStr( bool  bAddUnit=false ) const;
  std::string getMSecStr( bool bAddUnit=false ) const;
  std::string getUSecStr( bool bAddUnit=false ) const;
  std::string getNSecStr( bool bAddUnit=false ) const;

  void getString( UInt64 uiTime, std::string& rcStr ) const;

public:
  // get time which is rounded to a dRndLevel more precise value than the requested resolution
  double getSecRnd( double dRndLevel = 1000.0 ) const;
  double getMSecRnd( double dRndLevel = 1000.0 ) const;
  double getUSecRnd( double dRndLevel = 1000.0 ) const;
  double getNSecRnd( double dRndLevel = 1000.0 ) const;


private:
  // void rdtsc();
  void getTs();

private:
  UInt64 m_uiTimeStamp;

private:
  static UInt64 m_uiFreq;
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4251 )
#endif

__inline
void TimeStamp::getTs()
{
#if defined( MSYS_WIN32 )
  LARGE_INTEGER liStamp;
  ::QueryPerformanceCounter( &liStamp );
  m_uiTimeStamp = liStamp.QuadPart;
#elif defined( MSYS_LINUX )
  struct timeval cCurrentTime;
  ::gettimeofday( &cCurrentTime, NULL );
  m_uiTimeStamp = (((UInt64)cCurrentTime.tv_usec) + (1000000*((UInt64)cCurrentTime.tv_sec)));
#endif
}

__inline
UInt64 TimeStamp::getSec() const 
{
  ROTR( 0 == m_uiFreq, 0 );
  return m_uiTimeStamp / m_uiFreq;
}

__inline
UInt64 TimeStamp::getMSec() const
{
  ROTR( 0 == m_uiFreq, 0 );
  return (UInt64)((1000.0 * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq);
}

__inline
UInt64 TimeStamp::getUSec() const
{
  ROTR( 0 == m_uiFreq, 0 );
  return (UInt64)((1000000.0 * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq);
}

__inline
UInt64 TimeStamp::getNSec() const
{
  ROTR( 0 == m_uiFreq, 0 );
  return (UInt64)((1000000000.0 * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq);
}

__inline
std::string TimeStamp::getSecStr( bool bAddUnit ) const
{
  std::string cStr;
  UInt64 uiTime = getSec();
  getString( uiTime, cStr );

  if( true == bAddUnit )
  {
    cStr.append( " s" );
  }

  return cStr;
}

__inline
std::string TimeStamp::getMSecStr( bool bAddUnit ) const
{
  std::string cStr;
  UInt64 uiTime = getMSec();
  getString( uiTime, cStr );

  if( true == bAddUnit )
  {
    cStr.append( " ms" );
  }

  return cStr;
}

__inline
std::string TimeStamp::getUSecStr( bool bAddUnit ) const
{
  std::string cStr;
  UInt64 uiTime = getUSec();
  getString( uiTime, cStr );

  if( true == bAddUnit )
  {
    cStr.append( " us" );
  }

  return cStr;
}

__inline
std::string TimeStamp::getNSecStr( bool bAddUnit ) const
{
  std::string cStr;
  UInt64 uiTime = getNSec();
  getString( uiTime, cStr );

  if( true == bAddUnit )
  {
    cStr.append( " ns" );
  }

  return cStr;
}

__inline
void TimeStamp::getString( UInt64 uiTime, std::string& rcStr ) const
{
  UInt uiDigit;
  int iNumDigits;

  iNumDigits = 0;
  UInt64 uiDummy = uiTime;

  if( 0 == uiTime )
  {
    iNumDigits = 1;
  }
  else
  {
    while( 0 != uiDummy )
    {
      uiDummy /= 10;
      iNumDigits++;
    }
  }

  rcStr.resize( iNumDigits );

  for( int i=iNumDigits-1; i>=0; i-- )
  {
    uiDigit = (UInt)(uiTime % 10);
    rcStr[i] = (char)('0' + uiDigit);
    uiTime /= 10;
  }
}

__inline
double TimeStamp::getSecRnd( double dRndLevel ) const
{
  ROTR( 0 == m_uiFreq, 0.0 );
  ROTR( 0.0 >= dRndLevel, 0.0 );
  const UInt64 uiResVal = (UInt64)(((dRndLevel * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq) + 0.5);
  return (double)(Int64)uiResVal / dRndLevel;
}

__inline
double TimeStamp::getMSecRnd( double dRndLevel ) const
{
  ROTR( 0 == m_uiFreq, 0.0 );
  ROTR( 0.0 >= dRndLevel, 0.0 );
  const UInt64 uiResVal = (UInt64)(((1000.0 * dRndLevel * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq) + 0.5);
  return (double)(Int64)uiResVal / dRndLevel;
}

__inline
double TimeStamp::getUSecRnd( double dRndLevel ) const
{
  ROTR( 0 == m_uiFreq, 0.0 );
  ROTR( 0.0 >= dRndLevel, 0.0 );
  const UInt64 uiResVal = (UInt64)(((1000000.0 * dRndLevel * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq) + 0.5);
  return (double)(Int64)uiResVal / dRndLevel;
}

__inline
double TimeStamp::getNSecRnd( double dRndLevel ) const
{
  ROTR( 0 == m_uiFreq, 0.0 );
  ROTR( 0.0 >= dRndLevel, 0.0 );
  const UInt64 uiResVal = (UInt64)(((1000000000.0 * dRndLevel * (double)(Int64)m_uiTimeStamp) / (double)(Int64)m_uiFreq) + 0.5);
  return (double)(Int64)uiResVal / dRndLevel;
}

#endif // VIDEOCODECTOOLSLIB_MSYSTIMESTAMP_H
