/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include <set>
#include "VideoCodecToolsLib/MSysDefines.h"

#ifndef VIDEOCODECTOOLSLIB_NEWANDDELETETRACKER_H
#define VIDEOCODECTOOLSLIB_NEWANDDELETETRACKER_H

// this is not yet thread safe
#define ENABLE_NEW_AND_DELETE_TRACKER 0

#if ENABLE_NEW_AND_DELETE_TRACKER //(HK)

#define NEW_AND_DELETE_TRACKER_MODIFIED 1

#if NEW_AND_DELETE_TRACKER_MODIFIED

void* operator new( size_t size );
void  operator delete(void *p);
void* operator new[]( size_t size );
void  operator delete[](void *p);

#else


class NewAndDeleteTracker
{
public:
  NewAndDeleteTracker( bool bArray ) : m_bTrackingActive( true ), m_i64Counter( 0 ), m_bArray( bArray ) {}
  virtual ~NewAndDeleteTracker();

  void activate() { m_bTrackingActive = true; }
  void deactivate() { m_bTrackingActive = false; }

  void addPtr( void *p );
  void erasePtr( void *p );

private:
  class PointerTracking
  {
  public:
    PointerTracking( void *p, Int64 iIdx ) : m_p( p ), m_iIdx( iIdx ) {}

    bool operator==( const PointerTracking &r ) const { return r.m_p == m_p; }
    bool operator<( const PointerTracking &r ) const { return m_p < r.m_p; }

  public:
    void *m_p;
    Int64 m_iIdx;
  };

  std::set<PointerTracking> m_acTracking;
  bool m_bTrackingActive;
  Int64 m_i64Counter;
  const bool m_bArray;
};

inline NewAndDeleteTracker& getNewAndDeleteTracker( bool bArray )
{
  static NewAndDeleteTracker x( true );
  static NewAndDeleteTracker y( false );
  return bArray ? x : y;
}

inline void * __cdecl operator new( size_t size )
{
  void *ptr = (void *)malloc(size);
  getNewAndDeleteTracker( false ).addPtr( ptr );
  return(ptr);
}

inline void __cdecl operator delete(void *p)
{
  if( p )
  {
    getNewAndDeleteTracker( false ).erasePtr( p );
    free(p);
  }
}

inline void * __cdecl operator new[]( size_t size )
{
  void *ptr = (void *)malloc(size);
  getNewAndDeleteTracker( true ).addPtr( ptr );
  return(ptr);
}

inline void __cdecl operator delete[](void *p)
{
  if( p )
  {
    getNewAndDeleteTracker( true ).erasePtr( p );
    free(p);
  }
}

#endif

#endif


#endif // VIDEOCODECTOOLSLIB_NEWANDDELETETRACKER_H
