/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _ACCESSUNIT_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _ACCESSUNIT_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "PlanePackage.h"
#include "ContextStore.h"
#include "CabacContextModel.h"
#include "WienerFilterParameter.h"
#include "ExtendedCabacTables.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class SequenceParameterSet;
class PictureParameterSet;
class PocAndFrameNum;

class AccessUnitParameters
{
public:
  AccessUnitParameters()
    : m_cPoc                        ( -1 )
    , m_uiMaxDPBSizeInFrameStores   ( 1 )
    , m_bStoreForReference          ( true )
    , m_uiMaxNumRefInterBlocks      ( 0 )
    , m_uiDeblockingFilterIdc       ( 0 )
    , m_iDeblockingAlphaC0OffsetDiv2( 0 )
    , m_iDeblockingBetaOffsetDiv2   ( 0 )
    , m_uiInterpolationMethod       ( 0 )
    , m_uiAdapInterpSwitch          ( 0 )
    , m_uiEntropySliceSizeX         ( 0 )
    , m_uiEntropySliceSizeY         ( 0 )
    , m_pcSequenceParameterSet      ( NULL )
    , m_pcPictureParameterSet       ( NULL )
    , m_pcPocAndFrameNum            ( NULL )
    , m_uiFrameNum                  ( 0 )
    , m_uiPlaneGrouping             ( 0 )
    , m_bIdr                        ( false )
    , m_eCabacTables                ( DEFAULT_CABAC_TABLES )
    , m_bVLCFlag                    ( false )
    , m_uiVLCMaxDelayParameter      ( 0 )
  {}

  AccessUnitParameters( PicOrderCnt cPicOrderCnt, UInt uiWidth, UInt uiHeight, const std::vector<Subsampling>& rcSubsampling, UInt uiPlaneGrouping )
    : m_cPoc                        ( cPicOrderCnt )
    , m_cAUStats                    ( PrimBlockSize( uiWidth, uiHeight ), rcSubsampling )
    , m_uiMaxDPBSizeInFrameStores   ( 1 )
    , m_bStoreForReference          ( true )
    , m_uiMaxNumRefInterBlocks      ( 0 )
    , m_uiDeblockingFilterIdc       ( 0 )
    , m_iDeblockingAlphaC0OffsetDiv2( 0 )
    , m_iDeblockingBetaOffsetDiv2   ( 0 )
    , m_uiInterpolationMethod       ( 0 )
    , m_uiAdapInterpSwitch          ( 0 )
    , m_uiEntropySliceSizeX         ( 0 )
    , m_uiEntropySliceSizeY         ( 0 )
    , m_pcSequenceParameterSet      ( NULL )
    , m_pcPictureParameterSet       ( NULL )
    , m_pcPocAndFrameNum            ( NULL )
    , m_uiFrameNum                  ( 0 )
    , m_uiPlaneGrouping             ( uiPlaneGrouping )
    , m_bIdr                        ( false )
    , m_eCabacTables                ( DEFAULT_CABAC_TABLES )
    , m_bVLCFlag                    ( false )
    , m_uiVLCMaxDelayParameter      ( 0 )
  {
    for( UInt uiPlaPaId = 0, uiPrimPlaneId = 0, uiNumPlanes = 1; uiPrimPlaneId < rcSubsampling.size(); uiPlaPaId++, uiPrimPlaneId += uiNumPlanes, uiNumPlanes = 1 )
    {
      for( UInt uiId = uiPrimPlaneId; ( uiId < rcSubsampling.size() - 1 ) && ( ( uiPlaneGrouping >> uiId ) & 1 ); uiId++, uiNumPlanes++ );
      m_acPlanePackageParameters.push_back( PlanePackageParameters( uiPlaPaId, uiPrimPlaneId, uiNumPlanes, m_cAUStats ) );
    }
  }

public:
  const PicOrderCnt&              getPicOrderCnt              ()                  const { return m_cPoc; }
  const PicBlockStats&            getAUStats                  ()                  const { return m_cAUStats; }
  UInt                            getNumPlanes                ()                  const { return UInt( m_cAUStats.getSubsampling().size() ); }
  const PrimBlockSize&            getPicSize                  ()                  const { return m_cAUStats.getPrimBlockSize(); }
  const std::vector<Subsampling>& getSubsampling              ()                  const { return m_cAUStats.getSubsampling(); }
  const Subsampling&              getSubsampling              ( UInt uiPlane )    const { return m_cAUStats.getSubsampling()[ uiPlane ]; }
  const BlockSize                 getPlaneSize                ( UInt uiPlane )    const { return m_cAUStats.getPrimBlockSize() * m_cAUStats.getSubsampling()[ uiPlane ]; }
  UInt                            getNumPlanePacks            ()                  const { return UInt( m_acPlanePackageParameters.size() ); }
  const PlanePackageParameters&   getPlaPaPars                ( UInt uiPlaPaId )  const { return m_acPlanePackageParameters[ uiPlaPaId ]; }
  UInt                            getMaxDPBSize               ()                  const { return m_uiMaxDPBSizeInFrameStores; }
  bool                            getRefPicFlag               ()                  const { return m_bStoreForReference; }
  UInt                            getMaxMvsPerBlk             ()                  const { return m_uiMaxNumRefInterBlocks; }
  UInt                            getRefListSize              ()                  const { return UInt( m_aiDeltaRefPicOrderCnt.size() ); }
  const std::vector<int>&         getRefDeltaPocList          ()                  const { return m_aiDeltaRefPicOrderCnt; }
  int                             getRefDeltaPoc              ( UInt uiIndex )    const { return m_aiDeltaRefPicOrderCnt[ uiIndex ]; }
  PicOrderCnt                     getRefPoc                   ( UInt uiIndex )    const { return m_cPoc + PicOrderCnt( m_aiDeltaRefPicOrderCnt[ uiIndex ] ); }
  UInt                            getDeblockingIdc            ()                  const { return m_uiDeblockingFilterIdc; }
  int                             getDeblockAlphaOffs         ()                  const { return m_iDeblockingAlphaC0OffsetDiv2; }
  int                             getDeblockBetaOffs          ()                  const { return m_iDeblockingBetaOffsetDiv2; }
  UInt                            getInterpolationMethod      ()                  const { return m_uiInterpolationMethod; }
  UInt                            getAdapInterpSwitch         ()                  const { return m_uiAdapInterpSwitch; }
  UInt                            getEntropySliceSizeX        ()                  const { return m_uiEntropySliceSizeX; }
  UInt                            getEntropySliceSizeY        ()                  const { return m_uiEntropySliceSizeY; }
  bool                            getEntropySliceFlag         ()                  const { return m_uiEntropySliceSizeX != 0 || m_uiEntropySliceSizeY != 0; }
  bool                            getCtxStatesReuseFlag       ()                  const { return m_bCtxStatesReuse; }

  void                            setPicOrderCnt              ( PicOrderCnt cPoc )            { m_cPoc = cPoc; }
  void                            setAUStats                  ( const PicBlockStats &r )      { m_cAUStats = r; }
  PlanePackageParameters&         getPlaPaPars                ( UInt uiPlaPaId )              { return m_acPlanePackageParameters[ uiPlaPaId ]; }

  WienerFilterParameters&         getWienerFilterParameters   ()                              { return m_cWienerFilterParameters; }
  const WienerFilterParameters&   getWienerFilterParameters   ()                        const { return m_cWienerFilterParameters; }
  std::vector<PlanePackageParameters>& getPlaPaParsVec        ()                              { return m_acPlanePackageParameters; }

  std::vector<int>&               getRefDeltaPocList          ()                              { return m_aiDeltaRefPicOrderCnt; }
  ExtendedCabacTables             getExtendedCabacTables      ()                        const { return m_eCabacTables; }
  bool                            getVLCFlag                  ()                        const { return m_bVLCFlag; }
  UInt                            getVLCMaxDelayParameter     ()                        const { return m_uiVLCMaxDelayParameter; }
  UInt                            getVLCMaxDelayInBits        ()                        const { return ( m_uiVLCMaxDelayParameter << VLC_DELAY_PARAMETER_SHIFT ); }
  void                            setMaxDPBSize               ( UInt uiSize )                 { m_uiMaxDPBSizeInFrameStores = uiSize; }
  void                            setRefPicFlag               ( bool bFlag )                  { m_bStoreForReference = bFlag; }
  void                            setMaxMvsPerBlk             ( UInt uiNumMvs )               { m_uiMaxNumRefInterBlocks = uiNumMvs; }
  void                            setDeblockingIdc            ( UInt uiIdc )                  { m_uiDeblockingFilterIdc = uiIdc; }
  void                            setDeblockAlphaOffs         ( int  iOffs )                  { m_iDeblockingAlphaC0OffsetDiv2 = iOffs; }
  void                            setDeblockBetaOffs          ( int  iOffs )                  { m_iDeblockingBetaOffsetDiv2 = iOffs; }
  void                            setInterpolationMethod      ( UInt uiMethod )               { m_uiInterpolationMethod = uiMethod; }
  void                            setAdapInterpSwitch         ( UInt uiSwitch )               { m_uiAdapInterpSwitch = uiSwitch; }
  void                            setEntropySliceSizeX        ( UInt uiSize )                 { m_uiEntropySliceSizeX = uiSize; }
  void                            setEntropySliceSizeY        ( UInt uiSize )                 { m_uiEntropySliceSizeY = uiSize; }
  void                            setCtxStatesReuse           ( bool bFlag )                  { m_bCtxStatesReuse = bFlag; }
  void                            setExtendedCabacTables      ( ExtendedCabacTables eTables ) { m_eCabacTables = eTables; }
  void                            setVLCFlag                  ( bool bFlag )                  { m_bVLCFlag = bFlag; }
  void                            setVLCMaxDelayParameter     ( UInt uiMaxDelayParameter )    { m_uiVLCMaxDelayParameter = uiMaxDelayParameter; }
  void                            setVLCMaxDelayInBits        ( UInt uiMaxDelayInBits )       { m_uiVLCMaxDelayParameter = uiMaxDelayInBits >> VLC_DELAY_PARAMETER_SHIFT; }

  bool isDirectModeFlagPresent( UInt uiPlaPaId ) const
  {
    ROFRS( getEntropySliceFlag(),                                           false );
    ROTRS( m_acPlanePackageParameters[ uiPlaPaId ].getInterPlaPaPredFlag(), false );
    ROFRS( m_uiMaxNumRefInterBlocks >  1,                                   false );
    ROFRS( m_uiMaxNumRefInterBlocks <= m_aiDeltaRefPicOrderCnt.size(),      false );
    return true;
  }

  void setSPS( const SequenceParameterSet* pcSPS)           { m_pcSequenceParameterSet = pcSPS;}
  void setPPS( const PictureParameterSet*  pcPPS)           { m_pcPictureParameterSet = pcPPS;}
  void setFrameNum(const UInt64 val)                        { m_uiFrameNum = val; }
  void setPocAndFrameNum( PocAndFrameNum* pcPocAndFrameNum) { m_pcPocAndFrameNum = pcPocAndFrameNum; }
  void setPlaneGrouping(const UInt val)                     { m_uiPlaneGrouping = val; }
  void setIdr(const bool val)                               { m_bIdr = val; }

  const SequenceParameterSet* getSPS()            const { return m_pcSequenceParameterSet;}
  const PictureParameterSet*  getPPS()            const { return m_pcPictureParameterSet;}
  UInt64                      getFrameNum()       const { return m_uiFrameNum; }
  UInt                        getPlaneGrouping()  const { return m_uiPlaneGrouping; }
  const PocAndFrameNum*       getPocAndFrameNum() const { return m_pcPocAndFrameNum; }
  PocAndFrameNum*             getPocAndFrameNum()       { return m_pcPocAndFrameNum; }
  bool                        getIdr()            const { return m_bIdr; }

  void getInterpolationParameters( UInt &ruiShift, UInt &ruiMargin, ReferencePicture<Pel>::UpsMethod &rcUpsMethod, bool bDecoder = false ) const;


private:
  PicOrderCnt                         m_cPoc;
  PicBlockStats                       m_cAUStats;
  std::vector<PlanePackageParameters> m_acPlanePackageParameters;
  UInt                                m_uiMaxDPBSizeInFrameStores;
  bool                                m_bStoreForReference;
  UInt                                m_uiMaxNumRefInterBlocks;
  std::vector<int>                    m_aiDeltaRefPicOrderCnt;
  UInt                                m_uiDeblockingFilterIdc;
  int                                 m_iDeblockingAlphaC0OffsetDiv2;
  int                                 m_iDeblockingBetaOffsetDiv2;
  UInt                                m_uiInterpolationMethod;
  UInt                                m_uiAdapInterpSwitch;
  UInt                                m_uiEntropySliceSizeX;
  UInt                                m_uiEntropySliceSizeY;
  bool                                m_bCtxStatesReuse;
  WienerFilterParameters              m_cWienerFilterParameters;
  const SequenceParameterSet*         m_pcSequenceParameterSet;
  const PictureParameterSet*          m_pcPictureParameterSet;
  PocAndFrameNum*                     m_pcPocAndFrameNum; 
  UInt64                              m_uiFrameNum;
  UInt                                m_uiPlaneGrouping;
  bool                                m_bIdr;
  ExtendedCabacTables                 m_eCabacTables;
  bool                                m_bVLCFlag;
  UInt                                m_uiVLCMaxDelayParameter;
};


class AccessUnit
{
public:
  AccessUnit() : m_pcRefPicture( 0 ) {}
  ~AccessUnit() { delete m_pcRefPicture; }

  ErrVal setRefPicture( const ReferencePicture<Pel>* pcRefPic )
  {
    ROT( m_pcRefPicture );
    ROF( pcRefPic ); 
    ROF( pcRefPic->getPicOrderCnt() == m_cPars.getPicOrderCnt() );
    m_pcRefPicture = pcRefPic;
    return Err::m_nOK; 
  }

  const ReferencePicture<Pel>& getRefPicture() const { return *m_pcRefPicture; }

  const PlanePackage* getPlanePackage( UInt uiPlanePackage ) const { AOT( uiPlanePackage >= m_apcPlanePackages.size() ); return m_apcPlanePackages[uiPlanePackage]; }
  PlanePackage* getPlanePackage( UInt uiPlanePackage ) { AOT( uiPlanePackage >= m_apcPlanePackages.size() ); return m_apcPlanePackages[uiPlanePackage]; }

  UInt getNumPlanePackages() const { return UInt( m_apcPlanePackages.size() ); }

  const AccessUnitParameters& getAUPars() const { return m_cPars; }
  AccessUnitParameters& getAUPars() { return m_cPars; }
  ContextStore<CabacContextModel>& getCtxStore( UInt uiPlaPaId ) { return *m_apcContextStore[uiPlaPaId]; }

  ErrVal createPlanePackages();
  ErrVal setupContextModels();

  template<class t_Pel>
  ErrVal extractPlanePackagePicBlock( UInt uiPlanePackageToExtract, const PicBlock<t_Pel> &rcPicBlock, PicBlock<t_Pel> &rcPlanePackagePicBlock ) const;

  void deleteSliceDataInPlanePackages()
  {
    for( UInt ui = 0; ui < UInt( m_apcPlanePackages.size() ); ui++ )
    {
      m_apcPlanePackages[ui]->deleteSliceData();
    }
  }

  const PlanePackage* getRefPlanePackage( UInt uiPlaPaId ) const
  {
    AOF  ( uiPlaPaId < m_apcPlanePackages.size() );
    ROFRS( m_apcPlanePackages[ uiPlaPaId ]->getPars().getInterPlaPaPredFlag(), 0 );
    return m_apcPlanePackages[ m_apcPlanePackages[ uiPlaPaId ]->getPars().getRefPlaPlaId() ];
  }

private:
  AutoDeletePtrVector<PlanePackage> m_apcPlanePackages;
  AccessUnitParameters              m_cPars;
  const ReferencePicture<Pel>*      m_pcRefPicture; // upsampled version of the encoded picture
  AutoDeletePtrVector< ContextStore<CabacContextModel> > m_apcContextStore;
};

template<class t_Pel>
ErrVal AccessUnit::extractPlanePackagePicBlock( UInt uiPlanePackageToExtract, const PicBlock<t_Pel> &rcPicBlock, PicBlock<t_Pel> &rcPlanePackagePicBlock ) const
{
  ROF( m_apcPlanePackages.size() > uiPlanePackageToExtract );

  UInt uiStartPlane = 0;
  for( UInt ui = 0; ui < uiPlanePackageToExtract; ui++ )
  {
    uiStartPlane += m_apcPlanePackages[ui]->getPars().getNumPlanes();
  }

  std::vector<Block<t_Pel> > acPlanes;
  for( UInt ui = 0; ui < m_apcPlanePackages[uiPlanePackageToExtract]->getPars().getNumPlanes(); ui++ )
  {
    acPlanes.push_back( rcPicBlock.getPlane( uiStartPlane + ui ) );
  }
  rcPlanePackagePicBlock = PicBlock<t_Pel>( acPlanes );

  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _ACCESSUNIT_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
