/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _BITWRITER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _BITWRITER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/BitWriterIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_UnsignedStoreType, class t_UnsignedWriteType>
class BitWriter
  : public BitWriterIf<t_UnsignedWriteType>
{
public:
  BitWriter( std::vector<t_UnsignedStoreType> &raData )
    : m_paData( &raData )
    , m_uiBitsLeftInBack( 0 ) {}

  ErrVal write( t_UnsignedWriteType uiBits, UInt uiNumberOfBits = 1 );

  ErrVal writeAlignOne();
  ErrVal writeAlignZero();

  size_t getNumberOfBitsWritten() const { return 8 * sizeof( t_UnsignedStoreType ) * m_paData->size() - size_t( m_uiBitsLeftInBack ); }

  void setVector( std::vector<t_UnsignedStoreType> &racVector )
  {
    m_paData = &racVector;
    m_uiBitsLeftInBack = 0;
  }

  std::vector<t_UnsignedStoreType>&       getVectorRef()       { return *m_paData; }
  const std::vector<t_UnsignedStoreType>& getVectorRef() const { return *m_paData; }

private:
  std::vector<t_UnsignedStoreType> *m_paData;
  UInt m_uiBitsLeftInBack;
};

template<class t_UnsignedStoreType, class t_UnsignedWriteType>
ErrVal BitWriter<t_UnsignedStoreType, t_UnsignedWriteType>::writeAlignZero()
{
  const int iNumBits = 8 - int(getNumberOfBitsWritten() & 0x7);
  return write( 0, iNumBits );
}

template<class t_UnsignedStoreType, class t_UnsignedWriteType>
ErrVal BitWriter<t_UnsignedStoreType, t_UnsignedWriteType>::writeAlignOne()
{
  const int iNumBits = 8 - int(getNumberOfBitsWritten() & 0x7);
  if( 8 != iNumBits )
  {
    return write( (1<<iNumBits)-1, iNumBits );
  }
  return Err::m_nOK;
}

template<class t_UnsignedStoreType, class t_UnsignedWriteType>
ErrVal BitWriter<t_UnsignedStoreType, t_UnsignedWriteType>::write( t_UnsignedWriteType uiBits, UInt uiNumberOfBits )
{
  ROT( uiNumberOfBits > 8 * sizeof( t_UnsignedWriteType ) );
  ROT( uiNumberOfBits == 0 );
  ROT( (uiBits >> 1) >> (uiNumberOfBits - 1) );

  if( 0 == m_uiBitsLeftInBack )
  {
    m_paData->push_back( 0 );
    m_uiBitsLeftInBack = 8 * sizeof( t_UnsignedStoreType );
  }

  while( uiNumberOfBits > m_uiBitsLeftInBack )
  {
    uiNumberOfBits -= m_uiBitsLeftInBack;
    m_paData->back() |= uiBits >> uiNumberOfBits;

    m_paData->push_back( 0 );
    m_uiBitsLeftInBack = 8 * sizeof( t_UnsignedStoreType );
  }

  m_uiBitsLeftInBack -= uiNumberOfBits;
  m_paData->back() |= t_UnsignedStoreType( uiBits ) << m_uiBitsLeftInBack;

  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _BITWRITER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
