/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_CABACCONTEXTMODEL_H__2D76E021_2277_44AD_95CC_EE831C7AC09F__INCLUDED_)
#define AFX_CABACCONTEXTMODEL_H__2D76E021_2277_44AD_95CC_EE831C7AC09F__INCLUDED_

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class CabacContextModel
{
public:
  typedef BaseClassEnd BaseType;
  typedef BaseClassEnd CapsType;

  FORCEINLINE void addBin( bool b, UInt uiBinIndex ) {}
  FORCEINLINE void addBits( double d ) {}

  FORCEINLINE void initCCModel( const CtxIni &rcInitItem, int iQp )
  {
    if( rcInitItem.m_sV2 == CTX_EQ_PROB_TOKEN )
    {
      setCabacState( m_ucEqualProbabilityState ); // token for EP-Bins
    }
    else
    {
      int iState = ( ( rcInitItem.m_sV1 * iQp ) >> 4 ) + rcInitItem.m_sV2;
      iState = gMin(gMax( 1, iState), 126 );

      if( iState>=64 )
      {
        setCabacState( iState - 64 );
        xSetMps( true );
      }
      else
      {
        setCabacState( 63 - iState );
        xSetMps( false );
      }
    }
  }
  FORCEINLINE bool isEpCtx() const { return getCabacState() == m_ucEqualProbabilityState; }
  FORCEINLINE void updateCtx( bool bIsLps );

  CabacContextModel()/* : m_ucState(0), m_ucMPS(0)*/ {}

  UChar getCabacState()         const { return m_ucState; }
  UChar getMps()                const { return m_ucMPS;   }
  void toggleMps()                    { m_ucMPS ^= 1;     }
  void setCabacState( UChar ucState ) { m_ucState = ucState; }

private:
  FORCEINLINE void xSetMps( bool bMps ) { m_ucMPS = bMps ? 1 : 0; }
  UChar m_ucState;
  UChar m_ucMPS;

  static const UChar m_ucEqualProbabilityState = 128;
  static const UChar m_aucACNextStateMPS64[64];
  static const UChar m_aucACNextStateLPS64[64];
};


FORCEINLINE void CabacContextModel::updateCtx( bool bIsLps )
{
  if( bIsLps )
  {
    if( !getCabacState() )
    {
      toggleMps();
    }
    setCabacState( m_aucACNextStateLPS64[getCabacState()] );
  }
  else
  {
    setCabacState( m_aucACNextStateMPS64[getCabacState()] );
  }
}



class ExactCabacModel
{
public:
  void setProbabilityOfOne( double dProb )
  {
    if( dProb > 0.5 )
    {
      m_dProb = 1.0 - dProb;
      m_bMps = true;
    }
    else
    {
      m_dProb = dProb;
      m_bMps = false;
    }
  }
  void setAlpha( double dAlpha ) { m_dAlpha = dAlpha; }
  void setMinProbVal( double dProb ) { m_dMinProbVal = dProb; }

  double simulateEncodingOfOneBin( bool bBin )
  {
    double dBits;
    const bool bIsMps = m_bMps == bBin;
    if( bIsMps )
    {
      dBits = log( 1.0 - m_dProb ) * sm_dMinusOneOverLog2;
      m_dProb = gMax( m_dMinProbVal, m_dProb * m_dAlpha );
    }
    else
    {
      dBits = log( m_dProb ) * sm_dMinusOneOverLog2;
      m_dProb = m_dProb * m_dAlpha + 1.0 - m_dAlpha;
      if( m_dProb > 0.5 )
      {
        m_dProb = 1.0 - m_dProb;
        m_bMps = !m_bMps;
      }
    }
    return dBits;
  }

private:
  double m_dProb;
  bool m_bMps;
  double m_dAlpha;
  double m_dMinProbVal;
  static const double sm_dMinusOneOverLog2;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_CABACCONTEXTMODEL_H__2D76E021_2277_44AD_95CC_EE831C7AC09F__INCLUDED_)
