/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_CABACMULTICODER_H__2D76E021_2277_44AD_95CC_EE831C7AC09F__INCLUDED_)
#define AFX_CABACMULTICODER_H__2D76E021_2277_44AD_95CC_EE831C7AC09F__INCLUDED_

#include "VideoCommonLib/BinStreamIf.h"
#include "BitWriter.h"
#include "CabacTables.h"

#include <fstream>

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<bool t_bEnc> class ProcessBit;


template<> class ProcessBit<true>
{
public:
  template<class t_BitProc>
  static ErrVal processBit( t_BitProc &rcBitProcessor, const bool &rbBit )
  {
    return rcBitProcessor.write( rbBit );
  }
};

template<> class ProcessBit<false>
{
public:
  template<class t_BitProc>
  static ErrVal processBit( t_BitProc &rcBitProcessor, bool &rbBit )
  {
    UInt uiBit = 0;
    RNOK( rcBitProcessor.read( uiBit ) );
    rbBit = uiBit ? true : false;
    return Err::m_nOK;
  }
};


template<class t_CabacCoder, class t_EqProbBitStream, class t_AggregateTables, bool t_bEncoder>
class CabacMultiCoder
  : public BinStreamIf<ProbBin&, t_bEncoder>
  , public CabaCoderStartFinishIf
{
public:
  typedef typename Const<t_bEncoder, ProbBin>::Type ProbBinType;

  CabacMultiCoder( PicOrderCnt cPoc ) : m_pcBitStream( NULL ) {}

  virtual ~CabacMultiCoder()
  {
    while( !m_apcCabacCoders.empty() )
    {
      delete m_apcCabacCoders.back();
      m_apcCabacCoders.pop_back();
    }
  }

  ErrVal finishCabac()
  {
    for( UInt ui = 0; ui < UInt( m_apcCabacCoders.size() ); ++ui )
    {
      RNOK( m_apcCabacCoders[ui]->finishCabac() );
    }
    return Err::m_nOK;
  }

  ErrVal printNumBinsWritten( std::ostream &rcOut ) const
  {
    rcOut << "EqProb     " << std::setw( 3 ) << 0 << " bins: " << m_pcBitStream->getNumberOfBitsWritten() << std::endl;
    for( UInt ui = 0; ui < UInt( m_apcCabacCoders.size() ); ++ui )
    {
      rcOut << "CabacCoder " << std::setw( 3 ) << ui + 1 << " bins: " << m_apcCabacCoders[ui]->getNumBinsWritten() << std::endl;
    }
    return Err::m_nOK;
  }

  ErrVal startCabac() { ROT( 1 ); }

  template<class t_CtorPars>
  ErrVal addCabacCoder( t_CtorPars &r )
  {
    m_apcCabacCoders.push_back( new t_CabacCoder( r ) );
    RNOK( m_apcCabacCoders.back()->startCabac() );
    return Err::m_nOK;
  }

  t_CabacCoder& getCabacCoder( UInt ui ) { return *m_apcCabacCoders[ui]; }

  void setEqProbBitStream( t_EqProbBitStream *p ) { m_pcBitStream = p; }

  ErrVal pushBack( ProbBinType &rcT )
  {
    const UInt uiIdx = t_AggregateTables::sm_auiStateToIdx[rcT.getStateIdx()];
    const UInt uiAgState = t_AggregateTables::sm_auiIdxToAggregateState[uiIdx];
    if( uiAgState != 0 )
    {
      RNOK( m_apcCabacCoders[uiIdx-1]->pushBack( rcT ) );
    }
    else
    {
      bool bBin = rcT.isLps();
      RNOK( ProcessBit<t_bEncoder>::processBit( *m_pcBitStream, bBin ) );
      if( !t_bEncoder )
      {
        getIfNotConst( rcT ) = ProbBin( rcT.getStateIdx(), bBin );
      }
    }
    return Err::m_nOK;
  }

  ErrVal pushBack( typename Const<t_bEncoder, bool>::Type &rcT )
  {
    return ProcessBit<t_bEncoder>::processBit( *m_pcBitStream, rcT );
  }

private:
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( CabacMultiCoder );
  std::vector<t_CabacCoder*> m_apcCabacCoders;
  t_EqProbBitStream *m_pcBitStream;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_CABACMULTICODER_H__2D76E021_2277_44AD_95CC_EE831C7AC09F__INCLUDED_)
