/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _CODEDPICTURESTORE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _CODEDPICTURESTORE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "AccessUnit.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class CodedPictureStoreBase
{
protected:
  CodedPictureStoreBase() {}
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( CodedPictureStoreBase );
public:
  template<class t_RefPlaPa>
  std::vector<const ReferencePlanePackage<Pel>*> getPlanePackageVec( UInt uiPlaPaId, const AccessUnitParameters& rcAUPars ) const
  {
    TOF( uiPlaPaId < rcAUPars.getNumPlanePacks() );

    UInt                uiPrimPlaneId     = rcAUPars.getPlaPaPars( uiPlaPaId ).getPrimPlaneId();
    UInt                uiNumPlanes       = rcAUPars.getPlaPaPars( uiPlaPaId ).getNumPlanes();
    const Subsampling&  rcPrimSubsampling = rcAUPars.getPlaPaPars( uiPlaPaId ).getPrimPlaneSubsampling();

    std::vector<const ReferencePlanePackage<Pel>*> cRefPlaPas;
    for( UInt ui = 0; ui < rcAUPars.getRefListSize(); ui++ )
    {
      const PicOrderCnt cCurrRefPoc = rcAUPars.getRefPoc( ui );
      TOF( cCurrRefPoc >= 0 );

      std::list<AccessUnit*>::const_iterator cIter;
      for( cIter = m_apcAUs.begin(); cIter != m_apcAUs.end(); cIter++ )
      {
        if( (*cIter)->getAUPars().getPicOrderCnt() == cCurrRefPoc )
        {
          const ReferencePicture<Pel>&  rcRefPic    = (*cIter)->getRefPicture();
          ReferencePlanePackage<Pel>*   pcRefPlaPa  = new t_RefPlaPa( rcRefPic, uiPrimPlaneId, uiNumPlanes, rcPrimSubsampling );
          cRefPlaPas.push_back( pcRefPlaPa );
          break;
        }
      }
      if( cIter == m_apcAUs.end() )
      {
        TOT( 1 );
      }
    }
    return cRefPlaPas;
  }

  ContextStore<CabacContextModel>* getRefCtxStore( UInt uiPlaPaId, const AccessUnitParameters& rcAUPars ) const
  {
    PicOrderCnt cDeltaMin = TypeLimits<PicOrderCnt>::m_iMax;
    ContextStore<CabacContextModel> *pcCtxStore = NULL;

    for( UInt ui = 0; ui < rcAUPars.getRefListSize(); ui++ )
    {
      const PicOrderCnt cCurrRefPoc = rcAUPars.getRefPoc( ui );
      TOF( cCurrRefPoc >= 0 );

      std::list<AccessUnit*>::const_iterator cIter;
      for( cIter = m_apcAUs.begin(); cIter != m_apcAUs.end(); cIter++ )
      {
        if( (*cIter)->getAUPars().getPicOrderCnt() == cCurrRefPoc )
        {
          if ( (*cIter)->getAUPars().getMaxMvsPerBlk() == rcAUPars.getMaxMvsPerBlk() )
          {
            if ( (*cIter)->getAUPars().getRefListSize() == rcAUPars.getRefListSize() )
            {
              const PicOrderCnt cDeltaCurr = gAbs( rcAUPars.getPicOrderCnt() - (*cIter)->getAUPars().getPicOrderCnt() );
              if ( cDeltaCurr < cDeltaMin )
              {
                cDeltaMin = cDeltaCurr;
                pcCtxStore = &(*cIter)->getCtxStore( uiPlaPaId );
              }
            }
          }
        }
      }
    }
    return pcCtxStore;
  }

  bool isCodable( PicOrderCnt cPoc, const std::vector<int> &raiRefDeltaPocList ) const
  {
    for( UInt ui = 0; ui < UInt( raiRefDeltaPocList.size() ); ui++ )
    {
      const PicOrderCnt cCurrRefPoc = cPoc + Int64( raiRefDeltaPocList[ui] );
      std::list<AccessUnit*>::const_iterator cIter;
      bool bFound = false;
      for( cIter = m_apcAUs.begin(); cIter != m_apcAUs.end(); cIter++ )
      {
        if( (*cIter)->getAUPars().getPicOrderCnt() == cCurrRefPoc )
        {
          bFound = true;
          break;
        }
      }
      if( !bFound )
      {
        return false;
      }
    }
    return true;
  }

  template<class T>
  static ErrVal addItemToStore( const T &rcItem, std::list<T> &rcStoreList, UInt uiCodedPictureStoreMaxSize, std::vector<T> *pcRemovedItems = NULL )
  {
    ROF( uiCodedPictureStoreMaxSize );
    while( !rcStoreList.empty() && rcStoreList.size() >= uiCodedPictureStoreMaxSize )
    {
      if( pcRemovedItems )
      {
        pcRemovedItems->push_back( rcStoreList.front() );
      }
      rcStoreList.pop_front();
    }

    rcStoreList.push_back( rcItem );

    return Err::m_nOK;
  }

  UInt getCurrentSize() const { return UInt( m_apcAUs.size() ); }

protected:
  virtual ~CodedPictureStoreBase()
  {
    while( !m_apcAUs.empty() )
    {
      delete m_apcAUs.back();
      m_apcAUs.pop_back();
    }
  }

  std::list<AccessUnit*> m_apcAUs;
};


class CodedPictureStore
  : public CodedPictureStoreBase
{
public:
  CodedPictureStore() : m_bInit( false ), m_uiCodedPictureStoreMaxSize( 0 ) {}

  bool isInitialized() const { return m_bInit; }

  ErrVal initCodedPictureStore( UInt uiCodedPictureStoreMaxSize )
  {
    ROT( m_bInit );
    m_bInit = true;
    m_uiCodedPictureStoreMaxSize = uiCodedPictureStoreMaxSize;
    return Err::m_nOK;
  }

//  UInt getCodedPictureStoreMaxSize() const { return m_uiCodedPictureStoreMaxSize; }

  ErrVal addAUAndResizeStore( AccessUnit* pcAU )
  {
    ROF( m_bInit );
    pcAU->deleteSliceDataInPlanePackages();
    if( m_uiCodedPictureStoreMaxSize )
    {
      AutoDeletePtrVector<AccessUnit> acRemovedAUs;
      RNOK( addItemToStore( pcAU, m_apcAUs, m_uiCodedPictureStoreMaxSize, &acRemovedAUs ) );
    }
    else
    {
      delete pcAU;
    }

    return Err::m_nOK;
  }

private:
  bool m_bInit;
  UInt m_uiCodedPictureStoreMaxSize;
};


class CodedPictureStoreMT
  : public CodedPictureStoreBase
{
public:
  ~CodedPictureStoreMT()
  {
    TOF( m_aiUsageCount.size() == m_abIsCoded.size() );
    for( std::map<PicOrderCnt,int>::iterator cIter = m_aiUsageCount.begin(); cIter != m_aiUsageCount.end(); cIter++ )
    {
      const PicOrderCnt cPoc = cIter->first;
      TOT( m_abIsCoded.end() == m_abIsCoded.find( cPoc ) );
      TOF( m_abIsCoded[cPoc] );
      TOF( cIter->second == 1 );
    }
  }

  ErrVal addAU( AccessUnit* pcAU )
  {
    const PicOrderCnt cPoc = pcAU->getAUPars().getPicOrderCnt();
    ROT( m_aiUsageCount.end() == m_aiUsageCount.find( cPoc ) );
    ROT( m_abIsCoded.end() == m_abIsCoded.find( cPoc ) );
    if( m_aiUsageCount[cPoc] == 0 )
    {
      delete pcAU;
      m_aiUsageCount.erase( cPoc );
      m_abIsCoded.erase( cPoc );
    }
    else
    {
      pcAU->deleteSliceDataInPlanePackages();
      m_abIsCoded[cPoc] = true;
      m_apcAUs.push_back( pcAU );
    }

    return Err::m_nOK;
  }

  ErrVal registerAU( PicOrderCnt cPoc )
  {
    ROF( m_aiUsageCount.end() == m_aiUsageCount.find( cPoc ) );
    ROF( m_abIsCoded.end() == m_abIsCoded.find( cPoc ) );
    m_aiUsageCount[cPoc] = 1;
    m_abIsCoded[cPoc] = false;
    return Err::m_nOK;
  }

  template<class t_Integer>
  ErrVal markAUsAsUsed( PicOrderCnt cPoc, const std::vector<t_Integer> &rcDeltaPocs )
  {
    for( UInt ui = 0; ui < UInt( rcDeltaPocs.size() ); ui++ )
    {
      PicOrderCnt cCurrPoc = cPoc + PicOrderCnt( rcDeltaPocs[ui] );
      ROT( m_aiUsageCount.end() == m_aiUsageCount.find( cCurrPoc ) );
      ROT( m_abIsCoded.end() == m_abIsCoded.find( cCurrPoc ) );
      m_aiUsageCount[cCurrPoc]++;
    }
    return Err::m_nOK;
  }

  template<class t_Integer>
  ErrVal markAUsAsUnused( PicOrderCnt cPoc, const std::vector<t_Integer> &rcDeltaPocs )
  {
    for( UInt ui = 0; ui < UInt( rcDeltaPocs.size() ); ui++ )
    {
      PicOrderCnt cCurrPoc = cPoc + PicOrderCnt( rcDeltaPocs[ui] );
      ROT( m_aiUsageCount.end() == m_aiUsageCount.find( cCurrPoc ) );
      ROT( m_abIsCoded.end() == m_abIsCoded.find( cCurrPoc ) );
      m_aiUsageCount[cCurrPoc]--;
      if( m_aiUsageCount[cCurrPoc] == 0 && m_abIsCoded[cCurrPoc] )
      {
        m_aiUsageCount.erase( cCurrPoc );
        m_abIsCoded.erase( cCurrPoc );
        bool bFound = false;
        for( std::list<AccessUnit*>::iterator cIter = m_apcAUs.begin(); cIter != m_apcAUs.end(); ++cIter )
        {
          if( (*cIter)->getAUPars().getPicOrderCnt() == cCurrPoc )
          {
            delete *cIter;
            m_apcAUs.erase( cIter );
            bFound = true;
            break;
          }
        }
        ROF( bFound );
      }
    }
    return Err::m_nOK;
  }

private:
  std::map<PicOrderCnt,int> m_aiUsageCount;
  std::map<PicOrderCnt,bool> m_abIsCoded;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _CODEDPICTURESTORE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
