/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_COMMONCLASSES_H__4CE634CE_B48D_4812_8098_9CAEA258BAA2__INCLUDED_)
#define AFX_COMMONCLASSES_H__4CE634CE_B48D_4812_8098_9CAEA258BAA2__INCLUDED_


// video namespace begin
VIDEO_NAMESPACE_BEGIN

//!\ingroup H264CommonLib
//!\DynBuf
template<class T>
class DynBuf
{
private:
  DynBuf( const DynBuf<T>& rcDynBuf ): m_uiBufferSize(0), m_pT(NULL) 
  {
    *this = rcDynBuf;
  }

public:
  DynBuf(): m_uiBufferSize(0), m_pT(NULL) {}
  ~DynBuf()
  {
    uninit();
  }

  bool operator != (const DynBuf<T>& rcDynBuf ) const
  {
    ROTRS( m_uiBufferSize != rcDynBuf.m_uiBufferSize, true );

    for( UInt n = 0; n < m_uiBufferSize; n++ )
    {
      ROTRS( m_pT[n] != rcDynBuf.m_pT[n], true);
    }
    return false;
  }

  bool operator == (const DynBuf<T>& rcDynBuf ) const
  {
    return !(*this != rcDynBuf);
  }

  DynBuf<T>& operator = (const DynBuf<T>& rcDynBuf ) 
  {
    uninit();
    if( rcDynBuf.size() )
    {
      init( rcDynBuf.size() );
      for( UInt n = 0; n < rcDynBuf.size(); n++ )
      {
        m_pT[n] = rcDynBuf.m_pT[n];
      }
    }
    return *this;
  }

  void setAll( const T& rcT )
  {
    for( UInt n = 0; n < m_uiBufferSize; n++ )
    {
      m_pT[n] = rcT;
    }
  }

  T& operator[] (const UInt uiOffset)
  {
    AOT_DBG(uiOffset >= m_uiBufferSize);
    return m_pT[uiOffset];
  }

  const T& operator[] (const UInt uiOffset) const
  {
    AOT_DBG(uiOffset >= m_uiBufferSize);
    return m_pT[uiOffset];
  }

  bool isInitialized() const { return m_pT != NULL; }

  ErrVal init( UInt uiBufferSize, bool bDeleteOldBuffer = false )
  {
    if( bDeleteOldBuffer )
    {
      uninit();
    }
    ROT( NULL != m_pT );
    m_pT = new T [ uiBufferSize ];
    ROT( NULL == m_pT );
    m_uiBufferSize = uiBufferSize;
    return Err::m_nOK;
  }

  ErrVal uninit()
  {
    delete [] m_pT;
    m_pT = NULL;
    m_uiBufferSize = 0;
    return Err::m_nOK;
  }

  T& get( UInt uiOffset ) const
  {
    AOT_DBG(uiOffset >= m_uiBufferSize);
    return m_pT[uiOffset];
  }

  void set( UInt uiOffset, T t )
  {
    AOT_DBG(uiOffset >= m_uiBufferSize);
    m_pT[uiOffset] = t;
  }

  UInt size() const
  {
    return m_uiBufferSize;
  }

protected:
  UInt m_uiBufferSize;
  T*   m_pT;
};


//!\ingroup H264CommonLib
//!\StatBuf
template< class T, UInt uiSize >
class StatBuf
{
private:
  StatBuf( const StatBuf<T, uiSize>& rcStatBuf ) 
  {
    *this = rcStatBuf;
  }
public:
  StatBuf() {}

  bool operator != (const StatBuf<T, uiSize>& rcStatBuf ) const
  {
    for( UInt n = 0; n < uiSize; n++ )
    {
      ROTRS( m_aT[n] != rcStatBuf.m_aT[n], true);
    }
    return false;
  }

  bool operator == (const StatBuf<T, uiSize>& rcStatBuf ) const
  {
    return !(*this != rcStatBuf);
  }

  void setAll( const T& rcT )
  {
    for( UInt n = 0; n < uiSize; n++ )
    {
      m_aT[n] = rcT;
    }
  }

  StatBuf<T, uiSize>& operator = (const StatBuf<T, uiSize>& rcStatBuf ) 
  {
    for( UInt n = 0; n < uiSize; n++ )
    {
      m_aT[n] = rcStatBuf.m_aT[n];
    }
    return *this;
  }

  T& operator[] (const UInt uiOffset)
  {
    AOT_DBG(uiOffset >= uiSize);
    return m_aT[uiOffset];
  }

  const T& operator[] (const UInt uiOffset) const
  {
    AOT_DBG(uiOffset >= uiSize);
    return m_aT[uiOffset];
  }

  ErrVal get( T& rcT, UInt uiOffset ) const
  {
    ROTRS(uiOffset >= uiSize, Err::m_nInvalidParameter );
    rcT = m_aT[uiOffset];
    return Err::m_nOK;
  }

  bool isValidOffset( UInt uiOffset ) const { return uiOffset < uiSize; }

  const T& get( UInt uiOffset ) const
  {
    AOT_DBG(uiOffset >= uiSize);
    return m_aT[uiOffset];
  }

  T& get( UInt uiOffset )
  {
    AOT_DBG(uiOffset >= uiSize);
    return m_aT[uiOffset];
  }

  void set( UInt uiOffset, T t )
  {
    AOT_DBG(uiOffset >= uiSize);
    m_aT[uiOffset] = t;
  }

  UInt size() const  { return uiSize; }

protected:
  T m_aT[uiSize];
};

// video namespace end
VIDEO_NAMESPACE_END


#endif // !defined(AFX_COMMONCLASSES_H__4CE634CE_B48D_4812_8098_9CAEA258BAA2__INCLUDED_)
