/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_COMMONDEFS_H__4CE634CE_B48D_4812_8098_9CAEA258BAA2__INCLUDED_)
#define AFX_COMMONDEFS_H__4CE634CE_B48D_4812_8098_9CAEA258BAA2__INCLUDED_

// main switches
#define DISABLE_CHANGES_THAT_BREAK_0_89_COMPATIBILITY 1

#define USE_CABAC12_TABLE_IN_ENCODER 1
#define USE_CABAC12_TABLE_IN_DECODER 1

#define COLLECT_CABAC_BITS 0
#define TRACK_CABAC_BITS 0
#define TRACE_CABAC_BITS_DETAILED 0

#define ALLOW_BOOST_LIBRARY 0
#define RANDOM_SLICE_PARSING 1

// adaptive loop filter
#define NO_DC_IN_WF 0

// spline based reference picture upsampling
#define USE_SEPARABLE_SPLINE_UPSAMPLING   1
#define USE_NON_INTERLEAVED 0
// other defines
#define CTX_EQ_PROB_TOKEN  TypeLimits<short>::m_iMin
#define DISABLE_DEFAULT_CPYCTOR_AND_ASOP(x) private: x( const x &r ); const x& operator=( const x &r )

#define FAST_MC 1

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class VideoCodecException
{
public:
  VideoCodecException() : m_acFile( "unspecified" ), m_uiLine( 0 ) {}
  VideoCodecException( const char *pcFile, UInt uiLine ) : m_acFile( pcFile ), m_uiLine( uiLine ) {}

  std::string exceptionFile() const { return m_acFile; }
  UInt        exceptionLine() const { return m_uiLine; }

private:
  std::string m_acFile;
  UInt        m_uiLine;
};

#define RETURN_ERR_ON_EXCEPTION \
catch( std::exception &cEx ) \
{ \
  std::cerr << "Abnormal program termination: " << cEx.what() << std::endl; \
  return Err::m_nERR; \
} \
catch( VideoCodecException &cEx ) \
{ \
  std::cerr << std::endl << "VideoCodecExcteption caught at: " << std::endl; \
  std::cerr << "File: " << cEx.exceptionFile() << std::endl; \
  std::cerr << "Line: " << cEx.exceptionLine() << std::endl; \
  return Err::m_nERR; \
} \
catch( ... ) \
{ \
  std::cerr << "Caught unknown exception" << std::endl; \
  return Err::m_nERR; \
}


#define TOT(  exp ) { if(                exp  ) { throw VideoCodecException( __FILE__, __LINE__ ); } }
#define TOF(  exp ) { if(              !(exp) ) { throw VideoCodecException( __FILE__, __LINE__ ); } }
#define TNOK( exp ) { if( Err::m_nOK != (exp) ) { throw VideoCodecException( __FILE__, __LINE__ ); } }

#if defined( _DEBUG )

  #define TOT_DBG( exp )    TOT( exp )
  #define TOF_DBG( exp )    TOF( exp )

#else

  #define TOT_DBG( exp )    ((void)0)
  #define TOF_DBG( exp )    ((void)0)

#endif



enum Profile
{
  INVALID_PROFILE = 0,
  BASELINE_PROFILE  = 66,
  MAIN_PROFILE      = 77,
  EXTENDED_PROFILE  = 88,
  HIGH_PROFILE      = 100,
  THIS_PROFILE      = 108,
};


enum SliceType
{
  INVALID_SLICE_TYPE  = -1,
  I_SLICE  = 0,
  P_SLICE  = 1,
  B_SLICE  = 2,
  N3_SLICE  = 3,
  N4_SLICE  = 4,
  N5_SLICE  = 5,
  N6_SLICE  = 6,
  N7_SLICE  = 7,
  N8_SLICE  = 8,
};


enum NalRefIdc
{
  NAL_REF_IDC_PRIORITY_LOWEST = 0,
  NAL_REF_IDC_PRIORITY_LOW,
  NAL_REF_IDC_PRIORITY_HIGH,
  NAL_REF_IDC_PRIORITY_HIGHEST
};

enum NalUnitType
{
  NAL_UNIT_UNSPECIFIED_0 = 0,
  NAL_UNIT_CODED_SLICE,
  NAL_UNIT_CODED_SLICE_DATAPART_A,
  NAL_UNIT_CODED_SLICE_DATAPART_B,
  NAL_UNIT_CODED_SLICE_DATAPART_C,
  NAL_UNIT_CODED_SLICE_IDR,
  NAL_UNIT_SEI,
  NAL_UNIT_SPS,
  NAL_UNIT_PPS,
  NAL_UNIT_ACCESS_UNIT_DELIMITER,
  NAL_UNIT_END_OF_SEQUENCE,
  NAL_UNIT_END_OF_STREAM,
  NAL_UNIT_FILLER_DATA,
  NAL_UNIT_SPS_EXTENSION,
  NAL_UNIT_RESERVED_14,
  NAL_UNIT_RESERVED_15,
  NAL_UNIT_RESERVED_16,
  NAL_UNIT_RESERVED_17,
  NAL_UNIT_RESERVED_18,
  NAL_UNIT_AUX_PICTURE,
  NAL_UNIT_RESERVED_20,
  NAL_UNIT_RESERVED_21,
  NAL_UNIT_RESERVED_22,
  NAL_UNIT_RESERVED_23,
  NAL_UNIT_UNSPECIFIED_24,
  NAL_UNIT_UNSPECIFIED_25,
  NAL_UNIT_UNSPECIFIED_26,
  NAL_UNIT_UNSPECIFIED_27,
  NAL_UNIT_UNSPECIFIED_28,
  NAL_UNIT_UNSPECIFIED_29,
  NAL_UNIT_UNSPECIFIED_30,
  NAL_UNIT_UNSPECIFIED_31,
  NAL_UNIT_INVALID,
};


//!Enumeration of picture representation types.
//##ModelId=407BE98B0128
enum PicStruct
{
  PS_NOT_SPECIFIED = -1,
  PS_FRAME       = 0, //!< frame	field_pic_flag shall be 0	1
  PS_TOP         = 1, //!< top field	field_pic_flag shall be 1, bottom_field_flag shall be 0	1
  PS_BOT         = 2, //!< bottom field	field_pic_flag shall be 1, bottom_field_flag shall be 1	1
  PS_TOP_BOT     = 3, //!< top field, bottom field, in that order	field_pic_flag shall be 0	2
  PS_BOT_TOP     = 4, //!< bottom field, top field, in that order	field_pic_flag shall be 0	2
  PS_TOP_BOT_TOP = 5, //!< top field, bottom field, top field repeated, in that order	field_pic_flag shall be 0	3
  PS_BOT_TOP_BOT = 6, //!< bottom field, top field, bottom field repeated, in that order	field_pic_flag shall be 0	3
  PS_2XFRAME     = 7, //!< frame doubling 	field_pic_flag shall be 0 fixed_frame_rate_flag shall be 1	2 
  PS_3XFRAME     = 8, //!< frame tripling 	field_pic_flag shall be 0 fixed_frame_rate_flag shall be 1	3 
};

//! Enumeration chrominance format types.
enum ChromaFormatIdc
{
  //chroma format idc's
  YUV400 = 0,    //!< no chroma components
  YUV420 = 1,    //!< horizontal and vertical sub sampled chroma components
  YUV422 = 2,    //!< horizontal sub sampled chroma components
  YUV444 = 3,     //!< full resolution chroma components
};


enum ListIdx
{
  LIST_0 = 0x00,
  LIST_1 = 0x01
};

enum SliceMode
{
  ONE_SLICE        = 0,
  FIXED_NUM_MB     = 1,
  RANDOM_NUM_MB    = 2,
  MULTI_SLICE_TYPE = 3,
};

enum PicType
{
  NOT_SPECIFIED   = 0x00,
  TOP_FIELD       = 0x01,
  BOT_FIELD       = 0x02,
  FRAME           = 0x03,
  MAX_FRAME_TYPE  = 0x04
};


typedef UInt64 VLCodeType; // depends on maximum codeword size 

// video namespace end
VIDEO_NAMESPACE_END

#define MIN_QP          0
#define MAX_QP          51
#define INIT_QP         26
#define QP_BITS         15
#define QP_SHIFT1       12

#ifndef FORCEINLINE
#if defined( MSYS_WIN32 )
#define FORCEINLINE __forceinline
#elif defined( MSYS_LINUX )
#define FORCEINLINE __inline
#endif
#endif

#define EXTRA_PRECISION_BITS            6 // [0..6]


//==== under investigation ====
#define CHECK_RDOQ_COST                 0 // only for value checking [and potential further investigation]

#define MULTICABAC_TRANSCODING_HACK     1
#define VLC_DELAY_PARAMETER_SHIFT       6 // adjusts quantization of delay parameter ( 6 => 2^6 Bit = 64 Bit = 8 Bytes )

#endif // !defined(AFX_COMMONDEFS_H__4CE634CE_B48D_4812_8098_9CAEA258BAA2__INCLUDED_)
