/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_COMMONTYPES_H__CBFE313E_2382_4ECC_9D41_416668E3507D__INCLUDED_)
#define AFX_COMMONTYPES_H__CBFE313E_2382_4ECC_9D41_416668E3507D__INCLUDED_

#include "VideoCodecToolsLib/MSysTimeStamp.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

// template toolbox start
template<bool t_bSelectA, class A, class B> class TypeSelect;
template<                 class A, class B> class TypeSelect<true,  A, B> { public: typedef A Type; };
template<                 class A, class B> class TypeSelect<false, A, B> { public: typedef B Type; };

template<bool t_bConst, class T> class Const;
template<               class T> class Const<true, T>  { public: typedef const T        Type; };
template<               class T> class Const<false, T> { public: typedef       T        Type; };
template<               class T> class Const<true, T*> { public: typedef const T* const Type; };
template<               class T> class Const<true, T&> { public: typedef const T&       Type; };

template<class T> T& getIfNotConst( T& r ) { return r; }
template<class T> T& getIfNotConst( const T& r ) { AOT( 1 ); T *p(NULL); return *p; }

template<class A, class B> class TypeCompare      { public: static const bool m_bEqual = false; };
template<class A>          class TypeCompare<A,A> { public: static const bool m_bEqual = true; };

class BaseClassEnd {};
template<class t_CapToFind, class A> class CheckCap                            { public: static const bool m_bHasCap = TypeCompare<typename A::CapsType, t_CapToFind>::m_bEqual ? true : CheckCap<t_CapToFind, typename A::BaseType>::m_bHasCap; };
template<class t_CapToFind>          class CheckCap<t_CapToFind, BaseClassEnd> { public: static const bool m_bHasCap = false; };

template<class t_Out, bool t_b> class TypeConvert               { public: template<class T> static t_Out& getIfTrue( T &r ) { return r; } };
template<class t_Out>           class TypeConvert<t_Out, false> { public: template<class T> static t_Out& getIfTrue( T &r ) { AOT( 1 ); t_Out *p(NULL); return *p; } };
// template toolbox end


template<class T>
class Ptr
{
public:
  Ptr() : m_pT( NULL ) {}
  Ptr( T *p ) : m_pT( p ) {}

  T* getPtr() { return m_pT; }
  T& operator*() { return *m_pT; }
  T* operator->() { return m_pT; }

  const T* getPtr() const { return m_pT; }
  const T& operator*() const { return *m_pT; }
  const T* operator->() const { return m_pT; }

private:
  T *m_pT;
};

template<class T>
class AutoDeletePtrVector
  : public std::vector<T*>
{
public:
  AutoDeletePtrVector() {}

  virtual ~AutoDeletePtrVector()
  {
    while( !std::vector<T*>::empty() )
    {
      delete std::vector<T*>::back();
      std::vector<T*>::pop_back();
    }
  }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( AutoDeletePtrVector );
};


class StopWatch
{
public:
  StopWatch() { m_cStart.stamp(); }
  
  void restart() { m_cStart.stamp(); }
  void click()   { m_cEnd.stamp(); }

  TimeStamp getTime() { return m_cEnd - m_cStart; }

private:
  TimeStamp m_cStart;
  TimeStamp m_cEnd;
};


class CtxIni
{
public:
  CtxIni() {}
  CtxIni( short v1, short v2 ) : m_sV1( v1 ), m_sV2( v2 ) {}

  short m_sV1;
  short m_sV2;
};


class CtxBin
{
public:
  CtxBin( UInt uiCtx, bool bBin ) : m_uiCtxBin( (uiCtx<<1) | (bBin ? 1 : 0) ) {}
  explicit CtxBin( UInt uiCtx )   : m_uiCtxBin(  uiCtx<<1                   ) {}

  void setBin( bool bBin )       { m_uiCtxBin = ( m_uiCtxBin & ~UInt(1) ) | (bBin ? 1 : 0); }
  bool getBin()            const { return m_uiCtxBin & 1 ? true : false; }
  UInt getCtx()            const { return m_uiCtxBin >> 1; }

private:
  UInt m_uiCtxBin;
};



#if COLLECT_CABAC_BITS

class EPCtxBin
{
public:
  EPCtxBin() {}
  EPCtxBin( bool b, UInt ui ) : m_bBin( b ), m_uiCtx( ui ) {}

  bool m_bBin;
  UInt m_uiCtx;
};

#else

typedef bool EPCtxBin;

#endif



template<UInt t_uiAccuracy>
class BinaryFraction
{
public:
  template<UInt t_ui> friend class BinaryFraction;

  BinaryFraction() {}
  BinaryFraction( UInt i ) : m_iValue( Int64( i ) << t_uiAccuracy ) {}
  BinaryFraction( int i ) : m_iValue( Int64( i ) << t_uiAccuracy ) {}
  BinaryFraction( Int64 i ) : m_iValue( i << t_uiAccuracy ) {}
  BinaryFraction( double d ) : m_iValue( Int64( d * (1 << t_uiAccuracy) ) ) {}

  BinaryFraction( UInt64 uiRaw, bool bDummy ) : m_iValue( uiRaw ) {}

  void  setNumerator( Int64 i ) { m_iValue = i; }
  Int64 getNumerator() const { return m_iValue; }
  UInt  getAccuracy () const { return t_uiAccuracy; }

  double getValueDouble() const { return double(m_iValue) / double(UInt64(1)<<t_uiAccuracy); }

  const BinaryFraction& operator=( const BinaryFraction &f ) { m_iValue = f.m_iValue; return *this; }

  template<UInt t_uiAccuracyRight>
  BinaryFraction<(t_uiAccuracyRight<t_uiAccuracy?t_uiAccuracy:t_uiAccuracyRight)> operator+( const BinaryFraction<t_uiAccuracyRight> &f ) const
  {
    const UInt uiAccRet = t_uiAccuracyRight<t_uiAccuracy?t_uiAccuracy:t_uiAccuracyRight;
    BinaryFraction<uiAccRet> x;
    x.m_iValue = (m_iValue << (uiAccRet-t_uiAccuracy)) + (f.m_iValue << (uiAccRet-t_uiAccuracyRight));
    return x;
  }

  template<UInt t_uiAccuracyRight>
  BinaryFraction<(t_uiAccuracyRight<t_uiAccuracy?t_uiAccuracy:t_uiAccuracyRight)> operator-( const BinaryFraction<t_uiAccuracyRight> &f ) const
  {
    const UInt uiAccRet = t_uiAccuracyRight<t_uiAccuracy?t_uiAccuracy:t_uiAccuracyRight;
    BinaryFraction<uiAccRet> x;
    x.m_iValue = (m_iValue << (uiAccRet-t_uiAccuracy)) - (f.m_iValue << (uiAccRet-t_uiAccuracyRight));
    return x;
  }

  BinaryFraction& operator+=( const BinaryFraction &f ) { m_iValue += f.m_iValue; return *this; }
  BinaryFraction& operator-=( const BinaryFraction &f ) { m_iValue -= f.m_iValue; return *this; }

  template<UInt t_uiAccuracyRight>
  BinaryFraction<t_uiAccuracy+t_uiAccuracyRight> operator*( const BinaryFraction<t_uiAccuracyRight> &rcRight ) const
  {
    BinaryFraction<t_uiAccuracy+t_uiAccuracyRight> cRet;
    cRet.m_iValue = m_iValue * rcRight.m_iValue;
    return cRet;
  }

  bool operator==( const BinaryFraction &rcF ) const { return m_iValue == rcF.m_iValue; }
  bool operator!=( const BinaryFraction &rcF ) const { return m_iValue != rcF.m_iValue; }
  bool operator<=( const BinaryFraction &rcF ) const { return m_iValue <= rcF.m_iValue; }
  bool operator>=( const BinaryFraction &rcF ) const { return m_iValue >= rcF.m_iValue; }
  bool operator<( const BinaryFraction &rcF ) const { return m_iValue < rcF.m_iValue; }
  bool operator>( const BinaryFraction &rcF ) const { return m_iValue > rcF.m_iValue; }

private:
  Int64 m_iValue;
};

typedef BinaryFraction<16> BitRate;
typedef Int64 PicOrderCnt;


// video namespace end
VIDEO_NAMESPACE_END

#endif //!defined(AFX_COMMONTYPES_H__CBFE313E_2382_4ECC_9D41_416668E3507D__INCLUDED_)
