/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _CONTEXTINIT_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _CONTEXTINIT_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class CtxAccess
{
public:
  void appendCtxInitCell( UInt uiCellSize, std::string acCellName )
  {
    m_aiCellOffset.push_back( m_aiCellOffset.empty() ? 0 : m_aiCellOffset.back() + m_aiCellSize.back() );
    m_aiCellSize.push_back( uiCellSize );
#if COLLECT_CABAC_BITS
    m_acCellNames.push_back( acCellName );
#endif
  }

  __inline int getContextOffset( UInt uiCtxArrayIdx, UInt uiOffset = 0 ) const
  {
    AOT( uiCtxArrayIdx >= m_aiCellSize.size() );

    AOT( int( uiOffset ) >= m_aiCellSize[uiCtxArrayIdx] );
    return m_aiCellOffset[uiCtxArrayIdx] + uiOffset;
  }

  int getNumCtxElements( UInt uiIdx ) const { return m_aiCellSize[uiIdx]; }
  int getNumCtxInitArrays() const { return int( m_aiCellOffset.size() ); }

#if COLLECT_CABAC_BITS
  const std::string& getCtxName( UInt uiIdx ) const { return m_acCellNames[uiIdx]; }
#endif

private:
  std::vector<int> m_aiCellOffset;
  std::vector<int> m_aiCellSize;
#if COLLECT_CABAC_BITS
  std::vector<std::string> m_acCellNames;
#endif
};


class CtxInitArray
{
public:
  CtxInitArray() {}

  explicit CtxInitArray( const CtxAccess &rcAccess, const std::vector<CtxIni> &racInit )
    : m_acInit( racInit )
    , m_cCtxAccess( rcAccess ) {}

  explicit CtxInitArray( const std::string &rcCellName, const std::vector<CtxIni> &r )
    : m_acInit( r )
  {
    m_cCtxAccess.appendCtxInitCell( UInt( r.size() ), rcCellName );
  }

  const CtxInitArray operator<<( const CtxInitArray &r ) const
  {
    CtxInitArray c = *this;
    for( UInt ui = 0; ui < UInt( r.getCtxAccess().getNumCtxInitArrays() ); ui++ )
    {
#if COLLECT_CABAC_BITS
      c.m_cCtxAccess.appendCtxInitCell( UInt( r.getCtxAccess().getNumCtxElements( ui ) ), r.getCtxAccess().getCtxName( ui ) );
#else
      c.m_cCtxAccess.appendCtxInitCell( UInt( r.getCtxAccess().getNumCtxElements( ui ) ), "" );
#endif
    }

    c.m_acInit.insert( c.m_acInit.end(), r.m_acInit.begin(), r.m_acInit.end() );
    return c;
  }

  const std::vector<CtxIni>& getCtxInit()   const { return m_acInit; }
  const CtxAccess&           getCtxAccess() const { return m_cCtxAccess; }

private:
  std::vector<CtxIni> m_acInit;
  CtxAccess m_cCtxAccess;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _CONTEXTINIT_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
