/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _CONTEXTSTORE_H__3DAA82B6_8DEB_443A_A8FD_B9DE8C107AB0__INCLUDED_
#define _CONTEXTSTORE_H__3DAA82B6_8DEB_443A_A8FD_B9DE8C107AB0__INCLUDED_

#include "VideoCommonLib/ContextInit.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


enum CtxNames
{
  CTX_START_TREE_SUBDIV,
  CTX_START_PRED_MODE_INTRA,
  CTX_START_INTER_PLAPA_PRED,
  CTX_START_DIRECT_MODE,
  CTX_START_MERGE,
  CTX_START_MV_PRED_IDX,
  CTX_START_INTRA_SMOOTHED,
  CTX_START_MVD,
  CTX_START_NUM_MVS,
  CTX_START_RES_MODE,
  CTX_START_RES_SKIP_MODE,
  CTX_START_POW2_DQP,
  CTX_START_POW2_SUBDIV,
};

enum CtxWiener
{
  CTX_START_WIENER_TREE_SUBDIV,
  CTX_WIENER_FLAG,
  CTX_WIENER_SEPARABLE_FLAG,
  CTX_WIENER_MAX_BLOCK,
  CTX_WIENER_MIN_BLOCK,
  CTX_WIENER_FILTER_USED_IN_FRAME,
  CTX_WIENER_FILTER_TYPE,
  CTX_WIENER_FILTER_LENGTH,
  CTX_WIENER_FILTER_SET_NUM,
  CTX_WIENER_QUADTREE_FLAG,
  CTX_WIENER_FILTER_COEFFS_SIGN,
  CTX_WIENER_FILTER_ABS_COEFFS

};

enum ResidualCtxNames
{
  RES_CTX_START_POW2_CBF,
  RES_CTX_START_POW2_ABS_GREATER_ONE,
  RES_CTX_START_POW2_COEFF_LEVEL,
  RES_CTX_START_POW2_SIGN,
};

class ResidualContextStartIndices
{
public:
  UInt m_uiResCtxStart;
  UInt m_uiStartTrafoPow2Sig;
  UInt m_uiStartTrafoPow2Last;
};

class ContextArrayStartIndices
{
public:
  ContextArrayStartIndices() {}

  UInt m_uiStartRefIdx;
  std::vector<ResidualContextStartIndices> m_acResCtx;
};

template<class T>
class ContextStore
{
public:
  ContextStore()
    : m_pcCtxInitStates( NULL )
    , m_pcCtxInit( NULL )
    , m_pcCtxIndices( NULL ) {}

  ~ContextStore()
  {
    delete m_pcCtxInit;
    delete m_pcCtxIndices;
    delete[] m_pcCtxInitStates;

    for( UInt ui = 0; ui < m_apcCtxStates.size(); ui++ )
    {
      delete[] m_apcCtxStates[ui];
    }
  }

  bool isCtxStates( UInt uiSlice ) const
  {
    return m_apcCtxStates[uiSlice] != NULL;
  }

  bool isCtxInitStates() const
  {
    return m_pcCtxInitStates != NULL;
  }

  const CtxInitArray& getCtxInit() const
  {
    return *m_pcCtxInit;
  }

  const ContextArrayStartIndices& getCtxIndices() const
  {
    return *m_pcCtxIndices;
  }

  ErrVal setCtxInit( CtxInitArray* pcCtxInit )
  {
    m_pcCtxInit = pcCtxInit;
    return Err::m_nOK;
  }

  ErrVal setCtxIndices( ContextArrayStartIndices* pcCtxIndices )
  {
    m_pcCtxIndices = pcCtxIndices;
    return Err::m_nOK;
  }

  ErrVal getCtxStates( UInt uiSlice, T *&pacCtxStates ) const
  {
    if ( m_apcCtxStates[uiSlice] == NULL )
      pacCtxStates = m_pcCtxInitStates;
    else
      pacCtxStates = m_apcCtxStates[uiSlice];
    return Err::m_nOK;
  }

  ErrVal setCtxStates( UInt uiSlice, T *pacCtxStates )
  {
    if ( m_apcCtxStates[uiSlice] != NULL )
      delete[] m_apcCtxStates[uiSlice];
    m_apcCtxStates[uiSlice] = pacCtxStates;
    return Err::m_nOK;
  }

  ErrVal setCtxInitStates( T* pcCtxInitStates )
  {
    m_pcCtxInitStates = pcCtxInitStates;
    return Err::m_nOK;
  }

  ErrVal clearCtxStates()
  {
    for (UInt uiRun = 0; uiRun < m_apcCtxStates.size(); uiRun++)
    {
      delete[] m_apcCtxStates[uiRun];
      m_apcCtxStates[uiRun] = NULL;
    }
    return Err::m_nOK;
  }

  ErrVal setNumEntropySlices( UInt uiSlices )
  {
    m_pcCtxInitStates = NULL;
    m_apcCtxStates.resize( uiSlices );
    return Err::m_nOK;
  }

private:
  T* m_pcCtxInitStates;
  CtxInitArray* m_pcCtxInit;
  std::vector<T*> m_apcCtxStates;
  ContextArrayStartIndices* m_pcCtxIndices;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // _CONTEXTSTORE_H__3DAA82B6_8DEB_443A_A8FD_B9DE8C107AB0__INCLUDED_
