/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _DEBLOCKINGFILTER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _DEBLOCKINGFILTER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "AccessUnit.h"
#include "LeafNodeIterator.h"
#include "EdgeTree.h"
#include "InterTranslational.h"
#include "SliceData.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class PredInfoInterTranslational;

struct EdgeDeblockPars
{
  UInt m_uiBs;
  int m_iQps;
};

struct BlockDeblockPars
{
  BlockDeblockPars()
    : m_bIntra( false )
    , m_cPredSize( 0, 0 )
    , m_pcPredInfoInterTranslational( NULL ) {}

  bool m_bIntra;
  PrimBlockSize m_cPredSize;
  bool m_bSig;
  int m_iQp;
  const PredInfoInterTranslational *m_pcPredInfoInterTranslational;
};

class DeblockingFilter
{
public:
  virtual ~DeblockingFilter() {}

  static ErrVal expandEdgeTree( Subsampling cSub, UInt uiPlane, QuadtreeGroup<BlockDeblockPars> &rcBlockTree, const SliceData &rcSlice );

  template<class t_Pel>
  static ErrVal deblockPlanePackage( const PlanePackage &rcPlaPa,
                                     const PicBlock<t_Pel> &rcPic,
                                     const std::vector<bool> &rabIsLuma,
                                     UInt uiLog2MinTrafoSize,
                                     int iDeblockingAlphaC0Offset,
                                     int iDeblockingBetaOffset )
  {
    for( UInt uiPlane = 0; uiPlane < rcPlaPa.getPars().getNumPlanes(); uiPlane++ )
    {
      QuadtreeGroupStructure cDeblockTreeStructure( rcPlaPa.getCurrentSlice().getQuadtreeGroupStructure(), uiLog2MinTrafoSize );
      QuadtreeGroup<BlockDeblockPars> cBlockTree( &cDeblockTreeStructure, cDeblockTreeStructure );
      RNOK( expandEdgeTree( rcPlaPa.getPars().getSubsampling( uiPlane ), uiPlane, cBlockTree, rcPlaPa.getCurrentSlice() ) );

      EdgeTree<EdgeDeblockPars> cEdgeTree( cDeblockTreeStructure );
      DeblockParDeriver cDeblockParDeriver( rcPlaPa.getCurrentSlice().getQuadtreeGroupStructure(), cBlockTree, cEdgeTree, rabIsLuma[uiPlane], uiPlane );
      RNOK( cDeblockParDeriver.processTree() );
      if( rabIsLuma[uiPlane] )
      {
        {
          Deblocker<t_Pel, true, true> cHDeblocker( rcPic, cEdgeTree, uiPlane, iDeblockingAlphaC0Offset, iDeblockingBetaOffset );
          RNOK( cHDeblocker.processTree() );
          Deblocker<t_Pel, true, false> cVDeblocker( rcPic, cEdgeTree, uiPlane, iDeblockingAlphaC0Offset, iDeblockingBetaOffset );
          RNOK( cVDeblocker.processTree() );
        }
      }
      else
      {
        {
          Deblocker<t_Pel, false, true> cHDeblocker( rcPic, cEdgeTree, uiPlane, iDeblockingAlphaC0Offset, iDeblockingBetaOffset );
          RNOK( cHDeblocker.processTree() );
          Deblocker<t_Pel, false, false> cVDeblocker( rcPic, cEdgeTree, uiPlane, iDeblockingAlphaC0Offset, iDeblockingBetaOffset );
          RNOK( cVDeblocker.processTree() );
        }
      }
    }
    return Err::m_nOK;
  }

private:
  __inline static int xClip(const int iVal)
  {
    if( (UInt)iVal <= 0xff ) return iVal;
    return ( iVal < 0 ) ? 0 : 0xff;
  }

  class DeblockParDeriver : public LeafNodeIterator
  {
  public:
    DeblockParDeriver( const QuadtreeGroupStructure &rcPredInfoTreeStructure, const QuadtreeGroup<BlockDeblockPars> &rcDeblockTree, EdgeTree<EdgeDeblockPars> &rcEdgeTree, bool bIsLuma, UInt uiPlane )
      : LeafNodeIterator( rcDeblockTree.getTreeStructure() )
      , m_rcPredInfoTreeStructure( rcPredInfoTreeStructure )
      , m_rcDeblockTree( rcDeblockTree )
      , m_rcEdgeTree( rcEdgeTree )
      , m_bIsLuma( bIsLuma )
      , m_uiPlane( uiPlane ) {}

    template<QTNavi::Edge t_eEdge>
    ErrVal xDeriveDeblockPars( const QTNavi& rcPos, const PrimPixelPos &rcPrimPixelPos, const PrimPixelPos &rcOffsetToNeighbor );
    ErrVal xProcessLeafNode( const QTNavi& rcPos );

    static bool xCheckIfFilteringIsNeccessary( const std::vector<PredInfoInterTranslational::MvAndRefPoc> &rcMvA, const std::vector<PredInfoInterTranslational::MvAndRefPoc> &rcMvB );
    static bool xCheckIfMvAreSufficientlySimilar( const MotionVector &rcMvA, const MotionVector &rcMvB );
    static bool xCheckIfMvAndRefIdxAreSufficientlySimilar( const PredInfoInterTranslational::MvAndRefPoc &rcMvAndRefIdxA, const PredInfoInterTranslational::MvAndRefPoc &rcMvAndRefIdxB );

  private:
    const QuadtreeGroupStructure &m_rcPredInfoTreeStructure;
    const QuadtreeGroup<BlockDeblockPars> &m_rcDeblockTree;
    EdgeTree<EdgeDeblockPars> &m_rcEdgeTree;
    const bool m_bIsLuma;
    const UInt m_uiPlane;
  };

  template<class t_Pel, bool m_bIsLuma, bool m_bDeblockHorizontalEdges>
  class Deblocker : public LeafNodeIterator
  {
  public:
    Deblocker( const PicBlock<t_Pel> &rcPic,
               const EdgeTree<EdgeDeblockPars> &rcEdgeTree,
               UInt uiPlane,
               int iDeblockingAlphaC0Offset,
               int iDeblockingBetaOffset)
      : LeafNodeIterator( rcEdgeTree.getQuadtreeGroupStructure() )
      , m_rcPic( rcPic )
      , m_rcEdgeTree( rcEdgeTree )
      , m_uiPlane( uiPlane )
      , m_iDeblockingAlphaC0Offset( iDeblockingAlphaC0Offset )
      , m_iDeblockingBetaOffset( iDeblockingBetaOffset )
    {}

  private:
    ErrVal xProcessLeafNode( const QTNavi &rcPos );
    template<QTNavi::Edge t_eEdge> ErrVal xDeblockEdge( const QTNavi &rcPos, const PrimPixelPos& rcPrimPixelPos, const std::vector<QTNavi>& rcNavis, UInt uiPlane );

    const PicBlock<t_Pel> &m_rcPic;
    const EdgeTree<EdgeDeblockPars> &m_rcEdgeTree;
    const UInt m_uiPlane;
    const int m_iDeblockingAlphaC0Offset;
    const int m_iDeblockingBetaOffset;
  };


  template<class t_Pel, bool m_bIsLuma>
  static void xDeblock( t_Pel* pFlt, const int& iOffset, const int& iIndexA, const int& iIndexB, const UChar& ucBs );

  typedef struct
  {
    UChar ucAlpha;
    UChar aucClip[3];
  } AlphaClip;

  static const UChar sm_aucBetaTab[52];
  static const AlphaClip sm_acAlphaClip[52];

};


template<QTNavi::Edge t_eEdge>
ErrVal DeblockingFilter::DeblockParDeriver::xDeriveDeblockPars( const QTNavi &rcPos, const PrimPixelPos &rcPrimPixelPos, const PrimPixelPos &rcOffsetToNeighbor )
{
  std::vector<QTNavi> cNeighbours; //th to move outside

  if( m_rcEdgeTree.hasEdgeThatIsStoredInCNav<t_eEdge>( rcPos, rcPrimPixelPos, cNeighbours ) )
  {
    EdgeDeblockPars &rcEdgePars = m_rcEdgeTree.getEdge1<t_eEdge>( rcPos, cNeighbours );
    const QuadtreeGroupStructure &rcTree = m_rcEdgeTree.getQuadtreeGroupStructure();
    QTNavi cNeighbor = rcTree.getNodeByPrimPixelPos( rcPrimPixelPos + rcOffsetToNeighbor );
    if( cNeighbor.isLeafNode() )
    {
      const BlockDeblockPars &rcBlockParsCurr = m_rcDeblockTree.getT( rcPos );
      const BlockDeblockPars &rcBlockParsNeighbour = m_rcDeblockTree.getT( cNeighbor );

      rcEdgePars.m_iQps = ( rcBlockParsCurr.m_iQp + rcBlockParsNeighbour.m_iQp + 1 ) >> 1;

      if( rcBlockParsCurr.m_bIntra || rcBlockParsNeighbour.m_bIntra )
      {
        const PrimBlockSize cMaxPredPrimBlkSizeForBs3( 8, 8 );
        const bool          bInterIntra     = ( !rcBlockParsCurr.m_bIntra || !rcBlockParsNeighbour.m_bIntra );
        const bool          bPredBlkBorder  = ( m_rcPredInfoTreeStructure.getNodeByPrimPixelPos( rcPrimPixelPos ) != m_rcPredInfoTreeStructure.getNodeByPrimPixelPos( cNeighbor.getPrimPixelPos() ) );
        const bool          bLargePredSize  = ( rcBlockParsCurr.m_cPredSize > cMaxPredPrimBlkSizeForBs3 || rcBlockParsNeighbour.m_cPredSize > cMaxPredPrimBlkSizeForBs3 );
        if( bInterIntra || ( bPredBlkBorder && bLargePredSize ) )
        {
          rcEdgePars.m_uiBs = 4;
        }
        else
        {
          rcEdgePars.m_uiBs = 3;
        }
      }
      else
      {
        if( rcBlockParsCurr.m_bSig || rcBlockParsNeighbour.m_bSig )
        {
          rcEdgePars.m_uiBs = 2;
        }
        else
        {
          rcEdgePars.m_uiBs = xCheckIfFilteringIsNeccessary( rcBlockParsCurr.m_pcPredInfoInterTranslational->getMvAndRefPocVec(), rcBlockParsNeighbour.m_pcPredInfoInterTranslational->getMvAndRefPocVec() ) ? 1 : 0;
        }
      }
    }
    else
    {
      rcEdgePars.m_uiBs = 0;
    }
  }
  return Err::m_nOK;
}

inline
bool DeblockingFilter::DeblockParDeriver::xCheckIfMvAreSufficientlySimilar( const MotionVector &rcMvA, const MotionVector &rcMvB )
{
  const int iMaxPrec = gMax( rcMvA.getPrecision(), rcMvB.getPrecision() );
  const MotionVector cDiff = rcMvA.getMvWithPrecisionIncreasedTo( iMaxPrec ) - rcMvB.getMvWithPrecisionIncreasedTo( iMaxPrec );
  const int iThreshold = 1 << cDiff.getPrecision();
  return abs( cDiff.getX() ) < iThreshold && abs( cDiff.getY() ) < iThreshold;
}

inline
bool DeblockingFilter::DeblockParDeriver::xCheckIfMvAndRefIdxAreSufficientlySimilar( const PredInfoInterTranslational::MvAndRefPoc &rcMvAndRefIdxA, const PredInfoInterTranslational::MvAndRefPoc &rcMvAndRefIdxB )
{
  if( rcMvAndRefIdxA.m_cRefPoc == rcMvAndRefIdxB.m_cRefPoc )
  {
    return xCheckIfMvAreSufficientlySimilar( rcMvAndRefIdxA.m_cMv, rcMvAndRefIdxB.m_cMv );
  }
  return false;
}

inline
bool DeblockingFilter::DeblockParDeriver::xCheckIfFilteringIsNeccessary( const std::vector<PredInfoInterTranslational::MvAndRefPoc> &rcMvA, const std::vector<PredInfoInterTranslational::MvAndRefPoc> &rcMvB )
{
  if( rcMvA.size() != rcMvB.size() )
  {
    return true;
  }

  if( rcMvA.size() == 1 )
  {
    return !xCheckIfMvAndRefIdxAreSufficientlySimilar( rcMvA[0], rcMvB[0] );
  }
  else if( rcMvA.size() == 2 )
  {
    const bool bDirectSimilar = xCheckIfMvAndRefIdxAreSufficientlySimilar( rcMvA[0], rcMvB[0] ) && xCheckIfMvAndRefIdxAreSufficientlySimilar( rcMvA[1], rcMvB[1] );
    const bool bCrossSimilar  = xCheckIfMvAndRefIdxAreSufficientlySimilar( rcMvA[0], rcMvB[1] ) && xCheckIfMvAndRefIdxAreSufficientlySimilar( rcMvA[1], rcMvB[0] );
    return !( bDirectSimilar || bCrossSimilar );
  }

  // TODO: determine according to marriage theorem
  return false;
}


inline
ErrVal DeblockingFilter::DeblockParDeriver::xProcessLeafNode( const QTNavi &rcPos )
{
  const PrimPixelPos cPrimPixelPos = rcPos.getPrimPixelPos();
  RNOK( xDeriveDeblockPars<QTNavi::TOP>( rcPos, cPrimPixelPos, PrimPixelPos( 0, -1 ) ) );
  RNOK( xDeriveDeblockPars<QTNavi::LEFT>( rcPos, cPrimPixelPos, PrimPixelPos( -1, 0 ) ) );
  const PrimBlockSize cPrimBlockSize = rcPos.getPrimBlockSize();
  RNOK( xDeriveDeblockPars<QTNavi::RIGHT>( rcPos, cPrimPixelPos, PrimPixelPos( cPrimBlockSize.getX(), 0 ) ) );
  RNOK( xDeriveDeblockPars<QTNavi::BOTTOM>( rcPos, cPrimPixelPos, PrimPixelPos( 0, cPrimBlockSize.getY() ) ) );
  return Err::m_nOK;
}


template<class t_Pel, bool m_bIsLuma>
void DeblockingFilter::xDeblock( t_Pel* pFlt, const int& iOffset, const int& iIndexA, const int& iIndexB, const UChar& ucBs )
{
  ROFVS( ucBs );
  const int iAlpha    = sm_acAlphaClip[ iIndexA ].ucAlpha;
  const int P0        = pFlt[  -iOffset];
  const int Q0        = pFlt[         0];
  const int iDelta    = Q0 - P0;
  const int iAbsDelta = abs( iDelta  );

  ROFVS( iAbsDelta < iAlpha );

  const int P1        = pFlt[-2*iOffset];
  const int Q1        = pFlt[   iOffset];
  const int iBeta     = sm_aucBetaTab[ iIndexB ];

  ROFVS( (abs(P0 - P1) < iBeta) && (abs(Q0 - Q1) < iBeta) );

  if( ucBs < 4 )
  {
     int C0 = sm_acAlphaClip[ iIndexA ].aucClip[ucBs-1];

    if( m_bIsLuma )
    {
      const int P2 = pFlt[-3*iOffset] ;
      const int Q2 = pFlt[ 2*iOffset] ;
      const int aq = (( abs( Q2 - Q0 ) < iBeta ) ? 1 : 0 );
      const int ap = (( abs( P2 - P0 ) < iBeta ) ? 1 : 0 );

      if( ap )
      {
        pFlt[-2*iOffset] = P1 + gMinMax((P2 + ((P0 + Q0 + 1)>>1) - (P1<<1)) >> 1, -C0, C0 );
      }

      if( aq  )
      {
        pFlt[   iOffset] = Q1 + gMinMax((Q2 + ((P0 + Q0 + 1)>>1) - (Q1<<1)) >> 1, -C0, C0 );
      }

      C0 += ap + aq -1;
    }

    C0++;
    const int iDiff = gMinMax(((iDelta << 2) + (P1 - Q1) + 4) >> 3, -C0, C0 ) ;
    pFlt[-iOffset] = xClip( P0 + iDiff );
    pFlt[       0] = xClip( Q0 - iDiff );
    return;
  }

  if( ! m_bIsLuma )
  {
    pFlt[         0] = ((Q1 << 1) + Q0 + P1 + 2) >> 2;
    pFlt[  -iOffset] = ((P1 << 1) + P0 + Q1 + 2) >> 2;
  }
  else
  {
    const int  P2       = pFlt[-3*iOffset] ;
    const int  Q2       = pFlt[ 2*iOffset] ;
    const bool bEnable  = (iAbsDelta < ((iAlpha >> 2) + 2));
    const bool aq       = bEnable & ( abs( Q2 - Q0 ) < iBeta );
    const bool ap       = bEnable & ( abs( P2 - P0 ) < iBeta );
    const int  PQ0      = P0 + Q0;

    if( aq )
    {
      pFlt[         0] = (P1 + ((Q1 + PQ0) << 1) +  Q2 + 4) >> 3;
      pFlt[   iOffset] = (PQ0 +Q1 + Q2 + 2) >> 2;
      pFlt[ 2*iOffset] = (((pFlt[ 3*iOffset] + Q2) <<1) + Q2 + Q1 + PQ0 + 4) >> 3;
    }
    else
    {
      pFlt[         0] = ((Q1 << 1) + Q0 + P1 + 2) >> 2;
    }

    if( ap )
    {
      pFlt[  -iOffset] = (Q1 + ((P1 + PQ0) << 1) +  P2 + 4) >> 3;
      pFlt[-2*iOffset] = (PQ0 +P1 + P2 + 2) >> 2;
      pFlt[-3*iOffset] = (((pFlt[-4*iOffset] + P2) <<1) + pFlt[-3*iOffset] + P1 + PQ0 + 4) >> 3;
    }
    else
    {
      pFlt[  -iOffset] = ((P1 << 1) + P0 + Q1 + 2) >> 2;
    }
  }
}


template<class t_Pel, bool m_bIsLuma, bool m_bDeblockHorizontalEdges>
ErrVal DeblockingFilter::Deblocker<t_Pel,m_bIsLuma,m_bDeblockHorizontalEdges>::xProcessLeafNode( const QTNavi &rcPos )
{
  const PrimPixelPos cPrimPixelPos = rcPos.getPrimPixelPos();

  if( m_bDeblockHorizontalEdges )
  {
    std::vector<QTNavi> cNeighbours; //th
    if( m_rcEdgeTree.hasEdgeThatIsStoredInCNav<QTNavi::TOP>( rcPos, cPrimPixelPos, cNeighbours ) )
    {
      RNOK( xDeblockEdge<QTNavi::TOP>( rcPos, cPrimPixelPos, cNeighbours, m_uiPlane ) );
    }

    cNeighbours.resize(0);
    if( m_rcEdgeTree.hasEdgeThatIsStoredInCNav<QTNavi::BOTTOM>( rcPos, cPrimPixelPos, cNeighbours ) )
    {
      RNOK( xDeblockEdge<QTNavi::BOTTOM>( rcPos, cPrimPixelPos, cNeighbours, m_uiPlane ) );
    }
  }
  else
  {
    std::vector<QTNavi> cNeighbours; //th
    if( m_rcEdgeTree.hasEdgeThatIsStoredInCNav<QTNavi::LEFT>( rcPos, cPrimPixelPos, cNeighbours ) )
    {
      RNOK( xDeblockEdge<QTNavi::LEFT>( rcPos, cPrimPixelPos, cNeighbours, m_uiPlane ) );
    }

    cNeighbours.resize(0);
    if( m_rcEdgeTree.hasEdgeThatIsStoredInCNav<QTNavi::RIGHT>( rcPos, cPrimPixelPos, cNeighbours ) )
    {
      RNOK( xDeblockEdge<QTNavi::RIGHT>( rcPos, cPrimPixelPos, cNeighbours, m_uiPlane ) );
    }
  }

  return Err::m_nOK;
}


template<class t_Pel, bool m_bIsLuma, bool m_bDeblockHorizontalEdges>
template<QTNavi::Edge t_eEdge>
ErrVal DeblockingFilter::Deblocker<t_Pel, m_bIsLuma, m_bDeblockHorizontalEdges>::xDeblockEdge( const QTNavi& rcPos, const PrimPixelPos& rcPrimPixelPos, const std::vector<QTNavi>& rcNavis, const UInt uiPlane )
{
  const EdgeDeblockPars& rcEdgeDeblockPars = m_rcEdgeTree.getEdge1<t_eEdge>( rcPos, rcNavis );
  ROFRS( rcEdgeDeblockPars.m_uiBs, Err::m_nOK );
  const UInt          uiBs        = rcEdgeDeblockPars.m_uiBs;
  const int           iQp         = rcEdgeDeblockPars.m_iQps;
  const Subsampling&  rcSub       = m_rcPic.getStats().getSubsampling()[ uiPlane ];
  const BlockSize     cBlockSize  = rcPos.getPrimBlockSize() * rcSub;
  const PixelPos      cQ0Pos      = rcPrimPixelPos * rcSub + PixelPos( t_eEdge == QTNavi::RIGHT  ? cBlockSize.getX() : 0,
                                                                       t_eEdge == QTNavi::BOTTOM ? cBlockSize.getY() : 0 );
  
  const int           iMaxSamplesToAccessDiv2 = ( m_bIsLuma ? 4 : 2 );
  const int           iMaxSamplesToAccess     = iMaxSamplesToAccessDiv2 << 1;
  const BlockSize     cFilterBlockSize        = ( m_bDeblockHorizontalEdges ? BlockSize( cBlockSize.getX(), iMaxSamplesToAccess ) : BlockSize( iMaxSamplesToAccess, cBlockSize.getY() ) );
  const PixelPos      cTopLeft                = cQ0Pos + ( m_bDeblockHorizontalEdges ? PixelPos( 0, -iMaxSamplesToAccessDiv2 ) : PixelPos( -iMaxSamplesToAccessDiv2, 0 ) );
  const Block<t_Pel>  cFilterBlock            ( cTopLeft, cFilterBlockSize, m_rcPic.getPlane( uiPlane ) );

  const PixelPos      cOffsetToQ0             = ( m_bDeblockHorizontalEdges ? PixelPos( 0, iMaxSamplesToAccessDiv2 ) : PixelPos( iMaxSamplesToAccessDiv2, 0 ) );
  const UInt          uiOffset                = ( m_bDeblockHorizontalEdges ? cFilterBlock.getStride() : 1 );
  const int           iAdd                    = ( m_bDeblockHorizontalEdges ? 1 : cFilterBlock.getStride() );
  const UInt          uiEdgeLength            = ( m_bDeblockHorizontalEdges ? cBlockSize.getX() : cBlockSize.getY() );
  const int           iIndexA                 = gMinMax( iQp + m_iDeblockingAlphaC0Offset, 0, 51 );
  const int           iIndexB                 = gMinMax( iQp + m_iDeblockingBetaOffset, 0, 51 );

  t_Pel *pcQ0 = &cFilterBlock.getEl( cOffsetToQ0 );
  for( UInt ui = 0; ui < uiEdgeLength; ui++ )
  {
    xDeblock<t_Pel,m_bIsLuma>( pcQ0, uiOffset, iIndexA, iIndexB, uiBs );
    pcQ0 += iAdd;
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _DEBLOCKINGFILTER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
