/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _EDGETREE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _EDGETREE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "QuadtreeGroup.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class T>
class EdgeTree
{
public:
  EdgeTree( QuadtreeGroupStructure &rcStructure ) : m_cTree( &rcStructure, rcStructure ){}
  EdgeTree( QuadtreeGroupStructure &rcStructure, const T &r ) : m_cTree( &rcStructure, rcStructure, EdgeNode<T>( r ) ){}

  void setAllEdgesTo( const T &rcT )
  {
    EdgeNode<T> cEdges( rcT );
    m_cTree.setAllElementsTo( cEdges );
  }


  template<QTNavi::Edge t_eEdge>
  bool hasEdge( const QTNavi& rcNav, const std::vector<QTNavi>& cNeighbours ) const
  {
    return cNeighbours.size() <= 1;
  }

  template<QTNavi::Edge t_eEdge>
  bool hasEdgeThatIsStoredInCNav( const QTNavi &rcNav, const PrimPixelPos& rcPrimPixelPos, std::vector<QTNavi>& rcNeighbours ) const
  {
    rcNav.getEdgeNeighbours<t_eEdge,true>( rcNeighbours, rcPrimPixelPos );
    return xEdgeIsStoredHere<t_eEdge>( rcNeighbours, rcNav );
  }


  template<QTNavi::Edge t_eEdge> void getEdges( const QTNavi &rcNav, const PrimPixelPos& rcPrimPixelPos, std::vector<const T*> &rcVec ) const { xGetEdges<t_eEdge, const T>( rcNav, rcPrimPixelPos, rcVec, m_cTree ); }
  template<QTNavi::Edge t_eEdge> void getEdges( const QTNavi &rcNav, const PrimPixelPos& rcPrimPixelPos, std::vector<T*>       &rcVec )       { xGetEdges<t_eEdge, T>      ( rcNav, rcPrimPixelPos, rcVec, m_cTree );}


  template<QTNavi::Edge t_eEdge>
  T& getEdge1( const QTNavi &rcNav, const std::vector<QTNavi>& rcNeighbours )
  {
    QTNavi cLocation = rcNav;
    ANOK( xGetEdgeLocation<t_eEdge>( cLocation, rcNeighbours ) );
    return m_cTree.getT( cLocation ).m_cEdges[ cLocation == rcNav ? t_eEdge : OPPOSITE_EDGE_OF( t_eEdge ) ];
  }

  template<QTNavi::Edge t_eEdge>
  const T& getEdge1( const QTNavi &rcNav, const std::vector<QTNavi>& rcNeighbours ) const
  {
    QTNavi cLocation = rcNav;
    ANOK( xGetEdgeLocation<t_eEdge>( cLocation, rcNeighbours ) );
    return m_cTree.getT( cLocation ).m_cEdges[ cLocation == rcNav ? t_eEdge : OPPOSITE_EDGE_OF( t_eEdge ) ];
  }

  const QuadtreeGroupStructure& getQuadtreeGroupStructure() const { return m_cTree.getTreeStructure(); }

private:
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( EdgeTree );

  template<QTNavi::Edge t_eEdge, class t_Ret, class t_TreeType>
  static void xGetEdges( const QTNavi &rcNav, const PrimPixelPos& rcPrimPixelPos, std::vector<t_Ret*> &rcVec, t_TreeType &rcTree )
  {
    QTNavi::Edge cActiveEdge;
    std::vector<QTNavi> acEdgesNavis;

    rcNav.getEdgeNeighbours<t_eEdge, false>( acEdgesNavis, rcPrimPixelPos );
    if( xEdgeIsStoredHere<t_eEdge>( acEdgesNavis, rcNav ) )
    {
      cActiveEdge = t_eEdge;
      acEdgesNavis.clear();
      acEdgesNavis.push_back( rcNav );
    }
    else
    {
      cActiveEdge = OPPOSITE_EDGE_OF( t_eEdge );
    }


    rcVec.reserve( rcVec.size() + acEdgesNavis.size() );
    for( UInt ui = 0; ui < acEdgesNavis.size(); ui++ )
    {
      rcVec.push_back( &rcTree.getT( acEdgesNavis[ui] ).m_cEdges[cActiveEdge] );
    }
  }

  template<QTNavi::Edge t_eEdge>
  static bool xEdgeIsStoredHere( const std::vector<QTNavi> &rcNeighbours, const QTNavi &rcNav )
  {
    bool bStoreHere = false;
    if( rcNeighbours.empty() )
    {
      bStoreHere = true;
    }
    else if( rcNeighbours.size() > 1 )
    {
      bStoreHere = false;
    }
    else
    {
      const bool bHor = t_eEdge == QTNavi::TOP || t_eEdge == QTNavi::BOTTOM;
      const int iNeighbourDim = bHor ? rcNeighbours[0].getPrimBlockSize().getX() : rcNeighbours[0].getPrimBlockSize().getY();
      const int iCurrentDim   = bHor ? rcNav          .getPrimBlockSize().getX() : rcNav          .getPrimBlockSize().getY();
      const bool bTopOrLeft = t_eEdge == QTNavi::TOP || t_eEdge == QTNavi::LEFT;

      bStoreHere = iCurrentDim < iNeighbourDim || ( bTopOrLeft && iCurrentDim == iNeighbourDim );
    }
    return bStoreHere;
  }

  template<QTNavi::Edge t_eEdge>
  ErrVal xGetEdgeLocation( QTNavi& rcNav, const std::vector<QTNavi>& rcNeighbours ) const
  {
    ROF( rcNav.isLeafNode() );

    if( rcNeighbours.empty() )
    {
      return Err::m_nOK;
    }

    if( rcNeighbours.size() != 1 )
    {
      return Err::m_nERR;
    }

    const bool bHor = t_eEdge == QTNavi::TOP || t_eEdge == QTNavi::BOTTOM;
    const int iNeighbourDim = bHor ? rcNeighbours[0].getPrimBlockSize().getX() : rcNeighbours[0].getPrimBlockSize().getY();
    const int iCurrentDim   = bHor ? rcNav          .getPrimBlockSize().getX() : rcNav          .getPrimBlockSize().getY();

    const bool bTopOrLeft = t_eEdge == QTNavi::TOP || t_eEdge == QTNavi::LEFT;
    bool bStoreHere = iCurrentDim < iNeighbourDim || ( bTopOrLeft && iCurrentDim == iNeighbourDim );

    if( !bStoreHere )
    {
      rcNav = rcNeighbours[0];
    }

    return Err::m_nOK;
  }

  template<class U>
  class EdgeNode
  {
  public:
    EdgeNode() {}
    EdgeNode( const U &r )
    {
      m_cEdges[0] = r;
      m_cEdges[1] = r;
      m_cEdges[2] = r;
      m_cEdges[3] = r;
    }

    U m_cEdges[4];
  };

  QuadtreeGroup<EdgeNode<T> > m_cTree;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _EDGETREE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
