/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _EXTENDEDCONTEXTMODEL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _EXTENDEDCONTEXTMODEL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_


// video namespace begin
VIDEO_NAMESPACE_BEGIN

class BitCounterContextModelCapIf
{
protected:
  virtual ~BitCounterContextModelCapIf() {}
public:
  virtual double getBitsCabac() const = 0;
  virtual UInt64 getBinsCabac() const = 0;
};

template<class t_ModelBase>
class BitCounterContextModel
  : public t_ModelBase, public BitCounterContextModelCapIf
{
public:
  typedef t_ModelBase BaseType;
  typedef BitCounterContextModelCapIf CapsType;

  BitCounterContextModel()
    : m_dBits( 0.0 )
    , m_uiOnes( 0 )
    , m_uiZeros( 0 ) {}

  double getBitsCabac() const { return m_dBits; }
  UInt64 getBinsCabac() const { return m_uiOnes + m_uiZeros; }

  FORCEINLINE void addBin( bool b, UInt uiBinIndex )
  {
    if( b )
    {
      m_uiOnes++;
    }
    else
    {
      m_uiZeros++; 
    }
    t_ModelBase::addBin( b, uiBinIndex );
  }

  FORCEINLINE void addBits( double d )
  {
    m_dBits += d;
    t_ModelBase::addBits( d );
  }

  FORCEINLINE void initCCModel( const CtxIni &rcInitItem, int iQp )
  {
    m_dBits = 0.0;
    m_uiOnes = 0;
    m_uiZeros = 0;
    t_ModelBase::initCCModel( rcInitItem, iQp );
  }

private:
  double m_dBits;
  UInt64 m_uiOnes;
  UInt64 m_uiZeros;
};


class IndexedBin
{
public:
  IndexedBin() {}
  IndexedBin( bool bBin, UInt uiBinIndex ) : m_uiIndexedBin( ( uiBinIndex << 1 ) | (bBin ? 1 : 0) ) {}

  bool getBin() const { return (m_uiIndexedBin & 1) ? true : false; }
  UInt64 getBinIndex() const { return m_uiIndexedBin >> 1; }

  bool operator<( const IndexedBin &r ) const { return m_uiIndexedBin < r.m_uiIndexedBin; }

private:
//  UInt64 m_uiIndexedBin;
  UInt m_uiIndexedBin;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _BLOCK_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
