/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_PICTUREPARAMETERSET_H__8ED333BE_D213_4BFF_A379_67DDDA7F090C__INCLUDED_)
#define AFX_PICTUREPARAMETERSET_H__8ED333BE_D213_4BFF_A379_67DDDA7F090C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VideoCommonLib/CommonClasses.h"
#include "VideoCommonLib/HeaderSymbolReadIf.h"
#include "VideoCommonLib/HeaderSymbolWriteIf.h"

// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

class AccessUnitParameters;
#if defined( WIN32 )
# pragma warning( disable: 4251 )
#endif

//!\ingroup H264CommonLib
//!\PictureParameterSet
class VIDEOCOMMONLIB_DECL PictureParameterSet
{
protected:
  PictureParameterSet();

public:
  virtual ~PictureParameterSet();

  ErrVal setAccessUnitParameters( const AccessUnitParameters &rcPars );

  bool operator != ( const PictureParameterSet& rcPPS ) const;
  bool operator == ( const PictureParameterSet& rcPPS ) const { return !(*this != rcPPS); }
  PictureParameterSet& operator = ( const PictureParameterSet& rcPPS );

  static ErrVal create( PictureParameterSet*& rpcPPS );
  ErrVal destroy();
  void setSeqParameterSetId( UInt uiSeqParameterSetId )                   { m_uiSeqParameterSetId = uiSeqParameterSetId; }
  void setPicParameterSetId( UInt uiPicParameterSetId )                   { m_uiPicParameterSetId = uiPicParameterSetId; }

  void setEntropyCodingModeFlag ( bool bEntropyCodingModeFlag )           { m_bEntropyCodingModeFlag = bEntropyCodingModeFlag ;}
  void setPicOrderPresentFlag ( bool bPicOrderPresentFlag )               { m_bPicOrderPresentFlag = bPicOrderPresentFlag ;}
  void setNumSliceGroups ( UInt uiNumSliceGroups )                        { m_uiNumSliceGroups = uiNumSliceGroups ;}
  void setMbSliceGroupMapType ( UInt uiMbSliceGroupMapType )              { m_uiMbSliceGroupMapType = uiMbSliceGroupMapType ;}
  void setRunLength ( UInt uiNumber, UInt uiRunLength )                   { m_auiRunLength[uiNumber] = uiRunLength ;}
  void setTopLeftMb ( UInt uiNumber, UInt uiTopLeftMb )                   { m_auiTopLeftMb[uiNumber] = uiTopLeftMb ;}
  void setBottomRightMb ( UInt uiNumber, UInt uiBottomRightMb )           { m_auiBottomRightMb[uiNumber] = uiBottomRightMb ;}
  void setSliceGroupChangeDirectionFlag ( bool bSliceGroupChangeDirectionFlag )  { m_bSliceGroupChangeDirectionFlag = bSliceGroupChangeDirectionFlag ;}
  void setSliceGroupChangeRate ( UInt uiSliceGroupChangeRate )            { m_uiSliceGroupChangeRate = uiSliceGroupChangeRate ;}
  void setPicSizeInMapUnits(UInt uiPicSizeInMapUnits)                     { m_uiPicSizeInMapUnits = uiPicSizeInMapUnits;}
//  void setSliceGroupId ( UInt uiNumber, UInt uiSliceGroupId )            { m_cSliceGroupIdBuffer.set(uiNumber, uiSliceGroupId) ;}
  void setNumRefIdxActive ( ListIdx eLstIdx, UInt uiNumRefIdxActive )     { m_auiNumRefIdxActive[eLstIdx] = uiNumRefIdxActive ;}
  void setWeightedPredFlag ( bool  bWeightedPredFlag )                    { m_bWeightedPredFlag = bWeightedPredFlag ;}
  void setWeightedBipredIdc ( int  uiWeightedBipredIdc )                  { m_uiWeightedBipredIdc = uiWeightedBipredIdc ;}
  void setPicInitQp ( int iPicInitQp )                                    { m_iPicInitQp = iPicInitQp ;}
  void setPicInitQs ( UInt uiPicInitQs )                                  { m_uiPicInitQs = uiPicInitQs ;}
  void setChromaQpIndexOffset ( int iChromaQpIndexOffset )                { m_iChromaQpIndexOffset = iChromaQpIndexOffset ;}
  void setDeblockingFilterParametersPresentFlag ( bool bDeblockingFilterParametersPresentFlag ) { m_bDeblockingFilterParametersPresentFlag = bDeblockingFilterParametersPresentFlag ;}
  void setConstrainedIntraPredFlag ( bool bConstrainedIntraPredFlag )     { m_bConstrainedIntraPredFlag = bConstrainedIntraPredFlag ;}
  void setRedundantPicCntPresentFlag ( bool bRedundantPicCntPresentFlag ) { m_bRedundantPicCntPresentFlag = bRedundantPicCntPresentFlag ;}
  UInt getPicParameterSetId()                       const { return m_uiPicParameterSetId;}
  UInt getSeqParameterSetId()                       const { return m_uiSeqParameterSetId;}
  bool getEntropyCodingModeFlag()                   const { return m_bEntropyCodingModeFlag;}
  bool getPicOrderPresentFlag()                     const { return m_bPicOrderPresentFlag;}
  UInt getNumSliceGroups()                          const { return m_uiNumSliceGroups;}
  UInt getMbSliceGroupMapType()                     const { return m_uiMbSliceGroupMapType;}
  UInt getRunLength(UInt uiNumber)                  const { return m_auiRunLength[uiNumber];}
  UInt getTopLeftMb(UInt uiNumber)                  const { return m_auiTopLeftMb[uiNumber];}
  UInt getBottomRightMb(UInt uiNumber)              const { return m_auiBottomRightMb[uiNumber];}
  bool getSliceGroupChangeDirectionFlag()           const { return m_bSliceGroupChangeDirectionFlag;}
  UInt getSliceGroupChangeRate()                    const { return m_uiSliceGroupChangeRate;}

  UInt getPicSizeInMapUnits()                       const { return  m_uiPicSizeInMapUnits;}
//  UInt getSliceGroupId(UInt uiMbAddress)            const { return m_cSliceGroupIdBuffer.get(uiMbAddress);}
  UInt getNumRefIdxActive( ListIdx eLstIdx )        const { return m_auiNumRefIdxActive[eLstIdx];}
  bool getWeightedPredFlag()                        const { return m_bWeightedPredFlag;}
  UInt getWeightedBipredIdc()                       const { return m_uiWeightedBipredIdc;}
  int  getPicInitQp()                               const { return m_iPicInitQp;}
  UInt getPicInitQs()                               const { return m_uiPicInitQs;}
   int getChromaQpIndexOffset()                      const { return m_iChromaQpIndexOffset;}
   int getChromaQpIndexOffset( bool bCb )            const { return bCb ? m_iChromaQpIndexOffset : m_iSecondChromaQpIndexOffset;}
  bool getDeblockingFilterParametersPresentFlag()   const { return m_bDeblockingFilterParametersPresentFlag;}
  bool getConstrainedIntraPredFlag()                const { return m_bConstrainedIntraPredFlag;}
  bool getRedundantPicCntPresentFlag()              const { return m_bRedundantPicCntPresentFlag;}

  bool check( Profile eProfile, UInt uiLevel )      const;
  ErrVal write( HeaderSymbolWriteIf* pcWriteIf )    const;
  ErrVal read( HeaderSymbolReadIf* pcReadIf );

  void setSecondChromaQpIndexOffset( int iSecondChromaQpIndexOffset) { m_iSecondChromaQpIndexOffset = iSecondChromaQpIndexOffset;}
  int getSecondChromaQpIndexOffset()                const { return m_iSecondChromaQpIndexOffset;}
  bool getTransform8x8ModeFlag()                    const { return  m_bTransform8x8ModeFlag; }
  void setTransform8x8ModeFlag( bool bTransform8x8ModeFlag ) { m_bTransform8x8ModeFlag = bTransform8x8ModeFlag; }


protected:
  ErrVal xWriteSliceGroupInfo( HeaderSymbolWriteIf* pcWriteIf )  const;
  ErrVal xReadSliceGroupInfo( HeaderSymbolReadIf* pcReadIf );
  bool xCheckProfile( Profile eProfile )            const;
  bool xCheckLevel( UInt uiLevel )                  const;
  ErrVal xDeriveChromaQpIndexOffsetFromLumaAndChromaQp( int& iChromaQpIndexOffset, int iLumaQp, int iChromaQp );

  ErrVal xReadFrext( HeaderSymbolReadIf* pcReadIf );
  ErrVal xWriteFrext( HeaderSymbolWriteIf* pcWriteIf ) const;


protected:
  UInt m_uiPicParameterSetId;
  UInt m_uiSeqParameterSetId;
  bool m_bEntropyCodingModeFlag;
  bool m_bPicOrderPresentFlag;
  UInt m_uiNumSliceGroups;
  UInt m_uiMbSliceGroupMapType;
  StatBuf<UInt,8> m_auiRunLength;   // Max of 8 slice groups in profile/level definition
  StatBuf<UInt,7> m_auiTopLeftMb;   // (Max of 8 slice groups in profile/level definition) - 1
  StatBuf<UInt,7> m_auiBottomRightMb;   // (Max of 8 slice groups in profile/level definition) - 1
  bool m_bSliceGroupChangeDirectionFlag;
  UInt m_uiSliceGroupChangeRate;
  UInt m_uiPicSizeInMapUnits;
  StatBuf<UInt,2> m_auiNumRefIdxActive;
  bool m_bWeightedPredFlag;
  UInt m_uiWeightedBipredIdc;
  int  m_iPicInitQp;
  UInt m_uiPicInitQs;
   int m_iChromaQpIndexOffset;
  bool m_bDeblockingFilterParametersPresentFlag;
  bool m_bConstrainedIntraPredFlag;
  bool m_bRedundantPicCntPresentFlag;

	bool m_bTransform8x8ModeFlag;
  int m_iSecondChromaQpIndexOffset;
};

#if defined( WIN32 )
# pragma warning( default: 4251 )
#endif

// h264 namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_PICTUREPARAMETERSET_H__8ED333BE_D213_4BFF_A379_67DDDA7F090C__INCLUDED_)
