/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_SEQUENCEPARAMETERSET_H__66281283_5BFB_429A_B722_6DDE7A11D086__INCLUDED_)
#define AFX_SEQUENCEPARAMETERSET_H__66281283_5BFB_429A_B722_6DDE7A11D086__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VideoCommonLib/CommonClasses.h"
#include "VideoCommonLib/HeaderSymbolReadIf.h"
#include "VideoCommonLib/HeaderSymbolWriteIf.h"

// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

class AccessUnitParameters;

#if defined( WIN32 )
# pragma warning( disable: 4251 )
#endif

class VUI;

//!\ingroup H264CommonLib
//!\SequenceParameterSet
class VIDEOCOMMONLIB_DECL SequenceParameterSet
{

public:
  class VIDEOCOMMONLIB_DECL FrameCroppingRect
  {
  public:

    bool operator != ( const FrameCroppingRect& rcFCR ) const;
    bool operator == ( const FrameCroppingRect& rcFCR ) const { return !(*this != rcFCR); }

    FrameCroppingRect();
    ~FrameCroppingRect();

    void setLeftOffset   ( UInt uiLeftOffset   )  { m_uiLeftOffset   = uiLeftOffset;   }
    void setRightOffset  ( UInt uiRightOffset  )  { m_uiRightOffset  = uiRightOffset;  }
    void setTopOffset    ( UInt uiTopOffset    )  { m_uiTopOffset    = uiTopOffset;    }
    void setBottomOffset ( UInt uiBottomOffset )  { m_uiBottomOffset = uiBottomOffset; }

    UInt getLeftOffset()                    const { return m_uiLeftOffset;   }
    UInt getRightOffset()                   const { return m_uiRightOffset;  }
    UInt getTopOffset()                     const { return m_uiTopOffset;    }
    UInt getBottomOffset()                  const { return m_uiBottomOffset; }

    ErrVal read( HeaderSymbolReadIf* pcReadIf );
    ErrVal write( HeaderSymbolWriteIf* pcWriteIf ) const;

    int getTopOffsetInLumaPelUnits( const ChromaFormatIdc eChromaFormatIdc, const bool bFrameMbsOnlyFlag ) const
    {
      return getTopOffset() * ((eChromaFormatIdc == YUV420) ? 2 : 1) *  (bFrameMbsOnlyFlag ? 1:2);
    }

    int getBotOffsetInLumaPelUnits( const ChromaFormatIdc eChromaFormatIdc, const bool bFrameMbsOnlyFlag ) const
    {
      return getBottomOffset() * ((eChromaFormatIdc == YUV420) ? 2 : 1) *  (bFrameMbsOnlyFlag ? 1:2);
    }

    int getLeftOffsetInLumaPelUnits( const ChromaFormatIdc eChromaFormatIdc ) const
    {
      return getLeftOffset() * (((eChromaFormatIdc == YUV420) || (eChromaFormatIdc == YUV422)) ? 2 : 1);
    }

    int getRightOffsetInLumaPelUnits( const ChromaFormatIdc eChromaFormatIdc ) const
    {
      return getRightOffset() * (((eChromaFormatIdc == YUV420) || (eChromaFormatIdc == YUV422)) ? 2 : 1);
    }

  private:
    UInt m_uiLeftOffset;
    UInt m_uiRightOffset;
    UInt m_uiTopOffset;
    UInt m_uiBottomOffset;
  };

public:
  SequenceParameterSet();

public:
  virtual ~SequenceParameterSet();

  static ErrVal create( SequenceParameterSet*& rpcSPS );
  ErrVal destroy();

  bool operator != ( const SequenceParameterSet& rcSPS ) const;
  bool operator == ( const SequenceParameterSet& rcSPS ) const { return !(*this != rcSPS); }
  SequenceParameterSet& operator = ( const SequenceParameterSet& rcSPS );

  UInt getMaxDPBSize() const;
  void setConstrainedSet0Flag( bool bConstrainedSet0Flag)                           { m_bConstrainedSet0Flag = bConstrainedSet0Flag; }
  void setConstrainedSet1Flag( bool bConstrainedSet1Flag)                           { m_bConstrainedSet1Flag = bConstrainedSet1Flag; }
  void setConstrainedSet2Flag( bool bConstrainedSet2Flag)                           { m_bConstrainedSet2Flag = bConstrainedSet2Flag; }

  void setSeqParameterSetId( UInt uiSeqParameterSetId )                             { m_uiSeqParameterSetId = uiSeqParameterSetId; }

  void setProfileIdc( Profile eProfile)                                             { m_eProfileIdc = eProfile;}
  void setLevelIdc(UInt uiLevelIdc)                                                 { m_uiLevelIdc = uiLevelIdc;}
  void setLog2MaxFrameNum(UInt uiLog2MaxFrameNum)                                   { m_uiLog2MaxFrameNum = uiLog2MaxFrameNum;}
  void setPicOrderCntType(UInt uiPicOrderCntType)                                   { m_uiPicOrderCntType = uiPicOrderCntType;}
  void setLog2MaxPicOrderCntLsb (UInt uiLog2MaxPicOrderCntLsb)                      { m_uiLog2MaxPicOrderCntLsb = uiLog2MaxPicOrderCntLsb;}
  void setNumRefFramesInPicOrderCntCycle(UInt uiNumRefFramesInPicOrderCntCycle)     { m_uiNumRefFramesInPicOrderCntCycle = uiNumRefFramesInPicOrderCntCycle;}
  void setDeltaPicOrderAlwaysZeroFlag(bool bDeltaPicOrderAlwaysZeroFlag)            { m_bDeltaPicOrderAlwaysZeroFlag = bDeltaPicOrderAlwaysZeroFlag;}
  void setOffsetForNonRefPic(int iOffsetForNonRefPic)                               { m_iOffsetForNonRefPic = iOffsetForNonRefPic;}
  void setOffsetForTopToBottomField(int iOffsetForTopToBottomField)                 { m_iOffsetForTopToBottomField = iOffsetForTopToBottomField;}
  void setOffsetForRefFrame(UInt uiRefFrame, int iOffsetForRefFrame)                { m_piOffsetForRefFrame[uiRefFrame] = iOffsetForRefFrame;}
  void setNumRefFrames(UInt uiNumRefFrames)                                         { m_uiNumRefFrames = uiNumRefFrames;}
  void setGapsInFrameNumValueAllowedFlag(bool bGapsInFrameNumValueAllowedFlag)      { m_bGapsInFrameNumValueAllowedFlag = bGapsInFrameNumValueAllowedFlag;}
  void setFrameWidthInMbs(UInt uiFrameWidthInMbs)                                   { m_uiFrameWidthInMbs = uiFrameWidthInMbs;}
  void setFrameHeightInMbs(UInt uiFrameHeightInMbs)                                 { m_uiFrameHeightInMbs = uiFrameHeightInMbs;}
  void setFrameMbsOnlyFlag(bool FrameMbsOnlyFlag)                                   { m_bFrameMbsOnlyFlag = FrameMbsOnlyFlag;}
  void setMbAdaptiveFrameFieldFlag(bool MbAdaptiveFrameFieldFlag)                   { m_bMbAdaptiveFrameFieldFlag = MbAdaptiveFrameFieldFlag;}
  void setDirect8x8InferenceFlag(bool bDirect8x8InferenceFlag)                      { m_bDirect8x8InferenceFlag = bDirect8x8InferenceFlag;}
  void setFrameCroppingFlag ( bool bFrameCroppingFlag )                             { m_bFrameCroppingFlag = bFrameCroppingFlag ;  }
  void setVuiParametersPresentFlag(bool bVuiParametersPresentFlag)                  { m_bVuiParametersPresentFlag = bVuiParametersPresentFlag;}
  void setSPSRegId( UInt uiSPSRegId)                                                { m_uiSPSRegId = uiSPSRegId; }
    
  Profile getProfileIdc()                     const { return m_eProfileIdc;}
  bool getConstrainedSet0Flag()               const { return m_bConstrainedSet0Flag; }
  bool getConstrainedSet1Flag()               const { return m_bConstrainedSet1Flag; }
  bool getConstrainedSet2Flag()               const { return m_bConstrainedSet2Flag; }
  UInt getLevelIdc()                          const { return m_uiLevelIdc;}
  UInt getSeqParameterSetId()                 const { return m_uiSeqParameterSetId;}
  UInt getLog2MaxFrameNum()                   const { return m_uiLog2MaxFrameNum;}
  UInt getPicOrderCntType()                   const { return m_uiPicOrderCntType;}
  UInt getLog2MaxPicOrderCntLsb()             const { return m_uiLog2MaxPicOrderCntLsb;}
  UInt getNumRefFramesInPicOrderCntCycle()    const { return m_uiNumRefFramesInPicOrderCntCycle;}
  bool getDeltaPicOrderAlwaysZeroFlag()       const { return m_bDeltaPicOrderAlwaysZeroFlag;}
   int getOffsetForNonRefPic()                const { return m_iOffsetForNonRefPic;}
   int getOffsetForTopToBottomField()         const { return m_iOffsetForTopToBottomField;}
  UInt getOffsetForRefFrame(UInt uiRefFrame)  const { return m_piOffsetForRefFrame[uiRefFrame];}
  UInt getNumRefFrames()                      const { return m_uiNumRefFrames;}
  bool getGapsInFrameNumValueAllowedFlag()    const { return m_bGapsInFrameNumValueAllowedFlag;}
  UInt getFrameWidthInMbs()                   const { return m_uiFrameWidthInMbs;}
  UInt getFrameHeightInMbs()                  const { return m_uiFrameHeightInMbs;}
  bool getFrameMbsOnlyFlag()                  const { return m_bFrameMbsOnlyFlag;}
  bool getMbAdaptiveFrameFieldFlag()          const { return m_bMbAdaptiveFrameFieldFlag;}
  bool getDirect8x8InferenceFlag()            const { return m_bDirect8x8InferenceFlag;}
  bool getFrameCroppingFlag()                 const { return m_bFrameCroppingFlag;  }
  bool getVuiParametersPresentFlag()          const { return m_bVuiParametersPresentFlag;}
  UInt getChromaFormatParameterX2()           const { return m_uiChromaFormatParameterX2; }
  void setInitState(bool bInit)                     { m_bInitDone = bInit; }
  bool getInitState()                         const { return m_bInitDone; }
  UInt  getSPSRegId()                         const { return m_uiSPSRegId; }
  UInt getMbInFrame()                         const { return m_uiFrameWidthInMbs * m_uiFrameHeightInMbs;}



  void setQPPrimeTransformBypassFlag( bool bEnable ) { m_bQPPrimeTransformBypassFlag = bEnable; }
  bool getQPPrimeTransformBypassFlag()        const {return m_bQPPrimeTransformBypassFlag;}
  ChromaFormatIdc getChromaFormatIdc()        const { return m_eChromaFormatIdc; }
  void setChromaFormatIdc( ChromaFormatIdc eCFIDC)  { m_eChromaFormatIdc = eCFIDC; }
  
  UInt getMbByteSize()                              const 
  { 
    return 256+128*((1 << m_eChromaFormatIdc )>>1); 
  }
  ErrVal write( HeaderSymbolWriteIf* pcWriteIf ) const;
  ErrVal read( HeaderSymbolReadIf* pcReadIf );

  bool check()              const;


  ErrVal initOffsetForRefFrame(UInt uiSize)
  {
    RNOK( m_piOffsetForRefFrame.uninit() );

    ROTRS( uiSize < 1, Err::m_nOK );

    RNOK( m_piOffsetForRefFrame.init( uiSize ) );

    return Err::m_nOK;
  }
  

  const FrameCroppingRect& getFrameCroppingRect() const { return m_cFrameCroppingRect; }
  FrameCroppingRect& getFrameCroppingRect() { return m_cFrameCroppingRect; }

  ErrVal setAccessUnitParameters( const AccessUnitParameters &rcPars, const UInt uiLog2MaxPicOrderCntLsb, const UInt uiLevelIdc, const UInt uiProfileIdc );
  ErrVal getAccessUnitParameters( AccessUnitParameters &rcPars, const UInt uiPlaneGrouping ) const;
protected:
  ErrVal xReadFrext( HeaderSymbolReadIf* pcReadIf );
  ErrVal xWriteFrext( HeaderSymbolWriteIf* pcWriteIf ) const;
  ErrVal xWritePicOrderCntInfo( HeaderSymbolWriteIf* pcWriteIf ) const;
  ErrVal xReadPicOrderCntInfo( HeaderSymbolReadIf* pcReadIf );

  bool xCheckProfile()      const;
  bool xCheckLevel()        const;
  bool xCheckIndependent()  const;

protected:

  Profile m_eProfileIdc;
  UInt m_uiLevelIdc;
  UInt m_uiSeqParameterSetId;
  UInt m_uiLog2MaxFrameNum;
  bool m_bFrameMbsOnlyFlag;
  bool m_bDeltaPicOrderAlwaysZeroFlag;
  bool m_bMbAdaptiveFrameFieldFlag;
  bool m_bDirect8x8InferenceFlag;

  UInt m_uiPicOrderCntType;
  UInt m_uiLog2MaxPicOrderCntLsb;
   int m_iOffsetForNonRefPic;
   int m_iOffsetForTopToBottomField;
  UInt m_uiNumRefFramesInPicOrderCntCycle;
  UInt m_uiNumRefFrames;
  UInt m_uiFrameWidthInMbs;
  UInt m_uiFrameHeightInMbs;
  UInt m_uiChromaFormatParameterX2;
  UInt  m_uiSPSRegId;
  VUI* m_pcVUI;

	ChromaFormatIdc m_eChromaFormatIdc;
  bool m_bConstrainedSet3Flag;
  bool m_bQPPrimeTransformBypassFlag;
  bool m_bResidueTransformFlag;
	UInt m_uiBitDepthLuma;
	UInt m_uiBitDepthChroma;
  FrameCroppingRect m_cFrameCroppingRect;
  DynBuf<int> m_piOffsetForRefFrame;
  bool m_bConstrainedSet0Flag;
  bool m_bConstrainedSet1Flag;
  bool m_bConstrainedSet2Flag;
  bool m_bFrameCroppingFlag;
  bool m_bGapsInFrameNumValueAllowedFlag;
  bool m_bVuiParametersPresentFlag;
  bool m_bInitDone;
};


#if defined( WIN32 )
# pragma warning( default: 4251 )
#endif

// h264 namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_SEQUENCEPARAMETERSET_H__66281283_5BFB_429A_B722_6DDE7A11D086__INCLUDED_)
