/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_SLICEHEADER_H__F31F1842_FFCD_42AD_A981_7BD2736A4431__INCLUDED_)
#define AFX_SLICEHEADER_H__F31F1842_FFCD_42AD_A981_7BD2736A4431__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VideoCommonLib/H264SliceHeaderBase.h"
#include "VideoCommonLib/PlanePackage.h"
#include "VideoCommonLib/ExtendedCabacTables.h"

// h264 namespace begin
VIDEO_NAMESPACE_BEGIN


#if defined( WIN32 )
# pragma warning( disable: 4275 )
# pragma warning( disable: 4251 )
#endif

extern const int g_aaiLumaToChromaQp[2][52];


//!\ingroup H264CommonLib
//!\SliceHeader
class VIDEOCOMMONLIB_DECL SliceHeader
: public SliceHeaderBase
{
public:
  SliceHeader( const SequenceParameterSet& rcSPS, const PictureParameterSet& rcPPS );
  virtual ~SliceHeader();

  //! setAccessUnitParameters copies and converts attributes of AccessUnitParameters 
  //! into attributes of SliceHeader
  ErrVal setAccessUnitParameters( const AccessUnitParameters& rcAUPars);
  //! getAccessUnitParameters copies and converts attributes of SliceHeader 
  //! into attributes of AccessUnitParameters
  ErrVal getAccessUnitParameters( AccessUnitParameters& rcAUPars) const;

  const PictureParameterSet& getPPS()           const { return m_rcPPS; }
  const SequenceParameterSet& getSPS()          const { return m_rcSPS; }

  bool check()          const;

  ErrVal writeSH( HeaderSymbolWriteIf* pcWriteIf ) const;
  ErrVal readSH( HeaderSymbolReadIf* pcReadIf, AccessUnitParameters& rcAUPars );

  const SliceHeaderPicture& getSliceHeaderPicture() const { return *this; }
  const SliceHeaderBase&    getSliceHeaderBase()    const { return *this; }

protected:
  ErrVal xReadPlanePackageParameters( HeaderSymbolReadIf* pcReadIf, PlanePackageParameters& rcPPP, const bool bInterPred ) const;
  ErrVal xWritePlanePackageParameters( HeaderSymbolWriteIf* pcWriteIf, const PlanePackageParameters& rcPlaPaPars, bool bInterPred) const;

protected:
  const PictureParameterSet&  m_rcPPS;
  const SequenceParameterSet& m_rcSPS;

  ExtendedCabacTables                 m_eExtendedCabacTables;
  UInt                                m_uiPlaneGrouping;
  UInt                                m_uiInterpolationMethod;
  UInt                                m_uiEntropySliceSizeX;
  UInt                                m_uiEntropySliceSizeY;
  bool                                m_bCtxStatesReuse;
  bool                                m_bEntropySliceFlag;
  bool                                m_bInLoopFilterFlag;
  std::vector<int>                    m_aiDeltaRefPicOrderCnt;
  std::vector<PlanePackageParameters> m_acPlanePackageParameters;
  bool                                m_bDummyFlag1;
  bool                                m_bDummyFlag2;
  bool                                m_bDummyFlag3;
  UInt                                m_uiVLCMaxDelayParameter;
};


#if defined( WIN32 )
# pragma warning( default: 4251 )
# pragma warning( default: 4275 )
#endif

typedef SliceHeader::DeblockingFilterParameter DFP;

// h264 namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_SLICEHEADER_H__F31F1842_FFCD_42AD_A981_7BD2736A4431__INCLUDED_)
