/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_SLICEHEADERBASE_H__2CC1FD0F_CACB_4799_84BE_FC5FC9B9C245__INCLUDED_)
#define AFX_SLICEHEADERBASE_H__2CC1FD0F_CACB_4799_84BE_FC5FC9B9C245__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VideoCommonLib/CommonClasses.h"
#include "VideoCommonLib/H264SequenceParameterSet.h"
#include "VideoCommonLib/H264PictureParameterSet.h"

// h264 namespace begin
VIDEO_NAMESPACE_BEGIN


#if defined( WIN32 )
# pragma warning( disable: 4275 )
# pragma warning( disable: 4251 )
#endif

//!\ingroup H264CommonLib
//!\SliceHeaderPicture
class VIDEOCOMMONLIB_DECL SliceHeaderPicture
{
protected:
	SliceHeaderPicture();
	virtual ~SliceHeaderPicture();
public:
  SliceHeaderPicture& operator = ( const SliceHeaderPicture& rcSHP );

  UInt getPicParameterSetId()                   const {return m_uiPicParameterSetId;}
  UInt getFrameNum()                            const {return m_uiFrameNum;}
  bool getFieldPicFlag()                        const {return m_bFieldPicFlag;}
  bool getBottomFieldFlag()                     const {return m_bBottomFieldFlag;}
  UInt getIdrPicId()                            const {return m_uiIdrPicId; }
  UInt getPicOrderCntLsb()                      const {return m_uiPicOrderCntLsb; }
   int getDeltaPicOrderCntBottom()              const {return m_iDeltaPicOrderCntBottom; }
   int getDeltaPicOrderCnt(UInt uiNumber)       const {return m_aiDeltaPicOrderCnt[uiNumber];}
  bool getSpForSwitchFlag()                     const {return m_bSpForSwitchFlag;}
  UInt getSliceGroupChangeCycle()               const {return m_uiSliceGroupChangeCycle;}
  bool getLongTermReferenceFlag()               const { return m_bLongTermReferenceFlag; }
  bool getAdaptiveRefPicBufferingModeFlag()     const { return m_bAdaptiveRefPicBufferingModeFlag; }
  NalUnitType getNalUnitType()                  const { return m_eNalUnitType; }
  bool isIdrNalUnit()                           const { return m_eNalUnitType == NAL_UNIT_CODED_SLICE_IDR || m_eNalUnitType == NAL_UNIT_UNSPECIFIED_30; }
  NalRefIdc getNalRefIdc()                      const { return m_eNalRefIdc; }
  bool isNalRefIdc()                            const { return m_eNalRefIdc != NAL_REF_IDC_PRIORITY_LOWEST; }
  PicType getPicType()                          const { return ( ! m_bFieldPicFlag ? FRAME : m_bBottomFieldFlag ? BOT_FIELD : TOP_FIELD ); }
  bool getOneSliceTypeIndication()              const {return m_bOneSliceTypeIndication;}

  void setFrameNum(UInt uiFrameNum)                                       { m_uiFrameNum = uiFrameNum; }
  void setFieldPicFlag(bool bFieldPicFlag)                                { m_bFieldPicFlag = bFieldPicFlag; }
  void setBottomFieldFlag(bool bBottomFieldFlag)                          { m_bBottomFieldFlag = bBottomFieldFlag; }
  void setIdrPicId(UInt uiIdrPicId)                                       { m_uiIdrPicId = uiIdrPicId; }
  void setPicOrderCntLsb (UInt uiPicOrderCntLsb)                          { m_uiPicOrderCntLsb = uiPicOrderCntLsb; }
  void setDeltaPicOrderCntBottom (int iDeltaPicOrderCntBottom)            { m_iDeltaPicOrderCntBottom = iDeltaPicOrderCntBottom; }
  void setDeltaPicOrderCnt( UInt uiNumber, int aiDeltaPicOrderCnt)        { m_aiDeltaPicOrderCnt[uiNumber] = aiDeltaPicOrderCnt; }
  void setSpForSwitchFlag(bool bSpForSwitchFlag)                          { m_bSpForSwitchFlag = bSpForSwitchFlag; }
  void setSliceGroupChangeCycle(UInt uiSliceGroupChangeCycle)             { m_uiSliceGroupChangeCycle = uiSliceGroupChangeCycle; }
  void setLongTermReferenceFlag (bool bLongTermReferenceFlag)             { m_bLongTermReferenceFlag = bLongTermReferenceFlag; }
  void setAdaptiveRefPicBufferingModeFlag (bool bAdaptiveRefPicBufferingModeFlag) { m_bAdaptiveRefPicBufferingModeFlag = bAdaptiveRefPicBufferingModeFlag;}
  void setNalUnitType( NalUnitType eNalUnitType )                         { m_eNalUnitType = eNalUnitType;}
  void setNalRefIdc( NalRefIdc eNalRefIdc)                                { m_eNalRefIdc   = eNalRefIdc; }
  void setOneSliceTypeIndication(bool bOneSliceTypeIndication)            { m_bOneSliceTypeIndication = bOneSliceTypeIndication;}

protected :
  void setPicParameterSetId( UInt uiId )                                  { m_uiPicParameterSetId = uiId; }
protected :
  UInt m_uiPicParameterSetId;
  UInt m_uiFrameNum;
  UInt m_uiIdrPicId;
  UInt m_uiPicOrderCntLsb;
   int m_iDeltaPicOrderCntBottom;
   int m_aiDeltaPicOrderCnt[2];
  UInt m_uiSliceGroupChangeCycle;
  NalRefIdc m_eNalRefIdc;
  NalUnitType m_eNalUnitType;
  bool m_bFieldPicFlag;
  bool m_bBottomFieldFlag;
  bool m_bSpForSwitchFlag;
  bool m_bLongTermReferenceFlag;
  bool m_bAdaptiveRefPicBufferingModeFlag;
  bool m_bOneSliceTypeIndication;
};


//!\ingroup H264CommonLib
//!\SliceHeaderBase
class VIDEOCOMMONLIB_DECL SliceHeaderBase : public SliceHeaderPicture
{
public:

  class VIDEOCOMMONLIB_DECL DeblockingFilterParameter
  {
  public:
    DeblockingFilterParameter( UInt uiDisableDeblockingFilterIdc = 0, int iSliceAlphaC0Offset = 0, int iSliceBetaOffset = 0) :
      m_uiDisableDeblockingFilterIdc( uiDisableDeblockingFilterIdc ),
      m_iSliceAlphaC0Offset( iSliceAlphaC0Offset ),
      m_iSliceBetaOffset( iSliceBetaOffset )
    {}

    DeblockingFilterParameter& operator = ( const DeblockingFilterParameter& rcDFP )
    {
      m_uiDisableDeblockingFilterIdc  = rcDFP.m_uiDisableDeblockingFilterIdc;
      m_iSliceAlphaC0Offset           = rcDFP.m_iSliceAlphaC0Offset;
      m_iSliceBetaOffset              = rcDFP.m_iSliceBetaOffset;
      return *this;
    }

    void reset() 
    {
      m_uiDisableDeblockingFilterIdc  = 0;
      m_iSliceAlphaC0Offset           = 0;
      m_iSliceBetaOffset              = 0;
    }

    ErrVal write( HeaderSymbolWriteIf* pcWriteIf ) const;
    ErrVal read( HeaderSymbolReadIf* pcReadIf );

    DeblockingFilterParameter* getCopy()  const { return new DeblockingFilterParameter(m_uiDisableDeblockingFilterIdc, m_iSliceAlphaC0Offset, m_iSliceBetaOffset );}
    void setDisableDeblockingFilterIdc(UInt uiDisableDeblockingFilterIdc)   { m_uiDisableDeblockingFilterIdc = uiDisableDeblockingFilterIdc; }
    void setSliceAlphaC0Offset( int iSliceAlphaC0Offset)                    { AOT_DBG( 1 & iSliceAlphaC0Offset);  m_iSliceAlphaC0Offset = iSliceAlphaC0Offset; }
    void setSliceBetaOffset( int iSliceBetaOffset)                          { AOT_DBG( 1 & iSliceBetaOffset);     m_iSliceBetaOffset = iSliceBetaOffset; }

    UInt getDisableDeblockingFilterIdc()          const { return m_uiDisableDeblockingFilterIdc;}
    int  getSliceAlphaC0Offset()                  const { return m_iSliceAlphaC0Offset;}
    int  getSliceBetaOffset()                     const { return m_iSliceBetaOffset;}


  private:
    UInt m_uiDisableDeblockingFilterIdc;
     int m_iSliceAlphaC0Offset;
     int m_iSliceBetaOffset;
  };

protected:
	SliceHeaderBase();
  virtual ~SliceHeaderBase() {};

public:

  SliceHeaderBase& operator = ( const SliceHeaderBase& rcSHB );

  UInt getFirstMbInSlice()                      const {return m_uiFirstMbInSlice;}
  SliceType getSliceType()                      const {return m_eSliceType;}
  UInt getRedundantPicCount()                   const {return m_uiRedundantPicCount;}
  bool getDirectSpatialMvPredFlag()             const {return m_bDirectSpatialMvPredFlag;}
  bool getNumRefIdxActiveOverrideFlag()         const {return m_bNumRefIdxActiveOverrideFlag;}
  UInt getNumRefIdxActive( ListIdx eLstIdx )    const {return m_auiNumRefIdxActive[eLstIdx];}

  UInt getCabacInitIdc()                        const {return m_uiCabacInitIdc; }
   int getSliceQpDelta()                        const {return m_iSliceQpDelta;}
   int getSliceQsDelta()                        const {return m_iSliceQsDelta;}

  bool getNoOutputOfPriorPicsFlag()             const { return m_bNoOutputOfPriorPicsFlag; }

  UInt getLumaLogWeightDenom()                  const {return m_uiLumaLogWeightDenom;}
  UInt getChromaLogWeightDenom()                const {return m_uiChromaLogWeightDenom;}

  UInt getSliceId()                             const { return m_uiSliceId; }

  UInt getSliceGroupId( UInt uiMbAddress )      const ;
  DeblockingFilterParameter&       getDeblockingFilterParameter()           { return m_cDeblockingFilterParameter; }
  const DeblockingFilterParameter& getDeblockingFilterParameter()     const { return m_cDeblockingFilterParameter; }


  void setFirstMbInSlice(UInt uiFirstMbInSlice)                           { m_uiFirstMbInSlice = uiFirstMbInSlice; }
  virtual void setSliceType(SliceType eSliceType)                         { m_eSliceType = eSliceType; }
  void setRedundantPicCount(UInt uiRedundantPicCount)                     { m_uiRedundantPicCount = uiRedundantPicCount; }
  void setDirectSpatialMvPredFlag(bool bDirectSpatialMvPredFlag)          { m_bDirectSpatialMvPredFlag = bDirectSpatialMvPredFlag; }
  void setNumRefIdxActiveOverrideFlag(bool bNumRefIdxActiveOverrideFlag)  { m_bNumRefIdxActiveOverrideFlag = bNumRefIdxActiveOverrideFlag; }
  void setNumRefIdxActive( ListIdx eLstIdx, UInt uiNumRefIdxActive)       { m_auiNumRefIdxActive[eLstIdx] = uiNumRefIdxActive; }

  void setCabacInitIdc (UInt uiCabacInitIdc)                              { m_uiCabacInitIdc = uiCabacInitIdc; }
  void setSliceQpDelta( int iSliceQpDelta)                                { m_iSliceQpDelta = iSliceQpDelta; }
  void setSliceQsDelta( int iSliceQsDelta)                                { m_iSliceQsDelta = iSliceQsDelta; }

  void setNoOutputOfPriorPicsFlag (bool bNoOutputOfPriorPicsFlag)         { m_bNoOutputOfPriorPicsFlag = bNoOutputOfPriorPicsFlag; }

  ErrVal setLumaLogWeightDenom( UInt uiLumaLogWeightDenom )               { ROF(uiLumaLogWeightDenom < 8);    m_uiLumaLogWeightDenom   = uiLumaLogWeightDenom;   return Err::m_nOK;}
  ErrVal setChromaLogWeightDenom( UInt uiChromaLogWeightDenom)            { ROF(uiChromaLogWeightDenom < 8);  m_uiChromaLogWeightDenom = uiChromaLogWeightDenom; return Err::m_nOK;}
  void setSliceId( UInt uiSliceId )                                       { m_uiSliceId = uiSliceId; }

protected:
  bool xCheckProfile()  const;
  bool xCheckLevel()    const;

protected:
  DeblockingFilterParameter m_cDeblockingFilterParameter;
  UInt m_uiFirstMbInSlice;
  SliceType m_eSliceType;
  UInt m_uiRedundantPicCount;
  bool m_bDirectSpatialMvPredFlag;
  bool m_bNumRefIdxActiveOverrideFlag;
  UInt m_auiNumRefIdxActive[2];
  UInt m_uiCabacInitIdc;
   int m_iSliceQpDelta;
   int m_iSliceQsDelta;
  UInt m_uiLumaLogWeightDenom;
  UInt m_uiChromaLogWeightDenom;
  bool m_bNoOutputOfPriorPicsFlag;
  UInt m_uiSliceId;
};

#if defined( WIN32 )
# pragma warning( default: 4251 )
# pragma warning( default: 4275 )
#endif

// h264 namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_SLICEHEADERBASE_H__2CC1FD0F_CACB_4799_84BE_FC5FC9B9C245__INCLUDED_)
