/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTERPLANEPACKAGE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTERPLANEPACKAGE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/PredictionCalculator.h"
#include "VideoCommonLib/InterCommon.h"
#include "VideoCommonLib/AccessUnit.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class PredInfoInterPlanePackage : virtual public PredInfo
{
public:
  virtual ~PredInfoInterPlanePackage() { QuadtreeGroupPointerDeleter::deleteAllPointers( m_cPredTree ); }

  const QuadtreeGroup<PredInfo*>& getPredTree()  const { return m_cPredTree; }

protected:
  PredInfoInterPlanePackage( const QTNavi &rcNav )
    : m_cQTGS     ( rcNav.getPrimBlockSize(), 1 )
    , m_cPredTree ( &m_cQTGS, m_cQTGS, 0 )
  {}
  PredInfoInterPlanePackage( const QTNavi &rcNav, const PlanePackageParameters& rcPlaPaPars, const PlanePackage& rcRefPlaPa )
    : m_cQTGS     ( rcNav.getPrimBlockSize(), 1 )
    , m_cPredTree ( &m_cQTGS, m_cQTGS, 0 )
  {
    TOF( Err::m_nOK == xCreatePredTree( rcNav, rcPlaPaPars, rcRefPlaPa ) );
  }
  PredInfoInterPlanePackage( const PredInfoInterPlanePackage& rcPredInfoInterPlanePackage ) 
    : m_cQTGS     ( rcPredInfoInterPlanePackage.m_cQTGS, TypeLimits<UInt>::m_iMax ) // simply copy the structure
    , m_cPredTree ( &m_cQTGS, m_cQTGS, 0 )
  {
    TOF( Err::m_nOK == xCreatePredTree( rcPredInfoInterPlanePackage.m_cPredTree ) );
  }

  const PredInfoInterPlanePackage& operator=( const PredInfoInterPlanePackage& rcPredInfoInterPlanePackage )
  {
    TOF( Err::m_nOK == xCreatePredTree( rcPredInfoInterPlanePackage.m_cPredTree ) );
    return *this; 
  }

  ErrVal xSetPredInfo   ( const QTNavi &rcNav, const PlanePackageParameters& rcPlaPaPars, const PlanePackage& rcRefPlaPa );
  bool   xIsIntra       () const;

private:
  ErrVal xCreatePredTree( const QuadtreeGroup<PredInfo*>& rcPredTreeSrc );
  ErrVal xCreatePredTree( const QTNavi &rcNav, const PlanePackageParameters& rcPlaPaPars, const PlanePackage& rcRefPlaPa );
  ErrVal xCopyQTNode    ( const QTNavi &rcNavDes, const QuadtreeGroup<PredInfo*>& rcPredTreeSrc, const QTNavi &rcNavSrc );
  ErrVal xDeriveQTNode  ( const QTNavi &rcNavPrd, const QTNavi &rcNavDes, const UInt uiLog2ResFactor, const PlanePackage& rcRefPlaPa );
  bool   xIsAllIntra    ( const QTNavi &rcNavPrd, const QTNavi &rcNavDes, const UInt uiLog2ResFactor, const PlanePackage& rcRefPlaPa );
  ErrVal xReplaceIntra  ( const QTNavi &rcNavDes );

private:
  QuadtreeGroupStructure    m_cQTGS;
  QuadtreeGroup<PredInfo*>  m_cPredTree;
};


#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif


class PredSyntaxInterPlanePackage : public PredSyntax, public PredInfoInterPlanePackage
{
public:
  PredSyntaxInterPlanePackage( const QTNavi &rcNav )
    : PredSyntax                ( false )
    , PredInfoInterPlanePackage ( rcNav  )
  {}
  PredSyntaxInterPlanePackage( const PredSyntaxInterPlanePackage& rcPredSyntaxInterPlanePackage )
    : PredSyntax                ( rcPredSyntaxInterPlanePackage )
    , PredInfoInterPlanePackage ( rcPredSyntaxInterPlanePackage )
  {}
  PredSyntaxInterPlanePackage( const QTNavi &rcNav, const PlanePackageParameters& rcPlaPaPars, const PlanePackage& rcRefPlaPa )
    : PredSyntax                ( true )
    , PredInfoInterPlanePackage ( rcNav, rcPlaPaPars, rcRefPlaPa )
  {}

  virtual ~PredSyntaxInterPlanePackage() {}

  const PredSyntaxInterPlanePackage& operator=( const PredSyntaxInterPlanePackage& rcPredSyntaxInterPlanePackage )
  {
    PredSyntax               ::operator=( rcPredSyntaxInterPlanePackage );
    PredInfoInterPlanePackage::operator=( rcPredSyntaxInterPlanePackage );
    return *this;
  }

  virtual PredInfo::PredictionMode  getPredictionMode           ()  const { AOT( xIsPredSyntaxOnly() ); return PredInfo::INTER_PLANE_PACKAGE; }
  virtual bool                      isIntraPredicted            ()  const { AOT( xIsPredSyntaxOnly() ); return xIsIntra(); }
  virtual bool                      oversizedPredictionSupported()  const { AOT( xIsPredSyntaxOnly() ); return false; } //???
  virtual PredInfo::PredictionMode  getCodingMode               ()  const { return PredInfo::INTER_PLANE_PACKAGE; }
  virtual bool                      isIntraCoded                ()  const { return false; }

  virtual const PredInfoInterPlanePackage*        getPredInfoInterPlanePackage      ()  const { return ( xIsPredInfoPresent   () ? this : 0 ); }
  virtual const PredSyntax*                       getPredSyntax                     ()  const { return ( xIsPredSyntaxPresent () ? this : 0 ); }
  virtual const PredSyntaxInterPlanePackage*      getPredSyntaxInterPlanePackage    ()  const { return this; }

protected:
  virtual ErrVal  xSetComplementData( const QTNavi&                 rcPos,
                                      const PlanePackageParameters& rcPlaPaPars,
                                      const AccessUnitParameters&   rcAUPars,
                                      const SliceData&              rcSliceData,
                                      const PlanePackage*           pcRefPlanePackage );
};


#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif


class InterPlanePackage : public virtual PredictionCalculator
{
public:
  InterPlanePackage( InterTranslational& rcInterTranslational )
    : m_rcInterTranslational( rcInterTranslational ) {}

  ErrVal calculatePrediction( const PredInfo &rcPredInfo,
                              const QTNavi &rcNav,
                              const PrimPixelPos& rcTopLeftPosOfPrediction,
                              const PicBlock<const Pel> *pcEncPic,
                              const CodingState &rcCodingState,
                              const SliceData &rcCurrSlice,
                              const PicBlock<DPel> &rcPredPicBlock,
                              bool &rbPredictionValid );
private:
  ErrVal xCalcPredForNode   ( const QuadtreeGroup<PredInfo*>&  rcPredTree,
                              const QTNavi&                    rcNavPredTree,
                              const PrimPixelPos&              rcTopLeftPos,
                              const CodingState&               rcCodingState,
                              const SliceData&                 rcCurrSlice,
                              const PicBlock<DPel>&            rcPredBlock,
                              bool&                            rbPredictionValid );
private:
  InterTranslational& m_rcInterTranslational;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTERPLANEPACKAGE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
