/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTERTRANSLATIONAL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTERTRANSLATIONAL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/PredictionCalculator.h"
#include "VideoCommonLib/InterCommon.h"
#include "VideoCommonLib/AccessUnit.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class PredInfoInterTranslational : virtual public PredInfo
{
public:
  class MvAndRefPoc
  {
  public:
    MvAndRefPoc()                                     : m_cRefPoc( 0 )                  {}
    MvAndRefPoc( MotionVector cMv, PicOrderCnt cPoc ) : m_cRefPoc( cPoc ), m_cMv( cMv ) {}
    bool operator==( const MvAndRefPoc &rcComp ) const
    {
      return m_cMv == rcComp.m_cMv && m_cRefPoc == rcComp.m_cRefPoc;
    }
    PicOrderCnt   m_cRefPoc;
    MotionVector  m_cMv;
  };

public:
  virtual ~PredInfoInterTranslational() {}

  UInt                            getNumMv          ()              const { return UInt( m_acMvAndRefPoc.size() ); }
  const MotionVector&             getMv             ( UInt uiIdx )  const { AOF( uiIdx < m_acMvAndRefPoc.size() ); return m_acMvAndRefPoc[uiIdx].m_cMv; }
  const PicOrderCnt&              getRefPoc         ( UInt uiIdx )  const { AOF( uiIdx < m_acMvAndRefPoc.size() ); return m_acMvAndRefPoc[uiIdx].m_cRefPoc; }
  const MvAndRefPoc&              getMvAndRefPoc    ( UInt uiIdx )  const { AOF( uiIdx < m_acMvAndRefPoc.size() ); return m_acMvAndRefPoc[uiIdx]; }
  const std::vector<MvAndRefPoc>& getMvAndRefPocVec ()              const { return m_acMvAndRefPoc; }

  void          setNumMv      ( UInt ui )     { m_acMvAndRefPoc.resize( ui ); }
  MotionVector& getMv         ( UInt uiIdx )  { AOF( uiIdx < m_acMvAndRefPoc.size() ); return m_acMvAndRefPoc[uiIdx].m_cMv; }
  PicOrderCnt&  getRefPoc     ( UInt uiIdx )  { AOF( uiIdx < m_acMvAndRefPoc.size() ); return m_acMvAndRefPoc[uiIdx].m_cRefPoc; }
  MvAndRefPoc&  getMvAndRefPoc( UInt uiIdx )  { AOF( uiIdx < m_acMvAndRefPoc.size() ); return m_acMvAndRefPoc[uiIdx]; }
  std::vector<MvAndRefPoc>& getMvAndRefPocVec ()          { return m_acMvAndRefPoc; }
  
  void reset    ()                            { m_acMvAndRefPoc.clear(); }
  void pushBack ( MvAndRefPoc cMvAndRefPoc )  { m_acMvAndRefPoc.push_back( cMvAndRefPoc ); }

  ErrVal setDirectModeData( const QTNavi&                 rcPos,
                            const PlanePackageParameters& rcPlaPaPars,
                            const AccessUnitParameters&   rcAUPars,
                            const SliceData&              rcSliceData );

  static bool compareMvAndRefPocVec( const std::vector<MvAndRefPoc>& acVecA,
                                     const std::vector<MvAndRefPoc>& acVecB );

  ErrVal xFindLeftAndTopMergeTargets( const QTNavi& rcPos,
                                      const QuadtreeGroup<PredInfo*> &rcTree,
                                      bool& rbLeftAvailable,
                                      bool& rbTopAvailable,
                                      const PredInfoInterTranslational*& rpcPredInfoITLeft,
                                      const PredInfoInterTranslational*& rpcPredInfoITTop );

  bool operator==( const PredInfoInterTranslational &rcComp ) const;

protected:
  PredInfoInterTranslational()  {}
  PredInfoInterTranslational( const PredInfoInterTranslational& rcPredInfoInterTranslational )
    : m_acMvAndRefPoc( rcPredInfoInterTranslational.m_acMvAndRefPoc )
  {}

  const PredInfoInterTranslational& operator=( const PredInfoInterTranslational& rcPredInfoInterTranslational )
  {
    m_acMvAndRefPoc = rcPredInfoInterTranslational.m_acMvAndRefPoc;
    return *this;
  }

private:
  std::vector<MvAndRefPoc> m_acMvAndRefPoc;
};

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif

class PredSyntaxInterTranslational : public PredSyntax, public PredInfoInterTranslational
{
public:
  class MvDiffAndRefIdx
  {
  public:
    MvDiffAndRefIdx()                                                        : m_uiRefIdx( 0 )       , m_uiMvPredIdx( 0 )                                 {}
    MvDiffAndRefIdx( MotionVector cMvDiff, UInt uiMvPredIdx, UInt uiRefIdx ) : m_uiRefIdx( uiRefIdx ), m_uiMvPredIdx( uiMvPredIdx ), m_cMvDiff( cMvDiff ) {}
    bool operator==( const MvDiffAndRefIdx &rcComp ) const
    {
      return m_cMvDiff == rcComp.m_cMvDiff && m_uiRefIdx == rcComp.m_uiRefIdx && m_uiMvPredIdx == rcComp.m_uiMvPredIdx;
    }
    UInt          m_uiRefIdx;
    UInt          m_uiMvPredIdx;
    MotionVector  m_cMvDiff;
  };

public:
  bool                getDirectModeFlag ()                const { return m_bDirectModeFlag; }
  UInt                getNumEntries     ()                const { return UInt( m_acMvDiffAndRefIdx.size() ); }
  const MotionVector& getMvDiff         ( UInt uiIdx )    const { AOF( uiIdx < m_acMvDiffAndRefIdx.size() ); return m_acMvDiffAndRefIdx[uiIdx].m_cMvDiff; }
  const UInt&         getRefIdx         ( UInt uiIdx )    const { AOF( uiIdx < m_acMvDiffAndRefIdx.size() ); return m_acMvDiffAndRefIdx[uiIdx].m_uiRefIdx; }
  const UInt&         getMvPredIdx      ( UInt uiIdx )    const { AOF( uiIdx < m_acMvDiffAndRefIdx.size() ); return m_acMvDiffAndRefIdx[uiIdx].m_uiMvPredIdx; }
  const bool&         getMergeFlag      ()                const { return m_bMergeFlag; }
  const bool&         getMergeLeftFlag  ()                const { return m_bMergeLeftFlag; }
  const PrimPixelPos& getMVAddress      ()                const { return m_iMVAddress; }

  void                setDirectModeFlag ( bool bDirect )        { m_bDirectModeFlag = bDirect; }
  void                setMergeFlag      ( bool bMerge )         { m_bMergeFlag = bMerge; }
  void                setMergeLeftFlag  ( bool bMergeLeft )     { m_bMergeLeftFlag = bMergeLeft; }
  void                setMVAddress      ( PrimPixelPos cMVAddress ) { m_iMVAddress = cMVAddress; }
  void                setNumEntries     ( UInt ui )             { m_acMvDiffAndRefIdx.resize( ui ); }
  MotionVector&       getMvDiff         ( UInt uiIdx )          { AOF( uiIdx < m_acMvDiffAndRefIdx.size() ); return m_acMvDiffAndRefIdx[uiIdx].m_cMvDiff; }
  UInt&               getRefIdx         ( UInt uiIdx )          { AOF( uiIdx < m_acMvDiffAndRefIdx.size() ); return m_acMvDiffAndRefIdx[uiIdx].m_uiRefIdx; }
  UInt&               getMvPredIdx      ( UInt uiIdx )          { AOF( uiIdx < m_acMvDiffAndRefIdx.size() ); return m_acMvDiffAndRefIdx[uiIdx].m_uiMvPredIdx; }
  bool&               getMergeFlag      ()                      { return m_bMergeFlag; }
  bool&               getMergeLeftFlag  ()                      { return m_bMergeLeftFlag; }

public:
  PredSyntaxInterTranslational( bool bEncoder )
    : PredSyntax        ( bEncoder )
    , m_bDirectModeFlag ( false )
    , m_bMergeFlag      ( false )
    , m_bMergeLeftFlag  ( false )
  {}
  PredSyntaxInterTranslational( const PredSyntaxInterTranslational& rcPredSyntaxInterTranslational )
    : PredSyntax              ( rcPredSyntaxInterTranslational )
    , PredInfoInterTranslational( rcPredSyntaxInterTranslational )
    , m_bDirectModeFlag       ( rcPredSyntaxInterTranslational.m_bDirectModeFlag )
    , m_acMvDiffAndRefIdx     ( rcPredSyntaxInterTranslational.m_acMvDiffAndRefIdx )
    , m_bMergeFlag            ( rcPredSyntaxInterTranslational.m_bMergeFlag )
    , m_bMergeLeftFlag        ( rcPredSyntaxInterTranslational.m_bMergeLeftFlag )
  {}
  PredSyntaxInterTranslational( const PredInfoInterTranslational& rcPredInfoInterTranslational )
    : PredSyntax              ( true )
    , PredInfoInterTranslational( rcPredInfoInterTranslational )
    , m_bDirectModeFlag       ( false )
    , m_bMergeFlag            ( false )
    , m_bMergeLeftFlag        ( false )
  {}

  virtual ~PredSyntaxInterTranslational() {}

  const PredSyntaxInterTranslational& operator=( const PredSyntaxInterTranslational& rcPredSyntaxInterTranslational )
  {
    PredSyntax              ::operator=( rcPredSyntaxInterTranslational );
    PredInfoInterTranslational::operator=( rcPredSyntaxInterTranslational );
    m_bDirectModeFlag                 =  rcPredSyntaxInterTranslational.m_bDirectModeFlag;
    m_acMvDiffAndRefIdx               =  rcPredSyntaxInterTranslational.m_acMvDiffAndRefIdx;
    m_bMergeFlag                      =  rcPredSyntaxInterTranslational.m_bMergeFlag;
    m_bMergeLeftFlag                  =  rcPredSyntaxInterTranslational.m_bMergeLeftFlag;
    return *this;
  }
  
  virtual PredInfo::PredictionMode  getPredictionMode           ()  const { AOT( xIsPredSyntaxOnly() ); return PredInfo::INTER_TRANSLATIONAL; }
  virtual bool                      isIntraPredicted            ()  const { AOT( xIsPredSyntaxOnly() ); return false; }
  virtual bool                      oversizedPredictionSupported()  const { AOT( xIsPredSyntaxOnly() ); return true; }
  virtual PredInfo::PredictionMode  getCodingMode               ()  const { return PredInfo::INTER_TRANSLATIONAL; }
  virtual bool                      isIntraCoded                ()  const { return false; }

  virtual const PredInfoInterTranslational*   getPredInfoInterTranslational   ()  const { return ( xIsPredInfoPresent   () ? this : 0 ); }
  virtual const PredSyntax*                 getPredSyntax                 ()  const { return ( xIsPredSyntaxPresent () ? this : 0 ); }
  virtual const PredSyntaxInterTranslational* getPredSyntaxInterTranslational ()  const { return this; }

protected:
  virtual ErrVal  xSetComplementData( const QTNavi&                 rcPos,
                                      const PlanePackageParameters& rcPlaPaPars,
                                      const AccessUnitParameters&   rcAUPars,
                                      const SliceData&              rcSliceData,
                                      const PlanePackage*           pcRefPlanePackage );

private:
  static UInt xGetMvCompCost( int iMvComp );

private:
  bool                          m_bDirectModeFlag;
  std::vector<MvDiffAndRefIdx>  m_acMvDiffAndRefIdx;
  bool                          m_bMergeFlag;
  bool                          m_bMergeLeftFlag;
  PrimPixelPos                  m_iMVAddress;
  std::vector<MotionVector>     m_acMvVector;
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif

class InterTranslational
  : public virtual PredictionCalculator
{
public:

  ErrVal calculatePrediction( const PredInfo &rcPredInfo,
                              const QTNavi &rcNav,
                              const PrimPixelPos& rcTopLeftPosOfPrediction,
                              const PicBlock<const Pel> *pcEncPic,
                              const CodingState &rcCodingState,
                              const SliceData &rcCurrSlice,
                              const PicBlock<DPel> &rcPredPicBlock,
                              bool &rbPredictionValid );

  static ErrVal getMvPred   ( PicOrderCnt cRefPoc,
                              const QTNavi &rcNav,
                              const QuadtreeGroup<PredInfo*> &rcTree,
                              MotionVector &rcMvPred,
                              std::vector<MotionVector> &rcMotionVectorVector); // NOTE: motion vector precision is specified by the input rcMvPred !!!!

  static ErrVal getMvPredXInd       ( const PicOrderCnt&              rcRefPoc,
                                      const QTNavi&                   rcNav,
                                      const QuadtreeGroup<PredInfo*>& rcTree,
                                      const int                       iMvPrecision,
                                      int&                            riMvPredX );
  static ErrVal getMvPredYInd       ( const PicOrderCnt&              rcRefPoc,
                                      const QTNavi&                   rcNav,
                                      const QuadtreeGroup<PredInfo*>& rcTree,
                                      const int                       iMvPrecision,
                                      int&                            riMvPredY );
  
  static ErrVal getMvPredXDep       ( const PicOrderCnt&              rcRefPoc,
                                      const QTNavi&                   rcNav,
                                      const QuadtreeGroup<PredInfo*>& rcTree,
                                      const int                       iMvPrecision,
                                      const int                       iMvY,
                                      int&                            riMvPredX );
  static ErrVal getMvPredYDep       ( const PicOrderCnt&              rcRefPoc,
                                      const QTNavi&                   rcNav,
                                      const QuadtreeGroup<PredInfo*>& rcTree,
                                      const int                       iMvPrecision,
                                      const int                       iMvX,
                                      int&                            riMvPredY );
  
  static ErrVal getYThresXPredLists ( const PicOrderCnt&              rcRefPoc,
                                      const QTNavi&                   rcNav,
                                      const QuadtreeGroup<PredInfo*>& rcTree,
                                      const int                       iMvAccuracy,
                                      int&                            riYPred,
                                      std::vector<int>&               racYThresList,
                                      std::vector<int>&               racXPredList );
  static ErrVal getXThresYPredLists ( const PicOrderCnt&              rcRefPoc,
                                      const QTNavi&                   rcNav,
                                      const QuadtreeGroup<PredInfo*>& rcTree,
                                      const int                       iMvAccuracy,
                                      int&                            riXPred,
                                      std::vector<int>&               racXThresList,
                                      std::vector<int>&               racYPredList );

  static ErrVal getMvPredLst( PicOrderCnt cRefPoc,
                              const QTNavi &rcNav,
                              const QuadtreeGroup<PredInfo*> &rcTree,
                              std::vector<MotionVector>& rcMvPredLst,
                              int iMvPrecision );

  static ErrVal getNumMvPred( UInt uiRefIdx,
                              const QTNavi &rcNav,
                              const QuadtreeGroup<PredInfo*> &rcTree,
                              UInt& ruiNumMvPred,
                              UInt* auiCtx );

  static ErrVal getMvD      ( const PredSyntaxInterTranslational& rcPredSyntax,
                              UInt uiRefIdx,
                              MotionVector& rcMvD ); // NOTE: motion vector precision is specified by the input rcMvD !!!!

protected:
  static ErrVal xInsertReferenceMvs( PicOrderCnt cRefPoc,
                                     const QTNavi &rcNav,
                                     const QTNavi &rcNeighbour,
                                     const QuadtreeGroup<PredInfo*> &rcTree,
                                     int iMvAccuracy,
                                     std::vector<MotionVector> &rcReferenceMvs,
                                     bool bFirstOnly = false );
  //----- independent motion vector component prediction -----
  static int    xGetMvPredXCompInd  ( const std::vector<MotionVector>&  rcMvPredCandLst );
  static int    xGetMvPredYCompInd  ( const std::vector<MotionVector>&  rcMvPredCandLst );

  //----- dependent motion vector component prediction -----
  static int    xGetMvPredXCompDep  ( const std::vector<int>&           rcYThresLst,
                                      const std::vector<int>&           rcXPredLst,
                                      const int                         iYComponent );
  static int    xGetMvPredYCompDep  ( const std::vector<int>&           rcXThresLst,
                                      const std::vector<int>&           rcYPredLst,
                                      const int                         iXComponent );
  
  //----- thresholds and predictors for dependent motion vector component prediction -----
  static ErrVal xSetYThresXPredLists( const std::vector<MotionVector>&  rcMvPredCandLst,
                                      std::vector<int>&                 rcYThresLst,
                                      std::vector<int>&                 rcXPredLst );
  static ErrVal xSetXThresYPredLists( const std::vector<MotionVector>&  rcMvPredCandLst,
                                      std::vector<int>&                 rcXThresLst,
                                      std::vector<int>&                 rcYPredLst );
  
  //----- get motion vector prediction candidates -----
  static ErrVal xGetMvPredCandLstStd( const PicOrderCnt&                rcRefPoc,
                                      const QTNavi&                     rcNav,
                                      const QuadtreeGroup<PredInfo*>&   rcTree,
                                      const int                         iMvPrecision,
                                      std::vector<MotionVector>&        rcMvPredCandLst );
  static ErrVal xGetMvPredCandLstExt( const PicOrderCnt&                rcRefPoc,
                                      const QTNavi&                     rcNav,
                                      const QuadtreeGroup<PredInfo*>&   rcTree,
                                      const int                         iMvPrecision,
                                      std::vector<MotionVector>&        rcMvPredCandLst );
  static ErrVal xInsertNeighbourMvs ( const PicOrderCnt&                rcRefPoc,
                                      const QTNavi&                     rcNav,
                                      const QTNavi&                     rcNeighbour,
                                      const QuadtreeGroup<PredInfo*>&   rcTree,
                                      const int                         iMvAccuracy,
                                      const bool                        bAvoidMultipleInsertion,
                                      std::vector<MotionVector>&        racNeighbourMvs );


  static ErrVal xIsRefIdxAvailable ( UInt uiRefIdx,
                                     const QTNavi &rcNav,
                                     const QTNavi &rcNeighbour,
                                     const QuadtreeGroup<PredInfo*> &rcTree,
                                     bool& rbAvailable );

private:
  ReuseBuffer m_cReuseBuffer;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTERTRANSLATIONAL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
