/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTRACOMMON_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTRACOMMON_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/QuadtreeGroup.h"
#include "VideoCommonLib/PicBlock.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class IntraCommon
{
protected:

  bool xHasLeft    ( const QTNavi &rcNav ) const { const QTNavi cCand = rcNav.getLeft();      return !cCand.liesOutside() && cCand.isCodedBefore( rcNav ); }
  bool xHasTop     ( const QTNavi &rcNav ) const { const QTNavi cCand = rcNav.getTop();       return !cCand.liesOutside() && cCand.isCodedBefore( rcNav ); }
  bool xHasTopLeft ( const QTNavi &rcNav ) const { const QTNavi cCand = rcNav.getTopLeft();   return !cCand.liesOutside() && cCand.isCodedBefore( rcNav ); }
  bool xHasTopRight( const QTNavi &rcNav ) const { const QTNavi cCand = rcNav.getTopRight();  return !cCand.liesOutside() && cCand.isCodedBefore( rcNav ); }

  template<class t_Pel>
  ErrVal xGetLeft( const QTNavi &rcNav, bool bSmooth, const PixelPos &rcPos, const BlockSize &rcSize, const Block<const Pel> &rcEncPic, AllocBlock<t_Pel> &rcDst ) const
  {
    const BlockSize cDstSize( 1, rcSize.getY() );
    RNOK( rcDst.allocateBlock( cDstSize ) );

    if( bSmooth )
    {
      const Pel* pSrc = &rcEncPic.getEl(rcPos + PixelPos( -1, 0 ) );
      const int iStride = rcEncPic.getStride();
      const int iTopLeftOffset = xHasTopLeft( rcNav ) ? -iStride : 0;

      if( cDstSize.getY() > 1 )
      {
        rcDst.getEl( 0 ) = t_Pel( ( ( (int)(pSrc[iTopLeftOffset] + 2 * pSrc[0] + pSrc[iStride]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
        int i = 1;
        for( ; i < cDstSize.getY()-1; i++ )
        {
          pSrc += iStride;
          rcDst.getEl( i ) = t_Pel( ( ( (int)(pSrc[-iStride] + 2 * pSrc[0] + pSrc[iStride]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
        }
        pSrc += iStride;
        rcDst.getEl( i ) = t_Pel( ( ( (int)(pSrc[-iStride] + 3 * pSrc[0]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
      }
      else
      {
        // th I think we can never get here 
        rcDst.getEl( 0 ) = t_Pel( ( ( (int)(pSrc[iTopLeftOffset] + 3 * pSrc[0]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
      }
    }
    else
    {
      const Block<const Pel> cSrc( rcPos + PixelPos( -1, 0 ), cDstSize, rcEncPic );
      RNOK( rcDst.assign( cSrc ) );
      rcDst <<= EXTRA_PRECISION_BITS;
    }
    return Err::m_nOK;
  }

  template<class t_Pel>
  ErrVal xGetTop( const QTNavi &rcNav, bool bSmooth, const PixelPos &rcPos, const BlockSize &rcSize, const Block<const Pel> &rcEncPic, AllocBlock<t_Pel> &rcDst ) const
  {
    const BlockSize cDstSize( rcSize.getX(), 1 );
    RNOK( rcDst.allocateBlock( cDstSize ) );
    
    if( bSmooth )
    {
      const int iTopLeftOffset = xHasTopLeft( rcNav ) ? -1 : 0;
      const int iTopRightOffset= cDstSize.getX() + (xHasTopRight( rcNav ) ? 0 : -1);

      const Pel* pSrc = &rcEncPic.getEl( rcPos + PixelPos( 0, -1 ) );
      if( cDstSize.getX() > 1 )
      {
        rcDst.getEl( 0 ) = t_Pel( ( ( (int)(pSrc[iTopLeftOffset] + 2 * pSrc[0] + pSrc[1]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
        int i = 1;
        for( ; i < cDstSize.getX()-1; i++ )
        {
          rcDst.getEl( i ) = t_Pel( ( ( (int)(pSrc[i-1] + 2 * pSrc[i] + pSrc[i+1]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
        }
        rcDst.getEl( i ) = t_Pel( ( ( (int)(pSrc[i-1] + 2 * pSrc[i] + pSrc[iTopRightOffset]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
      }
      else
      {
        // th I think we can never get here 
        rcDst.getEl( 0 ) = t_Pel( ( ( (int)(pSrc[iTopLeftOffset] + 2 * pSrc[0] + pSrc[iTopRightOffset]) << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
      }
    }
    else
    {
      const Block<const Pel> cSrc( rcPos + PixelPos( 0, -1 ), cDstSize, rcEncPic );
      RNOK( rcDst.assign( cSrc ) );
      rcDst <<= EXTRA_PRECISION_BITS;
    }
    return Err::m_nOK;
  }

  DPel xGetTopLeft( const QTNavi &rcNav, bool bSmooth, const PixelPos &rcPos, const BlockSize &rcSize, const Block<const Pel> &rcEncPic ) const
  {
    if( bSmooth )
    {
      int iTopLeft, iLeftVal, iTopVal;
      iTopLeft = iLeftVal = iTopVal = rcEncPic.getEl( rcPos + PixelPos( -1, -1 ) );
      if( xHasLeft( rcNav ) )
      {
        iLeftVal = rcEncPic.getEl( rcPos + PixelPos( -1, 0 ) );
      }
      if( xHasTop( rcNav ) )
      {
        iTopVal = rcEncPic.getEl( rcPos + PixelPos( 0, -1 ) );
      }
      int iTemp = iLeftVal + 2 * iTopLeft + iTopVal;
      return ( ( iTemp << EXTRA_PRECISION_BITS ) + 2 ) >> 2;
    }
    return (int)rcEncPic.getEl( rcPos + PixelPos( -1, -1 ) ) << EXTRA_PRECISION_BITS;
  }

  template<class t_Pel>
  ErrVal xGetTopRight( const QTNavi &rcNav, bool bSmooth, const PixelPos &rcPos, const BlockSize &rcSize, BlockSize cTopRightSize, const Block<const Pel> &rcEncPic, AllocBlock<t_Pel> &rcDst ) const
  {
    const BlockSize cDstSize( cTopRightSize.getX(), 1 );
    RNOK( rcDst.allocateBlock( cDstSize ) );
    const Block<const Pel> cSrc( rcPos + PixelPos( rcSize.getX(), -1 ), cDstSize, rcEncPic );
    if( bSmooth )
    {
      int iLastOfTop = rcEncPic.getEl( rcPos + PixelPos( rcSize.getX() - ( xHasTop( rcNav ) ? 1 : 0 ), -1 ) );
      int iTemp        = iLastOfTop + 2 * (int)cSrc.getEl( 0 ) + (int)cSrc.getEl( gMin( 1, cDstSize.getX() - 1 ) );
      rcDst.getEl( 0 ) = t_Pel( ( ( iTemp << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
      for( int i = 1; i < cDstSize.getX(); i++ )
      {
        int iTemp        = (int)cSrc.getEl( i - 1 ) + 2 * (int)cSrc.getEl( i ) + (int)cSrc.getEl( gMin( i + 1, cDstSize.getX() - 1 ) );
        rcDst.getEl( i ) = t_Pel( ( ( iTemp << EXTRA_PRECISION_BITS ) + 2 ) >> 2 );
      }
    }
    else
    {
      RNOK( rcDst.assign( cSrc ) );
      rcDst <<= EXTRA_PRECISION_BITS;
    }
    return Err::m_nOK;
  }

};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTRACOMMON_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
