/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTRADC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTRADC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/IntraCommon.h"
#include "VideoCommonLib/PredictionCalculator.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class PredInfoIntraDC : virtual public PredInfo
{
public:
  virtual ~PredInfoIntraDC() {}

  bool operator==( const PredInfoIntraDC& rcPredInfoIntraDC ) const { return true; }

protected:
  PredInfoIntraDC()                                           {}
  PredInfoIntraDC( const PredInfoIntraDC& rcPredInfoIntraDC ) {}

  const PredInfoIntraDC& operator=( const PredInfoIntraDC& rcPredInfoIntraDC ) { return *this; }
};

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif


class PredSyntaxIntraDC : public PredSyntax, public PredInfoIntraDC
{
public:
  PredSyntaxIntraDC( bool bEncoder )
    : PredSyntax      ( bEncoder )
    , PredInfoIntraDC ()
  {}
  PredSyntaxIntraDC( const PredSyntaxIntraDC& rcPredSyntaxIntraDC )
    : PredSyntax      ( rcPredSyntaxIntraDC )
    , PredInfoIntraDC ( rcPredSyntaxIntraDC )
  {}
  PredSyntaxIntraDC( const PredInfoIntraDC& rcPredInfoIntraDC )
    : PredSyntax      ( true )
    , PredInfoIntraDC ( rcPredInfoIntraDC )
  {}

  virtual ~PredSyntaxIntraDC() {}

  const PredSyntaxIntraDC& operator=( const PredSyntaxIntraDC& rcPredSyntaxIntraDC )
  {
    PredSyntax     ::operator=( rcPredSyntaxIntraDC );
    PredInfoIntraDC::operator=( rcPredSyntaxIntraDC );
    return *this;
  }

  virtual PredInfo::PredictionMode  getPredictionMode           ()  const { AOT( xIsPredSyntaxOnly() ); return PredInfo::INTRA_DC; }
  virtual bool                      isIntraPredicted            ()  const { AOT( xIsPredSyntaxOnly() ); return true; }
  virtual bool                      oversizedPredictionSupported()  const { AOT( xIsPredSyntaxOnly() ); return false; }
  virtual PredInfo::PredictionMode  getCodingMode               ()  const { return PredInfo::INTRA_DC; }
  virtual bool                      isIntraCoded                ()  const { return true; }

  virtual const PredInfoIntraDC*   getPredInfoIntraDC   ()  const { return ( xIsPredInfoPresent   () ? this : 0 ); }
  virtual const PredSyntax*        getPredSyntax        ()  const { return ( xIsPredSyntaxPresent () ? this : 0 ); }
  virtual const PredSyntaxIntraDC* getPredSyntaxIntraDC ()  const { return this; }

protected:
  virtual ErrVal  xSetComplementData( const QTNavi&                 rcPos,
                                      const PlanePackageParameters& rcPlaPaPars,
                                      const AccessUnitParameters&   rcAUPars,
                                      const SliceData&              rcSliceData,
                                      const PlanePackage*           pcRefPlanePackage );
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif

class IntraDC
  : public virtual PredictionCalculator
  , private IntraCommon
{
public:
  ErrVal calculatePrediction( const PredInfo &rcPredInfo,
                              const QTNavi &rcNav,
                              const PrimPixelPos& rcTopLeftPosOfPrediction,
                              const PicBlock<const Pel> *pcEncPic,
                              const CodingState &rcCodingState,
                              const SliceData &rcCurrSlice,
                              const PicBlock<DPel> &rcPredPicBlock,
                              bool &rbPredictionValid );

private:
  template<class t_Pel>
  void xCalcSumOverElements( const Block<const t_Pel> &rcBlock, int &riSum, int &riNumElements ) const;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTRADC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
