/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTRADIRECTIONAL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTRADIRECTIONAL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/IntraCommon.h"
#include "VideoCommonLib/PredictionCalculator.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class PredInfoIntraDirectional : virtual public PredInfo
{
public:
  typedef enum
  {
    INTRA_NORTH           = 0,
    INTRA_WEST            = 1,
    INTRA_NORTHEAST       = 3,
    INTRA_NORTHWEST       = 4,
    INTRA_NORTH_NORTHWEST = 5,
    INTRA_WEST_NORTHWEST  = 6,
    INTRA_NORTH_NORTHEAST = 7,
    INTRA_WEST_SOUTHWEST  = 8,
    INTRA_UNSPECIFIED     = 9
  } IntraDirections;

public:
  virtual ~PredInfoIntraDirectional() {}

  IntraDirections getIntraDirection       ()  const { return m_eIntraDirection; }
  bool            getSmoothedRefPixelFlag ()  const { return m_bSmoothRefPixel; }

  void  setIntraDirection       ( IntraDirections eIntraDirection ) { m_eIntraDirection = eIntraDirection;  }
  void  setSmoothedRefPixelFlag ( bool            bSmoothRefPixel ) { m_bSmoothRefPixel = bSmoothRefPixel;  }

  bool operator==( const PredInfoIntraDirectional& rcPredInfoIntraDirectional ) const
  { 
    ROFRS( m_eIntraDirection == rcPredInfoIntraDirectional.m_eIntraDirection, false );
    ROFRS( m_bSmoothRefPixel == rcPredInfoIntraDirectional.m_bSmoothRefPixel, false );
    return true;
  }

protected:
  PredInfoIntraDirectional( IntraDirections eIntraDirection, bool bSmoothRefPixel )
    : m_eIntraDirection( eIntraDirection )
    , m_bSmoothRefPixel( bSmoothRefPixel )
  {}
  PredInfoIntraDirectional( const PredInfoIntraDirectional& rcPredInfoIntraDirectional )
    : m_eIntraDirection( rcPredInfoIntraDirectional.m_eIntraDirection )
    , m_bSmoothRefPixel( rcPredInfoIntraDirectional.m_bSmoothRefPixel )
  {}

  const PredInfoIntraDirectional& operator=( const PredInfoIntraDirectional& rcPredInfoIntraDirectional )
  {
    m_eIntraDirection = rcPredInfoIntraDirectional.m_eIntraDirection;
    m_bSmoothRefPixel = rcPredInfoIntraDirectional.m_bSmoothRefPixel;
    return *this;
  }

private:
  IntraDirections m_eIntraDirection;
  bool            m_bSmoothRefPixel;
};

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif

class PredSyntaxIntraDirectional : public PredSyntax, public PredInfoIntraDirectional
{
public:
  PredSyntaxIntraDirectional( bool bEncoder, PredInfoIntraDirectional::IntraDirections eIntraDirection = PredInfoIntraDirectional::INTRA_UNSPECIFIED, bool bSmoothRefPixel = false )
    : PredSyntax              ( bEncoder )
    , PredInfoIntraDirectional( eIntraDirection, bSmoothRefPixel )
  {}
  PredSyntaxIntraDirectional( const PredSyntaxIntraDirectional& rcPredSyntaxIntraDirectional )
    : PredSyntax              ( rcPredSyntaxIntraDirectional )
    , PredInfoIntraDirectional( rcPredSyntaxIntraDirectional )
  {}
  PredSyntaxIntraDirectional( const PredInfoIntraDirectional& rcPredInfoIntraDirectional )
    : PredSyntax              ( true )
    , PredInfoIntraDirectional( rcPredInfoIntraDirectional )
  {}
  
  virtual ~PredSyntaxIntraDirectional() {}

  const PredSyntaxIntraDirectional& operator=( const PredSyntaxIntraDirectional& rcPredSyntaxIntraDirectional )
  {
    PredSyntax              ::operator=( rcPredSyntaxIntraDirectional );
    PredInfoIntraDirectional::operator=( rcPredSyntaxIntraDirectional );
    return *this;
  }

  virtual PredInfo::PredictionMode  getPredictionMode           ()  const { AOT( xIsPredSyntaxOnly() ); return PredInfo::INTRA_DIRECTIONAL; }
  virtual bool                      isIntraPredicted            ()  const { AOT( xIsPredSyntaxOnly() ); return true; }
  virtual bool                      oversizedPredictionSupported()  const { AOT( xIsPredSyntaxOnly() ); return false; }
  virtual PredInfo::PredictionMode  getCodingMode               ()  const { return PredInfo::INTRA_DIRECTIONAL; }
  virtual bool                      isIntraCoded                ()  const { return true; }

  virtual const PredInfoIntraDirectional*   getPredInfoIntraDirectional   ()  const { return ( xIsPredInfoPresent   () ? this : 0 ); }
  virtual const PredSyntax*                 getPredSyntax                 ()  const { return ( xIsPredSyntaxPresent () ? this : 0 ); }
  virtual const PredSyntaxIntraDirectional* getPredSyntaxIntraDirectional ()  const { return this; }

protected:
  virtual ErrVal  xSetComplementData( const QTNavi&                 rcPos,
                                      const PlanePackageParameters& rcPlaPaPars,
                                      const AccessUnitParameters&   rcAUPars,
                                      const SliceData&              rcSliceData,
                                      const PlanePackage*           pcRefPlanePackage );
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif

class IntraDirectional
  : public virtual PredictionCalculator
  , private IntraCommon
{
public:
  ErrVal calculatePrediction( const PredInfo &rcPredInfo,
                              const QTNavi &rcNav,
                              const PrimPixelPos& rcTopLeftPosOfPrediction,
                              const PicBlock<const Pel> *pcEncPic,
                              const CodingState &rcCodingState,
                              const SliceData &rcCurrSlice,
                              const PicBlock<DPel> &rcPredPicBlock,
                              bool &rbPredictionValid );
  
private:
  ErrVal xPredWestSouthwest ( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredWest          ( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredWestNorthwest ( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredNorthwest     ( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredNorthNorthwest( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredNorth         ( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredNorthNortheast( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
  ErrVal xPredNortheast     ( const QTNavi                &rcNav,
                              const PicBlock<const Pel>   *pcEncPic,
                              const PicBlock<DPel>        &rcPredPicBlock,
                              bool                        &rbPredictionValid,
                              bool                         bSmoothRefPixel ) const;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTRADIRECTIONAL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
