/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _LEAFNODEITERATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _LEAFNODEITERATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/QuadtreeGroup.h"
#include "VideoCommonLib/TraceCommon.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class LeafNodeIterator
  : protected TraceBase
{
public:
  ErrVal processTree();

protected:
  LeafNodeIterator( const QuadtreeGroupStructure &rcTree, TraceCommon *pcTrace = NULL )
    : m_rcTree( rcTree )
    , m_uiSliceId( 0 )
    , m_bUseSliceId( false )
  {
    TraceBase::setTraceCommonPtr( pcTrace );
  }

  LeafNodeIterator( bool bUseSliceId, UInt uiSliceId, const QuadtreeGroupStructure &rcTree, TraceCommon *pcTrace = NULL )
    : m_rcTree( rcTree )
    , m_uiSliceId( uiSliceId )
    , m_bUseSliceId( bUseSliceId )
  {
    TraceBase::setTraceCommonPtr( pcTrace );
  }

private:
  ErrVal xProcessNode( const QTNavi &rcNav );

  virtual ErrVal xProcessEachNode( const QTNavi &rcPos ) { return Err::m_nOK; }
  virtual ErrVal xProcessLeafNode( const QTNavi &rcPos ) = 0;

  const QuadtreeGroupStructure &m_rcTree;
  const UInt m_uiSliceId;
  const bool m_bUseSliceId;
};




class LeafNodeCollector
  : private LeafNodeIterator
{
public:
  static ErrVal collectAllLeafNodes( const QuadtreeGroupStructure &rcTree, std::vector<QTNavi> &racLeafNodes )
  {
    LeafNodeCollector cCollector( rcTree );
    cCollector.m_pacLeafNodes = &racLeafNodes;
    cCollector.processTree();
    return Err::m_nOK;
  }

private:
  LeafNodeCollector( const QuadtreeGroupStructure &rcTree )
    : LeafNodeIterator( rcTree )
    , m_pacLeafNodes( NULL ) {}

  ErrVal xProcessLeafNode( const QTNavi &rcPos ) { m_pacLeafNodes->push_back( rcPos ); return Err::m_nOK; }

  std::vector<QTNavi> *m_pacLeafNodes;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _LEAFNODEITERATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
