/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_LEVELLIMIT_H__0218A97F_0462_44A7_B8F7_63030F6E01D6__INCLUDED_)
#define AFX_LEVELLIMIT_H__0218A97F_0462_44A7_B8F7_63030F6E01D6__INCLUDED_


VIDEO_NAMESPACE_BEGIN

typedef double FloatVal;

//!\ingroup VideoCommonLib
//!\LevelLimit
class LevelLimit
{
public:
  typedef struct
  {
    bool bValid;
    UInt uiMaxMbPerSec;
    UInt uiMaxFrameSize;
    UInt uiMaxDPBSizeX2;
    UInt uiMaxBitRate;
    UInt uiMaxCPBSize;
    UInt uiMaxVMvRange;
    UInt uiMinComprRatio;
    UInt uiMaxMvsPer2Mb;
    //----------
    UInt uiMaxSubMbRectSize_BE;
    UInt uiSliceRate_M;
    bool bMinLumaBiPredSizeIs8x8_ME;
    bool bDirect8x8InferenceFlag_M;
    bool bFrameMbsOnlyFlag;
  } LevelLimitStruct;
  
  //----- get level limits -----
  static UInt getMaxDPBSizeInFrames( UInt uiProfileIdc, UInt uiLevelIdc, UInt uiMbInFrame, UInt uiBytePerMb = 384 )
  {
    if((uiProfileIdc>=HIGH_PROFILE)
         &&(uiLevelIdc==42))
    {
      uiLevelIdc++;
    }
    return gMin( 16, (int)(sm_aLevelLimit[uiLevelIdc].uiMaxDPBSizeX2 / (2*uiBytePerMb*uiMbInFrame)) );
  }

  static UInt getMaxFrameSizeInMb( UInt uiProfileIdc, UInt uiLevelIdc )
  {
    if((uiProfileIdc>=HIGH_PROFILE)
      &&(uiLevelIdc==42))
    {
      uiLevelIdc++;
    }
    return sm_aLevelLimit[uiLevelIdc].uiMaxFrameSize;
  }

  static bool getFrameMbsOnlyFlag( UInt uiProfileIdc, UInt uiLevelIdc)
  {
    return ( uiProfileIdc==BASELINE_PROFILE ? true : sm_aLevelLimit[uiLevelIdc].bFrameMbsOnlyFlag );
  }

  static bool getDirect8x8InferenceFlag( UInt uiProfileIdc, UInt uiLevelIdc )
  {
    return ( uiProfileIdc==BASELINE_PROFILE ? false : uiProfileIdc==EXTENDED_PROFILE ? true : sm_aLevelLimit[uiLevelIdc].bDirect8x8InferenceFlag_M );
  }

  static bool getMinBiPredSizeIs8x8( UInt uiProfileIdc, UInt uiLevelIdc )
  {
    return ( uiProfileIdc==BASELINE_PROFILE ? false : sm_aLevelLimit[uiLevelIdc].bMinLumaBiPredSizeIs8x8_ME );
  }

  static UInt getMaxMVsPer2Mb( UInt uiLevelIdc )
  {
    return sm_aLevelLimit[uiLevelIdc].uiMaxMvsPer2Mb;
  }

  static short getMaxIntMvVer( UInt uiLevelIdc, bool bField )
  {
    return sm_aLevelLimit[uiLevelIdc].uiMaxVMvRange / (bField?2:1);
  }

  static short getMaxIntMvHor()
  {
    return 8192;
  }

  static bool  getConstrainedBlockSearch( UInt uiProfileIdc, UInt uiLevelIdc )
  {
    return ( uiProfileIdc==MAIN_PROFILE ? false : sm_aLevelLimit[uiLevelIdc].uiMaxSubMbRectSize_BE != 0 );
  }

  static FloatVal getMaxFrameRate( UInt uiLevelIdc, UInt uiMbInFrame )
  {
    return gMin( (FloatVal)172.0, (FloatVal) sm_aLevelLimit[uiLevelIdc].uiMaxMbPerSec / uiMbInFrame );
  }

  static UInt getMaxMBPerSec( UInt uiProfileIdc, UInt uiLevelIdc )
  {
    if((uiProfileIdc>=HIGH_PROFILE)
         &&(uiLevelIdc==42))
    {
      uiLevelIdc++;
    }
    return sm_aLevelLimit[uiLevelIdc].uiMaxMbPerSec;
  }

  static UInt getMinCR( UInt uiLevelIdc )
  {
    return sm_aLevelLimit[uiLevelIdc].uiMinComprRatio;
  }

  static UInt getSliceRate( UInt uiProfileIdc, UInt uiLevelIdc )
  {
    return ( (uiProfileIdc==MAIN_PROFILE)||(uiProfileIdc>=HIGH_PROFILE) ? sm_aLevelLimit[uiLevelIdc].uiSliceRate_M : 0 );
  }

  static UInt getMaxBitRate( UInt uiLevelIdc )
  {
    return sm_aLevelLimit[uiLevelIdc].uiMaxBitRate;
  }

  static UInt getMaxCpbSize( UInt uiLevelIdc )
  {
    return sm_aLevelLimit[uiLevelIdc].uiMaxCPBSize;
  }
  

  static ErrVal getLevelLimit( const LevelLimitStruct*& rpcLevelLimitStruct, UInt uiLevelIdc );

  static UInt   getLevelIdc( UInt uiMbY, UInt uiMbX, UInt uiMvRangeSubPel, UInt uiNumRefPic, float fOutFreq, float fTargetRate, float fBufferDelay );

  private:
    static const LevelLimitStruct sm_aLevelLimit[52];
};

VIDEO_NAMESPACE_END

#endif // !defined(AFX_LEVELLIMIT_H__0218A97F_0462_44A7_B8F7_63030F6E01D6__INCLUDED_)
