/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _MOMS_H__
#define _MOMS_H__

#include "Splines.h"
#include "InterCommon.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_TSrc, class t_TDst>
class CubicMOMSFixed
  : public CubicSplineFixed<t_TSrc, t_TDst>
{
public:
  CubicMOMSFixed() : CubicSplineFixed<t_TSrc, t_TDst>( 924844032, short( 5.25 * (1<<2) ), short(11276*-1), 12791 ) {}
  virtual ~CubicMOMSFixed() {}
private:
  __inline int xGetSplineVal( int iX, int iY ) const { return int(sm_aai16Q15SplVal8PelLUT[iX][iY]); }
  const short* xGetSplineBuf()      const { return sm_aai16Q15SplVal8PelLUT[0]; }
  const int xGetSplineBufStride() const { return 32; }
  const CoeffT* xGetFilterCoeffs( const int iFilter ) const { return sm_aai16Q15SplVal4PelLUT[iFilter]; }

#if USE_SEPARABLE_SPLINE_UPSAMPLING
  // table look-up for quarter-pel kernel values
  __inline short xGetKernel( int iDx, int iXX ) const { return sm_aai16Q15SplVal4PelLUT[iDx][iXX]; }
#endif

  static const short sm_aai16Q15SplVal8PelLUT[32][32];
  static const CoeffT sm_aai16Q15SplVal4PelLUT[4][4];
};

// Q15 LUT 4x4 MOMS value for 16 different (dx,dxx)
template<class t_TSrc, class t_TDst>
const CoeffT CubicMOMSFixed<t_TSrc, t_TDst>::sm_aai16Q15SplVal4PelLUT[4][4]=
{
  { 6242,   20285,  6242,     0,  },
  { 2889,   19078,  10520,  280,  },
  { 1073,   15311,  15311,  1073, },
  { 280,    10520,  19078,  2889, },
};

// Q15 LUT 4x4 MOMS value for every 32 1/8 Pel (dx,dy) pairs
template<class t_TSrc, class t_TDst>
const short CubicMOMSFixed<t_TSrc, t_TDst>::sm_aai16Q15SplVal8PelLUT[32][32]=
{
  {     1189,    3864,    1189,       0,     827,    3828,    1566,      21,     550,    3634,    2004,      53,     347,    3318,    2466,     111,     204,    2916,    2916,     204,     111,    2466,    3318,     347,      53,    2004,    3634,     550,      21,    1566,    3828,     827, },
  {     3864,   12557,    3864,       0,    2687,   12441,    5089,      67,    1789,   11810,    6513,     174,    1127,   10783,    8015,     359,     664,    9478,    9478,     664,     359,    8015,   10783,    1127,     174,    6513,   11810,    1789,      67,    5089,   12441,    2687, },
  {     1189,    3864,    1189,       0,     827,    3828,    1566,      21,     550,    3634,    2004,      53,     347,    3318,    2466,     111,     204,    2916,    2916,     204,     111,    2466,    3318,     347,      53,    2004,    3634,     550,      21,    1566,    3828,     827, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
  {      827,    2687,     827,       0,     575,    2663,    1089,      14,     383,    2528,    1394,      37,     241,    2308,    1715,      77,     142,    2029,    2029,     142,      77,    1715,    2308,     241,      37,    1394,    2528,     383,      14,    1089,    2663,     575, },
  {     3828,   12441,    3828,       0,    2663,   12326,    5042,      66,    1772,   11701,    6452,     172,    1117,   10683,    7941,     356,     658,    9391,    9391,     658,     356,    7941,   10683,    1117,     172,    6452,   11701,    1772,      66,    5042,   12326,    2663, },
  {     1566,    5089,    1566,       0,    1089,    5042,    2063,      27,     725,    4786,    2639,      70,     457,    4370,    3248,     146,     269,    3841,    3841,     269,     146,    3248,    4370,     457,      70,    2639,    4786,     725,      27,    2063,    5042,    1089, },
  {       21,      67,      21,       0,      14,      66,      27,       0,      10,      63,      35,       1,       6,      58,      43,       2,       4,      51,      51,       4,       2,      43,      58,       6,       1,      35,      63,      10,       0,      27,      66,      14, },
  {      550,    1789,     550,       0,     383,    1772,     725,      10,     255,    1682,     928,      25,     161,    1536,    1142,      51,      95,    1350,    1350,      95,      51,    1142,    1536,     161,      25,     928,    1682,     255,      10,     725,    1772,     383, },
  {     3634,   11810,    3634,       0,    2528,   11701,    4786,      63,    1682,   11108,    6125,     163,    1060,   10141,    7538,     338,     625,    8914,    8914,     625,     338,    7538,   10141,    1060,     163,    6125,   11108,    1682,      63,    4786,   11701,    2528, },
  {     2004,    6513,    2004,       0,    1394,    6452,    2639,      35,     928,    6125,    3378,      90,     585,    5592,    4157,     186,     344,    4916,    4916,     344,     186,    4157,    5592,     585,      90,    3378,    6125,     928,      35,    2639,    6452,    1394, },
  {       53,     174,      53,       0,      37,     172,      70,       1,      25,     163,      90,       2,      16,     149,     111,       5,       9,     131,     131,       9,       5,     111,     149,      16,       2,      90,     163,      25,       1,      70,     172,      37, },
  {      347,    1127,     347,       0,     241,    1117,     457,       6,     161,    1060,     585,      16,     101,     968,     720,      32,      60,     851,     851,      60,      32,     720,     968,     101,      16,     585,    1060,     161,       6,     457,    1117,     241, },
  {     3318,   10783,    3318,       0,    2308,   10683,    4370,      58,    1536,   10141,    5592,     149,     968,    9259,    6883,     309,     570,    8139,    8139,     570,     309,    6883,    9259,     968,     149,    5592,   10141,    1536,      58,    4370,   10683,    2308, },
  {     2466,    8015,    2466,       0,    1715,    7941,    3248,      43,    1142,    7538,    4157,     111,     720,    6883,    5116,     229,     424,    6050,    6050,     424,     229,    5116,    6883,     720,     111,    4157,    7538,    1142,      43,    3248,    7941,    1715, },
  {      111,     359,     111,       0,      77,     356,     146,       2,      51,     338,     186,       5,      32,     309,     229,      10,      19,     271,     271,      19,      10,     229,     309,      32,       5,     186,     338,      51,       2,     146,     356,      77, },
  {      204,     664,     204,       0,     142,     658,     269,       4,      95,     625,     344,       9,      60,     570,     424,      19,      35,     501,     501,      35,      19,     424,     570,      60,       9,     344,     625,      95,       4,     269,     658,     142, },
  {     2916,    9478,    2916,       0,    2029,    9391,    3841,      51,    1350,    8914,    4916,     131,     851,    8139,    6050,     271,     501,    7154,    7154,     501,     271,    6050,    8139,     851,     131,    4916,    8914,    1350,      51,    3841,    9391,    2029, },
  {     2916,    9478,    2916,       0,    2029,    9391,    3841,      51,    1350,    8914,    4916,     131,     851,    8139,    6050,     271,     501,    7154,    7154,     501,     271,    6050,    8139,     851,     131,    4916,    8914,    1350,      51,    3841,    9391,    2029, },
  {      204,     664,     204,       0,     142,     658,     269,       4,      95,     625,     344,       9,      60,     570,     424,      19,      35,     501,     501,      35,      19,     424,     570,      60,       9,     344,     625,      95,       4,     269,     658,     142, },
  {      111,     359,     111,       0,      77,     356,     146,       2,      51,     338,     186,       5,      32,     309,     229,      10,      19,     271,     271,      19,      10,     229,     309,      32,       5,     186,     338,      51,       2,     146,     356,      77, },
  {     2466,    8015,    2466,       0,    1715,    7941,    3248,      43,    1142,    7538,    4157,     111,     720,    6883,    5116,     229,     424,    6050,    6050,     424,     229,    5116,    6883,     720,     111,    4157,    7538,    1142,      43,    3248,    7941,    1715, },
  {     3318,   10783,    3318,       0,    2308,   10683,    4370,      58,    1536,   10141,    5592,     149,     968,    9259,    6883,     309,     570,    8139,    8139,     570,     309,    6883,    9259,     968,     149,    5592,   10141,    1536,      58,    4370,   10683,    2308, },
  {      347,    1127,     347,       0,     241,    1117,     457,       6,     161,    1060,     585,      16,     101,     968,     720,      32,      60,     851,     851,      60,      32,     720,     968,     101,      16,     585,    1060,     161,       6,     457,    1117,     241, },
  {       53,     174,      53,       0,      37,     172,      70,       1,      25,     163,      90,       2,      16,     149,     111,       5,       9,     131,     131,       9,       5,     111,     149,      16,       2,      90,     163,      25,       1,      70,     172,      37, },
  {     2004,    6513,    2004,       0,    1394,    6452,    2639,      35,     928,    6125,    3378,      90,     585,    5592,    4157,     186,     344,    4916,    4916,     344,     186,    4157,    5592,     585,      90,    3378,    6125,     928,      35,    2639,    6452,    1394, },
  {     3634,   11810,    3634,       0,    2528,   11701,    4786,      63,    1682,   11108,    6125,     163,    1060,   10141,    7538,     338,     625,    8914,    8914,     625,     338,    7538,   10141,    1060,     163,    6125,   11108,    1682,      63,    4786,   11701,    2528, },
  {      550,    1789,     550,       0,     383,    1772,     725,      10,     255,    1682,     928,      25,     161,    1536,    1142,      51,      95,    1350,    1350,      95,      51,    1142,    1536,     161,      25,     928,    1682,     255,      10,     725,    1772,     383, },
  {       21,      67,      21,       0,      14,      66,      27,       0,      10,      63,      35,       1,       6,      58,      43,       2,       4,      51,      51,       4,       2,      43,      58,       6,       1,      35,      63,      10,       0,      27,      66,      14, },
  {     1566,    5089,    1566,       0,    1089,    5042,    2063,      27,     725,    4786,    2639,      70,     457,    4370,    3248,     146,     269,    3841,    3841,     269,     146,    3248,    4370,     457,      70,    2639,    4786,     725,      27,    2063,    5042,    1089, },
  {     3828,   12441,    3828,       0,    2663,   12326,    5042,      66,    1772,   11701,    6452,     172,    1117,   10683,    7941,     356,     658,    9391,    9391,     658,     356,    7941,   10683,    1117,     172,    6452,   11701,    1772,      66,    5042,   12326,    2663, },
  {      827,    2687,     827,       0,     575,    2663,    1089,      14,     383,    2528,    1394,      37,     241,    2308,    1715,      77,     142,    2029,    2029,     142,      77,    1715,    2308,     241,      37,    1394,    2528,     383,      14,    1089,    2663,     575, },
};

template<class t_TSrc, class t_TDst>
class CubicMOMS
  : public CubicSpline<t_TSrc, t_TDst>
{
public:
  CubicMOMS() : CubicSpline<t_TSrc, t_TDst>( 27.5625, -1.625+(sqrt(pow(13.0/4.0,2)-4.0))/2 ) {}
  virtual ~CubicMOMS() {}
private:
  void xSplineVal ( const Block<double> &rcSplVal, double dDx, double dDy) const;

  static const double sm_aadSplValQPelLUT[16][16];
};

template<class t_TSrc, class t_TDst>
const double CubicMOMS<t_TSrc, t_TDst>::sm_aadSplValQPelLUT[16][16]=
{
  {     3.628117913832207e-02,    1.179138321995466e-01,    3.628117913832207e-02,    0.000000000000000e+00,    1.679421768707488e-02,    1.108985260770976e-01,    6.115362811791390e-02,    1.629818594104363e-03,    6.235827664399120e-03,    8.900226757369624e-02,    8.900226757369624e-02,    6.235827664399120e-03,    1.629818594104363e-03,    6.115362811791390e-02,    1.108985260770976e-01,    1.679421768707488e-02, },
  {     1.179138321995466e-01,    3.832199546485261e-01,    1.179138321995466e-01,    0.000000000000000e+00,    5.458120748299330e-02,    3.604202097505669e-01,    1.987492913832200e-01,    5.296910430839173e-03,    2.026643990929712e-02,    2.892573696145125e-01,    2.892573696145125e-01,    2.026643990929712e-02,    5.296910430839173e-03,    1.987492913832200e-01,    3.604202097505669e-01,    5.458120748299330e-02, },
  {     3.628117913832207e-02,    1.179138321995466e-01,    3.628117913832207e-02,    0.000000000000000e+00,    1.679421768707488e-02,    1.108985260770976e-01,    6.115362811791390e-02,    1.629818594104363e-03,    6.235827664399120e-03,    8.900226757369624e-02,    8.900226757369624e-02,    6.235827664399120e-03,    1.629818594104363e-03,    6.115362811791390e-02,    1.108985260770976e-01,    1.679421768707488e-02, },
  {     0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00,    0.000000000000000e+00, },
  {     1.679421768707488e-02,    5.458120748299330e-02,    1.679421768707488e-02,    0.000000000000000e+00,    7.773885921556150e-03,    5.133388804740656e-02,    2.830744114051876e-02,    7.544277476615902e-04,    2.886506164966001e-03,    4.119831526360552e-02,    4.119831526360552e-02,    2.886506164966001e-03,    7.544277476615902e-04,    2.830744114051876e-02,    5.133388804740656e-02,    7.773885921556150e-03, },
  {     1.108985260770976e-01,    3.604202097505669e-01,    1.108985260770976e-01,    0.000000000000000e+00,    5.133388804740656e-02,    3.389769400598782e-01,    1.869246640713577e-01,    4.981769726119775e-03,    1.906068416950120e-02,    2.720479467828799e-01,    2.720479467828799e-01,    1.906068416950120e-02,    4.981769726119775e-03,    1.869246640713577e-01,    3.389769400598782e-01,    5.133388804740656e-02, },
  {     6.115362811791390e-02,    1.987492913832200e-01,    6.115362811791390e-02,    0.000000000000000e+00,    2.830744114051876e-02,    1.869246640713577e-01,    1.030773067690618e-01,    2.747135638109499e-03,    1.051077983276647e-02,    1.500174939767574e-01,    1.500174939767574e-01,    1.051077983276647e-02,    2.747135638109499e-03,    1.030773067690618e-01,    1.869246640713577e-01,    2.830744114051876e-02, },
  {     1.629818594104363e-03,    5.296910430839173e-03,    1.629818594104363e-03,    0.000000000000000e+00,    7.544277476615902e-04,    4.981769726119775e-03,    2.747135638109499e-03,    7.321450715703418e-05,    2.801250708616879e-04,    3.998148738662260e-03,    3.998148738662260e-03,    2.801250708616879e-04,    7.321450715703418e-05,    2.747135638109499e-03,    4.981769726119775e-03,    7.544277476615902e-04, },
  {     6.235827664399120e-03,    2.026643990929712e-02,    6.235827664399120e-03,    0.000000000000000e+00,    2.886506164966001e-03,    1.906068416950120e-02,    1.051077983276647e-02,    2.801250708616879e-04,    1.071782879818601e-03,    1.529726473922908e-02,    1.529726473922908e-02,    1.071782879818601e-03,    2.801250708616879e-04,    1.051077983276647e-02,    1.906068416950120e-02,    2.886506164966001e-03, },
  {     8.900226757369624e-02,    2.892573696145125e-01,    8.900226757369624e-02,    0.000000000000000e+00,    4.119831526360552e-02,    2.720479467828799e-01,    1.500174939767574e-01,    3.998148738662260e-03,    1.529726473922908e-02,    2.183336876417234e-01,    2.183336876417234e-01,    1.529726473922908e-02,    3.998148738662260e-03,    1.500174939767574e-01,    2.720479467828799e-01,    4.119831526360552e-02, },
  {     8.900226757369624e-02,    2.892573696145125e-01,    8.900226757369624e-02,    0.000000000000000e+00,    4.119831526360552e-02,    2.720479467828799e-01,    1.500174939767574e-01,    3.998148738662260e-03,    1.529726473922908e-02,    2.183336876417234e-01,    2.183336876417234e-01,    1.529726473922908e-02,    3.998148738662260e-03,    1.500174939767574e-01,    2.720479467828799e-01,    4.119831526360552e-02, },
  {     6.235827664399120e-03,    2.026643990929712e-02,    6.235827664399120e-03,    0.000000000000000e+00,    2.886506164966001e-03,    1.906068416950120e-02,    1.051077983276647e-02,    2.801250708616879e-04,    1.071782879818601e-03,    1.529726473922908e-02,    1.529726473922908e-02,    1.071782879818601e-03,    2.801250708616879e-04,    1.051077983276647e-02,    1.906068416950120e-02,    2.886506164966001e-03, },
  {     1.629818594104363e-03,    5.296910430839173e-03,    1.629818594104363e-03,    0.000000000000000e+00,    7.544277476615902e-04,    4.981769726119775e-03,    2.747135638109499e-03,    7.321450715703418e-05,    2.801250708616879e-04,    3.998148738662260e-03,    3.998148738662260e-03,    2.801250708616879e-04,    7.321450715703418e-05,    2.747135638109499e-03,    4.981769726119775e-03,    7.544277476615902e-04, },
  {     6.115362811791390e-02,    1.987492913832200e-01,    6.115362811791390e-02,    0.000000000000000e+00,    2.830744114051876e-02,    1.869246640713577e-01,    1.030773067690618e-01,    2.747135638109499e-03,    1.051077983276647e-02,    1.500174939767574e-01,    1.500174939767574e-01,    1.051077983276647e-02,    2.747135638109499e-03,    1.030773067690618e-01,    1.869246640713577e-01,    2.830744114051876e-02, },
  {     1.108985260770976e-01,    3.604202097505669e-01,    1.108985260770976e-01,    0.000000000000000e+00,    5.133388804740656e-02,    3.389769400598782e-01,    1.869246640713577e-01,    4.981769726119775e-03,    1.906068416950120e-02,    2.720479467828799e-01,    2.720479467828799e-01,    1.906068416950120e-02,    4.981769726119775e-03,    1.869246640713577e-01,    3.389769400598782e-01,    5.133388804740656e-02, },
  {     1.679421768707488e-02,    5.458120748299330e-02,    1.679421768707488e-02,    0.000000000000000e+00,    7.773885921556150e-03,    5.133388804740656e-02,    2.830744114051876e-02,    7.544277476615902e-04,    2.886506164966001e-03,    4.119831526360552e-02,    4.119831526360552e-02,    2.886506164966001e-03,    7.544277476615902e-04,    2.830744114051876e-02,    5.133388804740656e-02,    7.773885921556150e-03, },
};

// return 4x4 spline values for a given position (dx,dy)
template<class t_TSrc, class t_TDst>
void CubicMOMS<t_TSrc, t_TDst>::xSplineVal(const Block<double> &rcSplVal, double dDx, double dDy) const
{
  int iX, iY;

  int iDx = int(dDx*4);
  int iDy = int(dDy*4);

  for(iY=-1; iY<=2; iY++)
  {
    for(iX=-1; iX<=2; iX++)
    {
      rcSplVal.getEl(iX+1, iY+1) = sm_aadSplValQPelLUT[iDx*4+iX+1][iDy*4+iY+1];
    }
  }
}

template<class t_TSrc, class t_TDst>
class QuinticMOMSFixed
  : public QuinticSplineFixed<t_TSrc, t_TDst>
{
public:
  QuinticMOMSFixed() : QuinticSplineFixed<t_TSrc, t_TDst>( int(pow((63360.0/850.0),2)*pow(2.0,6)+0.5) // c0*c0
                                                         , 6                                          // Q c0
                                                         , short(-0.478368402923858*pow(2.0,15)+0.5)  // pole1 (Q15)
                                                         , short( 0.620320200316728*pow(2.0,15)+0.5)  // pole1/(pole1*pole1-1) (Q15)
                                                         , short(-0.070202438211750*pow(2.0,15)+0.5)  // pole2 (Q15)
                                                         , short( 0.070550136256718*pow(2.0,15)+0.5)  // pole2/(pole2*pole2-1) (Q15)
                                                         , 7                                          // Q RowCausal2
                                                         , 9                                          // Q RowAntiCausal2
                                                         , 10                                         // Q ColCausal2
                                                         , 13                                         // Q ColAntiCausal2
                                                         , 12                                         // SumShift
                                                         , 15 ) {}                                    // Q SplVal LUT
  virtual ~QuinticMOMSFixed() {}
private:
  __inline int xGetSplineVal( int iX, int iY ) const { return int(sm_aai16Q15SplVal8PelLUT[iX][iY]); }
  const short* xGetSplineBuf()        const { return sm_aai16Q15SplVal8PelLUT[0]; }
  const int xGetSplineBufStride()   const { return 48; }

  static const short sm_aai16Q15SplVal8PelLUT[48][48];
};

// Q15 LUT 4x4 MOMS value for every 32 1/8 Pel (dx,dy) pairs
template<class t_TSrc, class t_TDst>
const short QuinticMOMSFixed<t_TSrc, t_TDst>::sm_aai16Q15SplVal8PelLUT[48][48]=
{
  {        6,     100,     229,     100,       6,       0,       3,      79,     226,     121,      10,       0,       2,      61,     217,     144,      15,       0,       1,      46,     204,     166,      23,       0,       0,      33,     187,     187,      33,       0,       0,      23,     166,     204,      46,       1,       0,      15,     144,     217,      61,       2,       0,      10,     121,     226,      79,       3, },
  {      100,    1681,    3860,    1681,     100,       0,      57,    1337,    3813,    2050,     164,       0,      31,    1031,    3671,    2430,     258,       1,      16,     769,    3444,    2804,     386,       3,       7,     555,    3149,    3149,     555,       7,       3,     386,    2804,    3444,     769,      16,       1,     258,    2430,    3671,    1031,      31,       0,     164,    2050,    3813,    1337,      57, },
  {      229,    3860,    8865,    3860,     229,       0,     131,    3071,    8755,    4708,     377,       1,      71,    2367,    8430,    5581,     592,       2,      36,    1767,    7910,    6438,     886,       6,      16,    1275,    7231,    7231,    1275,      16,       6,     886,    6438,    7910,    1767,      36,       2,     592,    5581,    8430,    2367,      71,       1,     377,    4708,    8755,    3071,     131, },
  {      100,    1681,    3860,    1681,     100,       0,      57,    1337,    3813,    2050,     164,       0,      31,    1031,    3671,    2430,     258,       1,      16,     769,    3444,    2804,     386,       3,       7,     555,    3149,    3149,     555,       7,       3,     386,    2804,    3444,     769,      16,       1,     258,    2430,    3671,    1031,      31,       0,     164,    2050,    3813,    1337,      57, },
  {        6,     100,     229,     100,       6,       0,       3,      79,     226,     121,      10,       0,       2,      61,     217,     144,      15,       0,       1,      46,     204,     166,      23,       0,       0,      33,     187,     187,      33,       0,       0,      23,     166,     204,      46,       1,       0,      15,     144,     217,      61,       2,       0,      10,     121,     226,      79,       3, },
  {        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0, },
  {        3,      57,     131,      57,       3,       0,       2,      46,     130,      70,       6,       0,       1,      35,     125,      83,       9,       0,       1,      26,     117,      95,      13,       0,       0,      19,     107,     107,      19,       0,       0,      13,      95,     117,      26,       1,       0,       9,      83,     125,      35,       1,       0,       6,      70,     130,      46,       2, },
  {       79,    1337,    3071,    1337,      79,       0,      46,    1064,    3033,    1631,     131,       0,      25,     820,    2920,    1933,     205,       1,      12,     612,    2740,    2230,     307,       2,       6,     442,    2505,    2505,     442,       6,       2,     307,    2230,    2740,     612,      12,       1,     205,    1933,    2920,     820,      25,       0,     131,    1631,    3033,    1064,      46, },
  {      226,    3813,    8755,    3813,     226,       0,     130,    3033,    8647,    4650,     373,       1,      70,    2338,    8325,    5512,     584,       2,      35,    1745,    7812,    6358,     875,       6,      16,    1259,    7141,    7141,    1259,      16,       6,     875,    6358,    7812,    1745,      35,       2,     584,    5512,    8325,    2338,      70,       1,     373,    4650,    8647,    3033,     130, },
  {      121,    2050,    4708,    2050,     121,       0,      70,    1631,    4650,    2501,     200,       0,      38,    1257,    4477,    2964,     314,       1,      19,     938,    4201,    3419,     471,       3,       9,     677,    3840,    3840,     677,       9,       3,     471,    3419,    4201,     938,      19,       1,     314,    2964,    4477,    1257,      38,       0,     200,    2501,    4650,    1631,      70, },
  {       10,     164,     377,     164,      10,       0,       6,     131,     373,     200,      16,       0,       3,     101,     359,     238,      25,       0,       2,      75,     337,     274,      38,       0,       1,      54,     308,     308,      54,       1,       0,      38,     274,     337,      75,       2,       0,      25,     238,     359,     101,       3,       0,      16,     200,     373,     131,       6, },
  {        0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0, },
  {        2,      31,      71,      31,       2,       0,       1,      25,      70,      38,       3,       0,       1,      19,      68,      45,       5,       0,       0,      14,      64,      52,       7,       0,       0,      10,      58,      58,      10,       0,       0,       7,      52,      64,      14,       0,       0,       5,      45,      68,      19,       1,       0,       3,      38,      70,      25,       1, },
  {       61,    1031,    2367,    1031,      61,       0,      35,     820,    2338,    1257,     101,       0,      19,     632,    2251,    1491,     158,       1,      10,     472,    2112,    1719,     237,       2,       4,     340,    1931,    1931,     340,       4,       2,     237,    1719,    2112,     472,      10,       1,     158,    1491,    2251,     632,      19,       0,     101,    1257,    2338,     820,      35, },
  {      217,    3671,    8430,    3671,     217,       0,     125,    2920,    8325,    4477,     359,       1,      68,    2251,    8016,    5307,     563,       2,      34,    1680,    7522,    6122,     843,       6,      15,    1212,    6876,    6876,    1212,      15,       6,     843,    6122,    7522,    1680,      34,       2,     563,    5307,    8016,    2251,      68,       1,     359,    4477,    8325,    2920,     125, },
  {      144,    2430,    5581,    2430,     144,       0,      83,    1933,    5512,    2964,     238,       0,      45,    1491,    5307,    3514,     373,       1,      23,    1113,    4980,    4053,     558,       4,      10,     803,    4553,    4553,     803,      10,       4,     558,    4053,    4980,    1113,      23,       1,     373,    3514,    5307,    1491,      45,       0,     238,    2964,    5512,    1933,      83, },
  {       15,     258,     592,     258,      15,       0,       9,     205,     584,     314,      25,       0,       5,     158,     563,     373,      39,       0,       2,     118,     528,     430,      59,       0,       1,      85,     483,     483,      85,       1,       0,      59,     430,     528,     118,       2,       0,      39,     373,     563,     158,       5,       0,      25,     314,     584,     205,       9, },
  {        0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0, },
  {        1,      16,      36,      16,       1,       0,       1,      12,      35,      19,       2,       0,       0,      10,      34,      23,       2,       0,       0,       7,      32,      26,       4,       0,       0,       5,      29,      29,       5,       0,       0,       4,      26,      32,       7,       0,       0,       2,      23,      34,      10,       0,       0,       2,      19,      35,      12,       1, },
  {       46,     769,    1767,     769,      46,       0,      26,     612,    1745,     938,      75,       0,      14,     472,    1680,    1113,     118,       0,       7,     352,    1577,    1283,     177,       1,       3,     254,    1441,    1441,     254,       3,       1,     177,    1283,    1577,     352,       7,       0,     118,    1113,    1680,     472,      14,       0,      75,     938,    1745,     612,      26, },
  {      204,    3444,    7910,    3444,     204,       0,     117,    2740,    7812,    4201,     337,       1,      64,    2112,    7522,    4980,     528,       2,      32,    1577,    7058,    5745,     791,       6,      15,    1137,    6452,    6452,    1137,      15,       6,     791,    5745,    7058,    1577,      32,       2,     528,    4980,    7522,    2112,      64,       1,     337,    4201,    7812,    2740,     117, },
  {      166,    2804,    6438,    2804,     166,       0,      95,    2230,    6358,    3419,     274,       0,      52,    1719,    6122,    4053,     430,       2,      26,    1283,    5745,    4676,     644,       5,      12,     926,    5251,    5251,     926,      12,       5,     644,    4676,    5745,    1283,      26,       2,     430,    4053,    6122,    1719,      52,       0,     274,    3419,    6358,    2230,      95, },
  {       23,     386,     886,     386,      23,       0,      13,     307,     875,     471,      38,       0,       7,     237,     843,     558,      59,       0,       4,     177,     791,     644,      89,       1,       2,     127,     723,     723,     127,       2,       1,      89,     644,     791,     177,       4,       0,      59,     558,     843,     237,       7,       0,      38,     471,     875,     307,      13, },
  {        0,       3,       6,       3,       0,       0,       0,       2,       6,       3,       0,       0,       0,       2,       6,       4,       0,       0,       0,       1,       6,       5,       1,       0,       0,       1,       5,       5,       1,       0,       0,       1,       5,       6,       1,       0,       0,       0,       4,       6,       2,       0,       0,       0,       3,       6,       2,       0, },
  {        0,       7,      16,       7,       0,       0,       0,       6,      16,       9,       1,       0,       0,       4,      15,      10,       1,       0,       0,       3,      15,      12,       2,       0,       0,       2,      13,      13,       2,       0,       0,       2,      12,      15,       3,       0,       0,       1,      10,      15,       4,       0,       0,       1,       9,      16,       6,       0, },
  {       33,     555,    1275,     555,      33,       0,      19,     442,    1259,     677,      54,       0,      10,     340,    1212,     803,      85,       0,       5,     254,    1137,     926,     127,       1,       2,     183,    1040,    1040,     183,       2,       1,     127,     926,    1137,     254,       5,       0,      85,     803,    1212,     340,      10,       0,      54,     677,    1259,     442,      19, },
  {      187,    3149,    7231,    3149,     187,       0,     107,    2505,    7141,    3840,     308,       0,      58,    1931,    6876,    4553,     483,       2,      29,    1441,    6452,    5251,     723,       5,      13,    1040,    5898,    5898,    1040,      13,       5,     723,    5251,    6452,    1441,      29,       2,     483,    4553,    6876,    1931,      58,       0,     308,    3840,    7141,    2505,     107, },
  {      187,    3149,    7231,    3149,     187,       0,     107,    2505,    7141,    3840,     308,       0,      58,    1931,    6876,    4553,     483,       2,      29,    1441,    6452,    5251,     723,       5,      13,    1040,    5898,    5898,    1040,      13,       5,     723,    5251,    6452,    1441,      29,       2,     483,    4553,    6876,    1931,      58,       0,     308,    3840,    7141,    2505,     107, },
  {       33,     555,    1275,     555,      33,       0,      19,     442,    1259,     677,      54,       0,      10,     340,    1212,     803,      85,       0,       5,     254,    1137,     926,     127,       1,       2,     183,    1040,    1040,     183,       2,       1,     127,     926,    1137,     254,       5,       0,      85,     803,    1212,     340,      10,       0,      54,     677,    1259,     442,      19, },
  {        0,       7,      16,       7,       0,       0,       0,       6,      16,       9,       1,       0,       0,       4,      15,      10,       1,       0,       0,       3,      15,      12,       2,       0,       0,       2,      13,      13,       2,       0,       0,       2,      12,      15,       3,       0,       0,       1,      10,      15,       4,       0,       0,       1,       9,      16,       6,       0, },
  {        0,       3,       6,       3,       0,       0,       0,       2,       6,       3,       0,       0,       0,       2,       6,       4,       0,       0,       0,       1,       6,       5,       1,       0,       0,       1,       5,       5,       1,       0,       0,       1,       5,       6,       1,       0,       0,       0,       4,       6,       2,       0,       0,       0,       3,       6,       2,       0, },
  {       23,     386,     886,     386,      23,       0,      13,     307,     875,     471,      38,       0,       7,     237,     843,     558,      59,       0,       4,     177,     791,     644,      89,       1,       2,     127,     723,     723,     127,       2,       1,      89,     644,     791,     177,       4,       0,      59,     558,     843,     237,       7,       0,      38,     471,     875,     307,      13, },
  {      166,    2804,    6438,    2804,     166,       0,      95,    2230,    6358,    3419,     274,       0,      52,    1719,    6122,    4053,     430,       2,      26,    1283,    5745,    4676,     644,       5,      12,     926,    5251,    5251,     926,      12,       5,     644,    4676,    5745,    1283,      26,       2,     430,    4053,    6122,    1719,      52,       0,     274,    3419,    6358,    2230,      95, },
  {      204,    3444,    7910,    3444,     204,       0,     117,    2740,    7812,    4201,     337,       1,      64,    2112,    7522,    4980,     528,       2,      32,    1577,    7058,    5745,     791,       6,      15,    1137,    6452,    6452,    1137,      15,       6,     791,    5745,    7058,    1577,      32,       2,     528,    4980,    7522,    2112,      64,       1,     337,    4201,    7812,    2740,     117, },
  {       46,     769,    1767,     769,      46,       0,      26,     612,    1745,     938,      75,       0,      14,     472,    1680,    1113,     118,       0,       7,     352,    1577,    1283,     177,       1,       3,     254,    1441,    1441,     254,       3,       1,     177,    1283,    1577,     352,       7,       0,     118,    1113,    1680,     472,      14,       0,      75,     938,    1745,     612,      26, },
  {        1,      16,      36,      16,       1,       0,       1,      12,      35,      19,       2,       0,       0,      10,      34,      23,       2,       0,       0,       7,      32,      26,       4,       0,       0,       5,      29,      29,       5,       0,       0,       4,      26,      32,       7,       0,       0,       2,      23,      34,      10,       0,       0,       2,      19,      35,      12,       1, },
  {        0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       2,       2,       0,       0,       0,       0,       1,       2,       1,       0,       0,       0,       1,       2,       1,       0, },
  {       15,     258,     592,     258,      15,       0,       9,     205,     584,     314,      25,       0,       5,     158,     563,     373,      39,       0,       2,     118,     528,     430,      59,       0,       1,      85,     483,     483,      85,       1,       0,      59,     430,     528,     118,       2,       0,      39,     373,     563,     158,       5,       0,      25,     314,     584,     205,       9, },
  {      144,    2430,    5581,    2430,     144,       0,      83,    1933,    5512,    2964,     238,       0,      45,    1491,    5307,    3514,     373,       1,      23,    1113,    4980,    4053,     558,       4,      10,     803,    4553,    4553,     803,      10,       4,     558,    4053,    4980,    1113,      23,       1,     373,    3514,    5307,    1491,      45,       0,     238,    2964,    5512,    1933,      83, },
  {      217,    3671,    8430,    3671,     217,       0,     125,    2920,    8325,    4477,     359,       1,      68,    2251,    8016,    5307,     563,       2,      34,    1680,    7522,    6122,     843,       6,      15,    1212,    6876,    6876,    1212,      15,       6,     843,    6122,    7522,    1680,      34,       2,     563,    5307,    8016,    2251,      68,       1,     359,    4477,    8325,    2920,     125, },
  {       61,    1031,    2367,    1031,      61,       0,      35,     820,    2338,    1257,     101,       0,      19,     632,    2251,    1491,     158,       1,      10,     472,    2112,    1719,     237,       2,       4,     340,    1931,    1931,     340,       4,       2,     237,    1719,    2112,     472,      10,       1,     158,    1491,    2251,     632,      19,       0,     101,    1257,    2338,     820,      35, },
  {        2,      31,      71,      31,       2,       0,       1,      25,      70,      38,       3,       0,       1,      19,      68,      45,       5,       0,       0,      14,      64,      52,       7,       0,       0,      10,      58,      58,      10,       0,       0,       7,      52,      64,      14,       0,       0,       5,      45,      68,      19,       1,       0,       3,      38,      70,      25,       1, },
  {        0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0,       0,       0,       0,       1,       0,       0, },
  {       10,     164,     377,     164,      10,       0,       6,     131,     373,     200,      16,       0,       3,     101,     359,     238,      25,       0,       2,      75,     337,     274,      38,       0,       1,      54,     308,     308,      54,       1,       0,      38,     274,     337,      75,       2,       0,      25,     238,     359,     101,       3,       0,      16,     200,     373,     131,       6, },
  {      121,    2050,    4708,    2050,     121,       0,      70,    1631,    4650,    2501,     200,       0,      38,    1257,    4477,    2964,     314,       1,      19,     938,    4201,    3419,     471,       3,       9,     677,    3840,    3840,     677,       9,       3,     471,    3419,    4201,     938,      19,       1,     314,    2964,    4477,    1257,      38,       0,     200,    2501,    4650,    1631,      70, },
  {      226,    3813,    8755,    3813,     226,       0,     130,    3033,    8647,    4650,     373,       1,      70,    2338,    8325,    5512,     584,       2,      35,    1745,    7812,    6358,     875,       6,      16,    1259,    7141,    7141,    1259,      16,       6,     875,    6358,    7812,    1745,      35,       2,     584,    5512,    8325,    2338,      70,       1,     373,    4650,    8647,    3033,     130, },
  {       79,    1337,    3071,    1337,      79,       0,      46,    1064,    3033,    1631,     131,       0,      25,     820,    2920,    1933,     205,       1,      12,     612,    2740,    2230,     307,       2,       6,     442,    2505,    2505,     442,       6,       2,     307,    2230,    2740,     612,      12,       1,     205,    1933,    2920,     820,      25,       0,     131,    1631,    3033,    1064,      46, },
  {        3,      57,     131,      57,       3,       0,       2,      46,     130,      70,       6,       0,       1,      35,     125,      83,       9,       0,       1,      26,     117,      95,      13,       0,       0,      19,     107,     107,      19,       0,       0,      13,      95,     117,      26,       1,       0,       9,      83,     125,      35,       1,       0,       6,      70,     130,      46,       2, },
};


template<class t_TSrc, class t_TDst>
class QuinticMOMS
  : public QuinticSpline<t_TSrc, t_TDst>
{
public:
  QuinticMOMS() : QuinticSpline<t_TSrc, t_TDst>( pow(63360.0/850.0,2), -0.478368402923858, -0.070202438211750 ) {}
  virtual ~QuinticMOMS() {}
private:
  double xSplineVal ( int iX, int iY ) const;

  static const double sm_aadSplValQPelLUT[24][24];
};

template<class t_TSrc, class t_TDst>
const double QuinticMOMS<t_TSrc, t_TDst>::sm_aadSplValQPelLUT[24][24]=
{
    {     1.799730655672635e-04,    3.038580545830567e-03,    6.977873351570774e-03,    3.038580545830567e-03,    1.799730655672635e-04,    2.117330183144427e-07,    5.606624158399822e-05,    1.863478835827468e-03,    6.635229778026604e-03,    4.393175613781326e-03,    4.657563987087749e-04,    1.644239220355039e-06,    1.280984760801015e-05,    1.003296907282983e-03,    5.691595265310535e-03,    5.691595265310535e-03,    1.003296907282983e-03,    1.280984760801015e-05,    1.644239220355039e-06,    4.657563987087749e-04,    4.393175613781326e-03,    6.635229778026604e-03,    1.863478835827468e-03,    5.606624158399822e-05, },
    {     3.038580545830567e-03,    5.130196401555008e-02,    1.178111299628227e-01,    5.130196401555008e-02,    3.038580545830567e-03,    3.574800642153862e-06,    9.465960387906066e-04,    3.146209973289635e-02,    1.120260971111371e-01,    7.417230968632975e-02,    7.863611856317768e-03,    2.776056123684333e-05,    2.162754388500784e-04,    1.693919284284602e-02,    9.609421606173797e-02,    9.609421606173797e-02,    1.693919284284602e-02,    2.162754388500784e-04,    2.776056123684333e-05,    7.863611856317768e-03,    7.417230968632975e-02,    1.120260971111371e-01,    3.146209973289635e-02,    9.465960387906066e-04, },
    {     6.977873351570774e-03,    1.178111299628227e-01,    2.705444637345679e-01,    1.178111299628227e-01,    6.977873351570774e-03,    8.209262766554433e-06,    2.173787126638090e-03,    7.225036295709929e-02,    2.572595677231297e-01,    1.703311712091620e-01,    1.805819750099703e-02,    6.375005617179353e-05,    4.966603973760146e-04,    3.889959161931806e-02,    2.206731924277497e-01,    2.206731924277497e-01,    3.889959161931806e-02,    4.966603973760146e-04,    6.375005617179353e-05,    1.805819750099703e-02,    1.703311712091620e-01,    2.572595677231297e-01,    7.225036295709929e-02,    2.173787126638090e-03, },
    {     3.038580545830567e-03,    5.130196401555008e-02,    1.178111299628227e-01,    5.130196401555008e-02,    3.038580545830567e-03,    3.574800642153862e-06,    9.465960387906066e-04,    3.146209973289635e-02,    1.120260971111371e-01,    7.417230968632975e-02,    7.863611856317768e-03,    2.776056123684333e-05,    2.162754388500784e-04,    1.693919284284602e-02,    9.609421606173797e-02,    9.609421606173797e-02,    1.693919284284602e-02,    2.162754388500784e-04,    2.776056123684333e-05,    7.863611856317768e-03,    7.417230968632975e-02,    1.120260971111371e-01,    3.146209973289635e-02,    9.465960387906066e-04, },
    {     1.799730655672635e-04,    3.038580545830567e-03,    6.977873351570774e-03,    3.038580545830567e-03,    1.799730655672635e-04,    2.117330183144427e-07,    5.606624158399822e-05,    1.863478835827468e-03,    6.635229778026604e-03,    4.393175613781326e-03,    4.657563987087749e-04,    1.644239220355039e-06,    1.280984760801015e-05,    1.003296907282983e-03,    5.691595265310535e-03,    5.691595265310535e-03,    1.003296907282983e-03,    1.280984760801015e-05,    1.644239220355039e-06,    4.657563987087749e-04,    4.393175613781326e-03,    6.635229778026604e-03,    1.863478835827468e-03,    5.606624158399822e-05, },
    {     2.117330183144427e-07,    3.574800642153862e-06,    8.209262766554433e-06,    3.574800642153862e-06,    2.117330183144427e-07,    2.490976686052444e-10,    6.596028421647318e-08,    2.192328042150118e-06,    7.806152680031854e-06,    5.168441898566633e-06,    5.479487043633034e-07,    1.934399082770771e-09,    1.507040895060124e-08,    1.180349302685947e-06,    6.695994429777574e-06,    6.695994429777574e-06,    1.180349302685947e-06,    1.507040895060124e-08,    1.934399082770771e-09,    5.479487043633034e-07,    5.168441898566633e-06,    7.806152680031854e-06,    2.192328042150118e-06,    6.596028421647318e-08, },
    {     5.606624158399822e-05,    9.465960387906066e-04,    2.173787126638090e-03,    9.465960387906066e-04,    5.606624158399822e-05,    6.596028421647318e-08,    1.746607713463892e-05,    5.805216145363855e-04,    2.067044835445900e-03,    1.368587263359903e-03,    1.450951045757431e-04,    5.122228321207130e-07,    3.990597195092379e-06,    3.125528067597572e-04,    1.773078400023016e-03,    1.773078400023016e-03,    3.125528067597572e-04,    3.990597195092379e-06,    5.122228321207130e-07,    1.450951045757431e-04,    1.368587263359903e-03,    2.067044835445900e-03,    5.805216145363855e-04,    1.746607713463892e-05, },
    {     1.863478835827468e-03,    3.146209973289635e-02,    7.225036295709929e-02,    3.146209973289635e-02,    1.863478835827468e-03,    2.192328042150118e-06,    5.805216145363855e-04,    1.929485037459150e-02,    6.870255959263855e-02,    4.548786093380831e-02,    4.822539355593967e-03,    1.702479745239392e-05,    1.326358465499409e-04,    1.038834642772830e-02,    5.893197010103732e-02,    5.893197010103732e-02,    1.038834642772830e-02,    1.326358465499409e-04,    1.702479745239392e-05,    4.822539355593967e-03,    4.548786093380831e-02,    6.870255959263855e-02,    1.929485037459150e-02,    5.805216145363855e-04, },
    {     6.635229778026604e-03,    1.120260971111371e-01,    2.572595677231297e-01,    1.120260971111371e-01,    6.635229778026604e-03,    7.806152680031854e-06,    2.067044835445900e-03,    6.870255959263855e-02,    2.446270172063970e-01,    1.619671785929972e-01,    1.717146238676410e-02,    6.061965440612839e-05,    4.722722371414244e-04,    3.698945447433082e-02,    2.098371901919363e-01,    2.098371901919363e-01,    3.698945447433082e-02,    4.722722371414244e-04,    6.061965440612839e-05,    1.717146238676410e-02,    1.619671785929972e-01,    2.446270172063970e-01,    6.870255959263855e-02,    2.067044835445900e-03, },
    {     4.393175613781326e-03,    7.417230968632975e-02,    1.703311712091620e-01,    7.417230968632975e-02,    4.393175613781326e-03,    5.168441898566633e-06,    1.368587263359903e-03,    4.548786093380831e-02,    1.619671785929972e-01,    1.072382243014565e-01,    1.136919930946706e-02,    4.013618161872602e-05,    3.126907348629484e-04,    2.449066193635791e-02,    1.389328866753697e-01,    1.389328866753697e-01,    2.449066193635791e-02,    3.126907348629484e-04,    4.013618161872602e-05,    1.136919930946706e-02,    1.072382243014565e-01,    1.619671785929972e-01,    4.548786093380831e-02,    1.368587263359903e-03, },
    {     4.657563987087749e-04,    7.863611856317768e-03,    1.805819750099703e-02,    7.863611856317768e-03,    4.657563987087749e-04,    5.479487043633034e-07,    1.450951045757431e-04,    4.822539355593967e-03,    1.717146238676410e-02,    1.136919930946706e-02,    1.205341600724647e-03,    4.255164157339250e-06,    3.315089661394457e-05,    2.596454935625505e-03,    1.472940912198996e-02,    1.472940912198996e-02,    2.596454935625505e-03,    3.315089661394457e-05,    4.255164157339250e-06,    1.205341600724647e-03,    1.136919930946706e-02,    1.717146238676410e-02,    4.822539355593967e-03,    1.450951045757431e-04, },
    {     1.644239220355039e-06,    2.776056123684333e-05,    6.375005617179353e-05,    2.776056123684333e-05,    1.644239220355039e-06,    1.934399082770771e-09,    5.122228321207130e-07,    1.702479745239392e-05,    6.061965440612839e-05,    4.013618161872602e-05,    4.255164157339250e-06,    1.502181787720521e-08,    1.170311445075070e-07,    9.166150053709269e-06,    5.199858174396109e-05,    5.199858174396109e-05,    9.166150053709269e-06,    1.170311445075070e-07,    1.502181787720521e-08,    4.255164157339250e-06,    4.013618161872602e-05,    6.061965440612839e-05,    1.702479745239392e-05,    5.122228321207130e-07, },
    {     1.280984760801015e-05,    2.162754388500784e-04,    4.966603973760146e-04,    2.162754388500784e-04,    1.280984760801015e-05,    1.507040895060124e-08,    3.990597195092379e-06,    1.326358465499409e-04,    4.722722371414244e-04,    3.126907348629484e-04,    3.315089661394457e-05,    1.170311445075070e-07,    9.117597415104046e-07,    7.141113281242375e-05,    4.051076630011120e-04,    4.051076630011120e-04,    7.141113281242375e-05,    9.117597415104046e-07,    1.170311445075070e-07,    3.315089661394457e-05,    3.126907348629484e-04,    4.722722371414244e-04,    1.326358465499409e-04,    3.990597195092379e-06, },
    {     1.003296907282983e-03,    1.693919284284602e-02,    3.889959161931806e-02,    1.693919284284602e-02,    1.003296907282983e-03,    1.180349302685947e-06,    3.125528067597572e-04,    1.038834642772830e-02,    3.698945447433082e-02,    2.449066193635791e-02,    2.596454935625505e-03,    9.166150053709269e-06,    7.141113281242375e-05,    5.593085170777340e-03,    3.172896960550094e-02,    3.172896960550094e-02,    5.593085170777340e-03,    7.141113281242375e-05,    9.166150053709269e-06,    2.596454935625505e-03,    2.449066193635791e-02,    3.698945447433082e-02,    1.038834642772830e-02,    3.125528067597572e-04, },
    {     5.691595265310535e-03,    9.609421606173797e-02,    2.206731924277497e-01,    9.609421606173797e-02,    5.691595265310535e-03,    6.695994429777574e-06,    1.773078400023016e-03,    5.893197010103732e-02,    2.098371901919363e-01,    1.389328866753697e-01,    1.472940912198996e-02,    5.199858174396109e-05,    4.051076630011120e-04,    3.172896960550094e-02,    1.799950262668510e-01,    1.799950262668510e-01,    3.172896960550094e-02,    4.051076630011120e-04,    5.199858174396109e-05,    1.472940912198996e-02,    1.389328866753697e-01,    2.098371901919363e-01,    5.893197010103732e-02,    1.773078400023016e-03, },
    {     5.691595265310535e-03,    9.609421606173797e-02,    2.206731924277497e-01,    9.609421606173797e-02,    5.691595265310535e-03,    6.695994429777574e-06,    1.773078400023016e-03,    5.893197010103732e-02,    2.098371901919363e-01,    1.389328866753697e-01,    1.472940912198996e-02,    5.199858174396109e-05,    4.051076630011120e-04,    3.172896960550094e-02,    1.799950262668510e-01,    1.799950262668510e-01,    3.172896960550094e-02,    4.051076630011120e-04,    5.199858174396109e-05,    1.472940912198996e-02,    1.389328866753697e-01,    2.098371901919363e-01,    5.893197010103732e-02,    1.773078400023016e-03, },
    {     1.003296907282983e-03,    1.693919284284602e-02,    3.889959161931806e-02,    1.693919284284602e-02,    1.003296907282983e-03,    1.180349302685947e-06,    3.125528067597572e-04,    1.038834642772830e-02,    3.698945447433082e-02,    2.449066193635791e-02,    2.596454935625505e-03,    9.166150053709269e-06,    7.141113281242375e-05,    5.593085170777340e-03,    3.172896960550094e-02,    3.172896960550094e-02,    5.593085170777340e-03,    7.141113281242375e-05,    9.166150053709269e-06,    2.596454935625505e-03,    2.449066193635791e-02,    3.698945447433082e-02,    1.038834642772830e-02,    3.125528067597572e-04, },
    {     1.280984760801015e-05,    2.162754388500784e-04,    4.966603973760146e-04,    2.162754388500784e-04,    1.280984760801015e-05,    1.507040895060124e-08,    3.990597195092379e-06,    1.326358465499409e-04,    4.722722371414244e-04,    3.126907348629484e-04,    3.315089661394457e-05,    1.170311445075070e-07,    9.117597415104046e-07,    7.141113281242375e-05,    4.051076630011120e-04,    4.051076630011120e-04,    7.141113281242375e-05,    9.117597415104046e-07,    1.170311445075070e-07,    3.315089661394457e-05,    3.126907348629484e-04,    4.722722371414244e-04,    1.326358465499409e-04,    3.990597195092379e-06, },
    {     1.644239220355039e-06,    2.776056123684333e-05,    6.375005617179353e-05,    2.776056123684333e-05,    1.644239220355039e-06,    1.934399082770771e-09,    5.122228321207130e-07,    1.702479745239392e-05,    6.061965440612839e-05,    4.013618161872602e-05,    4.255164157339250e-06,    1.502181787720521e-08,    1.170311445075070e-07,    9.166150053709269e-06,    5.199858174396109e-05,    5.199858174396109e-05,    9.166150053709269e-06,    1.170311445075070e-07,    1.502181787720521e-08,    4.255164157339250e-06,    4.013618161872602e-05,    6.061965440612839e-05,    1.702479745239392e-05,    5.122228321207130e-07, },
    {     4.657563987087749e-04,    7.863611856317768e-03,    1.805819750099703e-02,    7.863611856317768e-03,    4.657563987087749e-04,    5.479487043633034e-07,    1.450951045757431e-04,    4.822539355593967e-03,    1.717146238676410e-02,    1.136919930946706e-02,    1.205341600724647e-03,    4.255164157339250e-06,    3.315089661394457e-05,    2.596454935625505e-03,    1.472940912198996e-02,    1.472940912198996e-02,    2.596454935625505e-03,    3.315089661394457e-05,    4.255164157339250e-06,    1.205341600724647e-03,    1.136919930946706e-02,    1.717146238676410e-02,    4.822539355593967e-03,    1.450951045757431e-04, },
    {     4.393175613781326e-03,    7.417230968632975e-02,    1.703311712091620e-01,    7.417230968632975e-02,    4.393175613781326e-03,    5.168441898566633e-06,    1.368587263359903e-03,    4.548786093380831e-02,    1.619671785929972e-01,    1.072382243014565e-01,    1.136919930946706e-02,    4.013618161872602e-05,    3.126907348629484e-04,    2.449066193635791e-02,    1.389328866753697e-01,    1.389328866753697e-01,    2.449066193635791e-02,    3.126907348629484e-04,    4.013618161872602e-05,    1.136919930946706e-02,    1.072382243014565e-01,    1.619671785929972e-01,    4.548786093380831e-02,    1.368587263359903e-03, },
    {     6.635229778026604e-03,    1.120260971111371e-01,    2.572595677231297e-01,    1.120260971111371e-01,    6.635229778026604e-03,    7.806152680031854e-06,    2.067044835445900e-03,    6.870255959263855e-02,    2.446270172063970e-01,    1.619671785929972e-01,    1.717146238676410e-02,    6.061965440612839e-05,    4.722722371414244e-04,    3.698945447433082e-02,    2.098371901919363e-01,    2.098371901919363e-01,    3.698945447433082e-02,    4.722722371414244e-04,    6.061965440612839e-05,    1.717146238676410e-02,    1.619671785929972e-01,    2.446270172063970e-01,    6.870255959263855e-02,    2.067044835445900e-03, },
    {     1.863478835827468e-03,    3.146209973289635e-02,    7.225036295709929e-02,    3.146209973289635e-02,    1.863478835827468e-03,    2.192328042150118e-06,    5.805216145363855e-04,    1.929485037459150e-02,    6.870255959263855e-02,    4.548786093380831e-02,    4.822539355593967e-03,    1.702479745239392e-05,    1.326358465499409e-04,    1.038834642772830e-02,    5.893197010103732e-02,    5.893197010103732e-02,    1.038834642772830e-02,    1.326358465499409e-04,    1.702479745239392e-05,    4.822539355593967e-03,    4.548786093380831e-02,    6.870255959263855e-02,    1.929485037459150e-02,    5.805216145363855e-04, },
    {     5.606624158399822e-05,    9.465960387906066e-04,    2.173787126638090e-03,    9.465960387906066e-04,    5.606624158399822e-05,    6.596028421647318e-08,    1.746607713463892e-05,    5.805216145363855e-04,    2.067044835445900e-03,    1.368587263359903e-03,    1.450951045757431e-04,    5.122228321207130e-07,    3.990597195092379e-06,    3.125528067597572e-04,    1.773078400023016e-03,    1.773078400023016e-03,    3.125528067597572e-04,    3.990597195092379e-06,    5.122228321207130e-07,    1.450951045757431e-04,    1.368587263359903e-03,    2.067044835445900e-03,    5.805216145363855e-04,    1.746607713463892e-05, },
};

// return 6x6 spline values for a given position (dx,dy)
template<class t_TSrc, class t_TDst>
double QuinticMOMS<t_TSrc, t_TDst>::xSplineVal ( int iX, int iY) const
{
  return sm_aadSplValQPelLUT[iX][iY];
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _MOMS_H__
