/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef __MODEPICTUREGENERATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define __MODEPICTUREGENERATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "LeafNodeIterator.h"
#include "VideoCommonLib/SliceData.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/ResTrafoPow2.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class ModePictureGeneratorParams
{
public:
  ModePictureGeneratorParams()
    : m_uiAlpha( 128 )
    , m_bResMode( false )
    , m_bBlockBorders( false )
    , m_uiBorderColor( 0 ) { }

  ModePictureGeneratorParams( const UInt uiAlpha, const bool bResMode, const bool bBlockBorders, const UInt uiBorderColor )
    : m_uiAlpha( uiAlpha )
    , m_bResMode( bResMode )
    , m_bBlockBorders( bBlockBorders )
    , m_uiBorderColor( uiBorderColor ) { }

  const UInt getAlpha() const { return m_uiAlpha; }
  const bool getResMode() const { return m_bResMode; }
  const bool getBlockBorders() const { return m_bBlockBorders; }
  const UInt getBorderColor() const { return m_uiBorderColor; }

private:
  const UInt m_uiAlpha;
  const bool m_bResMode;
  const bool m_bBlockBorders;
  const UInt m_uiBorderColor;
};

class ModePictureGenerator
{
  friend class SliceDataPlotter;
  friend class TrafoPow2Plotter;

public:
  ModePictureGenerator( const ModePictureGeneratorParams &rcMPGParams, const SliceData& rcSliceData )
    : m_rcMPGParams( rcMPGParams )
    , m_rcSliceData( rcSliceData )
  {
    for ( UInt uiGreyScaleIndex = 0; uiGreyScaleIndex < 128; uiGreyScaleIndex++ ) // 0 to 127: grey scale
    {
      xAddColor( uiGreyScaleIndex, uiGreyScaleIndex << 1, uiGreyScaleIndex << 1, uiGreyScaleIndex << 1 );
    }

    xAddColor( 128, 153, 204, 255 ); // INTRA_NORTH           = 0
    xAddColor( 129, 153, 153, 255 ); // INTRA_WEST            = 1
    xAddColor( 130, 204, 153, 255 ); // INTRA_DC              = 2
    xAddColor( 131, 255, 153, 255 ); // INTRA_NORTHEAST       = 3
    xAddColor( 132, 153, 255, 255 ); // INTRA_NORTHWEST       = 4
    xAddColor( 133,  92, 173, 255 ); // INTRA_NORTH_NORTHWEST = 5
    xAddColor( 134,  31, 143, 255 ); // INTRA_WEST_NORTHWEST  = 6
    xAddColor( 135, 255, 143,  31 ); // INTRA_NORTH_NORTHEAST = 7
    xAddColor( 136, 255, 173,  92 ); // INTRA_WEST_SOUTHWEST  = 8
    xAddColor( 137, 204, 255, 153 );
    xAddColor( 138, 255, 255, 153 );

    xAddColor( 140 + PredInfo::INTRA_DC,          255, 100,   0 ); // orange
    xAddColor( 140 + PredInfo::INTRA_DIRECTIONAL, 240, 255,   0 ); // yellow
    xAddColor( 140 + PredInfo::INTRA_PLANE      , 255, 230, 140 ); // khaki
    xAddColor( 140 + PredInfo::INTER_TRANSLATIONAL,   0, 220, 255 ); // cyan
    //xAddColor( 140 + PredInfo::INTER_RESIDUAL   ,   0, 100, 255 ); // blue
  }

  ErrVal generatePicture( const Block<const Pel> &rcBackgroundPicture, const PicBlock<Pel> &rcModePicture, const Subsampling &rcPrimPlaneSubsampling );

private:
  class TrafoPow2Plotter : public LeafNodeIterator
  {
  public:
    TrafoPow2Plotter( ModePictureGenerator &rcPicGen, const ModePictureGeneratorParams &rcMPGParams, const ResInfoTrafoPow2 &rcResInfo, const PrimPixelPos& rcOrigin )
      : LeafNodeIterator( rcResInfo.getTree().getTreeStructure() )
      , m_rcPicGen( rcPicGen )
      , m_rcMPGParams( rcMPGParams )
      , m_rcResInfo( rcResInfo )
      , m_cOrigin( rcOrigin ) {}

  private:
    ErrVal xProcessLeafNode( const QTNavi &rcPos );

    ModePictureGenerator             &m_rcPicGen;
    const ModePictureGeneratorParams &m_rcMPGParams;
    const ResInfoTrafoPow2           &m_rcResInfo;
    const PrimPixelPos                m_cOrigin;
  };

  class SliceDataPlotter : public LeafNodeIterator
  {
  public:
    SliceDataPlotter( ModePictureGenerator &rcPicGen, const ModePictureGeneratorParams &rcMPGParams, const SliceData &rcSliceData )
      : LeafNodeIterator( rcSliceData.getQuadtreeGroupStructure() )
      , m_rcPicGen( rcPicGen )
      , m_rcMPGParams( rcMPGParams )
      , m_rcSliceData( rcSliceData ) {}
  private:
    ErrVal xProcessLeafNode( const QTNavi &rcPos );

    ModePictureGenerator             &m_rcPicGen;
    const ModePictureGeneratorParams &m_rcMPGParams;
    const SliceData                  &m_rcSliceData;
  };

  void xPlotBlockArea( const PixelPos& cPixelPos, const BlockSize& cBlockSize, UInt uiColor );
  void xPlotBlockBorder( const PixelPos& cPixelPos, const BlockSize& cBlockSize, UInt uiColor );
  void xPlotBlockBorders( const PixelPos& cPixelPos, const BlockSize& cBlockSize, UInt uiColor );
  void xDrawLine( const PrimPixelPos& rcEndA, const PrimPixelPos& rcEndB, UInt uiColor );

  void xAddColor( UInt uiKey, double dR, double dG, double dB )
  {
    double dY  =       0.299000*dR +  0.587000*dG + 0.114000*dB;
    double dCb = 128 - 0.168736*dR  - 0.331264*dG + 0.500000*dB;
    double dCr = 128 + 0.500000*dR  - 0.418688*dG - 0.081312*dB;
    std::vector<Pel> cYCbCrVec;
    cYCbCrVec.push_back( Pel( dY  ) );
    cYCbCrVec.push_back( Pel( dCb ) );
    cYCbCrVec.push_back( Pel( dCr ) );
    m_aacPaletteMap.insert( std::make_pair( uiKey, cYCbCrVec) );
  }

  const ModePictureGeneratorParams   &m_rcMPGParams;
  const SliceData                    &m_rcSliceData;
  const Block<Pel>                   *m_pcColorMap;
  std::map< UInt, std::vector<Pel> >  m_aacPaletteMap;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _MODEPICTUREGENERATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
