/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_PARAMETERIO_H__6CF9086D_DB71_40C0_91BD_6F205082CB8C__INCLUDED_)
#define AFX_PARAMETERIO_H__6CF9086D_DB71_40C0_91BD_6F205082CB8C__INCLUDED_

#include "VideoCodecToolsLib/ToolLibInitOptionParser.h"
#include "VideoCodecToolsLib/ToolLibParser.h"

class ParameterIoIf
{
protected:
  ParameterIoIf(){}
  virtual ~ParameterIoIf(){}
public:
  virtual ErrVal value (const std::string& rcOption, std::string& rVal, bool bWarnIfNotAvail = false) = 0;
  virtual ErrVal value( const std::string& rcOption, bool&   rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, char&   rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, short&  rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, int&  rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, Int64&  rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, UChar&  rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, UShort& rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, UInt& rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, UInt64& rVal, bool bWarnIfNotAvail = false ) = 0;
  virtual ErrVal value( const std::string& rcOption, float&  rVal, bool bWarnIfNotAvail = false ) = 0;
};


class ParameterRead : public ParameterIoIf
{
private:
  class PlusEqualMerger : protected ToolLibParser
  {
  public:
    ErrVal mergePlusEqual( std::string &rcString, int iErrId )
    {
      RNOK( ToolLibParser::init( rcString, iErrId ));

      std::stringstream cNewString;

      std::set<std::string> cCheckDoubleOccurences;

      std::string cOption;
      std::string cValue;
      ErrVal nRet = Err::m_nOK;
      while( Err::m_nEndOfFile != peekToken( cOption ) )
      {
        if( Err::m_nOK != (nRet = getToken( cOption )) )
        {
          std::cout << "Error in config file." << std::endl;
          break;
        }
        else if( Err::m_nOK != (nRet = getExpected( "=" )) )
        {
          std::cout << "'=' expected after parameter: " << cOption << std::endl;
          break;
        }
        else if (Err::m_nOK != (nRet = getString(cValue)) )
        {
          std::cout << "unexpected value after '=' in config file." << std::endl;
          break;
        }
        else
        {
          std::string cPlusToken;
          while( Err::m_nOK == peekToken( cPlusToken ) && cPlusToken == "+" )
          {
            getExpected( "+" );
            std::string cAppendString;
            if( Err::m_nOK != getString( cAppendString ) )
            {
              std::cout << "unexpected value after '+' in config file." << std::endl;
              RNOK( ToolLibParser::uninit() );
              return Err::m_nERR;
            }
            cValue += cAppendString;
          }
          if( cCheckDoubleOccurences.end() != cCheckDoubleOccurences.find( cOption ) )
          {
            std::cout << "Option already defined: " << cOption << std::endl;
            nRet = Err::m_nERR;
            break;
          }
          cCheckDoubleOccurences.insert( cOption );
          cNewString << cOption << " = \"" << cValue << "\"" << std::endl; 
        }
      }

      rcString.clear();
      if( Err::m_nOK == nRet )
      {
        rcString = cNewString.str();
      }
      RNOK( ToolLibParser::uninit() );
      return Err::m_nOK;
    }
  };
public:

  ParameterRead(  const std::string& rcFileContent, bool bPrintWarnings = true, int iErrId = 0, bool bReturnAlwaysSuccess = false )
    : m_bPrintWarnings( bPrintWarnings )
    , m_bRAS( bReturnAlwaysSuccess )
  {
    std::string cCopy = rcFileContent;
    PlusEqualMerger cMerger;
    ANOK( cMerger.mergePlusEqual( cCopy, iErrId ) );
    m_cInitOptionParser.init( cCopy, iErrId );
    m_cInitOptionParser.setIgnoreCase( false );
  }
  virtual ~ParameterRead()
  {
    m_cInitOptionParser.uninit( m_bPrintWarnings );
  }
  ErrVal value( const std::string& rcOption, std::string& rVal, bool bWarnIfNotAvail = false) { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, bool&   rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, char&   rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, short&  rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, int&  rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, Int64&  rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, UChar&  rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, UShort& rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, UInt& rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, UInt64& rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }
  ErrVal value( const std::string& rcOption, float&  rVal, bool bWarnIfNotAvail = false )     { ErrVal nRet = m_cInitOptionParser.getValue( rcOption, rVal, bWarnIfNotAvail ); return m_bRAS ? Err::m_nOK : nRet; }

private:
  ToolLibInitOptionParser m_cInitOptionParser;
  bool m_bPrintWarnings;
  bool m_bRAS;
};


#endif // !defined(AFX_PARAMETERIO_H__6CF9086D_DB71_40C0_91BD_6F205082CB8C__INCLUDED_)
