/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PLANEPACKAGE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PLANEPACKAGE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/PicBlock.h"
#include "VideoCommonLib/ReferencePicture.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class SliceData;
class ContextArrayStartIndices;
class CtxInitArray;
class AccessUnitParameters;
class PredInfo;

class PlanePackageParameters
{
public:
  PlanePackageParameters()
    : m_iPlaPaId                    ( -1 )
    , m_uiPrimPlaneId               ( 0 )
    , m_iRefPlaPaId                 ( -1 )
    , m_bAdaptiveInterPlaPaPredFlag ( false )
    , m_uiLog2MinPredBlkSize        ( 0 )
    , m_uiLog2MaxPredBlkSize        ( 0 )
    , m_uiMvAccuracy                ( 0 )
    , m_bMvCompetition              ( false )
    , m_uiTransformMode             ( 0 )
    , m_uiNumQuantizers             ( 0 )
    , m_uiLog2MinPow2TrafoBlkSize   ( 0 )
    , m_uiLog2MaxPow2TrafoBlkSize   ( 0 )
  {}
  PlanePackageParameters( UInt uiPlaPaId, UInt uiPrimPlaneId, UInt uiNumPlanes, const PicBlockStats& rcAUStats )
    : m_iPlaPaId                    ( (int)uiPlaPaId )
    , m_uiPrimPlaneId               ( uiPrimPlaneId )
    , m_cPicBlockStats              ( xGetPlaPaPicBlockStats( uiPrimPlaneId, uiNumPlanes, rcAUStats ) )
    , m_cPrimPlaneSubsampling       ( rcAUStats.getSubsampling()[ uiPrimPlaneId ] )
    , m_iRefPlaPaId                 ( -1 )
    , m_bAdaptiveInterPlaPaPredFlag ( false )
    , m_uiLog2MinPredBlkSize        ( 0 )
    , m_uiLog2MaxPredBlkSize        ( 0 )
    , m_uiMvAccuracy                ( 0 )
    , m_bMvCompetition              ( false )
    , m_uiTransformMode             ( 0 )
    , m_uiNumQuantizers             ( 0 )
    , m_uiLog2MinPow2TrafoBlkSize   ( 0 )
    , m_uiLog2MaxPow2TrafoBlkSize   ( 0 )
  {}

  int                             getPlaPaId                    ()                const { return m_iPlaPaId; }
  UInt                            getPrimPlaneId                ()                const { return m_uiPrimPlaneId; }
  const PicBlockStats&            getPlaPaStats                 ()                const { return m_cPicBlockStats; }
  UInt                            getNumPlanes                  ()                const { return UInt( m_cPicBlockStats.getSubsampling().size() ); }
  const std::vector<Subsampling>& getSubsampling                ()                const { return m_cPicBlockStats.getSubsampling(); }
  const Subsampling&              getSubsampling                ( UInt uiPlane )  const { return m_cPicBlockStats.getSubsampling()[ uiPlane ]; }
  const Subsampling&              getPrimPlaneSubsampling       ()                const { return m_cPrimPlaneSubsampling; }

  bool                            getInterPlaPaPredFlag         ()                const { return m_iRefPlaPaId >= 0; }
  int                             getRefPlaPlaId                ()                const { return m_iRefPlaPaId; }
  bool                            getAdaptiveInterPlaPaPredFlag ()                const { return m_bAdaptiveInterPlaPaPredFlag; }
  UInt                            getLog2MinPredBlkSize         ()                const { return m_uiLog2MinPredBlkSize; }
  UInt                            getLog2MaxPredBlkSize         ()                const { return m_uiLog2MaxPredBlkSize; }
  UInt                            getNumPredBlkSizes            ()                const { return m_uiLog2MaxPredBlkSize - m_uiLog2MinPredBlkSize + 1; }
  UInt                            getMvAccuracy                 ()                const { return m_uiMvAccuracy; }
  bool                            getMvCompetition              ()                const { return m_bMvCompetition; }
  UInt                            getTransformMode              ()                const { return m_uiTransformMode; }
  UInt                            getNumQuantizers              ()                const { return m_uiNumQuantizers; }
  UInt                            getLog2MinPow2TrafoBlkSize    ()                const { return m_uiLog2MinPow2TrafoBlkSize; }
  UInt                            getLog2MaxPow2TrafoBlkSize    ()                const { return m_uiLog2MaxPow2TrafoBlkSize; }
  UInt                            getNumPow2TrafoBlkSizes       ()                const { return m_uiLog2MaxPow2TrafoBlkSize - m_uiLog2MinPow2TrafoBlkSize + 1; }
  UInt                            getNumPow2SigCtxClasses       ()                const { return UInt( m_auiNumPow2SigContexts .size() ); }
  UInt                            getNumPow2LastCtxClasses      ()                const { return UInt( m_auiNumPow2LastContexts.size() ); }
  bool                            getCommonPow2SigCtxFlag       ()                const { return m_auiNumPow2SigContexts .size() == 1; }
  bool                            getCommonPow2LastCtxFlag      ()                const { return m_auiNumPow2LastContexts.size() == 1; }
  UInt                            getNumPow2SigCtx              ( UInt uiIndex )  const { return xGetVectorEntry( m_auiNumPow2SigContexts,  uiIndex ); }
  UInt                            getNumPow2LastCtx             ( UInt uiIndex )  const { return xGetVectorEntry( m_auiNumPow2LastContexts, uiIndex ); }
  int                             getPlaneQp                    ( UInt uiPlane )  const { return xGetVectorEntry( m_aiPlaneQps, uiPlane ); }
  const std::vector<UInt>&        getNumPow2SigCtxs             ()                const { return m_auiNumPow2SigContexts; }
  const std::vector<UInt>&        getNumPow2LastCtxs            ()                const { return m_auiNumPow2LastContexts; }
  const std::vector<int>&         getPlaneQps                   ()                const { return m_aiPlaneQps; }

  std::vector<UInt>&              getNumPow2SigCtxs             ()                      { return m_auiNumPow2SigContexts; }
  std::vector<UInt>&              getNumPow2LastCtxs            ()                      { return m_auiNumPow2LastContexts; }
  std::vector<int>&               getPlaneQps                   ()                      { return m_aiPlaneQps; }

  void  setNoInterPlaPaPredFlag       ()                      { m_iRefPlaPaId = -1; }
  void  setRefPlaPaId                 ( int  iRefId )         { m_iRefPlaPaId = iRefId; }
  void  setAdaptiveInterPlaPaPredFlag ( bool bFlag )          { m_bAdaptiveInterPlaPaPredFlag = bFlag; }
  void  setInterPlaPaPredMode         ( UInt uiPrdMode )      { m_iRefPlaPaId = ( uiPrdMode > 0 ? 0 : -1 ); m_bAdaptiveInterPlaPaPredFlag = ( uiPrdMode > 1 ); }
  void  setLog2MinPredBlkSize         ( UInt uiLog2BlkSize )  { m_uiLog2MinPredBlkSize = uiLog2BlkSize; }
  void  setLog2MaxPredBlkSize         ( UInt uiLog2BlkSize )  { m_uiLog2MaxPredBlkSize = uiLog2BlkSize; }
  void  setMvAccuracy                 ( UInt uiMvAccuracy )   { m_uiMvAccuracy = uiMvAccuracy; }
  void  setMvCompetition              ( bool bMvCompetition ) { m_bMvCompetition = bMvCompetition; }
  void  setTransformMode              ( UInt uiMode )         { m_uiTransformMode = uiMode; m_uiNumQuantizers = xGetNumQuantizers( m_uiTransformMode ); }
  void  setLog2MinPow2TrafoBlkSize    ( UInt uiLog2BlkSize )  { m_uiLog2MinPow2TrafoBlkSize = uiLog2BlkSize; }
  void  setLog2MaxPow2TrafoBlkSize    ( UInt uiLog2BlkSize )  { m_uiLog2MaxPow2TrafoBlkSize = uiLog2BlkSize; }
  void  setNumCommonPow2SigCtx        ( UInt uiNumCtx )       { m_auiNumPow2SigContexts .clear(); m_auiNumPow2SigContexts .push_back( uiNumCtx ); }
  void  setNumCommonPow2LastCtx       ( UInt uiNumCtx )       { m_auiNumPow2LastContexts.clear(); m_auiNumPow2LastContexts.push_back( uiNumCtx ); }

  void  setNumPow2SigCtx  ( const std::vector<UInt>& racNum ) { m_auiNumPow2SigContexts  = racNum; }
  void  setNumPow2LastCtx ( const std::vector<UInt>& racNum ) { m_auiNumPow2LastContexts = racNum; }
  void  setPlaneQps       ( const std::vector<int>&  racQps ) { m_aiPlaneQps = racQps; }

private:
  static PicBlockStats  xGetPlaPaPicBlockStats( UInt uiPrimPlaneId, UInt uiNumPlanes, const PicBlockStats& rcAUStats )
  {
    BlockSize                 cBlkSizePrim  = rcAUStats.getPrimBlockSize() * rcAUStats.getSubsampling()[ uiPrimPlaneId ];
    PrimBlockSize             cPrimBlkSize( cBlkSizePrim.getX(), cBlkSizePrim.getY() ); //  = PrimBlockSize( cBlkSizePrim.getX(), cBlkSizePrim.getY() );
    std::vector<Subsampling>  acPlaPaSub;
    acPlaPaSub.reserve( uiPrimPlaneId + uiNumPlanes );
    for( UInt uiPlaneId = uiPrimPlaneId; uiPlaneId < uiPrimPlaneId + uiNumPlanes; uiPlaneId++ )
    {
      BlockSize   cPlaneBlockSize   = rcAUStats.getPrimBlockSize() * rcAUStats.getSubsampling()[ uiPlaneId ];
      Subsampling cPlaneSubsampling = cPlaneBlockSize / cPrimBlkSize;
      acPlaPaSub.push_back( cPlaneSubsampling );
    }
    return PicBlockStats( cPrimBlkSize, acPlaPaSub );
  }
  
  static UInt xGetNumQuantizers( UInt uiTransformMode )
  {
    UInt uiNumQuantizers = 0;
    for( ; uiTransformMode; uiTransformMode >>= 1 )
    {
      if( uiTransformMode & 1 )
      {
        uiNumQuantizers++;
      }
    }
    return uiNumQuantizers;
  }
  
  template<typename T>
  static T xGetVectorEntry( const std::vector<T>& cVector, UInt uiIndex )
  {
    AOF( uiIndex < cVector.size() );
    return cVector[ uiIndex ];
  }

private:
  //--- general ---
  int               m_iPlaPaId;
  UInt              m_uiPrimPlaneId;
  PicBlockStats     m_cPicBlockStats;
  Subsampling       m_cPrimPlaneSubsampling;
  //--- inter-plane-package prediction ---
  int               m_iRefPlaPaId;
  bool              m_bAdaptiveInterPlaPaPredFlag;
  //--- prediction ---
  UInt              m_uiLog2MinPredBlkSize;
  UInt              m_uiLog2MaxPredBlkSize;
  UInt              m_uiMvAccuracy;
  bool              m_bMvCompetition;
  //--- transform ---
  UInt              m_uiTransformMode;
  UInt              m_uiNumQuantizers;
  UInt              m_uiLog2MinPow2TrafoBlkSize;
  UInt              m_uiLog2MaxPow2TrafoBlkSize;
  //--- entropy coding ---
  std::vector<UInt> m_auiNumPow2SigContexts;
  std::vector<UInt> m_auiNumPow2LastContexts;
  //--- quantization parameters ---
  std::vector<int>  m_aiPlaneQps;
};


class PlanePackage
{
public:
  PlanePackage( const PlanePackageParameters& rcPlaPaPars, UInt uiNumSlices );
  ~PlanePackage();

  const SliceData&  getSlice        ( UInt uiSliceIdx ) const { return *m_apcSlices[uiSliceIdx]; }
  SliceData&        getCurrentSlice ()                        { return *m_apcSlices.back(); }
  const SliceData&  getCurrentSlice () const                  { return *m_apcSlices.back(); }

  const PlanePackageParameters& getPars       () const { return m_cPars; }

  void deleteSliceData();

private:
  PlanePackageParameters        m_cPars;
  std::vector<SliceData*>       m_apcSlices;    //!< contains all slices of the access unit
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _PLANEPACKAGE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
