/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_POCANDFRAMENUM_H__F31F1842_FFCD_42AD_A981_7BD2736A4431__INCLUDED_)
#define AFX_POCANDFRAMENUM_H__F31F1842_FFCD_42AD_A981_7BD2736A4431__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VideoCommonLib/H264SliceHeader.h"

// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

#if defined( WIN32 )
# pragma warning( disable: 4275 )
# pragma warning( disable: 4251 )
#endif

class VIDEOCOMMONLIB_DECL PocAndFrameNum
{
public :
  PocAndFrameNum() : m_uiFrameNum(0), m_iPoc(0) {}
  ~PocAndFrameNum() {}

  void reset()
  {
    m_uiFrameNum = 0;
    m_iPoc = 0;
  }

  void incFrameNum() { m_uiFrameNum++; }

  UInt64 getFrameNum() const { return m_uiFrameNum; }
  Int64 getPOC()       const { return m_iPoc; }

  UInt64 calculateFrameNum( const UInt uiFrameNumLsb, const UInt uiLogMaxFrameNum)
  {
    Int64 iMaxFrameNumLsb      = 1 << uiLogMaxFrameNum;
    const UInt64 uiLsbMask      = iMaxFrameNumLsb-1;

    Int64 iCurrFrameNumMsb     = m_uiFrameNum & ~uiLsbMask;
    Int64 iCurrFrameNumLsb     = uiFrameNumLsb;
    Int64 iDiffFrameNumLsb     = (m_uiFrameNum & uiLsbMask) - iCurrFrameNumLsb;

    if( iDiffFrameNumLsb >= ( iMaxFrameNumLsb >> 1 ) )
    {
      iCurrFrameNumMsb += iMaxFrameNumLsb;
    }

    const UInt64 uiFrameNum = iCurrFrameNumMsb + iCurrFrameNumLsb;
    // check if FrameNum is illegally decreasing
    AOT( uiFrameNum < m_uiFrameNum );

    // update the FrameNum
    m_uiFrameNum = uiFrameNum;

//    std::cout << " FrameNum " << (int)m_uiFrameNum << "   ";
    return m_uiFrameNum;
  }

  Int64 calculatePoc( const UInt uiPicOrderCntLsb, const UInt uiLog2MaxPicOrderCntLsb, bool bUpdate, bool bReset )
  {
    const Int64 iMaxPocLsb      = 1 << uiLog2MaxPicOrderCntLsb;
    const UInt64 uiLsbMask      = iMaxPocLsb-1;

    Int64 iCurrPocMsb     = m_iPoc & ~uiLsbMask;
    Int64 iCurrPocLsb     = uiPicOrderCntLsb;
    Int64 iDiffPocLsb     = (m_iPoc & uiLsbMask) - iCurrPocLsb;

    if( bReset )
    {
      iCurrPocMsb  = 0;
      iCurrPocLsb  = 0;
    }
    else if( iDiffPocLsb >= ( iMaxPocLsb >> 1 ) )
    {
      iCurrPocMsb += iMaxPocLsb;
    }
    else if( -iDiffPocLsb > ( iMaxPocLsb >> 1 ) )
    {
      iCurrPocMsb -= iMaxPocLsb;
    }
    else // there's no wrap of the poc
    {

    }

    const Int64 iCurrPoc = iCurrPocMsb + iCurrPocLsb;

//    std::cout << " Poc " << (int)iCurrPoc << "   ";
    if( bUpdate)
    {
      m_iPoc = iCurrPoc;
    }

    return iCurrPoc;
  }

private:
  UInt64 m_uiFrameNum;
  Int64 m_iPoc;
};


// h264 namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_POCANDFRAMENUM_H__F31F1842_FFCD_42AD_A981_7BD2736A4431__INCLUDED_)
