/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PREDICTIONCALCULATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PREDICTIONCALCULATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/QuadtreeGroup.h"
#include "VideoCommonLib/PicBlock.h"
#include "VideoCommonLib/CodingState.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4251 )
#endif

class IntraDC;
class IntraDirectional;
class InterTranslational;
class InterPlanePackage;

class PredInfoIntraDC;
class PredInfoIntraDirectional;
class PredInfoInterTranslational;
class PredInfoInterPlanePackage;

class PredSyntax;
class PredSyntaxIntraDC;
class PredSyntaxIntraDirectional;
class PredSyntaxInterTranslational;
class PredSyntaxInterPlanePackage;

class PredInfo
{
public:
  typedef enum
  {
    PREDICTION_MODE_UNSET      = 0,
    INTRA_DC                   = 1,
    INTRA_DIRECTIONAL          = 2,
    INTRA_PLANE                = 3,
    INTER_TRANSLATIONAL          = 4,
    INTER_PLANE_PACKAGE        = 5,
  } PredictionMode;

public:
  virtual ~PredInfo() {}

  virtual PredictionMode  getPredictionMode           ()  const = 0;
  virtual bool            isIntraPredicted            ()  const = 0;
  virtual bool            oversizedPredictionSupported()  const = 0;

  virtual const PredSyntax*                       getPredSyntax                     ()  const { return 0; }
  virtual const PredInfoIntraDC*                  getPredInfoIntraDC                ()  const { return 0; }
  virtual const PredInfoIntraDirectional*         getPredInfoIntraDirectional       ()  const { return 0; }
  virtual const PredInfoInterTranslational*         getPredInfoInterTranslational       ()  const { return 0; }
  virtual const PredInfoInterPlanePackage*        getPredInfoInterPlanePackage      ()  const { return 0; }

  virtual ErrVal  setPredInfo   ( const QTNavi&                 rcPos,
                                  const PlanePackageParameters& rcPlaPaPars,
                                  const AccessUnitParameters&   rcAUPars,
                                  const SliceData&              rcSliceData,
                                  const PlanePackage*           pcRefPlanePackage )  = 0;
  virtual ErrVal  setPredSyntax ( const QTNavi&                 rcPos,
                                  const PlanePackageParameters& rcPlaPaPars,
                                  const AccessUnitParameters&   rcAUPars,
                                  const SliceData&              rcSliceData,
                                  const PlanePackage*           pcRefPlanePackage )  = 0;

protected:
  PredInfo()                              {}
  PredInfo( const PredInfo& rcPredInfo )  {}
  
  const PredInfo& operator=( const PredInfo& rcPredInfo )  { return *this; }
};


class PredSyntax : virtual public PredInfo
{
private:
  typedef enum
  {
    PRED_INFO_PRESENT   = 1,
    PRED_SYNTAX_PRESENT = 2
  } Status;

public:
  virtual ~PredSyntax() {}

  virtual PredInfo::PredictionMode  getCodingMode ()  const = 0;
  virtual bool                      isIntraCoded  ()  const = 0;

  virtual const PredSyntaxIntraDC*                  getPredSyntaxIntraDC                ()  const { return 0; }
  virtual const PredSyntaxIntraDirectional*         getPredSyntaxIntraDirectional       ()  const { return 0; }
  virtual const PredSyntaxInterTranslational*         getPredSyntaxInterTranslational       ()  const { return 0; }
  virtual const PredSyntaxInterPlanePackage*        getPredSyntaxInterPlanePackage      ()  const { return 0; }

  virtual ErrVal setPredInfo( const QTNavi&                 rcPos,
                              const PlanePackageParameters& rcPlaPaPars,
                              const AccessUnitParameters&   rcAUPars,
                              const SliceData&              rcSliceData,
                              const PlanePackage*           pcRefPlanePackage )
  {
    ROF ( xIsPredSyntaxOnly () );
    RNOK( xSetComplementData( rcPos, rcPlaPaPars, rcAUPars, rcSliceData, pcRefPlanePackage ) );
    RNOK( xSetComplete      () );
    return Err::m_nOK;
  }

  virtual ErrVal setPredSyntax( const QTNavi&                 rcPos,
                                const PlanePackageParameters& rcPlaPaPars,
                                const AccessUnitParameters&   rcAUPars,
                                const SliceData&              rcSliceData,
                                const PlanePackage*           pcRefPlanePackage )
  {
    ROF ( xIsPredInfoOnly   () );
    RNOK( xSetComplementData( rcPos, rcPlaPaPars, rcAUPars, rcSliceData, pcRefPlanePackage ) );
    RNOK( xSetComplete      () );
    return Err::m_nOK;
  }

protected:
  PredSyntax( bool              bEncoder     ) : m_eStatus ( bEncoder ? PRED_INFO_PRESENT : PRED_SYNTAX_PRESENT ) {}
  PredSyntax( const PredSyntax& rcPredSyntax ) : m_eStatus ( rcPredSyntax.m_eStatus )                             {}
  
  const PredSyntax& operator=( const PredSyntax& rcPredSyntax ) { m_eStatus = rcPredSyntax.m_eStatus; return *this; }
  
  bool  xIsPredInfoOnly     ()  const { return   m_eStatus                         == PRED_INFO_PRESENT;   }
  bool  xIsPredSyntaxOnly   ()  const { return   m_eStatus                         == PRED_SYNTAX_PRESENT; }
  bool  xIsPredInfoPresent  ()  const { return ( m_eStatus & PRED_INFO_PRESENT   ) == PRED_INFO_PRESENT;   }
  bool  xIsPredSyntaxPresent()  const { return ( m_eStatus & PRED_SYNTAX_PRESENT ) == PRED_SYNTAX_PRESENT; }
  
  virtual ErrVal  xSetComplementData( const QTNavi&                 rcPos,
                                      const PlanePackageParameters& rcPlaPaPars,
                                      const AccessUnitParameters&   rcAUPars,
                                      const SliceData&              rcSliceData,
                                      const PlanePackage*           pcRefPlanePackage )  = 0;

private:
  ErrVal xSetComplete()
  { 
    ROF( xIsPredInfoOnly() || xIsPredSyntaxOnly() );
    m_eStatus = Status( PRED_INFO_PRESENT + PRED_SYNTAX_PRESENT ); 
    return Err::m_nOK;
  }

private:
  Status m_eStatus;
};



class PredictionCalculator
{
public:
  virtual ~PredictionCalculator() {}

  virtual ErrVal calculatePrediction( const PredInfo &rcPredInfo,
                                      const QTNavi& rcNav,
                                      const PrimPixelPos& rcTopLeftPosOfPrediction,
                                      const PicBlock<const Pel> *pcEncPic,
                                      const CodingState &rcCodingState,
                                      const SliceData &rcCurrSlice,
                                      const PicBlock<DPel> &rcPredPicBlock,
                                      bool &rbPredictionValid ) = 0;

protected:
  PredictionCalculator() {}
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4251 )
#endif


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _PREDICTIONCALCULATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
