/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_PROBBIN_H__CBFE313E_2382_4ECC_9D41_416668E3507D__INCLUDED_)
#define AFX_PROBBIN_H__CBFE313E_2382_4ECC_9D41_416668E3507D__INCLUDED_

#include "VideoCommonLib/ExtendedCabacTables.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class ProbBin
{
private:
  typedef UChar StateType;

  enum
  {
    IS_LPS_FLAG_BIT_MASK  = 1<<6
  };


public:
  ProbBin() {}
  ProbBin( UShort uiProb, bool bIsLps ) : m_uiProbBin( bIsLps ? uiProb | IS_LPS_FLAG_BIT_MASK : uiProb ) { TOT_DBG( uiProb >= IS_LPS_FLAG_BIT_MASK ); }

  UShort getStateIdx()               const { return m_uiProbBin & ( IS_LPS_FLAG_BIT_MASK - 1 ); }
  bool   isLps()                     const { return m_uiProbBin & IS_LPS_FLAG_BIT_MASK ? true : false; }

  UChar getProbabilityIdxUnlimited() const { return isLps() ? 128 + getStateIdx() : 127 - getStateIdx(); }
  UChar getProbabilityIdx()          const { return isLps() ? 64  + getStateIdx() : 63  - getStateIdx(); }

  double getProbBinRate() const
  {
    return sm_adEntropy[getProbabilityIdx()];
//    return sm_adEntropy256[getProbabilityIdxUnlimited()];
  }

  template<class t_Tables>
  BitRate getProbBinBitRate() const
  {
    CAOF(( TypeCompare<t_Tables, DefaultCabacTables>::m_bEqual ));
    return sm_acBitsFromRawProbBin[m_uiProbBin];
  }

  static const double sm_adEntropy[128];
  static const double sm_adProbability[128];

  static const double sm_adEntropy256[256];
  static const double sm_adProbability256[256];

  static const BitRate sm_acBitsFromRawProbBin[130];

private:
  StateType m_uiProbBin;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif //!defined(AFX_PROBBIN_H__CBFE313E_2382_4ECC_9D41_416668E3507D__INCLUDED_)
