/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _REFERENCEPICTURE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _REFERENCEPICTURE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "PicBlock.h"
#include "InterCommon.h"
#include "InterpolationIf.h"
// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_PelIn>
class ReferencePicture
{
public:
  typedef typename TypeLimits<t_PelIn>::DType DType;

  typedef enum
  {
    NO_UPSAMPLING,
    SIX_TAP_H264,
    QUADRATIC_SPLINE_FLOAT,
    CUBIC_SPLINE_FLOAT,
    CUBIC_SPLINE_FIXED,
    CUBIC_MOMS_FLOAT,
    CUBIC_MOMS_FIXED,
    QUINTIC_SPLINE_FLOAT,
    QUINTIC_SPLINE_FIXED,
    QUINTIC_MOMS_FLOAT,
    QUINTIC_MOMS_FIXED,
    NUM_UPS_METHODS,
    SIX_TAP_H264_SLOW,
    CUBIC_MOMS_FIXED_MC,
    CUBIC_SPLINE_FIXED_MC,
  } UpsMethod;

  ReferencePicture( const PicOrderCnt cPoc, const PicBlockStats& rcAUStats, UInt uiUpsShift, UInt uiMargin, typename ReferencePicture<t_PelIn>::UpsMethod cUpsMethod );
  virtual ~ReferencePicture()
  {
    delete m_pcUpsPic;
    delete m_pcInterpolationIf;
  }

  const PicOrderCnt&      getPicOrderCnt()  const { return m_cPoc; }
  const UInt              getUpsShift   ()  const { return m_uiUpsShift; }
  const PrimPixelPos&     getPrimOrigin ()  const { return m_cPrimOrigin; }
  const PicBlockStats&    getStats      ()  const { return m_pcUpsPic->getStats(); }
  const PicBlock<DType>&  getPicBlock   ()  const { return *m_pcUpsPic; }
  UpsMethod               getUpsMethod  ()  const { return m_eUpsMethod; }
  const InterpolationIf<t_PelIn, DType>* getInterpolationIf() const { return m_pcInterpolationIf; } 
  virtual ErrVal assignPlanePackage( UInt uiPrimPlaneId, const PicBlock<const t_PelIn> &rcPic, UpsMethod cUpsMethod, const UInt uiMargin = 0 );

  virtual const PrimBlockSize&              get4x4BlockL1NormsPrimOrigin()  const { return *(PrimBlockSize*)0; }
  virtual const PicBlock<UInt>*             get4x4BlockL1NormsPic       ()  const { return 0; }
  virtual const PicBlock<DType>*            getDecPic                   ()  const { return 0; }
  virtual const ReferencePicture<t_PelIn>*  getOrigRefPicture           ()  const { return 0; }

protected:
  static Subsampling              xGetMinSubsampling( const PicBlockStats& rcAUStats );
  static std::vector<Subsampling> xGetUpsSubsampling( const PicBlockStats& rcAUStats, UInt uiUpsShift );

private:
  PicOrderCnt       m_cPoc;
  UInt              m_uiUpsShift;
  PrimPixelPos      m_cPrimOrigin;
  PicBlock<DType>*  m_pcUpsPic;
  InterpolationIf<t_PelIn, DType>* m_pcInterpolationIf;
  UpsMethod         m_eUpsMethod;

  template<bool t_bHor>
  class Clipper
  {
  public:
    Clipper( const BlockSize &rcSize, int iStepH, int iStepV )
      : m_iLastH( ( ( rcSize.getX() - 1 ) >> iStepH ) << iStepH )
      , m_iLastV( ( ( rcSize.getY() - 1 ) >> iStepV ) << iStepV )
      , m_iStep( 1 << (t_bHor ? iStepH : iStepV) ) {}


    PixelPos get( const PixelPos& cPos ) const
    {
      if( t_bHor )
      {
        return PixelPos( cPos.getX(), gMinMax( cPos.getY(), 0, m_iLastV ) );
      }
      return PixelPos( gMinMax( cPos.getX(), 0, m_iLastH ), cPos.getY() );
    }

    PixelPos get( const PixelPos& cPos, const int iOff ) const
    {
      const int iStep = iOff * m_iStep;
      if( t_bHor )
      {
        return PixelPos( gMinMax( cPos.getX() + iStep, 0, m_iLastH ), cPos.getY() );
      }
      return PixelPos( cPos.getX(), gMinMax( cPos.getY() + iStep, 0, m_iLastV ) );
    }

  private:
    int m_iLastH;
    int m_iLastV;
    int m_iStep;
  };


  template<bool t_bClipDest, bool t_bCopySrc, int t_iRightShiftResult, bool t_bHor, int t_iTapShift, int t_iTap0, int t_iTap1, int t_iTap2, int t_iTap3, int t_iTap4, int t_iTap5, int t_iTap6, int t_iTap7, int t_iTap8, int t_iTap9, class t_TDst, class t_TSrc>
  void xUpsampleHV( const Block<t_TDst> &rcDst, const Block<const t_TSrc> &rcSrc, const int iStepHDst, const int iStepVDst, const int iStepHSrc, const int iStepVSrc, const PixelPos& cSrcOff )
  {
    CAOT( !t_bCopySrc && t_iRightShiftResult != 0 );
    typedef int IntType;
    CAOT( sizeof( IntType ) >= sizeof( Int64 ) );
    const Clipper<t_bHor> cClp( rcSrc.getSize(), iStepHSrc, iStepVSrc );
    const PixelPos cStep( t_bHor ? 1<<iStepHDst : 0, t_bHor ? 0 : 1<<iStepVDst );
    PixelPos cDstPos, cSrcPos;
    const int iDstVInc = (t_bHor ? 1 : 2) << iStepVDst;
    const int iDstHInc = (t_bHor ? 2 : 1) << iStepHDst;

    for( cDstPos.getY() = 0, cSrcPos.getY() = cSrcOff.getY(); cDstPos.getY() < rcDst.getSize().getY(); cDstPos.getY() += iDstVInc, cSrcPos.getY() += 1<<iStepVSrc )
    {
      for( cDstPos.getX() = 0, cSrcPos.getX() = cSrcOff.getX(); cDstPos.getX() < rcDst.getSize().getX(); cDstPos.getX() += iDstHInc, cSrcPos.getX() += 1<<iStepHSrc )
      {
        const PixelPos cSrcClp = cClp.get( cSrcPos );
        if( t_bCopySrc )
        {
          Int64 iVal = rcSrc.getEl( cClp.get( cSrcClp, 0 ) );
          if( t_iRightShiftResult > 0 )
          {
            const UInt uiSh = gAbs( t_iRightShiftResult );
            iVal = ( (iVal<<1) + (Int64(1)<<uiSh) ) >> (uiSh+1);
          }
          else if( t_iRightShiftResult < 0 )
          {
            const UInt uiSh = gAbs( t_iRightShiftResult );
            iVal <<= uiSh;
          }
          if( t_bClipDest )
          {
            rcDst.getEl( cDstPos ) = t_TDst( gMinMax<Int64>( iVal, Int64( TypeLimits<t_TDst>::m_iMin ), Int64( TypeLimits<t_TDst>::m_iMax ) ) );
          }
          else
          {
            rcDst.getEl( cDstPos ) = t_TDst( iVal );
          }
        }

        if( cDstPos + cStep < rcDst.getSize() )
        {
          Int64 iVal          = t_iTap0 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp,  0 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  1 ) ) ) );
          if( t_iTap1 ) iVal += t_iTap1 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -1 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  2 ) ) ) );
          if( t_iTap2 ) iVal += t_iTap2 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -2 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  3 ) ) ) );
          if( t_iTap3 ) iVal += t_iTap3 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -3 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  4 ) ) ) );
          if( t_iTap4 ) iVal += t_iTap4 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -4 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  5 ) ) ) );
          if( t_iTap5 ) iVal += t_iTap5 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -5 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  6 ) ) ) );
          if( t_iTap6 ) iVal += t_iTap6 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -6 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  7 ) ) ) );
          if( t_iTap7 ) iVal += t_iTap7 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -7 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  8 ) ) ) );
          if( t_iTap8 ) iVal += t_iTap8 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -8 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp,  9 ) ) ) );
          if( t_iTap9 ) iVal += t_iTap9 * ( IntType( rcSrc.getEl( cClp.get( cSrcClp, -9 ) ) ) + IntType( rcSrc.getEl( cClp.get( cSrcClp, 10 ) ) ) );

          const int iShift = t_iRightShiftResult + t_iTapShift;
          if( iShift > 0 )
          {
            const UInt uiShift = gAbs( iShift );
            iVal = ( (iVal<<1) + (Int64(1)<<uiShift) ) >> (uiShift+1);
          }
          else if( iShift < 0 )
          {
            const UInt uiShift = gAbs( iShift );
            iVal <<= uiShift;
          }
          if( t_bClipDest)
          {
            rcDst.getEl( cDstPos + cStep ) = t_TDst( gMinMax<Int64>( iVal, Int64( TypeLimits<t_TDst>::m_iMin ), Int64( TypeLimits<t_TDst>::m_iMax ) ) );
          }
          else
          {
            rcDst.getEl( cDstPos + cStep ) = t_TDst( iVal );
          }
        }
      }
    }
  }

  template<bool t_bCopySrc, int t_iRightShiftResult, bool t_bHor, int t_iTapShift, class t_TDst, class t_TSrc>
  void xUpsampleHH264( const Block<t_TDst> &rcDst, const Block<const t_TSrc> &rcSrc )
  {
    CAOT( !t_bCopySrc && t_iRightShiftResult != 0 );
    typedef int IntType;
    CAOT( sizeof( IntType ) >= sizeof( Int64 ) );

    const int iYSize    = rcDst.getSize().getY();
    const int iXSrcSize = rcSrc.getSize().getX()-1;
    const int iMargin   = (rcDst.getSize().getX()/2 - rcSrc.getSize().getX())/2;

    t_TDst* pDst0       = rcDst.getPtr();
    const t_TSrc* pSrc0 = rcSrc.getPtr();
    for( int y = 0; y < iYSize; y++ )
    {
      int xD = 0;
      int xS;
      for( xS = -iMargin; xS < 3; xS++)
      {
        {
//          if( t_bCopySrc )
          {
            int iVal = pSrc0[gMax(0,xS)];
            if( t_iRightShiftResult > 0 )
            {
              const UInt uiSh = gAbs( t_iRightShiftResult );
              iVal = ( iVal + (int(1)<<(uiSh-1)) ) >> uiSh;
            }
            else if( t_iRightShiftResult < 0 )
            {
              const UInt uiSh = gAbs( t_iRightShiftResult );
              iVal <<= uiSh;
            }
            pDst0[xD++] = t_TDst( iVal );
          }

          {
            int iVal  = pSrc0[gMax(0,xS)  ] + pSrc0[gMax(0,xS+1)];
            iVal        = iVal<<2;
            iVal       -= pSrc0[gMax(0,xS-1)] + pSrc0[gMax(0,xS+2)];
            iVal       += iVal<<2;
            iVal       += pSrc0[gMax(0,xS-2)] + pSrc0[gMax(0,xS+3)];

            const int iShift = t_iRightShiftResult + t_iTapShift;
            if( iShift > 0 )
            {
              const UInt uiShift = gAbs( iShift );
              iVal = ( iVal + (int(1)<<(uiShift-1)) ) >> uiShift;
            }
            else if( iShift < 0 )
            {
              const UInt uiShift = gAbs( iShift );
              iVal <<= uiShift;
            }
            pDst0[xD++] = t_TDst( iVal );
          }
        }
      }
  
      for( xS = 3; xS < iXSrcSize-3; xS++)
      {
        {
  //        if( t_bCopySrc )
          {
            int iVal = pSrc0[xS];
            if( t_iRightShiftResult > 0 )
            {
              const UInt uiSh = gAbs( t_iRightShiftResult );
              iVal = ( (iVal<<1) + (int(1)<<uiSh) ) >> (uiSh+1);
            }
            else if( t_iRightShiftResult < 0 )
            {
              const UInt uiSh = gAbs( t_iRightShiftResult );
              iVal <<= uiSh;
            }
            pDst0[xD++] = t_TDst( iVal );
          }

          {
            int iVal  = pSrc0[xS  ] + pSrc0[xS+1];
            iVal        = iVal<<2;
            iVal       -= pSrc0[xS-1] + pSrc0[xS+2];
            iVal       += iVal<<2;
            iVal       += pSrc0[xS-2] + pSrc0[xS+3];

            const int iShift = t_iRightShiftResult + t_iTapShift;
            if( iShift > 0 )
            {
              const UInt uiShift = gAbs( iShift );
              iVal = ( (iVal<<1) + (int(1)<<uiShift) ) >> (uiShift+1);
            }
            else if( iShift < 0 )
            {
              const UInt uiShift = gAbs( iShift );
              iVal <<= uiShift;
            }
            pDst0[xD++] = t_TDst( iVal );
          }
        }
      }
      
      for( xS = iXSrcSize-3; xS <= iXSrcSize+iMargin; xS++)
      {
        {
    //      if( t_bCopySrc )
          {
            int iVal = pSrc0[gMin(xS,  iXSrcSize)];
            if( t_iRightShiftResult > 0 )
            {
              const UInt uiSh = gAbs( t_iRightShiftResult );
              iVal = ( (iVal<<1) + (int(1)<<uiSh) ) >> (uiSh+1);
            }
            else if( t_iRightShiftResult < 0 )
            {
              const UInt uiSh = gAbs( t_iRightShiftResult );
              iVal <<= uiSh;
            }
            pDst0[xD++] = t_TDst( iVal );
          }

          {
            int iVal  = pSrc0[gMin(xS,  iXSrcSize)] + pSrc0[gMin(xS+1,iXSrcSize)];
            iVal        = iVal<<2;
            iVal       -= pSrc0[gMin(xS-1,iXSrcSize)] + pSrc0[gMin(xS+2,iXSrcSize)];
            iVal       += iVal<<2;
            iVal       += pSrc0[gMin(xS-2,iXSrcSize)] + pSrc0[gMin(xS+3,iXSrcSize)];

            const int iShift = t_iRightShiftResult + t_iTapShift;
            if( iShift > 0 )
            {
              const UInt uiShift = gAbs( iShift );
              iVal = ( (iVal<<1) + (int(1)<<uiShift) ) >> (uiShift+1);
            }
            else if( iShift < 0 )
            {
              const UInt uiShift = gAbs( iShift );
              iVal <<= uiShift;
            }
            pDst0[xD++] = t_TDst( iVal );
          }
        }
      }
      pDst0 += rcDst.getStride();
      pSrc0 += rcSrc.getStride();
    }
  }

  template<bool t_bCopySrc, int t_iRightShiftResult, bool t_bHor1, int t_iTapShift, class t_TDst, class t_TSrc>
  void xUpsampleVH264( const Block<t_TDst> &rcDst, const Block<const t_TSrc> &rcSrc )
  {
    CAOT( !t_bCopySrc && t_iRightShiftResult != 0 );
    typedef int IntType;
    CAOT( sizeof( IntType ) >= sizeof( Int64 ) );

    const t_TSrc* pSrcL[3];
    const t_TSrc* pSrcR[3];
    const t_TSrc* pSrcStart = rcSrc.getPtr();
    const t_TSrc* pSrcLast  = pSrcStart + (rcSrc.getSize().getY()-1)*rcSrc.getStride();

    const int iStride = rcSrc.getStride();
    pSrcL[2] = pSrcL[1] = pSrcL[0] = pSrcR[0] = &rcSrc.getEl( 0 ); 
    pSrcR[1] = pSrcR[0] + iStride;
    pSrcR[2] = pSrcR[1] + iStride;

    t_TDst* pDst0 = rcDst.getPtr();
    t_TDst* pDst1 = pDst0 + rcDst.getStride();

    int yS = (rcSrc.getSize().getY() - rcDst.getSize().getY()/2)/2;
    for( int y = 0; y < rcDst.getSize().getY(); y += 2, yS++ )
    {
      if( yS > -3 )
      {
        pSrcL[2] = pSrcL[1];
        pSrcL[1] = pSrcL[0];
        pSrcL[0] = pSrcR[0];
        pSrcR[0] = pSrcR[1];
        pSrcR[1] = pSrcR[2];
        const t_TSrc* pcTemp = pSrcR[2] + iStride;
        pSrcR[2] = pcTemp < pSrcLast ? pcTemp : pSrcLast;
      }
      else
      {
        pSrcL[2] = pSrcStart + gMax(0, yS-2)*rcSrc.getStride();
        pSrcL[1] = pSrcStart + gMax(0, yS-1)*rcSrc.getStride();
        pSrcL[0] = pSrcStart + gMax(0, yS)  *rcSrc.getStride();
        pSrcR[0] = pSrcStart + gMax(0, yS+1)*rcSrc.getStride();
        pSrcR[1] = pSrcStart + gMax(0, yS+2)*rcSrc.getStride();
        pSrcR[2] = pSrcStart + gMax(0, yS+3)*rcSrc.getStride();

      }

      for( int x = 0; x < rcDst.getSize().getX(); x++ )
      {
        {

          Int64 iVal  = pSrcL[0][x] + pSrcR[0][x];
          iVal        = iVal<<2;
          iVal       -= pSrcL[1][x] + pSrcR[1][x];
          iVal       += iVal<<2;
          iVal       += pSrcL[2][x] + pSrcR[2][x];

          const int iShift = t_iRightShiftResult + t_iTapShift;
          if( iShift > 0 )
          {
            const UInt uiShift = gAbs( iShift );
            iVal = ( iVal + (Int64(1)<<(uiShift-1)) ) >> uiShift;
          }
          else if( iShift < 0 )
          {
            const UInt uiShift = gAbs( iShift );
            iVal <<= uiShift;
          }

          const UInt64 i2 = iVal + TypeLimits<t_TDst>::getMax()+1;
          pDst1[x] = t_TDst( ( i2 <= (UInt64)(2*TypeLimits<t_TDst>::getMax()+1) ) ? iVal : iVal < 0 ? TypeLimits<t_TDst>::getMin(): TypeLimits<t_TDst>::getMax());

//          pDst1[x] = t_TDst( iVal < TypeLimits<t_TDst>::getMin() ? TypeLimits<t_TDst>::getMin(): (iVal > TypeLimits<t_TDst>::getMax() ? TypeLimits<t_TDst>::getMax() : iVal));
        }

        if( t_bCopySrc )        
        {
          Int64 iVal = pSrcL[0][x];
          if( t_iRightShiftResult > 0 )
          {
            const UInt uiSh = gAbs( t_iRightShiftResult );
            iVal = ( iVal + (Int64(1)<<(uiSh-1)) ) >> uiSh;
          }
          else if( t_iRightShiftResult < 0 )
          {
            const UInt uiSh = gAbs( t_iRightShiftResult );
            iVal <<= uiSh;
          }
          const UInt64 i2 = iVal + TypeLimits<t_TDst>::getMax()+1;
          pDst0[x] = t_TDst( ( i2 <= (UInt64)(2*TypeLimits<t_TDst>::getMax()+1) ) ? iVal : iVal < 0 ? TypeLimits<t_TDst>::getMin(): TypeLimits<t_TDst>::getMax());
        }
      }
      pDst0 += 2*rcDst.getStride();
      pDst1 += 2*rcDst.getStride();
    }
  }
};


template<class t_PelIn>
class ReferencePlanePackage
{
public:
  typedef typename TypeLimits<t_PelIn>::DType DType;

  ReferencePlanePackage( const ReferencePicture<t_PelIn>& rcRefPic, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling& rcPrimPlaneSubsampling );
  virtual ~ReferencePlanePackage() {}

  const PicOrderCnt&      getPicOrderCnt        ()  const { return m_cPoc; }
  const UInt              getUpsShift           ()  const { return m_uiUpsShift; }
  const PrimPixelPos&     getOrigin             ()  const { return m_cOrigin; }
  const PicBlockStats&    getStats              ()  const { return m_cUpsPlaPa.getStats(); }
  const PicBlock<DType>&  getPicBlock           ()  const { return m_cUpsPlaPa; }
  const ReferencePicture<t_PelIn>& getRefPic    ()  const { return m_rcRefPic; }

  virtual ErrVal getL1Norm4x4Block( const PicBlock<UInt> &cDst, const MotionVector cPosVecToTopLeft ) const { return Err::m_nERR; }

  virtual const PrimBlockSize&                  get4x4BlockL1NormsOrigin  ()  const { return *(PrimBlockSize*)0; }
  virtual const PicBlock<UInt>&                 get4x4BlockL1NormsPlaPa   ()  const { return *(PicBlock<UInt>*)0; }
  virtual const PicBlock<DType>&                getDecRefPlaPa            ()  const { return *(PicBlock<DType>*)0; }
  virtual const ReferencePlanePackage<t_PelIn>& getOrigRefPlaPa           ()  const { return *(ReferencePlanePackage<t_PelIn>*)0; }

private:
  PicOrderCnt        m_cPoc;
  UInt               m_uiUpsShift;
  PrimPixelPos       m_cOrigin;
  PicBlock<DType>    m_cUpsPlaPa;
#if FAST_MC
  const ReferencePicture<t_PelIn>& m_rcRefPic;
#endif
};


template<class t_PelIn>
Subsampling
ReferencePicture<t_PelIn>::xGetMinSubsampling( const PicBlockStats& rcAUStats )
{
  ShiftScale cMinX = rcAUStats.getSubsampling()[ 0 ].getX();
  ShiftScale cMinY = rcAUStats.getSubsampling()[ 0 ].getY();
  for( UInt ui = 1; ui < rcAUStats.getNumPlanes(); ui++ )
  {
    if( rcAUStats.getSubsampling()[ ui ].getX() < cMinX )   cMinX = rcAUStats.getSubsampling()[ ui ].getX();
    if( rcAUStats.getSubsampling()[ ui ].getY() < cMinY )   cMinY = rcAUStats.getSubsampling()[ ui ].getY();
  }
  return Subsampling( cMinX, cMinY );
}

template<class t_PelIn>
std::vector<Subsampling>
ReferencePicture<t_PelIn>::xGetUpsSubsampling( const PicBlockStats& rcAUStats, UInt uiUpsShift )
{
  std::vector<Subsampling> acSub;
  acSub.reserve( rcAUStats.getNumPlanes() );
  for( UInt ui = 0; ui < rcAUStats.getNumPlanes(); ui++ )
  {
    ShiftScale cRatX = rcAUStats.getSubsampling()[ ui ].getX() << uiUpsShift;
    ShiftScale cRatY = rcAUStats.getSubsampling()[ ui ].getY() << uiUpsShift;
    acSub.push_back( Subsampling( cRatX, cRatY ) );
  }
  return acSub;
}

template<class t_PelIn>
ReferencePicture<t_PelIn>::ReferencePicture( const PicOrderCnt cPoc, const PicBlockStats &rcAUStats, UInt uiUpsShift, UInt uiMargin, UpsMethod cUpsMethod )
: m_cPoc        ( cPoc )
, m_uiUpsShift  ( uiUpsShift )
, m_cPrimOrigin ( PixelPos( (int)uiMargin, (int)uiMargin ) / xGetMinSubsampling( rcAUStats ) )
, m_pcUpsPic    ( new AllocPicBlock<DType>( PicBlockStats( rcAUStats.getPrimBlockSize() + m_cPrimOrigin * 2, xGetUpsSubsampling( rcAUStats, uiUpsShift ) ) ) )
, m_eUpsMethod  ( cUpsMethod )
{
  m_pcInterpolationIf =  gCreateInterpolationFilter(cUpsMethod);
/*
  switch ( cUpsMethod )
  {
  case CUBIC_SPLINE_FIXED_MC:
  case CUBIC_SPLINE_FIXED:
    m_pcInterpolationIf = new CubicSplineFixed<t_PelIn,DType>();
    break;
  case QUADRATIC_SPLINE_FLOAT:
    m_pcInterpolationIf = new QuadraticSpline<t_PelIn,DType>();
    break;
  case CUBIC_SPLINE_FLOAT:
    m_pcInterpolationIf = new CubicSpline<t_PelIn,DType>();
    break;
  case CUBIC_MOMS_FLOAT:
    m_pcInterpolationIf = new CubicMOMS<t_PelIn,DType>();
    break;
  case CUBIC_MOMS_FIXED_MC:
  case CUBIC_MOMS_FIXED:
    m_pcInterpolationIf = new CubicMOMSFixed<t_PelIn,DType>();
    break;
  case QUINTIC_SPLINE_FLOAT:
    m_pcInterpolationIf = new QuinticSpline<t_PelIn,DType>();
    break;
  case QUINTIC_SPLINE_FIXED:
    m_pcInterpolationIf = new QuinticSplineFixed<t_PelIn,DType>();
    break;
  case QUINTIC_MOMS_FLOAT:
    m_pcInterpolationIf = new QuinticMOMS<t_PelIn,DType>();
    break;
  case QUINTIC_MOMS_FIXED:
    m_pcInterpolationIf = new QuinticMOMSFixed<t_PelIn,DType>();
    break;
  default:
    m_pcInterpolationIf = NULL;
  }
*/
  if ( m_pcInterpolationIf )
    m_pcInterpolationIf->allocCoeffs( rcAUStats.getPrimBlockSize(), xGetUpsSubsampling( rcAUStats, 0 ) );
}


template<class t_PelIn>
ErrVal ReferencePicture<t_PelIn>::assignPlanePackage( UInt uiPrimPlaneId, const PicBlock<const t_PelIn> &rcPic, UpsMethod cUpsMethod, const UInt uiMargin )
{
  ROF( m_pcUpsPic );
  ROF( uiPrimPlaneId + rcPic.getStats().getNumPlanes() <= m_pcUpsPic->getStats().getNumPlanes() );
  
  if( cUpsMethod == NO_UPSAMPLING )
  {
    ROF( m_uiUpsShift  == 0 );
    ROF( m_cPrimOrigin == PrimPixelPos( 0, 0 ) );
    for( UInt uiInPlane = 0; uiInPlane < rcPic.getStats().getNumPlanes(); uiInPlane++ )
    {
      const Block<DType>&         rcUpsPlane  = m_pcUpsPic->getPlane( uiInPlane + uiPrimPlaneId );
      const Block<const t_PelIn>& rcInPlane   = rcPic      .getPlane( uiInPlane );
      RNOK( rcUpsPlane.assign( rcInPlane ) );
      rcUpsPlane *= 1 << EXTRA_PRECISION_BITS;
    }
    return Err::m_nOK;
  }
  if( cUpsMethod == SIX_TAP_H264_SLOW )
  {
    ROF( m_uiUpsShift == 1 );
    for( UInt uiInPlane = 0; uiInPlane < rcPic.getStats().getNumPlanes(); uiInPlane++ )
    {
      const int                   iTapShift   = 5;
      const Block<DType>&         rcUpsPlane  = m_pcUpsPic->getPlane( uiInPlane + uiPrimPlaneId );
      const Block<const t_PelIn>& rcInPlane   = rcPic      .getPlane( uiInPlane );
      Subsampling                 cSub        = m_pcUpsPic->getStats().getSubsampling()[ uiInPlane + uiPrimPlaneId ] << m_uiUpsShift;
      PixelPos                    cOffset     = PixelPos( 0, 0 ) - m_cPrimOrigin * cSub;
      ROT( rcUpsPlane.getSize().getY() % 2 );
      // H.264 interpolation filter taps
      AllocBlock<int> cTmpPlane( BlockSize( rcUpsPlane.getSize().getX(), rcUpsPlane.getSize().getY() / 2 ) );
      xUpsampleHV<false, true,-iTapShift,                      true,  iTapShift, 20, -5, 1, 0,0,0,0,0,0,0,int,t_PelIn>( cTmpPlane,  rcInPlane, 0, 0, 0, 0, cOffset );
      xUpsampleHV<true,  true, iTapShift-EXTRA_PRECISION_BITS, false, iTapShift, 20, -5, 1, 0,0,0,0,0,0,0,DType,int  >( rcUpsPlane, cTmpPlane, 0, 0, 0, 0, PixelPos( 0, 0 ) );
    }
    return Err::m_nOK;
  }
  if( cUpsMethod == SIX_TAP_H264 )
  {
    ROF( m_uiUpsShift == 1 );
    for( UInt uiInPlane = 0; uiInPlane < rcPic.getStats().getNumPlanes(); uiInPlane++ )
    {
      const int                   iTapShift   = 5;
      const Block<DType>&         rcUpsPlane  = m_pcUpsPic->getPlane( uiInPlane + uiPrimPlaneId );
      const Block<const t_PelIn>& rcInPlane   = rcPic      .getPlane( uiInPlane );
      Subsampling                 cSub        = m_pcUpsPic->getStats().getSubsampling()[ uiInPlane + uiPrimPlaneId ] << m_uiUpsShift;//Subsampling( Ratio( 1 << m_uiUpsShift,1 ), Ratio( 1 << m_uiUpsShift,1 ) );
      PixelPos                    cOffset     = PixelPos( 0, 0 ) - m_cPrimOrigin * cSub;
      ROT( rcUpsPlane.getSize().getY() % 2 );
      // H.264 interpolation filter taps
      AllocBlock<int> cTmpPlane( BlockSize( rcUpsPlane.getSize().getX(), rcInPlane.getSize().getY() ) );
      xUpsampleHH264<true,-iTapShift,                      true,  iTapShift, int,t_PelIn>( cTmpPlane, rcInPlane );
      xUpsampleVH264<true, iTapShift-EXTRA_PRECISION_BITS, false, iTapShift, DType,int  >( rcUpsPlane, cTmpPlane );
    }
    return Err::m_nOK;
  }

  if( m_pcInterpolationIf )
  {
    for( UInt uiInPlane = 0; uiInPlane < rcPic.getStats().getNumPlanes(); uiInPlane++ )
    {
      const Block<DType>&         rcUpsPlane  = m_pcUpsPic->getPlane( uiInPlane + uiPrimPlaneId );
      const Block<const t_PelIn>& rcInPlane   = rcPic      .getPlane( uiInPlane );

      ROT( rcUpsPlane.getSize().getY() % 2 );

      // compute filter coeffs
      m_pcInterpolationIf->calcCoeffs(rcInPlane, uiInPlane + uiPrimPlaneId);

      if( cUpsMethod != CUBIC_MOMS_FIXED_MC && cUpsMethod != CUBIC_SPLINE_FIXED_MC )
      {
        // interpolate using filter coeffs
        m_pcInterpolationIf->interpolate(uiInPlane + uiPrimPlaneId, rcUpsPlane, 1<<m_uiUpsShift);
      }
      //otherwise we do the interpolation on the fly while accessing the prediction data
    }
    return Err::m_nOK;
  }

  return Err::m_nERR;
}


template<class t_PelIn>
ReferencePlanePackage<t_PelIn>::ReferencePlanePackage( const ReferencePicture<t_PelIn> &rcRefPic, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling &rcPrimPlaneSubsampling )
: m_cPoc                 ( rcRefPic.getPicOrderCnt() )
, m_uiUpsShift           ( rcRefPic.getUpsShift() )
, m_cOrigin              ( ( rcRefPic.getPrimOrigin() * rcPrimPlaneSubsampling ).getX(), ( rcRefPic.getPrimOrigin() * rcPrimPlaneSubsampling ).getY() )
, m_cUpsPlaPa            ( rcRefPic.getPicBlock(), uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling )
#if FAST_MC
, m_rcRefPic             (rcRefPic)
#endif
{
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _REFERENCEPICTURE_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
