/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _REFERENCEPICTUREACCESSOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _REFERENCEPICTUREACCESSOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "ReferencePicture.h"
#include "VideoCommonLib/InterCommon.h"
#include <iomanip>


// video namespace begin
VIDEO_NAMESPACE_BEGIN

enum IpolType
{
  IPOLTYPE_COPY,
  IPOLTYPE_HOR,
  IPOLTYPE_VER,
  IPOLTYPE_BILIN,
  IPOLTYPE_LIN_UPRIGHT, // only for testing
  IPOLTYPE_LIN_UPLEFT   // only for testing
};

template<class t_HandlePredictionIf, class t_PelIn>
class ReferencePlanePackageAccessor : public t_HandlePredictionIf
{
public:
  ReferencePlanePackageAccessor( const ReferencePlanePackage<t_PelIn> &rcRefPlaPa, const t_HandlePredictionIf &rcHandlePredIf )
    : t_HandlePredictionIf( rcHandlePredIf ), m_rcRefPlaPa( rcRefPlaPa ) {}

  ErrVal getPrediction( const PicBlockStats &rcPicBlockStats, const MotionVector &rcMotionPositionVector );
  ErrVal getPlanePrediction( UInt uiPlane, const BlockSize &rcBlockSize, const MotionVector &rcMotionPositionVector );

  ErrVal getUpsampledBlock( PicBlock< int > &rcBlockDst, const MotionVector &rcTopLeftPos );

private:
  typedef typename TypeLimits<t_PelIn>::DType DType;
  const ReferencePlanePackage<t_PelIn>& m_rcRefPlaPa;

  template<IpolType t_IpolType, UInt t_uiAcc>
  __inline ErrVal xGetPredBilinearly( const PixelPos &rcPosInUps,
                                      const BinaryFraction<t_uiAcc> &rcOffX,
                                      const BinaryFraction<t_uiAcc> &rcOffY,
                                      const BlockSize& rcBlockSize,
                                      const Block<DType> &rcIn );

  template<IpolType t_IpolType, bool t_bCheckX, bool t_bCheckY, UInt t_uiAcc>
  __inline ErrVal xGetPredBilinearlyChecked( const PixelPos &rcPosInUps,
                                             const BinaryFraction<t_uiAcc> &rcOffX,
                                             const BinaryFraction<t_uiAcc> &rcOffY,
                                             const BlockSize& rcBlockSize,
                                             const Block<DType> &rcIn );
};


template<class t_HandlePredictionIf, class t_PelIn>
template<IpolType t_IpolType, UInt t_uiAcc>
ErrVal ReferencePlanePackageAccessor<t_HandlePredictionIf,t_PelIn>::xGetPredBilinearly( const PixelPos &rcPosInUps,
                                                                                        const BinaryFraction<t_uiAcc> &rcOffX,
                                                                                        const BinaryFraction<t_uiAcc> &rcOffY,
                                                                                        const BlockSize& rcBlockSize,
                                                                                        const Block<DType> &rcIn )
{
  const bool bCheckX = rcPosInUps.getX() < 0 || (rcPosInUps.getX() + ( (rcBlockSize.getX() - 1) << m_rcRefPlaPa.getUpsShift() )) >= (rcIn.getSize().getX() - 1);
  const bool bCheckY = rcPosInUps.getY() < 0 || (rcPosInUps.getY() + ( (rcBlockSize.getY() - 1) << m_rcRefPlaPa.getUpsShift() )) >= (rcIn.getSize().getY() - 1);
  if( bCheckX )
  {
    if( bCheckY )
    {
      RNOK( (xGetPredBilinearlyChecked<t_IpolType, true, true>( rcPosInUps, rcOffX, rcOffY, rcBlockSize, rcIn )) );
    }
    else
    {
      RNOK( (xGetPredBilinearlyChecked<t_IpolType, true, false>( rcPosInUps, rcOffX, rcOffY, rcBlockSize, rcIn )) );
    }
  }
  else
  {
    if( bCheckY )
    {
      RNOK( (xGetPredBilinearlyChecked<t_IpolType, false, true>( rcPosInUps, rcOffX, rcOffY, rcBlockSize, rcIn )) );
    }
    else
    {
      RNOK( (xGetPredBilinearlyChecked<t_IpolType, false, false>( rcPosInUps, rcOffX, rcOffY, rcBlockSize, rcIn )) );
    }
  }
  return Err::m_nOK;
};

template<class t_HandlePredictionIf, class t_PelIn>
template<IpolType t_IpolType, bool t_bCheckX, bool t_bCheckY, UInt t_uiAcc>
ErrVal ReferencePlanePackageAccessor<t_HandlePredictionIf,t_PelIn>::xGetPredBilinearlyChecked( const PixelPos &rcPosInUps,
                                                                                               const BinaryFraction<t_uiAcc> &rcOffX,
                                                                                               const BinaryFraction<t_uiAcc> &rcOffY,
                                                                                               const BlockSize& rcBlockSize,
                                                                                               const Block<DType> &rcIn )
{
  typedef Int64 IntType;

  CAOF( 8 * sizeof( DType ) + 2 * t_uiAcc < 8 * sizeof( IntType ) );

  const int iAddInStep = 1 << m_rcRefPlaPa.getUpsShift();
  for( int yIn = rcPosInUps.getY(), yOut = 0; yOut < rcBlockSize.getY(); yOut++, yIn += iAddInStep )
  {
    int yInChecked = yIn;
    int yOff = 1;
    if( t_bCheckY )
    {
      if( yInChecked < 0 )
      {
        yInChecked = 0;
      }
      else if( yInChecked >= int( rcIn.getSize().getY() ) - 1 )
      {
        yInChecked = rcIn.getSize().getY() - 1;
        yOff = 0;
      }
    }

    t_HandlePredictionIf::setActiveLine( yOut );

    const DType* pcRow0 = &rcIn.getEl( PixelPos( 0, yInChecked ) );
    const DType* pcRow1;
    if( t_IpolType != IPOLTYPE_COPY && t_IpolType != IPOLTYPE_HOR )
    {
      pcRow1 = &rcIn.getEl( PixelPos( 0, yInChecked+yOff ) );
    }

    for( int xIn = rcPosInUps.getX(), xOut = 0; xOut < rcBlockSize.getX(); xOut++, xIn += iAddInStep )
    {
      int xInChecked = xIn;
      int xOff = 1;
      if( t_bCheckX )
      {
        if( xInChecked < 0 )
        {
          xInChecked = 0;
        }
        else if( xInChecked >= int( rcIn.getSize().getX() ) - 1 )
        {
          xInChecked = rcIn.getSize().getX() - 1;
          xOff = 0;
        }
      }

      DType cOut = 0;
      if( t_IpolType == IPOLTYPE_COPY )
      {
        cOut = DType( pcRow0[ xInChecked ] );
      }
      else if( t_IpolType == IPOLTYPE_HOR )
      {
        const IntType iLeft   = IntType( pcRow0[ xInChecked ] );
        const IntType iRight  = IntType( pcRow0[ xInChecked + xOff] );
        IntType iValue  = ( (IntType(1)<<t_uiAcc) - IntType( rcOffX.getNumerator() ) ) * iLeft + IntType( rcOffX.getNumerator() ) * iRight;
        cOut            = DType( ( iValue + (IntType(1)<<(t_uiAcc-1)) ) >> t_uiAcc );
      }
      else if( t_IpolType == IPOLTYPE_VER )
      {
        const IntType iTop    = IntType( pcRow0[ xInChecked ] );
        const IntType iBottom = IntType( pcRow1[ xInChecked ] );
        IntType iValue  = ( (IntType(1)<<t_uiAcc) - IntType( rcOffY.getNumerator() ) ) * iTop + IntType( rcOffY.getNumerator() ) * iBottom;
        cOut            = DType( ( iValue + (IntType(1)<<(t_uiAcc-1)) ) >> t_uiAcc );
      }
      else if( t_IpolType == IPOLTYPE_BILIN )
      {
        const IntType iTopLeft      = IntType( pcRow0[ xInChecked ] );
        const IntType iTopRight     = IntType( pcRow0[ xInChecked + xOff] );
        const IntType iBottomLeft   = IntType( pcRow1[ xInChecked ] );
        const IntType iBottomRight  = IntType( pcRow1[ xInChecked + xOff] );
        IntType iLeft         = ( (IntType(1)<<t_uiAcc) - IntType( rcOffY.getNumerator() ) ) * iTopLeft  + IntType( rcOffY.getNumerator() ) * iBottomLeft;
        IntType iRight        = ( (IntType(1)<<t_uiAcc) - IntType( rcOffY.getNumerator() ) ) * iTopRight + IntType( rcOffY.getNumerator() ) * iBottomRight;
        IntType iValue        = ( (IntType(1)<<t_uiAcc) - IntType( rcOffX.getNumerator() ) ) * iLeft     + IntType( rcOffX.getNumerator() ) * iRight;
        cOut                  = DType( ( iValue + (IntType(1)<<(2*t_uiAcc-1)) ) >> (2*t_uiAcc) );
      }
      /* ----- for testing -----
      else if( t_IpolType == IPOLTYPE_LIN_UPRIGHT )
      {
        ROF( ( rcOffX.getNumerator() << 1 ) == ( 1 << t_uiAcc ) );
        ROF( ( rcOffY.getNumerator() << 1 ) == ( 1 << t_uiAcc ) );
        IntType iTopRight     = IntType( rcIn.getEl( PixelPos( xInChecked + xOff, yInChecked        ) ) );
        IntType iBottomLeft   = IntType( rcIn.getEl( PixelPos( xInChecked,        yInChecked + yOff ) ) );
        cOut                  = DType( ( iTopRight + iBottomLeft + IntType(1) ) >> 1 );
      }
      else if( t_IpolType == IPOLTYPE_LIN_UPLEFT )
      {
        ROF( ( rcOffX.getNumerator() << 1 ) == ( 1 << t_uiAcc ) );
        ROF( ( rcOffY.getNumerator() << 1 ) == ( 1 << t_uiAcc ) );
        IntType iTopLeft      = IntType( rcIn.getEl( PixelPos( xInChecked,        yInChecked        ) ) );
        IntType iBottomRight  = IntType( rcIn.getEl( PixelPos( xInChecked + xOff, yInChecked + yOff ) ) );
        cOut                  = DType( ( iTopLeft + iBottomRight + IntType(1) ) >> 1 );
      }
      */
      else
      {
        ROT(1);
      }

      t_HandlePredictionIf::setPredictionSample( xOut, cOut );
    }
  }
  return Err::m_nOK;
}


template<class t_HandlePredictionIf, class t_PelIn>
ErrVal ReferencePlanePackageAccessor<t_HandlePredictionIf,t_PelIn>::getPlanePrediction( UInt uiPlane, const BlockSize &rcBlockSize, const MotionVector &rcMotionPositionVector )
{
  t_HandlePredictionIf::setActivePlane( uiPlane );

  PixelPos cPosInUps;
  const UInt uiMaxAccOverrun = 15;
  const int iChromaSubsamplingPrecisionIncrease = 1;
  ROF( int( uiMaxAccOverrun ) >= rcMotionPositionVector.getPrecision() + iChromaSubsamplingPrecisionIncrease - int( m_rcRefPlaPa.getUpsShift() ) );

  BinaryFraction<uiMaxAccOverrun> cOffX, cOffY;
  RNOK( rcMotionPositionVector.plusPrimResPos( m_rcRefPlaPa.getOrigin() ).getNearestTopLeftPosAndOffset( m_rcRefPlaPa.getStats().getSubsampling()[uiPlane], cPosInUps, cOffX, cOffY ) );

#if FAST_MC
  const ReferencePicture<Pel>::UpsMethod eUpsMethod = m_rcRefPlaPa.getRefPic().getUpsMethod();
  if( ReferencePicture<Pel>::CUBIC_MOMS_FIXED_MC == eUpsMethod || ReferencePicture<Pel>::CUBIC_SPLINE_FIXED_MC == eUpsMethod)
  {
    const InterpolationIf<t_PelIn, DType>* pcInterpolationIf = m_rcRefPlaPa.getRefPic().getInterpolationIf();

    UInt uiThePlaneINeed = uiPlane;
    if( m_rcRefPlaPa.getPicBlock().getStats().getNumPlanes() == 2 )
    {
      uiThePlaneINeed++;
    }

    AllocBlock<short> cTempBlock( rcBlockSize );
    pcInterpolationIf->interpolateBlock( cTempBlock, uiThePlaneINeed, cPosInUps, cOffX, cOffY );

    // let's make or own fast copy
    // 
    {
      const short* pSrc = &cTempBlock.getEl(0);
      const int iSrcStride = cTempBlock.getStride();
      const int iHeight = rcBlockSize.getY();
      const int iWidth = rcBlockSize.getX();
      for( int y = 0; y < iHeight; y++ )
      {
        t_HandlePredictionIf::setActiveLine( y );
        for( int x = 0; x < iWidth; x++ )
        {
          t_HandlePredictionIf::setPredictionSample( x, pSrc[x] );
        }
        pSrc += iSrcStride;
      }
    }
    return Err::m_nOK;
  }
#endif
  const Block<DType> &rcSrc = m_rcRefPlaPa.getPicBlock().getPlane( uiPlane );


  if( BinaryFraction<uiMaxAccOverrun>( 0 ) == cOffX )
  {
    if( BinaryFraction<uiMaxAccOverrun>( 0 ) == cOffY )
    {
      RNOK( ( xGetPredBilinearly<IPOLTYPE_COPY >( cPosInUps, cOffX, cOffY, rcBlockSize, rcSrc ) ) );
    }
    else
    {
      RNOK( ( xGetPredBilinearly<IPOLTYPE_VER  >( cPosInUps, cOffX, cOffY, rcBlockSize, rcSrc ) ) );
    }
  }
  else
  {
    if( BinaryFraction<uiMaxAccOverrun>( 0 ) == cOffY )
    {
      RNOK( ( xGetPredBilinearly<IPOLTYPE_HOR  >( cPosInUps, cOffX, cOffY, rcBlockSize, rcSrc ) ) );
    }
    else
    {
      RNOK( ( xGetPredBilinearly<IPOLTYPE_BILIN>( cPosInUps, cOffX, cOffY, rcBlockSize, rcSrc )) );
    }
  }

  return Err::m_nOK;
}


template<class t_HandlePredictionIf, class t_PelIn>
ErrVal ReferencePlanePackageAccessor<t_HandlePredictionIf,t_PelIn>::getPrediction( const PicBlockStats &rcPicBlockStats, const MotionVector &rcMotionPositionVector )
{
  for( UInt uiPl = 0; uiPl < m_rcRefPlaPa.getStats().getNumPlanes(); uiPl++ )
  {
    RNOK( getPlanePrediction( uiPl, rcPicBlockStats.getPrimBlockSize() * rcPicBlockStats.getSubsampling()[uiPl], rcMotionPositionVector ) );
  }

  return Err::m_nOK;
}

template<class t_HandlePredictionIf, class t_PelIn>
ErrVal ReferencePlanePackageAccessor<t_HandlePredictionIf,t_PelIn>::getUpsampledBlock( PicBlock< int > &rcPicBlkDst, const MotionVector &rcTopLeftPos )
{
  for( UInt uiPlane = 0; uiPlane < rcPicBlkDst.getStats().getNumPlanes(); uiPlane++ )
  {
    const Subsampling   cAddUpScale   = rcPicBlkDst.getStats().getSubsampling()[ uiPlane ] / m_rcRefPlaPa.getPicBlock().getStats().getSubsampling()[ uiPlane ];
    const int           iAddLog2StepX = (int)m_rcRefPlaPa.getUpsShift() - cAddUpScale.getX().getDownShift();  ROF( iAddLog2StepX >= 0 );
    const int           iAddLog2StepY = (int)m_rcRefPlaPa.getUpsShift() - cAddUpScale.getX().getDownShift();  ROF( iAddLog2StepY >= 0 );
    const Block<DType>& rcBlkSrc      = m_rcRefPlaPa.getPicBlock().getPlane( uiPlane );
    const Block<int>&   rcBlkDst      = rcPicBlkDst               .getPlane( uiPlane );

    if( rcTopLeftPos.getPrecision() < 0 )
    {
      ROT( 1 ); // currently disabled ===> CURRENT VERSION DOES NOT WORK CORRECTLY
      const int       iUpsDist    = rcTopLeftPos.getPrecision() - int( m_rcRefPlaPa.getUpsShift() );
      const UInt      uiLeftShift = -iUpsDist;
      const int       iStep       = 1 << uiLeftShift;
      const PixelPos  cOriginOff  ( m_rcRefPlaPa.getOrigin() * m_rcRefPlaPa.getStats().getSubsampling()[ uiPlane ] );
      const PixelPos  cTopLeftOff ( rcTopLeftPos.getX() << uiLeftShift, rcTopLeftPos.getY() << uiLeftShift );
      const PixelPos  cTotalOff   ( cOriginOff + cTopLeftOff );
      const int       iLastYSrc   = rcBlkSrc.getSize().getY() - 1;
      const int       iLastXSrc   = rcBlkSrc.getSize().getX() - 1;
      const int       iLastYDst   = rcBlkDst.getSize().getY();
      const int       iLastXDst   = rcBlkDst.getSize().getX();
      for( int y = 0; y < iLastYDst; y++ )
      {
        const int iYClip = gMinMax( y * iStep + cTotalOff.getY(), 0, iLastYSrc );
        for( int x = 0; x < iLastXDst; x++ )
        {
          const int iXClip = gMinMax( x * iStep + cTotalOff.getX(), 0, iLastXSrc );
          rcBlkDst.getEl( PixelPos( x, y ) ) = rcBlkSrc.getEl( PixelPos( iXClip, iYClip ) );
        }
      }
    }
    else
    {
      const int iShiftX = rcTopLeftPos.getPrecision() + iAddLog2StepX;
      const int iShiftY = rcTopLeftPos.getPrecision() + iAddLog2StepY;
      const int iStepX  = 1 << iShiftX;
      const int iStepY  = 1 << iShiftY;
      const int iXEnd   = gMin( iStepX, rcBlkDst.getSize().getX() );
      const int iYEnd   = gMin( iStepY, rcBlkDst.getSize().getY() );
      const int iAddX   = iStepX - 1;
      const int iAddY   = iStepY - 1;
      // fill the destination block polyphasewise
      for( int y = 0; y < iYEnd; y++ )
      {
        for( int x = 0; x < iXEnd; x++ )
        {
          Block<int>    cBlkDstCurr     ( PixelPos( x, y ), rcBlkDst.getSize() - BlockSize( x, y ), rcBlkDst );
          BlockSize     cPolyPhaseSize  ( ( cBlkDstCurr.getSize().getX() + iAddX ) >> iShiftX, ( cBlkDstCurr.getSize().getY() + iAddY ) >> iShiftY );
          MotionVector  cOffset         = rcTopLeftPos + MotionVector( x, y, rcTopLeftPos.getPrecision() );
          t_HandlePredictionIf::setBlock( cBlkDstCurr, iStepX, iStepY );
          RNOK( getPlanePrediction      ( uiPlane, cPolyPhaseSize, cOffset ) );
        }
      }
    }
  }

  return Err::m_nOK;
}


template<class t_PelOut>
class AddPredictionToPicBlock
{
public:
  AddPredictionToPicBlock( const PicBlock<t_PelOut>& rcBlock ) : m_rcBlock( rcBlock ), m_pcActivePlane( NULL ) {}

protected:
  void setActivePlane( UInt uiPlane ) { m_pcActivePlane = &m_rcBlock.getPlane( uiPlane ); } 
  void setActiveLine(int y) { m_pcPelOut = &(m_pcActivePlane->getEl( PixelPos( 0, y ))); }

  void setPredictionSample( int x, t_PelOut cSample )
  {
    m_pcPelOut[x] += cSample;
  }

private:
public:
  const PicBlock<t_PelOut>& m_rcBlock;
  const Block<t_PelOut> *m_pcActivePlane;
  t_PelOut* m_pcPelOut;
};


template<class t_PelOut>
class SetPredictionToPicBlock
{
public:
  SetPredictionToPicBlock( const PicBlock<t_PelOut>& rcBlock ) : m_rcBlock( rcBlock ), m_pcActivePlane( NULL ) {}

protected:
  FORCEINLINE void setActivePlane( UInt uiPlane ) { m_pcActivePlane = &m_rcBlock.getPlane( uiPlane ); } 
  void setActiveLine(int y) { m_pcPelOut = &(m_pcActivePlane->getEl( PixelPos( 0, y ))); }

  void setPredictionSample( int x, t_PelOut cSample )
  {
    m_pcPelOut[x] = cSample;
  }

private:
public:
  const PicBlock<t_PelOut>& m_rcBlock;
  const Block<t_PelOut> *m_pcActivePlane;
  t_PelOut* m_pcPelOut;
};


template<class t_PelOut>
class SetPredictionToBlockWithStepSize
{
public:
  void setBlock( const Block<t_PelOut>& rcBlock, int iStepX, int iStepY )
  {
    m_cBlock = rcBlock;
    m_iStepX = iStepX;
    m_iStepY = iStepY;
  }

protected:
  void setActivePlane( UInt uiPlane ) {}
  void setActiveLine(int y) { m_pcPelOut = &(m_cBlock.getEl( PixelPos( 0, y * m_iStepY))); }

  void setPredictionSample( int x, t_PelOut cSample )
  {
    m_pcPelOut[x*m_iStepX] = cSample;
  }

private:
public:
  Block<t_PelOut> m_cBlock;
  int m_iStepX;
  int m_iStepY;
  t_PelOut* m_pcPelOut;
};


template<class t_PelOut>
class SetPredictionToBlock
{
public:
  void setBlock( const Block<t_PelOut>& rcBlock, int iStepX, int iStepY )
  {
    m_cBlock = rcBlock;
  }

protected:
  void setActivePlane( UInt uiPlane ) {}
  void setActiveLine(int y) { m_pcPelOut = &(m_cBlock.getEl( PixelPos( 0, y ))); }

  void setPredictionSample( int x, t_PelOut cSample )
  {
    m_pcPelOut[x] = cSample;
  }

private:
public:
  Block<t_PelOut> m_cBlock;
  t_PelOut* m_pcPelOut;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _REFERENCEPICTUREACCESSOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
