/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RESIDUALRECONSTRUCTOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RESIDUALRECONSTRUCTOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "LeafNodeIterator.h"
#include "PicBlock.h"
#include "SliceData.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN

class CodingState;

class ResidualReconstructor
  : public LeafNodeIterator
{
public:
  ResidualReconstructor( const PicBlock<Int> &rcResPic, const CodingState &rcCodingState, const PlanePackageParameters& rcPlaPaPars, const SliceData &rcSliceData )
    : LeafNodeIterator( rcSliceData.getQuadtreeGroupStructure() )
    , m_cResMng( rcSliceData.getQuadtreeGroupStructure().getLog2MaxBlockSize(), rcPlaPaPars.getIntegerPow2TrafoFlag() )
    , m_rcResPic( rcResPic )
    , m_rcCodingState( rcCodingState )
    , m_rcSlice( rcSliceData )
  {
  }

private:
  ErrVal xProcessLeafNode( const QTNavi &rcPos )
  {
    const PicBlock<Int> cDecPicBlock( rcPos.getPrimPixelPos(), rcPos.getPrimBlockSize(), m_rcResPic );
    AllocPicBlock<DPel> cTempBlock( cDecPicBlock.getStats() );
    const ResInfo &rcResInfo = *m_rcSlice.getResInfoTree().getT( rcPos );
    RNOK( m_cResMng.getCalc( rcResInfo.getResidualMode() ).reconstructResidual( rcResInfo, cTempBlock, false ) );
    cDecPicBlock += cTempBlock;

    return Err::m_nOK;
  }

  ResidualCalculatorManager m_cResMng;
  const PicBlock<Int> &m_rcResPic;
  const CodingState &m_rcCodingState;
  const SliceData &m_rcSlice;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _RESIDUALRECONSTRUCTOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
