/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SYNTAXELEMENTCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SYNTAXELEMENTCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "ContextDataDeriver.h"

#include "SyntaxElementCoderIf.h"
#include "SliceData.h"
#include "ContextInit.h"

#include "TraceCommon.h"
#include "IntraDC.h"
#include "IntraDirectional.h"
#include "InterTranslational.h"
#include "InterPlanePackage.h"
#include "ResTrafoPow2.h"
#include "ResTrafoSkip.h"
#include "InStreamIf.h"
#include "AccessUnit.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_BinProcessor, bool t_bIsEncoder>
class SyntaxElementCoder
  : public SyntaxElementCoderIf<t_bIsEncoder>
  , virtual protected TraceBase
  , public t_BinProcessor
{
protected:
  SyntaxElementCoder( const CtxAccess &rcCtxAccess,
                      const ContextArrayStartIndices &rcCtxArrayStartIndices,
                      const PlanePackageParameters& rcPlaPaPars,
                      TraceCommon *pcTrace,
                      typename t_BinProcessor::CtorPars &rcBinProcessorPars )
    : t_BinProcessor( rcBinProcessorPars )
    , m_cCtxAccess( rcCtxAccess )
    , m_cCtxArrayStartIndices( rcCtxArrayStartIndices )
  {
    TraceBase::setTraceCommonPtr( pcTrace );

    m_aaauiScanPattern.resize( rcPlaPaPars.getLog2MaxPow2TrafoBlkSize() + 1 );
    for( UInt uiLog2TSize = 0; uiLog2TSize < m_aaauiScanPattern.size(); uiLog2TSize++ )
    {
      m_aaauiScanPattern[ uiLog2TSize ].resize( 2 );
      TNOK( ResTrafoPow2::setScanPattern( m_aaauiScanPattern[ uiLog2TSize ][ 0 ], uiLog2TSize, false ) );
      TNOK( ResTrafoPow2::setScanPattern( m_aaauiScanPattern[ uiLog2TSize ][ 1 ], uiLog2TSize, true  ) );
    }
  }

  virtual ~SyntaxElementCoder() {}

public:
  ErrVal processPredData              ( const QTNavi &rcPos,
                                        const PlanePackageParameters& rcPlaPaPars,
                                        const AccessUnitParameters &rcAUPars,
                                        const SliceData &rcSliceData,
                                        typename Const<t_bIsEncoder,PredSyntax*>::Type &rpcPredSyntax );
  ErrVal processResData               ( const QTNavi &rcPos,
                                        const PlanePackageParameters &rcPlaPaPars,
                                        const AccessUnitParameters &rcAUPars,
                                        const SliceData &rcCurrSlice,
                                        const PredSyntax* pcPredSyntax,
                                        typename Const<t_bIsEncoder, ResInfo*>::Type &rpcResInfo );
  ErrVal processResidualHead          ( const QTNavi &rcPos,
                                        const PlanePackageParameters &rcPlaPaPars,
                                        const SliceData &rcContextSlice,
                                        typename Const<t_bIsEncoder, ResInfoTrafoPow2>::Type &rcResInfo,
                                        bool& rbHasNonzeroCoeff );
  ErrVal processResTreeSubdivisionFlag( UInt uiLog2BlockSize,
                                        typename Const<t_bIsEncoder,bool>::Type &rbFlag );
  ErrVal processResidualLeaf          ( const QTNavi &rcSubLeafPos,
                                        const QTNavi &rcCtxNav,
                                        const PlanePackageParameters &rcPlaPaPars,
                                        const AccessUnitParameters &rcAUPars,
                                        const SliceData &rcSliceData,
                                        const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcTreeCurr,
                                        typename Const<t_bIsEncoder,ResInfoTrafoPow2Leaf>::Type &rcLeaf );
  ErrVal processTreeSubdivisionFlag   ( const QTNavi &rcPos,
                                        typename Const<t_bIsEncoder, bool>::Type& rbFlag );

  ErrVal processWienerTreeSubdivisionFlag ( const QTNavi &rcPos,
                                          typename Const<t_bIsEncoder, bool>::Type& rbFlag );
  ErrVal processWienerParameters      ( typename Const<t_bIsEncoder, WienerFilterParameters>::Type &rcWFParameters, UInt uiNumOfPlanes ) ;


  ErrVal processWienerFlag            ( typename Const<t_bIsEncoder, bool>::Type &rbFlag );

protected:
  FORCEINLINE ErrVal xProcessBin( typename Const<t_bIsEncoder,bool>::Type &bBin, UInt uiCtx );
  FORCEINLINE ErrVal xProcessEqProbBin( typename Const<t_bIsEncoder,bool>::Type &bBin, UInt uiCtx );

  virtual ErrVal xProcessUnarySymbol           ( typename Const<t_bIsEncoder,UInt>::Type &ruiSymbol,
                                                 UInt uiCtxStart,
                                                 UInt uiNumCtx,
                                                 UInt uiSymbolMax ) = 0;
  virtual ErrVal xProcessDQp                   ( UInt uiDQpCtx,
                                                 const CtxNames eContext,
                                                 bool bNonZeroCoeffsPresent,
                                                 const std::vector<int>& raiPredQp,
                                                 typename Const<t_bIsEncoder,std::vector<int> >::Type& raiQp ) = 0;
  virtual ErrVal xProcessSigLastPow2           ( const Block<typename Const<t_bIsEncoder,int>::Type> &rcBlock,
                                                 UInt uiPlane,
                                                 int iLog2BlockSize,
                                                 int iLog2PlaneBlockDim ) = 0;
  virtual ErrVal xProcessAbsGreaterOnePow2     ( typename Const<t_bIsEncoder,int>::Type &rcCoeff, UInt uiCtx ) = 0;
  virtual ErrVal xProcessCoeffLevelPow2        ( typename Const<t_bIsEncoder,int>::Type &rcCoeff,
                                                 UInt uiCtx,
                                                 UInt uiEpCtx ) = 0;
  virtual ErrVal xProcessSignPow2               ( typename Const<t_bIsEncoder,int>::Type &rcCoeff, UInt uiEpCtx ) = 0;
  virtual ErrVal xProcessPredMode               ( const bool bIsIntraOnlySlice,
                                                  bool &rbIsIntra,
                                                  typename Const<t_bIsEncoder,PredSyntax*>::Type& rpcPredSyntax,
                                                  const PredSyntax *pcLeftData,
                                                  const PredSyntax *pcTopData ) = 0;
  virtual ErrVal xProcessResMode                ( const QTNavi &rcNav,
                                                  bool bIsIntra,
                                                  bool bLeftIsAvailable,
                                                  bool bTopIsAvailable,
                                                  ResInfo::ResidualMode eLeftMode,
                                                  ResInfo::ResidualMode eTopMode,
                                                  const SliceData &rcCurrSlice,
                                                  const PlanePackageParameters &rcPlaPaPars,
                                                  typename Const<t_bIsEncoder, ResInfo*>::Type &rpcResInfo ) = 0;
  virtual ErrVal xProcessMvDComponent           ( typename Const<t_bIsEncoder,int>::Type &riMvDComponent,
                                                  UInt uiLocalCtx,
                                                  UInt uiCtxOffset ) = 0;
  virtual ErrVal xProcessInterPlaPaPredFlag     ( bool& rbIsInterPlaPaPred,
                                                  typename Const<t_bIsEncoder,PredSyntax*>::Type& rcPredSyntax,
                                                  const QTNavi &rcPos,
                                                  const PredSyntax* pcLeft,
                                                  const PredSyntax* pcTop,
                                                  const PlanePackageParameters& rcPlaPaPars,
                                                  const AccessUnitParameters& rcAUPars ) = 0;
  virtual ErrVal xInferInterPlaPaPredFlag       ( bool& rbIsInterPlaPaPred,
                                                  typename Const<t_bIsEncoder,PredSyntax*>::Type& rcPredSyntax,
                                                  const QTNavi &rcPos,
                                                  const PredSyntax* pcLeft,
                                                  const PredSyntax* pcTop,
                                                  const PlanePackageParameters& rcPlaPaPars,
                                                  const AccessUnitParameters& rcAUPars ) = 0;
  virtual ErrVal xProcessDirectModeFlag         ( bool& rbIsDirect,
                                                  typename Const<t_bIsEncoder,PredSyntax*>::Type& rcPredSyntax,
                                                  const QTNavi &rcPos,
                                                  const PredSyntax* pcLeft,
                                                  const PredSyntax* pcTop,
                                                  const PlanePackageParameters& rcPlaPaPars,
                                                  const AccessUnitParameters& rcAUPars ) = 0;
  virtual ErrVal xProcessNumMvs                 ( typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type& rcPredSyntax,
                                                  UInt uiContextStart,
                                                  UInt uiMaxNumContexts,
                                                  UInt uiMaxNumMvs ) = 0;
  virtual ErrVal xProcessRefIdxs                ( typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type& rcPredSyntax,
                                                  UInt uiRefListSize ) = 0;
  virtual ErrVal xProcessMvPrecision            ( const PlanePackageParameters &rcPlaPaPars,
                                                  typename Const<t_bIsEncoder,int>::Type &riPrecision ) = 0;
  virtual ErrVal xProcessMvPredictor            ( const QTNavi &rcPos,
                                                  typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type& rcPredSyntax,
                                                  const PlanePackageParameters& rcPlaPaPars,
                                                  const SliceData& rcSliceData,
                                                  const UInt uiIdx ) = 0;
  virtual ErrVal xProcessResTrafoSkip           ( const QTNavi &rcPos,
                                                  const SliceData &rcContextSlice,
                                                  UInt uiNumPlanes,
                                                  typename Const<t_bIsEncoder,ResInfoTrafoSkip>::Type &rcResInfo ) = 0;
  virtual ErrVal xProcessInterTranslational       ( const QTNavi &rcPos,
                                                  const PlanePackageParameters &rcPlaPaPars,
                                                  const AccessUnitParameters &rcAUPars,
                                                  typename Const<t_bIsEncoder,PredSyntax*>::Type& rpcPredSyntax,
                                                  const SliceData& rcSliceData,
                                                  const PredSyntax *pcLeft,
                                                  const PredSyntax *pcTop ) = 0;

  virtual ErrVal xProcessWienerParameters       ( typename Const<t_bIsEncoder, WienerFilterParameters>::Type &rcWFParameters ,  UInt uiNumOfPlanes ) = 0;

  // common functions
  ErrVal xGetIntraModeIdx( const PredSyntax& rcPredSyntax, UInt &ruiIntraModeIdx );

  ErrVal xProcessInterTranslational( const QTNavi &rcPos,
                                   const PlanePackageParameters &rcPlaPaPars,
                                   const AccessUnitParameters &rcAUPars,
                                   typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type& rpcPredSyntax,
                                   const SliceData& rcSliceData,
                                   const PredSyntaxInterTranslational *pcLeft,
                                   const PredSyntaxInterTranslational *pcTop );

  ErrVal xProcessMergeFlags     ( typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type& rcPredSyntax,
                                  const PredSyntaxInterTranslational* pcLeft,
                                  const PredSyntaxInterTranslational* pcTop,
                                  const PlanePackageParameters& rcPlaPaPars,
                                  const AccessUnitParameters&   rcAUPars );

  ErrVal xProcessLeafNodesPow2( const QTNavi &rcSubNav,
                                const QTNavi &rcCtxNav,
                                const SliceData &rcContextSlice,
                                const PlanePackageParameters &rcPlaPaPars,
                                const AccessUnitParameters &rcAUPars,
                                typename Const<t_bIsEncoder, QuadtreeGroup<ResInfoTrafoPow2Leaf*> >::Type &rcTree,
                                bool& rbHasNonzeroCoeff );

  void xGetNextZigZagScanPosition(  UInt &ruiScanPosX,
                                      UInt &ruiScanPosY,
                                      UInt const uiSizeMinus1 );
  void xGetPrevZigZagScanPosition(  UInt &ruiScanPosX,
                                      UInt &ruiScanPosY,
                                      UInt const uiSizeMinus1 );

  ErrVal xProcessResTrafoPow2         ( const QTNavi &rcPos,
                                        const SliceData &rcContextSlice,
                                        const PlanePackageParameters &rcPlaPaPars,
                                        const AccessUnitParameters &rcAUPars,
                                        typename Const<t_bIsEncoder,ResInfoTrafoPow2>::Type &rcResInfo );

  ErrVal xGetSkipModeCtx( bool bIsIntra,
                          bool bLeftIsAvailable,
                          bool bTopIsAvailable,
                          ResInfo::ResidualMode eLeftMode,
                          ResInfo::ResidualMode eTopMode,
                          UInt &ruiCtx );


  ErrVal xProcessWienerFlagForQuadtreeUse( typename Const<t_bIsEncoder, bool>::Type &rbFlag );
  ErrVal xProcessWienerPlaneSet( typename Const<t_bIsEncoder, UInt>::Type &ruiParameterSet ) ;
  ErrVal xProcessWienerFilterLength( typename Const<t_bIsEncoder, UInt>::Type &ruiFilterLength ) ;
  ErrVal xProcessWienerFilterType( typename Const<t_bIsEncoder, UInt>::Type &ruiFilterType ) ;
  ErrVal xProcessWienerFilterCoeffs( typename Const<t_bIsEncoder, UInt>::Type &raiCoefficients, const UInt &ruiNumberOfCoefficients ) ;
  ErrVal xProcessWienerBlockSize( typename Const<t_bIsEncoder, UInt>::Type &ruiBlockMax, typename Const<t_bIsEncoder, UInt>::Type &ruiBlockMin  );
  ErrVal xProcessWienerSeparableFlag( typename Const<t_bIsEncoder, bool>::Type &rbFlag );
  // mvd functions
  UInt xGetMvDLocalCtx( UInt uiAbsSum ) { return uiAbsSum < 3 ? 0 : uiAbsSum > 32 ? 2 : 1; }

  static const UInt m_uiNumMvDUnaryCtxPerComponent = 4;
  const CtxAccess m_cCtxAccess;
  bool m_bActivateTraceProcessBin;
  const ContextArrayStartIndices m_cCtxArrayStartIndices;
  std::vector<std::vector<std::vector<UInt> > > m_aaauiScanPattern;
};


template<class t_BinProcessor, bool t_bIsEncoder>
FORCEINLINE ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xProcessBin( typename Const<t_bIsEncoder,bool>::Type &bBin, UInt uiCtx )
{
  if( t_bIsEncoder )
  {
    CtxBin cCtxBin( uiCtx, bBin );
    RNOK(( t_BinProcessor::TypeCall::pushBack( cCtxBin ) ));
  }
  else
  {
    CtxBin cCtxBin( uiCtx );
    RNOK(( t_BinProcessor::TypeCall::pushBack( cCtxBin ) ));
    getIfNotConst( bBin ) = cCtxBin.getBin();
  }
  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
FORCEINLINE ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xProcessEqProbBin( typename Const<t_bIsEncoder,bool>::Type &bBin, UInt uiCtx )
{
#if COLLECT_CABAC_BITS
  if( t_bIsEncoder )
  {
    EPCtxBin cEPCtxBin( bBin, uiCtx );
    return t_BinProcessor::BoolCall::pushBack( cEPCtxBin );
  }
  else
  {
    EPCtxBin cEPCtxBin( false, uiCtx );
    RNOK( t_BinProcessor::BoolCall::pushBack( cEPCtxBin ) );
    getIfNotConst( bBin ) = cEPCtxBin.m_bBin;
  }
  return Err::m_nOK;
#else
  return t_BinProcessor::BoolCall::pushBack( bBin );
#endif
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processTreeSubdivisionFlag( const QTNavi &rcPos, typename Const<t_bIsEncoder, bool>::Type &rbFlag )
{
  if( ! rcPos.isMaxLayer() )
  {
    RNOK( xProcessBin( rbFlag, m_cCtxAccess.getContextOffset( CTX_START_TREE_SUBDIV, rcPos.getLayer() ) ) );

    TRACE_NUMBER;
    TRACE_TO_STR( "is_not_leaf_node_flag_pred_tree: " );
    TRACE_TO_STR( rbFlag );
    TRACE_TO_STR( "(" );
    TRACE_TO_STR( rcPos.getPrimPixelPos().getX() );
    TRACE_TO_STR( ", " );
    TRACE_TO_STR( rcPos.getPrimPixelPos().getY() );
    TRACE_TO_STR( ")" );
    TRACE_TO_STR( std::endl );
  }
  else
  {
    ROT( rbFlag );
  }

  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processWienerTreeSubdivisionFlag( const QTNavi &rcPos, typename Const<t_bIsEncoder, bool>::Type &rbFlag )
{
  if( ! rcPos.isMaxLayer() )
  {
    RNOK( xProcessBin( rbFlag, m_cCtxAccess.getContextOffset( CTX_START_WIENER_TREE_SUBDIV, rcPos.getLayer() ) ) );

    TRACE_NUMBER;
    TRACE_TO_STR( "is_not_leaf_node_flag_lf_tree: " );
    TRACE_TO_STR( rbFlag );
    TRACE_TO_STR( "(" );
    TRACE_TO_STR( rcPos.getPrimPixelPos().getX() );
    TRACE_TO_STR( ", " );
    TRACE_TO_STR( rcPos.getPrimPixelPos().getY() );
    TRACE_TO_STR( ")" );
    TRACE_TO_STR( std::endl );
  }
  else
  {
    ROT( rbFlag );
  }

  return Err::m_nOK;
}

template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor,t_bIsEncoder>::xProcessWienerSeparableFlag( typename Const<t_bIsEncoder, bool>::Type &rbFlag )
{
  RNOK( xProcessBin( rbFlag, m_cCtxAccess.getContextOffset( CTX_WIENER_SEPARABLE_FLAG ) ) );
  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processWienerFlag( typename Const<t_bIsEncoder, bool>::Type &rbFlag )
{
  RNOK( xProcessBin( rbFlag, m_cCtxAccess.getContextOffset( CTX_WIENER_FLAG ) ) );

  TRACE_NUMBER;
  TRACE_TO_STR( "quadtree_based_inloop_filter_mode_map[ y0 ][ x0 ]: " );
  TRACE_TO_STR( rbFlag );
  TRACE_TO_STR( std::endl   );

  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor,t_bIsEncoder>::xProcessWienerFlagForQuadtreeUse( typename Const<t_bIsEncoder, bool>::Type &rbFlag )
{
  RNOK( xProcessBin( rbFlag, m_cCtxAccess.getContextOffset( CTX_WIENER_QUADTREE_FLAG ) ) );
  return Err::m_nOK;
}


template<class t_BinProcessor,bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor,t_bIsEncoder>::processWienerParameters( typename Const<t_bIsEncoder, WienerFilterParameters>::Type &rcWFParameters, UInt uiNumOfPlanes )
{
  RNOK( xProcessWienerParameters( rcWFParameters , uiNumOfPlanes ) );
    return Err::m_nOK ;
}


template<class t_BinProcessor,bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor,t_bIsEncoder>::xProcessWienerPlaneSet( typename Const<t_bIsEncoder, UInt>::Type &ruiParameterSet)
{
  UInt uiIdxOfFilterTypes = 3;
  xProcessUnarySymbol( ruiParameterSet, m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_SET_NUM), 1, uiIdxOfFilterTypes  );
  return Err::m_nOK ;
}


template<class t_BinProcessor,bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor,t_bIsEncoder>::xProcessWienerFilterType( typename Const<t_bIsEncoder, UInt>::Type &ruiFilterType )
{
  UInt uiIdxOfFilterTypes = 3;
  xProcessUnarySymbol( ruiFilterType, m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_TYPE), 1, uiIdxOfFilterTypes  );
  return Err::m_nOK;
}


template<class t_BinProcessor,bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor,t_bIsEncoder>::xProcessWienerBlockSize( typename Const<t_bIsEncoder, UInt>::Type &ruiBlockMax, typename Const<t_bIsEncoder, UInt>::Type &ruiBlockMin  )
{
  UInt uiMaxValue = 8;
  xProcessUnarySymbol( ruiBlockMax, m_cCtxAccess.getContextOffset(CTX_WIENER_MAX_BLOCK), 1, uiMaxValue  );
  xProcessUnarySymbol( ruiBlockMin, m_cCtxAccess.getContextOffset(CTX_WIENER_MIN_BLOCK), 1, uiMaxValue  );
  return Err::m_nOK ;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xProcessResTrafoPow2( const QTNavi &rcPos,
                                                                               const SliceData &rcContextSlice,
                                                                               const PlanePackageParameters &rcPlaPaPars,
                                                                               const AccessUnitParameters &rcAUPars,
                                                                               typename Const<t_bIsEncoder,ResInfoTrafoPow2>::Type &rcResInfo )
{
  bool bHasNonzeroCoeff = false;
  // recursively process leaf nodes
  for( QTNavi cNav = rcResInfo.getTree().getTreeStructure().getFirstQuadtreeRoot(); !cNav.liesOutside(); cNav = rcResInfo.getTree().getTreeStructure().getNextQuadtreeRoot( cNav ) )
  {
    RNOK( xProcessLeafNodesPow2( cNav, rcPos, rcContextSlice, rcPlaPaPars, rcAUPars, rcResInfo.getTree(), bHasNonzeroCoeff ) );
  }
  // process plane QPs
  RNOK( processResidualHead( rcPos, rcPlaPaPars, rcContextSlice, rcResInfo, bHasNonzeroCoeff ) );
  return Err::m_nOK;
}

template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xProcessLeafNodesPow2( const QTNavi &rcSubNav,
                                                                                const QTNavi &rcCtxNav,
                                                                                const SliceData &rcContextSlice,
                                                                                const PlanePackageParameters &rcPlaPaPars,
                                                                                const AccessUnitParameters &rcAUPars,
                                                                                typename Const<t_bIsEncoder, QuadtreeGroup<ResInfoTrafoPow2Leaf*> >::Type &rcTree,
                                                                                bool& rbHasNonzeroCoeff )
{
  if ( rcSubNav.liesOutside() )
  {
    // for blocks outside the picture, do not process anything
    return Err::m_nOK;
  }
  bool bSubdiv = false;
  if( t_bIsEncoder )
  {
    bSubdiv = rcSubNav.isInternalNode();
  }

  RNOK( processResTreeSubdivisionFlag( rcSubNav.getLog2BlockSize(), bSubdiv ) );

  if ( bSubdiv )
  {
    TRACE_INDENT_INC;
    if( !t_bIsEncoder )
    {
      bool bSuccess = false;
      RNOK( getIfNotConst( rcTree.getTreeStructure() ).appendChildNodes( rcSubNav, bSuccess ) );
      ROF( bSuccess );
    }
    for ( UInt uiChildIndex=0; uiChildIndex < 4; uiChildIndex++ )
    {
      RNOK( xProcessLeafNodesPow2( rcSubNav.getChild( uiChildIndex ), rcCtxNav, rcContextSlice, rcPlaPaPars, rcAUPars, rcTree, rbHasNonzeroCoeff ) );
    }
    TRACE_INDENT_DEC;
  }
  else
  {
    if( !t_bIsEncoder )
    {
      /* allocate memory for leaf */
      AOF( rcSubNav.isLeafNode() );
      AOT( rcTree.getT( rcSubNav ) );
      const PrimBlockSize cPaddedSize( 1 << rcSubNav.getLog2BlockSize(), 1 << rcSubNav.getLog2BlockSize() );
      ResInfoTrafoPow2Leaf* pLeaf = new ResInfoTrafoPow2Leaf( PicBlockStats( cPaddedSize, rcPlaPaPars.getSubsampling() ) );
      pLeaf->getCoeff().assign( 0 );

      getIfNotConst( rcTree.getT( rcSubNav ) ) = pLeaf;
    }

    RNOK( processResidualLeaf( rcSubNav, rcCtxNav, rcPlaPaPars, rcAUPars, rcContextSlice, rcTree, *rcTree.getT( rcSubNav ) ) );

    // determine coded block flags for the later coding of delta qps
    for( UInt ui = 0; ui < rcPlaPaPars.getNumPlanes(); ui++ )
    {
      rbHasNonzeroCoeff = rbHasNonzeroCoeff || rcTree.getT( rcSubNav )->getCbf( ui );
    }
  }

  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processResTreeSubdivisionFlag( UInt uiLog2BlockSize,
                                                                                        typename Const<t_bIsEncoder,bool>::Type &rbFlag )
{
  UInt uiUsedCtx = uiLog2BlockSize;
  RNOK( xProcessBin( rbFlag, m_cCtxAccess.getContextOffset( CTX_START_POW2_SUBDIV, uiUsedCtx ) ) );
  
  TRACE_NUMBER;
  TRACE_TO_STR( "is_not_leaf_node_flag_res_tree: " );
  TRACE_TO_STR( rbFlag );
  TRACE_TO_STR( "(" );
  TRACE_TO_STR( uiUsedCtx );
  TRACE_TO_STR( ")" );
  TRACE_TO_STR( std::endl );

  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processResidualHead( const QTNavi &rcPos,
                                                                              const PlanePackageParameters &rcPlaPaPars,
                                                                              const SliceData &rcContextSlice,
                                                                              typename Const<t_bIsEncoder, ResInfoTrafoPow2>::Type &rcResInfo,
                                                                              bool& rbHasNonzeroCoeff )
{
  std::vector<int>  aiPredQp;
  UInt              uiDQpCtx = 0;
  RNOK( rcContextSlice.derivePredQPs( rcPos, aiPredQp, &uiDQpCtx ) );
  RNOK( xProcessDQp( uiDQpCtx, CTX_START_POW2_DQP, rbHasNonzeroCoeff, aiPredQp, rcResInfo.getQps() ) );
  if( rbHasNonzeroCoeff )
  {
    TRACE_NUMBER;
    TRACE_TO_STR( "delta_qp (TrafoPow2): " );
    TRACE_TO_STR( rcResInfo.getQps()[0] - aiPredQp[0] );
    TRACE_TO_STR( std::endl   );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processResidualLeaf( const QTNavi &rcSubLeafPos,
                                                                              const QTNavi &rcCtxNav, 
                                                                              const PlanePackageParameters &rcPlaPaPars,
                                                                              const AccessUnitParameters &rcAUPars,
                                                                              const SliceData &rcSliceData,
                                                                              const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcTreeCurr,
                                                                              typename Const<t_bIsEncoder,ResInfoTrafoPow2Leaf>::Type &rcLeaf )
{
  int iLog2PrimaryBlockSize = rcSubLeafPos.getLog2BlockSize();
  UInt uiNumPlanes = rcLeaf.getCoeff().getStats().getNumPlanes();

  std::vector<UInt> auiPow2CbfCtx( uiNumPlanes, 0 );
  RNOK( ContextDataDeriver::derivePow2CbfCtx( rcSubLeafPos,
                                              rcCtxNav,
                                              rcAUPars,
                                              rcPlaPaPars,
                                              rcSliceData,
                                              rcTreeCurr,
                                              auiPow2CbfCtx ) );

  for( UInt uiPlane = 0; uiPlane < uiNumPlanes; uiPlane++ )
  {
    RNOK( xProcessBin( rcLeaf.getCbf( uiPlane ), m_cCtxAccess.getContextOffset( m_cCtxArrayStartIndices.m_acResCtx[uiPlane].m_uiResCtxStart + RES_CTX_START_POW2_CBF, auiPow2CbfCtx[uiPlane] ) ) );

    TRACE_NUMBER;
    TRACE_TO_STR( "coded_block_flag (TrafoPow2): " );
    TRACE_TO_STR( rcLeaf.getCbf( uiPlane ) );
    TRACE_TO_STR( std::endl );
    if( rcLeaf.getCbf( uiPlane ) )
    {
      const Block<typename Const<t_bIsEncoder, int>::Type > &rcCurr = rcLeaf.getCoeff().getPlane( uiPlane );
      const int iPlaneBlockDim = rcCurr.getSize().getX();
      ROF( iPlaneBlockDim == rcCurr.getSize().getY() );
      const int iLog2PlaneBlockDim = gCeilLog2( rcCurr.getSize().getX() );
      ROF( 1 << iLog2PlaneBlockDim == iPlaneBlockDim );
      RNOK( xProcessSigLastPow2( rcCurr, uiPlane, iLog2PrimaryBlockSize, iLog2PlaneBlockDim ) );

      // Encoding of the Levels
      int c1 = 1;
      int c2 = 0;
      //----- encode significant coefficients -----
      UInt uiSigNum = 0;
      const UInt uiSize       = rcCurr.getSize().getX();
      const UInt uiSizeMinus1 = uiSize-1;
      UInt uiScanPosX   = uiSizeMinus1;
      UInt uiScanPosY   = uiSizeMinus1;

      const int* piCur      = rcCurr.getPtr();
      const int iCurStride  = rcCurr.getStride();
      for ( UInt uiY = 0; uiY < uiSize; uiY++ )
      {
        for ( UInt uiX = 0; uiX < uiSize; uiX++ )
        {
          if( piCur[uiX] )
          {
            uiSigNum++;
          }
        }
        piCur += iCurStride;
      }

      if ( iLog2PrimaryBlockSize > 2 && uiSigNum > uiSize )
      {
        UInt uiBlkPosX = 0;
        UInt uiBlkPosY = 0;
        UInt uiPrevBlkNumOnes = 0;
        const UInt uiNumOfSets = 4;
        const UInt uiNum4x4Blk = uiSize * uiSize / 16;

        for ( UInt uiSubBlk = 0; uiSubBlk < uiNum4x4Blk; uiSubBlk++ )
        {
          UInt uiSetIdx = 0;
          UInt uiOffsetAbs = 0;
          UInt uiOffsetUnary = 0;

          UInt uiSubPosX = uiBlkPosX * 4;
          UInt uiSubPosY = uiBlkPosY * 4;

          uiSigNum = 0;
          piCur = &rcCurr.getEl( PixelPos( uiSubPosX, uiSubPosY ) );
          for ( UInt uiY = 0; uiY < 4; uiY++ )
          {
            for ( UInt uiX = 0; uiX < 4; uiX++ )
            {
              if( piCur[uiX] )
              {
                uiSigNum++;
              }
            }
            piCur += iCurStride;
          }

          if ( uiSigNum > 0 )
          {
            c1 = 1;
            c2 = 0;

            if ( uiBlkPosX == 0 && uiBlkPosY == 0)
            {
              uiSetIdx = uiNumOfSets + 1; // 5
            }
            else
            {
              uiSetIdx = uiPrevBlkNumOnes / uiNumOfSets + 1; // 1 - 5
              uiPrevBlkNumOnes = 0;
            }

            uiOffsetAbs = uiSetIdx * 5;
            uiOffsetUnary = uiSetIdx * 6;

            uiSubPosX = 3;
            uiSubPosY = 3;
            
            const UInt uiResCtxStart = m_cCtxArrayStartIndices.m_acResCtx[uiPlane].m_uiResCtxStart;

            for ( UInt ui = 0; ui < 16; ui++ )
            {
              typename Const<t_bIsEncoder, int>::Type &cCoeff = rcCurr.getEl( PixelPos( uiSubPosX + uiBlkPosX * 4, uiSubPosY + uiBlkPosY * 4 ) );

              if( cCoeff )
              {
                TRACE_NUMBER;
                TRACE_TO_STR( "LevelCoding: " );
                TRACE_TO_STR( std::endl );
                UInt uiCtx = gMin (c1, 4);
                RNOK( xProcessAbsGreaterOnePow2( cCoeff, m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_ABS_GREATER_ONE, uiCtx + uiOffsetAbs ) ) );

                if( abs( cCoeff ) > 1 )
                {
                  uiCtx  = gMin (c2,4);
                  RNOK( xProcessCoeffLevelPow2( cCoeff,
                                                m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_COEFF_LEVEL, 1 + uiCtx + uiOffsetUnary ),
                                                m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_COEFF_LEVEL ) ) );
                  c1     = 0;
                  c2++;
                  uiPrevBlkNumOnes++;
                }
                else if( c1 )
                {
                  c1++;
                }
                RNOK( xProcessSignPow2( cCoeff, m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_SIGN ) ) );

                TRACE_NUMBER;
                TRACE_TO_STR( "Coeff: " );
                TRACE_TO_STR( cCoeff );
                TRACE_TO_STR( std::endl   );
              }
              xGetPrevZigZagScanPosition( uiSubPosX, uiSubPosY, 3 );
            }
          }
          xGetNextZigZagScanPosition( uiBlkPosX, uiBlkPosY, uiSize / 4 - 1 );
        }
      }
      else
      {
        const UInt uiResCtxStart = m_cCtxArrayStartIndices.m_acResCtx[uiPlane].m_uiResCtxStart;
      for( UInt i = 0; i < uiSize*uiSize; i++ )
      {
        PixelPos cCoeffPos( uiScanPosX, uiScanPosY );
        typename Const<t_bIsEncoder, int>::Type &cCoeff = rcCurr.getEl( cCoeffPos );
        if( cCoeff )
        {
          TRACE_NUMBER;
          TRACE_TO_STR( "LevelCoding: " );
          TRACE_TO_STR( std::endl   );
          UInt uiCtx    = gMin (c1, 4); // evtl. mehr modelle
          RNOK( xProcessAbsGreaterOnePow2( cCoeff, m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_ABS_GREATER_ONE, uiCtx ) ) );

          if( abs( cCoeff ) > 1 )
          {
            uiCtx  = gMin (c2,4);
            RNOK( xProcessCoeffLevelPow2( cCoeff,
                                          m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_COEFF_LEVEL, 1 + uiCtx ),
                                          m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_COEFF_LEVEL ) ) );
            c1     = 0;
            c2++;
          }
          else if( c1 )
          {
            c1++;
          }
          RNOK( xProcessSignPow2( cCoeff, m_cCtxAccess.getContextOffset( uiResCtxStart + RES_CTX_START_POW2_SIGN ) ) );

          TRACE_NUMBER;
          TRACE_TO_STR( "Coeff: " );
          TRACE_TO_STR( cCoeff );
          TRACE_TO_STR( std::endl   );

        }
        xGetPrevZigZagScanPosition( uiScanPosX, uiScanPosY, uiSizeMinus1 );
      }
    } // end if block is coded
  } // end for all layers
}
  return Err::m_nOK;
}

template<class t_BinProcessor, bool t_bIsEncoder>
void SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xGetNextZigZagScanPosition( UInt &ruiScanPosX,
                                                                                     UInt &ruiScanPosY,
                                                                                     UInt const uiSizeMinus1 )
{
  // Go to next zig-zag scan position.
  // uiSizeMinus1 is the block size minus 1.
  // The block is expected to have the same power of two as SizeX and SizeY.
  if ( ruiScanPosX == uiSizeMinus1 )
  {
    if ( ( ruiScanPosY & 1 ) == 0 )
    {
      ruiScanPosX -= 1;
    }
    ruiScanPosY += 1;
  }
  else if ( ruiScanPosY == uiSizeMinus1 )
  {
    if ( ruiScanPosX & 1 )
    {
      ruiScanPosY -= 1;
    }
    ruiScanPosX += 1;
  }
  else if ( ruiScanPosY == 0 )
  {
    if ( ruiScanPosX & 1 )
    {
      ruiScanPosX -= 1;
      ruiScanPosY += 1;
    }
    else
    {
      ruiScanPosX += 1;
    }
  }
  else if ( ruiScanPosX == 0 )
  {
    if ( ruiScanPosY & 1 )
    {
      ruiScanPosY += 1;
    }
    else
    {
      ruiScanPosX += 1;
      ruiScanPosY -= 1;
    }
  }
  else if ( ( ruiScanPosX ^ ruiScanPosY ) & 1 )
  {
    ruiScanPosX -= 1;
    ruiScanPosY += 1;
  }
  else
  {
    ruiScanPosX += 1;
    ruiScanPosY -= 1;
  }
  
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xGetSkipModeCtx( bool bIsIntra,
                                                                          bool bLeftIsAvailable,
                                                                          bool bTopIsAvailable,
                                                                          ResInfo::ResidualMode eLeftMode,
                                                                          ResInfo::ResidualMode eTopMode,
                                                                          UInt &ruiCtx )
{
  if( bIsIntra )
  {
    ruiCtx = 4;
    return Err::m_nOK;
  }

  {
    bool bLeftMode = false;
    bool bTopMode = false;
    if( bLeftIsAvailable )
    {
      bLeftMode = eLeftMode == ResInfo::RES_TRAFO_SKIP;
    }
    if( bTopIsAvailable )
    {
      bTopMode = eTopMode == ResInfo::RES_TRAFO_SKIP;
    }
    if ( bLeftIsAvailable && bTopIsAvailable )
    {
      ruiCtx = ( bLeftMode ? 2 : 0 ) + ( bTopMode ? 1 : 0 );
    }
    else if ( bLeftIsAvailable )
    {
      ruiCtx = bLeftMode ? 3 : 0;
    }
    else if ( bTopIsAvailable )
    {
      ruiCtx = bTopMode ? 3 : 0;
    }
    else
    {
      ruiCtx = 1;
    }
  }
  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
void SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xGetPrevZigZagScanPosition( UInt &ruiScanPosX,
                                                                                     UInt &ruiScanPosY,
                                                                                     UInt const uiSizeMinus1 )
{
  // Go to previous zig-zag scan position.
  // uiSizeMinus1 is the block size minus 1.
  // The block is expected to have the same power of two as SizeX and SizeY.
  if( ruiScanPosX == 0 )
  {
    if( ruiScanPosY & 1 )
    {
      ruiScanPosX++;
    }
    ruiScanPosY--;
  }
  else if( ruiScanPosY == 0 )
  {
    if( ( ruiScanPosX & 1 ) == 0 )
    {
      ruiScanPosY++;
    }
    ruiScanPosX--;
  }
  else if( ruiScanPosY == uiSizeMinus1 )
  {
    if( ruiScanPosX & 1 )
    {
      ruiScanPosX--;
    }
    else
    {
      ruiScanPosX++;
      ruiScanPosY--;
    }
  }
  else if( ruiScanPosX == uiSizeMinus1 )
  {
    if( ruiScanPosY & 1 )
    {
      ruiScanPosY++;
      ruiScanPosX--;
    }
    else
    {
      ruiScanPosY--;
    }
  }
  else if ( ( ruiScanPosX ^ ruiScanPosY ) & 1 )
  {
    ruiScanPosX += 1;
    ruiScanPosY -= 1;
  }
  else
  {
    ruiScanPosX -= 1;
    ruiScanPosY += 1;
  }
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processResData( const QTNavi &rcPos,
                                                                         const PlanePackageParameters &rcPlaPaPars,
                                                                         const AccessUnitParameters &rcAUPars,
                                                                         const SliceData &rcCurrSlice,
                                                                         const PredSyntax* pcPredSyntax,
                                                                         typename Const<t_bIsEncoder, ResInfo*>::Type &rpcResInfo )
{
  bool bLeftIsLeaf = false, bTopIsLeaf = false;
  ResInfo::ResidualMode eLeftMode = ResInfo::ResidualMode(0), eTopMode = ResInfo::ResidualMode(0);
  // if mode is signaled, find context model index to use
  if( rcPlaPaPars.getNumQuantizers() > 1 )
  {
    const QuadtreeGroup<ResInfo*> &rcTree( rcCurrSlice.getResInfoTree() );
    const PrimPixelPos  cNavPos        = rcPos.getPrimPixelPos();
    const QTNavi cLeft = rcPos.findLeafNode( cNavPos + PrimPixelPos( -1,  0 ) );
    const QTNavi cTop  = rcPos.findLeafNode( cNavPos + PrimPixelPos(  0, -1 ) );
    bLeftIsLeaf = cLeft.isLeafNode() && cLeft.isParsedBefore( rcPos );
    bTopIsLeaf  = cTop .isLeafNode() && cTop.isParsedBefore( rcPos );
    eLeftMode = bLeftIsLeaf ? rcTree.getT( cLeft )->getResidualMode() : ResInfo::ResidualMode( 0 );
    eTopMode  = bTopIsLeaf  ? rcTree.getT( cTop  )->getResidualMode() : ResInfo::ResidualMode( 0 );
  }

  // The decoder creates the ResInfo object here.
  RNOK( xProcessResMode( rcPos, pcPredSyntax->isIntraCoded(), bLeftIsLeaf, bTopIsLeaf, eLeftMode, eTopMode, rcCurrSlice, rcPlaPaPars, rpcResInfo ) );

  const ResInfo::ResidualMode eResMode = rpcResInfo->getResidualMode();
  if ( eResMode == ResInfo::RES_TRAFO_POW2 )
  {
    ROF( rcPlaPaPars.getTransformMode() & ResInfo::RES_TRAFO_POW2 );
    TRACE_NUMBER;
    TRACE_TO_STR( "TRAFO_POW2 " );
    TRACE_TO_STR( std::endl );
  }
  else if ( eResMode == ResInfo::RES_TRAFO_SKIP )
  {
    ROF( rcPlaPaPars.getTransformMode() & ResInfo::RES_TRAFO_SKIP );
    TRACE_NUMBER;
    TRACE_TO_STR( "TRAFO_SKIP " );
    TRACE_TO_STR( std::endl );
  }
  else
  {
    ROT( 1 );
  }

  // process coefficients/parameters depending on mode
  if(eResMode == ResInfo::RES_TRAFO_POW2)
  {
    RNOK( xProcessResTrafoPow2( rcPos, rcCurrSlice, rcPlaPaPars, rcAUPars, *rpcResInfo->getResInfoTrafoPow2() ) );
  }
  else if(eResMode == ResInfo::RES_TRAFO_SKIP )
  {
    RNOK( xProcessResTrafoSkip( rcPos, rcCurrSlice, rcPlaPaPars.getNumPlanes(), *rpcResInfo->getResInfoTrafoSkip() ) );
  }
  else
  {
    TOT( 1 );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::processPredData( const QTNavi &rcPos,
                                                                          const PlanePackageParameters& rcPlaPaPars,
                                                                          const AccessUnitParameters &rcAUPars,
                                                                          const SliceData &rcSliceData,
                                                                          typename Const<t_bIsEncoder,PredSyntax*>::Type &rpcPredSyntax )
{
  const QuadtreeGroup<PredInfo*> &rcTree( rcSliceData.getPredInfoTree() );
  const PrimPixelPos  cNavPos        = rcPos.getPrimPixelPos();
  const QTNavi cLeftPos = rcPos.findLeafNode( cNavPos + PrimPixelPos( -1,  0 ) );
  const QTNavi cTopPos  = rcPos.findLeafNode( cNavPos + PrimPixelPos(  0, -1 ) );

  const PredSyntax* pcLeft = ( cLeftPos.isLeafNode() && cLeftPos.isParsedBefore( rcPos ) ? rcTree.getT( cLeftPos )->getPredSyntax() : 0 );
  const PredSyntax* pcTop  = ( cTopPos. isLeafNode() && cTopPos.isParsedBefore ( rcPos ) ? rcTree.getT( cTopPos  )->getPredSyntax() : 0 );

  bool   bIsDirect = false;
  RNOK ( xProcessDirectModeFlag( bIsDirect, rpcPredSyntax, rcPos, pcLeft, pcTop, rcPlaPaPars, rcAUPars ) );
  ROTRS( bIsDirect, Err::m_nOK );

  bool   bIsInterPlaPaPred = false;
  RNOK ( xProcessInterPlaPaPredFlag( bIsInterPlaPaPred, rpcPredSyntax, rcPos, pcLeft, pcTop, rcPlaPaPars, rcAUPars ) );
  ROTRS( bIsInterPlaPaPred, Err::m_nOK );

  bool   bIsIntra  = false;
  RNOK( xProcessPredMode( rcAUPars.getMaxMvsPerBlk() == 0, bIsIntra, rpcPredSyntax, pcLeft, pcTop ) );

  if( !bIsIntra )
  {
    bool   bIsInterPlaPaPred = false;
    RNOK ( xInferInterPlaPaPredFlag( bIsInterPlaPaPred, rpcPredSyntax, rcPos, pcLeft, pcTop, rcPlaPaPars, rcAUPars ) );
    ROTRS( bIsInterPlaPaPred, Err::m_nOK );

    RNOK ( xProcessInterTranslational( rcPos, rcPlaPaPars, rcAUPars, rpcPredSyntax, rcSliceData, pcLeft, pcTop ) );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xGetIntraModeIdx( const PredSyntax& rcPredSyntax, UInt& ruiIntraModeIdx )
{
  if( rcPredSyntax.getCodingMode() == PredInfo::INTRA_DC )
  {
    ruiIntraModeIdx = 2;
  }
  else if( rcPredSyntax.getCodingMode() == PredInfo::INTRA_PLANE )
  {
    ROT( 1 ); // signalling currently not implemented
//    ruiIntraModeIdx = 1;
  }
  else
  {
    ruiIntraModeIdx = rcPredSyntax.getPredSyntaxIntraDirectional()->getIntraDirection();
  }
  return Err::m_nOK;
}

template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xProcessMergeFlags( typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type&  rcPredSyntax,
                                                                             const PredSyntaxInterTranslational*                               pcLeft,
                                                                             const PredSyntaxInterTranslational*                               pcTop,
                                                                             const PlanePackageParameters&                                   rcPlaPaPars,
                                                                             const AccessUnitParameters&                                     rcAUPars )
{
  typename Const<t_bIsEncoder,bool>::Type& rbMerge( rcPredSyntax.getMergeFlag() );
  typename Const<t_bIsEncoder,bool>::Type& rbMergeLeft( rcPredSyntax.getMergeLeftFlag() );

  bool bInferMergeFlag = false;
  UInt uiCtx = 10000;
  if ( pcLeft == NULL && pcTop == NULL )
  {
    bInferMergeFlag = true;
  }
  else if ( pcLeft == NULL )
  {
    // only top inter block is available
    uiCtx = pcTop->getMergeFlag() ? 2 : 0;
  }
  else if( pcTop == NULL )
  {
    // only left inter block is available
    uiCtx = pcLeft->getMergeFlag() ? 2 : 0;
  }
  else
  {
    // both blocks are inter translational
    UInt  uiLeft  = pcLeft->getMergeFlag() ? 1 : 0;
    UInt  uiTop   = pcTop->getMergeFlag() ? 1 : 0;
    uiCtx   = uiLeft + uiTop;
  }
  
  if ( bInferMergeFlag )
  {
    if ( t_bIsEncoder )
    {
      AOT( rbMerge );
    }
    else
    {
      getIfNotConst( rbMerge ) = false;
    }
  }
  else
  {
    RNOK( xProcessBin( rbMerge, m_cCtxAccess.getContextOffset( CTX_START_MERGE, uiCtx ) ) );
  }

  TRACE_NUMBER;
  TRACE_TO_STR( "merge_flag: " );
  TRACE_TO_STR( rbMerge );
  TRACE_TO_STR( std::endl );

  if ( rbMerge )
  {
    if ( pcLeft && pcTop )
    {
      bool bSameVectors = PredInfoInterTranslational::compareMvAndRefPocVec( pcLeft->getPredInfoInterTranslational()->getMvAndRefPocVec(), pcTop->getPredInfoInterTranslational()->getMvAndRefPocVec() );
      if ( !bSameVectors )
      {
        // two different merge candidates exist. merge left flag is present
        UInt uiCtx = 3;
        RNOK( xProcessBin( rbMergeLeft, m_cCtxAccess.getContextOffset( CTX_START_MERGE, uiCtx ) ) );
        TRACE_NUMBER;
        TRACE_TO_STR( "merge_left_flag: " );
        TRACE_TO_STR( rbMergeLeft );
        TRACE_TO_STR( std::endl );
      }
    }
  }

  return Err::m_nOK;
}

template<class t_BinProcessor, bool t_bIsEncoder>
ErrVal SyntaxElementCoder<t_BinProcessor, t_bIsEncoder>::xProcessInterTranslational( const QTNavi &rcPos,
                                                                                   const PlanePackageParameters &rcPlaPaPars,
                                                                                   const AccessUnitParameters &rcAUPars,
                                                                                   typename Const<t_bIsEncoder,PredSyntaxInterTranslational>::Type &rcPredSyntax,
                                                                                   const SliceData &rcSliceData,
                                                                                   const PredSyntaxInterTranslational *pcLeft,
                                                                                   const PredSyntaxInterTranslational *pcTop )
{
  if( !rcAUPars.getEntropySliceFlag() )
  {
    RNOK ( xProcessMergeFlags( rcPredSyntax, pcLeft, pcTop, rcPlaPaPars, rcAUPars ) );
    ROTRS( rcPredSyntax.getMergeFlag(), Err::m_nOK );
  }

  // number of motion vectors
  RNOK( xProcessNumMvs( rcPredSyntax, m_cCtxAccess.getContextOffset( CTX_START_NUM_MVS ), m_cCtxAccess.getNumCtxElements( CTX_START_NUM_MVS ), rcAUPars.getMaxMvsPerBlk() ) );

  TRACE_NUMBER;
  TRACE_TO_STR( "num_mvs_minus1: " );
  TRACE_TO_STR( rcPredSyntax.getNumEntries() - 1 );
  TRACE_TO_STR( std::endl );

  // reference indices
  TRACE_NUMBER;
  TRACE_TO_STR( "Beginning RefIdx " );
  TRACE_TO_STR( std::endl );
  RNOK( xProcessRefIdxs( rcPredSyntax, rcAUPars.getRefListSize() ) );
  TRACE_NUMBER;
  TRACE_TO_STR( "Ending RefIdx " );
  TRACE_TO_STR( std::endl );

  for( UInt uiIdx = 0; uiIdx < rcPredSyntax.getNumEntries(); uiIdx++ )
  {
    // motion vector precision
    RNOK( xProcessMvPrecision( rcPlaPaPars, rcPredSyntax.getMvDiff( uiIdx ).getPrecision() ) );

    // motion vector predictor
    RNOK( xProcessMvPredictor( rcPos, rcPredSyntax, rcPlaPaPars, rcSliceData, uiIdx ) );
    
    // determine variables for context determination
    MotionVector cMvDiffLeft ( 0, 0, rcPredSyntax.getMvDiff( uiIdx ).getPrecision() ); // specify precision for motion vector differences used in context determination !!!!
    MotionVector cMvDiffAbove( 0, 0, rcPredSyntax.getMvDiff( uiIdx ).getPrecision() ); // specify precision for motion vector differences used in context determination !!!!
    if( pcLeft )
    {
      if ( pcLeft->getMergeFlag() )
      {
        // This block stores the merge information only.
        // Get syntax object of block that stores MV information for this block.
        const PrimPixelPos cRegionOrigin( pcLeft->getMVAddress() );
        pcLeft = rcSliceData.getPredInfoTree().getT( rcPos.findLeafNode( cRegionOrigin ) )->getPredSyntax()->getPredSyntaxInterTranslational();
      }
      RNOK( InterTranslational::getMvD( *pcLeft, rcPredSyntax.getRefIdx( uiIdx ), cMvDiffLeft  ) );
    }
    if( pcTop )
    {
      if ( pcTop->getMergeFlag() )
      {
        // This block stores the merge information only.
        // Get syntax object of block that stores MV information for this block.
        const PrimPixelPos cRegionOrigin( pcTop->getMVAddress() );
        pcTop = rcSliceData.getPredInfoTree().getT( rcPos.findLeafNode( cRegionOrigin ) )->getPredSyntax()->getPredSyntaxInterTranslational();
      }
      RNOK( InterTranslational::getMvD( *pcTop,  rcPredSyntax.getRefIdx( uiIdx ), cMvDiffAbove ) );
    }
    const UInt uiAbsSumHor = abs( cMvDiffAbove.getX() ) + abs( cMvDiffLeft.getX() );
    const UInt uiAbsSumVer = abs( cMvDiffAbove.getY() ) + abs( cMvDiffLeft.getY() );
    
    // horizontal motion vector difference
    RNOK( xProcessMvDComponent( rcPredSyntax.getMvDiff( uiIdx ).getX(), xGetMvDLocalCtx( uiAbsSumHor ),  1 ) );
    TRACE_NUMBER;
    TRACE_TO_STR( "mvd[i][0]: " );
    TRACE_TO_STR( rcPredSyntax.getMvDiff( uiIdx ).getX() );
    TRACE_TO_STR( " above " );
    TRACE_TO_STR( cMvDiffAbove.getX() );
    TRACE_TO_STR( " left " );
    TRACE_TO_STR( cMvDiffLeft .getX() );
    TRACE_TO_STR( std::endl );
    
    // vertical motion vector difference
    RNOK( xProcessMvDComponent( rcPredSyntax.getMvDiff( uiIdx ).getY(), xGetMvDLocalCtx( uiAbsSumVer ), 12 ) );
    TRACE_NUMBER;
    TRACE_TO_STR( "mvd[i][1]: " );
    TRACE_TO_STR( rcPredSyntax.getMvDiff( uiIdx ).getY() );
    TRACE_TO_STR( " above " );
    TRACE_TO_STR( cMvDiffAbove.getY() );
    TRACE_TO_STR( " left " );
    TRACE_TO_STR( cMvDiffLeft .getY() );
    TRACE_TO_STR( std::endl );
  }
  return Err::m_nOK;

}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _SYNTAXELEMENTCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
