/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_TRACECABAC_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_)
#define AFX_TRACECABAC_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_

#include "VideoCodecToolsLib/MSysLargeFile.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/CabacContextModel.h"
#include "VideoCommonLib/ContextInit.h"
#include "VideoCommonLib/ExtendedContextModels.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class TraceCabacCategorizerIf
{
protected:
  TraceCabacCategorizerIf() {}
  virtual ~TraceCabacCategorizerIf() {}

public:
  virtual UInt getCategoryIdx() const = 0;
  virtual UInt getNumCategories() const = 0;
  virtual std::string getCategoryName() const = 0;
};

class TraceCabacCategorizer
  : public TraceCabacCategorizerIf
{
public:
  TraceCabacCategorizer() : m_pcAUPars( NULL ) {}
  void setAUPars( const AccessUnitParameters &rcAUPars ) { m_pcAUPars = &rcAUPars; }

  UInt getCategoryIdx() const;
  UInt getNumCategories() const;
  std::string getCategoryName() const;

private:
  const AccessUnitParameters *m_pcAUPars;
};


template<class> class ProbabilityAllocator;

class BinsAndBits
{
public:
  BinsAndBits() : m_uiBins( 0 ), m_dBits( 0.0 ) {}

  UInt64 m_uiBins;
  double m_dBits;
};


class CabacTraceCategoryElement
{
public:
  CabacTraceCategoryElement() : m_dCatElementBits( 0.0 ), m_uiCatElementBins( 0 ) {}


  double m_dCatElementBits;
  UInt64 m_uiCatElementBins;
  std::string m_acCtxName;
  std::vector<BinsAndBits> m_acDetailedBinsAndBits;
};

class CabacTraceCategory
{
public:
  CabacTraceCategory() : m_dSum( 0.0 ), m_uiMergingStart( 0 ), m_uiMergingEnd( 0 ) {}

  std::vector<CabacTraceCategoryElement> m_acCategoryElements;
  double m_dSum;
  std::string m_acCatName;
  std::string m_acCtxIdentifyer;
  UInt m_uiMergingStart;
  UInt m_uiMergingEnd;
};


class TraceCabac
{
public:
  TraceCabac() : m_dSum( 0.0 ) {}

  template<class t_ContextModel>
  ErrVal storeCabacBits( const TraceCabacCategorizerIf &rcCat,
                         const std::string &racCtxIdentifyer,
                         const CtxAccess &rcCtxInit,
                         const ProbabilityAllocator<t_ContextModel> &rcProbAlloc );
  ErrVal traceStoredCabacBits( LargeFile &rcTraceFile );

  ErrVal traceStoredCabacBits( const std::string &rcFilename )
  {
    LargeFile cFile;
    RNOK( cFile.open( rcFilename, LargeFile::OM_WRITEONLY ) );
    std::cout << "Tracing stored CABAC bits..";
    RNOK( traceStoredCabacBits( cFile ) );
    RNOK( cFile.close() );
    std::cout << "Done." << std::endl;
    return Err::m_nOK;
  }

  static ErrVal processExactCabacBits( const std::vector<IndexedBin> &rcVec, const CabacContextModel &rcInitCCModel, double &rdBits );
  static ErrVal processSteadyStateBits( const std::vector<IndexedBin> &rcVec, const CabacContextModel &rcInitCCModel, double &rdBits );

private:

  UInt xGetMaxCtxNameLength() const;
  ErrVal xResizeCategoriesIfNecessary( UInt uiNumCategories );
  ErrVal xInitCategoryIfNecessary( UInt uiCategoryIdx, std::string cCategoryName, const CtxAccess &rcCtxInit );
  ErrVal xResizeCategoryElementIfNecessary( CabacTraceCategoryElement &rcCatElement, UInt uiNumCtxElementsDetailed );

  UInt xGetOneAfterLastNonemptyBinsAndBitsElement( const std::vector<BinsAndBits> &racBinsAndBits );
  UInt xGetNumNonemptyBinsAndBitsElements( const std::vector<BinsAndBits> &racBinsAndBits );

  std::vector<CabacTraceCategory> m_acTraceCats;
  double m_dSum;
};


template<class t_ContextModel>
ErrVal TraceCabac::storeCabacBits( const TraceCabacCategorizerIf &rcCat,
                                   const std::string &racCtxIdentifyer,
                                   const CtxAccess &rcCtxInit,
                                   const ProbabilityAllocator<t_ContextModel> &rcProbAlloc )
{
  RNOK( xResizeCategoriesIfNecessary( rcCat.getNumCategories() ) );
  RNOK( xInitCategoryIfNecessary( rcCat.getCategoryIdx(), rcCat.getCategoryName(), rcCtxInit ) );
  CabacTraceCategory &rcCatData = m_acTraceCats[rcCat.getCategoryIdx()];

  rcCatData.m_acCtxIdentifyer = racCtxIdentifyer;

  for( UInt ui = 0; ui < UInt( rcCtxInit.getNumCtxInitArrays() ); ui++ )
  {
    CabacTraceCategoryElement &rcCatElement = rcCatData.m_acCategoryElements[ui];
    const double dBits = rcProbAlloc.getBits( rcCtxInit, ui );
    rcCatElement.m_dCatElementBits += dBits;
    rcCatElement.m_uiCatElementBins += rcProbAlloc.getBins( rcCtxInit, ui );
    rcCatData.m_dSum += dBits;
    m_dSum += dBits;

    if( CheckCap<BitCounterContextModelCapIf, t_ContextModel>::m_bHasCap )
    {
      const UInt uiNumCtxElementsDetailed = rcCtxInit.getNumCtxElements( ui );
      RNOK( xResizeCategoryElementIfNecessary( rcCatElement, uiNumCtxElementsDetailed ) );

      for( UInt k = 0; k < uiNumCtxElementsDetailed; k++ )
      {
        BinsAndBits &rcBinsAndBits = rcCatElement.m_acDetailedBinsAndBits[k];

        const bool bHasCounter = CheckCap<BitCounterContextModelCapIf, t_ContextModel>::m_bHasCap;
        if( bHasCounter )
        {
          const BitCounterContextModelCapIf &rcCounterModelIf = TypeConvert<const BitCounterContextModelCapIf, bHasCounter>::getIfTrue( rcProbAlloc.getCtxDetailed( rcCtxInit, ui, k ) );
          rcBinsAndBits.m_uiBins += rcCounterModelIf.getBinsCabac();
          rcBinsAndBits.m_dBits += rcCounterModelIf.getBitsCabac();
        }
      }
    }
  }
  return Err::m_nOK;
}



// video namespace end
VIDEO_NAMESPACE_END


#endif // !defined(AFX_TRACECABAC_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_)
