/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_TRACECOMMON_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_)
#define AFX_TRACECOMMON_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_

#include "VideoCodecToolsLib/MSysLargeFile.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#define TRACE_DEC_ENABLED       0
#define TRACE_ENC_ENABLED       0
#define TRACE_NUM_INDENT_SPACES 4
#define TRACE_HEADERS           0
#define TRACE_PROB_BINS         0




#if TRACE_ENC_ENABLED || TRACE_DEC_ENABLED

#define TRACE_TO_STR( msg )                                               \
{                                                                         \
  TraceBase *pcTraceBase = this;                                          \
  TraceCommon *pcTraceCommon = pcTraceBase->getTraceCommon();             \
  if( pcTraceCommon && pcTraceCommon->isActive() )                        \
  {                                                                       \
    std::stringstream cMsg;                                               \
    cMsg << (msg);                                                        \
    pcTraceCommon->trace( cMsg.str(), false, true );                      \
  }                                                                       \
}

#define TRACE_NUMBER                                                      \
{                                                                         \
  TraceBase *pcTraceBase = this;                                          \
  TraceCommon *pcTraceCommon = pcTraceBase->getTraceCommon();             \
  if( pcTraceCommon && pcTraceCommon->isActive() )                        \
  {                                                                       \
    std::string cMsg;                                                     \
    pcTraceCommon->trace( cMsg, true, true );                             \
  }                                                                       \
}

#if TRACE_NUM_INDENT_SPACES

#define TRACE_INDENT_INC                                                  \
{                                                                         \
  TraceBase *pcTraceBase = this;                                          \
  TraceCommon *pcTraceCommon = pcTraceBase->getTraceCommon();             \
  if( pcTraceCommon && pcTraceCommon->isActive() )                        \
  {                                                                       \
    pcTraceCommon->incIndent();                                           \
  }                                                                       \
}

#define TRACE_INDENT_DEC                                                  \
{                                                                         \
  TraceBase *pcTraceBase = this;                                          \
  TraceCommon *pcTraceCommon = pcTraceBase->getTraceCommon();             \
  if( pcTraceCommon && pcTraceCommon->isActive() )                        \
  {                                                                       \
    pcTraceCommon->decIndent();                                           \
  }                                                                       \
}
#else

#define TRACE_INDENT_INC
#define TRACE_INDENT_DEC

#endif

#else

#define TRACE_TO_STR( msg )
#define TRACE_NUMBER
#define TRACE_INDENT_INC
#define TRACE_INDENT_DEC

#endif


class TraceCommon
{
protected:
  virtual ErrVal xPutString( const std::string &rcString, bool bFlush = false ) = 0;

public:
  ErrVal trace( const std::string &rcMsg, bool bPrintNumber, bool bFlush = false )
  {
    if( bPrintNumber )
    {
//      if( m_uiCount == 10191 )
//        printf("");

      std::stringstream cStr;
      cStr << std::setw(8) << (int)(m_uiCount++) << ":";
      for( UInt ui = 0; ui < (TRACE_NUM_INDENT_SPACES) * m_uiIndent; ui++ )
      {
        if( ui % (TRACE_NUM_INDENT_SPACES) == 0 )
        {
          cStr << ".";
        }
        else
        {
          cStr << " ";
        }
      }
      RNOK( xPutString( cStr.str(), bFlush ) );
    }
    if( 0 != rcMsg.length() )
    {
      RNOK( xPutString( rcMsg, bFlush ) );
    }
    return Err::m_nOK;
  }

  void incIndent()  { m_uiIndent++; }
  void decIndent()  { AOT_DBG( m_uiIndent == 0); m_uiIndent--; }

  void activate()         { m_bActive = true; }
  void deactivate()       { m_bActive = false; }
  bool isActive()   const { return m_bActive; }

  void setOStream( std::ostream &rcOut ) { m_pcOStream = &rcOut; }
  void resetCount() { m_uiCount = 0; m_uiIndent = 0; }

protected:
  TraceCommon( std::ostream &rcOutput )
    : m_uiCount( 0 )
    , m_uiIndent( 0 )
    , m_bActive( true )
    , m_pcOStream( &rcOutput ) {}

  virtual ~TraceCommon() {}

  UInt m_uiCount;
  UInt m_uiIndent;
  bool m_bActive;
  std::ostream *m_pcOStream;
};


class TraceBase
{
public:
  virtual ~TraceBase() {}

  TraceBase() : m_pcTrace( NULL ) {}

  TraceCommon* getTraceCommon() { return m_pcTrace; }

  void setTraceCommonPtr( TraceCommon *p ) { m_pcTrace = p; }

private:
  TraceCommon *m_pcTrace;
};



class TraceDec
  : public TraceCommon
{
private:
  ErrVal xPutString( const std::string &rcString, bool bFlush )
  {
    *m_pcOStream << rcString;
    if( bFlush )
    {
      m_pcOStream->flush();
    }
    return Err::m_nOK;
  }

public:
  TraceDec( std::ostream &rcOutput ) : TraceCommon( rcOutput ) {}
};

// video namespace end
VIDEO_NAMESPACE_END


#endif // !defined(AFX_TRACECOMMON_H__B87E26CF_023E_4DC7_8F94_D3E38F59ABA1__INCLUDED_)
