/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VLC_INIT_TABLES_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_VLC_INIT_TABLES_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_


#include "VideoCommonLib/VLCodeword.h"
#include "VideoCommonLib/VLCInitTablesData.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_ValType>
class VLCInitTableEntry
{
public:
  VLCInitTableEntry ( const char* pcBinString, const char* pcBitString         ) : m_cBinWord( pcBinString                    ), m_cBitWord( pcBitString                    ) {}
  VLCInitTableEntry ( const VLCInitTableEntry<t_ValType>&  rcVLCInitTableEntry ) : m_cBinWord( rcVLCInitTableEntry.m_cBinWord ), m_cBitWord( rcVLCInitTableEntry.m_cBitWord ) {}
  ~VLCInitTableEntry()  {}

  FORCEINLINE const VLCodeword<t_ValType>&  getBinWord()              const { return m_cBinWord; }
  FORCEINLINE const VLCodeword<t_ValType>&  getBitWord()              const { return m_cBitWord; }
  FORCEINLINE const VLCodeword<t_ValType>&  operator[]( UInt uiIdx )  const { TOF_DBG( uiIdx < 2 ); return ( uiIdx ? m_cBitWord : m_cBinWord ); }

  FORCEINLINE VLCInitTableEntry<t_ValType>& operator= ( const VLCInitTableEntry<t_ValType>& rcVLCInitTableEntry )
  {
    m_cBinWord  = rcVLCInitTableEntry.m_cBinWord;
    m_cBitWord  = rcVLCInitTableEntry.m_cBitWord;
    return *this;
  }

private:
  VLCodeword<t_ValType> m_cBinWord;
  VLCodeword<t_ValType> m_cBitWord;
};


template<class t_ValType>
class VLCInitTable
{
public:
  VLCInitTable  ( const UInt uiState, const UInt uiNumCodePairs, const char** papcBinBitStringArray ) 
  : m_uiState   ( uiState )  
  {
    for( UInt uiIdx = 0; uiIdx < ( uiNumCodePairs << 1 ); uiIdx += 2 )
    {
      m_acVLCInitTableEntries.push_back( VLCInitTableEntry<t_ValType>( papcBinBitStringArray[ uiIdx ], papcBinBitStringArray[ uiIdx + 1 ] ) );
    }
  }
  VLCInitTable  ( const VLCInitTable<t_ValType>&  rcVLCInitTable ) 
  : m_uiState               ( rcVLCInitTable.m_uiState )
  , m_acVLCInitTableEntries ( rcVLCInitTable.m_acVLCInitTableEntries ) 
  {}
  ~VLCInitTable ()  {}

  FORCEINLINE const UInt                          getState      ()              const { return m_uiState; }
  FORCEINLINE const UInt                          getNumEntries ()              const { return UInt( m_acVLCInitTableEntries.size() ); }
  FORCEINLINE const VLCInitTableEntry<t_ValType>& operator[]    ( UInt uiIdx )  const { return m_acVLCInitTableEntries[ uiIdx ]; }

  FORCEINLINE VLCInitTable<t_ValType>& operator= ( const VLCInitTable<t_ValType>& rcVLCInitTable )
  {
    m_uiState               = rcVLCInitTable.m_uiState;
    m_acVLCInitTableEntries = rcVLCInitTable.m_acVLCInitTableEntries;
    return *this;
  }

  FORCEINLINE VLCInitTable<t_ValType>& operator<< ( const VLCInitTableEntry<t_ValType>& rcVLCInitTableEntry )
  {
    m_acVLCInitTableEntries.push_back( rcVLCInitTableEntry );
    return *this;
  }

private:
  UInt                                        m_uiState;
  std::vector<VLCInitTableEntry<t_ValType> >  m_acVLCInitTableEntries;
};


template<class t_ValType>
class VLCInitTableSet
{
public:
  ~VLCInitTableSet()  {}

  static const UInt*                    getStateQuantTable();
  static const VLCInitTable<t_ValType>& getVLCInitTable   ( UInt uiIdx );

private:
  VLCInitTableSet ()  {}
  VLCInitTableSet ( const VLCInitTableSet<t_ValType>& rcVLCInitTableSet ) : m_acVLCInitTables( rcVLCInitTableSet.m_acVLCInitTables ) {}

  FORCEINLINE const VLCInitTable<t_ValType>& operator[]    ( UInt uiIdx )  const { return m_acVLCInitTables[ uiIdx ]; }

  FORCEINLINE VLCInitTableSet<t_ValType>& operator=  ( const VLCInitTableSet<t_ValType>& rcVLCInitTableSet )
  {
    m_acVLCInitTables = rcVLCInitTableSet.m_acVLCInitTables;
    return *this;
  }

  FORCEINLINE VLCInitTableSet<t_ValType>& operator<< ( const VLCInitTable<t_ValType>& rcVLCInitTable )
  {
    m_acVLCInitTables.push_back( rcVLCInitTable );
    return *this;
  }

private:
  std::vector<VLCInitTable<t_ValType> >   m_acVLCInitTables;
};


template<class t_ValType>
const UInt*
VLCInitTableSet<t_ValType>::getStateQuantTable()
{
  static const UInt* pauiStateQuant = VLCInitTablesData::m_auiStateQuantTable;
  return pauiStateQuant;
}


template<class t_ValType>
const VLCInitTable<t_ValType>& 
VLCInitTableSet<t_ValType>::getVLCInitTable( UInt uiIdx )
{
#define ADD_TABLE(i) << VLCInitTable<t_ValType>( i, VLCInitTablesData::m_auiStateTableSizes[i], VLCInitTablesData::m_aaacStateInitTable[i] )
  static const VLCInitTableSet<t_ValType> acInitTableSet = VLCInitTableSet<t_ValType>()
  ADD_TABLE(  0 )  ADD_TABLE(  1 )  ADD_TABLE(  2 )  ADD_TABLE(  3 )  ADD_TABLE(  4 )  ADD_TABLE(  5 )  ADD_TABLE(  6 )  ADD_TABLE(  7 )
  ADD_TABLE(  8 )  ADD_TABLE(  9 )  ADD_TABLE( 10 )  ADD_TABLE( 11 )  ADD_TABLE( 12 )  ADD_TABLE( 13 )  ADD_TABLE( 14 )  ADD_TABLE( 15 )
  ADD_TABLE( 16 )  ADD_TABLE( 17 )  ADD_TABLE( 18 )  ADD_TABLE( 19 )  ADD_TABLE( 20 )  ADD_TABLE( 21 )  ADD_TABLE( 22 )  ADD_TABLE( 23 )
  ADD_TABLE( 24 )  ADD_TABLE( 25 )  ADD_TABLE( 26 )  ADD_TABLE( 27 )  ADD_TABLE( 28 )  ADD_TABLE( 29 )  ADD_TABLE( 30 )  ADD_TABLE( 31 )
  ADD_TABLE( 32 )  ADD_TABLE( 33 )  ADD_TABLE( 34 )  ADD_TABLE( 35 )  ADD_TABLE( 36 )  ADD_TABLE( 37 )  ADD_TABLE( 38 )  ADD_TABLE( 39 )
  ADD_TABLE( 40 )  ADD_TABLE( 41 )  ADD_TABLE( 42 )  ADD_TABLE( 43 )  ADD_TABLE( 44 )  ADD_TABLE( 45 )  ADD_TABLE( 46 )  ADD_TABLE( 47 )
  ADD_TABLE( 48 )  ADD_TABLE( 49 )  ADD_TABLE( 50 )  ADD_TABLE( 51 )  ADD_TABLE( 52 )  ADD_TABLE( 53 )  ADD_TABLE( 54 )  ADD_TABLE( 55 )
  ADD_TABLE( 56 )  ADD_TABLE( 57 )  ADD_TABLE( 58 )  ADD_TABLE( 59 )  ADD_TABLE( 60 )  ADD_TABLE( 61 )  ADD_TABLE( 62 )  ADD_TABLE( 63 );
#undef ADD_TABLE
  return acInitTableSet[ uiIdx ];
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_VLC_INIT_TABLES_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
