/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VLCODEWORD_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_VLCODEWORD_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_


// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_ValType>
class VLCBit
{
public:
  VLCBit ( const bool&              rbBit ) : m_tBit( t_ValType( rbBit ? 1 : 0 ) ) {}
  VLCBit ( const t_ValType&         rtBit ) : m_tBit( t_ValType( 1 ) & rtBit )     {}
  VLCBit ( const VLCBit<t_ValType>& rcBit ) : m_tBit( rcBit.m_tBit )               {}
  ~VLCBit() {}

  FORCEINLINE operator bool     ()  const { return m_tBit == t_ValType( 1 ); }
  FORCEINLINE operator t_ValType()  const { return m_tBit; }

  FORCEINLINE VLCBit<t_ValType>&  operator= ( const VLCBit<t_ValType>& rcVLCBit )
  {
    m_tBit = rcVLCBit.m_tBit;
    return *this;
  }

private:
  t_ValType m_tBit;
};



template<class t_ValType>
class VLCodeword
{
public:

public:
  VLCodeword ();
  VLCodeword ( const t_ValType&             rtValue,
               const UInt&                  ruiLength    );
  VLCodeword ( const VLCodeword<t_ValType>& rcVLCodeword );
  VLCodeword ( const char*                  pcBitString  );
  ~VLCodeword() {}

  FORCEINLINE const t_ValType         getValue  ()  const { return m_tValue;   }
  FORCEINLINE const UInt              getLength ()  const { return m_uiLength; }

  FORCEINLINE const VLCBit<t_ValType> operator[]( const UInt uiIdx     ) const;
  FORCEINLINE const VLCBit<t_ValType> getFront  ( const UInt uiIdx = 0 ) const;
  FORCEINLINE const VLCBit<t_ValType> getBack   ( const UInt uiIdx = 0 ) const;

  FORCEINLINE void                    remFront  ( const UInt uiNum = 1 );
  FORCEINLINE void                    remBack   ( const UInt uiNum = 1 );

  FORCEINLINE const VLCBit<t_ValType> popFront  ();
  FORCEINLINE const VLCBit<t_ValType> popBack   ();

  FORCEINLINE void                    pushFront ( const VLCBit<t_ValType>&  rcBit );
  FORCEINLINE void                    pushBack  ( const VLCBit<t_ValType>&  rcBit );

  FORCEINLINE void                    clear     ()
  {
    m_tValue    = 0;
    m_uiLength  = 0;
  }

  FORCEINLINE VLCodeword<t_ValType>&  operator= ( const VLCodeword<t_ValType>& rcVLCodeword )
  {
    m_tValue    = rcVLCodeword.m_tValue;
    m_uiLength  = rcVLCodeword.m_uiLength;
    return *this;
  }

private:
  t_ValType m_tValue;
  UInt      m_uiLength;
};



template<class t_ValType>
FORCEINLINE
VLCodeword<t_ValType>::VLCodeword()
: m_tValue  ( 0 )
, m_uiLength( 0 )
{
}

template<class t_ValType>
FORCEINLINE
VLCodeword<t_ValType>::VLCodeword( const t_ValType& rtValue, const UInt& ruiLength )
: m_tValue  ( rtValue   )
, m_uiLength( ruiLength )
{
  TOF_DBG( m_uiLength <= ( UInt( sizeof( m_tValue ) ) << 3 ) );
}

template<class t_ValType>
FORCEINLINE
VLCodeword<t_ValType>::VLCodeword( const VLCodeword<t_ValType>& rcVLCodeword )
: m_tValue  ( rcVLCodeword.m_tValue   )
, m_uiLength( rcVLCodeword.m_uiLength )
{
}

template<class t_ValType>
FORCEINLINE
VLCodeword<t_ValType>::VLCodeword( const char* pcBitString )
: m_tValue  ( 0 )
, m_uiLength( 0 )
{
  const UInt uiLength  =   UInt( ::strlen( pcBitString ) );
  TOF_DBG(   uiLength <= ( UInt( sizeof  ( m_tValue    ) ) << 3 ) );
  for( UInt uiIdx = 0; uiIdx < uiLength; uiIdx++ )
  {
    const char cBit = pcBitString[ uiIdx ];
    TOF_DBG (  cBit == '0' || cBit == '1' );
    pushBack(  VLCBit<t_ValType>( ( cBit == '1' ) ) );
  }
  TOF_DBG( uiLength == m_uiLength );
}


template<class t_ValType>
FORCEINLINE
const VLCBit<t_ValType>
VLCodeword<t_ValType>::operator[]( const UInt uiIdx ) const
{
  TOF_DBG( uiIdx < m_uiLength );
  return VLCBit<t_ValType>( m_tValue >> ( m_uiLength - uiIdx - 1 ) );
}

template<class t_ValType>
FORCEINLINE
const VLCBit<t_ValType>
VLCodeword<t_ValType>::getFront( const UInt uiIdx ) const
{
  TOF_DBG( uiIdx < m_uiLength );
  return VLCBit<t_ValType>( m_tValue >> ( m_uiLength - uiIdx - 1 ) );
}

template<class t_ValType>
FORCEINLINE
const VLCBit<t_ValType>
VLCodeword<t_ValType>::getBack( const UInt uiIdx ) const
{
  TOF_DBG( uiIdx < m_uiLength );
  return VLCBit<t_ValType>( m_tValue >> uiIdx );
}


template<class t_ValType>
FORCEINLINE
void
VLCodeword<t_ValType>::remFront( const UInt uiNum )
{
  TOF_DBG( uiNum < m_uiLength );
  m_uiLength  -= uiNum;
  m_tValue    &= ( ( t_ValType( 1 ) << m_uiLength ) - 1 );
}

template<class t_ValType>
FORCEINLINE
void
VLCodeword<t_ValType>::remBack( const UInt uiNum )
{
  TOF_DBG( uiNum < m_uiLength );
  m_uiLength  -= uiNum;
  m_tValue   >>= uiNum;
}


template<class t_ValType>
FORCEINLINE
const VLCBit<t_ValType>
VLCodeword<t_ValType>::popFront()
{
  TOF_DBG( m_uiLength > 0 );
  m_uiLength    --;
  t_ValType tBit = (   m_tValue       >> m_uiLength );
  m_tValue      &= ( ( t_ValType( 1 ) << m_uiLength ) - 1 );
  return VLCBit<t_ValType>( tBit );
}

template<class t_ValType>
FORCEINLINE
const VLCBit<t_ValType>
VLCodeword<t_ValType>::popBack()
{
  TOF_DBG( m_uiLength > 0 );
  m_uiLength    --;
  t_ValType tBit = m_tValue;
  m_tValue     >>= 1;
  return VLCBit<t_ValType>( tBit );
}


template<class t_ValType>
FORCEINLINE
void
VLCodeword<t_ValType>::pushFront( const VLCBit<t_ValType>& rcBit )
{
  TOF_DBG( m_uiLength < ( UInt( sizeof( m_tValue ) ) << 3 ) );
  m_tValue  += ( t_ValType( rcBit ) << m_uiLength );
  m_uiLength++;
}

template<class t_ValType>
FORCEINLINE
void
VLCodeword<t_ValType>::pushBack( const VLCBit<t_ValType>& rcBit )
{
  TOF_DBG( m_uiLength < ( UInt( sizeof( m_tValue ) ) << 3 ) );
  m_tValue <<= 1;
  m_tValue  += t_ValType( rcBit );
  m_uiLength++;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_VLCODEWORD_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
