/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VLC_CODING_TABLES_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_VLC_CODING_TABLES_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_


#include "VideoCommonLib/VLCInitTables.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


enum StateTransition
{
  STATE_TRANSITION_BIT0   = 0,
  STATE_TRANSITION_BIT1   = 1,
  STATE_TRANSITION_END    = 2,
  STATE_TRANSITION_RESET  = 3
};

class VLCTreeNode
{
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( VLCTreeNode );
public:
  template<class t_ValType>
  VLCTreeNode( const VLCInitTable<t_ValType>& rcVLCInitTable, const UInt uiCodeTreeIdx, UInt& ruiNumNodes );
  virtual ~VLCTreeNode();

  const VLCTreeNode*  getRootNode         ()  const { return m_pcRoot; }
  const VLCTreeNode*  getNextNode         ()  const { return xGetNextNode( m_pcRoot ); }
  const UInt          getNodeIdx          ()  const { return m_uiNodeIdx; }
  const UInt          getNode0Idx         ()  const { return ( m_pcChild0 ? m_pcChild0->m_uiNodeIdx : m_pcRoot->m_uiNodeIdx ); }
  const UInt          getNode1Idx         ()  const { return ( m_pcChild1 ? m_pcChild1->m_uiNodeIdx : m_pcRoot->m_uiNodeIdx ); }
  const UInt          getMinLengthNodeIdx ()  const { return m_pcMinLengthLeafNode->m_uiNodeIdx; }
  const UInt          getRootNodeIdx      ()  const { return m_pcRoot->m_uiNodeIdx; }
  const UInt          getCodewordIdx      ()  const { return m_uiCodeIdx; }

private:
  VLCTreeNode( VLCTreeNode* pcRootNode, VLCTreeNode* pcParentNode, UInt& ruiNextNodeIdx );

  template<class t_ValType>
  ErrVal              xAddCodeWord      ( VLCTreeNode* pcRootNode, UInt& ruiNextNodeIdx, const UInt uiCodeIdx, const UInt uiCodeLength, VLCodeword<t_ValType>& rcVLCodeword );
  ErrVal              xCheckInitTreeRoot( const UInt uiNumCodewords, UInt& ruiNumNodes );
  ErrVal              xCheckInitTreeNode( const VLCTreeNode* pcRootNode, const VLCTreeNode* pcParentNode, UInt& ruiMaxNodeIdx, UInt& ruiMaxCodeIdx, UInt& ruiNumNodes );
  ErrVal              xAddMinLengthNodes();
  ErrVal              xAddMinLengthNode ();

  //===== basic navigation =====
  const VLCTreeNode*  xGetNextLeafNode  ( const VLCTreeNode* pcSubTreeRoot )  const;
  VLCTreeNode*        xGetNextLeafNode  ( const VLCTreeNode* pcSubTreeRoot );
  const VLCTreeNode*  xGetNextNode      ( const VLCTreeNode* pcSubTreeRoot )  const;
  VLCTreeNode*        xGetNextNode      ( const VLCTreeNode* pcSubTreeRoot );
  const VLCTreeNode*  xGetNextChild1Node( const VLCTreeNode* pcSubTreeRoot,
                                          const VLCTreeNode* pcPrevNode    )  const;
  VLCTreeNode*        xGetNextChild1Node( const VLCTreeNode* pcSubTreeRoot,
                                          const VLCTreeNode* pcPrevNode    );

private:
  UInt              m_uiNodeIdx;
  UInt              m_uiCodeIdx;
  UInt              m_uiCodeLength;
  VLCTreeNode*      m_pcRoot;
  VLCTreeNode*      m_pcParent;
  VLCTreeNode*      m_pcChild0;
  VLCTreeNode*      m_pcChild1;
  VLCTreeNode*      m_pcMinLengthLeafNode;
};


template<class t_ValType>
VLCTreeNode::VLCTreeNode( const VLCInitTable<t_ValType> &rcVLCInitTable, const UInt uiCodeTreeIdx, UInt& ruiNumNodes )
: m_uiNodeIdx           ( 0 )
, m_uiCodeIdx           ( TypeLimits<UInt>::getMax() )
, m_uiCodeLength        ( TypeLimits<UInt>::getMax() )
, m_pcRoot              ( 0 )
, m_pcParent            ( 0 )
, m_pcChild0            ( 0 )
, m_pcChild1            ( 0 )
, m_pcMinLengthLeafNode ( 0 )
{
  m_pcRoot = this;
  const UInt uiNumTableEntries = UInt( rcVLCInitTable.getNumEntries() );
  TOF_DBG  ( uiCodeTreeIdx     < 2 );
  TOF_DBG  ( uiNumTableEntries > 1 );
  for( UInt uiNextNodeIdx = m_uiNodeIdx + 1, uiCodeIdx = 0; uiCodeIdx < uiNumTableEntries; uiCodeIdx++ )
  {
    VLCodeword<t_ValType> cSrcCodeWord = rcVLCInitTable[ uiCodeIdx ][     uiCodeTreeIdx ];
    VLCodeword<t_ValType> cDesCodeWord = rcVLCInitTable[ uiCodeIdx ][ 1 - uiCodeTreeIdx ];
    TNOK( xAddCodeWord( this, uiNextNodeIdx, uiCodeIdx, cDesCodeWord.getLength(), cSrcCodeWord ) );
  }
  TNOK( xCheckInitTreeRoot( uiNumTableEntries, ruiNumNodes ) );
  TNOK( xAddMinLengthNodes() );
}

template<class t_ValType>
ErrVal
VLCTreeNode::xAddCodeWord( VLCTreeNode* pcRootNode, UInt& ruiNextNodeIdx, const UInt uiCodeIdx, const UInt uiCodeLength, VLCodeword<t_ValType>& rcVLCodeword )
{
  if( rcVLCodeword.getLength() == 0 )
  { // new leaf node (check and update member data)
    ROF( m_uiCodeIdx    == TypeLimits<UInt>::getMax() );
    ROF( m_uiCodeLength == TypeLimits<UInt>::getMax() );
    ROF( m_pcChild0     == 0 );
    ROF( m_pcChild1     == 0 );
    m_uiCodeIdx         =  uiCodeIdx;
    m_uiCodeLength      =  uiCodeLength;
    return Err::m_nOK;
  }
  VLCTreeNode*& rpcChild = ( bool( rcVLCodeword.popFront() ) ? m_pcChild1 : m_pcChild0 );
  if( !rpcChild )
  {
    rpcChild = new VLCTreeNode( pcRootNode, this, ruiNextNodeIdx );
  }
  RNOK( rpcChild->xAddCodeWord( pcRootNode, ruiNextNodeIdx, uiCodeIdx, uiCodeLength, rcVLCodeword ) );
  return Err::m_nOK;
}





template<class t_ValType>
class VLCodingTable
{
public:
  VLCodingTable ( const VLCInitTable<t_ValType>&  rcVLCInitTable, 
                  const bool                      bIsEncoder );
  VLCodingTable ( const VLCodingTable<t_ValType>& rcVLCodingTable );
  virtual ~VLCodingTable();
public:
  FORCEINLINE const UInt  getMaxCodewordLength()  const { return m_uiMaxCodewordLenght; }
  FORCEINLINE const VLCodeword<t_ValType>&  updateState( UInt& ruiState, const StateTransition eStateTransition ) const
  {
    TOF_DBG( ruiState < m_uiNumStates );
    ruiState = m_pauiStateTransition[ ( ruiState << 2 ) + UInt( eStateTransition ) ];
    return     m_pacCodewordTable   [   ruiState ];
  }
private:
  UInt                    m_uiNumStates;
  UInt                    m_uiMaxCodewordLenght;
  UInt*                   m_pauiStateTransition;
  VLCodeword<t_ValType>*  m_pacCodewordTable;
};



template<class t_ValType>
VLCodingTable<t_ValType>::VLCodingTable( const VLCInitTable<t_ValType>& rcVLCInitTable, const bool bIsEncoder )
: m_uiNumStates         ( 0 )
, m_uiMaxCodewordLenght ( 0 )
, m_pauiStateTransition ( 0 )
, m_pacCodewordTable    ( 0 )
{
  const UInt        uiCodetreeIdx = ( bIsEncoder ? 0 : 1 );
  const UInt        uiCodewordIdx = ( bIsEncoder ? 1 : 0 );
  const VLCTreeNode cVLCTree      ( rcVLCInitTable, uiCodetreeIdx, m_uiNumStates );
  m_pauiStateTransition           = new UInt                  [ m_uiNumStates << 2 ];
  m_pacCodewordTable              = new VLCodeword<t_ValType> [ m_uiNumStates ];
  for( const VLCTreeNode* pcNextNode = cVLCTree.getRootNode(); pcNextNode; pcNextNode = pcNextNode->getNextNode() )
  {
    const UInt  uiNodeIdx       = pcNextNode->getNodeIdx();
    const UInt  auiTEntries[4]  = { pcNextNode->getNode0Idx(), pcNextNode->getNode1Idx(), pcNextNode->getMinLengthNodeIdx(), pcNextNode->getRootNodeIdx() };
    const UInt  uiCodeIdx       = pcNextNode->getCodewordIdx();
    TOF_DBG( uiNodeIdx      < m_uiNumStates );
    TOF_DBG( auiTEntries[0] < m_uiNumStates );
    TOF_DBG( auiTEntries[1] < m_uiNumStates );
    TOF_DBG( auiTEntries[2] < m_uiNumStates );
    TOF_DBG( auiTEntries[3] < m_uiNumStates );
    TOF_DBG( uiCodeIdx      < rcVLCInitTable.getNumEntries() || uiCodeIdx == TypeLimits<UInt>::getMax() );
    m_pacCodewordTable[ uiNodeIdx ] = ( uiCodeIdx != TypeLimits<UInt>::getMax() ? rcVLCInitTable[ uiCodeIdx ][ uiCodewordIdx ] : VLCodeword<t_ValType>() );
    UInt*       pauiCurrTableRow    = m_pauiStateTransition + ( uiNodeIdx << 2 );
    ::memcpy(   pauiCurrTableRow, auiTEntries, sizeof( UInt ) << 2 );
  }
  for( UInt uiTableEntryIdx = 0; uiTableEntryIdx < rcVLCInitTable.getNumEntries(); uiTableEntryIdx++ )
  {
    UInt uiCodewordLength = rcVLCInitTable[ uiTableEntryIdx ].getBitWord().getLength();
    m_uiMaxCodewordLenght = gMax<UInt>( m_uiMaxCodewordLenght, uiCodewordLength );
  }
  TOF_DBG( m_uiMaxCodewordLenght > 0 );
}

template<class t_ValType>
VLCodingTable<t_ValType>::VLCodingTable( const VLCodingTable<t_ValType>& rcVLCodingTable )
: m_uiNumStates         ( rcVLCodingTable.m_uiNumStates                     )
, m_uiMaxCodewordLenght ( rcVLCodingTable.m_uiMaxCodewordLenght             )
, m_pauiStateTransition ( new UInt                  [ m_uiNumStates << 2 ]  )
, m_pacCodewordTable    ( new VLCodeword<t_ValType> [ m_uiNumStates ]       )
{
  ::memcpy( m_pauiStateTransition, rcVLCodingTable.m_pauiStateTransition, ( m_uiNumStates << 2 ) * sizeof( UInt ) );
  for( UInt uiIdx = 0; uiIdx < m_uiNumStates; uiIdx++ )
  {
    m_pacCodewordTable[ uiIdx ] = rcVLCodingTable.m_pacCodewordTable[ uiIdx ];
  }
}

template<class t_ValType>
VLCodingTable<t_ValType>::~VLCodingTable()
{
  delete [] m_pauiStateTransition;
  delete [] m_pacCodewordTable;
}





template<class t_ValType, bool t_bIsEncoder>
class VLCodingTables
{
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( VLCodingTables );
public:
  ~VLCodingTables() {}

  static const UInt                       getNumVLCTables ()                     { return sm_cVLCodingTables.m_uiNumVLCTables; }
  static const UInt*                      getState2IdxMap ()                     { return sm_cVLCodingTables.m_auiState2Idx; }
  static const VLCodingTable<t_ValType>&  getVLCodingTable( const UInt  uiIdx )  { return sm_cVLCodingTables.m_acVLCodingTables[ uiIdx ]; }

private:
  VLCodingTables();

private:
  UInt                                    m_uiNumVLCTables;
  UInt                                    m_auiState2Idx[64];
  std::vector<VLCodingTable<t_ValType> >  m_acVLCodingTables;

private:
  static const VLCodingTables<t_ValType,t_bIsEncoder> sm_cVLCodingTables;
};


template<class t_ValType, bool t_bIsEncoder>
VLCodingTables<t_ValType,t_bIsEncoder>::VLCodingTables()
{
  //===== generate intermediate QState2Idx array =====
  const UInt*       pauiStateQuantTable = VLCInitTableSet<t_ValType>::getStateQuantTable();
  UInt              auiQState2Idx[64];
  std::vector<UInt> acQuantStates;
  for( UInt uiIdx = 0; uiIdx < 64; uiIdx++ )
  {
    UInt  uiQuantState  = pauiStateQuantTable[ uiIdx ];
    UInt  uiNextIdx     = UInt( acQuantStates.size() );
    bool  bIsInsideSet  = false;
    for( UInt uiSetIdx  = 0; !bIsInsideSet && uiSetIdx < uiNextIdx; uiSetIdx++ )
    {
      if( acQuantStates[ uiSetIdx ] == uiQuantState )
      {
        bIsInsideSet    = true;
      }
    }
    if( !bIsInsideSet )
    {
      acQuantStates.push_back( uiQuantState );
      auiQState2Idx[ uiQuantState ] = uiNextIdx;
    }
  }
  //===== set number of VLC tables and State2Idx array =====
  m_uiNumVLCTables = UInt( acQuantStates.size() );
  for( UInt uiIdx  = 0; uiIdx < 64; uiIdx++ )
  {
    m_auiState2Idx[ uiIdx ] = auiQState2Idx[ pauiStateQuantTable[ uiIdx ] ];
  }
  //===== set VL coding tables =====
  for( UInt uiIdx = 0; uiIdx < m_uiNumVLCTables; uiIdx++ )
  {
    m_acVLCodingTables.push_back( VLCodingTable<t_ValType>( VLCInitTableSet<t_ValType>::getVLCInitTable( acQuantStates[ uiIdx ] ), t_bIsEncoder ) );
  }
}


template<class t_ValType, bool t_bIsEncoder>
const VLCodingTables<t_ValType,t_bIsEncoder>
VLCodingTables<t_ValType,t_bIsEncoder>::sm_cVLCodingTables;


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_VLC_CODING_TABLES_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
