/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef WIENERFILTER_H_
#define WIENERFILTER_H_


#include "WienerFilterParameter.h"
#include <math.h>



// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_Pel>
class WienerFilter
{
public:
  virtual ~WienerFilter() {}
  static ErrVal filterBlockSeparableHor(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs);
  static ErrVal filterBlockSeparableVer(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs);
  static ErrVal filterPicBlockSeparableNew( const PicBlock<t_Pel> &rcPicBlock, PicBlock<t_Pel>& rcDestPicBlock , const WienerFilterParameters &rcWFparam ) ;
  static ErrVal filterBlockSeparableNew(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock, const FilterCoefficients &rcCoeffs);

private:
  template<UInt uiFilterLength>
  static void xFilterBlockSeparableHor(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs);
  template<UInt uiFilterLength>
  static void xFilterBlockSeparableVer(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs);
  static UChar xClip( const int iX )
  {
    if( 0xFF >= UInt(iX) ){ return iX; }
    return iX < 0 ? 0x00 : 0xFF;
  }
};




template<class t_Pel>
ErrVal WienerFilter<t_Pel>::filterBlockSeparableNew(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs)
{

  int iAccu = 0;

  UInt uiFilterOverlap = rcCoeffs.getFilterLength() >> 1;
  UInt uiBitShift      = rcCoeffs.getNumberOfQuantBits() ;
  int iDcOffset        = rcCoeffs.getQuantDcOffset() ;

  const t_Pel *pcSourceData = &rcBlock.getEl(0);
  t_Pel *pcDestData         = &rcFilteredBlock.getEl(0);

  UInt uiSourceStride      = rcBlock.getStride() ;
  UInt uiDestStride        = rcFilteredBlock.getStride() ;

  const int iAdd = 1 << (uiBitShift - 1 );


  if ( rcCoeffs.getIsHorizontalFilter())
    {
    for (int y = 0; y < rcBlock.getSize().getY(); y++)
    {
//      for (int x = coeff.getMarginOverlap(); x < rcBlock.getSize().getX()-coeff.getMarginOverlap(); x++)
      for (int x = 0 ; x < rcBlock.getSize().getX() ; x++)
      {
        iAccu = 0 ;
        for (UInt i = 0; i < rcCoeffs.getFilterLength(); i++)
        {
          //iAccu += rcBlock.getEl(PixelPos( x - uiFilterOverlap + i, y )) * coeff.getQuantFilterCoeff(i) ;
//          iAccu += pcSourceData[ x - uiFilterOverlap + i + uiSourceStride* y ] * coeff.getQuantFilterCoeff(i) ;
          iAccu += pcSourceData[ std::max( 0 , std::min( rcBlock.getSize().getX() - 1 , int(x - uiFilterOverlap + i ) ))
                                + uiSourceStride *  std::max( 0 , std::min( rcBlock.getSize().getY() - 1 , y )) ]
                                * rcCoeffs.getQuantFilterCoeff(i);
    }
//        rcFilteredBlock.getEl(x,y) = UChar( std::min(255,std::max(0, ((iAccu  + iDcOffset + iAdd)>>uiBitShift) ))) ;
        pcDestData[x + uiDestStride * y ] = UChar( std::min(255,std::max(0, ((iAccu  + iDcOffset + iAdd)>>uiBitShift) )));
//        CHECK( pcDestData[x + uiDestStride * y ] == UChar( std::min(255,std::max(0, ((iAccu  + iDcOffset + iAdd)>>uiBitShift) ))) );
  }
    }
  }
  else if( rcCoeffs.getIsVerticalFilter() )
  {
    for (int y = 0 ; y < rcBlock.getSize().getY() ; y++)
//    for (int y = coeff.getMarginOverlap(); y < rcBlock.getSize().getY()-coeff.getMarginOverlap(); y++)
    {
      for (int x = 0; x < rcBlock.getSize().getX(); x++)
  {
        iAccu = 0 ;
        for (UInt i = 0; i < rcCoeffs.getFilterLength(); i++)
        {
//          iAccu += rcBlock.getEl(PixelPos( x, y - uiFilterOverlap + i )) * coeff.getQuantFilterCoeff(i) ;
//          iAccu += pcSourceData[ x + uiSourceStride* (y- uiFilterOverlap + i) ] * coeff.getQuantFilterCoeff(i) ;
            iAccu += pcSourceData[ std::max( 0 , std::min( rcBlock.getSize().getX() - 1 , x))
                                  + uiSourceStride *  std::max( 0 , std::min( rcBlock.getSize().getY() - 1 , int(y - uiFilterOverlap + i ))) ]
                                  * rcCoeffs.getQuantFilterCoeff(i);
  }
//        rcFilteredBlock.getEl(x,y) = UChar( std::min(255,std::max(0, ((iAccu  + iDcOffset + iAdd)>>uiBitShift) ))) ;
        pcDestData[x + uiDestStride * y ] = UChar( std::min(255,std::max(0, ((iAccu  + iDcOffset + iAdd)>>uiBitShift) )));
//        CHECK( pcDestData[x + uiDestStride * y ] == UChar( std::min(255,std::max(0, ((iAccu  + iDcOffset + iAdd)>>uiBitShift) ))) );
      }
    }
  }
  return 0;
}


template<class t_Pel>
ErrVal WienerFilter<t_Pel>::filterBlockSeparableHor(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs)
{
  ROT( ! rcCoeffs.getIsHorizontalFilter() );

  const UInt uiFilterLength = rcCoeffs.getFilterLength();
  switch( uiFilterLength )
  {
  case 1:
    xFilterBlockSeparableHor<1>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 3:
    xFilterBlockSeparableHor<3>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 5:
    xFilterBlockSeparableHor<5>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 7:
    xFilterBlockSeparableHor<7>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 9:
    xFilterBlockSeparableHor<9>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 11:
    xFilterBlockSeparableHor<11>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 13:
    xFilterBlockSeparableHor<13>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  default:
    RNOK( filterBlockSeparableNew( rcBlock, rcFilteredBlock, rcCoeffs ));
    break;
  }


  return 0;
}

template<class t_Pel>
template<UInt uiFilterLength>
void WienerFilter<t_Pel>::xFilterBlockSeparableHor(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs)
{
  int aiCoeffs[uiFilterLength];
  for (UInt i = 0; i < uiFilterLength; i++)
  {
    aiCoeffs[i] = rcCoeffs.getQuantFilterCoeff(i);
  }

  const UInt uiBitShift     = rcCoeffs.getNumberOfQuantBits() ;
  const int iAdd            = rcCoeffs.getQuantDcOffset() + (1 << (uiBitShift - 1 ));

  const int iFilterOverlap  = uiFilterLength >> 1;
  const int iSizeX          = rcBlock.getSize().getX();
  const int iSizeY          = rcBlock.getSize().getY();

  const int* piCoeffs       = &aiCoeffs[0] + iFilterOverlap;
  const UInt uiSourceStride = rcBlock.getStride() ;
  const UInt uiDestStride   = rcFilteredBlock.getStride() ;

  const t_Pel *pcSourceData = rcBlock.getPtr();
  t_Pel *pcDestData         = rcFilteredBlock.getPtr();

  for (int y = 0; y < iSizeY; y++)
  {
    // left border checked src access
    for (int x = 0 ; x < iFilterOverlap; x++)
    {
      int iAccu = 0;
      const int* piCoeffTmp = &aiCoeffs[0];
      for( int i = x-iFilterOverlap; i <= 0; i++)
      {
        iAccu += *piCoeffTmp++;
      }

      iAccu *= pcSourceData[ 0 ];

      for( int i = 1; i <= iFilterOverlap+x; i++)
      {
        iAccu += *piCoeffTmp++ * pcSourceData[ i ];
      }

      pcDestData[x] = xClip( (iAccu + iAdd)>>uiBitShift);
    }

    // middle unchecked src access
    const int iUncheckedEnd = iSizeX-iFilterOverlap;
    for (int x = iFilterOverlap ; x < iUncheckedEnd; x++)
    {
      int iAccu = 0;
      for( int i = -iFilterOverlap; i <= iFilterOverlap; i++)
      {
        iAccu += piCoeffs[i] * pcSourceData[ x + i ];
      }
      pcDestData[x] = xClip( (iAccu + iAdd)>>uiBitShift);
    }

    // right border checked src access

    const t_Pel *pcSrc = pcSourceData + iSizeX-1;
    t_Pel *pcDes = pcDestData  + iSizeX-1;
    for (int x = -iFilterOverlap+1 ; x <= 0; x++)
    {
      int iAccu = 0;
      const int* piCoeffTmp = &aiCoeffs[0];
      for( int i = x-iFilterOverlap; i < 0; i++)
      {
        iAccu += *piCoeffTmp++ * pcSrc[ i ];
      }

      int iAccu1 = 0;
      for( int i = 0; i <= iFilterOverlap+x; i++)
      {
        iAccu1 += *piCoeffTmp++;
      }
      iAccu += iAccu1 * pcSrc[ 0 ];

      pcDes[x] = xClip( (iAccu + iAdd)>>uiBitShift);
    }
    pcDestData += uiDestStride;
    pcSourceData += uiSourceStride;
  }
}

template<class t_Pel>
ErrVal WienerFilter<t_Pel>::filterBlockSeparableVer(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs)
{
  ROT( ! rcCoeffs.getIsVerticalFilter() )

  const UInt uiFilterLength = rcCoeffs.getFilterLength();
  switch( uiFilterLength )
  {
  case 1:
    xFilterBlockSeparableVer<1>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 3:
    xFilterBlockSeparableVer<3>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 5:
    xFilterBlockSeparableVer<5>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 7:
    xFilterBlockSeparableVer<7>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 9:
    xFilterBlockSeparableVer<9>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 11:
    xFilterBlockSeparableVer<11>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  case 13:
    xFilterBlockSeparableVer<13>( rcBlock, rcFilteredBlock, rcCoeffs);
    break;
  default:
    RNOK( filterBlockSeparableNew( rcBlock, rcFilteredBlock, rcCoeffs ));
    break;
  }

  return 0;
}


template<class t_Pel>
template<UInt uiFilterLength>
void WienerFilter<t_Pel>::xFilterBlockSeparableVer(const Block<t_Pel> &rcBlock, const Block<t_Pel> &rcFilteredBlock,  const FilterCoefficients &rcCoeffs)
{
  int aiCoeffs[uiFilterLength];
  for (UInt i = 0; i < uiFilterLength; i++)
  {
    aiCoeffs[i] = rcCoeffs.getQuantFilterCoeff(i);
  }

  const UInt uiBitShift     = rcCoeffs.getNumberOfQuantBits() ;
  const int iAdd            = rcCoeffs.getQuantDcOffset() + (1 << (uiBitShift - 1 ));

  const int  iFilterOverlap = uiFilterLength >> 1;
  const int iSizeX          = rcBlock.getSize().getX();
  const int iSizeY          = rcBlock.getSize().getY();

  const UInt uiSourceStride = rcBlock.getStride() ;
  const UInt uiDestStride   = rcFilteredBlock.getStride() ;
  const int* piCoeffs       = &aiCoeffs[0] + iFilterOverlap;

  const t_Pel *pcSourceData = rcBlock.getPtr();
  t_Pel *pcDestData         = rcFilteredBlock.getPtr();


  for (int y = 0 ; y < iFilterOverlap; y++)
  {
    for (int x = 0; x < iSizeX; x++)
    {
      const t_Pel *pcSrc = pcSourceData + x;
      int iAccu = 0;
      const int* piCoeffTmp = &aiCoeffs[0];
      for( int i = y-iFilterOverlap; i <= 0; i++)
      {
        iAccu += *piCoeffTmp++;
      }

      iAccu *= *pcSrc;
      for( int i = 1; i <= iFilterOverlap+y; i++)
      {
        pcSrc += uiSourceStride;
        iAccu += *piCoeffTmp++ * *pcSrc;
      }
      pcDestData[x] = xClip( (iAccu + iAdd)>>uiBitShift);
    }
    pcDestData += uiDestStride;
  }

  const int iUncheckedEnd = iSizeY-iFilterOverlap;
  for (int y = iFilterOverlap ; y < iUncheckedEnd; y++)
  {
    for (int x = 0; x < iSizeX; x++)
    {
      int iAccu = 0;
      const t_Pel *pcSrc = pcSourceData + x;
      for( int i = -iFilterOverlap; i <= iFilterOverlap; i++)
      {
        iAccu += piCoeffs[i] * pcSrc[0];
        pcSrc += uiSourceStride;
      }
      pcDestData[x] = xClip( (iAccu + iAdd)>>uiBitShift);
    }
    pcSourceData += uiSourceStride;
    pcDestData += uiDestStride;
  }

  // this is not the right positon
  pcSourceData = &rcBlock.getEl(0,iSizeY-2*iFilterOverlap);
  for (int y = -iFilterOverlap+1; y <= 0; y++)
  {
    for (int x = 0; x < iSizeX; x++)
    {
      const t_Pel *pcSrc = pcSourceData + x;
      int iAccu = 0;
      const int* piCoeffTmp = &aiCoeffs[0];
      for( int i = y-iFilterOverlap; i < 0; i++)
      {
        iAccu += *piCoeffTmp++ * *pcSrc;
        pcSrc += uiSourceStride;
      }

      int iAccu1 = 0;
      for( int i = 0; i <= iFilterOverlap+y; i++)
      {
        iAccu1 += *piCoeffTmp++;
      }
      iAccu += iAccu1 * *pcSrc;
      pcDestData[x] = xClip( (iAccu + iAdd)>>uiBitShift);
    }
    pcDestData += uiDestStride;
    pcSourceData += uiSourceStride;
  }
}

template<class t_Pel>
ErrVal WienerFilter<t_Pel>::filterPicBlockSeparableNew( const PicBlock<t_Pel> &rcPicBlock, PicBlock<t_Pel>& rcDestPicBlock , const WienerFilterParameters &rcWFparam )
{
  for ( UInt uiPlane=0 ; uiPlane < rcPicBlock.getStats().getNumPlanes() ; uiPlane ++)
  {
    if ( rcWFparam.getFilterPlaneFlag( uiPlane ))
    {
      AllocBlock<t_Pel> cHorizontallyFilteredBlock(rcPicBlock.getPlane( uiPlane ).getSize());
      const FilterCoefficients cHorFilterCoeff = rcWFparam.getHorizontalCoeffOfPlane( uiPlane );
      const FilterCoefficients cVerFilterCoeff = rcWFparam.getVerticalCoeffOfPlane( uiPlane );
      RNOK( filterBlockSeparableHor(rcPicBlock.getPlane( uiPlane), cHorizontallyFilteredBlock,cHorFilterCoeff ));
      RNOK( filterBlockSeparableVer(cHorizontallyFilteredBlock, rcDestPicBlock.getPlane(uiPlane ),cVerFilterCoeff ));
    } 
    else
    {
      rcDestPicBlock.getPlane( uiPlane ).assign( rcPicBlock.getPlane( uiPlane ));
    }
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END


#endif /* WIENERFILTER_H_ */
