/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VIDEODECODERIF_H__0366BFA9_45D9_4834_B404_8DE3914C1E58__INCLUDED_)
#define AFX_VIDEODECODERIF_H__0366BFA9_45D9_4834_B404_8DE3914C1E58__INCLUDED_

#include "VideoCodecToolsLib/VideoCodecPicture.h"


//!\ingroup H264VideoLibExternalInterfaces
//!The VideoEncoderIf is an abstract interface class declaring 
//!available encoding methods. 
//!Note all list delivered by the encoder are destroyed with 
//!the next call to the interface.
class VideoDecoderIf 
{
protected:
  VideoDecoderIf() {}
	virtual ~VideoDecoderIf() {}

public:
  virtual ErrVal getPictureStats( const std::vector<UChar> &rBitstreamPacket,
                                  VideoCodecPicture<VideoCodecPel> &rcStats,
                                  UInt &ruiPlanePackages,
                                  UInt64 *puiPoc = NULL ) = 0;

  virtual ErrVal prepareForNewStream() = 0;
  virtual ErrVal createOverlaidDecToDispBuffer() = 0;

  virtual ErrVal decodePicture( const std::vector<UChar> &rcBitstreamPacket,
                                const VideoCodecPicture<VideoCodecPel> *pcReconstructedPic,
                                VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> &rcReconstructedPicSinkIf,
                                std::vector<VideoCodecPicture<VideoCodecPel>*> apcOverlayPic,
                                std::vector<VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*>*> apcOverlayPicSinkIf ) = 0;
};

#endif // !defined(AFX_VIDEODECODERIF_H__0366BFA9_45D9_4834_B404_8DE3914C1E58__INCLUDED_)
