/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_ENCODERPARAMETER_H__8403A680_A65D_466E_A411_05C3A7C0D59F__INCLUDED_)
#define AFX_ENCODERPARAMETER_H__8403A680_A65D_466E_A411_05C3A7C0D59F__INCLUDED_


class ParameterIoIf;


#if defined( MSYS_WIN32 )
# pragma warning( disable: 4275 )
# pragma warning( disable: 4251 )
#endif

// return vector element
#define RETVECELEM(v,e) {if(e<v.size()) return v[e]; return v[v.size()-1]; }


class VideoEncoderParameter
{
public:
  class SPSParams
  {
    friend class VideoEncoderParameter;
  public:
    SPSParams() 
      : m_uiNumberReferenceFrames( 0 )
      , m_uiCodedPictureStoreSize( 0 )
      , m_uiInterpolationMethod  ( 0 )
      , m_uiLevelIdc( 51 ) 
      , m_uiProfileIdc( 108 ) 
      , m_uiAdapInterpSwitch     ( 0 )
    {}

    ErrVal check() const ;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    UInt getNumberReferenceFrames()                         const { return m_uiNumberReferenceFrames; }
    UInt getCodedPictureStoreSize()                         const { return m_uiCodedPictureStoreSize; }
    UInt getInterpolationMethod  ()                         const { return m_uiInterpolationMethod;   }
    UInt getLevelIdc()                                      const { return m_uiLevelIdc; }
    UInt getProfileIdc()                                    const { return m_uiProfileIdc; }
    UInt getAdapInterpSwitch     ()                         const { return m_uiAdapInterpSwitch;      }

  protected:
    UInt m_uiNumberReferenceFrames;
    UInt m_uiCodedPictureStoreSize;
    UInt m_uiInterpolationMethod;
    UInt m_uiLevelIdc; 
    UInt m_uiProfileIdc; 
    UInt m_uiAdapInterpSwitch;
  };


  class GeneralParams
  {
    friend class VideoEncoderParameter;
  public:
    GeneralParams()
      : m_uiTotalInputFrames     ( TypeLimits<UInt>::m_iMax )
      , m_uiInputPicSizeX        ( 0 )
      , m_uiInputPicSizeY        ( 0 )
      , m_uiChromaFormatIdc      ( 1 )
      , m_uiPlaneGrouping        ( 0 )
      , m_bDisableTwoPassEncoding( false )
    {}

    ErrVal check() const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    const std::string& getInputName()                       const { return m_cInputName; }
    const std::string& getOutputName()                      const { return m_cOutputName; }
    const std::string& getResidualName()                    const { return m_cResidualName; }
    const std::string& getPredictionName()                  const { return m_cPredictionName; }
    const std::string& getBitstreamName()                   const { return m_cBitstreamName; }

    UInt  getTotalInputFrames          ()  const { return m_uiTotalInputFrames; }
    UInt  getInputPicSizeX             ()  const { return m_uiInputPicSizeX; }
    UInt  getInputPicSizeY             ()  const { return m_uiInputPicSizeY; }
    UInt  getChromaFormatIdc           ()  const { return m_uiChromaFormatIdc; }
    UInt  getPlaneGrouping             ()  const { return m_uiPlaneGrouping; }
    bool  getDisableTwoPassEncodingFlag()  const { return m_bDisableTwoPassEncoding; }
    UInt  getNumberOfPlanes            ()  const { return ( m_uiChromaFormatIdc > 0 ? 3 : 1 ); }
    UInt  getNumberOfPlanePacks        ()  const { return ( getNumberOfPlanes() == 1 ? 1 : sm_auiNumPP[ m_uiPlaneGrouping ] ); }

    UInt  getNumPlanesInPP  ( UInt uiPP ) const { return ( uiPP >= getNumberOfPlanePacks() ? 0 : sm_aauiNumP[m_uiPlaneGrouping][uiPP] ); }
    UInt  getInterPlanePred ( UInt uiPP ) const { RETVECELEM( m_auiInterPlanePred, uiPP ); }

  protected:
    UInt m_uiTotalInputFrames;
    UInt m_uiInputPicSizeX;
    UInt m_uiInputPicSizeY;
    UInt m_uiChromaFormatIdc;
    UInt m_uiPlaneGrouping;
    std::vector<UInt> m_auiInterPlanePred;
    std::string m_cInputName;
    std::string m_cOutputName;
    std::string m_cResidualName;
    std::string m_cPredictionName;
    std::string m_cBitstreamName;
    bool m_bDisableTwoPassEncoding;

  private:
    static const UInt sm_auiNumPP[4];
    static const UInt sm_aauiNumP[4][3];
  };

  class PrecodedPictureParsingParams
  {
    friend class VideoEncoderParameter;
  public:
    PrecodedPictureParsingParams() {}

    ErrVal check() const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    const std::string& getPrecodedBitstreamName() const { return m_cPrecodedBitstreamName; }
    UInt getNumFramesToParse() const { return m_uiNumFramesToParse; }

  private:
    std::string m_cPrecodedBitstreamName;
    UInt m_uiNumFramesToParse;
  };

  class MultithreadingControlParams
  {
    friend class VideoEncoderParameter;
  public:
    MultithreadingControlParams()
      : m_uiNumAccessUnitEncodingThreads( 1 )
      , m_uiNumPicsInMultithreadingQueue( 0 ) {}

    ErrVal check() const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    UInt getNumAccessUnitEncodingThreads() const { return m_uiNumAccessUnitEncodingThreads; }
    UInt getNumPicsInMultithreadingQueue() const { return m_uiNumPicsInMultithreadingQueue; }

  private:
    UInt m_uiNumAccessUnitEncodingThreads;
    UInt m_uiNumPicsInMultithreadingQueue;
  };

  class EntropySliceParams
  {
    friend class VideoEncoderParameter;
  public:
    EntropySliceParams()
      : m_uiEntropySliceSizeX( 0 )
      , m_uiEntropySliceSizeY( 0 )
      , m_bContextStatesReuse( false ) {}

    ErrVal check() const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    UInt getEntropySliceSizeX()  const { return m_uiEntropySliceSizeX; }
    UInt getEntropySliceSizeY()  const { return m_uiEntropySliceSizeY; }
    bool getContextStatesReuse() const { return m_bContextStatesReuse; }

  protected:
    UInt m_uiEntropySliceSizeX;
    UInt m_uiEntropySliceSizeY;
    bool m_bContextStatesReuse;
  };

  class MotionVectorSearchParams
  {
    friend class VideoEncoderParameter;
  public:
    MotionVectorSearchParams() 
      : m_uiSearchModeFullPel ( 2 )
      , m_uiSearchModeSubPel  ( 1 )
      , m_bSearchPrimPlaneOnly( true )
      , m_uiSearchRange       ( 32 )
      , m_uiMaxIterations     (  0 )
      , m_uiIterSearchRange   (  8 )
    {}

    ErrVal check  () const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    UInt getSearchModeFullPel   ()  const { return m_uiSearchModeFullPel; }
    UInt getSearchModeSubPel    ()  const { return m_uiSearchModeSubPel; }
    bool getPrimPlaneOnlySearch ()  const { return m_bSearchPrimPlaneOnly; }
    UInt getSearchRange         ()  const { return m_uiSearchRange; }
    UInt getMaxIterations       ()  const { return m_uiMaxIterations; }
    UInt getIterSearchRange     ()  const { return m_uiIterSearchRange; }

  protected:
    UInt  m_uiSearchModeFullPel;
    UInt  m_uiSearchModeSubPel;
    bool  m_bSearchPrimPlaneOnly;
    UInt  m_uiSearchRange;   // no limit
    UInt  m_uiMaxIterations;
    UInt  m_uiIterSearchRange;
  };

  class DeblockingFilterParams
  {
    friend class VideoEncoderParameter;
  public:
    DeblockingFilterParams()
      : m_uiFilterIdc     ( 0 )
      , m_iAlphaOffsetDiv2( 0 )
      , m_iBetaOffsetDiv2 ( 0 )
    {}

    ErrVal check  () const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    UInt getFilterIdc       () const { return m_uiFilterIdc; }
    int  getAlphaOffsetDiv2 () const { return m_iAlphaOffsetDiv2; }
    int  getBetaOffsetDiv2  () const { return m_iBetaOffsetDiv2; }

  protected:
    UInt  m_uiFilterIdc;   // 0: Filter All Edges, 1: Filter No Edges
    int   m_iAlphaOffsetDiv2;
    int   m_iBetaOffsetDiv2;
  };

  class InLoopWienerFilterParams
  {
  friend class VideoEncoderParameter;
  public:
    InLoopWienerFilterParams()
      : m_bDisableFilter(true)
      , m_bUseBlocks(true)
      , m_uiFilterType(0)
      , m_uiMinFilterLength(7)
      , m_uiMaxFilterLength(7)
      , m_uiPower2MaxBlock(6)
      , m_uiPower2MinBlock(2) {}

    ErrVal check () const;
    ErrVal process ( ParameterIoIf* pcParameterIoIf );

    bool getUseFilter() const {return !m_bDisableFilter;}
    bool getDisableFilterFlag() const {return m_bDisableFilter; }
    UInt getFilterType() const {return m_uiFilterType;}
    UInt getMinFilterLength() const {return m_uiMinFilterLength;}
    UInt getMaxFilterLength() const {return m_uiMaxFilterLength;}
    UInt getPower2MaxBlcok() const {return m_uiPower2MaxBlock;}
    UInt getPower2MinBlock() const {return m_uiPower2MinBlock;}
    bool getUseBlocks() const {return m_bUseBlocks;}

  protected:
  bool m_bDisableFilter;
  bool m_bUseBlocks;
  UInt m_uiFilterType; //0:not symmetric, 1: point symmetric,2: axially symmetic,3: radial symmetric
  UInt m_uiMinFilterLength;
  UInt m_uiMaxFilterLength;
  UInt m_uiPower2MaxBlock;
  UInt m_uiPower2MinBlock;
  };




  class QuadtreeParams
  {
    friend class VideoEncoderParameter;
  public:
    QuadtreeParams()
      : m_bDisableIntraInInter           ( false )
      , m_uiVLCMode                      ( 0 )
      , m_uiVLCMaxDelay                  ( 0 )
      , m_uiMultiCodewordArithmeticCoding( 0 )
      , m_bFastModeDecision              ( 0 )
      , m_uiFMDThresholdHighQP           ( 50 )
      , m_uiFMDThresholdLowQP            ( 125 ) {}

    ErrVal check  ( UInt uiNumPlaPa, UInt uiChromaFormatIdc, UInt uiPlaneGrouping ) const;
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    bool  getDisableIntraInInter              () const { return m_bDisableIntraInInter; }
    UInt  getMultiCodewordArithmeticCodingMode() const { return m_uiMultiCodewordArithmeticCoding; }
    UInt  getNumCtxEntriesSig                 () const { return UInt( m_auiTrafoPow2ContextModelsSig.size() ); }
    UInt  getNumCtxEntriesLast                () const { return UInt( m_auiTrafoPow2ContextModelsLast.size() ); }
    bool  getFastModeDecision                 () const { return m_bFastModeDecision; }
    UInt  getFMDThresholdHighQP               () const { return m_uiFMDThresholdHighQP; }
    UInt  getFMDThresholdLowQP                () const { return m_uiFMDThresholdLowQP; }
    UInt  getVLCMode                          () const { return m_uiVLCMode; }
    UInt  getVLCMaxDelay                      () const { return m_uiVLCMaxDelay; }
    UInt  getMvCompetition              ( UInt uiPP  ) const { RETVECELEM( m_auiMvCompetition,              uiPP ); }
    UInt  getLog2MvAccuracy             ( UInt uiPP  ) const { RETVECELEM( m_auiLog2MvAccuracy,             uiPP ); }
    UInt  getLog2MaxPredictionBlockSize ( UInt uiPP  ) const { RETVECELEM( m_auiLog2MaxPredictionBlockSize, uiPP ); }
    UInt  getLog2MinPredictionBlockSize ( UInt uiPP  ) const { RETVECELEM( m_auiLog2MinPredictionBlockSize, uiPP ); }
    UInt  getLog2MaxTransformBlockSize  ( UInt uiPP  ) const { RETVECELEM( m_auiLog2MaxTransformBlockSize,  uiPP ); }
    UInt  getLog2MinTransformBlockSize  ( UInt uiPP  ) const { RETVECELEM( m_auiLog2MinTransformBlockSize,  uiPP ); }
    bool  getTrafoPow2SubdivideResidual ( UInt uiPP  ) const { RETVECELEM( m_abTrafoPow2SubdivideResidual,  uiPP ); }
    UInt  getQuantizationMode           ( UInt uiPP  ) const { RETVECELEM( m_auiQuantizationMode,           uiPP ); }
    UInt  getTrafoPow2ContextModelsSig  ( UInt uiIdx ) const { RETVECELEM( m_auiTrafoPow2ContextModelsSig,  uiIdx ); }
    UInt  getTrafoPow2ContextModelsLast ( UInt uiIdx ) const { RETVECELEM( m_auiTrafoPow2ContextModelsLast, uiIdx ); }

  private:
    static bool xCheckIfPowerOfTwoWithEvenExponent( UInt uiVal );

  protected:
    std::vector<UInt> m_auiMvCompetition;
    std::vector<UInt> m_auiLog2MvAccuracy;
    std::vector<UInt> m_auiLog2MaxPredictionBlockSize;
    std::vector<UInt> m_auiLog2MinPredictionBlockSize;
    std::vector<UInt> m_auiLog2MaxTransformBlockSize;
    std::vector<UInt> m_auiLog2MinTransformBlockSize;
    std::vector<bool> m_abTrafoPow2SubdivideResidual;
    std::vector<UInt> m_auiQuantizationMode;
    bool              m_bDisableIntraInInter;
    UInt              m_uiVLCMode;
    UInt              m_uiVLCMaxDelay;
    UInt              m_uiMultiCodewordArithmeticCoding;
    std::vector<UInt> m_auiTrafoPow2ContextModelsSig;
    std::vector<UInt> m_auiTrafoPow2ContextModelsLast;
    bool              m_bFastModeDecision;
    UInt              m_uiFMDThresholdHighQP;
    UInt              m_uiFMDThresholdLowQP;
  };

  class QuantParams
  {
    friend class VideoEncoderParameter;
  public:
    QuantParams()
      : m_uiLambdaCalculationMode( 0 )
      , m_iPPSQuant              ( 26 )
    {}

    ErrVal check  ( UInt uiNumPlanePacks, UInt uiNumLayers ) const; 
    ErrVal process( ParameterIoIf* pcParameterIoIf );

    UInt  getLambdaCalculationMode()                const { return m_uiLambdaCalculationMode; }
    int   getPPSQuant             ()                const { return m_iPPSQuant; }
    int   getQpOffset             ( UInt uiLayer )  const { RETVECELEM( m_aiLayerQpOffsets, uiLayer ); }
    int   getQp                   ( UInt uiLayer )  const { return m_iPPSQuant + getQpOffset( uiLayer ); }
    UInt  getMaxAbsDeltaQP        ( UInt uiPlane )  const { RETVECELEM( m_auiMaxAbsQPDelta, uiPlane ); }
    
    const std::pair<UInt,UInt>& getLayerDeadzonesInter( UInt uiLayer ) const { RETVECELEM( m_apuiuiLayerDeadzonesInter, uiLayer ); }
    const std::pair<UInt,UInt>& getLayerDeadzonesIntra( UInt uiLayer ) const { RETVECELEM( m_apuiuiLayerDeadzonesIntra, uiLayer ); }

  protected:
    UInt                                m_uiLambdaCalculationMode;
    int                                 m_iPPSQuant;
    std::vector<int>                    m_aiLayerQpOffsets;
    std::vector<UInt>                   m_auiMaxAbsQPDelta;
    std::vector<std::pair<UInt,UInt> >  m_apuiuiLayerDeadzonesInter;
    std::vector<std::pair<UInt,UInt> >  m_apuiuiLayerDeadzonesIntra;
  };

  class GOPSettings
  {
    friend class VideoEncoderParameter;
  public:
    GOPSettings()
      : m_iLayerChangeStartPoc( 0 )
      , m_iLayerOffset( 0 ) {}

    ErrVal check();
    ErrVal process( ParameterIoIf* pcParameterIoIf );
    
    std::string getGOPFormatString    () const { return m_cInputFormatString; }
    Int64       getLayerChangeStartPoc() const { return m_iLayerChangeStartPoc; }
    int         getLayerOffset        () const { return m_iLayerOffset; }
    UInt        getNumLayers          () const;

  protected:
    std::string m_cInputFormatString;
    Int64 m_iLayerChangeStartPoc;
    int m_iLayerOffset;
  };


private:
  template<class T>
  static ErrVal xParseStringOfTs    ( const std::string &racString, std::vector<T> &ratVec );
  static ErrVal xParseStringOfRatios( const std::string &racString, std::vector<std::pair<UInt,UInt> > &rauiuiVec );

  VideoEncoderParameter( const VideoEncoderParameter &r );
  const VideoEncoderParameter& operator=( const VideoEncoderParameter &r );

public:
  virtual ~VideoEncoderParameter() {}

protected:
  VideoEncoderParameter();

public:
  static  ErrVal create ( VideoEncoderParameter*& rpcEncoderParameter );
  virtual ErrVal destroy();

public:
  const MotionVectorSearchParams&     getMotionVectorSearchParams    () const { return m_cMotionVectorSearchParams;}
  const DeblockingFilterParams&       getDeblockingFilterParams      () const { return m_cLoopFilterParams;}
  const InLoopWienerFilterParams&     getInLoopWienerParams          () const { return m_cInLoopWienerParams;}
  const QuadtreeParams&               getQuadtreeParams              () const { return m_cQuadtreeParams; }
  const QuantParams&                  getQuantParams                 () const { return m_cQuantParams; }
  const GOPSettings&                  getGOPSettings                 () const { return m_cGOPSettings; }
  const GeneralParams&                getGeneralParams               () const { return m_cGeneralParams; }
  const PrecodedPictureParsingParams& getPrecodedPictureParsingParams() const { return m_cPrecodedPictureParsingParams; }
  const MultithreadingControlParams&  getMultithreadingControlParams () const { return m_cMultithreadingControlParams; }
  const EntropySliceParams            getEntropySliceParams          () const { return m_cEntropySliceParams; }
  const SPSParams&                    getSPSParams                   () const { return m_cSPSParams; }

  ErrVal checkEncoderParameter();
  ErrVal initCommandLineArgs  ( int argc, char**argv, bool bPrintOptionsOnFail = true );
  ErrVal initConfigFile       ( const std::string& rcConfigFileName, bool bIgnoreMissingParameters = false );

protected:
  virtual ErrVal xProcessConfigFileParameters ( ParameterIoIf* pcParameterIoIf );
  virtual ErrVal xProcessArgList              ( const std::string& rcCommand, const std::string& rcParameter );
  virtual void   xPrintHelp();

  ErrVal xFileName2ContentString( const std::string& rcFileName, std::string& rcContentString );

protected:
  MotionVectorSearchParams     m_cMotionVectorSearchParams;
  DeblockingFilterParams       m_cLoopFilterParams;
  InLoopWienerFilterParams     m_cInLoopWienerParams;
  QuadtreeParams               m_cQuadtreeParams;
  QuantParams                  m_cQuantParams;
  GOPSettings                  m_cGOPSettings;
  GeneralParams                m_cGeneralParams;
  PrecodedPictureParsingParams m_cPrecodedPictureParsingParams;
  MultithreadingControlParams  m_cMultithreadingControlParams;
  EntropySliceParams           m_cEntropySliceParams;
  SPSParams                    m_cSPSParams;
};

#if defined( MSYS_WIN32 )
# pragma warning( default: 4275 )
# pragma warning( default: 4251 )
#endif


#endif // !defined(AFX_ENCODERPARAMETER_H__8403A680_A65D_466E_A411_05C3A7C0D59F__INCLUDED_)
