/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCodecToolsLibPch.h"
#include "VideoCodecToolsLib/NewAndDeleteTracker.h"

#if ENABLE_NEW_AND_DELETE_TRACKER

#if NEW_AND_DELETE_TRACKER_MODIFIED

#include "VideoCodecToolsLib/MSysDefines.h"

class NewAndDeleteTracker
{
public:
  NewAndDeleteTracker( bool bArray ) : m_bTrackingActive( true ), m_i64Counter( 0 ), m_bArray( bArray ) {}
  virtual ~NewAndDeleteTracker();

  void activate() { m_bTrackingActive = true; }
  void deactivate() { m_bTrackingActive = false; }

  void addPtr( void *p );
  void erasePtr( void *p );

private:
  class PointerTracking
  {
  public:
    PointerTracking( void *p, Int64 iIdx ) : m_p( p ), m_iIdx( iIdx ) {}

    bool operator==( const PointerTracking &r ) const { return r.m_p == m_p; }
    bool operator<( const PointerTracking &r ) const { return m_p < r.m_p; }

  public:
    void *m_p;
    Int64 m_iIdx;
  };

  std::set<PointerTracking> m_acTracking;
  bool m_bTrackingActive;
  Int64 m_i64Counter;
  const bool m_bArray;
};

inline NewAndDeleteTracker& getNewAndDeleteTracker( bool bArray )
{
  static NewAndDeleteTracker x( true );
  static NewAndDeleteTracker y( false );
  return bArray ? x : y;
}

void* operator new( size_t size )
{
  void *ptr = (void *)malloc(size);
  getNewAndDeleteTracker( false ).addPtr( ptr );
  return(ptr);
}

void operator delete(void *p)
{
  if( p )
  {
    getNewAndDeleteTracker( false ).erasePtr( p );
    free(p);
  }
}

void* operator new[]( size_t size )
{
  void *ptr = (void *)malloc(size);
  getNewAndDeleteTracker( true ).addPtr( ptr );
  return(ptr);
}

void operator delete[](void *p)
{
  if( p )
  {
    getNewAndDeleteTracker( true ).erasePtr( p );
    free(p);
  }
}

#endif

NewAndDeleteTracker::~NewAndDeleteTracker()
{
  m_bTrackingActive = false;
  std::cout << std::endl << "new";
  if( m_bArray )
    std::cout << "[]";

  if( !m_acTracking.empty() )
  {      
    std::cout  << ": Memory Leaks found:" << std::endl;
    for( std::set<PointerTracking>::iterator p = m_acTracking.begin(); p != m_acTracking.end(); p++ )
    {
      std::cout << "Number of new" << ( m_bArray ? "[]" : "" ) << ": " << p->m_iIdx << std::endl;
    }
  }
  else
  {
    std::cout << ": No memory leaks found." << std::endl;
  }
}


void NewAndDeleteTracker::addPtr( void *p )
{
  if( m_bTrackingActive )
  {
    bool bActiveState = m_bTrackingActive;
    m_bTrackingActive = false;
//    if( !m_bArray && m_i64Counter == 1666771 )
//      std::cout << " ";
    m_acTracking.insert( PointerTracking( p, m_i64Counter++ ) );
    m_bTrackingActive = bActiveState;
  }
}

void NewAndDeleteTracker::erasePtr( void *p )
{
  if( m_bTrackingActive )
  {
    bool bActiveState = m_bTrackingActive;
    m_bTrackingActive = false;
    std::set<PointerTracking>::iterator cElementToErase = m_acTracking.find( PointerTracking( p, 0 ) );
    if( cElementToErase == m_acTracking.end() )
    {
      std::cout << std::endl << "delete"  << ( m_bArray ? "[]" : "" ) << " called for invalid pointer." << std::endl;
    }
    else
    {
      m_acTracking.erase( cElementToErase );
    }
    m_bTrackingActive = bActiveState;
  }
}

#endif
