/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/ContextTables.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class T>
class VecInit
{
public:
  VecInit operator<<( const T &rcT ) const
  {
    VecInit x;
    x.m_acT = m_acT;
    x.m_acT.push_back( rcT );
    return x;
  }

  operator std::vector<T>() const { return m_acT; }

private:
  std::vector<T> m_acT;
};


#define CTX_EQ_PR CTX_EQ_PROB_TOKEN, CTX_EQ_PROB_TOKEN

const CtxInitArray ContextTables::m_cCtxAccessSig4x4
= CtxInitArray
( "CTX_START_POW2_SIG", VecInit<CtxIni>() << // contexts are in raster scan
CtxIni( -13, 108) << CtxIni( -15, 100) << CtxIni( -10,  88) << CtxIni( -16,  84) <<
CtxIni( -13, 101) << CtxIni( -12,  94) << CtxIni( -10,  86) << CtxIni(   0,  72) <<
CtxIni( -13,  91) << CtxIni(  -7,  83) << CtxIni(   1,  70) << CtxIni(  -5,  74) <<
CtxIni( -13,  87) << CtxIni( -19,  94) << CtxIni(  18,  59) << CtxIni(  18,  59)
);

const CtxInitArray ContextTables::m_cCtxAccessLast4x4
= CtxInitArray
( "CTX_START_POW2_LAST", VecInit<CtxIni>() << // contexts are in raster scan
CtxIni(  26, -19) << CtxIni(  22, -17) << CtxIni(  33, -23) << CtxIni(  37, -27) <<
CtxIni(  26, -17) << CtxIni(  28, -20) << CtxIni(  33, -23) << CtxIni(  41,  -6) <<
CtxIni(  30, -25) << CtxIni(  40, -28) << CtxIni(  40, -15) << CtxIni(  38,   1) <<
CtxIni(  38, -17) << CtxIni(  33, -11) << CtxIni(  41,  17) << CtxIni(  41,  17)
);


const CtxInitArray ContextTables::m_cCtxAccessSig
= CtxInitArray
( "CTX_START_POW2_SIG", VecInit<CtxIni>() <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
);

const CtxInitArray ContextTables::m_cCtxAccessLast
= CtxInitArray
( "CTX_START_POW2_LAST", VecInit<CtxIni>() <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
);

const CtxInitArray ContextTables::m_cResidualContexts
= CtxInitArray
( "RES_CTX_START_POW2_CBF", VecInit<CtxIni>() <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 )
)
<< CtxInitArray
( "RES_CTX_START_POW2_ABS_GREATER_ONE", VecInit<CtxIni>() <<
CtxIni( -12,  92) << CtxIni( -15,  55) << CtxIni( -10,  60) << CtxIni(  -6,  62) << CtxIni(  -4,  65) <<
CtxIni( -12,  92) << CtxIni( -15,  55) << CtxIni( -10,  60) << CtxIni(  -6,  62) << CtxIni(  -4,  65) <<
CtxIni( -12,  92) << CtxIni( -15,  55) << CtxIni( -10,  60) << CtxIni(  -6,  62) << CtxIni(  -4,  65) <<
CtxIni( -12,  92) << CtxIni( -15,  55) << CtxIni( -10,  60) << CtxIni(  -6,  62) << CtxIni(  -4,  65) <<
CtxIni( -12,  92) << CtxIni( -15,  55) << CtxIni( -10,  60) << CtxIni(  -6,  62) << CtxIni(  -4,  65) <<
CtxIni( -12,  92) << CtxIni( -15,  55) << CtxIni( -10,  60) << CtxIni(  -6,  62) << CtxIni(  -4,  65)
)
<< CtxInitArray
( "RES_CTX_START_POW2_COEFF_LEVEL", VecInit<CtxIni>() <<
CtxIni(CTX_EQ_PR) << CtxIni( -12,  73) << CtxIni(  -8,  76) << CtxIni(  -7,  80) << CtxIni(  -9,  88) << CtxIni( -17, 110) <<
CtxIni(CTX_EQ_PR) << CtxIni( -12,  73) << CtxIni(  -8,  76) << CtxIni(  -7,  80) << CtxIni(  -9,  88) << CtxIni( -17, 110) <<
CtxIni(CTX_EQ_PR) << CtxIni( -12,  73) << CtxIni(  -8,  76) << CtxIni(  -7,  80) << CtxIni(  -9,  88) << CtxIni( -17, 110) <<
CtxIni(CTX_EQ_PR) << CtxIni( -12,  73) << CtxIni(  -8,  76) << CtxIni(  -7,  80) << CtxIni(  -9,  88) << CtxIni( -17, 110) <<
CtxIni(CTX_EQ_PR) << CtxIni( -12,  73) << CtxIni(  -8,  76) << CtxIni(  -7,  80) << CtxIni(  -9,  88) << CtxIni( -17, 110) <<
CtxIni(CTX_EQ_PR) << CtxIni( -12,  73) << CtxIni(  -8,  76) << CtxIni(  -7,  80) << CtxIni(  -9,  88) << CtxIni( -17, 110)
)
<< CtxInitArray
( "RES_CTX_START_POW2_SIGN", VecInit<CtxIni>() <<
CtxIni(CTX_EQ_PR)
);


const CtxInitArray ContextTables::m_cCtxAccess
= CtxInitArray
( "CTX_START_TREE_SUBDIV", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_PRED_MODE_INTRA", VecInit<CtxIni>() <<
CtxIni(  13,  41) << CtxIni(   3,  62) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_INTER_PLAPA_PRED", VecInit<CtxIni>() <<
 CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
 )
<< CtxInitArray
( "CTX_START_DIRECT_MODE", VecInit<CtxIni>() <<
CtxIni(  26,  67) << CtxIni(  16,  90) << CtxIni(   9, 104)
)
<< CtxInitArray
( "CTX_START_MERGE", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_MV_PRED_IDX", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_INTRA_SMOOTHED", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_MVD", VecInit<CtxIni>() <<
CtxIni(CTX_EQ_PR) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_NUM_MVS", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_RES_MODE", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_RES_SKIP_MODE", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,   2) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_POW2_DQP", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_START_POW2_SUBDIV", VecInit<CtxIni>() <<
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << 
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << 
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << 
CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 ) << CtxIni( 0, 64 )
);


const CtxInitArray ContextTables::m_cCtxWiener
= CtxInitArray
( "CTX_START_TREE_SUBDIV", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FLAG", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_SEPARABLE_FLAG", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_MAX_BLOCK", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_MIN_BLOCK", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FILTER_USED_IN_FRAME", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FILTER_TYPE", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FILTER_LENGTH", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_QUADTREE_FLAG", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FILTER_SET_NUM", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FILTER_COEFFS_SIGN", VecInit<CtxIni>() <<
CtxIni(   0,  64)
)
<< CtxInitArray
( "CTX_WIENER_FILTER_ABS_COEFFS", VecInit<CtxIni>() <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) << CtxIni(   0,  64) <<
CtxIni(   0,  64)
);


// video namespace end
VIDEO_NAMESPACE_END
