/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/DeblockingFilter.h"
#include "VideoCommonLib/ResTrafoPow2.h"
#include "VideoCommonLib/ResTrafoSkip.h"
#include "VideoCommonLib/InterPlanePackage.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


const UChar DeblockingFilter::sm_aucBetaTab[52]  =
{
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  2,  2,  2,  3,
  3,  3,  3,  4,  4,  4,  6,  6,  7,  7,
  8,  8,  9,  9, 10, 10, 11, 11, 12, 12,
  13, 13, 14, 14, 15, 15, 16, 16, 17, 17,
  18, 18
};

const DeblockingFilter::AlphaClip DeblockingFilter::sm_acAlphaClip[52] =
{
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
                 
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   0, {  0, 0, 0 } },
  {   4, {  0, 0, 0 } },
  {   4, {  0, 0, 1 } },
  {   5, {  0, 0, 1 } },
  {   6, {  0, 0, 1 } },
                 
  {   7, {  0, 0, 1 } },
  {   8, {  0, 1, 1 } },
  {   9, {  0, 1, 1 } },
  {  10, {  1, 1, 1 } },
  {  12, {  1, 1, 1 } },
  {  13, {  1, 1, 1 } },
  {  15, {  1, 1, 1 } },
  {  17, {  1, 1, 2 } },
  {  20, {  1, 1, 2 } },
  {  22, {  1, 1, 2 } },
                 
  {  25, {  1, 1, 2 } },
  {  28, {  1, 2, 3 } },
  {  32, {  1, 2, 3 } },
  {  36, {  2, 2, 3 } },
  {  40, {  2, 2, 4 } },
  {  45, {  2, 3, 4 } },
  {  50, {  2, 3, 4 } },
  {  56, {  3, 3, 5 } },
  {  63, {  3, 4, 6 } },
  {  71, {  3, 4, 6 } },
                 
  {  80, {  4, 5, 7 } },
  {  90, {  4, 5, 8 } },
  { 101, {  4, 6, 9 } },
  { 113, {  5, 7,10 } },
  { 127, {  6, 8,11 } },
  { 144, {  6, 8,13 } },
  { 162, {  7,10,14 } },
  { 182, {  8,11,16 } },
  { 203, {  9,12,18 } },
  { 226, { 10,13,20 } },
                   
  { 255, { 11,15,23 } },
  { 255, { 13,17,25 } }
} ;


class LeafNodeExpander : public LeafNodeIterator
{
public:
  LeafNodeExpander( Subsampling cSub, UInt uiPlane, const SliceData &rcSlice, QuadtreeGroup<BlockDeblockPars> &rcBlockTree )
    : LeafNodeIterator( rcSlice.getQuadtreeGroupStructure() )
    , m_cSub( cSub )
    , m_uiPlane( uiPlane )
    , m_rcBlockTree( rcBlockTree )
    , m_rcSlice( rcSlice ) {}

private:
  ErrVal xProcessLeafNode( const QTNavi &rcPos );

  ErrVal xExpandTrafoPow2( const QTNavi &rcNavPow2, const QTNavi &rcNav, const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcQuadtree, const bool bIntra, const PrimBlockSize& rcPredSize, const PredInfoInterTranslational *const pcPredInfoInterTranslational, const std::vector<int> &raiQps );
  ErrVal xExpandTrafoSkip( const QTNavi &rcPredNav, const QTNavi &rcNav, const ResInfoTrafoSkip &rcResInfo, const bool bIntra, const PrimBlockSize& rcPredSize, const PredInfoInterTranslational *const pcPredInfoInterTranslational );
  ErrVal xExpandForTrafoPow2RootNodes( const UInt uiLog2TrafoPow2RootNodeSize, const QTNavi &rcNav );
  ErrVal xExpandInterPlaPaPred( const QTNavi &rcEdgeTreeNav, const PrimBlockSize& rcPredSize, const PredInfoInterTranslational* pcPredInfo );
  ErrVal xExpandInterPlaPaPred( const QTNavi &rcEdgeTreeNav, const QTNavi &rcPredTreeNav, const QuadtreeGroup<PredInfo*>& rcPredTree, const PrimBlockSize& rcPredSize );

  const Subsampling m_cSub;
  const UInt m_uiPlane;
  QuadtreeGroup<BlockDeblockPars> &m_rcBlockTree;
  const SliceData &m_rcSlice;
};



ErrVal
LeafNodeExpander::xExpandInterPlaPaPred( const QTNavi&                    rcEdgeTreeNav,
                                         const PrimBlockSize&             rcPredSize,
                                         const PredInfoInterTranslational*  pcPredInfo )
{
  if( rcEdgeTreeNav.isLeafNode() )
  {
    BlockDeblockPars& rcPars            = m_rcBlockTree.getT( rcEdgeTreeNav );
    rcPars.m_cPredSize                  = rcPredSize;
    rcPars.m_pcPredInfoInterTranslational = pcPredInfo;
    return Err::m_nOK;
  }
  for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
  {
    const QTNavi &rcChildNav = rcEdgeTreeNav.getChild( uiIdx );
    if( rcChildNav.isLeafOrInternalNode() )
    {
      RNOK( xExpandInterPlaPaPred( rcChildNav, rcPredSize, pcPredInfo ) );
    }
  }
  return Err::m_nOK;
}

ErrVal
LeafNodeExpander::xExpandInterPlaPaPred( const QTNavi&                    rcEdgeTreeNav,
                                         const QTNavi&                    rcPredTreeNav,
                                         const QuadtreeGroup<PredInfo*>& rcPredTree,
                                         const PrimBlockSize&            rcPredSize )
{
  if( rcPredTreeNav.isLeafNode() ) // same pred info -> assign to all child nodes as appropriate
  {
    const PredInfo*                 pcPredInfo  = rcPredTree.getT( rcPredTreeNav );
    ROF ( pcPredInfo );
    const PredInfoInterTranslational* pcPredTrans = pcPredInfo->getPredInfoInterTranslational();
    ROF ( pcPredTrans );
    RNOK( xExpandInterPlaPaPred( rcEdgeTreeNav, rcPredSize, pcPredTrans ) );
    return Err::m_nOK;
  }
  if( rcEdgeTreeNav.isLeafNode() ) // add child nodes with same info [problem: sig info is inferred also inside a block -> stronger deblocking]
  {
    BlockDeblockPars& rcBasePars  = m_rcBlockTree.getT( rcEdgeTreeNav );
    bool              bOK         = false;
    RNOK( m_rcBlockTree.getTreeStructure().appendChildNodes( rcEdgeTreeNav, bOK ) );
    ROF ( bOK );
    for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
    {
      const QTNavi &rcChildNav = rcEdgeTreeNav.getChild( uiIdx );
      if( rcChildNav.isLeafOrInternalNode() )
      {
        BlockDeblockPars& rcChildPars = m_rcBlockTree.getT( rcChildNav );
        rcChildPars                   = rcBasePars;
      }
    }
  }
  for( UInt uiIdx = 0; uiIdx < 4; uiIdx++ )
  {
    const QTNavi &rcChildNavEdge = rcEdgeTreeNav.getChild( uiIdx );
    const QTNavi &rcChildNavPred = rcPredTreeNav.getChild( uiIdx );
    if( rcChildNavEdge.isLeafOrInternalNode() )
    {
      const PrimBlockSize &rcChildPredSize = rcChildNavPred.getPrimBlockSize();
      RNOK( xExpandInterPlaPaPred( rcChildNavEdge, rcChildNavPred, rcPredTree, rcChildPredSize ) );
    }
  }
  return Err::m_nOK;
}


ErrVal LeafNodeExpander::xExpandTrafoPow2( const QTNavi &rcNavPow2, const QTNavi &rcNav, const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcQuadtree, const bool bIntra, const PrimBlockSize& rcPredSize, const PredInfoInterTranslational *const pcPredInfoInterTranslational, const std::vector<int> &raiQps )
{
  if( !rcNavPow2.isLeafNode() )
  {
    bool bSuccess = false;
    RNOK( m_rcBlockTree.getTreeStructure().appendChildNodes( rcNav, bSuccess ) );
    ROF( bSuccess );

    for( UInt ui = 0; ui < 4; ui++ )
    {
      const QTNavi &rcChild = rcNav.getChild( ui );
      const QTNavi &rcChildPow2 = rcNavPow2.getChild( ui );
      if( rcChild.isLeafNode() )
      {
        RNOK( xExpandTrafoPow2( rcChildPow2, rcChild, rcQuadtree, bIntra, rcPredSize, pcPredInfoInterTranslational, raiQps ) );
      }
    }
  }
  else
  {
    BlockDeblockPars &rcPars = m_rcBlockTree.getT( rcNav );
    rcPars.m_bIntra = bIntra;
    rcPars.m_cPredSize = rcPredSize;
    rcPars.m_iQp = raiQps[m_uiPlane];
    rcPars.m_pcPredInfoInterTranslational = pcPredInfoInterTranslational;
    rcPars.m_bSig = rcQuadtree.getT( rcNavPow2 )->getCbf( m_uiPlane );
  }

  return Err::m_nOK;
}


ErrVal LeafNodeExpander::xExpandTrafoSkip( const QTNavi &rcPredNav, const QTNavi &rcNav, const ResInfoTrafoSkip &rcResInfo, const bool bIntra, const PrimBlockSize& rcPredSize, const PredInfoInterTranslational *const pcPredInfoInterTranslational )
{
  BlockDeblockPars &rcPars = m_rcBlockTree.getT( rcNav );
  rcPars.m_bIntra = bIntra;
  rcPars.m_cPredSize = rcPredSize;
  rcPars.m_pcPredInfoInterTranslational = pcPredInfoInterTranslational;

  rcPars.m_bSig = false;
  rcPars.m_iQp = rcResInfo.getQps()[m_uiPlane];

  return Err::m_nOK;
}

ErrVal LeafNodeExpander::xExpandForTrafoPow2RootNodes( const UInt uiLog2TrafoPow2RootNodeSize, const QTNavi &rcNav )
{
  if( rcNav.getLog2BlockSize() > uiLog2TrafoPow2RootNodeSize )
  {
    bool bSuccess = false;
    RNOK( m_rcBlockTree.getTreeStructure().appendChildNodes( rcNav, bSuccess ) );
    ROF( bSuccess );

    for( UInt ui = 0; ui < 4; ui++ )
    {
      const QTNavi &rcChild = rcNav.getChild( ui );
      ROF( rcChild.getLog2BlockSize() + 1 == rcNav.getLog2BlockSize() ); // was false due to bug in QuadTreeGroup.cpp
      if( rcChild.isLeafNode() )
      {
        RNOK( xExpandForTrafoPow2RootNodes( uiLog2TrafoPow2RootNodeSize, rcChild ) );
      }
    }
  }

  return Err::m_nOK;
}


ErrVal LeafNodeExpander::xProcessLeafNode( const QTNavi &rcPos )
{
  const QTNavi &rcEdgeTreeNav = m_rcBlockTree.getTreeStructure().getNodeByPrimPixelPos( rcPos.getPrimPixelPos() );
  const ResInfo *pResInfo = m_rcSlice.getResInfoTree().getT( rcPos );
  const bool bIntra = m_rcSlice.getPredInfoTree().getT( rcPos )->isIntraPredicted();
  const PrimBlockSize& rcPredSize = rcPos.getPrimBlockSize();
  const PredInfoInterTranslational *pcPredInfoInterTranslational = m_rcSlice.getPredInfoTree().getT( rcPos )->getPredInfoInterTranslational();
  switch( pResInfo->getResidualMode() )
  {
  case ResInfo::RES_TRAFO_SKIP:
    {
      RNOK( xExpandTrafoSkip( rcPos, rcEdgeTreeNav, *pResInfo->getResInfoTrafoSkip(), bIntra, rcPredSize, pcPredInfoInterTranslational ) );
      break;
    }
  case ResInfo::RES_TRAFO_POW2:
    {
      const QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcQuadtree = pResInfo->getResInfoTrafoPow2()->getTree();
      RNOK( xExpandForTrafoPow2RootNodes( rcQuadtree.getTreeStructure().getLog2MaxBlockSize(), rcEdgeTreeNav ) );
      for( QTNavi cNavPow2 = rcQuadtree.getTreeStructure().getFirstQuadtreeRoot(); !cNavPow2.liesOutside(); cNavPow2 = rcQuadtree.getTreeStructure().getNextQuadtreeRoot( cNavPow2 ) )
      {
        const QTNavi cSubEdgeTreeNav = m_rcBlockTree.getTreeStructure().getNodeByPrimPixelPos( rcPos.getPrimPixelPos() + cNavPow2.getPrimPixelPos() );
        RNOK( xExpandTrafoPow2( cNavPow2, cSubEdgeTreeNav, rcQuadtree, bIntra, rcPredSize, pcPredInfoInterTranslational, pResInfo->getResInfoTrafoPow2()->getQps() ) );
      }
      break;
    }
  default:
    ROT( 1 );
  }
  if( m_rcSlice.getPredInfoTree().getT( rcPos )->getPredictionMode() == PredInfo::INTER_PLANE_PACKAGE )
  {
    const PredInfo*                   pcPredInfo  = m_rcSlice.getPredInfoTree().getT( rcPos );
    ROF( pcPredInfo );
    const PredInfoInterPlanePackage*  pcPredIPP   = pcPredInfo->getPredInfoInterPlanePackage();
    ROF( pcPredIPP );
    const QuadtreeGroup<PredInfo*>&   rcPredTree  = pcPredIPP->getPredTree();
    const QTNavi                      cPredNav    = rcPredTree.getTreeStructure().getFirstQuadtreeRoot();
    RNOK( xExpandInterPlaPaPred( rcEdgeTreeNav, cPredNav, rcPredTree, cPredNav.getPrimBlockSize() ) );
  }
  return Err::m_nOK;
}


ErrVal DeblockingFilter::expandEdgeTree( Subsampling cSub, UInt uiPlane, QuadtreeGroup<BlockDeblockPars> &rcBlockTree, const SliceData &rcSlice )
{
  LeafNodeExpander cExpander( cSub, uiPlane, rcSlice, rcBlockTree );
  RNOK( cExpander.processTree() );
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
