/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/H264PictureParameterSet.h"
#include "VideoCommonLib/AccessUnit.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

extern const int g_aaiLumaToChromaQp[2][52];

ErrVal PictureParameterSet::setAccessUnitParameters( const AccessUnitParameters &rcPars )
{
  m_uiPicParameterSetId                       = 0;
  m_uiSeqParameterSetId                       = 0;
  m_bEntropyCodingModeFlag                    = !rcPars.getVLCFlag();
  m_bPicOrderPresentFlag                      = true; // poc mode 0

  m_uiNumSliceGroups                          = 1;
  m_uiMbSliceGroupMapType                     = 0;
  m_bSliceGroupChangeDirectionFlag            = false;
  m_uiSliceGroupChangeRate                    = 0;
  m_uiPicSizeInMapUnits                       = 0;
  m_uiPicInitQs                               = 26;
  m_bConstrainedIntraPredFlag                 = false;
  m_bRedundantPicCntPresentFlag               = false;

  m_bWeightedPredFlag                         = false;
  m_uiWeightedBipredIdc                       = 0;
  m_iPicInitQp                                = 26;
  m_bDeblockingFilterParametersPresentFlag    = rcPars.getDeblockingIdc()    != 0 ||
                                                rcPars.getDeblockAlphaOffs() != 0 ||
                                                rcPars.getDeblockBetaOffs()  != 0;

  m_bTransform8x8ModeFlag                     = false;
  m_iChromaQpIndexOffset                      = 0;
  m_iSecondChromaQpIndexOffset                = 0;

  std::vector<int> cExtQpVector;
  for( UInt uiPlaPaId = 0; uiPlaPaId < rcPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    const std::vector<int> &rcQpVector = rcPars.getPlaPaPars( uiPlaPaId ).getPlaneQps();
    cExtQpVector.insert( cExtQpVector.end(), rcQpVector.begin(), rcQpVector.end());
  }

  m_auiNumRefIdxActive. setAll( 1 );
  ROT(cExtQpVector.size() > 3);
  ROT(cExtQpVector.size() < 1);

  if( cExtQpVector.size() > 1 )
  {
    RNOK( xDeriveChromaQpIndexOffsetFromLumaAndChromaQp( m_iChromaQpIndexOffset, cExtQpVector[0], cExtQpVector[1] ) );
    if( cExtQpVector.size() > 2 )
    {
      RNOK( xDeriveChromaQpIndexOffsetFromLumaAndChromaQp( m_iSecondChromaQpIndexOffset, cExtQpVector[0], cExtQpVector[2] ) );
    }
    else
    {
      m_iSecondChromaQpIndexOffset = m_iChromaQpIndexOffset;
    }
  }

  m_iPicInitQp = cExtQpVector[0];
  return Err::m_nOK;
}

ErrVal PictureParameterSet::xDeriveChromaQpIndexOffsetFromLumaAndChromaQp( int& iChromaQpIndexOffset, int iLumaQp, int iChromaQp )
{
  int iBestIdx = 99;
  for( int i = -12; i <= 12; i++ )
  {
    if( g_aaiLumaToChromaQp[0][gMinMax( iLumaQp+i, 0, 51 )] == iChromaQp && abs( i ) < abs( iBestIdx ) )
    {
      iBestIdx = i;
    }
  }
  if( iBestIdx == 99 )
  {
    return Err::m_nERR;
  }
  iChromaQpIndexOffset = iBestIdx;
  return Err::m_nOK;
}


PictureParameterSet::PictureParameterSet( )
  : m_uiPicParameterSetId                       ( -1 )
  , m_uiSeqParameterSetId                       ( -1 )
  , m_bEntropyCodingModeFlag                    ( false )
  , m_bPicOrderPresentFlag                      ( false )
  , m_uiNumSliceGroups                          ( 1 )
  , m_uiMbSliceGroupMapType                     ( 0 )
  , m_bSliceGroupChangeDirectionFlag            ( false )
  , m_uiSliceGroupChangeRate                    ( 0 )
  , m_uiPicSizeInMapUnits                       ( 0 )
  , m_bWeightedPredFlag                         ( false )
  , m_uiWeightedBipredIdc                       ( 0 )
  , m_iPicInitQp                                ( 26 )
  , m_uiPicInitQs                               ( 26 )
  , m_iChromaQpIndexOffset                      ( 0 )
  , m_bDeblockingFilterParametersPresentFlag    ( false )
  , m_bConstrainedIntraPredFlag                 ( false )
  , m_bRedundantPicCntPresentFlag               ( false )
  , m_bTransform8x8ModeFlag                     ( false )
  , m_iSecondChromaQpIndexOffset                ( 0 )
{
  m_auiRunLength.       setAll( 0 );
  m_auiTopLeftMb.       setAll( 0 );
  m_auiBottomRightMb.   setAll( 0 );
  m_auiNumRefIdxActive. setAll( 0 );
}

PictureParameterSet::~PictureParameterSet()
{
}


ErrVal PictureParameterSet::create( PictureParameterSet*& rpcPPS )
{
  rpcPPS = new PictureParameterSet;

  ROT( NULL == rpcPPS );

  return Err::m_nOK;
}

ErrVal PictureParameterSet::destroy()
{
  delete this;
  return Err::m_nOK;
}


bool PictureParameterSet::check( Profile eProfile, UInt uiLevel ) const
{
  ROFRS( xCheckProfile( eProfile ), false );
  ROFRS( xCheckLevel(   uiLevel ),  false );
  return true;
}


bool PictureParameterSet::xCheckLevel( UInt uiLevel ) const
{
  return true;
}

bool PictureParameterSet::xCheckProfile( Profile eProfile ) const
{
  switch( eProfile )
  {
  case BASELINE_PROFILE:
    {
      ROTRS( getWeightedPredFlag(),       false);
      ROTRS( getEntropyCodingModeFlag(),  false);
      ROTRS( getNumSliceGroups() >= 8,    false);
    }
    break;
  case MAIN_PROFILE:
    {

    }
    break;
  case EXTENDED_PROFILE:
    {
      ROTRS( getEntropyCodingModeFlag(),  false);
      ROTRS( getNumSliceGroups() >= 8,    false);
    }
    break;
  default:
    {
      AOT(1);
      return false;
    }
  }
  return true;
}


ErrVal PictureParameterSet::write( HeaderSymbolWriteIf* pcWriteIf ) const
{
  RNOK( pcWriteIf->writeUvlc( getPicParameterSetId(),  "PPS: pic_parameter_set_id"));

  RNOK( pcWriteIf->writeUvlc( getSeqParameterSetId(),  "PPS: seq_parameter_set_id"));


  RNOK( pcWriteIf->writeFlag( getEntropyCodingModeFlag(),  "PPS: entropy_coding_mode_flag"));

  RNOK( pcWriteIf->writeFlag( getPicOrderPresentFlag(),"PPS: pic_order_present_flag"));

  RNOK( xWriteSliceGroupInfo( pcWriteIf ));

  RNOK( pcWriteIf->writeUvlc( getNumRefIdxActive(LIST_0)-1, "PPS: num_ref_idx_l0_active_minus1"));

  RNOK( pcWriteIf->writeUvlc( getNumRefIdxActive(LIST_1)-1, "PPS: num_ref_idx_l1_active_minus1"));

  RNOK( pcWriteIf->writeFlag( getWeightedPredFlag(),"PPS: weighted_pred_flag"));

  RNOK( pcWriteIf->writeCode( getWeightedBipredIdc(),2,"PPS: weighted_bipred_idc"));

  RNOK( pcWriteIf->writeSvlc(((int) getPicInitQp())-26,  "PPS: pic_init_qp_minus26"));

  RNOK( pcWriteIf->writeSvlc(((int) getPicInitQs())-26,  "PPS: pic_init_qs_minus26"));

  RNOK( pcWriteIf->writeSvlc( getChromaQpIndexOffset(),  "PPS: chroma_qp_index_offset"));

  RNOK( pcWriteIf->writeFlag( getDeblockingFilterParametersPresentFlag(),"PPS: deblocking_filter_parameters_present_flag"));

  RNOK( pcWriteIf->writeFlag( getConstrainedIntraPredFlag(),"PPS: constrained_intra_pred_flag"));

  RNOK( pcWriteIf->writeFlag( getRedundantPicCntPresentFlag(),"PPS: redundant_pic_cnt_present_flag"));

  RNOK( xWriteFrext( pcWriteIf ) );
  return Err::m_nOK;
}


ErrVal PictureParameterSet::xWriteSliceGroupInfo( HeaderSymbolWriteIf* pcWriteIf ) const
{
  UInt uiSliceGroupsMinus1;
  UInt uiMbMapType;

  RNOK( pcWriteIf->writeUvlc(uiSliceGroupsMinus1=( getNumSliceGroups()-1),  "PPS: num_slice_groups_minus1"));

  ROTRS( 0 == uiSliceGroupsMinus1, Err::m_nOK );

  RNOK( pcWriteIf->writeUvlc(uiMbMapType =  getMbSliceGroupMapType(),  "PPS: mb_slice_group_map_type"));

  switch( uiMbMapType )
  {
  case 0:
    {
      for( UInt n = 0; n <= uiSliceGroupsMinus1; n++)
      {
        RNOK( pcWriteIf->writeUvlc( getRunLength( n )-1,  "PPS: run_length_minus1"));
      }
      break;
    }
  case 1:
    {
      break;
    }
  case 2:
    {
      for( UInt n = 0; n < uiSliceGroupsMinus1; n++)
      {
        RNOK( pcWriteIf->writeUvlc( getTopLeftMb( n ),  "PPS: top_left_mb"));
        RNOK( pcWriteIf->writeUvlc( getBottomRightMb( n ),  "PPS: bottom_right_mb"));
      }
      break;
    }
  case 3:
  case 4:
  case 5:
    {
      // for these modes uiSliceGroupsMinus1 has to be 1
      ROT(uiSliceGroupsMinus1!=1);

      RNOK( pcWriteIf->writeFlag( getSliceGroupChangeDirectionFlag(),"PPS: slice_group_change_direction_flag"));
      RNOK( pcWriteIf->writeUvlc( getSliceGroupChangeRate()-1,  "PPS: slice_group_change_rate_minus1"));
      break;
    }
  case 6:
    {
      UInt uiPicSizeInMapUnits =  getPicSizeInMapUnits();
      ROT (uiPicSizeInMapUnits<1);
      RNOK( pcWriteIf->writeUvlc(uiPicSizeInMapUnits-1,  "PPS: pic_size_in_map_units1"));

      AOT(1);
      break;
    }
  default:
    {
      AOT(1);
      return Err::m_nERR;
    }
  }

  return Err::m_nOK;
}


ErrVal PictureParameterSet::read( HeaderSymbolReadIf* pcReadIf )
{
  UInt uiTmp;
   int  iTmp;

  RNOK( pcReadIf->getUvlc( m_uiPicParameterSetId,  "PPS: pic_parameter_set_id"));
  ROF( 0 == m_uiPicParameterSetId );
  ROTR(m_uiPicParameterSetId>255, Err::m_nInvalidParameter );

  RNOK( pcReadIf->getUvlc( m_uiSeqParameterSetId,  "PPS: seq_parameter_set_id"));
  ROF( 0 == m_uiSeqParameterSetId );
  ROTR(m_uiSeqParameterSetId>31, Err::m_nInvalidParameter);

  RNOK( pcReadIf->getFlag(m_bEntropyCodingModeFlag, "PPS: entropy_coding_mode_flag"));

  RNOK( pcReadIf->getFlag( m_bPicOrderPresentFlag,"PPS: pic_order_present_flag"));
  ROF( m_bPicOrderPresentFlag );

  RNOK( xReadSliceGroupInfo( pcReadIf ) );

  RNOK( pcReadIf->getUvlc(uiTmp,  "PPS: num_ref_idx_l0_active_minus1"));
  ROTR(uiTmp>31, Err::m_nInvalidParameter);
   setNumRefIdxActive( LIST_0, uiTmp+1);

  RNOK( pcReadIf->getUvlc(uiTmp,  "PPS: num_ref_idx_l1_active_minus1"));
  ROTR(uiTmp>31, Err::m_nInvalidParameter);
   setNumRefIdxActive( LIST_1, uiTmp+1);

  RNOK( pcReadIf->getFlag( m_bWeightedPredFlag, "PPS: weighted_pred_flag"));
  ROT( m_bWeightedPredFlag );

  RNOK( pcReadIf->getCode( m_uiWeightedBipredIdc, 2, "PPS: weighted_bipred_idc"));
  ROF( 0 == m_uiWeightedBipredIdc );
  ROTR( m_uiWeightedBipredIdc>2, Err::m_nInvalidParameter);

  RNOK( pcReadIf->getSvlc(iTmp,  "PPS: pic_init_qp_minus26"));
  // QP range 0..51
  ROTR((iTmp<(-(26)))||(iTmp>25), Err::m_nInvalidParameter);
  setPicInitQp((UInt)(iTmp+26));

  RNOK( pcReadIf->getSvlc(iTmp,  "PPS: pic_init_qs_minus26"));
  // QP range 0..51
  ROTR((iTmp<(-26))||(iTmp>25), Err::m_nInvalidParameter);
   setPicInitQs((UInt)(iTmp+26));

  RNOK( pcReadIf->getSvlc(iTmp,  "PPS: chroma_qp_index_offset"));
  // QP range 0..51
  ROTR((iTmp<(-12))||(iTmp>12), Err::m_nInvalidParameter);
  setChromaQpIndexOffset(iTmp);
  setSecondChromaQpIndexOffset(iTmp);
  RNOK( pcReadIf->getFlag( m_bDeblockingFilterParametersPresentFlag,"PPS: deblocking_filter_parameters_present_flag"));

  RNOK( pcReadIf->getFlag( m_bConstrainedIntraPredFlag,"PPS: constrained_intra_pred_flag"));
  ROT( m_bConstrainedIntraPredFlag );

  RNOK( pcReadIf->getFlag( m_bRedundantPicCntPresentFlag,"PPS: redundant_pic_cnt_present_flag"));
  ROT( m_bRedundantPicCntPresentFlag );

  RNOK( xReadFrext( pcReadIf ) );

  return Err::m_nOK;
}



ErrVal PictureParameterSet::xReadSliceGroupInfo( HeaderSymbolReadIf* pcReadIf )
{
  UInt uiSliceGroupsMinus1;
  UInt uiMbMapType;
  UInt uiTmp;
  bool bFlag;

  RNOK( pcReadIf->getUvlc(uiSliceGroupsMinus1,  "PPS: num_slice_groups_minus1"));

  ROT( 0 != uiSliceGroupsMinus1 );
  // Note: Allowed range specified by profile/level
  //       max. range defined for baseline / XXX is 0..7
  ROTR(uiSliceGroupsMinus1>7, Err::m_nInvalidParameter );
  setNumSliceGroups(uiSliceGroupsMinus1+1);

  ROTRS( 0 == uiSliceGroupsMinus1, Err::m_nOK ); // one slice group no add. info

  RNOK( pcReadIf->getUvlc(uiMbMapType,  "PPS: mb_slice_group_map_type"));
  ROTR(uiMbMapType>6, Err::m_nInvalidParameter)
  setMbSliceGroupMapType(uiMbMapType);

  switch( uiMbMapType )
  {
  case 0:
    {
      for( UInt i = 0; i <= uiSliceGroupsMinus1; i++)
      {
        RNOK( pcReadIf->getUvlc(uiTmp,  "PPS: run_length_minus1"));
        // !KS Note: no range specified so far
        setRunLength(i,uiTmp+1);
      }
      break;
    }
  case 1:// dispersed
    {
      // nothing to read
      break;
    }
  case 2: // foreground + leftover
    {
      for( UInt i = 0; i < uiSliceGroupsMinus1; i++)
      {
        RNOK( pcReadIf->getUvlc( uiTmp, "PPS: top_left_mb"));
        setTopLeftMb(i,uiTmp);

        RNOK( pcReadIf->getUvlc( uiTmp, "PPS: bottom_right_mb"));
        setBottomRightMb(i,uiTmp);
      }
      break;
    }
  case 3:
  case 4:
  case 5:
    {
      // for these modes uiSliceGroupsMinus1 has to be 1
      ROTR(uiSliceGroupsMinus1!=1, Err::m_nInvalidParameter);

      RNOK( pcReadIf->getFlag(bFlag,"PPS: slice_group_change_direction_flag"));
      setSliceGroupChangeDirectionFlag(bFlag);
      RNOK( pcReadIf->getUvlc(uiTmp,  "PPS: slice_group_change_rate_minus1"));
      setSliceGroupChangeRate(uiTmp+1);
      break;
    }
  case 6: // fmo
    {
      UInt uiNumBits = gCeilLog2( uiSliceGroupsMinus1+1 );
      UInt uiPicSizeInMapUnits;

      RNOK( pcReadIf->getUvlc(uiPicSizeInMapUnits,  "PPS: pic_size_in_map_units_minus1"));
      setPicSizeInMapUnits(uiPicSizeInMapUnits+1);

//      RNOK( m_cSliceGroupIdBuffer.init( getPicSizeInMapUnits() ) );
      for( UInt uiMbIndex = 0; uiMbIndex <= uiPicSizeInMapUnits; uiMbIndex++)
      {
        RNOK( pcReadIf->getCode(uiTmp,uiNumBits,"PPS: slice_group_id"));
//         setSliceGroupId( uiMbIndex, uiTmp );
      }
      break;
    }
  }

  return Err::m_nOK;
}


ErrVal PictureParameterSet::xReadFrext( HeaderSymbolReadIf* pcReadIf )
{
  ROTRS( ! pcReadIf->moreRBSPData(), Err::m_nOK );

  RNOK( pcReadIf->getFlag( m_bTransform8x8ModeFlag, "PPS: transform_8x8_mode_flag") );
  ROT( m_bTransform8x8ModeFlag );

  bool b;
  RNOK( pcReadIf->getFlag( b, "PPS: scaling_matrix_present_flag") );
  ROT(b);

  RNOK( pcReadIf->getSvlc( m_iSecondChromaQpIndexOffset, "PPS: second_chroma_qp_index_offset") );
  ROTR((m_iSecondChromaQpIndexOffset<(-12))||(m_iSecondChromaQpIndexOffset>12), Err::m_nInvalidParameter);

  return Err::m_nOK;
}


ErrVal PictureParameterSet::xWriteFrext( HeaderSymbolWriteIf* pcWriteIf ) const
{
  bool bDefault = (( false == m_bTransform8x8ModeFlag) 
                && ( m_iChromaQpIndexOffset == m_iSecondChromaQpIndexOffset));

  ROTRS( bDefault, Err::m_nOK );

  RNOK( pcWriteIf->writeFlag( m_bTransform8x8ModeFlag, "PPS: transform_8x8_mode_flag") );

  RNOK( pcWriteIf->writeFlag( false, "PPS: scaling_matrix_present_flag") );

  RNOK( pcWriteIf->writeSvlc( m_iSecondChromaQpIndexOffset, "PPS: second_chroma_qp_index_offset") );

  return Err::m_nOK;
}


PictureParameterSet& PictureParameterSet::operator = ( const PictureParameterSet& rcPPS ) 
{
  m_uiPicParameterSetId                     = rcPPS.m_uiPicParameterSetId;
  m_uiSeqParameterSetId                     = rcPPS.m_uiSeqParameterSetId;
  m_bEntropyCodingModeFlag                  = rcPPS.m_bEntropyCodingModeFlag;
  m_bPicOrderPresentFlag                    = rcPPS.m_bPicOrderPresentFlag;
  m_uiNumSliceGroups                        = rcPPS.m_uiNumSliceGroups;
  m_uiMbSliceGroupMapType                   = rcPPS.m_uiMbSliceGroupMapType;
  m_auiRunLength                            = rcPPS.m_auiRunLength;
  m_auiTopLeftMb                            = rcPPS.m_auiTopLeftMb;
  m_auiBottomRightMb                        = rcPPS.m_auiBottomRightMb;
  m_bSliceGroupChangeDirectionFlag          = rcPPS.m_bSliceGroupChangeDirectionFlag;
  m_uiSliceGroupChangeRate                  = rcPPS.m_uiSliceGroupChangeRate;
  m_uiPicSizeInMapUnits                     = rcPPS.m_uiPicSizeInMapUnits;
  m_auiNumRefIdxActive                      = rcPPS.m_auiNumRefIdxActive;
  m_bWeightedPredFlag                       = rcPPS.m_bWeightedPredFlag;
  m_uiWeightedBipredIdc                     = rcPPS.m_uiWeightedBipredIdc;
  m_iPicInitQp                              = rcPPS.m_iPicInitQp;
  m_uiPicInitQs                             = rcPPS.m_uiPicInitQs;
  m_iChromaQpIndexOffset                    = rcPPS.m_iChromaQpIndexOffset;
  m_bDeblockingFilterParametersPresentFlag  = rcPPS.m_bDeblockingFilterParametersPresentFlag;
  m_bConstrainedIntraPredFlag               = rcPPS.m_bConstrainedIntraPredFlag;
  m_bRedundantPicCntPresentFlag             = rcPPS.m_bRedundantPicCntPresentFlag,

  m_iSecondChromaQpIndexOffset              = rcPPS.m_iSecondChromaQpIndexOffset;
  m_iSecondChromaQpIndexOffset              = rcPPS.m_iSecondChromaQpIndexOffset;
	m_bTransform8x8ModeFlag                   = rcPPS.m_bTransform8x8ModeFlag;


  return *this;
}

bool PictureParameterSet::operator != ( const PictureParameterSet& rcPPS ) const
{
  ROTRS( m_uiPicParameterSetId            != rcPPS.m_uiPicParameterSetId,             true );
  ROTRS( m_uiSeqParameterSetId            != rcPPS.m_uiSeqParameterSetId,             true );

  ROTRS( m_bEntropyCodingModeFlag         != rcPPS.m_bEntropyCodingModeFlag,          true );
  ROTRS( m_bPicOrderPresentFlag           != rcPPS.m_bPicOrderPresentFlag,            true );
  ROTRS( m_uiNumSliceGroups               != rcPPS.m_uiNumSliceGroups,                true );
  ROTRS( m_uiMbSliceGroupMapType          != rcPPS.m_uiMbSliceGroupMapType,           true );

  ROTRS( m_auiRunLength                   != rcPPS.m_auiRunLength,                    true );
  ROTRS( m_auiTopLeftMb                   != rcPPS.m_auiTopLeftMb,                    true );
  ROTRS( m_auiBottomRightMb               != rcPPS.m_auiBottomRightMb,                true );


  ROTRS( m_bSliceGroupChangeDirectionFlag != rcPPS.m_bSliceGroupChangeDirectionFlag,  true );
  ROTRS( m_uiSliceGroupChangeRate         != rcPPS.m_uiSliceGroupChangeRate,          true );
  ROTRS( m_uiPicSizeInMapUnits            != rcPPS.m_uiPicSizeInMapUnits,             true );


  ROTRS( m_auiNumRefIdxActive             != rcPPS.m_auiNumRefIdxActive,              true );
  ROTRS( m_bWeightedPredFlag              != rcPPS.m_bWeightedPredFlag,               true );
  ROTRS( m_uiWeightedBipredIdc            != rcPPS.m_uiWeightedBipredIdc,             true );
  ROTRS( m_iPicInitQp                     != rcPPS.m_iPicInitQp,                      true );
  ROTRS( m_uiPicInitQs                    != rcPPS.m_uiPicInitQs,                     true );
  ROTRS( m_iChromaQpIndexOffset           != rcPPS.m_iChromaQpIndexOffset,            true );
  ROTRS( m_bConstrainedIntraPredFlag      != rcPPS.m_bConstrainedIntraPredFlag,       true );
  ROTRS( m_bRedundantPicCntPresentFlag    != rcPPS.m_bRedundantPicCntPresentFlag,     true );

  ROTRS( m_bDeblockingFilterParametersPresentFlag != rcPPS.m_bDeblockingFilterParametersPresentFlag, true );

  ROTRS( m_iSecondChromaQpIndexOffset     != rcPPS.m_iSecondChromaQpIndexOffset,      true );
  ROTRS( m_bTransform8x8ModeFlag          != rcPPS.m_bTransform8x8ModeFlag,           true );
  
  return false;
}




// h264 namespace end
VIDEO_NAMESPACE_END
