/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/H264SequenceParameterSet.h"
#include "VideoCommonLib/LevelLimit.h"
#include "VideoCommonLib/AccessUnit.h"
//#include <cmath>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

ErrVal SequenceParameterSet::setAccessUnitParameters( const AccessUnitParameters &rcPars, const UInt uiLog2MaxPicOrderCntLsb, const UInt uiLevelIdc, const UInt uiProfileIdc )
{

  m_eProfileIdc                       = Profile(uiProfileIdc);
  m_bConstrainedSet0Flag              = false;
  m_bConstrainedSet1Flag              = false;
  m_bConstrainedSet2Flag              = false;
  m_uiLevelIdc                        = uiLevelIdc;
  m_uiSeqParameterSetId               = 0;
  m_uiLog2MaxFrameNum                 = 4; // tbd
  m_uiPicOrderCntType                 = 0;
  m_uiLog2MaxPicOrderCntLsb           = uiLog2MaxPicOrderCntLsb;
  m_bDeltaPicOrderAlwaysZeroFlag      = true;
  m_iOffsetForNonRefPic               = 0;
  m_iOffsetForTopToBottomField        = 0;
  m_uiNumRefFramesInPicOrderCntCycle  = 0;
//  m_piOffsetForRefFrame             ;
  m_uiNumRefFrames                    = rcPars.getMaxDPBSize();
  m_bGapsInFrameNumValueAllowedFlag   = false;
  m_uiFrameWidthInMbs                 = (rcPars.getPicSize().getX()+15)/16;
  m_uiFrameHeightInMbs                = (rcPars.getPicSize().getY()+15)/16;
  m_uiChromaFormatParameterX2         = 0; //??
  m_bFrameMbsOnlyFlag                 = true;
  m_bMbAdaptiveFrameFieldFlag         = false;
  m_bDirect8x8InferenceFlag           = true;
  m_bFrameCroppingFlag                = false;
//  m_cFrameCroppingRect                ;
  m_bVuiParametersPresentFlag         = false;
  m_pcVUI                             = NULL;
  m_bInitDone                         = true;


  m_bConstrainedSet3Flag              = false;
  m_uiBitDepthLuma                    = 8;
  m_uiBitDepthChroma                  = 8;
  m_bQPPrimeTransformBypassFlag       = false;
  //m_cSeqScalingMatrix                 ;
  m_bResidueTransformFlag             = false;

  if( 1 == rcPars.getSubsampling().size() )
  {
    m_eChromaFormatIdc = YUV400;
  }
  else if( 3 == rcPars.getSubsampling().size() )
  {
    if( 0 == rcPars.getSubsampling(1).getX().getDownShift() )
    {
      if( 0 == rcPars.getSubsampling(1).getY().getDownShift() )
      {
        m_eChromaFormatIdc = YUV444;
      }
      else
      {
        AOT(1);
      }
    }
    else if( 1 == rcPars.getSubsampling(1).getX().getDownShift() )
    {
      if( 0 == rcPars.getSubsampling(1).getY().getDownShift() )
      {
        m_eChromaFormatIdc = YUV422;
      }
      else if( 1 == rcPars.getSubsampling(1).getY().getDownShift() )
      {
        m_eChromaFormatIdc = YUV420;
      }
      else
      {
        AOT(1);
      }
    }
    else
    {
      AOT(1);
    }
  }
  else
  {
    AOT(1);
  }

  AOT( getMaxDPBSize() < rcPars.getMaxDPBSize() );

  int iXDiff = 16*m_uiFrameWidthInMbs  - rcPars.getPicSize().getX();
  int iYDiff = 16*m_uiFrameHeightInMbs - rcPars.getPicSize().getY();


  if( 0 != iXDiff || 0 != iYDiff )
  {
    m_bFrameCroppingFlag = true;
    m_cFrameCroppingRect.setLeftOffset(0);
    m_cFrameCroppingRect.setTopOffset(0);
    if( m_eChromaFormatIdc == YUV420 || m_eChromaFormatIdc == YUV422 )
    {
      AOT( 0 != (iXDiff & 1) );
      iXDiff /= 2;
    }

    if( m_eChromaFormatIdc == YUV420 )
    {
      if( m_bFrameMbsOnlyFlag )
      {
        AOT( 0 != (iYDiff & 1) );
        iYDiff /= 2;
      }
      else
      {
        AOT( 0 != (iXDiff & 3) );
        iYDiff /= 4;
      }
    }
    else
    {
      if( false == m_bFrameMbsOnlyFlag )
      {
        AOT( 0 != (iYDiff & 1) );
        iYDiff /= 2;
      }
    }

    m_cFrameCroppingRect.setRightOffset( iXDiff );  //((m_eChromaFormatIdc == YUV420 || m_eChromaFormatIdc == YUV422) ? 2 : 1);
    m_cFrameCroppingRect.setBottomOffset( iYDiff);  //(m_eChromaFormatIdc == YUV420 ? 2 : 1) *  (m_bFrameMbsOnlyFlag ? 1:2);
  }



  return Err::m_nOK;
}
ErrVal SequenceParameterSet::getAccessUnitParameters( AccessUnitParameters &rcPars, const UInt uiPlaneGrouping ) const
{
  std::vector<Subsampling> cSubsampling( 1, Subsampling( ShiftScale(0), ShiftScale(0)) );
  switch( m_eChromaFormatIdc )
  {
  case YUV400:
    {
    }
    break;
  case YUV420:
    {
      cSubsampling .push_back( Subsampling( ShiftScale(1), ShiftScale(1) ) );
      cSubsampling .push_back( Subsampling( ShiftScale(1), ShiftScale(1) ) );
    }
    break;
  case YUV422:
    {
      cSubsampling .push_back( Subsampling( ShiftScale(1), ShiftScale(0) ) );
      cSubsampling .push_back( Subsampling( ShiftScale(1), ShiftScale(0) ) );
    }
    break;
  case YUV444:
    {
      cSubsampling .push_back( Subsampling( ShiftScale(0), ShiftScale(0) ) );
      cSubsampling .push_back( Subsampling( ShiftScale(0), ShiftScale(0) ) );
    }
    break;

  default:
    AOT(1);
    break;
  }

  if( m_bFrameCroppingFlag )
  {
    AOT( 0 != m_cFrameCroppingRect.getLeftOffset() )
    AOT( 0 != m_cFrameCroppingRect.getTopOffset() )
    rcPars = AccessUnitParameters( PicOrderCnt( 0 ), m_uiFrameWidthInMbs*16-m_cFrameCroppingRect.getRightOffsetInLumaPelUnits( m_eChromaFormatIdc), m_uiFrameHeightInMbs*16-m_cFrameCroppingRect.getBotOffsetInLumaPelUnits( m_eChromaFormatIdc, m_bFrameMbsOnlyFlag), cSubsampling, uiPlaneGrouping );
  }
  else
  {
    rcPars = AccessUnitParameters( PicOrderCnt( 0 ), m_uiFrameWidthInMbs*16, m_uiFrameHeightInMbs*16, cSubsampling, uiPlaneGrouping );
  }

  rcPars.setMaxDPBSize( getMaxDPBSize() );

  const int iPlaneQp = 26;
  for( UInt uiPlaPaId = 0; uiPlaPaId < rcPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    for( UInt ui = 0; ui < rcPars.getPlaPaPars( uiPlaPaId ).getNumPlanes(); ui++ )
    {
      rcPars.getPlaPaPars( uiPlaPaId ).getPlaneQps().push_back( iPlaneQp );
    }
  }
    
  return Err::m_nOK;
}


SequenceParameterSet::FrameCroppingRect::FrameCroppingRect()
: m_uiLeftOffset    ( 0 )
, m_uiRightOffset   ( 0 )
, m_uiTopOffset     ( 0 )
, m_uiBottomOffset  ( 0 )
{
}

SequenceParameterSet::FrameCroppingRect::~FrameCroppingRect()
{
}

ErrVal SequenceParameterSet::FrameCroppingRect::write( HeaderSymbolWriteIf* pcWriteIf ) const
{
  RNOK( pcWriteIf->writeUvlc( m_uiLeftOffset,  "SPS: frame_cropping_rect_left_offset"   ) );
  RNOK( pcWriteIf->writeUvlc( m_uiRightOffset, "SPS: frame_cropping_rect_right_offset"  ) );
  RNOK( pcWriteIf->writeUvlc( m_uiTopOffset,   "SPS: frame_cropping_rect_top_offset"    ) );
  RNOK( pcWriteIf->writeUvlc( m_uiBottomOffset,"SPS: frame_cropping_rect_bottom_offset" ) );
  return Err::m_nOK;
}

ErrVal SequenceParameterSet::FrameCroppingRect::read( HeaderSymbolReadIf* pcReadIf )
{
  RNOK( pcReadIf->getUvlc( m_uiLeftOffset,  "SPS: frame_cropping_rect_left_offset"  ) );
  RNOK( pcReadIf->getUvlc( m_uiRightOffset, "SPS: frame_cropping_rect_right_offset" ) );
  RNOK( pcReadIf->getUvlc( m_uiTopOffset,   "SPS: frame_cropping_rect_top_offset"   ) );
  RNOK( pcReadIf->getUvlc( m_uiBottomOffset,"SPS: frame_cropping_rect_bottom_offset") );
  return Err::m_nOK;
}


bool SequenceParameterSet::FrameCroppingRect::operator != ( const FrameCroppingRect& rcFCR ) const
{
  ROTRS( m_uiLeftOffset   != rcFCR.m_uiLeftOffset,    true ) ;
  ROTRS( m_uiRightOffset  != rcFCR.m_uiRightOffset,   true ) ;
  ROTRS( m_uiTopOffset    != rcFCR.m_uiTopOffset,     true ) ;
  ROTRS( m_uiBottomOffset != rcFCR.m_uiBottomOffset,  true ) ;
  return false;
}


SequenceParameterSet& SequenceParameterSet::operator = ( const SequenceParameterSet& rcSPS )
{
  m_eProfileIdc                       = rcSPS.m_eProfileIdc;
  m_bConstrainedSet0Flag              = rcSPS.m_bConstrainedSet0Flag;
  m_bConstrainedSet1Flag              = rcSPS.m_bConstrainedSet1Flag;
  m_bConstrainedSet2Flag              = rcSPS.m_bConstrainedSet2Flag;
  m_uiLevelIdc                        = rcSPS.m_uiLevelIdc;
  m_uiSeqParameterSetId               = rcSPS.m_uiSeqParameterSetId;
  m_uiLog2MaxFrameNum                 = rcSPS.m_uiLog2MaxFrameNum;
  m_uiPicOrderCntType                 = rcSPS.m_uiPicOrderCntType;
  m_uiLog2MaxPicOrderCntLsb           = rcSPS.m_uiLog2MaxPicOrderCntLsb;
  m_bDeltaPicOrderAlwaysZeroFlag      = rcSPS.m_bDeltaPicOrderAlwaysZeroFlag;
  m_iOffsetForNonRefPic               = rcSPS.m_iOffsetForNonRefPic;
  m_iOffsetForTopToBottomField        = rcSPS.m_iOffsetForTopToBottomField;
  m_uiNumRefFramesInPicOrderCntCycle  = rcSPS.m_uiNumRefFramesInPicOrderCntCycle;
  m_piOffsetForRefFrame               = rcSPS.m_piOffsetForRefFrame;
  m_uiNumRefFrames                    = rcSPS.m_uiNumRefFrames;
  m_bGapsInFrameNumValueAllowedFlag   = rcSPS.m_bGapsInFrameNumValueAllowedFlag;
  m_uiFrameWidthInMbs                 = rcSPS.m_uiFrameWidthInMbs;
  m_uiFrameHeightInMbs                = rcSPS.m_uiFrameHeightInMbs;
  m_uiChromaFormatParameterX2         = rcSPS.m_uiChromaFormatParameterX2;
  m_bFrameMbsOnlyFlag                 = rcSPS.m_bFrameMbsOnlyFlag;
  m_bMbAdaptiveFrameFieldFlag         = rcSPS.m_bMbAdaptiveFrameFieldFlag;
  m_bDirect8x8InferenceFlag           = rcSPS.m_bDirect8x8InferenceFlag;
  m_bFrameCroppingFlag                = rcSPS.m_bFrameCroppingFlag;
  m_cFrameCroppingRect                = rcSPS.m_cFrameCroppingRect;
  m_bVuiParametersPresentFlag         = rcSPS.m_bVuiParametersPresentFlag;
  m_bInitDone                         = rcSPS.m_bInitDone;


	m_bConstrainedSet3Flag              = rcSPS.m_bConstrainedSet3Flag;
	m_eChromaFormatIdc                  = rcSPS.m_eChromaFormatIdc;
	m_uiBitDepthLuma                    = rcSPS.m_uiBitDepthLuma;
	m_uiBitDepthChroma                  = rcSPS.m_uiBitDepthChroma;
	m_bQPPrimeTransformBypassFlag       = rcSPS.m_bQPPrimeTransformBypassFlag;
  m_bResidueTransformFlag             = rcSPS.m_bResidueTransformFlag;

  return *this;
}


SequenceParameterSet::SequenceParameterSet()
  : m_eProfileIdc                           ( BASELINE_PROFILE )
  , m_uiLevelIdc                            ( 51 )
  , m_uiSeqParameterSetId                   ( -1 )
  , m_uiLog2MaxFrameNum                     ( 4 )
  , m_bFrameMbsOnlyFlag                     ( true )
  , m_bDeltaPicOrderAlwaysZeroFlag          ( false )
  , m_bMbAdaptiveFrameFieldFlag             ( false )
  , m_bDirect8x8InferenceFlag               ( false )
  , m_uiPicOrderCntType                     ( 0 )
  , m_uiLog2MaxPicOrderCntLsb               ( 4 )
  , m_iOffsetForNonRefPic                   ( 0 )
  , m_iOffsetForTopToBottomField            ( 0 )
  , m_uiNumRefFramesInPicOrderCntCycle      ( 0 )
  , m_uiNumRefFrames                        ( 0 )
  , m_uiFrameWidthInMbs                     ( 0 )
  , m_uiFrameHeightInMbs                    ( 0 )
  , m_uiChromaFormatParameterX2             ( 3 )
  , m_uiSPSRegId                            ( 0 )
  , m_pcVUI                                 ( NULL )
  , m_eChromaFormatIdc                      ( YUV420 )
	, m_bConstrainedSet3Flag                  ( false )
	, m_bQPPrimeTransformBypassFlag           ( false )
  , m_bResidueTransformFlag                 ( false )
  , m_uiBitDepthLuma                        ( 8 )
  , m_uiBitDepthChroma                      ( 8 )
  , m_bConstrainedSet0Flag                  ( false )
  , m_bConstrainedSet1Flag                  ( false )
  , m_bConstrainedSet2Flag                  ( false )
  , m_bFrameCroppingFlag                    ( false )
  , m_bGapsInFrameNumValueAllowedFlag       ( false )
  , m_bVuiParametersPresentFlag             ( false )
  , m_bInitDone                             ( false )
{
}

SequenceParameterSet::~SequenceParameterSet()
{
  m_piOffsetForRefFrame.uninit();
}


ErrVal SequenceParameterSet::create( SequenceParameterSet*& rpcSPS )
{
  rpcSPS = new SequenceParameterSet;
  ROT( NULL == rpcSPS);
  return Err::m_nOK;
}

ErrVal SequenceParameterSet::destroy()
{
  delete this;
  return Err::m_nOK;
}


UInt SequenceParameterSet::getMaxDPBSize() const
{
//  UInt uiFrameSize = getMbByteSize()*getMbInFrame();
  UInt uiFrameSize = 384*getMbInFrame();
  const LevelLimit::LevelLimitStruct *pcLevelLimit;
  ANOK( LevelLimit::getLevelLimit( pcLevelLimit, getLevelIdc() ) );
  return gMin(16, (int)(pcLevelLimit->uiMaxDPBSizeX2 / ( 2*uiFrameSize)) );
}

bool SequenceParameterSet::check() const
{
  ROFRS( xCheckProfile(),     false );
  ROFRS( xCheckLevel(),       false );
  ROFRS( xCheckIndependent(), false );
  return true;
}

bool SequenceParameterSet::xCheckLevel() const
{

//  ROFRS( getFrameMbsOnlyFlag(), false);
  return true;
}




bool SequenceParameterSet::xCheckIndependent() const
{
  //  find the limits for the current level
  const LevelLimit::LevelLimitStruct* pcLevelLimit;
  ROTR( LevelLimit::getLevelLimit( pcLevelLimit, getLevelIdc() ), false);

//a)	The difference between  consecutive removal times satisfies the constraint that tr(n) ? tr(n?1) >= gMax( PicWidthInMbs * PicHeightInMbs  MaxMBPS, FR), where MaxMBPS is specified in Table A?1. If picture n-1 is a frame, FR = 1  172.  If picture n-1 is a field, FR = 1  (172 * 2). [Ed. (GJS): Also include the similar constraint on output times.]
//b)	the total number of bytes of NAL unit data (i.e., the total of the NumBytesInNALunit variables for the picture) associated with coded picture n is <= PicWidthInMbs * PicHeightInMbs * 256 * ChromaFormatParameter * MaxMBPS * ( tr(n) ? tr(n?1) )  MinCR.  [Ed. Note (JB/GJS): Need to count the non-VCL bits in this calculation too - take all bits associated with the picture.]
  ROTRS( getMbInFrame() > pcLevelLimit->uiMaxFrameSize, false ) ; //c
  UInt uiSize = (UInt)sqrt( (double) pcLevelLimit->uiMaxFrameSize * 8 );
  ROTRS( getFrameWidthInMbs()  > uiSize,                false ) ; //d
  ROTRS( getFrameHeightInMbs() > uiSize,                false ) ; //e
//f)	The total number of CABAC bins [Ed(AG): How counted?] for any coded picture <= 32 * BytesInPic / 3 + 96 * ( frame_width_in_mbs_minus1+1 ) * ( frame_height_in_mbs_minus1+1 ) where BytesInPic is the total sum of NumBytesInNALunit of all NAL units that contain primary slices and slice data partitions, for any coded picture.
//g)	DPB size max_dec_frame_buffering <= 1024 * MaxDPB / ( ( frame_width_in_mbs_minus1+1 ) * ( frame_height_in_mbs_minus1+1 ) * 256 * ChromaFormatParameter ). MaxDPB is given in Table A?1. max-dec_frame_buffering is a sequence VUI parameter. DPB size is used by the HRD (see Annex C).
//h)	bit_rate[ k ] <= MaxBR and cpb_size[ k ] <= the level limit given, where bit_rate[ k ] is given by Equation E-1 and cpb_size[ k ] is given by Equation E-2 if vcl_hrd_parameters_present_flag is equal to 1. MaxBR is specified in Table A?1. The bitstream shall satisfy these conditions for at least one value of k in the range 0 to cpb_cnt_minus1. [Ed. Note (GJS/AG): When HRD parameters not present, the bitstream must still obey something.]
//i)	and bit_rate[ k ] <= 1.2 * MaxBR for bit_rate_value[ k ] and cpb_size[ k ] <= 1.2 * MaxCPB, where bit_rate[ k ] is given by Equation E-1 and cpb_size[ k ] is given by Equation E-2 if nal_hrd_parameters_present_flag equal to 1. MaxBR and MaxCPB are specified in Table A?1.  The bitstream shall satisfy these conditions for at least one value of k in the range 0 to cpb_cnt_minus1. [Ed. Note (GJS/AG): Consider when HRD parameters not present.]
//j)	required CPB buffer size is <= the level limit given
//k)	vertical motion vector component range does not exceed MaxVmvR in units of luma frame samples, where MaxVmvR is specified in Table A?1
//l)	horizontal motion vector range does not exceed [-2048, 2047.75] in units of luma samples
//m)	number of motion vectors per two consecutive macroblocks in decoding order (including from the last macroblock of a slice to the first macroblock of the next slice in decoding order) does not exceed MaxMvsPer2Mb, where  MaxMvsPer2Mb is specified in Table A?1.
//n)	number of bits of macroblock_layer( ) data for any macroblock is not greater than 128 + 2048 * ChromaFormatParameter. [Ed. Note (AG): How counted for CABAC?]

  return true;
}


bool SequenceParameterSet::xCheckProfile() const
{
  Profile eProfile = getProfileIdc();
  switch( eProfile )
  {
  case BASELINE_PROFILE:
    {
      ROFRS( getFrameMbsOnlyFlag(), false);
    }
    break;
  case MAIN_PROFILE:
    {
      /* !KS Todo: check constrained flags 
      ROTRS( getArbitrarySliceOrderAllowedFlag(),   false);
      ROTRS( getMoreThanOneSliceGroupAllowedFlag(), false);
      ROTRS( getRedundantSlicesAllowedFlag(),       false);
      */
    }
    break;
  case EXTENDED_PROFILE:
    {

    }
    break;
  default:
    {
      AOT(1);
      return false;
    }
  }
  return true;
}


ErrVal SequenceParameterSet::write( HeaderSymbolWriteIf* pcWriteIf ) const
{
  UInt uiLevelIdc           = getLevelIdc();

  RNOK( pcWriteIf->writeCode( getProfileIdc(), 8,"SPS: profile_idc"));

  RNOK( pcWriteIf->writeFlag( m_bConstrainedSet0Flag, "SPS: constrained_set0_flag" ) );
  RNOK( pcWriteIf->writeFlag( m_bConstrainedSet1Flag, "SPS: constrained_set1_flag" ) );
  RNOK( pcWriteIf->writeFlag( m_bConstrainedSet2Flag, "SPS: constrained_set2_flag" ) );
  bool bConstrainedSet3Flag = m_bConstrainedSet3Flag;
  if ( ((m_eProfileIdc == BASELINE_PROFILE) || (m_eProfileIdc == MAIN_PROFILE) || (m_eProfileIdc == EXTENDED_PROFILE)) && (9 == uiLevelIdc) )
  {
    bConstrainedSet3Flag = 1;
    uiLevelIdc = 11;
  }
  RNOK( pcWriteIf->writeFlag( bConstrainedSet3Flag,   "SPS: constrained_set3_flag" ) );
  RNOK( pcWriteIf->writeCode( 0, 4,                   "SPS: reserved_zero_4bits" ) );
  RNOK( pcWriteIf->writeCode( uiLevelIdc, 8,          "SPS: level_idc" ) );

  RNOK( pcWriteIf->writeUvlc( getSeqParameterSetId(), "SPS: seq_parameter_set_id"));
  RNOK( xWriteFrext( pcWriteIf ) );
  
  UInt uiTmp = gMax(((int) getLog2MaxFrameNum())-4,0);
  RNOK( pcWriteIf->writeUvlc( uiTmp,                  "SPS: log2_max_frame_num_minus_4"));

  RNOK (xWritePicOrderCntInfo( pcWriteIf ) );

  RNOK( pcWriteIf->writeUvlc( getNumRefFrames(),      "SPS: num_ref_frames"));
  RNOK( pcWriteIf->writeFlag( getGapsInFrameNumValueAllowedFlag(),"SPS: gaps_in_frame_num_value_allowed_flag"));
  RNOK( pcWriteIf->writeUvlc( getFrameWidthInMbs()-1, "SPS: pic_height_in_map_units_minus1"));

  UInt uiHeight = ( getFrameHeightInMbs()-1) >> (1- getFrameMbsOnlyFlag());

  RNOK( pcWriteIf->writeUvlc( uiHeight,               "SPS: frame_height_in_mbs_minus_1"));
  RNOK( pcWriteIf->writeFlag( getFrameMbsOnlyFlag(),  "SPS: frame_mbs_only_flag"));

  if( ! getFrameMbsOnlyFlag() )
  {
    RNOK( pcWriteIf->writeFlag( getMbAdaptiveFrameFieldFlag(),"SPS: mb_adaptive_frame_field_flag"));
  }

  RNOK( pcWriteIf->writeFlag( getDirect8x8InferenceFlag(),  "SPS: direct_8x8_inference_flag"));
  RNOK( pcWriteIf->writeFlag( getFrameCroppingFlag(),       "SPS: frame_cropping_flag"));
  if( getFrameCroppingFlag() )
  {
    RNOK( m_cFrameCroppingRect.write( pcWriteIf ) );
  }

  RNOK( pcWriteIf->writeFlag( getVuiParametersPresentFlag(),"SPS: vui_parameters_present_flag"));
  if( getVuiParametersPresentFlag() )
  {
    ROT(1);
  }

  return Err::m_nOK;
}


ErrVal SequenceParameterSet::xWritePicOrderCntInfo( HeaderSymbolWriteIf* pcWriteIf ) const
{
  RNOK( pcWriteIf->writeUvlc( getPicOrderCntType(),"SPS: pic_order_cnt_type"));

  if( 0 == getPicOrderCntType() )
  {
    ROT ( getLog2MaxPicOrderCntLsb() < 4);
    RNOK( pcWriteIf->writeUvlc(getLog2MaxPicOrderCntLsb() - 4,"SPS: log2_max_pic_order_cnt_lsb_minus4"));
    return Err::m_nOK;
  }

  if( 1== getPicOrderCntType() )
  {
    RNOK( pcWriteIf->writeFlag( getDeltaPicOrderAlwaysZeroFlag(),   "SPS: delta_pic_order_always_zero_flag"));
    RNOK( pcWriteIf->writeSvlc( getOffsetForNonRefPic(),            "SPS: offset_for_non_ref_pic"));
    RNOK( pcWriteIf->writeSvlc( getOffsetForTopToBottomField(),     "SPS: offset_for_top_to_bottom_field"));
    RNOK( pcWriteIf->writeUvlc( getNumRefFramesInPicOrderCntCycle(),"SPS: num_ref_frames_in_pic_order_cnt_cycle"));

    UInt uiNumRefFrames = getNumRefFramesInPicOrderCntCycle();
    for( UInt i = 0; i < uiNumRefFrames; i++)
    {
      RNOK( pcWriteIf->writeSvlc( getOffsetForRefFrame(i),"SPS: offset_for_ref_frame"));
    }
  }

  return Err::m_nOK;
}


ErrVal SequenceParameterSet::xReadFrext( HeaderSymbolReadIf* pcReadIf )
{
  ROTRS( m_eProfileIdc < HIGH_PROFILE, Err::m_nOK );

  UInt uiChromaFormatIdc;
  RNOK( pcReadIf->getUvlc( uiChromaFormatIdc,             "SPS: chroma_format_idc"));
  m_eChromaFormatIdc = ChromaFormatIdc (uiChromaFormatIdc);
  if( m_eChromaFormatIdc == YUV444 )
  {
    RNOK( pcReadIf->getFlag( m_bResidueTransformFlag,         "SPS: residue_transform_flag"));
    ROT( m_bResidueTransformFlag );
  }
  RNOK( pcReadIf->getUvlc( m_uiBitDepthLuma,                "SPS: bit_depth_luma_minus8"));
  m_uiBitDepthLuma  +=8;
  RNOK( pcReadIf->getUvlc( m_uiBitDepthChroma,              "SPS: bit_depth_chroma_minus8"));
  m_uiBitDepthChroma+=8;
  RNOK( pcReadIf->getFlag( m_bQPPrimeTransformBypassFlag,   "SPS: qpprime_y_zero_transform_bypass_flag"));
  ROT( m_bQPPrimeTransformBypassFlag );

  bool b;
  RNOK( pcReadIf->getFlag( b,   "SPS: scaling_matrix_present_flag"));
  ROT(b);

  ROTR( m_uiBitDepthLuma   > 8, Err::m_nDataNotAvailable );
  ROTR( m_uiBitDepthChroma > 8, Err::m_nDataNotAvailable );
//  ROTR( m_eChromaFormatIdc > YUV420, Err::m_nDataNotAvailable );


  return Err::m_nOK;
}

ErrVal SequenceParameterSet::xWriteFrext( HeaderSymbolWriteIf* pcWriteIf ) const
{
  ROTRS( m_eProfileIdc < HIGH_PROFILE, Err::m_nOK );

  RNOK( pcWriteIf->writeUvlc( m_eChromaFormatIdc,            "SPS: chroma_format_idc"));
  if ( m_eChromaFormatIdc == YUV444 )
  {
    RNOK( pcWriteIf->writeFlag( m_bResidueTransformFlag,      "SPS: residue_transform_flag"));
  }
  RNOK( pcWriteIf->writeUvlc( m_uiBitDepthLuma-8,             "SPS: bit_depth_luma_minus8"));
  RNOK( pcWriteIf->writeUvlc( m_uiBitDepthChroma-8,           "SPS: bit_depth_chroma_minus8"));
  RNOK( pcWriteIf->writeFlag( m_bQPPrimeTransformBypassFlag,  "SPS: qpprime_y_zero_transform_bypass_flag"));

  RNOK( pcWriteIf->writeFlag( false,   "SPS: scaling_matrix_present_flag"));

  return Err::m_nOK;
}


ErrVal SequenceParameterSet::read( HeaderSymbolReadIf* pcReadIf )
{
  UInt uiTmp;

  UInt uiDummy;
  RNOK( pcReadIf->getCode( uiDummy, 8, "SPS: profile_idc"));
  m_eProfileIdc = (Profile)uiDummy;

  ROTR( m_eProfileIdc != BASELINE_PROFILE 
        && m_eProfileIdc != MAIN_PROFILE 
        && m_eProfileIdc != EXTENDED_PROFILE 
        && m_eProfileIdc != HIGH_PROFILE 
        && m_eProfileIdc != THIS_PROFILE 
/*        && m_eProfileIdc != HIGH_10_PROFILE 
        && m_eProfileIdc != HIGH_422_PROFILE 
        && m_eProfileIdc != HIGH_444_PROFILE
*/        , Err::m_nInvalidParameter );

  RNOK( pcReadIf->getFlag( m_bConstrainedSet0Flag,  "SPS: constrained_set0_flag" ) );
  RNOK( pcReadIf->getFlag( m_bConstrainedSet1Flag,  "SPS: constrained_set1_flag" ) );
  RNOK( pcReadIf->getFlag( m_bConstrainedSet2Flag,  "SPS: constrained_set2_flag" ) );

  ROTR( ((m_eProfileIdc >= HIGH_PROFILE ) && (m_bConstrainedSet0Flag || m_bConstrainedSet1Flag ||m_bConstrainedSet2Flag)), 
            Err::m_nInvalidParameter );

  RNOK( pcReadIf->getFlag( m_bConstrainedSet3Flag,  "SPS: constrained_set3_flag"   ) );
  RNOK( pcReadIf->getCode( uiTmp, 4,                "SPS: reserved_zero_4bits"   ) );

  ROT( 0 != uiTmp) //SPS: reserved_zero_4bits not zero

  RNOK( pcReadIf->getCode( m_uiLevelIdc,  8,        "SPS: level_idc"));

  {
    // level limit check
    // check for level 1b
    if ( ((m_eProfileIdc == BASELINE_PROFILE) || (m_eProfileIdc == MAIN_PROFILE) || (m_eProfileIdc == EXTENDED_PROFILE)) && (true == m_bConstrainedSet3Flag) )
    {
      // in Baseline, Main and Extended level 1.b is indicted by constrained_set3_flag plus level 1.1
      ROTR (11 != m_uiLevelIdc, Err::m_nERR);
      m_uiLevelIdc = 9;
    }
    const LevelLimit::LevelLimitStruct* pcLevelLimitStruct;
    ErrVal nRet = LevelLimit::getLevelLimit( pcLevelLimitStruct, m_uiLevelIdc);
//    ERRLOG_CMSG( Err::m_nOK != nRet, "SPS: level not supported" );
    RNOK( nRet );
  }

  RNOK( pcReadIf->getUvlc( m_uiSeqParameterSetId,   "SPS: seq_parameter_set_id"));
  ROF( 0 == m_uiSeqParameterSetId );
  ROTR( m_uiSeqParameterSetId > 31, Err::m_nInvalidParameter );

  RNOK( xReadFrext( pcReadIf ) );
  
  RNOK( pcReadIf->getUvlc( uiTmp,                   "SPS: log2_max_frame_num_minus_4"));
  ROTR(uiTmp>12, Err::m_nInvalidParameter);
  setLog2MaxFrameNum(uiTmp+4);

  RNOK (xReadPicOrderCntInfo( pcReadIf ) );

  RNOK( pcReadIf->getUvlc( m_uiNumRefFrames,        "SPS: num_ref_frames"));

  RNOK( pcReadIf->getFlag( m_bGapsInFrameNumValueAllowedFlag,"SPS: gaps_in_frame_num_value_allowed_flag"));
  ROT( m_bGapsInFrameNumValueAllowedFlag );
  // Note: frame width/height restrictions are profile/level specific (no check here)

  RNOK( pcReadIf->getUvlc( uiTmp,                   "SPS: pic_width_in_mbs_minus1"));
  setFrameWidthInMbs(uiTmp+1);
  RNOK( pcReadIf->getUvlc( uiTmp,                   "SPS: pic_height_in_map_units_minus1"));
  RNOK( pcReadIf->getFlag( m_bFrameMbsOnlyFlag,     "SPS: frame_mbs_only_flag"));

  if( getFrameMbsOnlyFlag() )
  {
    setFrameHeightInMbs(uiTmp+1);
    setMbAdaptiveFrameFieldFlag(0);
  }
  else
  {
    ROT( 1 );
    setFrameHeightInMbs((uiTmp+1)<<1);
    RNOK( pcReadIf->getFlag( m_bMbAdaptiveFrameFieldFlag, "SPS: mb_adaptive_frame_field_flag"));
  }

  RNOK( pcReadIf->getFlag( m_bDirect8x8InferenceFlag,     "SPS: direct_8x8_inference_flag"));
  if( m_eProfileIdc == BASELINE_PROFILE )
  {
//    ROTR( m_bDirect8x8InferenceFlag,   Err::m_nInvalidParameter );
  }
  else if( m_eProfileIdc == EXTENDED_PROFILE )
  {
//    ROTR( ! m_bDirect8x8InferenceFlag,   Err::m_nInvalidParameter );
  }
  else
  {
    if ( m_uiLevelIdc >= 30 && ! m_bDirect8x8InferenceFlag) 
    {
      ROT(1) 
//      ERRLOG_MSG("direct_8x8_inference_flag not set in level > 3.0");
    }
    //ROTR( m_uiLevelIdc >= 30 && ! m_bDirect8x8InferenceFlag, Err::m_nInvalidParameter );
  }
  ROTRS( m_bMbAdaptiveFrameFieldFlag && ! m_bDirect8x8InferenceFlag, Err::m_nInvalidParameter );

  RNOK( pcReadIf->getFlag( m_bFrameCroppingFlag,          "SPS: frame_cropping_flag"));

  if( getFrameCroppingFlag() )
  {
    RNOK( m_cFrameCroppingRect.read( pcReadIf ) );
  }

  RNOK( pcReadIf->getFlag( m_bVuiParametersPresentFlag,   "SPS: vui_parameters_present_flag"));

  ROT(m_bVuiParametersPresentFlag)

  return Err::m_nOK;
}

ErrVal SequenceParameterSet::xReadPicOrderCntInfo( HeaderSymbolReadIf* pcReadIf )
{
  RNOK( pcReadIf->getUvlc( m_uiPicOrderCntType,"SPS: pic_order_cnt_type"));
  ROF( 0 == m_uiPicOrderCntType );
  ROTR(m_uiPicOrderCntType>2, Err::m_nInvalidParameter );

  ROTRS( 2 == m_uiPicOrderCntType, Err::m_nOK );

  if( 0 == m_uiPicOrderCntType )
  {
    UInt uiTmp;
    RNOK( pcReadIf->getUvlc(uiTmp,"SPS: log2_max_pic_order_cnt_lsb_minus4"));
    setLog2MaxPicOrderCntLsb(4+uiTmp);
  }
  else if( 1 == m_uiPicOrderCntType )
  {
    RNOK( pcReadIf->getFlag( m_bDeltaPicOrderAlwaysZeroFlag,      "SPS: delta_pic_order_always_zero_flag"));
    RNOK( pcReadIf->getSvlc( m_iOffsetForNonRefPic,               "SPS: offset_for_non_ref_pic"));
    RNOK( pcReadIf->getSvlc( m_iOffsetForTopToBottomField,        "SPS: offset_for_top_to_bottom_field"));
    RNOK( pcReadIf->getUvlc( m_uiNumRefFramesInPicOrderCntCycle,  "SPS: num_ref_frames_in_pic_order_cnt_cycle"));
    RNOK( initOffsetForRefFrame( m_uiNumRefFramesInPicOrderCntCycle ) );
  }

  for( UInt i = 0; i < m_uiNumRefFramesInPicOrderCntCycle; i++)
  {
    int  iTmp;
    RNOK( pcReadIf->getSvlc(iTmp,"SPS: offset_for_ref_frame"));
    setOffsetForRefFrame(i,iTmp);
  }

  return Err::m_nOK;

}

bool SequenceParameterSet::operator != ( const SequenceParameterSet& rcSPS ) const
{
  ROTRS( m_eProfileIdc                      != rcSPS.m_eProfileIdc,                       true ) ;
  ROTRS( m_bConstrainedSet0Flag             != rcSPS.m_bConstrainedSet0Flag,              true ) ;
  ROTRS( m_bConstrainedSet1Flag             != rcSPS.m_bConstrainedSet1Flag,              true ) ;
  ROTRS( m_bConstrainedSet2Flag             != rcSPS.m_bConstrainedSet2Flag,              true ) ;
  ROTRS( m_uiLevelIdc                       != rcSPS.m_uiLevelIdc,                        true ) ;
  ROTRS( m_uiSeqParameterSetId              != rcSPS.m_uiSeqParameterSetId,               true ) ;
  ROTRS( m_uiLog2MaxFrameNum                != rcSPS.m_uiLog2MaxFrameNum,                 true ) ;
  ROTRS( m_uiPicOrderCntType                != rcSPS.m_uiPicOrderCntType,                 true ) ;
  ROTRS( m_uiLog2MaxPicOrderCntLsb          != rcSPS.m_uiLog2MaxPicOrderCntLsb,           true ) ;
  ROTRS( m_bDeltaPicOrderAlwaysZeroFlag     != rcSPS.m_bDeltaPicOrderAlwaysZeroFlag,      true ) ;
  ROTRS( m_iOffsetForNonRefPic              != rcSPS.m_iOffsetForNonRefPic,               true ) ;
  ROTRS( m_iOffsetForTopToBottomField       != rcSPS.m_iOffsetForTopToBottomField,        true ) ;
  ROTRS( m_uiNumRefFramesInPicOrderCntCycle != rcSPS.m_uiNumRefFramesInPicOrderCntCycle,  true ) ;
  ROTRS( m_piOffsetForRefFrame              != rcSPS.m_piOffsetForRefFrame,               true ) ;
  ROTRS( m_uiNumRefFrames                   != rcSPS.m_uiNumRefFrames,                    true ) ;
  ROTRS( m_bGapsInFrameNumValueAllowedFlag  != rcSPS.m_bGapsInFrameNumValueAllowedFlag,   true ) ;
  ROTRS( m_uiFrameWidthInMbs                != rcSPS.m_uiFrameWidthInMbs,                 true ) ;
  ROTRS( m_uiFrameHeightInMbs               != rcSPS.m_uiFrameHeightInMbs,                true ) ;
  ROTRS( m_uiChromaFormatParameterX2        != rcSPS.m_uiChromaFormatParameterX2,         true ) ;
  ROTRS( m_bFrameMbsOnlyFlag                != rcSPS.m_bFrameMbsOnlyFlag,                 true ) ;
  ROTRS( m_bMbAdaptiveFrameFieldFlag        != rcSPS.m_bMbAdaptiveFrameFieldFlag,         true ) ;
  ROTRS( m_bDirect8x8InferenceFlag          != rcSPS.m_bDirect8x8InferenceFlag,           true ) ;
  ROTRS( m_bFrameCroppingFlag               != rcSPS.m_bFrameCroppingFlag,                true ) ;
  ROTRS( m_cFrameCroppingRect               != rcSPS.m_cFrameCroppingRect,                true ) ;
  ROTRS( m_bVuiParametersPresentFlag        != rcSPS.m_bVuiParametersPresentFlag,         true ) ;

  return false;
}



// h264 namespace end
VIDEO_NAMESPACE_END
