/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/H264PictureParameterSet.h"
#include "VideoCommonLib/H264SliceHeader.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/ResidualCalculator.h"
#include "VideoCommonLib/PocAndFrameNum.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
// h264 namespace begin
VIDEO_NAMESPACE_BEGIN


const int g_aaiLumaToChromaQp[2][52] = {
{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,29,30,31,32,32,33,34,34,35,35,36,36,37,37,37,38,38,38,39,39,39,39},
{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,26,27,28,28,29,29,30,30,31,32,32,33,33,34,34,35,35,35,36,36,36,37,37,37,37},
};

ErrVal SliceHeader::setAccessUnitParameters( const AccessUnitParameters& rcAUPars)
{
  //th since stored arguments and syntak arguments differ 
  m_eNalRefIdc    = rcAUPars.getRefPicFlag() ? (rcAUPars.getIdr() ? NAL_REF_IDC_PRIORITY_HIGHEST : NAL_REF_IDC_PRIORITY_HIGH) : NAL_REF_IDC_PRIORITY_LOWEST;
  m_eNalUnitType  = rcAUPars.getIdr() ? NAL_UNIT_UNSPECIFIED_30 : NAL_UNIT_UNSPECIFIED_31;
  m_uiPicOrderCntLsb = (int)rcAUPars.getPicOrderCnt();

  const UInt uiPocMask = (1<<getSPS().getLog2MaxPicOrderCntLsb())-1;
  m_uiPicOrderCntLsb = (int)rcAUPars.getPicOrderCnt() & uiPocMask;

  if( rcAUPars.getIdr() )
  {
    m_uiFrameNum = 0;
    m_uiIdrPicId = (UInt)rcAUPars.getFrameNum(); // contain idr pic id in this special case
  }
  else
  {
    const UInt uiFrameNumMask = (1<<getSPS().getLog2MaxFrameNum())-1;
    m_uiFrameNum = (int)rcAUPars.getFrameNum() & uiFrameNumMask;
    m_uiIdrPicId = 0;
  }


  std::vector<int> cExtQpVector;
  for( UInt uiPlaPaId = 0; uiPlaPaId < rcAUPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    const std::vector<int> &rcQpVector = rcAUPars.getPlaPaPars( uiPlaPaId ).getPlaneQps();
    cExtQpVector.insert( cExtQpVector.end(), rcQpVector.begin(), rcQpVector.end());
  }
  
  m_iSliceQpDelta = cExtQpVector[0] - getPPS().getPicInitQp();
  std::vector<int> cH264QpVector;
  UInt uiIntraPic = ( rcAUPars.getMaxMvsPerBlk() == 0 ? 1 : 0 );
  cH264QpVector.push_back( getPPS().getPicInitQp() + m_iSliceQpDelta );
  cH264QpVector.push_back( g_aaiLumaToChromaQp[uiIntraPic][gMinMax( getPPS().getPicInitQp() + m_iSliceQpDelta + getPPS().getChromaQpIndexOffset( 0 ), 0, 51 )] );
  cH264QpVector.push_back( g_aaiLumaToChromaQp[uiIntraPic][gMinMax( getPPS().getPicInitQp() + m_iSliceQpDelta + getPPS().getChromaQpIndexOffset( 1 ), 0, 51 )] );

  // check if this is correct
  ROT( cExtQpVector.size() > cH264QpVector.size() )
  for( int i = 0; i < (int)cExtQpVector.size(); i++ )
  {
    AOT( cExtQpVector[i] != cH264QpVector[i]);
  }

  m_eSliceType = SliceType( rcAUPars.getMaxMvsPerBlk());
  setNumRefIdxActive( LIST_0, rcAUPars.getRefListSize() );
  setNumRefIdxActiveOverrideFlag( m_rcPPS.getNumRefIdxActive( LIST_0 ) != getNumRefIdxActive( LIST_0 ));

  m_aiDeltaRefPicOrderCnt = rcAUPars.getRefDeltaPocList();

  m_cDeblockingFilterParameter.setDisableDeblockingFilterIdc( rcAUPars.getDeblockingIdc() );
  m_cDeblockingFilterParameter.setSliceAlphaC0Offset( rcAUPars.getDeblockAlphaOffs() );
  m_cDeblockingFilterParameter.setSliceBetaOffset( rcAUPars.getDeblockBetaOffs() );
  m_eExtendedCabacTables    = rcAUPars.getExtendedCabacTables();
  m_uiPlaneGrouping         = rcAUPars.getPlaneGrouping();
  m_uiInterpolationMethod   = rcAUPars.getInterpolationMethod();
  m_uiEntropySliceSizeX     = rcAUPars.getEntropySliceSizeX();
  m_uiEntropySliceSizeY     = rcAUPars.getEntropySliceSizeY();
  m_bCtxStatesReuse         = rcAUPars.getCtxStatesReuseFlag();
  m_bEntropySliceFlag       = rcAUPars.getEntropySliceFlag();
  m_bInLoopFilterFlag       = rcAUPars.getWienerFilterParameters().getUseFilterFlag();
  m_uiVLCMaxDelayParameter  = rcAUPars.getVLCMaxDelayParameter();
  for( UInt uiPlaPaId = 0; uiPlaPaId < rcAUPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    // just copy once more 
    m_acPlanePackageParameters.push_back(rcAUPars.getPlaPaPars( uiPlaPaId ));
  }
  return Err::m_nOK;
}

ErrVal SliceHeader::getAccessUnitParameters( AccessUnitParameters& rcAUPars) const
{
  rcAUPars.setRefPicFlag( NAL_REF_IDC_PRIORITY_LOWEST != m_eNalRefIdc );
  
  PocAndFrameNum* pcPocAndFrameNum = rcAUPars.getPocAndFrameNum();
  // do this only once per frame
  rcAUPars.setPicOrderCnt( pcPocAndFrameNum->calculatePoc( m_uiPicOrderCntLsb, getSPS().getLog2MaxPicOrderCntLsb(), true, false ) );
  rcAUPars.setFrameNum( pcPocAndFrameNum->calculateFrameNum( m_uiFrameNum, getSPS().getLog2MaxFrameNum() ));

  std::vector<int> cH264QpVector;
  UInt uiIntraPic = ( m_eSliceType == I_SLICE ? 1 : 0 );
  cH264QpVector.push_back( getPPS().getPicInitQp() + m_iSliceQpDelta );
  cH264QpVector.push_back( g_aaiLumaToChromaQp[uiIntraPic][gMinMax( getPPS().getPicInitQp() + m_iSliceQpDelta + getPPS().getChromaQpIndexOffset( 0 ), 0, 51 )] );
  cH264QpVector.push_back( g_aaiLumaToChromaQp[uiIntraPic][gMinMax( getPPS().getPicInitQp() + m_iSliceQpDelta + getPPS().getChromaQpIndexOffset( 1 ), 0, 51 )] );

  int i = 0;
  for( UInt uiPlaPaId = 0; uiPlaPaId < rcAUPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    rcAUPars.getPlaPaPars( uiPlaPaId ).getPlaneQps().clear();
    for( UInt ui = 0; ui < rcAUPars.getPlaPaPars( uiPlaPaId ).getNumPlanes(); ui++ )
    {
      rcAUPars.getPlaPaPars( uiPlaPaId ).getPlaneQps().push_back( cH264QpVector[i++] );
    }
  }

  rcAUPars.getRefDeltaPocList()         = m_aiDeltaRefPicOrderCnt;
  rcAUPars.setMaxMvsPerBlk              ( m_eSliceType );
  rcAUPars.setDeblockingIdc             ( m_cDeblockingFilterParameter.getDisableDeblockingFilterIdc( ) );
  rcAUPars.setDeblockAlphaOffs          ( m_cDeblockingFilterParameter.getSliceAlphaC0Offset( ) );
  rcAUPars.setDeblockBetaOffs           ( m_cDeblockingFilterParameter.getSliceBetaOffset( ) );
  rcAUPars.setExtendedCabacTables       ( m_eExtendedCabacTables );
  rcAUPars.setPlaneGrouping             ( m_uiPlaneGrouping );
  rcAUPars.setInterpolationMethod       ( m_uiInterpolationMethod  );
  rcAUPars.setEntropySliceSizeX         ( m_uiEntropySliceSizeX );
  rcAUPars.setEntropySliceSizeY         ( m_uiEntropySliceSizeY );
  rcAUPars.setCtxStatesReuse            ( m_bCtxStatesReuse );
  rcAUPars.setIdr                       ( isIdrNalUnit() );
  rcAUPars.setVLCFlag                   ( ! getPPS().getEntropyCodingModeFlag() );
  rcAUPars.setVLCMaxDelayParameter      ( m_uiVLCMaxDelayParameter );

  return Err::m_nOK;
}


SliceHeader::SliceHeader( const SequenceParameterSet& rcSPS, const PictureParameterSet& rcPPS )
  : m_rcPPS                             ( rcPPS )
  , m_rcSPS                             ( rcSPS )
  , m_eExtendedCabacTables              ( DEFAULT_CABAC_TABLES )
  , m_uiPlaneGrouping                   ( 0 )
  , m_uiInterpolationMethod             ( 0 )
  , m_uiEntropySliceSizeX               ( 0 )
  , m_uiEntropySliceSizeY               ( 0 )
  , m_bCtxStatesReuse                   ( false )
  , m_bEntropySliceFlag                 ( false )
  , m_bInLoopFilterFlag                 ( false )
  , m_bDummyFlag1                       ( false )
  , m_bDummyFlag2                       ( false )
  , m_bDummyFlag3                       ( false )
{
  setPicParameterSetId( rcPPS.getPicParameterSetId() );
  setNumRefIdxActive  ( LIST_0, m_rcPPS.getNumRefIdxActive( LIST_0 ) );
  setNumRefIdxActive  ( LIST_1, m_rcPPS.getNumRefIdxActive( LIST_1 ) );

  setNalUnitType      ( NAL_UNIT_UNSPECIFIED_31 );
}

SliceHeader::~SliceHeader()
{
}



bool SliceHeader::check() const
{
  ROFRS( getSPS().check(),  false );
  ROFRS( getPPS().check( getSPS().getProfileIdc(), getSPS().getLevelIdc() ),  false );
  return true;
}




ErrVal SliceHeader::writeSH( HeaderSymbolWriteIf* pcWriteIf ) const
{
  const bool bL0InfoPresent = (m_eSliceType != I_SLICE);
  RNOK( pcWriteIf->writeUvlc( getFirstMbInSlice(),                         "SH: first_mb_in_slice" ) );
  RNOK( pcWriteIf->writeUvlc( m_eSliceType,                                "SH: slice_type" ) );
  RNOK( pcWriteIf->writeUvlc( getPicParameterSetId(),                      "SH: pic_parameter_set_id" ) );

  if( YUV400 == getSPS().getChromaFormatIdc() )
  {
  }
  else if( YUV420 == getSPS().getChromaFormatIdc() || YUV422 == getSPS().getChromaFormatIdc() || YUV444 == getSPS().getChromaFormatIdc())
  {
    RNOK( pcWriteIf->writeCode( m_uiPlaneGrouping, 2, "SH: plane_grouping" ) );
  }
  else // here we can add alpha and stereo stuff
  {
    ROT(1)
  }


  RNOK( pcWriteIf->writeCode( getFrameNum(), getSPS().getLog2MaxFrameNum(),"SH: frame_num" ) );

  if (isIdrNalUnit())
  {
    RNOK( pcWriteIf->writeUvlc( getIdrPicId(),                             "SH: idr_pic_id" ) );
  }

  if( 0 == getSPS().getPicOrderCntType() )
  {
    RNOK( pcWriteIf->writeCode( getPicOrderCntLsb(), getSPS().getLog2MaxPicOrderCntLsb(), "SH: pic_order_cnt_lsb" ) );
  }
  else
  {
    ROT(1)
  }

  if( bL0InfoPresent )
  {
/*
    RNOK( pcWriteIf->writeFlag( getNumRefIdxActiveOverrideFlag(), "SH: num_ref_idx_active_override_flag" ) );

    if( getNumRefIdxActiveOverrideFlag() )
    {
      const int iNumActive = getNumRefIdxActive(LIST_0)-1;
      AOT_DBG( 0 == getNumRefIdxActive(LIST_0) );
      RNOK( pcWriteIf->writeUvlc( iNumActive,   "SH: num_ref_idx_l0_active_minus1" ) );
    }
*/
    for( UInt i = 0; i < getNumRefIdxActive( LIST_0); i++)
    {
      RNOK( pcWriteIf->writeSvlc( m_aiDeltaRefPicOrderCnt[i],       "SH: delta_ref_pic_order_cnt" ) );
    }
    RNOK( pcWriteIf->writeSvlc( 0,                                  "SH: delta_ref_pic_order_cnt" ) );

//    RNOK( getRplrBuffer( LIST_0 ).write( pcWriteIf ) );
  }

/*
  if( (getPPS().getWeightedPredFlag() && ((m_eSliceType == P_SLICE)||(m_eSliceType == SP_SLICE))) ||
      ((getPPS().getWeightedBipredIdc()==1) && (m_eSliceType == B_SLICE)))
  {
    const ChromaFormatIdc eChromaFormatIdc = getSPS().getChromaFormatIdc();
    RNOK( pcWriteIf->writeUvlc( getLumaLogWeightDenom(),          "PWT: luma_log_weight_denom" ) );
    if ( YUV400 != eChromaFormatIdc)
    {
      RNOK( pcWriteIf->writeUvlc( getChromaLogWeightDenom(),        "PWT: chroma_log_weight_denom" ) );
    }

    if( bL0InfoPresent )
    {
      RNOK( m_acPredWeightTable[LIST_0].write( pcWriteIf, getNumRefIdxActive(LIST_0), eChromaFormatIdc ) );
      if( bL1InfoPresent )
      {
        RNOK( m_acPredWeightTable[LIST_1].write( pcWriteIf, getNumRefIdxActive(LIST_1), eChromaFormatIdc ) );
      }
    }
  }
*/
  if( getNalRefIdc() )
  {
    if( isIdrNalUnit() )
    {
      RNOK( pcWriteIf->writeFlag( getNoOutputOfPriorPicsFlag(),   "DRPM: no_output_of_prior_pics_flag" ) );
      RNOK( pcWriteIf->writeFlag( getLongTermReferenceFlag(),     "DRPM: long_term_reference_flag" ) );
    }
    else
    {
      RNOK( pcWriteIf->writeFlag( getAdaptiveRefPicBufferingModeFlag(),"DRPM: adaptive_ref_pic_buffering_mode_flag" ) );

      if( getAdaptiveRefPicBufferingModeFlag() )
      {
        ROT(1);
      }
    }
  }

  if( getPPS().getEntropyCodingModeFlag()  && bL0InfoPresent)// (eSliceType != I_SLICE) )
  {
    RNOK( pcWriteIf->writeUvlc( getCabacInitIdc(),                "SH: cabac_init_idc" ) );
  }

  RNOK( pcWriteIf->writeSvlc( getSliceQpDelta(),                  "SH: slice_qp_delta" ) );

  if( getPPS().getDeblockingFilterParametersPresentFlag() )
  {
    RNOK( getDeblockingFilterParameter().write( pcWriteIf ) );
  }

  pcWriteIf->writeUvlc( m_uiInterpolationMethod,        "SH: interpolation_method" );
  pcWriteIf->writeFlag( false,                          "SH: reserved_flag_0" );

  for( UInt uiPlaPaId = 0; uiPlaPaId < m_acPlanePackageParameters.size(); uiPlaPaId++ )
  {
    RNOK( xWritePlanePackageParameters( pcWriteIf, m_acPlanePackageParameters[uiPlaPaId], bL0InfoPresent ) );
  }

  pcWriteIf->writeFlag       ( m_bEntropySliceFlag, "SH: use_entropy_slices" );
  if( m_bEntropySliceFlag )
  {
    pcWriteIf->writeFlag     ( m_bCtxStatesReuse,               "SH: context_states_reuse" );
    pcWriteIf->writeUvlc     ( m_uiEntropySliceSizeX,           "SH: entropy_slice_size_x" );
    pcWriteIf->writeUvlc     ( m_uiEntropySliceSizeY,           "SH: entropy_slice_size_y" );
  }

  pcWriteIf->writeFlag       ( m_bInLoopFilterFlag,             "SH: quadtree_based_inloop_filter_enabled" );

  if( getPPS().getEntropyCodingModeFlag() )
  {
    TOF( UInt( m_eExtendedCabacTables ) < 2 );
    pcWriteIf->writeFlag( UInt( m_eExtendedCabacTables ) != 0, "SH: multi_codeword_arithmetic_coding_flag" );
  }
  else
  {
    pcWriteIf->writeUvlc( m_uiVLCMaxDelayParameter, "SH: vlc_max_delay_parameter" );
  }

  pcWriteIf->writeFlag       ( m_bDummyFlag1,             "SH: reserved_flag_1" );
  pcWriteIf->writeFlag       ( m_bDummyFlag2,             "SH: reserved_flag_2" );
  pcWriteIf->writeFlag       ( m_bDummyFlag3,             "SH: reserved_flag_3" );

  return Err::m_nOK;
}


ErrVal SliceHeader::xWritePlanePackageParameters( HeaderSymbolWriteIf* pcWriteIf, const PlanePackageParameters& rcPlaPaPars, bool bInterPred) const
{
  //  for( UInt uiPlaPaId = 1; uiPlaPaId < rcPars.getNumPlanePacks(); uiPlaPaId++ )
  {
    if( 0 < rcPlaPaPars.getPlaPaId() )
    {
      pcWriteIf->writeFlag     ( rcPlaPaPars.getInterPlaPaPredFlag(),          "SH: inter_plane_package_pred_flag" );
      if( rcPlaPaPars.getInterPlaPaPredFlag() )
      {
        UInt uiDeltaPlaPaId = rcPlaPaPars.getPlaPaId() - rcPlaPaPars.getRefPlaPlaId();
        pcWriteIf->writeUvlc   ( uiDeltaPlaPaId - 1,                           "SH: delta_ref_plane_package_id_minus1" );
        pcWriteIf->writeFlag   ( rcPlaPaPars.getAdaptiveInterPlaPaPredFlag(),  "SH: adaptive_plane_package_pred_flag" );
      }
    }

    pcWriteIf->writeUvlc     ( rcPlaPaPars.getLog2MinPredBlkSize () - 2,     "SH: log2_min_pred_block_size_minus2" );
    pcWriteIf->writeUvlc     ( rcPlaPaPars.getNumPredBlkSizes    () - 1,     "SH: log2_diff_max_min_pred_block_size" );
    if( bInterPred )
    {
      pcWriteIf->writeUvlc   ( rcPlaPaPars.getMvAccuracy(),                  "SH: log2_mv_accuracy" );
      pcWriteIf->writeFlag   ( rcPlaPaPars.getMvCompetition(),               "SH: mv_competition_flag" );
    }
    {
      pcWriteIf->writeFlag( (rcPlaPaPars.getTransformMode() & ResInfo::RES_TRAFO_SKIP) != 0, "SH: residual_skip_allowed_flag" );
      pcWriteIf->writeFlag( 0, "SH: residual_mode_reserved_flag_0" );
      pcWriteIf->writeFlag( (rcPlaPaPars.getTransformMode() & ResInfo::RES_TRAFO_POW2) != 0, "SH: residual_pow2_transform_allowed_flag" );
      pcWriteIf->writeFlag( 0, "SH: residual_mode_reserved_flag_1" );
    }
    if( rcPlaPaPars.getTransformMode() & ResInfo::RES_TRAFO_POW2 )
    {
      pcWriteIf->writeUvlc   ( rcPlaPaPars.getLog2MinPow2TrafoBlkSize() - 2, "SH: log2_min_transform_block_size_minus2" );
      pcWriteIf->writeUvlc   ( rcPlaPaPars.getNumPow2TrafoBlkSizes   () - 1, "SH: log2_diff_max_min_transform_block_size" );
      pcWriteIf->writeFlag   ( true,                                         "SH: pow2_transform_reserved_flag" );

      if( getPPS().getEntropyCodingModeFlag() )
      {
        pcWriteIf->writeFlag   ( rcPlaPaPars.getCommonPow2SigCtxFlag(),        "SH: common_pow2_sig_context_flag" );
        int iLog2NumPow2SigCtxDiv2 = 2;
        for( UInt ui = 0; ui < rcPlaPaPars.getNumPow2SigCtxClasses(); ui++ )
        {
          UInt uiLog2NumPow2SigCtxDiv2  = gCeilLog2( rcPlaPaPars.getNumPow2SigCtx( ui ) ) >> 1;
          ROF( rcPlaPaPars.getNumPow2SigCtx( ui ) == UInt( 1 << ( uiLog2NumPow2SigCtxDiv2 << 1 ) ) );
          pcWriteIf->writeSvlc ( iLog2NumPow2SigCtxDiv2-uiLog2NumPow2SigCtxDiv2,                      "SH: log2_num_pow2_sig_contexts_div2_diff" );
          iLog2NumPow2SigCtxDiv2 = uiLog2NumPow2SigCtxDiv2;
        }
        pcWriteIf->writeFlag   ( rcPlaPaPars.getCommonPow2LastCtxFlag(),       "SH: common_pow2_last_context_flag" );
        int iLog2NumPow2LastCtxDiv2 = 2;
        for( UInt ui = 0; ui < rcPlaPaPars.getNumPow2LastCtxClasses(); ui++ )
        {
          UInt uiLog2NumPow2LastCtxDiv2 = gCeilLog2( rcPlaPaPars.getNumPow2LastCtx( ui ) ) >> 1;
          ROF( rcPlaPaPars.getNumPow2LastCtx( ui ) == UInt( 1 << ( uiLog2NumPow2LastCtxDiv2 << 1 ) ) );
          pcWriteIf->writeSvlc ( iLog2NumPow2LastCtxDiv2-uiLog2NumPow2LastCtxDiv2,                     "SH: log2_num_pow2_last_contexts_div2_diff" );
          iLog2NumPow2LastCtxDiv2 = uiLog2NumPow2LastCtxDiv2;
        }
      }
    }
  }

  return Err::m_nOK;
}


ErrVal SliceHeader::readSH( HeaderSymbolReadIf* pcReadIf, AccessUnitParameters& rcAUPars )
{
  UInt uiSliceType;

  RNOK( pcReadIf->getUvlc( m_uiFirstMbInSlice,                         "SH: first_mb_in_slice" ) );
  ROF( 0 == m_uiFirstMbInSlice );
  RNOK( pcReadIf->getUvlc( uiSliceType,                                "SH: slice_type" ) );
  RNOK( pcReadIf->getUvlc( m_uiPicParameterSetId,                      "SH: pic_parameter_set_id" ) );
  if( YUV400 == getSPS().getChromaFormatIdc() )
  {
    m_uiPlaneGrouping = 0;
  }
  else if( YUV420 == getSPS().getChromaFormatIdc() || YUV422 == getSPS().getChromaFormatIdc() || YUV444 == getSPS().getChromaFormatIdc())
  {
    RNOK( pcReadIf->getCode( m_uiPlaneGrouping, 2, "SH: plane_grouping" ) );
  }
  else // here we can add alpha and stereo stuff
  {
    AOT(1)
  }
  RNOK( pcReadIf->getCode( m_uiFrameNum, getSPS().getLog2MaxFrameNum(), "SH: frame_num" ) );
  m_eSliceType = SliceType( uiSliceType);
  const bool bL0InfoPresent = I_SLICE != m_eSliceType;

   int iTmp;


  if (isIdrNalUnit())
  {
    AOT(0 != m_uiFrameNum);
    RNOK( pcReadIf->getUvlc( m_uiIdrPicId,                             "SH: idr_pic_id" ) );
    ROTR( m_uiIdrPicId > 65535, Err::m_nInvalidParameter );
  }

  if( 0 == getSPS().getPicOrderCntType() )
  {
    RNOK( pcReadIf->getCode( m_uiPicOrderCntLsb, getSPS().getLog2MaxPicOrderCntLsb() ,"SH: pic_order_cnt_lsb" ) );
  }
  else
  {
    AOT(1);
  }

  if (getPPS().getRedundantPicCntPresentFlag())
  {
    AOT(1);
  }

  if( bL0InfoPresent )
  {
    while( 1 )
    {
      int iTemp;
      RNOK( pcReadIf->getSvlc( iTemp,                                     "SH: delta_ref_pic_order_cnt" ) );
      m_aiDeltaRefPicOrderCnt.push_back( iTemp );

      if( 0 == iTemp )
      {
        m_aiDeltaRefPicOrderCnt.pop_back();
        setNumRefIdxActive( LIST_0, UInt( m_aiDeltaRefPicOrderCnt.size() ) );
        break;
      }
    }
  }

  if( getNalRefIdc() )
  {
    if( isIdrNalUnit())
    {
      RNOK( pcReadIf->getFlag( m_bNoOutputOfPriorPicsFlag,              "DRPM: no_output_of_prior_pics_flag" ) );
      RNOK( pcReadIf->getFlag( m_bLongTermReferenceFlag,                "DRPM: long_term_reference_flag" ) );
    }
    else
    {
      RNOK( pcReadIf->getFlag( m_bAdaptiveRefPicBufferingModeFlag,      "DRPM: adaptive_ref_pic_buffering_mode_flag" ) );
      if( m_bAdaptiveRefPicBufferingModeFlag )
      {
        ROT(1);
      }
    }
  }

  if( getPPS().getEntropyCodingModeFlag()  && bL0InfoPresent )
  {
    RNOK( pcReadIf->getUvlc( m_uiCabacInitIdc,                          "SH: cabac_init_idc" ) );
    ROF( 0 == m_uiCabacInitIdc );
    ROTR( m_uiCabacInitIdc > 2, Err::m_nInvalidParameter );
  }

  RNOK( pcReadIf->getSvlc( iTmp,                                        "SH: slice_qp_delta" ) );
  ROTR((getPPS().getPicInitQp()+iTmp) > 51, Err::m_nInvalidParameter);

  ROTR(((int)getPPS().getPicInitQp()+iTmp) < 0, Err::m_nInvalidParameter);
  setSliceQpDelta(iTmp);
  
  if( getPPS().getDeblockingFilterParametersPresentFlag() )
  {
    RNOK( getDeblockingFilterParameter().read( pcReadIf ) );
  }

  pcReadIf->getUvlc     ( m_uiInterpolationMethod,       "SH: interpolation_method" );
  {
    bool bReservedFlag0;
    pcReadIf->getFlag( bReservedFlag0,                   "SH: reserved_flag_0" );
  }

  PocAndFrameNum* pcPocAndFrameNum = rcAUPars.getPocAndFrameNum();
  RNOK( getSPS().getAccessUnitParameters( rcAUPars, m_uiPlaneGrouping ) );

  rcAUPars.setPocAndFrameNum( pcPocAndFrameNum );
  rcAUPars.setSPS(&getSPS());
  rcAUPars.setPPS(&getPPS());


  RNOK( getAccessUnitParameters( rcAUPars ) );

  UInt uiNumPlanePackages = rcAUPars.getNumPlanePacks();
  for( UInt uiPlaPaId = 0; uiPlaPaId < uiNumPlanePackages; uiPlaPaId++ )
  {
    RNOK( xReadPlanePackageParameters( pcReadIf, rcAUPars.getPlaPaPars(uiPlaPaId), bL0InfoPresent ) );
  }

  pcReadIf->getFlag       ( m_bEntropySliceFlag, "SH: use_entropy_slices" );
  if( m_bEntropySliceFlag )
  {
    pcReadIf->getFlag     ( m_bCtxStatesReuse,               "SH: context_states_reuse" );
    pcReadIf->getUvlc     ( m_uiEntropySliceSizeX,           "SH: entropy_slice_size_x" );
    pcReadIf->getUvlc     ( m_uiEntropySliceSizeY,           "SH: entropy_slice_size_y" );
  }

  pcReadIf->getFlag       ( m_bInLoopFilterFlag,             "SH: quadtree_based_inloop_filter_enabled" );

  {
    bool bFlag = false;
    if( getPPS().getEntropyCodingModeFlag() )
    {
      pcReadIf->getFlag( bFlag,                              "SH: multi_codeword_arithmetic_coding_flag" );
    }
    else
    {
      pcReadIf->getUvlc( m_uiVLCMaxDelayParameter,           "SH: vlc_max_delay_parameter" );
      rcAUPars.setVLCMaxDelayParameter( m_uiVLCMaxDelayParameter );
    }
    m_eExtendedCabacTables = bFlag ? AGGREGATE_CABAC_TABLES12 : DEFAULT_CABAC_TABLES;
    rcAUPars.setExtendedCabacTables( m_eExtendedCabacTables );
  }

  pcReadIf->getFlag       ( m_bDummyFlag1,             "SH: reserved_flag_1" );
  pcReadIf->getFlag       ( m_bDummyFlag2,             "SH: reserved_flag_2" );
  pcReadIf->getFlag       ( m_bDummyFlag3,             "SH: reserved_flag_3" );

  rcAUPars.setEntropySliceSizeX( m_uiEntropySliceSizeX );
  rcAUPars.setEntropySliceSizeY( m_uiEntropySliceSizeY );
  rcAUPars.setCtxStatesReuse( m_bCtxStatesReuse );
  rcAUPars.getWienerFilterParameters().setUseFilterFlag( m_bInLoopFilterFlag );

  return Err::m_nOK;
}

ErrVal SliceHeader::xReadPlanePackageParameters( HeaderSymbolReadIf* pcReadIf, PlanePackageParameters& rcPPP, const bool bInterPred ) const
{
  if( 0 < rcPPP.getPlaPaId() )
  {
    bool bPlaPaPred;
    pcReadIf->getFlag     ( bPlaPaPred,   "SH: inter_plane_package_pred_flag" );
    if( bPlaPaPred )
    {
      UInt  uiDeltaId;
      bool  bAdaptFlag;
      pcReadIf->getUvlc   ( uiDeltaId,    "SH delta_ref_plane_package_id_minus1" );
      pcReadIf->getFlag   ( bAdaptFlag,   "SH: adaptive_plane_package_pred_flag" );
      ROF( rcPPP.getPlaPaId() > (int)uiDeltaId );
      UInt  uiRefPlaPaId = rcPPP.getPlaPaId() - ( uiDeltaId + 1 );
      rcPPP.setRefPlaPaId                 ( uiRefPlaPaId );
      rcPPP.setAdaptiveInterPlaPaPredFlag ( bAdaptFlag );
    }
  }
  UInt uiLog2MinPrd, uiLog2MaxPrd;
  pcReadIf->getUvlc     ( uiLog2MinPrd, "SH: log2_min_pred_block_size_minus2" );
  uiLog2MinPrd += 2;
  pcReadIf->getUvlc     ( uiLog2MaxPrd, "SH: log2_diff_max_min_pred_block_size" );
  uiLog2MaxPrd += uiLog2MinPrd;
  rcPPP.setLog2MinPredBlkSize( uiLog2MinPrd );
  rcPPP.setLog2MaxPredBlkSize( uiLog2MaxPrd );
  if( bInterPred )
  {
    UInt uiLog2MvAcc;
    pcReadIf->getUvlc   ( uiLog2MvAcc,  "SH: log2_mv_accuracy" );
    rcPPP.setMvAccuracy( uiLog2MvAcc );
    bool bMvCompetition;
    pcReadIf->getFlag   ( bMvCompetition, "SH: mv_competition_flag" );
    rcPPP.setMvCompetition( bMvCompetition );
  }

  UInt uiTrafoMode = 0;
  {
    bool bFlag;
    pcReadIf->getFlag( bFlag, "SH: residual_skip_allowed_flag" );
    uiTrafoMode |= bFlag ? ResInfo::RES_TRAFO_SKIP : 0;
    pcReadIf->getFlag( bFlag, "SH: residual_mode_reserved_flag_0" );
    pcReadIf->getFlag( bFlag, "SH: residual_pow2_transform_allowed_flag" );
    uiTrafoMode |= bFlag ? ResInfo::RES_TRAFO_POW2 : 0;
    pcReadIf->getFlag( bFlag, "SH: residual_mode_reserved_flag_1" );
  }
  rcPPP.setTransformMode( uiTrafoMode );

  if( uiTrafoMode & ResInfo::RES_TRAFO_POW2 )
  {
    UInt uiLog2MinTSz, uiLog2MaxTSz;
    pcReadIf->getUvlc   ( uiLog2MinTSz, "SH: log2_min_transform_block_size_minus2" );
    uiLog2MinTSz += 2;
    pcReadIf->getUvlc   ( uiLog2MaxTSz, "SH: log2_diff_max_min_transform_block_size" );
    uiLog2MaxTSz += uiLog2MinTSz;
    {
      bool bReservedPow2Flag;
      pcReadIf->getFlag( bReservedPow2Flag, "SH: pow2_transform_reserved_flag" );
    }
    rcPPP.setLog2MinPow2TrafoBlkSize( uiLog2MinTSz );
    rcPPP.setLog2MaxPow2TrafoBlkSize( uiLog2MaxTSz );
    if( getPPS().getEntropyCodingModeFlag() )
    {
      bool bComSig;
      pcReadIf->getFlag   ( bComSig,      "SH: common_pow2_sig_context_flag" );
      UInt uiNumSigClasses  = ( bComSig ? 1 : uiLog2MaxTSz - uiLog2MinTSz + 1 );
      UInt uiLog2NumPow2SigCtxDiv2 = 2;
      for( UInt ui = 0; ui < uiNumSigClasses; ui++ )
      {
        int iDiffCtx;
        pcReadIf->getSvlc ( iDiffCtx,     "SH: log2_num_pow2_sig_contexts_div2_diff" );
        uiLog2NumPow2SigCtxDiv2 += iDiffCtx;
        UInt uiNumCtx = ( 1 << ( uiLog2NumPow2SigCtxDiv2 << 1 ) );
        rcPPP.getNumPow2SigCtxs().push_back( uiNumCtx );
      }
      bool bComLast;
      pcReadIf->getFlag   ( bComLast,     "SH: common_pow2_last_context_flag" );
      UInt uiNumLastClasses = ( bComLast ? 1 : uiLog2MaxTSz - uiLog2MinTSz + 1 );
      UInt uiLog2NumPow2LastCtxDiv2 = 2;
      for( UInt ui = 0; ui < uiNumLastClasses; ui++ )
      {
        int iDiffCtx;
        pcReadIf->getSvlc ( iDiffCtx,     "SH: log2_num_pow2_last_contexts_div2_diff" ); 
        uiLog2NumPow2LastCtxDiv2 += iDiffCtx;
        UInt uiNumCtx = ( 1 << ( uiLog2NumPow2LastCtxDiv2 << 1 ) );
        rcPPP.getNumPow2LastCtxs().push_back( uiNumCtx );
      }
    }
  }

  return Err::m_nOK;
}

// h264 namespace end
VIDEO_NAMESPACE_END
