/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCommonLib/H264SliceHeaderBase.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

SliceHeaderPicture::SliceHeaderPicture()
  : m_uiPicParameterSetId               ( 0 )
  , m_uiFrameNum                        ( 0 )
  , m_uiIdrPicId                        ( 0 )
  , m_uiPicOrderCntLsb                  ( 0 )
  , m_iDeltaPicOrderCntBottom           ( 0 )
  , m_uiSliceGroupChangeCycle           ( 0 )
  , m_eNalRefIdc                        ( NAL_REF_IDC_PRIORITY_LOW )
  , m_eNalUnitType                      ( NAL_UNIT_INVALID )
  , m_bFieldPicFlag                     ( false )
  , m_bBottomFieldFlag                  ( false )
  , m_bSpForSwitchFlag                  ( false )
  , m_bLongTermReferenceFlag            ( false )
  , m_bAdaptiveRefPicBufferingModeFlag  ( false )
  , m_bOneSliceTypeIndication           ( false )
{
  m_aiDeltaPicOrderCnt[LIST_0] = 0;
  m_aiDeltaPicOrderCnt[LIST_1] = 0;
}


SliceHeaderPicture::~SliceHeaderPicture()
{
}

SliceHeaderPicture& SliceHeaderPicture::operator = ( const SliceHeaderPicture& rcSHP )
{
  m_bOneSliceTypeIndication           = rcSHP.m_bOneSliceTypeIndication;

  m_uiFrameNum                        = rcSHP.m_uiFrameNum;
  m_bFieldPicFlag                     = rcSHP.m_bFieldPicFlag;
  m_bBottomFieldFlag                  = rcSHP.m_bBottomFieldFlag; //false;
  m_uiIdrPicId                        = rcSHP.m_uiIdrPicId;
  m_uiPicOrderCntLsb                  = rcSHP.m_uiPicOrderCntLsb;
  m_iDeltaPicOrderCntBottom           = rcSHP.m_iDeltaPicOrderCntBottom;

  m_bSpForSwitchFlag                  = rcSHP.m_bSpForSwitchFlag;
  m_uiSliceGroupChangeCycle           = rcSHP.m_uiSliceGroupChangeCycle;

  m_eNalRefIdc                        = rcSHP.m_eNalRefIdc;
  m_eNalUnitType                      = rcSHP.m_eNalUnitType;
  m_aiDeltaPicOrderCnt[0]             = rcSHP.m_aiDeltaPicOrderCnt[0];
  m_aiDeltaPicOrderCnt[1]             = rcSHP.m_aiDeltaPicOrderCnt[1];
  m_bLongTermReferenceFlag            = rcSHP.m_bLongTermReferenceFlag;
  m_bAdaptiveRefPicBufferingModeFlag  = rcSHP.m_bAdaptiveRefPicBufferingModeFlag;
  m_eNalRefIdc                        = rcSHP.m_eNalRefIdc;
  m_eNalUnitType                      = rcSHP.m_eNalUnitType;


  return *this;
}


SliceHeaderBase::SliceHeaderBase():
  m_uiFirstMbInSlice                  ( 0 ),
  m_eSliceType                        ( P_SLICE ),
  m_uiRedundantPicCount               ( 0 ),
  m_bDirectSpatialMvPredFlag          ( false ),
  m_bNumRefIdxActiveOverrideFlag      ( false ),
  m_uiCabacInitIdc                    ( 0 ),
  m_iSliceQpDelta                     ( 0 ),
  m_iSliceQsDelta                     ( 0 ),
  m_uiLumaLogWeightDenom              ( 0 ),
  m_uiChromaLogWeightDenom            ( 0 ),
  m_bNoOutputOfPriorPicsFlag          ( false ),
  m_uiSliceId                         ( 0 )
{
}


SliceHeaderBase& SliceHeaderBase::operator = ( const SliceHeaderBase& rcSHB )
{
  m_eSliceType                        = rcSHB.m_eSliceType;
  m_uiRedundantPicCount               = rcSHB.m_uiRedundantPicCount;
  m_bDirectSpatialMvPredFlag          = rcSHB.m_bDirectSpatialMvPredFlag;
  m_bNumRefIdxActiveOverrideFlag      = false;
  m_uiCabacInitIdc                    = rcSHB.m_uiCabacInitIdc;
  m_iSliceQpDelta                     = rcSHB.m_iSliceQpDelta;
  m_iSliceQsDelta                     = rcSHB.m_iSliceQsDelta;
  m_uiLumaLogWeightDenom              = rcSHB.m_uiLumaLogWeightDenom;
  m_uiChromaLogWeightDenom            = rcSHB.m_uiChromaLogWeightDenom;
  m_bNoOutputOfPriorPicsFlag          = rcSHB.m_bNoOutputOfPriorPicsFlag;
  m_auiNumRefIdxActive[LIST_0]        = rcSHB.m_auiNumRefIdxActive[LIST_0];
  m_auiNumRefIdxActive[LIST_1]        = rcSHB.m_auiNumRefIdxActive[LIST_1];
  m_cDeblockingFilterParameter        = rcSHB.m_cDeblockingFilterParameter;

  return *this;
}





ErrVal SliceHeaderBase::DeblockingFilterParameter::write( HeaderSymbolWriteIf* pcWriteIf ) const
{
  RNOK( pcWriteIf->writeUvlc( getDisableDeblockingFilterIdc(),  "SH: disable_deblocking_filter_idc" ) );

  ROTRS( 1 == getDisableDeblockingFilterIdc(), Err::m_nOK );

  RNOK( pcWriteIf->writeSvlc( getSliceAlphaC0Offset() >> 1,     "SH: slice_alpha_c0_offset_div2" ) );
  RNOK( pcWriteIf->writeSvlc( getSliceBetaOffset() >> 1,        "SH: slice_beta_offset_div2" ) );
  return Err::m_nOK;
}

ErrVal SliceHeaderBase::DeblockingFilterParameter::read( HeaderSymbolReadIf* pcReadIf )
{
  RNOK( pcReadIf->getUvlc( m_uiDisableDeblockingFilterIdc,      "SH: disable_deblocking_filter_idc" ) );

  ROTR( m_uiDisableDeblockingFilterIdc > 2, Err::m_nInvalidParameter );
  ROTRS( 1 == getDisableDeblockingFilterIdc(), Err::m_nOK );

  int iTmp;
  RNOK( pcReadIf->getSvlc( iTmp,                                "SH: slice_alpha_c0_offset_div2" ) );
  ROTR( (iTmp < -6) || (iTmp >  6) , Err::m_nInvalidParameter);
  setSliceAlphaC0Offset( iTmp << 1);

  RNOK( pcReadIf->getSvlc( iTmp,                                "SH: slice_beta_offset_div2" ) );
  ROTR( (iTmp < -6) || (iTmp >  6) , Err::m_nInvalidParameter);
  setSliceBetaOffset( iTmp << 1 );

  return Err::m_nOK;
}

// h264 namespace end
VIDEO_NAMESPACE_END

