/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoCommonLibPch.h"
#include "VideoCodecToolsLib/MSysLargeFile.h"
#include "VideoCommonLib/TraceCabac.h"
#include "VideoCommonLib/ProbabilityAllocator.h"
#include "VideoCommonLib/SyntaxElementCoder.h"

//#include <fstream>


// video namespace begin
VIDEO_NAMESPACE_BEGIN


UInt TraceCabacCategorizer::getCategoryIdx() const
{
  return UInt( m_pcAUPars->getPicOrderCnt() );
}

UInt TraceCabacCategorizer::getNumCategories() const
{
  return 300;
}

std::string TraceCabacCategorizer::getCategoryName() const
{
  std::stringstream cCat;
  cCat << "Frame-Nr " << getCategoryIdx() << ":";
  return cCat.str();
}


UInt TraceCabac::xGetMaxCtxNameLength() const
{
  UInt uiMaxNameLength = 0;
  for( UInt ui = 0; ui < m_acTraceCats.size(); ui++ )
  {
    for( UInt uiCell = 0; uiCell < UInt( m_acTraceCats[ui].m_acCategoryElements.size() ); uiCell++ )
    {
      UInt uiCurrStringLength = UInt( m_acTraceCats[ui].m_acCategoryElements[uiCell].m_acCtxName.length() );
      if( uiCurrStringLength > uiMaxNameLength )
      {
        uiMaxNameLength = uiCurrStringLength;
      }
    }
  }
  return uiMaxNameLength;
}

UInt TraceCabac::xGetOneAfterLastNonemptyBinsAndBitsElement( const std::vector<BinsAndBits> &racBinsAndBits )
{
  for( int i = int( racBinsAndBits.size() ) - 1; i >= 0; i-- )
  {
    if( racBinsAndBits[i].m_uiBins != 0 )
    {
      return i+1;
    }
  }
  return 0;
}


UInt TraceCabac::xGetNumNonemptyBinsAndBitsElements( const std::vector<BinsAndBits> &racBinsAndBits )
{
  UInt uiNumNonempty = 0;
  for( UInt i = 0; i < UInt( racBinsAndBits.size() ); i++ )
  {
    if( racBinsAndBits[i].m_uiBins != 0 )
    {
      uiNumNonempty++;
    }
  }
  return uiNumNonempty;
}


ErrVal TraceCabac::traceStoredCabacBits( LargeFile &rcTraceFile )
{
  std::stringstream acCaBits;
  acCaBits << std::fixed << std::setprecision( 2 );

  const UInt uiMaxNameLength = xGetMaxCtxNameLength() + 1;

  for( UInt ui = 0; ui < m_acTraceCats.size(); ui++ )
  {
    const CabacTraceCategory &rcCatData = m_acTraceCats[ui];
    if( !rcCatData.m_acCategoryElements.empty() )
    {
      acCaBits << rcCatData.m_acCatName << std::endl;
      for( UInt uiCell = 0; uiCell < UInt( rcCatData.m_acCategoryElements.size() ); uiCell++ )
      {
        const CabacTraceCategoryElement &rcCatElement = rcCatData.m_acCategoryElements[uiCell];

        acCaBits << std::setw( 4 ) << uiCell << " ";
        acCaBits << rcCatElement.m_acCtxName << std::setw( uiMaxNameLength - rcCatElement.m_acCtxName.length() ) << " " << "\t";

        const double dBits = rcCatElement.m_dCatElementBits;
        acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 12 ) << dBits << "\t";
        acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 7 ) << 100 * dBits / rcCatData.m_dSum << "\t";
        acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 7 ) << 100 * dBits / m_dSum;
        bool bTraceNumBins = true;
        if( bTraceNumBins )
        {
          acCaBits << "\t" << std::setw( 9 ) << rcCatElement.m_uiCatElementBins;
        }
#if TRACE_CABAC_BITS_DETAILED
        if( !rcCatElement.m_acDetailedBinsAndBits.empty() )
        {
          const UInt uiOneAfterLastNonempty = xGetOneAfterLastNonemptyBinsAndBitsElement( rcCatElement.m_acDetailedBinsAndBits );
          const UInt uiNumNonempty = xGetNumNonemptyBinsAndBitsElements( rcCatElement.m_acDetailedBinsAndBits );
          acCaBits << "\t" << std::setw( 8 ) << uiNumNonempty;
          acCaBits << "\t(\t";
          std::vector<double> adRealCabacBits;
          for( UInt k = 0; k < uiOneAfterLastNonempty; k++ )
          {
            const double dBitsDetailed = rcCatElement.m_acDetailedBinsAndBits[k].m_dBits;
            adRealCabacBits.push_back( dBitsDetailed );
            const UInt64 uiBins = rcCatElement.m_acDetailedBinsAndBits[k].m_uiBins;
            acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 12 )<< dBitsDetailed << "\t";
            if( uiBins != 0 )
            {
              acCaBits << std::fixed << std::setprecision( 2 ) << std::setw( 6 )<< double( uiBins )/dBitsDetailed << "\t|\t";
            }
            else
            {
              acCaBits << " --.-- " << "\t|\t";
            }
          }
          acCaBits << ")" << std::endl;
        }
        else
#endif
        {
          acCaBits << std::endl;
        }
      }
      acCaBits << "Sum" << std::setw( uiMaxNameLength + 2 ) << " " << "\t";
      acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 12 ) << rcCatData.m_dSum << "\t";
      acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 7 ) << 100.0 << "\t";
      acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 7 ) << 100 * rcCatData.m_dSum / m_dSum << std::endl;
      acCaBits << std::endl;
    }
  }
  acCaBits << "TotalSum" << std::setw( uiMaxNameLength - 8 ) << "\t";
  acCaBits << std::fixed << std::setprecision( 3 ) << std::setw( 12 ) << m_dSum << std::endl << std::endl;

  RNOK( rcTraceFile.write( acCaBits.str().c_str(), UInt( acCaBits.str().length() ) ) );
  return Err::m_nOK;
}


ErrVal TraceCabac::processExactCabacBits( const std::vector<IndexedBin> &rcVec, const CabacContextModel &rcInitCCModel, double &rdBits )
{
  rdBits = 0.0;
  if( !rcVec.empty() )
  {
    if( rcInitCCModel.isEpCtx() )
    {
      rdBits = double( rcVec.size() );
    }
    else
    {
      ExactCabacModel cModel;
        const double dAlpha = pow( 0.0375, 1.0/63.0 );
    //  const double dAlpha = pow( 2.0, -1.0/16.0 );
    //  const double dAlpha = pow( 2.0, -1.0/13.3 );
      cModel.setAlpha( dAlpha );

//      const double dMinProbVal = 0.5 * pow( dAlpha, 63 );
//      const double dMinProbVal = 0.5 * pow( dAlpha, 127 );
      const double dMinProbVal = 0.5 * pow( dAlpha, 80 );
      cModel.setMinProbVal( dMinProbVal );
      const bool bIsLps = true != (rcInitCCModel.getMps() != 0);
      cModel.setProbabilityOfOne( ProbBin::sm_adProbability[ bIsLps ? 64 + rcInitCCModel.getCabacState() : 63 - rcInitCCModel.getCabacState()] );

      for( UInt ui = 0; ui < UInt( rcVec.size() ); ui++ )
      {
        rdBits += cModel.simulateEncodingOfOneBin( rcVec[ui].getBin() );
      }
    }
  }
  return Err::m_nOK;
}


ErrVal TraceCabac::processSteadyStateBits( const std::vector<IndexedBin> &rcVec, const CabacContextModel &rcInitCCModel, double &rdBits )
{
  rdBits = 0.0;
  if( !rcVec.empty() )
  {
    if( rcInitCCModel.isEpCtx() )
    {
      rdBits = double( rcVec.size() );
    }
    else
    {
      UInt uiNumOnes = 0;
      UInt uiNumZeros = 0;
      for( UInt ui = 0; ui < UInt( rcVec.size() ); ui++ )
      {
        if( rcVec[ui].getBin() )
        {
          uiNumOnes++;
        }
        else
        {
          uiNumZeros++;
        }
      }
      rdBits = 7.0;
      if( uiNumOnes != 0 && uiNumZeros != 0 )
      {
        double dProb = double( uiNumOnes ) / double( uiNumOnes + uiNumZeros );
        double dMinusOneOverLog2 = -1.0 / log( 2.0 );
        rdBits += (double( uiNumOnes ) * log( dProb ) + double( uiNumZeros ) * log( 1.0 - dProb ) ) * dMinusOneOverLog2;
      }
    }
  }
  return Err::m_nOK;
}


ErrVal TraceCabac::xResizeCategoriesIfNecessary( UInt uiNumCategories )
{
  if( m_acTraceCats.empty() )
  {
    m_acTraceCats.resize( uiNumCategories );
  }
  ROF( m_acTraceCats.size() == uiNumCategories );
  return Err::m_nOK;
}


ErrVal TraceCabac::xInitCategoryIfNecessary( UInt uiCategoryIdx, std::string cCategoryName, const CtxAccess &rcCtxInit )
{
  CabacTraceCategory &rcCatData = m_acTraceCats[uiCategoryIdx];

  if( rcCatData.m_acCategoryElements.empty() )
  {
    rcCatData.m_acCatName = cCategoryName;
    rcCatData.m_acCategoryElements.resize( rcCtxInit.getNumCtxInitArrays() );
    for( UInt ui = 0; ui < UInt( rcCtxInit.getNumCtxInitArrays() ); ui++ )
    {
#if COLLECT_CABAC_BITS
      rcCatData.m_acCategoryElements[ui].m_acCtxName = rcCtxInit.getCtxName( ui );
#endif
    }
  }
  else
  {
    ROF( rcCatData.m_acCatName == cCategoryName );
    ROF( rcCtxInit.getNumCtxInitArrays() == int( rcCatData.m_acCategoryElements.size() ) );
  }
  return Err::m_nOK;
}


ErrVal TraceCabac::xResizeCategoryElementIfNecessary( CabacTraceCategoryElement &rcCatElement, UInt uiNumCtxElementsDetailed )
{
  if( rcCatElement.m_acDetailedBinsAndBits.empty() )
  {
    rcCatElement.m_acDetailedBinsAndBits.resize( uiNumCtxElementsDetailed );
  }
  if( rcCatElement.m_acDetailedBinsAndBits.size() != uiNumCtxElementsDetailed )
  {
    std::cout << "Number of Context Models has changed for: " << rcCatElement.m_acCtxName << std::endl;
    if( rcCatElement.m_acDetailedBinsAndBits.size() < uiNumCtxElementsDetailed )
    {
      rcCatElement.m_acDetailedBinsAndBits.resize( uiNumCtxElementsDetailed );
    }
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
