/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _ACCESSUNITDECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _ACCESSUNITDECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/PicBlock.h"
#include "VideoCommonLib/CodedPictureStore.h"
#include "ProbabilityAllocatorDec.h"
#include "BitReader.h"
#include "CabaDecoder.h"
#include "VLCDecoder.h"
#include "UvlcReader.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class ContextArrayStartIndices;
class CtxInitArray;
class CodingState;
template<class> class BitReaderIf;
class TraceCommon;
class TraceCabac;


class TraceCabacVecIf
{
public:
  virtual TraceCabac& getTraceCabac( UInt ui ) = 0;

protected:
  TraceCabacVecIf() {}
  virtual ~TraceCabacVecIf() {}
};


class CabacPassThrough
: public ProbBinStreamDecIf
{
public:
  CabacPassThrough( MultiVectorBitReader<UChar, UInt>::InitPars &rcPars ) { TNOK( m_cBitReader.init( rcPars ) ); }

  ErrVal startCabac() { return Err::m_nOK; }
  ErrVal finishCabac() { return Err::m_nOK; }

  ErrVal pushBack( ProbBin &rcT )
  {
    UInt uiBits = 0;
    RNOK( m_cBitReader.read( uiBits ) );
    bool bBit = uiBits ? true: false;
    rcT = ProbBin( rcT.getStateIdx(), bBit == rcT.isLps() );

    return Err::m_nOK;
  }

  ErrVal pushBack( bool &rcT )
  {
    UInt uiBits = 0;
    RNOK( m_cBitReader.read( uiBits ) );
    rcT = uiBits ? true : false;
    return Err::m_nOK;
  }

private:
  MultiVectorBitReader<UChar, UInt> m_cBitReader;
};


class AccessUnitDecoder
{
public:
  AccessUnitDecoder() {}
  virtual ~AccessUnitDecoder() {}

  static ErrVal decodeNalUnit( AccessUnitParameters& rcAUPars, 
                              std::vector<UChar> *pcRawData, 
                              const std::vector<UChar> &rcNalUnit, 
                              bool& rbSizeInfoAvailable, 
                              TraceCommon* pcTrace, 
                              SequenceParameterSet*& rpcSPS, 
                              PictureParameterSet*& rpcPPS);

  static ErrVal parseAccessUnit( const std::vector<UChar> &raucBitstream,
                                 const CodedPictureStoreBase &rcCodedPictureStore,
                                 AccessUnit *pcAU,
                                 TraceCabacVecIf *pcCabacTraceVecIf,
                                 PicOrderCnt &rcPoc,
                                 TraceCommon *pcTrace );

  static ErrVal decodeAccessUnit( const PicBlock<Pel> &rcDec,
                                  const std::vector<PicBlock<Pel> > &racOverlaidDec,
                                  const CodedPictureStore &rcCodedPictureStore,
                                  const AccessUnit *pcAU );

  static ErrVal decodePlanePackage( UInt uiPlaPaId,
                                    const PicBlock<Pel> &rcDec,
                                    const CodingState &rcCodingState,
                                    const AccessUnit &rcAU );

private:
  static ErrVal xParsePlanePackage( UInt uiPlaPaId,
                                    const CodedPictureStoreBase &rcCodedPictureStore,
                                    ProbBinStreamDecIf &rcProbBinInput,
                                    AccessUnit &rcAU,
                                    TraceCabacVecIf *pcCabacTraceVecIf,
                                    TraceCommon *pcTrace );

  static ErrVal xParsePlanePackages( const std::vector< std::vector<UChar> > &rcBitstream,
                                     const CodedPictureStoreBase &rcCodedPictureStore,
                                     AccessUnit &rcAU,
                                     TraceCabacVecIf *pcCabacTraceVecIf,
                                     TraceCommon *pcTrace );

  static ErrVal xParseWienerParameters( ProbBinStreamDecIf& rcProbBinStreamDecIf,
                                        AccessUnitParameters &rcAUPars,
                                        TraceCabacVecIf *pcCabacTraceVecIf,
                                        TraceCommon *pcTrace );

  template<class t_CabacTablesType>
  static ErrVal xReadMultiplex( BitReader<UChar,UInt> &rcBitReader,
                                const std::vector<UChar>::const_iterator &rcEnd,
                                std::vector<std::list<std::vector<UChar> > > &raaaucBins,
                                std::list<std::vector<UChar>::const_iterator> &raacIterEpStart,
                                std::list<std::vector<UChar>::const_iterator> &raacIterEpEnd,
                                PicOrderCnt cPoc );

};


template<class t_CabacTablesType>
ErrVal AccessUnitDecoder::xReadMultiplex( BitReader<UChar,UInt> &rcBitReader,
                                          const std::vector<UChar>::const_iterator &rcEnd,
                                          std::vector<std::list<std::vector<UChar> > > &raaaucBins,
                                          std::list<std::vector<UChar>::const_iterator> &raacIterEpStart,
                                          std::list<std::vector<UChar>::const_iterator> &raacIterEpEnd,
                                          PicOrderCnt cPoc )
{
  UvlcReader cPacketReader( rcBitReader );
  std::vector<UInt> auiTotalSizes( t_CabacTablesType::eNumAggregateStates );
  int uiLastSize = 0;
  for( UInt ui = 0; ui < t_CabacTablesType::eNumAggregateStates; ++ui )
  {
    int iCode;
    RNOK( cPacketReader.getSvlc( iCode, "diff_fragment_size_sum_in_bytes" ) );
    uiLastSize += iCode;
    auiTotalSizes[ui] = uiLastSize;
  }

  bool bFragmented = false;
  std::vector<UInt> auiFragmentSizes( t_CabacTablesType::eNumAggregateStates );
  for( UInt ui = 0; ui < t_CabacTablesType::eNumAggregateStates; ++ui )
  {
    if( auiTotalSizes[ui] )
    {
      UInt uiCode;
      RNOK( cPacketReader.getUvlc( uiCode, "fragment_size_in_bytes" ) );
      if( uiCode )
      {
        bFragmented = true;
      }
      auiFragmentSizes[ui] = uiCode ? uiCode : auiTotalSizes[ui];
    }
  }

  if( bFragmented )
  {
    bool bFragmentsAreIndep = false;
    RNOK( cPacketReader.getFlag( bFragmentsAreIndep, "fragments_are_independently_decodable" ) );
    ROF( bFragmentsAreIndep );
  }

//  std::vector<UChar>::const_iterator cBegin = rcEnd - rcBitReader.getNumRemainingBits() / 8;

  std::vector<UChar>::const_iterator cBegin = rcEnd - UInt(rcBitReader.getNumRemainingBitsToEnd() / 8);
  std::vector<std::vector<UChar>::const_iterator> acSubstreamsBegin( t_CabacTablesType::eNumAggregateStates );
  std::vector<std::vector<UChar>::const_iterator> acSubstreamsEnd( t_CabacTablesType::eNumAggregateStates );

  ROF( raaaucBins.empty() );
  raaaucBins.resize( t_CabacTablesType::eNumAggregateStates );

  bool bDone = false;
  while( !bDone )
  {
    for( UInt ui = 0; ui < t_CabacTablesType::eNumAggregateStates; ++ui )
    {
      acSubstreamsBegin[ui] = cBegin;
      if( auiTotalSizes[ui] > auiFragmentSizes[ui] )
      {
        cBegin += auiFragmentSizes[ui];
        auiTotalSizes[ui] -= auiFragmentSizes[ui];
      }
      else
      {
        cBegin += auiTotalSizes[ui];
        auiTotalSizes[ui] = 0;
      }
      acSubstreamsEnd[ui] = cBegin;
    }

    if( acSubstreamsBegin[0] != acSubstreamsEnd[0] )
    {
      raacIterEpStart.push_back( acSubstreamsBegin[0] );
      raacIterEpEnd.push_back( acSubstreamsEnd[0] );
    }

    bDone = true;
    for( UInt ui = 1; ui < t_CabacTablesType::eNumAggregateStates; ++ui )
    {
      if( acSubstreamsBegin[ui] != acSubstreamsEnd[ui] )
      {
        bDone = false;
        UInt uiNumBinsDecoded = 0;
        raaaucBins[ui].push_back( std::vector<UChar>() );
        bool bIsCabacFragment = auiTotalSizes[ui] != 0;
        RNOK( SingleStateCabacDecoder<typename t_CabacTablesType::Tables>::decodeCabac( t_CabacTablesType::sm_auiIdxToAggregateState[ui],
                                                                                        acSubstreamsBegin[ui],
                                                                                        acSubstreamsEnd[ui],
                                                                                        raaaucBins[ui].back(),
                                                                                        uiNumBinsDecoded,
                                                                                        bIsCabacFragment ) );
      }
    }
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _ACCESSUNITDECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
