/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _CABADECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _CABADECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/InStreamIf.h"
#include "VideoCommonLib/ExtendedCabacTables.h"
#include "ProbabilityAllocatorDec.h"
#include "BitReader.h"
#include "VideoCommonLib/BitWriterIf.h"
#include "MultiMpsCabaDecoder.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_UnsignedReadType> class BitReaderIf;


template<class t_Tables>
class SingleStateCabacDecoder
{
public:
  static ErrVal decodeCabac( const UInt uiState,
                             std::vector<UChar>::const_iterator cBegin,
                             std::vector<UChar>::const_iterator cEnd,
                             std::vector<UChar> &raucDecodedBins,
                             UInt &ruiNumBinsDecoded,
                             bool bIsCabacFragment )
  {
    BitReader<UChar, UInt>::InitPars cReaderPars = { false, cBegin, cEnd };
//    std::vector<UChar> aucTest( 1000, 255 );
//    BitReader<UChar, UInt>::CtorPars cReaderParsTest = { aucTest.begin(), aucTest.end() };
//    MultiMpsCabaDecoder<BitReader<UChar, UInt>, t_Tables, UInt64> cMMpsDecoderTest( cReaderParsTest, raucDecodedBins );
//    RNOK( cMMpsDecoderTest.startCabac() );
//    RNOK( cMMpsDecoderTest.decodeAllBins( uiState ) );

    MultiMpsCabaDecoder<BitReader<UChar, UInt>, t_Tables, UInt64> cMMpsDecoder( cReaderPars, raucDecodedBins, bIsCabacFragment );
    RNOK( cMMpsDecoder.startCabac() );
    RNOK( cMMpsDecoder.decodeAllBins( uiState ) );
    return Err::m_nOK;
  }
};


#if USE_CABAC12_TABLE_IN_DECODER
template<class t_Tables = AggregateCabacTables12>
#else
template<class t_Tables = DefaultCabacTables>
#endif
class CabaDecoder
  : public ProbBinStreamDecIf
  , public CabaCoderStartFinishIf
{
public:
  typedef typename t_Tables::Tables TablesType;

  CabaDecoder( BitReaderIf<UInt> &rcBitReaderIf );
  virtual ~CabaDecoder() {}

  ErrVal startCabac();
  ErrVal finishCabac();

  ErrVal pushBack( ProbBin &rcT );
  ErrVal pushBack( bool &rcT );

private:
  UInt xGetTerminateBufferBit();
  ErrVal xReadBit( UInt& ruiValue );

protected:
  BitReaderIf<UInt>* m_pcBitReaderIf;

private:
  UInt m_uiRange;
  UInt m_uiValue;
  UInt m_uiNumZerosAtEnd;
  bool m_bEndOfBits;
};


template<class t_Tables>
ErrVal CabaDecoder<t_Tables>::xReadBit( UInt& ruiValue )
{
  UInt uiBit = 0;
  RNOKS( m_pcBitReaderIf->read( uiBit, 1 ) );
  ruiValue = (ruiValue << 1) | uiBit;
  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaDecoder<t_Tables>::pushBack( ProbBin &rcT )
{
  UInt uiIdx = t_Tables::sm_auiStateToIdx[rcT.getStateIdx()];
  if( !TypeCompare<t_Tables,DefaultCabacTables>::m_bEqual && uiIdx == 0 )
  {
    bool bIsLps;
    RNOK( pushBack( bIsLps ) );
    rcT = ProbBin( rcT.getStateIdx(), bIsLps );
    return Err::m_nOK;
  }

  UInt uiLPS = TablesType::sm_auiLPSTable[t_Tables::sm_auiIdxToAggregateState[uiIdx]][(m_uiRange>>(6+TablesType::sm_uiDifBits)) & 0x03];
  m_uiRange -= uiLPS;

  if( m_uiValue < m_uiRange )
  {
    rcT = ProbBin( rcT.getStateIdx(), false );
  }
  else
  {
    m_uiValue -= m_uiRange,
    m_uiRange  = uiLPS,
    rcT = ProbBin( rcT.getStateIdx(), true );
  }

  while( m_uiRange < TablesType::sm_uiQuarter )
  {
    m_uiRange += m_uiRange;
    RNOKS( xReadBit( m_uiValue ) );
  }

  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaDecoder<t_Tables>::pushBack( bool &rcT )
{
  UInt uiValue = m_uiValue;

  RNOKS( xReadBit( uiValue ) );

  if( uiValue >= m_uiRange )
  {
    rcT = true;
    uiValue -= m_uiRange;
  }
  else
  {
    rcT = false;
  }
  m_uiValue = uiValue;
  return Err::m_nOK;
}


template<class t_Tables>
UInt CabaDecoder<t_Tables>::xGetTerminateBufferBit()
{
  UInt uiBit;
  UInt uiRange = m_uiRange-2;
  UInt uiValue = m_uiValue;

  if( uiValue >= uiRange )
  {
    uiBit = 1;
  }
  else
  {
    uiBit = 0;

    while( uiRange < TablesType::sm_uiQuarter )
    {
      uiRange += uiRange;
      RNOKS( xReadBit( uiValue ) );
    }

    m_uiRange = uiRange;
    m_uiValue = uiValue;
  }

  return uiBit;
}


template<class t_Tables>
CabaDecoder<t_Tables>::CabaDecoder( BitReaderIf<UInt> &rcBitReaderIf )
: m_pcBitReaderIf  ( &rcBitReaderIf )
, m_uiRange        ( 0 )
, m_uiValue        ( 0 )
, m_uiNumZerosAtEnd( 0 )
, m_bEndOfBits     ( false )
{
}


template<class t_Tables>
ErrVal CabaDecoder<t_Tables>::finishCabac()
{
  ROF( xGetTerminateBufferBit() );
  return Err::m_nOK;
}


template<class t_Tables>
ErrVal CabaDecoder<t_Tables>::startCabac()
{
  m_uiRange         = TablesType::sm_uiHalf - 2;
  m_uiValue         = 0;
  m_uiNumZerosAtEnd = TablesType::sm_uiBits - 1;
  m_bEndOfBits      = false;

  for( UInt n = 0; n < TablesType::sm_uiBits - 1; n++ )
  {
    RNOKS( xReadBit( m_uiValue ) );
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _CABADECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
