/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _MULTIMPSCABADECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _MULTIMPSCABADECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_


#include "ProbabilityAllocatorDec.h"
#include "VideoCommonLib/CabacTables.h"
#include "VideoCommonLib/BitWriter.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_BitReader, class t_Tables = CabacTables<0>, class t_UInt = UInt, UInt t_uiRenormReserve = 8 * (sizeof( t_UInt ) - 2)>
class MultiMpsCabaDecoder
  : public t_BitReader
  , private BitWriter<UChar, UInt>
{
public:
  MultiMpsCabaDecoder( typename t_BitReader::InitPars &rcBitReaderPars,
                       std::vector<UChar> &raucDecodedBins,
                       bool bIsCabacFragment )
    : BitWriter<UChar, UInt>( raucDecodedBins )
    , m_raucDecodedBins( raucDecodedBins )
    , m_bIsCabacFragment( bIsCabacFragment )
  {
    TNOK( t_BitReader::init( rcBitReaderPars ) );
    TOF( raucDecodedBins.empty() );
  }

  virtual ~MultiMpsCabaDecoder() {}

  ErrVal startCabac();
  ErrVal decodeAllBins( UInt uiState );

private:
  ErrVal xDecodeAllBinsMultiMps( UInt uiLog2Lps );
  ErrVal xDecodeAllBins( UInt uiState );
  ErrVal xReload();

  std::vector<UChar> &m_raucDecodedBins;
  UInt m_uiRange;
  t_UInt m_uiValue;
  int m_iBitsLeft;
  bool m_bArithmeticDecodingFinished;
  static const UInt sm_uiRenormTable32[32];
  bool m_bIsCabacFragment;
};


template<class t_BitReader, class t_Tables, class t_UInt, UInt t_uiRenormReserve>
const UInt MultiMpsCabaDecoder <t_BitReader, t_Tables, t_UInt, t_uiRenormReserve>::sm_uiRenormTable32[32] =
{
  6,5,4,4,
  3,3,3,3,
  2,2,2,2,
  2,2,2,2,
  1,1,1,1,
  1,1,1,1,
  1,1,1,1,
  1,1,1,1
};

template<class t_BitReader, class t_Tables, class t_UInt, UInt t_uiRenormReserve>
ErrVal MultiMpsCabaDecoder<t_BitReader, t_Tables, t_UInt, t_uiRenormReserve>::xDecodeAllBins( UInt uiState )
{
  while( true )
  {
    if( m_iBitsLeft <= 0 )
    {
      if( Err::m_nOK != xReload() )
      {
        break;
      }
    }

    UInt uiLPS = t_Tables::sm_auiLPSTable[uiState][(m_uiRange>>(6+t_Tables::sm_uiDifBits)) & 0x03];
    m_uiRange -= uiLPS;

    if( (m_uiValue >> m_iBitsLeft) < m_uiRange )
    {
      RNOK( write( 0 ) );
      if( m_uiRange < t_Tables::sm_uiQuarter )
      {
        m_uiRange <<= 1;
        m_iBitsLeft--;
      }
    }
    else
    {
      RNOK( write( 1 ) );
      m_uiValue -= t_UInt(m_uiRange) << m_iBitsLeft;
      int iRenorm = sm_uiRenormTable32[(uiLPS>>3) & 0x1F];
      m_uiRange = uiLPS << iRenorm;
      m_iBitsLeft -= iRenorm;
    }
  }
  return Err::m_nOK;
}


template<class t_BitReader, class t_Tables, class t_UInt, UInt t_uiRenormReserve>
ErrVal MultiMpsCabaDecoder<t_BitReader, t_Tables, t_UInt, t_uiRenormReserve>::xDecodeAllBinsMultiMps( UInt uiLog2Lps )
{
  const UInt uiLps     = 1 << uiLog2Lps;
  const int  iLpsPlus1 = 1 + uiLps;
  while( true )
  {
    if( m_iBitsLeft <= 0 )
    {
      if( Err::m_nOK != xReload() )
      {
        break;
      }
    }

    int iMpsRange = int( m_uiRange ) - iLpsPlus1 - int( m_uiValue >> m_iBitsLeft );
    if( iMpsRange >= 0 )
    {
      int iNumMpsMax = 1 + ( iMpsRange >> uiLog2Lps );
      int iNumMpsUntilShift = 1 + ((m_uiRange - t_Tables::sm_uiQuarter) >> uiLog2Lps);
      if( iNumMpsMax < iNumMpsUntilShift )
      {
        RNOK( write( 0, iNumMpsMax ) );
        m_uiRange -= iNumMpsMax << uiLog2Lps;
      }
      else
      {
        RNOK( write( 0, iNumMpsUntilShift ) );
        m_uiRange -= iNumMpsUntilShift << uiLog2Lps;
        m_uiRange <<= 1;
        m_iBitsLeft--;
      }
    }
    else
    {
      m_uiRange -= uiLps;
      RNOK( write( 1 ) );
      m_uiValue -= t_UInt(m_uiRange) << m_iBitsLeft;
      int iRenorm = sm_uiRenormTable32[(uiLps>>3) & 0x1F];
      m_uiRange = uiLps << iRenorm;
      m_iBitsLeft -= iRenorm;
    }
  }
  return Err::m_nOK;
}


template<class t_BitReader, class t_Tables, class t_UInt, UInt t_uiRenormReserve>
ErrVal MultiMpsCabaDecoder<t_BitReader, t_Tables, t_UInt, t_uiRenormReserve>::decodeAllBins( UInt uiState )
{
  UInt uiLps = t_Tables::sm_auiLPSTable[uiState][0];
  bool bOneRLpsOnly = uiLps == t_Tables::sm_auiLPSTable[uiState][1] &&
                      uiLps == t_Tables::sm_auiLPSTable[uiState][2] &&
                      uiLps == t_Tables::sm_auiLPSTable[uiState][3];
  bool bIsLog2Lps = false;
  UInt uiLog2Lps = 0;
  if( bOneRLpsOnly )
  {
    uiLog2Lps = gCeilLog2( uiLps );
    bIsLog2Lps = uiLps == 1u << uiLog2Lps;
  }

  if( bIsLog2Lps )
  {
    RNOK( xDecodeAllBinsMultiMps( uiLog2Lps ) );
  }
  else
  {
    RNOK( xDecodeAllBins( uiState ) );
  }

  if( m_bIsCabacFragment && (getNumberOfBitsWritten() & 7) )
  {
    m_raucDecodedBins.pop_back();
  }
  return Err::m_nOK;
}


template<class t_BitReader, class t_Tables, class t_UInt, UInt t_uiRenormReserve>
ErrVal MultiMpsCabaDecoder<t_BitReader, t_Tables, t_UInt, t_uiRenormReserve>::startCabac()
{
  // force byte alignment
  UInt uiNumRemainingBits = UInt( t_BitReader::getNumRemainingBits() );
  UInt uiStartingReserve = 8 + ( uiNumRemainingBits & 0x03 );
  if( uiStartingReserve < 10 )
  {
    uiStartingReserve += 8;
  }
  
  t_BitReader::template readWithSpecifiedReadTyp<t_UInt>( m_uiValue, gMin( uiStartingReserve, uiNumRemainingBits ) );
  if( uiNumRemainingBits < uiStartingReserve )
  {
    m_uiValue <<= uiStartingReserve - uiNumRemainingBits;
  }
  m_iBitsLeft = uiStartingReserve - 9;
  m_uiRange = t_Tables::sm_uiHalf - 2;
  m_bArithmeticDecodingFinished = false;
  return Err::m_nOK;
}


template<class t_BitReader, class t_Tables, class t_UInt, UInt t_uiRenormReserve>
ErrVal MultiMpsCabaDecoder<t_BitReader, t_Tables, t_UInt, t_uiRenormReserve>::xReload()
{
  t_UInt uiNextBits = 0;
  UInt uiNumRemainingBits = UInt( t_BitReader::getNumRemainingBits() );
  if( uiNumRemainingBits >= t_uiRenormReserve )
  {
    t_BitReader::template readWithSpecifiedReadTyp<t_UInt>( uiNextBits, t_uiRenormReserve );
    m_uiValue = (m_uiValue << t_uiRenormReserve) | uiNextBits;
    m_iBitsLeft += t_uiRenormReserve;
  }
  else
  {
    if( uiNumRemainingBits > 0 )
    {
      t_BitReader::template readWithSpecifiedReadTyp<t_UInt>( uiNextBits, uiNumRemainingBits );
      m_uiValue = ( m_uiValue << uiNumRemainingBits ) | uiNextBits;
      m_iBitsLeft += uiNumRemainingBits;
    }

    if( m_bArithmeticDecodingFinished )
    {
      return Err::m_nERR;
    }

    if( m_iBitsLeft < 1 )
    {
      m_bArithmeticDecodingFinished = true;
      if( m_bIsCabacFragment )
      {
        return Err::m_nERR;
      }
      else
      {
        m_uiValue   <<= 10;
        m_iBitsLeft  += 10;
      }
    }
  }
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _MULTIMPSCABADECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
