/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PROBABILITYALLOCATORDEC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PROBABILITYALLOCATORDEC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/ProbabilityAllocator.h"
#include "VideoCommonLib/CabacContextModel.h"
#include "VideoCommonLib/BinStreamIf.h"
#include "SyntaxElementDecoder.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef BinStreamIf<ProbBin&, false> ProbBinStreamDecIf;

template<class t_ContextModel = CabacContextModel>
class ProbabilityAllocatorDec
  : public ProbabilityAllocator<t_ContextModel>
  , private CtxBinStreamDecIf
{
public:
  ProbabilityAllocatorDec( ProbBinStreamDecIf &rcProbBinReadIf, UInt uiNumCtx, TraceCommon *pcTrace )
    : ProbabilityAllocator<t_ContextModel>( uiNumCtx, pcTrace )
    , m_rcProbBinReadIf  ( rcProbBinReadIf   ) {}

  virtual ~ProbabilityAllocatorDec() {}

  ErrVal pushBack( CtxBin &rcT );
  ErrVal pushBack( EPCtxBin &rcT );

  CtxBinStreamDecIf& getInStreamIf() { return *this; }

private:
  ProbBinStreamDecIf &m_rcProbBinReadIf;
};


template<class t_ContextModel>
ErrVal ProbabilityAllocatorDec<t_ContextModel>::pushBack( CtxBin &rcT )
{
  t_ContextModel &rcCCModel = ProbabilityAllocator<t_ContextModel>::m_pacContext[rcT.getCtx()];

  ProbBin cProbBin( rcCCModel.getCabacState(), true );
  static_cast<ProbBinStreamDecIf::TypeCall&>( m_rcProbBinReadIf ).pushBack( cProbBin );

  rcT.setBin( (rcCCModel.getMps() != 0) != cProbBin.isLps() );

  TRACE_NUMBER;
  TRACE_TO_STR( "Bin: ");
  TRACE_TO_STR( rcT.getBin() );
  TRACE_TO_STR( " Ctx: ");
  TRACE_TO_STR( rcT.getCtx() );
  TRACE_TO_STR( "Prob: " );
  TRACE_TO_STR( (UInt)rcCCModel.getCabacState() );
  TRACE_TO_STR( " isMps: " );
  TRACE_TO_STR( rcCCModel.getMps() ? 1 : 0 );
  TRACE_TO_STR( std::endl );

#if COLLECT_CABAC_BITS
  rcCCModel.addBits( cProbBin.getProbBinRate() );
  rcCCModel.addBin( rcT.getBin(), ProbabilityAllocator<t_ContextModel>::m_uiBinIndex++ );
#endif

  rcCCModel.updateCtx( cProbBin.isLps() );

  return Err::m_nOK;
}

template<class t_ContextModel>
ErrVal ProbabilityAllocatorDec<t_ContextModel>::pushBack( EPCtxBin &rcT )
{
#if COLLECT_CABAC_BITS
  static_cast<ProbBinStreamDecIf::BoolCall&>( m_rcProbBinReadIf ).pushBack( rcT.m_bBin );
  t_ContextModel &rcCCModel = ProbabilityAllocator<t_ContextModel>::m_pacContext[rcT.m_uiCtx];
  rcCCModel.addBits( 1.0 );
  rcCCModel.addBin( rcT.m_bBin, ProbabilityAllocator<t_ContextModel>::m_uiBinIndex++ );
#else
  static_cast<ProbBinStreamDecIf::BoolCall&>( m_rcProbBinReadIf ).pushBack( rcT );
#endif

  TRACE_NUMBER;
  TRACE_TO_STR( "Bin: ");
  TRACE_TO_STR( rcT );
  TRACE_TO_STR( " EqProb");
  TRACE_TO_STR( std::endl );

  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _PROBABILITYALLOCATORDEC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
