/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SYNTAXELEMENTDECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SYNTAXELEMENTDECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/SyntaxElementCoder.h"
#include "VideoCommonLib/BinStreamIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef SyntaxElementCoderIf<false> SyntaxElementDecoderIf;

#if COLLECT_CABAC_BITS
  typedef BinStreamIf<CtxBin&, false, EPCtxBin&> CtxBinStreamDecIf;
#else
  typedef BinStreamIf<CtxBin&, false> CtxBinStreamDecIf;
#endif

template<class t_BinProcessor = BinStreamProcessor<CtxBin&, false, EPCtxBin&> >
class SyntaxElementDecoder
  : public SyntaxElementCoder<t_BinProcessor, false>
{
private:
  typedef SyntaxElementCoder<t_BinProcessor, false> BaseCall;

public:
  SyntaxElementDecoder( const CtxAccess &rcCtxAccess,
                        const ContextArrayStartIndices &rcCtxArrayStartIndices,
                        const PlanePackageParameters& rcPlaPaPars,
                        TraceCommon *pcTrace,
                        typename t_BinProcessor::CtorPars &rcBinProcessorPars )
    : SyntaxElementCoder<t_BinProcessor, false>( rcCtxAccess, rcCtxArrayStartIndices, rcPlaPaPars, pcTrace, rcBinProcessorPars ) {}
  ~SyntaxElementDecoder() {}

private:
  ErrVal xProcessUnarySymbol           ( UInt &ruiSymbol,
                                         UInt uiCtxStart,
                                         UInt uiNumCtx,
                                         UInt uiSymbolMax );
  ErrVal xProcessDQp                   ( UInt uiDQpCtx,
                                         const CtxNames eContext,
                                         bool bNonZeroCoeffsPresent,
                                         const std::vector<int>& raiPredQp,
                                         std::vector<int> & raiQp );
  ErrVal xProcessSigLastPow2           ( const Block<int> &rcBlock,
                                         UInt uiPlane,
                                         int iLog2BlockSize,
                                         int iLog2PlaneBlockDim );
  ErrVal xProcessAbsGreaterOnePow2     ( int &rcCoeff, UInt uiCtx );
  ErrVal xProcessCoeffLevelPow2        ( int &rcCoeff,
                                         UInt uiCtx,
                                         UInt uiEpCtx );
  ErrVal xProcessSignPow2               ( int &rcCoeff, UInt uiEpCtx );
  ErrVal xProcessPredMode               ( const bool bIsIntraOnlySlice,
                                          bool &rbIsIntra,
                                          PredSyntax*& rpcPredSyntax,
                                          const PredSyntax *pcLeftData,
                                          const PredSyntax *pcTopData );
  ErrVal xProcessResMode                ( const QTNavi &rcNav,
                                          bool bIsIntra,
                                          bool bLeftIsAvailable,
                                          bool bTopIsAvailable,
                                          ResInfo::ResidualMode eLeftMode,
                                          ResInfo::ResidualMode eTopMode,
                                          const SliceData &rcCurrSlice,
                                          const PlanePackageParameters &rcPlaPaPars,
                                           ResInfo* &rpcResInfo );
  ErrVal xProcessMvDComponent           ( int &riMvDComponent,
                                          UInt uiLocalCtx,
                                          UInt uiCtxOffset );
  ErrVal xProcessInterPlaPaPredFlag     ( bool& rbIsInterPlaPaPred,
                                          PredSyntax*& rcPredSyntax,
                                          const QTNavi &rcPos,
                                          const PredSyntax* pcLeft,
                                          const PredSyntax* pcTop,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const AccessUnitParameters& rcAUPars );
  ErrVal xInferInterPlaPaPredFlag       ( bool& rbIsInterPlaPaPred,
                                          PredSyntax*& rcPredSyntax,
                                          const QTNavi &rcPos,
                                          const PredSyntax* pcLeft,
                                          const PredSyntax* pcTop,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const AccessUnitParameters& rcAUPars );
  ErrVal xProcessDirectModeFlag         ( bool& rbIsDirect,
                                          PredSyntax*& rcPredSyntax,
                                          const QTNavi &rcPos,
                                          const PredSyntax* pcLeft,
                                          const PredSyntax* pcTop,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const AccessUnitParameters& rcAUPars );
  ErrVal xProcessNumMvs                 ( PredSyntaxInterTranslational& rcPredSyntax,
                                          UInt uiContextStart,
                                          UInt uiMaxNumContexts,
                                          UInt uiMaxNumMvs );
  ErrVal xProcessRefIdxs                ( PredSyntaxInterTranslational& rcPredSyntax,
                                          UInt uiRefListSize );
  ErrVal xProcessMvPrecision            ( const PlanePackageParameters &rcPlaPaPars,
                                          int &riPrecision );
  ErrVal xProcessMvPredictor            ( const QTNavi &rcNav,
                                          PredSyntaxInterTranslational& rcPredSyntax,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const SliceData& rcSliceData,
                                          const UInt uiIdx );
  ErrVal xProcessResTrafoSkip           ( const QTNavi &rcPos,
                                          const SliceData &rcContextSlice,
                                          UInt uiNumPlanes,
                                          ResInfoTrafoSkip &rcResInfo );
  ErrVal xReadEpExGolomb                ( UInt &ruiSymbol, UInt uiCount, UInt uiEpCtx );
  ErrVal xProcessInterTranslational       ( const QTNavi &rcPos,
                                          const PlanePackageParameters &rcPlaPaPars,
                                          const AccessUnitParameters &rcAUPars,
                                          PredSyntax *&rpcPredSyntax,
                                          const SliceData& rcSliceData,
                                          const PredSyntax *pcLeft,
                                          const PredSyntax *pcTop );
  ErrVal xSetMVAddress                  ( const QTNavi& rcPos,
                                          PredSyntaxInterTranslational& rcPredSyntax,
                                          const PredSyntaxInterTranslational* pcTop,
                                          const PredSyntaxInterTranslational* pcLeft );

  ErrVal xProcessWienerParameters       ( WienerFilterParameters& rcWFParam , UInt uiNumOfPlanes );
  ErrVal xProcessWienerFilterLength     ( UInt &ruiFilterLength ) ;
  ErrVal xProcessWienerFilterCoeffs     ( FilterCoefficients& rcCoeffs );
};


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessSigLastPow2( const Block<int>&  rcBlock,
                                                                  UInt uiPlane,
                                                                  int iLog2PrimaryBlockSize,
                                                                  int iLog2PlaneBlockDim )
{
  ROT( iLog2PlaneBlockDim > iLog2PrimaryBlockSize );
  const UInt  uiPlaneBlkSize    = 1 << iLog2PlaneBlockDim;
  const UInt  uiMaxNumCoeff     = uiPlaneBlkSize * uiPlaneBlkSize;
  const UInt  uiMaxNumCoeffM1   = uiMaxNumCoeff  - 1;
  UInt        uiDownLeft        = 1;
  UInt        uiNumSigTopRight  = 0;
  UInt        uiNumSigBotLeft   = 0;
  bool        bLastReceived     = false;

  rcBlock.assign( 0 );
  std::vector<std::vector<UInt> >&  raauiScanPattern = BaseCall::m_aaauiScanPattern[ iLog2PlaneBlockDim ];

  for( UInt uiScanPos = 0; uiScanPos < uiMaxNumCoeffM1; uiScanPos++ )
  {
    UInt  uiBlkPos  = raauiScanPattern[ uiDownLeft ][ uiScanPos ];
    UInt  uiPosY    = uiBlkPos >> iLog2PlaneBlockDim;
    UInt  uiPosX    = uiBlkPos - ( uiPosY << iLog2PlaneBlockDim );

    //===== code significance flag =====
    bool bSig = false;
    UInt  uiCtxSigBase  = BaseCall::m_cCtxArrayStartIndices.m_acResCtx[ uiPlane ].m_uiStartTrafoPow2Sig + UInt( iLog2PrimaryBlockSize );
    UInt  uiCtxSigInc   = ResTrafoPow2::getSigCtxInc( rcBlock, uiPosX, uiPosY, UInt( iLog2PlaneBlockDim ), ( uiDownLeft > 0 ) );
    UInt  uiCtxSig      = BaseCall::m_cCtxAccess.getContextOffset( uiCtxSigBase, uiCtxSigInc );
    RNOK( BaseCall::xProcessBin( bSig, uiCtxSig ) );
    TRACE_NUMBER;
    TRACE_TO_STR( "Sig: " );
    TRACE_TO_STR( bSig );
    TRACE_TO_STR( std::endl );

    if( bSig )
    {
      rcBlock.getEl( uiPosX, uiPosY ) = 1;
      if( uiPosX > uiPosY )
      {
        uiNumSigTopRight++;
      }
      else if( uiPosY > uiPosX )
      {
        uiNumSigBotLeft++;
      }

      //===== code last flag =====
      bool  bLast         = false;
      UInt  uiCtxLastBase = BaseCall::m_cCtxArrayStartIndices.m_acResCtx[ uiPlane ].m_uiStartTrafoPow2Last + UInt( iLog2PrimaryBlockSize );
      UInt  uiCtxLastInc  = ResTrafoPow2::getLastCtxInc( rcBlock, uiPosX, uiPosY, UInt( iLog2PlaneBlockDim ) );
      UInt  uiCtxLast     = BaseCall::m_cCtxAccess.getContextOffset( uiCtxLastBase, uiCtxLastInc );
      RNOK( BaseCall::xProcessBin( bLast, uiCtxLast ) );
      TRACE_NUMBER;
      TRACE_TO_STR( "Last: " );
      TRACE_TO_STR( bLast );
      TRACE_TO_STR( std::endl );

      if( bLast )
      {
        bLastReceived = true;
        break;
      }
    }

    //===== update scan direction =====
    if( ( uiDownLeft == 1 && ( uiPosX == 0 || uiPosY == uiPlaneBlkSize - 1 ) ) ||
        ( uiDownLeft == 0 && ( uiPosY == 0 || uiPosX == uiPlaneBlkSize - 1 ) )   )
    {
      uiDownLeft = ( uiNumSigTopRight >= uiNumSigBotLeft ? 1 : 0 );
    }
  }
  if( ! bLastReceived )
  {
    rcBlock.getEl( uiPlaneBlkSize - 1, uiPlaneBlkSize - 1 ) = 1;
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessAbsGreaterOnePow2( int &rcCoeff, UInt uiCtx )
{
  bool bSymbol = false;
  RNOK( BaseCall::xProcessBin( bSymbol, uiCtx ) );
  if( bSymbol )
  {
    rcCoeff = 2;
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xReadEpExGolomb( UInt &ruiSymbol, UInt uiCount, UInt uiEpCtx )
{
  ruiSymbol = 0;
  bool bBit = true;

  while( bBit )
  {
    RNOK( BaseCall::xProcessEqProbBin( bBit, uiEpCtx ) );
    ruiSymbol += (bBit ? 1 : 0) << uiCount++;
  }

  uiCount--;
  while( uiCount-- )
  {
    RNOK( BaseCall::xProcessEqProbBin( bBit, uiEpCtx ) );
    ruiSymbol += (bBit ? 1 : 0) << uiCount;
  }

  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessCoeffLevelPow2( int &rcCoeff,
                                                                     UInt uiCtx,
                                                                     UInt uiEpCtx )
{
  bool bSymbol = false;
  UInt uiCount = 0;
  do
  {
    RNOK( BaseCall::xProcessBin( bSymbol, uiCtx ) );
    uiCount++;
  }
  while( bSymbol && (uiCount != 13));

  rcCoeff = uiCount-1;

  if( bSymbol )
  {
    UInt uiSymbol;
    RNOK( xReadEpExGolomb( uiSymbol, 0, uiEpCtx ) );
    rcCoeff += uiSymbol+1;
  }
  rcCoeff += 2;
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessSignPow2( int &rcCoeff, UInt uiEpCtx )
{
  bool bSign = false;
  RNOK( BaseCall::xProcessEqProbBin( bSign, uiEpCtx ) );
  if( bSign )
  {
    rcCoeff = -rcCoeff;
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessUnarySymbol( UInt &ruiSymbol,
                                                                  UInt uiCtxStart,
                                                                  UInt uiNumCtx,
                                                                  UInt uiSymbolMax )
{
  bool bBin = false;

  ROT( uiNumCtx == 0 );

  TRACE_NUMBER;
  TRACE_TO_STR( "Unary: " );

  UInt ui = 0;
  for( ; ui < uiSymbolMax; ui++ )
  {
    const UInt uiCtx = uiCtxStart + gMin( ui, uiNumCtx - 1 );
    RNOK( BaseCall::xProcessBin( bBin, uiCtx ) );

    TRACE_TO_STR( bBin ? 1 : 0 );
    TRACE_TO_STR( "," );
    TRACE_TO_STR( uiCtx );
    TRACE_TO_STR( " " );

    if( ! bBin )
    {
      break;
    }
  }

  TRACE_TO_STR( std::endl );

  ruiSymbol = ui;
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessDQp( UInt uiDQpCtx,
                                                          const CtxNames eContext,
                                                          bool bNonZeroCoeffsPresent,
                                                          const std::vector<int>& raiPredQp,
                                                          std::vector<int>& raiQp )
{
  ROF( raiPredQp.size() == raiQp.size() );
  int iDQp = 0;
  if( bNonZeroCoeffsPresent )
  {
    bool bBin = false;
    RNOK( BaseCall::xProcessBin( bBin,  BaseCall::m_cCtxAccess.getContextOffset( eContext, uiDQpCtx ) ) );
    if( bBin )
    {
      UInt uiDQp;
      RNOK( xProcessUnarySymbol( uiDQp,  BaseCall::m_cCtxAccess.getContextOffset( eContext, 3 ), 2, TypeLimits<UInt>::m_iMax ) );
      iDQp = ( uiDQp + 2 ) / 2;
      if( uiDQp & 1 )
      {
        iDQp = -iDQp;
      }
    }
  }
  for( UInt uiPlane = 0; uiPlane < raiPredQp.size(); uiPlane++ )
  {
    raiQp[ uiPlane ] = raiPredQp[ uiPlane ] + iDQp;
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessPredMode( const bool bIsIntraOnlySlice,
                                                               bool &rbIsIntra,
                                                               PredSyntax *&rpcPredSyntax,
                                                               const PredSyntax *pcLeftData,
                                                               const PredSyntax *pcTopData )
{
  UInt  uiLeftIntra = ( pcLeftData && pcLeftData->isIntraCoded() ? 1 : 0 );
  UInt  uiTopIntra  = ( pcTopData  && pcTopData ->isIntraCoded() ? 1 : 0 );

  if( !bIsIntraOnlySlice )
  {
    UInt uiIntraCtx = 2 + uiLeftIntra + uiTopIntra;
    RNOK( BaseCall::xProcessBin( rbIsIntra, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, uiIntraCtx ) ) );

    TRACE_NUMBER;
    TRACE_TO_STR( "intra_flag: " );
    TRACE_TO_STR( rbIsIntra );
    TRACE_TO_STR( std::endl );
  }
  else
  {
    rbIsIntra = true;
  }

  if( rbIsIntra )
  {
    UInt  uiLeftModeIdx = 0;
    UInt  uiTopModeIdx  = 0;
    if( uiLeftIntra )
    {
      RNOK( BaseCall::xGetIntraModeIdx( *pcLeftData, uiLeftModeIdx ) );
    }
    if( uiTopIntra )
    {
      RNOK( BaseCall::xGetIntraModeIdx( *pcTopData,  uiTopModeIdx  ) );
    }

    UInt uiMostProbable = 2; // assume INTRA_DC
    if( uiLeftIntra && uiTopIntra )
    {
      uiMostProbable = gMin( uiLeftModeIdx, uiTopModeIdx );
    }
    bool bIsMostProbable        = false;
    UInt uiMostProbableFlagCtx  = 0;
    if( uiLeftIntra && uiTopIntra )
    {
      uiMostProbableFlagCtx = 5 + ( uiLeftModeIdx == uiTopModeIdx ? 0 : 1 );
    }

    RNOK( BaseCall::xProcessBin( bIsMostProbable, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, uiMostProbableFlagCtx ) ) );
    
    TRACE_NUMBER;
    TRACE_TO_STR( "most_probable_ipred_mode_flag: " );
    TRACE_TO_STR( bIsMostProbable );
    TRACE_TO_STR( std::endl );

    UInt uiIntraIdx = 0;
    if( !bIsMostProbable )
    {
      bool b0 = false, b1 = false, b2 = false;

      RNOK( BaseCall::xProcessBin( b2, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, 1 ) ) );
      RNOK( BaseCall::xProcessBin( b1, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, 1 ) ) );
      RNOK( BaseCall::xProcessBin( b0, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, 1 ) ) );
      uiIntraIdx += b0 ? 4 : 0;
      uiIntraIdx += b1 ? 2 : 0;
      uiIntraIdx += b2 ? 1 : 0;

      TRACE_NUMBER;
      TRACE_TO_STR( "intra_pred_mode: " );
      TRACE_TO_STR( uiIntraIdx );
      TRACE_TO_STR( std::endl );

      if( uiIntraIdx >= uiMostProbable )
      {
        uiIntraIdx++;
      }
    }
    else
    {
      uiIntraIdx = uiMostProbable;
    }

    if( uiIntraIdx == 2 )
    {
      rpcPredSyntax = new PredSyntaxIntraDC( false );
      TRACE_NUMBER;
      TRACE_TO_STR( "INTRA_DC " );
      TRACE_TO_STR( std::endl );
    }
    else
    {
      PredSyntaxIntraDirectional *pInfo = new PredSyntaxIntraDirectional( false, PredInfoIntraDirectional::IntraDirections( uiIntraIdx ) );

      TRACE_NUMBER;
      TRACE_TO_STR( "INTRA_DIRECTIONAL: " );
      TRACE_TO_STR( pInfo->getIntraDirection() );
      TRACE_TO_STR( std::endl );

      bool bSmoothedFlag = false;
      RNOK( BaseCall::xProcessBin( bSmoothedFlag, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_INTRA_SMOOTHED, pInfo->getIntraDirection() ) ) );
      pInfo->setSmoothedRefPixelFlag( bSmoothedFlag );
      TRACE_NUMBER;
      TRACE_TO_STR( "smooth_ref_samples_flag: " );
      TRACE_TO_STR( pInfo->getSmoothedRefPixelFlag() );
      TRACE_TO_STR( std::endl );

      rpcPredSyntax = pInfo;
    }
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessResMode( const QTNavi &rcNav,
                                                              bool bIsIntra,
                                                              bool bLeftIsAvailable,
                                                              bool bTopIsAvailable,
                                                              ResInfo::ResidualMode eLeftMode,
                                                              ResInfo::ResidualMode eTopMode,
                                                              const SliceData &rcCurrSlice,
                                                              const PlanePackageParameters &rcPlaPaPars,
                                                              ResInfo *&rpcResInfo )
{
  ResInfo::ResidualMode eResMode;

  UInt uiNumQuantizers = rcPlaPaPars.getNumQuantizers();
  UInt uiQuantizers    = rcPlaPaPars.getTransformMode();
  if( uiNumQuantizers > 1 && (uiQuantizers & ResInfo::RES_TRAFO_SKIP) )
  {
    if( !bIsIntra )
    {
      bool bIsSkip = false;
      UInt uiSkipCtx = 0;
      RNOK( BaseCall::xGetSkipModeCtx( bIsIntra, bLeftIsAvailable, bTopIsAvailable, eLeftMode, eTopMode, uiSkipCtx  ) );
      RNOK( BaseCall::xProcessBin( bIsSkip, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_RES_SKIP_MODE, uiSkipCtx ) ) );

      TRACE_NUMBER;
      TRACE_TO_STR( "trafo_skip_flag: " );
      TRACE_TO_STR( bIsSkip );
      TRACE_TO_STR( std::endl );

      if( bIsSkip )
      {
        rpcResInfo = new ResInfoTrafoSkip( rcPlaPaPars.getNumPlanes() );
        return Err::m_nOK;
      }
    }
    uiNumQuantizers--;
    uiQuantizers &= ~UInt( ResInfo::RES_TRAFO_SKIP );
  }

  ROT( uiNumQuantizers > 1 );

  eResMode = ResInfo::ResidualMode( uiQuantizers );

  ROF( eResMode == ResInfo::RES_TRAFO_POW2 );

  // mode RES_TRAFO_POW2
  QuadtreeGroupStructure* pcQuadTreeGroupStructure = new QuadtreeGroupStructure( rcNav.getPrimBlockSize(), gMin( rcNav.getLog2BlockSize(), rcPlaPaPars.getLog2MaxPow2TrafoBlkSize() ), 0 );
  rpcResInfo = new ResInfoTrafoPow2( rcPlaPaPars.getNumPlanes(), *pcQuadTreeGroupStructure );

  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessResTrafoSkip( const QTNavi &rcPos,
                                                                   const SliceData &rcContextSlice,
                                                                   UInt uiNumPlanes,
                                                                   ResInfoTrafoSkip &rcResInfo )
{
  RNOK( rcContextSlice.derivePredQPs( rcPos, rcResInfo.getQps() ) );
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xSetMVAddress( const QTNavi& rcPos,
                                                            PredSyntaxInterTranslational& rcPredSyntax,
                                                            const PredSyntaxInterTranslational* pcTop,
                                                            const PredSyntaxInterTranslational* pcLeft )
{
  PrimPixelPos cMVAddressTop = pcTop ? pcTop->getMVAddress() : PrimPixelPos( 0, 0 );
  PrimPixelPos cMVAddressLeft = pcLeft ? pcLeft->getMVAddress() : PrimPixelPos( 0, 0 );
  PrimPixelPos cMVAddress;

  if ( !rcPredSyntax.getDirectModeFlag() && rcPredSyntax.getMergeFlag() )
  {
    // copy MV address
    if ( pcTop && pcLeft )
    {
      const PredInfoInterTranslational* pcPredInfoITTop = pcTop->getPredInfoInterTranslational();
      const PredInfoInterTranslational* pcPredInfoITLeft = pcLeft->getPredInfoInterTranslational();
      bool bSameVectors = PredInfoInterTranslational::compareMvAndRefPocVec( pcPredInfoITTop->getMvAndRefPocVec(), pcPredInfoITLeft->getMvAndRefPocVec() );
      if ( !bSameVectors && rcPredSyntax.getMergeLeftFlag() )
      {
        cMVAddress = cMVAddressLeft;
      }
      else
      {
        cMVAddress = cMVAddressTop;
      }
    }
    else if ( pcTop )
    {
      cMVAddress = cMVAddressTop;
    }
    else if ( pcLeft )
    {
      cMVAddress = cMVAddressLeft;
    }
    else
    {
      return Err::m_nERR;
    }
  }
  else
  {
    // Block is not merged. Save Address of current block.
    cMVAddress = rcPos.getPrimPixelPos();
  }

  rcPredSyntax.setMVAddress( cMVAddress );

  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessInterTranslational( const QTNavi &rcPos,
                                                                       const PlanePackageParameters &rcPlaPaPars,
                                                                       const AccessUnitParameters &rcAUPars,
                                                                       PredSyntax *&rpcPredSyntax,
                                                                       const SliceData &rcSliceData,
                                                                       const PredSyntax *pcLeft,
                                                                       const PredSyntax *pcTop )
{
  PredSyntaxInterTranslational*       pcTransCurr = new PredSyntaxInterTranslational( false );
  const PredSyntaxInterTranslational* pcTransLeft = ( pcLeft ? pcLeft->getPredSyntaxInterTranslational() : 0 );
  const PredSyntaxInterTranslational* pcTransTop  = ( pcTop  ? pcTop ->getPredSyntaxInterTranslational() : 0 );
  RNOK(( SyntaxElementCoder<t_BinProcessor,false>::xProcessInterTranslational( rcPos, rcPlaPaPars, rcAUPars, *pcTransCurr, rcSliceData, pcTransLeft, pcTransTop ) ));

  xSetMVAddress( rcPos, *pcTransCurr, pcTransTop, pcTransLeft );

  rpcPredSyntax = pcTransCurr;
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessMvDComponent( int &riMvDComponent,
                                                                   UInt uiLocalCtx,
                                                                   UInt uiCtxOffset )
{
  bool bSymbol    = false;
  riMvDComponent  = 0;
  // unary part
  do 
  {
    RNOK( BaseCall::xProcessBin( bSymbol, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MVD, uiCtxOffset + uiLocalCtx ) ) );
    if( UInt( riMvDComponent ) <  BaseCall::m_uiNumMvDUnaryCtxPerComponent )
    {
      uiLocalCtx = 3 + riMvDComponent;
    }
    riMvDComponent++;
  } while( bSymbol && riMvDComponent < 9 );
  riMvDComponent--;
  // exp-golomb part
  if( bSymbol )
  {
    UInt uiExpGolomb;
    RNOK( xReadEpExGolomb( uiExpGolomb, 3,  BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MVD ) ) );
    riMvDComponent += uiExpGolomb + 1;
  }
  // sign
  if( riMvDComponent != 0 )
  {
    RNOK( BaseCall::xProcessEqProbBin( bSymbol, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MVD ) ) );
    if( bSymbol )
    {
      riMvDComponent = -riMvDComponent;
    }
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessInterPlaPaPredFlag ( bool &rbIsInterPlaPaPred,
                                                                          PredSyntax *&rpcPredSyntax,
                                                                          const QTNavi &rcPos,
                                                                          const PredSyntax *pcLeft,
                                                                          const PredSyntax *pcTop,
                                                                          const PlanePackageParameters &rcPlaPaPars,
                                                                          const AccessUnitParameters &rcAUPars )
{
  rbIsInterPlaPaPred = false;
  ROFRS( rcPlaPaPars.getInterPlaPaPredFlag(),         Err::m_nOK );
  ROFRS( rcPlaPaPars.getAdaptiveInterPlaPaPredFlag(), Err::m_nOK );
  const UInt  uiLeft        = ( !pcLeft || pcLeft->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE ? 1 : 0 );
  const UInt  uiTop         = ( !pcTop  || pcTop ->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE ? 1 : 0 );
  const UInt  uiCtx         = uiLeft + uiTop;
  RNOK( BaseCall::xProcessBin( rbIsInterPlaPaPred,  BaseCall::m_cCtxAccess.getContextOffset( CTX_START_INTER_PLAPA_PRED, uiCtx ) ) );
  TRACE_NUMBER;
  TRACE_TO_STR( "inter_pla_pa_pred_flag: " );
  TRACE_TO_STR( rbIsInterPlaPaPred );
  TRACE_TO_STR( std::endl );
  if( rbIsInterPlaPaPred )
  {
    rpcPredSyntax = new PredSyntaxInterPlanePackage( rcPos );
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xInferInterPlaPaPredFlag ( bool &rbIsInterPlaPaPred,
                                                                        PredSyntax *&rpcPredSyntax,
                                                                        const QTNavi &rcPos,
                                                                        const PredSyntax *pcLeft,
                                                                        const PredSyntax *pcTop,
                                                                        const PlanePackageParameters &rcPlaPaPars,
                                                                        const AccessUnitParameters &rcAUPars )
{
  ROF( rcAUPars.getMaxMvsPerBlk() > 0 );
  rbIsInterPlaPaPred = ( rcPlaPaPars.getInterPlaPaPredFlag() && !rcPlaPaPars.getAdaptiveInterPlaPaPredFlag() );
  if( rbIsInterPlaPaPred )
  {
    rpcPredSyntax = new PredSyntaxInterPlanePackage( rcPos );
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessDirectModeFlag( bool &rbIsDirect,
                                                                     PredSyntax *&rpcPredSyntax,
                                                                     const QTNavi &rcPos,
                                                                     const PredSyntax *pcLeft,
                                                                     const PredSyntax *pcTop,
                                                                     const PlanePackageParameters &rcPlaPaPars,
                                                                     const AccessUnitParameters &rcAUPars )
{
  ROFRS( rcAUPars.isDirectModeFlagPresent( rcPlaPaPars.getPlaPaId() ), Err::m_nOK );
  const UInt  uiLeft  = ( !pcLeft ? 0 : ( !pcLeft->getPredSyntaxInterTranslational() ? 1 : ( pcLeft->getPredSyntaxInterTranslational()->getDirectModeFlag() ? 0 : 1 ) ) );
  const UInt  uiTop   = ( !pcTop  ? 0 : ( !pcTop ->getPredSyntaxInterTranslational() ? 1 : ( pcTop ->getPredSyntaxInterTranslational()->getDirectModeFlag() ? 0 : 1 ) ) );
  const UInt  uiCtx   = uiLeft + uiTop;
  RNOK( BaseCall::xProcessBin( rbIsDirect,  BaseCall::m_cCtxAccess.getContextOffset( CTX_START_DIRECT_MODE, uiCtx ) ) );
  if( rbIsDirect )
  {
    PredSyntaxInterTranslational* pcPredSyntax = new PredSyntaxInterTranslational( false );
    pcPredSyntax->setDirectModeFlag( true );
    pcPredSyntax->setNumEntries( rcAUPars.getMaxMvsPerBlk() );
    for( UInt uiIdx = 0; uiIdx < pcPredSyntax->getNumEntries(); uiIdx++ )
    {
      pcPredSyntax->getRefIdx   ( uiIdx ) = uiIdx;
      pcPredSyntax->getMvPredIdx( uiIdx ) = 0;
      pcPredSyntax->getMvDiff   ( uiIdx ) = MotionVector( 0, 0, (int)rcPlaPaPars.getMvAccuracy() );
    }

    const PredSyntaxInterTranslational*  pcTopIT  = ( !pcTop  ? 0 : pcTop->getPredSyntaxInterTranslational() );
    const PredSyntaxInterTranslational*  pcLeftIT = ( !pcLeft ? 0 : pcLeft->getPredSyntaxInterTranslational() );
    xSetMVAddress( rcPos, *pcPredSyntax, pcTopIT, pcLeftIT );

    rpcPredSyntax = pcPredSyntax;
  }
  TRACE_NUMBER;
  TRACE_TO_STR( "direct_mode_flag: " );
  TRACE_TO_STR( rbIsDirect );
  TRACE_TO_STR( std::endl );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessNumMvs( PredSyntaxInterTranslational& rcPredSyntax,
                                                             UInt uiContextStart,
                                                             UInt uiMaxNumContexts,
                                                             UInt uiMaxNumMvs )
{
  UInt uiTemp = 0;
  RNOK( xProcessUnarySymbol( uiTemp, uiContextStart, uiMaxNumContexts, uiMaxNumMvs - 1 ) );
  rcPredSyntax.setNumEntries( uiTemp + 1 );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessRefIdxs( PredSyntaxInterTranslational &rcPredSyntax,
                                                              UInt uiRefListSize )
{
  RNOK( xProcessUnarySymbol( rcPredSyntax.getRefIdx( 0 ),
                             BaseCall::m_cCtxAccess.getContextOffset (  BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ),
                             BaseCall::m_cCtxAccess.getNumCtxElements(  BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ),
                             uiRefListSize - 1 ) );

  if( uiRefListSize - 1 > 0 )
  {
    TRACE_NUMBER;
    TRACE_TO_STR( "ref_idx: " );
    TRACE_TO_STR( rcPredSyntax.getRefIdx( 0 ) );
    TRACE_TO_STR( std::endl );
  }

  for( UInt uiIdx = 1; uiIdx < rcPredSyntax.getNumEntries(); uiIdx++ )
  {
    UInt  uiLastRefIdx  = rcPredSyntax.getRefIdx( uiIdx - 1 );
    UInt  uiRemListSize = uiRefListSize - uiLastRefIdx;
    UInt  uiSymbol      = 0;
    RNOK( xProcessUnarySymbol( uiSymbol,
                               BaseCall::m_cCtxAccess.getContextOffset (  BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ) + uiLastRefIdx,
                               BaseCall::m_cCtxAccess.getNumCtxElements(  BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ) - uiLastRefIdx,
                               uiRemListSize - 1 ) );
    if( uiRemListSize - 1 > 0 )
    {
      TRACE_NUMBER;
      TRACE_TO_STR( "delta_ref_idx_minus1: " );
      TRACE_TO_STR( uiSymbol );
      TRACE_TO_STR( std::endl );
    }
    rcPredSyntax.getRefIdx( uiIdx ) = uiLastRefIdx + ( uiSymbol == uiRemListSize - 1 ? 0 : uiSymbol + 1 );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessMvPrecision( const PlanePackageParameters& rcPlaPaPars, int& riPrecision )
{
  riPrecision = (int)rcPlaPaPars.getMvAccuracy();
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessMvPredictor( const QTNavi &rcNav,
                                                                  PredSyntaxInterTranslational& rcPredSyntax,
                                                                  const PlanePackageParameters& rcPlaPaPars,
                                                                  const SliceData& rcSliceData,
                                                                  const UInt uiIdx )
{
  ROFRS( rcPlaPaPars.getMvCompetition(), Err::m_nOK );
  const QuadtreeGroup<PredInfo*> &rcTree( rcSliceData.getPredInfoTree() );
  UInt  uiNumMvPred = 0;
  UInt  auiCtx[2]   = { 0, 0 }; 
  UInt  uiVal       = 0;
  RNOK ( InterTranslational::getNumMvPred( rcPredSyntax.getRefIdx( uiIdx ), rcNav, rcTree, uiNumMvPred, auiCtx ) );
  if( uiNumMvPred > 1 )
  {
    bool  bFlag = false;
    RNOK( BaseCall::xProcessBin( bFlag, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MV_PRED_IDX, auiCtx[0] ) ) );
    uiVal += ( bFlag ? 1 : 0 );
    if( bFlag && uiNumMvPred > 2 )
    {
      RNOK( BaseCall::xProcessBin( bFlag, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MV_PRED_IDX, auiCtx[1] ) ) );
      uiVal += ( bFlag ? 1 : 0 );
    }
  }
  rcPredSyntax.getMvPredIdx( uiIdx ) = uiVal;
  TRACE_NUMBER;
  TRACE_TO_STR( "mv_predictor_idc: " );
  TRACE_TO_STR( rcPredSyntax.getMvPredIdx( uiIdx ) );
  TRACE_TO_STR( std::endl );
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessWienerParameters( WienerFilterParameters &rcWFParam , UInt uiNumOfPlanes )
{
  {
    bool bSeparableFlag = false;
    BaseCall::xProcessWienerSeparableFlag( bSeparableFlag );
    TOF( bSeparableFlag );
  }
    TRACE_NUMBER;
    TRACE_TO_STR( "quadtree_based_inloop_filter_separable_flag: " );
    TRACE_TO_STR( true );
    TRACE_TO_STR( std::endl );


    bool bUseQuadtreeFlag = false;
    BaseCall::xProcessWienerFlagForQuadtreeUse( bUseQuadtreeFlag );
    rcWFParam.setUseBlocksFlag( bUseQuadtreeFlag );
    TRACE_NUMBER;
    TRACE_TO_STR( "quadtree_based_inloop_filter_quadtree_used_flag: " );
    TRACE_TO_STR( rcWFParam.getUseBlocksFlag() );
    TRACE_TO_STR( std::endl   );

    if( bUseQuadtreeFlag )
    {
      UInt uiMaxBlock = 0;
      UInt uiMinBlock = 0;
      BaseCall::xProcessWienerBlockSize( uiMaxBlock , uiMinBlock );
      rcWFParam.setPower2MaxBlock( uiMaxBlock );
      rcWFParam.setPower2MinBlock( uiMinBlock );
      TRACE_NUMBER;
      TRACE_TO_STR( "quadtree_based_inloop_filter_log2_blocksize_max: " );
      TRACE_TO_STR(  uiMaxBlock );
      TRACE_TO_STR( std::endl   );
      TRACE_NUMBER;
      TRACE_TO_STR( "quadtree_based_inloop_filter_log2_blocksize_min: " );
      TRACE_TO_STR(  uiMinBlock );
      TRACE_TO_STR( std::endl   );
    }

    UInt uiNumOfSets = 0;
    UInt uiSetOfPlane = 0;
    for ( UInt uiPlane=0; uiPlane < uiNumOfPlanes ; uiPlane++)
    {
      BaseCall::xProcessWienerPlaneSet( uiSetOfPlane );
      TRACE_TO_STR( "quadtree_based_inloop_filter_allowed_modes_idx[ " );
      TRACE_TO_STR( uiPlane ) ;
      TRACE_TO_STR( " ]: " ) ;
      TRACE_TO_STR( uiSetOfPlane );
      TRACE_TO_STR( std::endl );

      if( uiSetOfPlane > 0 )
      {
        rcWFParam.setFilterPlaneFlag(uiPlane, true);
        rcWFParam.setFilterSetForPlane( uiPlane , uiSetOfPlane - 1 );
        rcWFParam.setFilterPlaneFlag(uiPlane, true);
      }
      else
      {
        rcWFParam.setFilterPlaneFlag(uiPlane, false);
      }
      uiNumOfSets = std::max( uiNumOfSets, uiSetOfPlane );
    }
    for ( UInt uiSetNumber=0; uiSetNumber < uiNumOfSets; uiSetNumber++)
    {
      UInt uiHorizontalFilterLength = 0;
      UInt uiVerticalFilterLength = 0;
      UInt uiFilterType = 0;

      xProcessWienerFilterLength( uiHorizontalFilterLength );
      TRACE_TO_STR( "filter_length_minus1_div2[ " );
      TRACE_TO_STR( uiSetNumber ) ;
      TRACE_TO_STR( " ][ 0 ]: " ) ;
      TRACE_TO_STR( ( uiHorizontalFilterLength - 1 ) / 2 );
      TRACE_TO_STR( std::endl );

      xProcessWienerFilterLength( uiVerticalFilterLength );
      TRACE_TO_STR( "filter_length_minus1_div2[ " );
      TRACE_TO_STR( uiSetNumber ) ;
      TRACE_TO_STR( " ][ 1 ]: " ) ;
      TRACE_TO_STR( ( uiVerticalFilterLength - 1 ) / 2 );
      TRACE_TO_STR( std::endl );

      BaseCall::xProcessWienerFilterType( uiFilterType );
      TRACE_TO_STR( "quadtree_based_inloop_filter_type[ " );
      TRACE_TO_STR( uiSetNumber ) ;
      TRACE_TO_STR( " ]: " ) ;
      TRACE_TO_STR( uiFilterType );
      TRACE_TO_STR( std::endl );

      rcWFParam.getHorizontalFilterCoeff( uiSetNumber ).setFilterTypeSep( uiFilterType , uiHorizontalFilterLength);
      rcWFParam.getHorizontalFilterCoeff( uiSetNumber ).setHorizontal();
      rcWFParam.getVerticalFilterCoeff( uiSetNumber ).setFilterTypeSep( uiFilterType , uiVerticalFilterLength);
      rcWFParam.getVerticalFilterCoeff( uiSetNumber ).setVertical();
      xProcessWienerFilterCoeffs( rcWFParam.getHorizontalFilterCoeff( uiSetNumber ));
      xProcessWienerFilterCoeffs( rcWFParam.getVerticalFilterCoeff( uiSetNumber ));
    }

  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessWienerFilterCoeffs( FilterCoefficients& rcCoeffs )
{
  for (UInt i = 0; i< rcCoeffs.getNumberOfCoeffs(); i++)
  {
    bool    thisSign ;
    UInt    thisAbsCoeff = 0;
    int     thisCoeff ;
    RNOK( BaseCall::xProcessBin( thisSign, BaseCall::m_cCtxAccess.getContextOffset( CTX_WIENER_FILTER_COEFFS_SIGN ) ) );
    TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_sign[ i ][ hv ][ " );
    TRACE_TO_STR( i ) ;
    TRACE_TO_STR( " ]: " ) ;
    TRACE_TO_STR( thisSign ? 1 : 0 );
    TRACE_TO_STR( std::endl );
    xReadEpExGolomb( thisAbsCoeff, 14, BaseCall::m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_ABS_COEFFS) );
    TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_abs[ i ][ hv ][ " );
    TRACE_TO_STR( i ) ;
    TRACE_TO_STR( " ]: " ) ;
    TRACE_TO_STR( thisAbsCoeff );
    TRACE_TO_STR( std::endl );
    thisCoeff = thisSign ? (-1 * thisAbsCoeff) : thisAbsCoeff ;
    rcCoeffs.pushNewQuantCoeffPredicted( i , thisCoeff );
  }

  bool bDcSign;
  UInt uiDcAbs = 0;
  int iDC;
  RNOK( BaseCall::xProcessBin( bDcSign, BaseCall::m_cCtxAccess.getContextOffset( CTX_WIENER_FILTER_COEFFS_SIGN ) ) );
  TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_dc_sign[ i ][ hv ]: " );
  TRACE_TO_STR( bDcSign ? 1 : 0 );
  TRACE_TO_STR( std::endl );
  xReadEpExGolomb( uiDcAbs, 14, BaseCall::m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_ABS_COEFFS) );
  TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_dc_abs[ i ][ hv ]: " );
  TRACE_TO_STR( uiDcAbs );
  TRACE_TO_STR( std::endl );
  iDC = bDcSign ? (-1 * uiDcAbs) : uiDcAbs ;
  rcCoeffs.pushBackQuantDcOffset( iDC );

  return Err::m_nOK ;
}


template<class t_BinProcessor>
ErrVal SyntaxElementDecoder<t_BinProcessor>::xProcessWienerFilterLength( UInt &ruiFilterLength)
{
    UInt uiMaxIdxOfFilterLength= 7;
    UInt iFilterLenghtIdx = 0;
    xProcessUnarySymbol( iFilterLenghtIdx, BaseCall::m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_LENGTH), 1, uiMaxIdxOfFilterLength );

    ruiFilterLength = iFilterLenghtIdx*2 + 1 ;

    return Err::m_nOK ;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _SYNTAXELEMENTDECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
