/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_UVLCREADER_H__EA98D347_89D5_4D2D_B6D5_FB3A374CD295__INCLUDED_)
#define AFX_UVLCREADER_H__EA98D347_89D5_4D2D_B6D5_FB3A374CD295__INCLUDED_


#include "VideoCommonLib/HeaderSymbolReadIf.h"
#include "VideoCommonLib/TraceCommon.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

template<class t_UnsignedReadType> class BitReaderIf;


//!\ingroup VideoDecoderLib
//!UvlcReader
class UvlcReader
: public HeaderSymbolReadIf
, public TraceBase
{
public:
  UvlcReader( BitReaderIf<UInt> &rcBitReaderIf ) : m_pcBitReaderIf( &rcBitReaderIf ) {}
  virtual ~UvlcReader() {}

  ErrVal  getUvlc( UInt& ruiCode, const char* pcTraceString );
  ErrVal  getUvlc64( UInt64& ruiCode, const char* pcTraceString );
  ErrVal  getCode( UInt& ruiCode, UInt uiLength, const char* pcTraceString );
  ErrVal  getSCode( int& riCode, UInt uiLength, const char* pcTraceString );
  ErrVal  getSvlc( int& riCode, const char* pcTraceString );
  ErrVal  getFlag( bool& rbFlag, const char* pcTraceString );
  ErrVal  readByteAlign();
  ErrVal  readByteAlignOnes();
  bool    moreRBSPData();

private:
  __inline ErrVal xGetFlag( UInt& ruiCode );
  __inline ErrVal xGetCode( UInt& ruiCode, UInt uiLength );
  ErrVal xGetUvlcCode( UInt& ruiVal);
  ErrVal xGetUvlc64Code( UInt64& ruiVal);
  ErrVal xGetSvlcCode( int& riVal);

  template<class T>
  void xTraceString( const char *pcTraceString, const T &r )
  {
    TRACE_NUMBER;
    TRACE_TO_STR( " " );
    TRACE_TO_STR( pcTraceString );
    TRACE_TO_STR( " " );
    TRACE_TO_STR( r );
    TRACE_TO_STR( std::endl );
  }

protected:
  BitReaderIf<UInt>* m_pcBitReaderIf;
};

// video namespace begin
VIDEO_NAMESPACE_END

#endif // !defined(AFX_UVLCREADER_H__EA98D347_89D5_4D2D_B6D5_FB3A374CD295__INCLUDED_)
