/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _VLC_STATE_DECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _VLC_STATE_DECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_


#include "VideoCommonLib/VLCodingTables.h"
#include "BitReader.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_ValType>
class VLCStateDecoder : private VLCodingTable<t_ValType>
{
public:
  VLCStateDecoder( const UInt uiDecoderId, BitReaderIf<UInt>& rcBitReaderIf, const UInt uiMaxDelayInBits, UInt& ruiMaxDelayBitCounter );
  virtual ~VLCStateDecoder()  {}

  ErrVal  getNextBin( bool& rbBin );

private:
  ErrVal  xReadNextCodeword();

private:
  const UInt              m_uiDecoderId;
  const UInt              m_uiMaxDelayInBits;
  UInt                    m_uiMaxBitCounterValue;
  UInt&                   m_ruiMaxDelayBitCounter;
  BitReaderIf<UInt>&      m_rcBitReaderIf;
  VLCodeword<t_ValType>   m_cCodeword;
};


template<class t_ValType>
VLCStateDecoder<t_ValType>::VLCStateDecoder( const UInt uiDecoderId, BitReaderIf<UInt>& rcBitReaderIf, const UInt uiMaxDelayInBits, UInt& ruiMaxDelayBitCounter )
: VLCodingTable<t_ValType>  ( VLCodingTables<t_ValType,false>::getVLCodingTable( uiDecoderId ) )
, m_uiDecoderId             ( uiDecoderId   )
, m_uiMaxDelayInBits        ( uiMaxDelayInBits ? uiMaxDelayInBits : TypeLimits<UInt>::getMax() >> 1 )
, m_uiMaxBitCounterValue    ( m_uiMaxDelayInBits )
, m_ruiMaxDelayBitCounter   ( ruiMaxDelayBitCounter )
, m_rcBitReaderIf           ( rcBitReaderIf )
, m_cCodeword               ( VLCodeword<t_ValType>() )
{
}


template<class t_ValType>
ErrVal
VLCStateDecoder<t_ValType>::getNextBin( bool& rbBin )
{
  if( m_cCodeword.getLength() == 0 || m_ruiMaxDelayBitCounter > m_uiMaxBitCounterValue )
  {
    m_uiMaxBitCounterValue    = m_ruiMaxDelayBitCounter + m_uiMaxDelayInBits;
    m_ruiMaxDelayBitCounter  += VLCodingTable<t_ValType>::getMaxCodewordLength();
    RNOK( xReadNextCodeword() );
  }
  rbBin = bool( m_cCodeword.popFront() );
  return Err::m_nOK;
}


template<class t_ValType>
ErrVal
VLCStateDecoder<t_ValType>::xReadNextCodeword()
{
  UInt  uiBit   = 0;
  UInt  uiState = 0;
  do 
  {
    RNOK( m_rcBitReaderIf.read( uiBit ) );
    m_cCodeword = VLCodingTable<t_ValType>::updateState( uiState, StateTransition( uiBit ) );
  } while( m_cCodeword.getLength() == 0 );
  return Err::m_nOK;
}



// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _VLC_STATE_DECODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
