/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoDecoderLibPch.h"
#include "VideoDecoderImpl.h"
#include "VideoCommonLib/TraceCabac.h"
#include "BitReader.h"
#include "../VideoEncoderLib/AccessUnitEncoder.h"
#include "VideoCommonLib/BitWriter.h"
#include "VideoCommonLib/H264SequenceParameterSet.h"
#include "VideoCommonLib/H264PictureParameterSet.h"
#include "VideoCommonLib/H264SliceHeader.h"
#include "H264NalUnitParser.h"
#include "UvlcReader.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


VideoDecoderImpl::VideoDecoderImpl()
: m_pcCodedPictureStore( new CodedPictureStore )
, m_pcDecToDispBuffer( new DecodingToDisplayOrderBuffer( 0 ) )
, m_pcSequenceParameterSet(NULL)
, m_pcPictureParameterSet(NULL)
{
}


VideoDecoderImpl::~VideoDecoderImpl()
{
  delete m_pcCodedPictureStore;
  delete m_pcDecToDispBuffer;
  delete m_pcSequenceParameterSet;
  delete m_pcPictureParameterSet;

  for (UInt ui = 0; ui < m_apcOverlaidDecToDispBuffer.size(); ui++ )
    delete m_apcOverlaidDecToDispBuffer[ui];
  m_apcOverlaidDecToDispBuffer.clear();
#if COLLECT_CABAC_BITS
  for( UInt ui = 0; ui < UInt( m_apcTraceCabac.size() ); ui++ )
  {
    if( m_apcTraceCabac[ui] != NULL )
    {
#if TRACK_CABAC_BITS
      std::stringstream acName;
      acName << "CabacTraceFile" << ui << ".txt";
      ANOK( m_apcTraceCabac[ui]->traceStoredCabacBits( acName.str() ) );
#endif
      delete m_apcTraceCabac[ui];
    }
  }
#endif
}


ErrVal VideoDecoderImpl::runAlternative( int argc, char **argv, bool &rbExit )
{
  rbExit = false;
  return Err::m_nOK;
}


ErrVal VideoDecoderImpl::create( VideoDecoderInstance *&rpcVideoDecoderInstance )
{
  try
  {
    rpcVideoDecoderInstance = new VideoDecoderImpl;
  }
  RETURN_ERR_ON_EXCEPTION;
  return Err::m_nOK;
}

ErrVal VideoDecoderImpl::prepareForNewStream()
{
  delete m_pcCodedPictureStore;
  delete m_pcDecToDispBuffer;
  delete m_pcSequenceParameterSet;
  delete m_pcPictureParameterSet;

  for (UInt ui = 0; ui < m_apcOverlaidDecToDispBuffer.size(); ui++ )
    delete m_apcOverlaidDecToDispBuffer[ui];

  m_pcCodedPictureStore = new CodedPictureStore;
  m_pcDecToDispBuffer = new DecodingToDisplayOrderBuffer( 0 );
  m_apcOverlaidDecToDispBuffer.clear();

  return Err::m_nOK;
}

ErrVal VideoDecoderImpl::createOverlaidDecToDispBuffer()
{
  m_apcOverlaidDecToDispBuffer.push_back( new DecodingToDisplayOrderBuffer( 0 ) );
  return Err::m_nOK;
}

TraceCabac& VideoDecoderImpl::getTraceCabac( UInt ui )
{
  while( m_apcTraceCabac.size() <= ui )
  {
    m_apcTraceCabac.push_back( NULL );
  }
  if( NULL == m_apcTraceCabac[ui] )
  {
    m_apcTraceCabac[ui] = new TraceCabac;
  }
  return *m_apcTraceCabac[ui];
}


ErrVal VideoDecoderImpl::destroy()
{
  try
  {
    delete this;
  }
  RETURN_ERR_ON_EXCEPTION;
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
