/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _VIDEODECODERIMPL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _VIDEODECODERIMPL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include <fstream>

#include "VideoCommonLib/DecodingToDisplayOrderBuffer.h"
#include "VideoCommonLib/ModePictureGenerator.h"
#include "VideoCommonLib/PocAndFrameNum.h"
#include "VideoDecoderLib/VideoDecoderIf.h"
#include "VideoDecoderLib/VideoDecoderInstance.h"
#include "AccessUnitDecoder.h"
#include "VideoCommonLib/TraceCommon.h"

#include "H264NalUnitParser.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class TraceCabac;

class VideoDecoderImpl
  : private VideoDecoderIf, VideoDecoderInstance, TraceCabacVecIf
{
protected:
  VideoDecoderImpl();
  virtual ~VideoDecoderImpl();

public:
  static ErrVal runAlternative( int argc, char **argv, bool &rbExit );

  static ErrVal create( VideoDecoderInstance *&rpcVideoDecoderInstance );
  ErrVal destroy();

  VideoDecoderIf* getVideoDecoderIf() { return this; }
  
  ErrVal getPictureStats( const std::vector<UChar> &rBitstreamPacket,
                          VideoCodecPicture<VideoCodecPel> &rcStats,
                          UInt &ruiPlanePackages,
                          UInt64 *puiPoc )
  {
    AccessUnitParameters cPars;

    bool bSizeAvailable = false;
    cPars.setPocAndFrameNum( &m_cPocAndFrameNum );
    RNOK( AccessUnitDecoder::decodeNalUnit( cPars, NULL, rBitstreamPacket, bSizeAvailable, NULL, m_pcSequenceParameterSet, m_pcPictureParameterSet ) );
    if( false == bSizeAvailable )
    {
      return Err::m_nSizeNotAvailable;
    }

    rcStats.m_acPlanes.clear();
    ruiPlanePackages = cPars.getNumPlanePacks();
    rcStats.m_acPlanes.reserve( cPars.getNumPlanes() );
    for( UInt ui = 0; ui < cPars.getNumPlanes(); ui++ )
    {
      const BlockSize cPlaneSize = cPars.getPlaneSize( ui );
      rcStats.m_acPlanes.push_back( VideoCodecPlane<VideoCodecPel>( NULL,cPlaneSize.getX(), cPlaneSize.getY(), 0 ) );
    }
    if( puiPoc )
    {
      *puiPoc = cPars.getPicOrderCnt();
    }
    return Err::m_nOK;
  }

  ErrVal prepareForNewStream();
  ErrVal createOverlaidDecToDispBuffer();

  ErrVal decodePicture( const std::vector<UChar> &rcBitstreamPacket,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedPic,
                        VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> &rcReconstructedPicSinkIf,
                        std::vector<VideoCodecPicture<VideoCodecPel>*> apcOverlayPic,
                        std::vector<VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*>*> apcOverlayPicSinkIf )
  {
    RNOK( xDecodePicture( rcBitstreamPacket, pcReconstructedPic, rcReconstructedPicSinkIf, apcOverlayPic, apcOverlayPicSinkIf ) );
    return Err::m_nOK;
  }

  TraceCabac& getTraceCabac( UInt ui );

private:
  template<class t_Pel>
  ErrVal xDecodePicture( const std::vector<UChar> &rcBitstreamPacket,
                         const VideoCodecPicture<t_Pel> *pcReconstructedPic,
                         VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                         std::vector<VideoCodecPicture<VideoCodecPel>*> apcOverlayPic,
                         std::vector<VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*>*> apcOverlayPicSinkIf );

  CodedPictureStore *m_pcCodedPictureStore;
  DecodingToDisplayOrderBuffer *m_pcDecToDispBuffer;
  std::vector<DecodingToDisplayOrderBuffer*> m_apcOverlaidDecToDispBuffer;
  std::vector<TraceCabac*> m_apcTraceCabac;
  SequenceParameterSet* m_pcSequenceParameterSet;
  PictureParameterSet* m_pcPictureParameterSet;
  PocAndFrameNum m_cPocAndFrameNum;
};

template<class t_Pel>
ErrVal VideoDecoderImpl::xDecodePicture( const std::vector<UChar> &rcBitstreamPacket,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedPic,
                                         VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                                         std::vector<VideoCodecPicture<VideoCodecPel>*> apcOverlayPic,
                                         std::vector<VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*>*> apcOverlayPicSinkIf )
{
  ROF( ( apcOverlayPic.empty() ) == ( apcOverlayPicSinkIf.empty() ) );
  CAOF( (TypeCompare<t_Pel, Pel>::m_bEqual) );
  PicOrderCnt cPoc = 0;
  
  AccessUnit *pcAU = new AccessUnit;

  TraceCommon *pcTrace = NULL;
#if TRACE_DEC_ENABLED
  std::stringstream cHeaderTrace;
  TraceDec cTraceDec( cHeaderTrace );
  pcTrace = &cTraceDec;
#endif

  bool bSizeInfoAvailable = false;
  std::vector<UChar> cRawData;
  pcAU->getAUPars().setPocAndFrameNum( &m_cPocAndFrameNum );
  RNOK( AccessUnitDecoder::decodeNalUnit( pcAU->getAUPars(), &cRawData, rcBitstreamPacket, bSizeInfoAvailable, pcTrace, m_pcSequenceParameterSet, m_pcPictureParameterSet ));

  // in case of SPS or PPS
  if( false == bSizeInfoAvailable )
  {
    delete pcAU;
    return Err::m_nOK;
  }

  if( !m_pcCodedPictureStore->isInitialized() )
  {
    RNOK( m_pcCodedPictureStore->initCodedPictureStore( pcAU->getAUPars().getMaxDPBSize() ) );
  }

  std::cout << "Decoding frame with POC = " << std::flush;


#if TRACE_DEC_ENABLED
  std::fstream cTraceFile;
  {
    std::stringstream acFilename;
    acFilename << "TraceDec_POC" << pcAU->getAUPars().getPicOrderCnt() << ".txt";
    cTraceFile.open( acFilename.str().c_str(), std::ios::out | std::ios::binary | std::ios::trunc );
    ROT( cTraceFile.fail() );
    std::string cStr = cHeaderTrace.str();
    cTraceFile << cHeaderTrace.str();
    cTraceDec.setOStream( cTraceFile );
  }
#endif


  RNOK( AccessUnitDecoder::parseAccessUnit( cRawData, *m_pcCodedPictureStore, pcAU, this, cPoc, pcTrace ) );

#if TRACE_DEC_ENABLED
  cTraceFile.close();
#endif

  std::vector<PicBlock<t_Pel> > acOverlayPicBlock;
  const PicBlock<Pel> cDec( pcReconstructedPic->m_acPlanes );
  if( !apcOverlayPic.empty() )
  {
    for ( UInt uiPlaPaId = 0; uiPlaPaId < apcOverlayPic.size(); uiPlaPaId++ )
    {
      acOverlayPicBlock.push_back( PicBlock<t_Pel>( apcOverlayPic[uiPlaPaId]->m_acPlanes ) );
    }
    RNOK( AccessUnitDecoder::decodeAccessUnit( cDec, acOverlayPicBlock, *m_pcCodedPictureStore, pcAU ) );
    for ( UInt uiPlaPaId = 0; uiPlaPaId < apcOverlayPic.size(); uiPlaPaId++ )
    {
      RNOK( m_apcOverlaidDecToDispBuffer[uiPlaPaId]->insertItem( cPoc, apcOverlayPic[uiPlaPaId] ) );
      RNOK( m_apcOverlaidDecToDispBuffer[uiPlaPaId]->outputItemsIfPossible( *apcOverlayPicSinkIf[uiPlaPaId] ) );
    }
  }
  else
  {
    RNOK( AccessUnitDecoder::decodeAccessUnit( cDec, acOverlayPicBlock, *m_pcCodedPictureStore, pcAU ) );
  }

  std::cout << pcAU->getAUPars().getPicOrderCnt() << std::endl;


  if( pcAU->getAUPars().getRefPicFlag() )
  {
    UInt uiShift;
    UInt uiMargin;
    ReferencePicture<Pel>::UpsMethod cUpsMethod;

    pcAU->getAUPars().getInterpolationParameters( uiShift, uiMargin, cUpsMethod, true );

    ReferencePicture<Pel>* pcRefPicture = new ReferencePicture<Pel>( pcAU->getAUPars().getPicOrderCnt(), pcAU->getAUPars().getAUStats(), uiShift, uiMargin, cUpsMethod );
    for( UInt uiRuns = 0; uiRuns < pcAU->getAUPars().getNumPlanePacks(); uiRuns++ )
    {
      PicBlock<Pel> cDecPlanePackage;
      RNOK( pcAU->extractPlanePackagePicBlock( uiRuns, cDec, cDecPlanePackage ) );
      RNOK( pcRefPicture->assignPlanePackage( pcAU->getPlanePackage( uiRuns )->getPars().getPrimPlaneId(), cDecPlanePackage, cUpsMethod ) );
    }

    RNOK( pcAU->setRefPicture( pcRefPicture ) );
    RNOK( m_pcCodedPictureStore->addAUAndResizeStore( pcAU ) );
  }
  else
  {
    delete pcAU;
  }

  RNOK( m_pcDecToDispBuffer->insertItem( cPoc, pcReconstructedPic ) );
  RNOK( m_pcDecToDispBuffer->outputItemsIfPossible( rcReconstructedPicSinkIf ) );
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _VIDEODECODERIMPL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
