/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include "VideoEncoderLibPch.h"
#include "AUCodingStarter.h"
#include "ReferencePictureEnc.h"

#include <fstream>

// video namespace begin
VIDEO_NAMESPACE_BEGIN


ErrVal AUCodingStarter::xEncodeItem( const CodedPictureStoreBase &rcCodedPictureStore, AccessUnitEncodingItem *pcAUCodingItem )
{
  std::vector<const CodingState*> apcCodStat;
  for( UInt uiPlap = 0; uiPlap < pcAUCodingItem->getAUParameters().getNumPlanePacks(); uiPlap++ )
  {
    apcCodStat.push_back( new CodingState( rcCodedPictureStore.getPlanePackageVec<ReferencePlanePackageEnc<Pel> >( uiPlap, pcAUCodingItem->getAUParameters() ), rcCodedPictureStore.getRefCtxStore( uiPlap, pcAUCodingItem->getAUParameters() ) ) );
  }
  RNOK( pcAUCodingItem->prepareCodingStates( apcCodStat ) );

#if ALLOW_BOOST_LIBRARY
  if( m_cAUEncFIFO.getNumThreads() > 1 )
  {
    m_cAUEncFIFO.getInput().pushBack( pcAUCodingItem );
    m_iItemsInFIFO++;
  }
  else
#endif
  {
    RNOK( pcAUCodingItem->encodeAU() );
    m_apcEncodedItems.push_back( pcAUCodingItem );
  }

  return Err::m_nOK;
}


ErrVal AUCodingStarter::checkForCodedItems( std::vector<AccessUnitEncodingItem*> &rapcCodedItems, bool bWaitForOne )
{
#if ALLOW_BOOST_LIBRARY
  if( m_cAUEncFIFO.getNumThreads() > 1 )
  {
    bool bWait = bWaitForOne;
    while( true )
    {
      bool bSuccess = true;
      AccessUnitEncodingItem *pcCodedItem = m_cAUEncFIFO.getOutput().tryPopFront( bSuccess, bWait && m_iItemsInFIFO > 0 );
      bWait = false;
      if( bSuccess )
      {
        m_iItemsInFIFO--;
        rapcCodedItems.push_back( pcCodedItem );
      }
      else
      {
        break;
      }
    }
  }
  else
#endif
  {
    rapcCodedItems.insert( rapcCodedItems.end(), m_apcEncodedItems.begin(), m_apcEncodedItems.end() );
    m_apcEncodedItems.clear();
  }

  return Err::m_nOK;
}

AUCodingStarter::~AUCodingStarter()
{
#if ALLOW_BOOST_LIBRARY
  TNOK( m_cAUEncFIFO.joinAllThreads() );
#endif
}


ErrVal AUCodingStarter::pushBackAUCodingItem( AccessUnitEncodingItem *pcAUCodingItem )
{
  m_apcNextPocs.push_back( pcAUCodingItem );
  return Err::m_nOK;
}


ErrVal AUCodingStarter::startCodableItems( const CodedPictureStoreBase &rcCodedPictureStore )
{
  std::multimap<int, AccessUnitEncodingItem*> apcCodableItems;
  for( std::list<AccessUnitEncodingItem*>::iterator cListIter = m_apcNextPocs.begin(); cListIter != m_apcNextPocs.end(); )
  {
    if( rcCodedPictureStore.isCodable( (*cListIter)->getAUParameters().getPicOrderCnt(), (*cListIter)->getAUParameters().getRefDeltaPocList() ) )
    {
      apcCodableItems.insert( std::make_pair( (*cListIter)->getCodingPriorityIndex(), *cListIter ) );
      m_apcNextPocs.erase( cListIter++ );
    }
    else
    {
      ++cListIter;
    }
  }
  std::multimap<int, AccessUnitEncodingItem*>::reverse_iterator cIter = apcCodableItems.rbegin();
  for( ; cIter != apcCodableItems.rend(); ++cIter )
  {
    RNOK( xEncodeItem( rcCodedPictureStore, cIter->second ) );
  }
  return Err::m_nOK;
}


AccessUnitEncodingItem::AccessUnitEncodingItem( const OriginalAndReconstructedPic &rcInputPics,
                                                const PicBlock<const Pel> &rcOrig,
                                                const PicBlock<Pel> &rcEnc,
                                                const PicBlock<Pel> *pcEncPre,
                                                const PicBlock<Pel> *pcEncRes,
                                                PicOrderCnt iLastKeyPicPoc0,
                                                PicOrderCnt iLastKeyPicPoc1,
                                                int iCodingPriorityIndex,
                                                const std::vector<PicOrderCnt> &raiPocsToRemoveFromStoreAfterCurrentPoc,
                                                const AccessUnitEncoderParameters &rcAUEstPars,
                                                const AccessUnitParameters &rcAUPars,
                                                AccessUnit *pcAccessUnit,
                                                const std::string &rcOut,
                                                bool bParsed,
#if MULTICABAC_TRANSCODING_HACK
                                                const UInt uiParsedPacketSize,
#endif
                                                Int64 iCodingOrder )
: m_cInputPics( rcInputPics )
, m_cOrig( rcOrig )
, m_cEnc( rcEnc )
, m_pcEncPre( pcEncPre )
, m_pcEncRes( pcEncRes )
, m_iLastKeyPicPoc0( iLastKeyPicPoc0 )
, m_iLastKeyPicPoc1( iLastKeyPicPoc1 )
, m_iCodingPriorityIndex( iCodingPriorityIndex )
, m_aiPocsToRemoveFromStoreAfterCurrentPoc( raiPocsToRemoveFromStoreAfterCurrentPoc )
, m_cAUEstPars( rcAUEstPars )
, m_cAUPars( rcAUPars )
, m_pcAU( pcAccessUnit )
, m_paaucBitstreamData( new std::list<std::vector<UChar> >() )
, m_bParsed( bParsed )
#if MULTICABAC_TRANSCODING_HACK
, m_uiParsedPacketSize( uiParsedPacketSize )
#endif
, m_iCodingOrder( iCodingOrder )
, m_uiEncodingDurationMSec( 0 )
, m_uiMEDurationMSec( 0 )
{
  m_cOut << rcOut;
}


ErrVal AccessUnitEncodingItem::prepareCodingStates( const std::vector<const CodingState*> &rapcCodingStates )
{
  ROF( m_acCodStat.empty() );
  m_acCodStat.insert( m_acCodStat.end(), rapcCodingStates.begin(), rapcCodingStates.end() );
  return Err::m_nOK;
}

ErrVal AccessUnitEncodingItem::encodeAU()
{
  TraceEnc *pcTraceEnc = NULL;
#if TRACE_ENC_ENABLED
  std::stringstream acFilename;
  acFilename << "TraceEnc_POC" << m_cAUPars.getPicOrderCnt() << ".txt";
  std::fstream cTraceFile;
  cTraceFile.open( acFilename.str().c_str(), std::ios::out | std::ios::binary | std::ios::trunc );
  ROT( cTraceFile.fail() );
  TraceEnc cTraceEnc( cTraceFile );
  pcTraceEnc = &cTraceEnc;
#endif

#if MULTICABAC_TRANSCODING_HACK
  RNOK( AccessUnitEncoder::encodeAccessUnit( m_cOrig,
                                             m_cEnc,
                                             m_pcEncPre,
                                             m_pcEncRes,
                                             m_cAUPars,
                                             m_cAUEstPars,
                                             m_acCodStat,
                                             m_iLastKeyPicPoc0,
                                             m_iLastKeyPicPoc1,
                                             m_pcAU,
                                             m_bParsed,
                                             m_uiParsedPacketSize,
                                             *m_paaucBitstreamData,
                                             m_adPSNR,
                                             m_cOut,
                                             m_uiEncodingDurationMSec,
                                             m_uiMEDurationMSec,
                                             pcTraceEnc ) );
#else
  RNOK( AccessUnitEncoder::encodeAccessUnit( m_cOrig,
                                             m_cEnc,
                                             m_pcEncPre,
                                             m_pcEncRes,
                                             m_cAUPars,
                                             m_cAUEstPars,
                                             m_acCodStat,
                                             m_iLastKeyPicPoc0,
                                             m_iLastKeyPicPoc1,
                                             m_pcAU,
                                             m_bParsed,
                                             *m_paaucBitstreamData,
                                             m_adPSNR,
                                             m_cOut,
                                             m_uiEncodingDurationMSec,
                                             m_uiMEDurationMSec,
                                             pcTraceEnc ) );
#endif

#if TRACE_ENC_ENABLED
  cTraceFile.close();
#endif

  delete m_pcEncPre;
  delete m_pcEncRes;
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END
