/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _AUCODINGSTARTER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _AUCODINGSTARTER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCodecToolsLib/VideoCodecPicture.h"
#include "AccessUnitEncoder.h"
#include "VideoCommonLib/MultiThreading.h"
#include "TraceEnc.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


struct OriginalAndReconstructedPic
{
  OriginalAndReconstructedPic( const VideoCodecPicture<const VideoCodecPel> *pcOriginalPic,
                               const VideoCodecPicture<VideoCodecPel> *pcReconstructedPic,
                               const VideoCodecPicture<VideoCodecPel> *pcReconstructedPrePic,
                               const VideoCodecPicture<VideoCodecPel> *pcReconstructedResPic )
    : m_pcOriginalPic( pcOriginalPic )
    , m_pcReconstructedPic( pcReconstructedPic )
    , m_pcReconstructedPrePic( pcReconstructedPrePic )
    , m_pcReconstructedResPic( pcReconstructedResPic ) { }

  OriginalAndReconstructedPic() : m_pcOriginalPic( NULL ), m_pcReconstructedPic( NULL ), m_pcReconstructedPrePic( NULL ), m_pcReconstructedResPic( NULL ) {}

  const VideoCodecPicture<const VideoCodecPel> *m_pcOriginalPic;
  const VideoCodecPicture<VideoCodecPel> *m_pcReconstructedPic;
  const VideoCodecPicture<VideoCodecPel> *m_pcReconstructedPrePic;
  const VideoCodecPicture<VideoCodecPel> *m_pcReconstructedResPic;
};


class AccessUnitEncodingItem
{
public:
  AccessUnitEncodingItem( const OriginalAndReconstructedPic &rcInputPics,
                          const PicBlock<const Pel> &rcOrig,
                          const PicBlock<Pel> &rcEnc,
                          const PicBlock<Pel> *pcEncPre,
                          const PicBlock<Pel> *pcEncRes,
                          PicOrderCnt iLastKeyPicPoc0,
                          PicOrderCnt iLastKeyPicPoc1,
                          int iCodingPriorityIndex,
                          const std::vector<PicOrderCnt> &raiPocsToRemoveFromStoreAfterCurrentPoc,
                          const AccessUnitEncoderParameters &rcAUEstPars,
                          const AccessUnitParameters &rcAUPars,
                          AccessUnit *pcAccessUnit,
                          const std::string &rcOut,
                          bool bParsed,
#if MULTICABAC_TRANSCODING_HACK
                          const UInt uiParsedPacketSize,
#endif
                          Int64 iCodingOrder );

  ErrVal prepareCodingStates( const std::vector<const CodingState*> &rapcCodingStates );

  ErrVal encodeAU();

  AccessUnit*                           getAccessUnit()                           const { return m_pcAU; }
  const AccessUnitParameters&           getAUParameters()                         const { return m_cAUPars; }
  const OriginalAndReconstructedPic&    getInputPics()                            const { return m_cInputPics; }
  Int64                                 getCodingOrder()                          const { return m_iCodingOrder; }
  std::list<std::vector<UChar> >*       getBitstreamData()                              { return m_paaucBitstreamData; }
  const std::vector<double>&            getAUPSNR()                               const { return m_adPSNR; }
  std::string                           getOutput()                               const { return m_cOut.str(); }
  const std::vector<PicOrderCnt>&       getPocsToRemoveFromStoreAfterCurrentPoc() const { return m_aiPocsToRemoveFromStoreAfterCurrentPoc; }
  int                                   getCodingPriorityIndex()                  const { return m_iCodingPriorityIndex; }
  UInt64                                getEncodingDurationMSec()                 const { return m_uiEncodingDurationMSec; }
  UInt64                                getMEDurationMSec()                       const { return m_uiMEDurationMSec; }

  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( AccessUnitEncodingItem );

private:
  const OriginalAndReconstructedPic m_cInputPics;
  const PicBlock<const Pel> m_cOrig;
  const PicBlock<Pel> m_cEnc;
  const PicBlock<Pel> *m_pcEncPre;
  const PicBlock<Pel> *m_pcEncRes;
  AutoDeletePtrVector<const CodingState> m_acCodStat;
  const PicOrderCnt m_iLastKeyPicPoc0;
  const PicOrderCnt m_iLastKeyPicPoc1;
  const int m_iCodingPriorityIndex;
  const std::vector<PicOrderCnt> m_aiPocsToRemoveFromStoreAfterCurrentPoc;
  const AccessUnitEncoderParameters m_cAUEstPars;
  const AccessUnitParameters m_cAUPars;
  AccessUnit *m_pcAU;
  std::list<std::vector<UChar> > *m_paaucBitstreamData;
  const bool m_bParsed;
#if MULTICABAC_TRANSCODING_HACK
  const UInt m_uiParsedPacketSize;
#endif
  const Int64 m_iCodingOrder;
  std::stringstream m_cOut;
  std::vector<double> m_adPSNR;
  UInt64 m_uiEncodingDurationMSec;
  UInt64 m_uiMEDurationMSec;
};


class AUEncodingWorker
{
public:
  // necessary typedefs
  typedef AccessUnitEncodingItem* WorkerInput;
  typedef AccessUnitEncodingItem* WorkerOutput;
  typedef VideoCodecException     WorkerException;

  // thread function
  WorkerOutput operator() ( WorkerInput &rcWorkItem )
  {
    if( Err::m_nOK != rcWorkItem->encodeAU() )
    {
      TOT( 1 ); // if coding stops here, set number of threads to one to find out the place where the error ocurred
    }
    return rcWorkItem;
  }
};


class AUCodingStarter
{
public:
#if ALLOW_BOOST_LIBRARY
  AUCodingStarter() : m_iItemsInFIFO( 0 ) {}
  void setNumThreads( UInt uiNumThreads )
  {
    TOF( m_cAUEncFIFO.getNumThreads() == 0 );
    for( UInt ui = 0; ui < uiNumThreads; ui++ )
    {
      m_cAUEncFIFO.addThread();
    }
  }
#endif

  UInt getNumCodingItems() const
  {
#if ALLOW_BOOST_LIBRARY
    if( m_cAUEncFIFO.getNumThreads() > 1 )
    {
      return UInt( m_iItemsInFIFO + m_apcNextPocs.size() );
    }
#endif
    return UInt( m_apcEncodedItems.size() + m_apcNextPocs.size() );
  }

  virtual ~AUCodingStarter();

  ErrVal pushBackAUCodingItem( AccessUnitEncodingItem *pcAUCodingItem );
  ErrVal startCodableItems( const CodedPictureStoreBase &rcCodedPictureStore );
  ErrVal checkForCodedItems( std::vector<AccessUnitEncodingItem*> &rapcCodedItems, bool bWaitForOne );

private:
  ErrVal xEncodeItem( const CodedPictureStoreBase &rcCodedPictureStore, AccessUnitEncodingItem *pcAUCodingItem );

#if ALLOW_BOOST_LIBRARY
  MultiThreadedFIFO<AUEncodingWorker> m_cAUEncFIFO;
  int m_iItemsInFIFO;
#endif
  std::vector<AccessUnitEncodingItem*> m_apcEncodedItems;
  std::list<AccessUnitEncodingItem*> m_apcNextPocs;
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _AUCODINGSTARTER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
