/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _ACCESSUNITENCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _ACCESSUNITENCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "AccessUnitEncoderParameters.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/BitWriter.h"
#include "SliceEstimator.h"
#include "SyntaxElementProcessingChainEnc.h"
#include "PredEstAndEnc.h"
#include "ResEstAndEnc.h"
#include "SubdivisionEstimator.h"
#include "SingleNodeEstimator.h"
#include "CabaEncoder.h"
#include "VLCEncoder.h"
#include "VideoCodecToolsLib/PacketParser.h"
#include "PredAndResEstConfig.h"
#include "TraceEnc.h"
#include "VideoCommonLib/ExtendedCabacTables.h"
#include "VideoCommonLib/BitWriter.h"
#include "VideoCommonLib/CabacMultiCoder.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class UvlcWriter;

template<class t_StoreType, class t_WriteType>
class MultiVectorBitWriter
{
public:
  MultiVectorBitWriter() : m_acVectors( 1 ), m_cBitWriter( m_acVectors.front() ) {}

  BitWriter<t_StoreType, t_WriteType>&  getBitWriter() { return m_cBitWriter; }
  std::list<std::vector<t_StoreType> >& getVectors()   { return m_acVectors; }

private:
  std::list<std::vector<t_StoreType> > m_acVectors;
  BitWriter<t_StoreType, t_WriteType> m_cBitWriter;
};


class ExtendedTablesCabaEncoder
{
public:

  ExtendedTablesCabaEncoder()
    : m_pcProbBinOut( NULL )
    , m_pcCabaEncoder( NULL )
    , m_cBitWriter( m_aucBitstream ) {}

  virtual ~ExtendedTablesCabaEncoder()
  {
    delete m_pcCabaEncoder;
  }

  ErrVal createEncoder( const AccessUnitParameters &rcAUPars )
  {
    TOT( m_pcCabaEncoder );
    if( rcAUPars.getVLCFlag() )
    {
      VLCEncoder<VLCodeType>* pcVLCEncoder = new VLCEncoder<VLCodeType>( m_cBitWriter, rcAUPars.getVLCMaxDelayInBits() );
      if( !rcAUPars.getEntropySliceFlag() )
      {
        RNOK( pcVLCEncoder->startCabac() );
      }
      m_pcProbBinOut  = pcVLCEncoder;
      m_pcCabaEncoder = pcVLCEncoder;
    }
    else if( rcAUPars.getExtendedCabacTables() == AGGREGATE_CABAC_TABLES12 )
    {
      TOT( rcAUPars.getEntropySliceFlag() );

      CabacMultiCoder<CabaEncoder<AggregateCabacTables12>, BitWriterIf<UInt>, AggregateCabacTables12, true> *pcCabaMultiEncoder;
      pcCabaMultiEncoder = new CabacMultiCoder<CabaEncoder<AggregateCabacTables12>, BitWriterIf<UInt>, AggregateCabacTables12, true>( rcAUPars.getPicOrderCnt() );
      for( UInt ui = 0; ui < AggregateCabacTables12::eNumAggregateStates; ++ui )
      {
        m_apcBitWriter.push_back( new MultiVectorBitWriter<UChar,UInt>() );
        if( ui == 0 )
        {
          pcCabaMultiEncoder->setEqProbBitStream( &m_apcBitWriter.back()->getBitWriter() );
        }
        else
        {
          RNOK( pcCabaMultiEncoder->addCabacCoder( m_apcBitWriter.back()->getBitWriter() ) );
          pcCabaMultiEncoder->getCabacCoder( ui - 1 ).setUseShortFinishFlag( true );
        }
      }
      m_pcProbBinOut = pcCabaMultiEncoder;
      m_pcCabaEncoder = pcCabaMultiEncoder;
    }
    else
    {
      CabaEncoder<> *pcCabaEncoderCurr = new CabaEncoder<>( m_cBitWriter );

      if( !rcAUPars.getEntropySliceFlag() )
      {
        RNOK( pcCabaEncoderCurr->startCabac() );
      }
      m_pcProbBinOut = pcCabaEncoderCurr;
      m_pcCabaEncoder = pcCabaEncoderCurr;
    }
    return Err::m_nOK;
  }

  ProbBinStreamEncIf&                             getProbBinStreamEncIf()     { return *m_pcProbBinOut; }
  CabaCoderStartFinishIf&                         getCabaCoderStartFinishIf() { return *m_pcCabaEncoder; }
  std::vector<MultiVectorBitWriter<UChar,UInt>*>& getMultiBitWriters()        { return m_apcBitWriter; }
  BitWriter<UChar,UInt>&                          getDefaultCabacBitWriter()  { return m_cBitWriter; }
  std::vector<UChar>&                             getDefaultCabacBitstream()  { return m_aucBitstream; }

private:
  ProbBinStreamEncIf *m_pcProbBinOut;
  CabaCoderStartFinishIf *m_pcCabaEncoder;
  AutoDeletePtrVector<MultiVectorBitWriter<UChar,UInt> > m_apcBitWriter;
  std::vector<UChar> m_aucBitstream;
  BitWriter<UChar,UInt> m_cBitWriter;
};


class AccessUnitEncoder
{
public:
  virtual ~AccessUnitEncoder() {}

  static ErrVal parsePrecodedAccessUnit( PacketParser<LargeFile> &rcPrecodedBitstreamParser,
                                         const CodedPictureStoreBase &rcCodedPictureStore,
                                         bool bParse,
#if MULTICABAC_TRANSCODING_HACK
                                         UInt& ruiParsedPacketSize,
#endif
                                         const AccessUnitParameters &rcAUPars,
                                         AccessUnit *pcAU,
                                         TraceEnc *pcTraceEnc,
                                         SequenceParameterSet*& rpcSequenceParameterSet,
                                         PictureParameterSet*& rpcPictureParameterSet);

  static ErrVal encodeAccessUnit( const PicBlock<const Pel> &rcOrig,
                                  const PicBlock<Pel> &rcEnc,
                                  const PicBlock<Pel> *pcEncPre,
                                  const PicBlock<Pel> *pcEncRes,
                                  const AccessUnitParameters &rcAUPars,
                                  const AccessUnitEncoderParameters &rcAUEstPar,
                                  const std::vector<const CodingState*> &rapcCodingState,
                                  PicOrderCnt iLastKeyPicPoc0,
                                  PicOrderCnt iLastKeyPicPoc1,
                                  AccessUnit *pcAU,
                                  bool bAUWasParsed,
#if MULTICABAC_TRANSCODING_HACK
                                  const UInt uiParsedPacketSize,
#endif
                                  std::list<std::vector<UChar> > &rcBitstreamData,
                                  std::vector<double> &radPSNR,
                                  std::ostream &rcOut,
                                  UInt64 &ruiEncodingDurationMSec,
                                  UInt64 &ruiMEDurationMSec,
                                  TraceEnc *pcTraceEnc );

  static ErrVal encodePlanePackage( UInt uiPlaPaId,
                                    const AccessUnit &rcAU,
                                    ContextStore<CabacContextModel> &rcContextStore,
                                    const ContextStore<CabacContextModel> *pcRefCtxStore,
                                    ProbBinStreamEncIf &rcProbBinOut,
                                    CabaCoderStartFinishIf &rcCabaEncoder,
                                    BitWriter<UChar,UInt> &rcBitWriter,
                                    std::list< std::vector<UChar> > &rcBitstreamData,
                                    TraceEnc *pcTrace );

private:

  template<bool t_bUseProbBinOut>
  static ErrVal xEstimatePlanePackage( UInt uiPlaPaId,
                                       ExtendedCabacTables eExtendedCabacTableIdx,
                                       const PicBlock<const Pel> &rcOrig,
                                       const PicBlock<Pel> &rcEnc,
                                       const PicBlock<Pel> *pcEncPre,
                                       const PicBlock<Pel> *pcEncRes,
                                       AccessUnit &rcAU,
                                       const SliceEstimator::Parameters &rcSliceEstPars,
                                       const AccessUnitEncoderParameters &rcAUEstPars,
                                       const CodingState &rcCodingState,
                                       ProbBinStreamEncIf *pcProbBinOut,
                                       UInt64 &ruiMEDuration,
                                       std::ostream &rcConsoleOutput,
                                       TraceEnc *pcTrace,
                                       BitRate &rcEstimatedBitrate );

  static ErrVal xPrintQpAndFrameType( std::ostream &rcOut,
                                      const PlanePackage &rcPlaPaCurr,
                                      const AccessUnitParameters &rcAUPars,
                                      bool bTestIntraModes,
                                      bool bTestInterModes );

  static ErrVal xSetMaxAllowedLog2TrafoBlockSize( VecResEstInfo &rcVecResEstInfo, UInt uiMax );

  static ErrVal xCalcPSNR( const PicBlock<const Pel> &rcOrig,
                           const PicBlock<const Pel> &rcEnc,
                           std::vector<double> &radPSNR );

  static ErrVal xEstimateInterpolationMethod( Block<double>* &rpcInLoopFiltMatrix,
                                              UInt           &ruiUpsMethodEst );

  static ErrVal xWriteMultiplex( std::vector<MultiVectorBitWriter<UChar,UInt>*> &racPackets,
                                 std::list<std::vector<UChar> > &rcBitstreamData,
                                 BitWriterIf<UInt> &rcBitWriter,
                                 UInt &ruiBytesInMultiplex );

};




template<bool t_bUseProbBinOut>
ErrVal AccessUnitEncoder::xEstimatePlanePackage( UInt uiPlaPaId,
                                                 ExtendedCabacTables eExtendedCabacTableIdx,
                                                 const PicBlock<const Pel> &rcOrig,
                                                 const PicBlock<Pel> &rcEnc,
                                                 const PicBlock<Pel> *pcEncPre,
                                                 const PicBlock<Pel> *pcEncRes,
                                                 AccessUnit &rcAU,
                                                 const SliceEstimator::Parameters &rcSliceEstPars,
                                                 const AccessUnitEncoderParameters &rcAUEstPars,
                                                 const CodingState &rcCodingState,
                                                 ProbBinStreamEncIf *pcProbBinOut,
                                                 UInt64 &ruiMEDuration,
                                                 std::ostream &rcConsoleOutput,
                                                 TraceEnc *pcTrace,
                                                 BitRate &rcEstimatedBitrate )
{
  const AccessUnitParameters& rcAUPars    =  rcAU.getAUPars          ();
  PlanePackage&               rcPlaPaCurr = *rcAU.getPlanePackage    ( uiPlaPaId );
  const PlanePackage*         pcRefPlaPa  =  rcAU.getRefPlanePackage ( uiPlaPaId );

  StateMultiplier cStateMultiplier;
  SyntaxElementProcessingChainEncIf<> *pcSEProcChainEnc;

#if 0
  if( eExtendedCabacTableIdx == AGGREGATE_CABAC_TABLES12 )
  {
    pcSEProcChainEnc = new SyntaxElementProcessingChainEnc<true, AggregateCabacTables12, t_bUseProbBinOut>( &cStateMultiplier,
                                                                                                            pcProbBinOut,
                                                                                                            rcAU.getCtxStore( uiPlaPaId ).getCtxIndices(),
                                                                                                            rcAU.getCtxStore( uiPlaPaId ).getCtxInit(),
                                                                                                            rcPlaPaCurr.getPars(),
                                                                                                            rcPlaPaCurr.getCurrentSlice().getSliceQps()[0],
                                                                                                            pcTrace );
  }
  else
#endif
  {
    TOF( eExtendedCabacTableIdx == DEFAULT_CABAC_TABLES );
    pcSEProcChainEnc = new SyntaxElementProcessingChainEnc<true, DefaultCabacTables, t_bUseProbBinOut>( &cStateMultiplier,
                                                                                                        pcProbBinOut,
                                                                                                        rcAU.getCtxStore( uiPlaPaId ).getCtxIndices(),
                                                                                                        rcAU.getCtxStore( uiPlaPaId ).getCtxInit(),
                                                                                                        rcPlaPaCurr.getPars(),
                                                                                                        rcPlaPaCurr.getCurrentSlice().getSliceQps()[0],
                                                                                                        pcTrace );
  }

  if( pcTrace )
  {
    cStateMultiplier.addStateIf( pcTrace );
  }

  const bool bInterFrame            = !rcCodingState.getPlanePackagesForPrediction().empty() && rcAUPars.getMaxMvsPerBlk() != 0;
  const bool bTestInterPlaPaPred    = rcPlaPaCurr.getPars().getInterPlaPaPredFlag();
  const bool bAlwaysInterPlaPaPred  = bTestInterPlaPaPred && !rcPlaPaCurr.getPars().getAdaptiveInterPlaPaPredFlag();
  ROT( bAlwaysInterPlaPaPred && !bInterFrame );
  const bool bTestInterModes        =  bInterFrame && !bAlwaysInterPlaPaPred;
  const bool bTestIntraModes        = !bInterFrame || !rcAUEstPars.getDisableIntraInInter();

  PredAndResEstConfig cPredAndResEstConfig( rcAUEstPars, rcPlaPaCurr, rcAUPars, rcSliceEstPars, bTestIntraModes, bTestInterModes, bTestInterPlaPaPred, pcRefPlaPa );

  RNOK( xPrintQpAndFrameType( rcConsoleOutput, rcPlaPaCurr, rcAUPars, bTestIntraModes, bTestInterModes ) );

  QuadtreeGroupStructure &rcStructure = rcPlaPaCurr.getCurrentSlice().getQuadtreeGroupStructure();
  RDOQ cRDOQ( pcSEProcChainEnc->getProbAlloc(), rcAU.getCtxStore( uiPlaPaId ).getCtxIndices(), rcAU.getCtxStore( uiPlaPaId ).getCtxInit().getCtxAccess(), rcAUEstPars, rcPlaPaCurr.getPars(), rcSliceEstPars );

  const ResidualEstimatorManager cResEstManager( rcPlaPaCurr.getPars(), rcAUEstPars, pcSEProcChainEnc->getSyntaxElementEncoderIf(), pcSEProcChainEnc->getBitRateEstimatorIf(), cStateMultiplier, rcSliceEstPars.getLambda(), cRDOQ );
  PredEstAndEnc cPredEstAndEnc( rcPlaPaCurr.getPars(), rcAUPars, rcAUEstPars, cPredAndResEstConfig.getVecPredEstInfo(), &pcSEProcChainEnc->getSyntaxElementEncoderIf(), rcOrig, &rcEnc, rcCodingState, rcPlaPaCurr.getCurrentSlice() );
  ResEstAndEnc cResEstAndEnc( rcPlaPaCurr.getPars(), rcAUPars, &pcSEProcChainEnc->getSyntaxElementEncoderIf(), rcCodingState, rcPlaPaCurr.getCurrentSlice(), cResEstManager, cPredEstAndEnc.getResEstInfoVecSrcIf() );
  SingleNodeEstimator<PredEstAndEnc::PredIf,ResEstAndEnc::ResIf> cSingleNodeEst( rcAUEstPars, &pcSEProcChainEnc->getSyntaxElementEncoderIf(), &pcSEProcChainEnc->getBitRateEstimatorIf(), &cPredEstAndEnc, &cResEstAndEnc, rcSliceEstPars.getLambda(), rcEnc, pcEncPre, pcEncRes, rcOrig, &cStateMultiplier );

  const ContextStore<CabacContextModel>* pcRefCtxStore = NULL;
  if ( rcAUPars.getCtxStatesReuseFlag() && rcCodingState.getRefCtxStore() != NULL )
  {
    pcRefCtxStore = rcCodingState.getRefCtxStore();
  }

  SubdivisionEstimator<SingleNodeEstimator_IfType, CabacContextModel> cLeafNodeEst( rcAUEstPars, rcStructure, cSingleNodeEst, pcSEProcChainEnc->getProbAlloc(), &rcAU.getCtxStore( uiPlaPaId ), pcRefCtxStore, rcSliceEstPars.getLambda(), &cStateMultiplier, pcTrace );
  BitRate cBRBefore = pcSEProcChainEnc->getBitRateEstimatorIf().getBitRateEstimate();
  RNOK( cLeafNodeEst.estimateTree() );
  rcEstimatedBitrate = pcSEProcChainEnc->getBitRateEstimatorIf().getBitRateEstimate() - cBRBefore;

  ruiMEDuration = cPredEstAndEnc.getTotalMEDuration();

  delete pcSEProcChainEnc;
  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _ACCESSUNITENCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
