/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _ACCESSUNITENCODERPARAMETERS_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _ACCESSUNITENCODERPARAMETERS_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

//#include "SequenceStructure.h"
#include "VideoCommonLib/Block.h"
//#include "VideoCommonLib/AccessUnit.h"
//#include "CabaEncoder.h"
//#include "SliceEstimator.h"
//#include "SyntaxElementProcessingChainEnc.h"
//#include "VideoCommonLib/ContextTables.h"
//#include "PredEstAndEnc.h"
//#include "ResEstAndEnc.h"
//#include "SubdivisionEstimator.h"
//#include "SingleNodeEstimator.h"

//#include "VideoCodecToolsLib/PacketParser.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

//class CodingState;
//template<class> class BitWriterIf;
//class TraceEnc;
//class TraceCommon;
//class CtxInitArray;
//class ContextArrayStartIndices;
//class RDOQ;

class AccessUnitEncoderParameters
{
public:
  void  setPrimQp                     ( int         iPrimQp )                   { m_iPrimQp = iPrimQp; }
  void  setDeadzoneInter              ( Ratio       cDeadzone )                 { m_cDeadzoneInter = cDeadzone; }
  void  setDeadzoneIntra              ( Ratio       cDeadzone )                 { m_cDeadzoneIntra = cDeadzone; }
  void  setDisableIntraInInter        ( bool        bDisableIntraInInter )      { m_bDisableIntraInInter = bDisableIntraInInter; }
  void  setNumberOfReferenceFrames    ( UInt        uiNumberOfReferenceFrames ) { m_uiNumberOfReferenceFrames = uiNumberOfReferenceFrames; }
  void  setSearchModeSubPel           ( UInt        uiSearchModeSubPel )        { m_uiSearchModeSubPel = uiSearchModeSubPel; }
  void  setSearchModeFullPel          ( UInt        uiSearchModeFullPel )       { m_uiSearchModeFullPel = uiSearchModeFullPel; }
  void  setPrimPlaneOnlySearch        ( bool        bSearchPrimPlaneOnly )      { m_bSearchPrimPlaneOnly = bSearchPrimPlaneOnly; }
  void  setSearchRange                ( UInt        uiSearchRange )             { m_uiSearchRange = uiSearchRange; }
  void  setIterSearchRange            ( UInt        uiIterSearchRange )         { m_uiIterSearchRange = uiIterSearchRange; }
  void  setMaxIterations              ( UInt        uiMaxIterations )           { m_uiMaxIterations = uiMaxIterations; }
  void  setLambdaCalculationMode      ( UInt        uiLambdaCalculationMode )   { m_uiLambdaCalculationMode = uiLambdaCalculationMode; }
  void  setNextPoc                    ( PicOrderCnt cNextPoc )                  { m_cNextPoc = cNextPoc; }
  void  setTwoPassEncodingFlag        ( bool        bTwoPassEncodingFlag )      { m_bTwoPassEncodingFlag = bTwoPassEncodingFlag; }

  void  setFastModeDecision           ( bool        bFastModeDecision,
                                        UInt        uiFMDThresholdHighQP,
                                        UInt        uiFMDThresholdLowQP,
                                        bool        bIntraPic,
                                        int         iPrimQp )
  {
    m_bFastModeDecision             = bFastModeDecision; 
    m_iFmdQStepPercentThreshold     = int( uiFMDThresholdHighQP );
    if( m_bFastModeDecision && ! bIntraPic )
    {
      const int   iMinQpAdjust      = 24;
      const int   iMaxQpAdjust      = 48;
      const int   iDeltaThreshold   = int( uiFMDThresholdLowQP ) - int( uiFMDThresholdHighQP );
      const int   iAddThreshold     = iDeltaThreshold * ( iMaxQpAdjust - gMinMax( iPrimQp, iMinQpAdjust, iMaxQpAdjust ) ) / ( iMaxQpAdjust - iMinQpAdjust );
      m_iFmdQStepPercentThreshold  += iAddThreshold;
    }
  }

  void  setDisableWienerFlag          (bool         bFlag )                     { m_bDisableWienerFilter = bFlag; }
  void  setWienerFilterType           (UInt         uiType )                    { m_uiWienerFilterType = uiType; }
  void  setWienerMinLength            (UInt         uiMinLength )               { m_uiWienerMinLength = uiMinLength; }
  void  setWienerMaxLength            (UInt         uiMaxLength )               { m_uiWienerMaxLength = uiMaxLength; }
  void  setWienerPower2MaxBlock       (UInt         uiMaxBlock )                { m_uiWienerPower2MaxBlock = uiMaxBlock; }
  void  setWienerPower2MinBlock       (UInt         uiMinBlock )                { m_uiWienerPower2MinBlock = uiMinBlock; }

  std::vector<bool>&        getTrafoPow2SubdivideResidual ()          { return m_abTrafoPow2SubdivideResidual; }
  std::vector<UInt>&        getQuantizationMode           ()          { return m_auiQuantizationMode; }
  std::vector<UInt>&        getMaxAbsQPDelta              ()          { return m_auiMaxAbsQPDelta; }

  const std::vector<bool>&  getTrafoPow2SubdivideResidual ()    const { return m_abTrafoPow2SubdivideResidual; }
  const std::vector<UInt>&  getQuantizationMode           ()    const { return m_auiQuantizationMode; }
  const std::vector<UInt>&  getMaxAbsQPDelta              ()    const { return m_auiMaxAbsQPDelta; }

  int         getPrimQp                     ()                  const { return m_iPrimQp; } 
  const Ratio& getDeadzoneInter             ()                  const { return m_cDeadzoneInter; }
  const Ratio& getDeadzoneIntra             ()                  const { return m_cDeadzoneIntra; }
  bool        getDisableIntraInInter        ()                  const { return m_bDisableIntraInInter; }
  UInt        getNumberOfReferenceFrames    ()                  const { return m_uiNumberOfReferenceFrames; }
  UInt        getSearchModeFullPel          ()                  const { return m_uiSearchModeFullPel; }
  UInt        getSearchModeSubPel           ()                  const { return m_uiSearchModeSubPel; }
  bool        getPrimPlaneOnlySearch        ()                  const { return m_bSearchPrimPlaneOnly; }
  UInt        getSearchRange                ()                  const { return m_uiSearchRange; }
  UInt        getIterSearchRange            ()                  const { return m_uiIterSearchRange; }
  UInt        getMaxIterations              ()                  const { return m_uiMaxIterations; }
  UInt        getLambdaCalculationMode      ()                  const { return m_uiLambdaCalculationMode; }
  PicOrderCnt getNextPoc                    ()                  const { return m_cNextPoc; }
  bool        getTwoPassEncodingFlag        ()                  const { return m_bTwoPassEncodingFlag; }
  bool        getTrafoPow2SubdivideResidual ( UInt  uiPlaPaId ) const { return m_abTrafoPow2SubdivideResidual[ uiPlaPaId ]; }
  UInt        getQuantizationMode           ( UInt  uiPlaPaId ) const { return m_auiQuantizationMode[ uiPlaPaId ]; }
  UInt        getMaxAbsQPDelta              ( UInt  uiPlaPaId ) const { return m_auiMaxAbsQPDelta[ uiPlaPaId ]; }
  bool        getFastModeDecision           ()                  const { return m_bFastModeDecision; }
  int         getFmdQStepPercentThreshold   ()                  const { return m_iFmdQStepPercentThreshold; }
  bool        getDisableWienerFlag          ()                  const { return m_bDisableWienerFilter; }
  UInt        getWienerFilterType           ()                  const { return m_uiWienerFilterType; }
  UInt        getWienerMinLength            ()                  const { return m_uiWienerMinLength; }
  UInt        getWinerMaxLength             ()                  const { return m_uiWienerMaxLength; }
  UInt        getWienerPower2MaxBlock       ()                  const { return m_uiWienerPower2MaxBlock; }
  UInt        getWienerPower2MinBlock       ()                  const { return m_uiWienerPower2MinBlock; }

private:
  int               m_iPrimQp;
  Ratio             m_cDeadzoneInter;
  Ratio             m_cDeadzoneIntra;
  bool              m_bDisableIntraInInter;
  UInt              m_uiNumberOfReferenceFrames;
  UInt              m_uiSearchModeFullPel;
  UInt              m_uiSearchModeSubPel;
  bool              m_bSearchPrimPlaneOnly;
  UInt              m_uiSearchRange;
  UInt              m_uiIterSearchRange;
  UInt              m_uiMaxIterations;
  UInt              m_uiLambdaCalculationMode;
  PicOrderCnt       m_cNextPoc;
  bool              m_bTwoPassEncodingFlag;
  std::vector<bool> m_abTrafoPow2SubdivideResidual;
  std::vector<UInt> m_auiQuantizationMode;
  std::vector<UInt> m_auiMaxAbsQPDelta;
  bool              m_bFastModeDecision;
  int               m_iFmdQStepPercentThreshold;
  bool              m_bDisableWienerFilter;
  UInt              m_uiWienerFilterType;
  UInt              m_uiWienerMinLength;
  UInt              m_uiWienerMaxLength;
  UInt              m_uiWienerPower2MaxBlock;
  UInt              m_uiWienerPower2MinBlock;
};

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _ACCESSUNITENCODERPARAMETERS_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
