/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _BITRATEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _BITRATEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/ExtendedCabacTables.h"
#include "ProbabilityAllocatorEnc.h"
#include "StateIf.h"
#include "RDCost.h"
#include "VideoCommonLib/InStreamIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#define USE_EXACT_CABAC_BITRATE 0

class BitRateEstimatorIf
{
protected:
  BitRateEstimatorIf() {}
  virtual ~BitRateEstimatorIf() {}

public:
  virtual BitRate getBitRateEstimate() = 0;
};


template<class t_BREst> class BREstStateIf;


template<class t_CabacTables = DefaultCabacTables, class t_ProbBinStreamEncIf = ProbBinStreamEncIf, bool t_bEnableForwarding = true>
class ProbBinBitRateEstimator
  : public BitRateEstimatorIf
  , public ProbBinStreamEncIf
{
public:
  typedef ProbBinBitRateEstimator TypeCall;
  typedef ProbBinBitRateEstimator BoolCall;

  typedef typename t_CabacTables::Tables CabacTablesType;
  template<class t_BREst> friend class BREstStateIf;

  struct CtorPars
  {
    t_ProbBinStreamEncIf *m_pcStreamOut;
  };

  ProbBinBitRateEstimator( CtorPars &rcPars )
    : m_pcStreamOut( rcPars.m_pcStreamOut )
  {
    if( t_bEnableForwarding )
    {
      TOF( m_pcStreamOut );
    }
    startEstimator();
  }

  void startEstimator()
  {
#if USE_EXACT_CABAC_BITRATE
    m_uiBits = 0;
    m_uiRange = CabacTablesType::sm_uiHalf - 2;
#else
    m_cBitRate = 0;
#endif
  }

  BitRate getBitRateEstimate();

  FORCEINLINE ErrVal pushBack( const ProbBin &rcT );
  FORCEINLINE ErrVal pushBack( const bool &rcT );

private:
  t_ProbBinStreamEncIf *m_pcStreamOut;

  class PBBREState
    : public State
  {
  public:
    PBBREState( ProbBinBitRateEstimator *pcStatesObject ) : m_pcStatesObject( pcStatesObject ) {}

#if USE_EXACT_CABAC_BITRATE
    UInt m_uiBits;
    UInt m_uiRange;
#else
    BitRate m_cBitRate;
#endif
    State *m_pcState;

  private:
    ProbBinBitRateEstimator *m_pcStatesObject;

    ErrVal setState();
    ErrVal deleteState();
  };

#if USE_EXACT_CABAC_BITRATE
  UInt m_uiBits;
  UInt m_uiRange;
private:
  static const BitRate sm_acBitRate[254];
#else
  BitRate m_cBitRate;
#endif
};

template<class t_BREst>
class BREstStateIf
  : public StateIf
{
public:
  BREstStateIf( t_BREst &rcBREst ) : m_rcBREst( rcBREst ) {}

  ErrVal getState( State *&rpcState )
  {
    typename t_BREst::PBBREState *pcPBBREState = new typename t_BREst::PBBREState( &m_rcBREst );
#if USE_EXACT_CABAC_BITRATE
    pcPBBREState->m_uiBits  = m_rcBREst.m_uiBits;
    pcPBBREState->m_uiRange = m_rcBREst.m_uiRange;
#else
    pcPBBREState->m_cBitRate = m_rcBREst.m_cBitRate;
#endif

    rpcState = pcPBBREState;
    return Err::m_nOK;
  }
private:
  t_BREst &m_rcBREst;
};


#if USE_EXACT_CABAC_BITRATE
template<class t_CabacTables, class t_InStreamConstIf_ProbBin, bool t_bEnableForwarding>
const BitRate ProbBinBitRateEstimator<t_CabacTables, t_InStreamConstIf_ProbBin, t_bEnableForwarding>::sm_acBitRate[254] =
{
  BitRate( 0.9943534369 ), BitRate( 0.9887288877 ), BitRate( 0.9831261814 ), BitRate( 0.9775451492 ), BitRate( 0.9719856238 ),
  BitRate( 0.9664474403 ), BitRate( 0.9609304353 ), BitRate( 0.9554344476 ), BitRate( 0.9499593175 ), BitRate( 0.9445048874 ),
  BitRate( 0.9390710014 ), BitRate( 0.9336575052 ), BitRate( 0.9282642464 ), BitRate( 0.9228910743 ), BitRate( 0.9175378398 ),
  BitRate( 0.9122043955 ), BitRate( 0.9068905956 ), BitRate( 0.9015962959 ), BitRate( 0.8963213539 ), BitRate( 0.8910656284 ),
  BitRate( 0.8858289801 ), BitRate( 0.8806112708 ), BitRate( 0.8754123641 ), BitRate( 0.8702321250 ), BitRate( 0.8650704199 ),
  BitRate( 0.8599271166 ), BitRate( 0.8548020845 ), BitRate( 0.8496951940 ), BitRate( 0.8446063174 ), BitRate( 0.8395353278 ),
  BitRate( 0.8344821001 ), BitRate( 0.8294465102 ), BitRate( 0.8244284354 ), BitRate( 0.8194277544 ), BitRate( 0.8144443468 ),
  BitRate( 0.8094780940 ), BitRate( 0.8045288780 ), BitRate( 0.7995965824 ), BitRate( 0.7946810920 ), BitRate( 0.7897822926 ),
  BitRate( 0.7849000712 ), BitRate( 0.7800343161 ), BitRate( 0.7751849164 ), BitRate( 0.7703517627 ), BitRate( 0.7655347464 ),
  BitRate( 0.7607337601 ), BitRate( 0.7559486975 ), BitRate( 0.7511794534 ), BitRate( 0.7464259234 ), BitRate( 0.7416880044 ),
  BitRate( 0.7369655942 ), BitRate( 0.7322585915 ), BitRate( 0.7275668962 ), BitRate( 0.7228904090 ), BitRate( 0.7182290316 ),
  BitRate( 0.7135826667 ), BitRate( 0.7089512180 ), BitRate( 0.7043345899 ), BitRate( 0.6997326880 ), BitRate( 0.6951454185 ),
  BitRate( 0.6905726887 ), BitRate( 0.6860144067 ), BitRate( 0.6814704816 ), BitRate( 0.6769408231 ), BitRate( 0.6724253420 ),
  BitRate( 0.6679239497 ), BitRate( 0.6634365587 ), BitRate( 0.6589630822 ), BitRate( 0.6545034340 ), BitRate( 0.6500575289 ),
  BitRate( 0.6456252826 ), BitRate( 0.6412066114 ), BitRate( 0.6368014322 ), BitRate( 0.6324096631 ), BitRate( 0.6280312226 ),
  BitRate( 0.6236660301 ), BitRate( 0.6193140055 ), BitRate( 0.6149750698 ), BitRate( 0.6106491444 ), BitRate( 0.6063361515 ),
  BitRate( 0.6020360141 ), BitRate( 0.5977486557 ), BitRate( 0.5934740006 ), BitRate( 0.5892119737 ), BitRate( 0.5849625007 ),
  BitRate( 0.5807255078 ), BitRate( 0.5765009220 ), BitRate( 0.5722886707 ), BitRate( 0.5680886822 ), BitRate( 0.5639008852 ),
  BitRate( 0.5597252092 ), BitRate( 0.5555615843 ), BitRate( 0.5514099410 ), BitRate( 0.5472702106 ), BitRate( 0.5431423250 ),
  BitRate( 0.5390262166 ), BitRate( 0.5349218182 ), BitRate( 0.5308290636 ), BitRate( 0.5267478868 ), BitRate( 0.5226782225 ),
  BitRate( 0.5186200059 ), BitRate( 0.5145731728 ), BitRate( 0.5105376596 ), BitRate( 0.5065134030 ), BitRate( 0.5025003405 ),
  BitRate( 0.4984984100 ), BitRate( 0.4945075498 ), BitRate( 0.4905276989 ), BitRate( 0.4865587967 ), BitRate( 0.4826007831 ),
  BitRate( 0.4786535986 ), BitRate( 0.4747171840 ), BitRate( 0.4707914808 ), BitRate( 0.4668764308 ), BitRate( 0.4629719763 ),
  BitRate( 0.4590780602 ), BitRate( 0.4551946258 ), BitRate( 0.4513216166 ), BitRate( 0.4474589770 ), BitRate( 0.4436066515 ),
  BitRate( 0.4397645852 ), BitRate( 0.4359327236 ), BitRate( 0.4321110126 ), BitRate( 0.4282993987 ), BitRate( 0.4244978285 ),
  BitRate( 0.4207062494 ), BitRate( 0.4169246088 ), BitRate( 0.4131528549 ), BitRate( 0.4093909361 ), BitRate( 0.4056388013 ),
  BitRate( 0.4018963996 ), BitRate( 0.3981636807 ), BitRate( 0.3944405947 ), BitRate( 0.3907270919 ), BitRate( 0.3870231231 ),
  BitRate( 0.3833286396 ), BitRate( 0.3796435927 ), BitRate( 0.3759679346 ), BitRate( 0.3723016174 ), BitRate( 0.3686445938 ),
  BitRate( 0.3649968168 ), BitRate( 0.3613582397 ), BitRate( 0.3577288163 ), BitRate( 0.3541085006 ), BitRate( 0.3504972471 ),
  BitRate( 0.3468950104 ), BitRate( 0.3433017457 ), BitRate( 0.3397174083 ), BitRate( 0.3361419541 ), BitRate( 0.3325753391 ),
  BitRate( 0.3290175197 ), BitRate( 0.3254684526 ), BitRate( 0.3219280949 ), BitRate( 0.3183964039 ), BitRate( 0.3148733374 ),
  BitRate( 0.3113588532 ), BitRate( 0.3078529097 ), BitRate( 0.3043554654 ), BitRate( 0.3008664794 ), BitRate( 0.2973859106 ),
  BitRate( 0.2939137187 ), BitRate( 0.2904498634 ), BitRate( 0.2869943048 ), BitRate( 0.2835470032 ), BitRate( 0.2801079192 ),
  BitRate( 0.2766770138 ), BitRate( 0.2732542482 ), BitRate( 0.2698395837 ), BitRate( 0.2664329823 ), BitRate( 0.2630344058 ),
  BitRate( 0.2596438166 ), BitRate( 0.2562611772 ), BitRate( 0.2528864505 ), BitRate( 0.2495195994 ), BitRate( 0.2461605873 ),
  BitRate( 0.2428093778 ), BitRate( 0.2394659347 ), BitRate( 0.2361302221 ), BitRate( 0.2328022044 ), BitRate( 0.2294818461 ),
  BitRate( 0.2261691121 ), BitRate( 0.2228639674 ), BitRate( 0.2195663773 ), BitRate( 0.2162763073 ), BitRate( 0.2129937233 ),
  BitRate( 0.2097185913 ), BitRate( 0.2064508775 ), BitRate( 0.2031905483 ), BitRate( 0.1999375705 ), BitRate( 0.1966919110 ),
  BitRate( 0.1934535369 ), BitRate( 0.1902224157 ), BitRate( 0.1869985148 ), BitRate( 0.1837818021 ), BitRate( 0.1805722456 ),
  BitRate( 0.1773698136 ), BitRate( 0.1741744744 ), BitRate( 0.1709861968 ), BitRate( 0.1678049496 ), BitRate( 0.1646307018 ),
  BitRate( 0.1614634227 ), BitRate( 0.1583030818 ), BitRate( 0.1551496488 ), BitRate( 0.1520030934 ), BitRate( 0.1488633859 ),
  BitRate( 0.1457304964 ), BitRate( 0.1426043954 ), BitRate( 0.1394850536 ), BitRate( 0.1363724417 ), BitRate( 0.1332665309 ),
  BitRate( 0.1301672922 ), BitRate( 0.1270746971 ), BitRate( 0.1239887173 ), BitRate( 0.1209093243 ), BitRate( 0.1178364903 ),
  BitRate( 0.1147701872 ), BitRate( 0.1117103875 ), BitRate( 0.1086570635 ), BitRate( 0.1056101880 ), BitRate( 0.1025697336 ),
  BitRate( 0.0995356736 ), BitRate( 0.0965079809 ), BitRate( 0.0934866289 ), BitRate( 0.0904715911 ), BitRate( 0.0874628413 ),
  BitRate( 0.0844603531 ), BitRate( 0.0814641006 ), BitRate( 0.0784740580 ), BitRate( 0.0754901996 ), BitRate( 0.0725124998 ),
  BitRate( 0.0695409333 ), BitRate( 0.0665754748 ), BitRate( 0.0636160993 ), BitRate( 0.0606627819 ), BitRate( 0.0577154979 ),
  BitRate( 0.0547742225 ), BitRate( 0.0518389315 ), BitRate( 0.0489096005 ), BitRate( 0.0459862053 ), BitRate( 0.0430687219 ),
  BitRate( 0.0401571265 ), BitRate( 0.0372513953 ), BitRate( 0.0343515048 ), BitRate( 0.0314574315 ), BitRate( 0.0285691522 ),
  BitRate( 0.0256866437 ), BitRate( 0.0228098829 ), BitRate( 0.0199388471 ), BitRate( 0.0170735134 ), BitRate( 0.0142138592 ),
  BitRate( 0.0113598622 ), BitRate( 0.0085114999 ), BitRate( 0.0056687501 ), BitRate( 0.0028315908 )
};
#endif

template<class t_CabacTables, class t_InStreamConstIf_ProbBin, bool t_bEnableForwarding>
BitRate ProbBinBitRateEstimator<t_CabacTables, t_InStreamConstIf_ProbBin, t_bEnableForwarding>::getBitRateEstimate()
{
#if USE_EXACT_CABAC_BITRATE
  if( m_uiRange == CabacTablesType::sm_uiHalf - 2 )
  {
    return (int)m_uiBits;
  }
//  AOT_DBG( m_uiRange < 256 || m_uiRange > 509 );

//   // generator loop for array sm_acBitRate
//   for( UInt ui = 256; ui < 510; ui++ )
//   {
//     static const double dLog2  = log( 2.0 );
//     std::cout <<  " BitRate( " << std::fixed << std::setprecision( 10 ) << std::setw( 12 ) << (log((double)510.0/(double)ui)/dLog2) << " )," << std::endl;
//   }
  return sm_acBitRate[m_uiRange-256] + BitRate( (int)m_uiBits );
#else
  return m_cBitRate;
#endif
}


template<class t_CabacTables, class t_InStreamConstIf_ProbBin, bool t_bEnableForwarding>
FORCEINLINE ErrVal ProbBinBitRateEstimator<t_CabacTables, t_InStreamConstIf_ProbBin, t_bEnableForwarding>::pushBack( const ProbBin &rcT )
{
#if USE_EXACT_CABAC_BITRATE
  const UInt uiConvertedIndex = t_CabacTables::sm_auiIdxToAggregateState[t_CabacTables::sm_auiStateToIdx[rcT.getStateIdx()]];
  UInt uiLPS = CabacTablesType::sm_auiLPSTable[uiConvertedIndex][(m_uiRange>>(6+CabacTablesType::sm_uiDifBits)) & 3];

  m_uiRange -= uiLPS;
  if( rcT.isLps() )
  {
    m_uiRange = uiLPS;
  }

  while( m_uiRange < CabacTablesType::sm_uiQuarter )
  {
    m_uiBits++;
    m_uiRange <<= 1;
  }
#else
  const UInt uiConvertedIndex = t_CabacTables::sm_auiIdxToAggregateState[t_CabacTables::sm_auiStateToIdx[rcT.getStateIdx()]];
  m_cBitRate += ProbBin( uiConvertedIndex, rcT.isLps() ).getProbBinBitRate<t_CabacTables>();
#endif

  // pass through
  if( t_bEnableForwarding && m_pcStreamOut )
  {
    static_cast<typename t_InStreamConstIf_ProbBin::TypeCall*>(m_pcStreamOut)->pushBack( rcT );
  }

  return Err::m_nOK;
}


template<class t_CabacTables, class t_InStreamConstIf_ProbBin, bool t_bEnableForwarding>
FORCEINLINE ErrVal ProbBinBitRateEstimator<t_CabacTables, t_InStreamConstIf_ProbBin, t_bEnableForwarding>::pushBack( const bool &rcT )
{
#if USE_EXACT_CABAC_BITRATE
  m_uiBits++;
#else
  m_cBitRate += 1;
#endif

  // pass through
  if( t_bEnableForwarding && m_pcStreamOut )
  {
    static_cast<typename t_InStreamConstIf_ProbBin::BoolCall*>(m_pcStreamOut)->pushBack( rcT );
  }

  return Err::m_nOK;
}


template<class t_CabacTables, class t_InStreamConstIf_ProbBin, bool t_bEnableForwarding>
ErrVal ProbBinBitRateEstimator<t_CabacTables, t_InStreamConstIf_ProbBin, t_bEnableForwarding>::PBBREState::setState()
{
#if USE_EXACT_CABAC_BITRATE
  m_pcStatesObject->m_uiBits  = m_uiBits;
  m_pcStatesObject->m_uiRange = m_uiRange;
#else
  m_pcStatesObject->m_cBitRate = m_cBitRate;
#endif

  return Err::m_nOK;
}


template<class t_CabacTables, class t_InStreamConstIf_ProbBin, bool t_bEnableForwarding>
ErrVal ProbBinBitRateEstimator<t_CabacTables, t_InStreamConstIf_ProbBin, t_bEnableForwarding>::PBBREState::deleteState()
{
  delete this;
  return Err::m_nOK;
}




// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _BITRATEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
