/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _FASTSEARCH_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _FASTSEARCH_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "FastFullSearch.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq, class t_MvMeasure = MvMeasure<> >
class FastLossySearch : public MotionSearchIf<t_OrigPel>, public t_MvMeasure
{
public:
  FastLossySearch() // search configuration -> could be set based on an input parameter
    : m_bStartWithDiamondSearch   ( true )
    , m_uiMaxFirstSearchRounds    ( TypeLimits<UInt>::m_iMax )
    , m_bEnableRasterSearch       ( true )
    , m_bAlwaysRasterSearch       ( false )
    , m_uiRasterDistance          ( 3 )
    , m_bEnableRasterRefinement   ( false )
    , m_bRasterRefinementDiamond  ( false )
    , m_bEnableStarRefinement     ( true )
    , m_bStarRefinementDiamond    ( true )
    , m_uiMaxStarRefinementRounds ( 2 )
  {}

  virtual ~FastLossySearch() {}

  virtual double getTriangleInequalityEliminationRatio() const { return 0.0; }

  virtual ErrVal searchMotion( int iNumPlanesToSearch,
                               const std::vector<SearchControl> &racSearchControl,
                               const QTNavi &rcNav,
                               const SliceData &rcSliceData,
                               const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                               const ReferencePlanePackage<Pel> &rcRefPlaPa,
                               const Lambda cLambda,
                               const MvMeasureInit&  rcMvMeasureInitInput,
                               MotionVector &rcBestPos,
                               const std::vector<MotionVector>& rcCandMvList );
private:
  ErrVal xSearchMvFixedPrecision( int iNumPlanesToSearch,
                                  const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                                  const ReferencePlanePackage<Pel> &rcRefPlaPa,
                                  const MotionVector &rcPosVecToSearchWindow,
                                  const MvMeasureInit&  rcMvMeasureInitInput,
                                  const SearchControl& rcSearchControl,
                                  const Lambda cLambda,
                                  MotionVector &rcPosVecToBestPos,
                                  const std::vector<MotionVector>& rcCandMvList );

  template<class t_RefPel>
  ErrVal xFastMvSearchFullPel   ( const SearchControl& rcSearchControl,
                                  const PicBlock<const t_OrigPel> &rcOrig,
                                  const PicBlock<const t_RefPel> &rcRef,
                                  PrimPixelPos &rcBestPos,
                                  const Lambda cLambda,
                                  const MvMeasureInit& rcMvMeasureInit,
                                  const PicBlock<const UInt> &rcNorms,
                                  const std::vector<MotionVector>& rcCandMvList );

private:
  const bool  m_bStartWithDiamondSearch;
  const UInt  m_uiMaxFirstSearchRounds;
  const bool  m_bEnableRasterSearch;
  const bool  m_bAlwaysRasterSearch;
  const UInt  m_uiRasterDistance;
  const bool  m_bEnableRasterRefinement;
  const bool  m_bRasterRefinementDiamond;
  const bool  m_bEnableStarRefinement;
  const bool  m_bStarRefinementDiamond;
  const UInt  m_uiMaxStarRefinementRounds;

  FullSearch<t_eDistMeasure,t_MvMeasure,t_OrigPel,int>  m_cSearch;
  Int64                                                 m_aiPseudoSAD4x4RowValues[ 4 * MAX_4X4_BLOCKS ];

private:
  template<class t_RefPel>
  class SearchStruct
  {
  public:
    SearchStruct( const bool                                bMultipleOf4x4Blks,
                  const PicBlock<const t_OrigPel>&          rcOrig,
                  const PicBlock<const t_RefPel>&           rcRef,
                  const std::vector<SearchScaleParameters>& racSearchScalePars,
                  const PicBlock<const UInt>&               rcRefNorms,
                  const PicBlock<const UInt>&               rcOrgNorms,
                  const Lambda                              cLambda,
                  const PrimPixelPos                        cMinPos,
                  const PrimPixelPos                        cMaxPos )
    : m_bMultipleOf4x4Blks( bMultipleOf4x4Blks )
    , m_rcOrig            ( rcOrig )
    , m_rcRef             ( rcRef )
    , m_racSearchScalePars( racSearchScalePars )
    , m_rcRefNorms        ( rcRefNorms )
    , m_rcOrgNorms        ( rcOrgNorms )
    , m_cLambda           ( cLambda )
    , m_cMinPos           ( cMinPos )
    , m_cMaxPos           ( cMaxPos )
    , m_cInitCost         ( xGetMaxCost() )
    , m_cBestPos          ( PrimPixelPos( -1, -1 ) )
    , m_cBestCost         ( m_cInitCost )
    , m_uiBestDistance    ( 0 )
    , m_uiBestRound       ( 0 )
    , m_uiBestPoint       ( 0 )
    {}

    const bool&                               isMultOf4x4Blks ()  const { return m_bMultipleOf4x4Blks; }
    const PicBlock<const t_OrigPel>&          getOrgPicBlock  ()  const { return m_rcOrig; }
    const PicBlock<const t_RefPel>&           getRefPicBlock  ()  const { return m_rcRef; }
    const std::vector<SearchScaleParameters>& getScalePars    ()  const { return m_racSearchScalePars; }
    const PicBlock<const UInt>&               getRefNorms     ()  const { return m_rcRefNorms; }
    const PicBlock<const UInt>&               getOrgNorms     ()  const { return m_rcOrgNorms; }
    const Lambda&                             getLambda       ()  const { return m_cLambda; }
    const PrimPixelPos&                       getMinPos       ()  const { return m_cMinPos; }
    const PrimPixelPos&                       getMaxPos       ()  const { return m_cMaxPos; }
    const bool                                pointFound      ()  const { return m_cInitCost != m_cBestCost; }

    PrimPixelPos&                             getBestPos      ()        { return m_cBestPos; }
    LambdaTimesRate&                          getBestCost     ()        { return m_cBestCost; }
    UInt&                                     getBestDistance ()        { return m_uiBestDistance; }
    UInt&                                     getBestRound    ()        { return m_uiBestRound; }
    UInt&                                     getBestPoint    ()        { return m_uiBestPoint; }

  private:
    static LambdaTimesRate xGetMaxCost()
    {
      LambdaTimesRate cInitCost;
      cInitCost.setNumerator( TypeLimits<Int64>::m_iMax );
      return cInitCost;
    }

  private:
    const bool                                m_bMultipleOf4x4Blks;
    const PicBlock<const t_OrigPel>&          m_rcOrig;
    const PicBlock<const t_RefPel>&           m_rcRef;
    const std::vector<SearchScaleParameters>& m_racSearchScalePars;
    const PicBlock<const UInt>&               m_rcRefNorms;
    const PicBlock<const UInt>&               m_rcOrgNorms;
    const Lambda                              m_cLambda;
    const PrimPixelPos                        m_cMinPos;
    const PrimPixelPos                        m_cMaxPos;
    const LambdaTimesRate                     m_cInitCost;
    PrimPixelPos                              m_cBestPos;
    LambdaTimesRate                           m_cBestCost;
    UInt                                      m_uiBestDistance;
    UInt                                      m_uiBestRound;
    UInt                                      m_uiBestPoint;
  };

private:
  template<class t_RefPel>  FORCEINLINE void  xCheckInsidePos ( SearchStruct<t_RefPel>& rcSearchStruct, int iCandX,   int iCandY,   UInt  uiDistance, UInt  uiPoint );
  template<class t_RefPel>  FORCEINLINE void  xCheckPos       ( SearchStruct<t_RefPel>& rcSearchStruct, int iCandX,   int iCandY,   UInt  uiDistance, UInt  uiPoint );
  template<class t_RefPel>  void              xSquareSearch   ( SearchStruct<t_RefPel>& rcSearchStruct, int iCenterX, int iCenterY, UInt  uiDistance );
  template<class t_RefPel>  void              xDiamondSearch  ( SearchStruct<t_RefPel>& rcSearchStruct, int iCenterX, int iCenterY, UInt  uiDistance );
  template<class t_RefPel>  void              xTwoPointSearch ( SearchStruct<t_RefPel>& rcSearchStruct );
  template<class t_RefPel>  ErrVal            xFullPelSearch  ( SearchStruct<t_RefPel>& rcSearchStruct, const std::vector<MotionVector>& rcCandMvList );
};



template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
ErrVal 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::searchMotion( int iNumPlanesToSearch,
                                                                                   const std::vector<SearchControl> &racSearchControl,
                                                                                   const QTNavi &rcNav,
                                                                                   const SliceData &rcSliceData,
                                                                                   const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                                                                                   const ReferencePlanePackage<Pel> &rcRefPlaPa,
                                                                                   const Lambda cLambda,
                                                                                   const MvMeasureInit&  rcMvMeasureInitInput,
                                                                                   MotionVector &rcBestPos,
                                                                                   const std::vector<MotionVector>& rcCandMvList )
{
  MotionVector cPosVecToBestPos = rcBestPos.plusPrimResPos( rcNav.getPrimPixelPos() );

  for( UInt ui = 0; ui < racSearchControl.size(); ui++ )
  {
    const SearchControl &rcCtrl = racSearchControl[ui];
    const ReferencePlanePackage<Pel> &rcRefPlaPaForME = rcRefPlaPa;
    cPosVecToBestPos = cPosVecToBestPos.getMvWithPrecisionIncreasedTo( rcCtrl.getPrecision() );

    MotionVector cPosVecToSearchWindow( cPosVecToBestPos.getX() - rcCtrl.getSearchWindowCenter().getX(),
                                        cPosVecToBestPos.getY() - rcCtrl.getSearchWindowCenter().getY(), cPosVecToBestPos.getPrecision() );
    
    if( rcCtrl.getPrecision() == 0 && iNumPlanesToSearch > 1 ) // decrease vector accuracy if required
    {
      int iPosVecPrecision = 0;
      for( UInt uiPlane = 0; uiPlane < rcOrigPicBlock.getStats().getNumPlanes(); uiPlane++ )
      {
        const Subsampling&  rcSub   = rcOrigPicBlock.getStats().getSubsampling()[ uiPlane ];
        const UInt          uiLog2X = rcSub.getX().getDownShift();
        const UInt          uiLog2Y = rcSub.getY().getDownShift();
        ROF( 0 <= rcSub.getX().getDownShift() );
        ROF( 0 <= rcSub.getY().getDownShift() );
        iPosVecPrecision = gMin( iPosVecPrecision, -int( gMax( uiLog2X, uiLog2Y ) ) );
      }
      cPosVecToSearchWindow = cPosVecToSearchWindow.getMvWithPrecisionDecreasedTo( iPosVecPrecision ).getMvWithPrecisionIncreasedTo( rcCtrl.getPrecision() );
    }

    std::vector<MotionVector> acCandInSearchWindowList;
    if( rcCtrl.getPrecision() == 0 )
    {
      for( UInt ui = 0; ui < rcCandMvList.size(); ui++ )
      {
        acCandInSearchWindowList.push_back( rcCandMvList[ ui ].plusPrimResPos( rcNav.getPrimPixelPos() ) - cPosVecToSearchWindow );
      }
    }

    RNOK( xSearchMvFixedPrecision( iNumPlanesToSearch,
                                   rcOrigPicBlock,
                                   rcRefPlaPaForME,
                                   cPosVecToSearchWindow,
                                   rcMvMeasureInitInput,
                                   rcCtrl,
                                   cLambda,
                                   cPosVecToBestPos,
                                   acCandInSearchWindowList ) );
  }
  rcBestPos = cPosVecToBestPos.minusPrimResPos( rcNav.getPrimPixelPos() );
  return Err::m_nOK;
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
ErrVal 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xSearchMvFixedPrecision( int iNumPlanesToSearch,
                                                                                              const PicBlock<const t_OrigPel> &rcOrigPicBlock,
                                                                                              const ReferencePlanePackage<Pel> &rcRefPlaPa,
                                                                                              const MotionVector &rcPosVecToSearchWindow,
                                                                                              const MvMeasureInit&  rcMvMeasureInitInput,
                                                                                              const SearchControl& rcSearchControl,
                                                                                              const Lambda cLambda,
                                                                                              MotionVector &rcPosVecToBestPos,
                                                                                              const std::vector<MotionVector>& rcCandMvList )
{
  ROF( iNumPlanesToSearch <= (int)rcOrigPicBlock.getStats().getNumPlanes() );
  const int iSearchPrecision = rcPosVecToSearchWindow.getPrecision();
  ROF( rcPosVecToBestPos.getPrecision() == iSearchPrecision );

  const PrimBlockSize cUpsBlockSize     = ( ( rcOrigPicBlock.getStats().getPrimBlockSize() - PrimBlockSize( 1, 1 ) ) << iSearchPrecision ) + PrimBlockSize( 1, 1 );
  const PrimBlockSize cSearchRegionSize = rcSearchControl.getSearchWindowSize() + cUpsBlockSize - PrimBlockSize( 1, 1 );

  const int           iMvPrecision  = rcMvMeasureInitInput.m_iMvScale;
  const MotionVector  cSWOrigin     = rcPosVecToSearchWindow.getMvWithPrecisionIncreasedTo( iMvPrecision );
  MvMeasureInit       cMvMeasureInit;
  cMvMeasureInit.m_iMvScale         = iMvPrecision - rcPosVecToSearchWindow.getPrecision();
  cMvMeasureInit.m_bMvCompetition   = rcMvMeasureInitInput.m_bMvCompetition;
  if( cMvMeasureInit.m_bMvCompetition )
  {
    for( UInt uiIdx = 0; uiIdx < rcMvMeasureInitInput.m_acMvCompOrigins.size(); uiIdx++ )
    {
      int iX = cSWOrigin.getX() - rcMvMeasureInitInput.m_acMvCompOrigins[uiIdx].getX();
      int iY = cSWOrigin.getY() - rcMvMeasureInitInput.m_acMvCompOrigins[uiIdx].getY();
      cMvMeasureInit.m_acMvCompOrigins.push_back( PrimPixelPos( iX, iY ) );
    }
  }
  else
  {
    cMvMeasureInit.m_iYOrigin = cSWOrigin.getY() - rcMvMeasureInitInput.m_iYOrigin;
    for( UInt uiIdx = 0; uiIdx < rcMvMeasureInitInput.m_aiYThreshold.size(); uiIdx++ )
    {
      cMvMeasureInit.m_aiYThreshold.push_back( rcMvMeasureInitInput.m_aiYThreshold[ uiIdx ] - cSWOrigin.getY() );
      cMvMeasureInit.m_aiXOrigin   .push_back( cSWOrigin.getX() - rcMvMeasureInitInput.m_aiXOrigin   [ uiIdx ] );
    }
  }

  PrimPixelPos  cBestPos  = rcSearchControl.getSearchWindowCenter();

  bool bFoundNewBestPos = false;
  if( iSearchPrecision == 0 )
  {
    bFoundNewBestPos = true;
    if( bTriangleUneq )
    {
      std::vector<Subsampling> acSRSub;
      for( int i = 0; i < iNumPlanesToSearch; i++ )
      {
        acSRSub.push_back( rcOrigPicBlock.getStats().getSubsampling()[ i ] );
      }
      PicBlockStats       cSearchRegionStats( cSearchRegionSize, acSRSub );
      AllocPicBlock<int>  cSearchRegion     ( cSearchRegionStats );
      AllocPicBlock<UInt> cNormRegion       ( cSearchRegionStats );

      ReferencePlanePackageAccessor<SetPredictionToBlock<int>,Pel> cRefPlaPaAccess( rcRefPlaPa, SetPredictionToBlock<int>() );
      RNOK( cRefPlaPaAccess.getUpsampledBlock( cSearchRegion, rcPosVecToSearchWindow ) );
      RNOK( rcRefPlaPa     .getL1Norm4x4Block( cNormRegion,   rcPosVecToSearchWindow ) );

      RNOK( xFastMvSearchFullPel( rcSearchControl,
                                  rcOrigPicBlock,
                                  cSearchRegion,
                                  cBestPos,
                                  cLambda,
                                  cMvMeasureInit,
                                  cNormRegion,
                                  rcCandMvList ) );
    }
    else
    {
      PicBlock<UInt>* pcNormRegion              = 0;
      PrimPixelPos    cPosVecToSearchWindowEnd  = PrimPixelPos( rcPosVecToSearchWindow.getX() + cSearchRegionSize.getX(), rcPosVecToSearchWindow.getY() + cSearchRegionSize.getY() );

      PrimBlockSize   cDecPicSize               = rcRefPlaPa.getDecRefPlaPa().getStats().getPrimBlockSize();
      bool            bSearchRegionInsidePic    = ( rcPosVecToSearchWindow  .getX() >= 0                  &&
                                                    rcPosVecToSearchWindow  .getY() >= 0                  &&
                                                    cPosVecToSearchWindowEnd.getX() <= cDecPicSize.getX() &&
                                                    cPosVecToSearchWindowEnd.getY() <= cDecPicSize.getY() );

      if( bSearchRegionInsidePic )
      {
        PrimPixelPos cSearchWindowStart( rcPosVecToSearchWindow.getX(), rcPosVecToSearchWindow.getY() );
        PicBlock<ReferencePlanePackage<Pel>::DType> cSearchPlaPa ( rcRefPlaPa.getDecRefPlaPa(), UInt( iNumPlanesToSearch ) ); 
        PicBlock<ReferencePlanePackage<Pel>::DType> cSearchRegion( cSearchWindowStart, cSearchRegionSize, cSearchPlaPa );

        RNOK( xFastMvSearchFullPel( rcSearchControl,
                                    rcOrigPicBlock,
                                    cSearchRegion,
                                    cBestPos,
                                    cLambda,
                                    cMvMeasureInit,
                                    *pcNormRegion,
                                    rcCandMvList ) );
      }
      else
      {
        std::vector<Subsampling> acSRSub;
        for( int i = 0; i < iNumPlanesToSearch; i++ )
        {
          acSRSub.push_back( rcOrigPicBlock.getStats().getSubsampling()[ i ] );
        }
        AllocPicBlock<int> cSearchRegion( PicBlockStats( cSearchRegionSize, acSRSub ) );
        ReferencePlanePackageAccessor<SetPredictionToBlock<int>,Pel> cRefPlaPaAccess( rcRefPlaPa, SetPredictionToBlock<int>() );
        RNOK( cRefPlaPaAccess.getUpsampledBlock( cSearchRegion, rcPosVecToSearchWindow ) );

        RNOK( xFastMvSearchFullPel( rcSearchControl,
                                    rcOrigPicBlock,
                                    cSearchRegion,
                                    cBestPos,
                                    cLambda,
                                    cMvMeasureInit,
                                    *pcNormRegion,
                                    rcCandMvList ) );
      }
    }
  }
  else
  {
    LambdaTimesRate cBestCostVal;
    cBestCostVal.setNumerator( TypeLimits<Int64>::m_iMax );
    std::vector<Subsampling> acSRSub( iNumPlanesToSearch, Subsampling( ShiftScale(0), ShiftScale(0) ) );
    AllocPicBlock<int> cSearchRegion( PicBlockStats( cSearchRegionSize, acSRSub ) );
    ReferencePlanePackageAccessor<SetPredictionToBlockWithStepSize<int>,Pel> cRefPlaPaAccess( rcRefPlaPa, SetPredictionToBlockWithStepSize<int>() );
    RNOK( cRefPlaPaAccess.getUpsampledBlock( cSearchRegion, rcPosVecToSearchWindow ) );

    bFoundNewBestPos = m_cSearch.searchMV( rcOrigPicBlock,
                                           cSearchRegion,
                                           cBestPos,
                                           cLambda,
                                           cBestCostVal,
                                           cMvMeasureInit,
                                           iSearchPrecision );
    ROF( bFoundNewBestPos );
  }

  if( bFoundNewBestPos )
  {
    rcPosVecToBestPos.getX() = rcPosVecToSearchWindow.getX() + cBestPos.getX();
    rcPosVecToBestPos.getY() = rcPosVecToSearchWindow.getY() + cBestPos.getY();
    rcPosVecToBestPos.getPrecision() = rcPosVecToSearchWindow.getPrecision();
  }

  return Err::m_nOK;
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
ErrVal 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xFastMvSearchFullPel( const SearchControl& rcSearchControl,
                                                                                           const PicBlock<const t_OrigPel> &rcOrig,
                                                                                           const PicBlock<const t_RefPel> &rcRef,
                                                                                           PrimPixelPos &rcBestPos,
                                                                                           const Lambda cLambda,
                                                                                           const MvMeasureInit& rcMvMeasureInit,
                                                                                           const PicBlock<const UInt> &rcNorms,
                                                                                           const std::vector<MotionVector>& rcCandMvList )
{
  CAOF( t_eDistMeasure == DIST_SAD || ! bTriangleUneq );
  ROF ( rcSearchControl.getFullPelSearchMode() == ( bTriangleUneq ? 3 : 2 ) );

  t_MvMeasure::init( rcMvMeasureInit );
  
  bool bBlockSizeIsMultipleOf4x4 = true;
  for( UInt uiPlane = 0; uiPlane < rcRef.getStats().getNumPlanes(); uiPlane++ )
  {
    if( ( rcOrig.getPlane( uiPlane ).getSize().getX() % 4 ) != 0 || ( rcOrig.getPlane( uiPlane ).getSize().getY() % 4 ) != 0 )
    {
      bBlockSizeIsMultipleOf4x4 = false;
    }
    ROF( rcOrig.getStats().getSubsampling()[ uiPlane ] == rcRef.getStats().getSubsampling()[ uiPlane ] );
  }

  const PrimBlockSize& cSizeInRef   = rcOrig.getStats().getPrimBlockSize();
  const PrimBlockSize cOrig4x4Size  = rcOrig.getStats().getPrimBlockSize() >> 2;

  std::vector<SearchScaleParameters> acSearchScalePars;
  for( UInt uiPlane = 0; uiPlane < rcRef.getStats().getNumPlanes(); uiPlane++ )
  {
    acSearchScalePars.push_back( SearchScaleParameters( cSizeInRef, rcRef.getStats().getSubsampling()[ uiPlane ], rcOrig.getStats().getSubsampling()[ uiPlane ] ) );
  }

  const int           iYMax   = rcRef.getStats().getPrimBlockSize().getY() - rcOrig.getStats().getPrimBlockSize().getY();
  const int           iXMax   = rcRef.getStats().getPrimBlockSize().getX() - rcOrig.getStats().getPrimBlockSize().getX();
  const PrimPixelPos  cMinPos = PrimPixelPos( 0, 0 );
  const PrimPixelPos  cMaxPos = PrimPixelPos( iXMax, iYMax );

  if( bTriangleUneq && bBlockSizeIsMultipleOf4x4 )
  {
    AllocPicBlock<UInt> cOrig4x4L1Norm( PicBlockStats( cOrig4x4Size, rcRef.getStats().getSubsampling() ) );
    for( UInt uiPlane = 0; uiPlane < acSearchScalePars.size(); uiPlane++ )
    {
      const Block<const t_OrigPel>& rcOrigBlk = rcOrig        .getPlane( uiPlane );
      const Block<UInt>&            rcNormBlk = cOrig4x4L1Norm.getPlane( uiPlane );
      for( int y = 0; y < acSearchScalePars[ uiPlane ].get4x4BlkSize().getY(); y++ )
      {
        for( int x = 0; x < acSearchScalePars[ uiPlane ].get4x4BlkSize().getX(); x++ )
        {
          UInt uiBlockNorm = 0;
          for( int yy = 0; yy < 4; yy++ )
          {
            for( int xx = 0; xx < 4; xx++ )
            {
              uiBlockNorm += abs( rcOrigBlk.getEl( PixelPos( ( x << 2 ) + xx, ( y << 2 ) + yy ) ) );
            }
          }
          rcNormBlk.getEl( PixelPos( x, y ) ) = uiBlockNorm;
        }
      }
    }
    ROF( acSearchScalePars.size()                            <= 4 );
    ROF( cOrig4x4L1Norm.getStats().getPrimBlockSize().getY() <= MAX_4X4_BLOCKS );

    SearchStruct<t_RefPel> cSearchStruct( bBlockSizeIsMultipleOf4x4, rcOrig, rcRef, acSearchScalePars, rcNorms, cOrig4x4L1Norm, cLambda, cMinPos, cMaxPos );
    RNOK( xFullPelSearch( cSearchStruct, rcCandMvList ) );
    rcBestPos = cSearchStruct.getBestPos();
  }
  else
  {
    PicBlock<UInt>* pcOrig4x4L1Norm = 0;
    SearchStruct<t_RefPel> cSearchStruct( bBlockSizeIsMultipleOf4x4, rcOrig, rcRef, acSearchScalePars, rcNorms, *pcOrig4x4L1Norm, cLambda, cMinPos, cMaxPos );
    RNOK( xFullPelSearch( cSearchStruct, rcCandMvList ) );
    rcBestPos = cSearchStruct.getBestPos();
  }
  return Err::m_nOK;
}



template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
FORCEINLINE void 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xCheckInsidePos( SearchStruct<t_RefPel>& rcSearchStruct,
                                                                                      int                     iCandX,
                                                                                      int                     iCandY,
                                                                                      UInt                    uiDistance,
                                                                                      UInt                    uiPoint )
{
  const PrimPixelPos  cCandPos  = PrimPixelPos( iCandX, iCandY );
  if( bTriangleUneq && rcSearchStruct.isMultOf4x4Blks() )
  {
    const BitRate cMvRate = t_MvMeasure::getRate( iCandX, iCandY );
    FullSearch<t_eDistMeasure,t_MvMeasure,t_OrigPel,t_RefPel>::xCheckCandidate( cCandPos,
                                                                                rcSearchStruct.getOrgPicBlock(),
                                                                                rcSearchStruct.getRefPicBlock(),
                                                                                rcSearchStruct.getScalePars  (),
                                                                                rcSearchStruct.getRefNorms   (),
                                                                                rcSearchStruct.getOrgNorms   (),
                                                                                rcSearchStruct.getLambda     () * cMvRate,
                                                                                m_aiPseudoSAD4x4RowValues,
                                                                                rcSearchStruct.getBestCost   (),
                                                                                rcSearchStruct.getBestPos    () );
    if( rcSearchStruct.getBestPos() == cCandPos )
    {
      rcSearchStruct.getBestDistance() = uiDistance;
      rcSearchStruct.getBestRound   () = 0;
      rcSearchStruct.getBestPoint   () = uiPoint;
    }
  }
  else
  {
    const BitRate   cMvRate = t_MvMeasure::getRate( iCandX, iCandY );
    LambdaTimesRate cCost   = rcSearchStruct.getLambda() * cMvRate;
    for( UInt uiPlane = 0; uiPlane < rcSearchStruct.getScalePars().size(); uiPlane++ )
    {
      const SearchScaleParameters&  rcSSP    = rcSearchStruct.getScalePars()[ uiPlane ];  
      Block<const t_RefPel>         cRefStart( rcSSP.getPixelPos( cCandPos ), rcSSP.getBlockSize(), rcSearchStruct.getRefPicBlock().getPlane( uiPlane ) );
      cCost += DistortionComputation<t_OrigPel,t_RefPel,t_eDistMeasure>::computeDist( rcSearchStruct.getOrgPicBlock().getPlane( uiPlane ), cRefStart/*, 0, 0*/ );
      if( cCost >= rcSearchStruct.getBestCost() )
      {
        return;
      }
    }
    rcSearchStruct.getBestCost    () = cCost;
    rcSearchStruct.getBestPos     () = cCandPos;
    rcSearchStruct.getBestDistance() = uiDistance;
    rcSearchStruct.getBestRound   () = 0;
    rcSearchStruct.getBestPoint   () = uiPoint;
  }
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
FORCEINLINE void 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xCheckPos( SearchStruct<t_RefPel>& rcSearchStruct,
                                                                                int                     iCandX,
                                                                                int                     iCandY,
                                                                                UInt                    uiDistance,
                                                                                UInt                    uiPoint )
{
  ROFVS( iCandX >= rcSearchStruct.getMinPos().getX() && iCandX <= rcSearchStruct.getMaxPos().getX() );
  ROFVS( iCandY >= rcSearchStruct.getMinPos().getY() && iCandY <= rcSearchStruct.getMaxPos().getY() );

  xCheckInsidePos( rcSearchStruct, iCandX, iCandY, uiDistance, uiPoint );
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
void 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xSquareSearch( SearchStruct<t_RefPel>& rcSearchStruct,
                                                                                    int                     iCenterX,
                                                                                    int                     iCenterY,
                                                                                    UInt                    uiDistance )
{ /*  8 point search,                 1 2 3
      search around the start point   4 0 5
      with the given distance         6 7 8
      */
  AOT( uiDistance == 0 );
  const int iLeftX    = iCenterX - (int)uiDistance;
  const int iRightX   = iCenterX + (int)uiDistance;
  const int iTopY     = iCenterY - (int)uiDistance;
  const int iBottomY  = iCenterY + (int)uiDistance;
  rcSearchStruct.getBestRound()  += 1;

  if( iTopY >= rcSearchStruct.getMinPos().getY() )
  {
    if( iLeftX >= rcSearchStruct.getMinPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iLeftX,    iTopY,    uiDistance, 1 ); // top left
    }
    {
      xCheckInsidePos( rcSearchStruct, iCenterX,  iTopY,    uiDistance, 2 ); // top center
    }
    if( iRightX <= rcSearchStruct.getMaxPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iRightX,   iTopY,    uiDistance, 3 ); // top right
    }
  }
  {
    if( iLeftX >= rcSearchStruct.getMinPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iLeftX,    iCenterY, uiDistance, 4 ); // center left
    }
    if( iRightX <= rcSearchStruct.getMaxPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iRightX,   iCenterY, uiDistance, 5 ); // center right
    }
  }
  if( iBottomY <= rcSearchStruct.getMaxPos().getY() )
  {
    if( iLeftX >= rcSearchStruct.getMinPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iLeftX,    iBottomY, uiDistance, 6 ); // bottom left
    }
    {
      xCheckInsidePos( rcSearchStruct, iCenterX,  iBottomY, uiDistance, 7 ); // bottom center
    }
    if( iRightX <= rcSearchStruct.getMaxPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iRightX,   iBottomY, uiDistance, 8 ); // bottom right
    }
  }
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
void 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xDiamondSearch( SearchStruct<t_RefPel>&  rcSearchStruct,
                                                                                     int                      iCenterX,
                                                                                     int                      iCenterY,
                                                                                     UInt                     uiDistance )
{ /*  8 point search,                 1 2 3
      search around the start point   4 0 5
      with the given distance         6 7 8
      */
  AOT( uiDistance == 0 );
  const int iLeftX    = iCenterX - (int)uiDistance;
  const int iRightX   = iCenterX + (int)uiDistance;
  const int iTopY     = iCenterY - (int)uiDistance;
  const int iBottomY  = iCenterY + (int)uiDistance;
  rcSearchStruct.getBestRound()  += 1;

  if( uiDistance == 1 )
  {
    if( iTopY >= rcSearchStruct.getMinPos().getY() )
    {
      xCheckInsidePos( rcSearchStruct, iCenterX,  iTopY,    uiDistance, 2 );  // top center
    }
    if( iLeftX >= rcSearchStruct.getMinPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iLeftX,    iCenterY, uiDistance, 4 );  // center left
    }
    if( iRightX <= rcSearchStruct.getMaxPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iRightX,   iCenterY, uiDistance, 5 );  // center right
    }
    if( iBottomY <= rcSearchStruct.getMaxPos().getY() )
    {
      xCheckInsidePos( rcSearchStruct, iCenterX,  iBottomY, uiDistance, 7 );  // bottom center
    }
  }
  else if( uiDistance <= 8 )
  {
    const UInt  uiHalfDistance  = uiDistance >> 1;
    const int   iHalfTopY       = iCenterY - (int)uiHalfDistance;
    const int   iHalfBottomY    = iCenterY + (int)uiHalfDistance;
    const int   iHalfLeftX      = iCenterX - (int)uiHalfDistance;
    const int   iHalfRightX     = iCenterX + (int)uiHalfDistance;
    
    if( iTopY    >= rcSearchStruct.getMinPos().getX() && 
        iLeftX   >= rcSearchStruct.getMinPos().getX() &&
        iBottomY <= rcSearchStruct.getMaxPos().getY() && 
        iRightX  <= rcSearchStruct.getMaxPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iCenterX,    iTopY,        uiDistance,     2 );  // top center
      xCheckInsidePos( rcSearchStruct, iHalfLeftX,  iHalfTopY,    uiHalfDistance, 1 );  // half top left
      xCheckInsidePos( rcSearchStruct, iHalfRightX, iHalfTopY,    uiHalfDistance, 3 );  // half top right
      xCheckInsidePos( rcSearchStruct, iLeftX,      iCenterY,     uiDistance,     4 );  // center left
      xCheckInsidePos( rcSearchStruct, iRightX,     iCenterY,     uiDistance,     5 );  // center right
      xCheckInsidePos( rcSearchStruct, iHalfLeftX,  iHalfBottomY, uiHalfDistance, 6 );  // half bottom left
      xCheckInsidePos( rcSearchStruct, iHalfRightX, iHalfBottomY, uiHalfDistance, 8 );  // half bottom right
      xCheckInsidePos( rcSearchStruct, iCenterX,    iBottomY,     uiDistance,     7 );  // bottom center
    }
    else
    {
      if( iTopY >= rcSearchStruct.getMinPos().getY() )
      {
        {
          xCheckInsidePos( rcSearchStruct, iCenterX,    iTopY,        uiDistance,     2 );  // top center
        }
      }
      if( iHalfTopY >= rcSearchStruct.getMinPos().getY() )
      {
        if( iHalfLeftX >= rcSearchStruct.getMinPos().getX() )
        {
          xCheckInsidePos( rcSearchStruct, iHalfLeftX,  iHalfTopY,    uiHalfDistance, 1 );  // half top left
        }
        if( iHalfRightX <= rcSearchStruct.getMaxPos().getX() )
        {
          xCheckInsidePos( rcSearchStruct, iHalfRightX, iHalfTopY,    uiHalfDistance, 3 );  // half top right
        }
      }
      {
        if( iLeftX >= rcSearchStruct.getMinPos().getX() )
        {
          xCheckInsidePos( rcSearchStruct, iLeftX,      iCenterY,     uiDistance,     4 );  // center left
        }
        if( iRightX <= rcSearchStruct.getMaxPos().getX() )
        {
          xCheckInsidePos( rcSearchStruct, iRightX,     iCenterY,     uiDistance,     5 );  // center right
        }
      }
      if( iHalfBottomY <= rcSearchStruct.getMaxPos().getY() )
      {
        if( iHalfLeftX >= rcSearchStruct.getMinPos().getX() )
        {
          xCheckInsidePos( rcSearchStruct, iHalfLeftX,  iHalfBottomY, uiHalfDistance, 6 );  // half bottom left
        }
        if( iHalfRightX <= rcSearchStruct.getMaxPos().getX() )
        {
          xCheckInsidePos( rcSearchStruct, iHalfRightX, iHalfBottomY, uiHalfDistance, 8 );  // half bottom right
        }
      }
      if( iBottomY <= rcSearchStruct.getMaxPos().getY() )
      {
        {
          xCheckInsidePos( rcSearchStruct, iCenterX,    iBottomY,     uiDistance,     7 );  // bottom center
        }
      }
    }
  }
  else // uiDistance > 8
  {
    if( iTopY    >= rcSearchStruct.getMinPos().getY() &&
        iLeftX   >= rcSearchStruct.getMinPos().getX() &&
        iBottomY <= rcSearchStruct.getMaxPos().getY() &&
        iRightX  <= rcSearchStruct.getMaxPos().getX() )
    {
      xCheckInsidePos( rcSearchStruct, iCenterX,  iTopY,    uiDistance,  0 );  // top center
      xCheckInsidePos( rcSearchStruct, iLeftX,    iCenterY, uiDistance,  0 );  // center left
      xCheckInsidePos( rcSearchStruct, iRightX,   iCenterY, uiDistance,  0 );  // center right
      xCheckInsidePos( rcSearchStruct, iCenterX,  iBottomY, uiDistance,  0 );  // bottom center

      for( UInt uiIndex = 1; uiIndex < 4; uiIndex++ )
      {
        const int iTestDistance = int( uiDistance * uiIndex ) >> 2;
        const int iTestTopY     = iCenterY - iTestDistance;
        const int iTestBottomY  = iCenterY + iTestDistance;
        const int iTestLeftX    = iCenterX - iTestDistance;
        const int iTestRightX   = iCenterX + iTestDistance;
        xCheckInsidePos( rcSearchStruct, iTestLeftX,  iTestTopY,    uiDistance,  0 );  // test top left
        xCheckInsidePos( rcSearchStruct, iTestRightX, iTestTopY,    uiDistance,  0 );  // test top right
        xCheckInsidePos( rcSearchStruct, iTestLeftX,  iTestBottomY, uiDistance,  0 );  // test bottom left
        xCheckInsidePos( rcSearchStruct, iTestRightX, iTestBottomY, uiDistance,  0 );  // test bottom right
      }
    }
    else
    {
      if( iTopY >= rcSearchStruct.getMinPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iCenterX,  iTopY,    uiDistance,  0 );  // top center
      }
      if( iLeftX >= rcSearchStruct.getMinPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iLeftX,    iCenterY, uiDistance,  0 );  // center left
      }
      if( iRightX <= rcSearchStruct.getMaxPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iRightX,   iCenterY, uiDistance,  0 );  // center right
      }
      if( iBottomY <= rcSearchStruct.getMaxPos().getY()  )
      {
        xCheckInsidePos( rcSearchStruct, iCenterX,  iBottomY, uiDistance,  0 );  // bottom center
      }

      for( UInt uiIndex = 1; uiIndex < 4; uiIndex++ )
      {
        const int iTestDistance = int( uiDistance * uiIndex ) >> 2;
        const int iTestTopY     = iCenterY - iTestDistance;
        const int iTestBottomY  = iCenterY + iTestDistance;
        const int iTestLeftX    = iCenterX - iTestDistance;
        const int iTestRightX   = iCenterX + iTestDistance;

        if( iTestTopY >= rcSearchStruct.getMinPos().getY() )
        {
          if( iTestLeftX >= rcSearchStruct.getMinPos().getX() )
          {
            xCheckInsidePos( rcSearchStruct, iTestLeftX,  iTestTopY,    uiDistance,  0 );  // test top left
          }
          if( iTestRightX <= rcSearchStruct.getMaxPos().getX() )
          {
            xCheckInsidePos( rcSearchStruct, iTestRightX, iTestTopY,    uiDistance,  0 );  // test top right
          }
        }
        if( iTestBottomY <= rcSearchStruct.getMaxPos().getY() )
        {
          if( iTestLeftX >= rcSearchStruct.getMinPos().getX() )
          {
            xCheckInsidePos( rcSearchStruct, iTestLeftX,  iTestBottomY, uiDistance,  0 );  // test bottom left
          }
          if( iTestRightX <= rcSearchStruct.getMaxPos().getX() )
          {
            xCheckInsidePos( rcSearchStruct, iTestRightX, iTestBottomY, uiDistance,  0 );  // test bottom right
          }
        }
      }
    }
  }
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
void 
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xTwoPointSearch( SearchStruct<t_RefPel>& rcSearchStruct )
{ /*  2 point search,                 1 2 3
      search around the start point   4 0 5
      given by stored point number    6 7 8
      */
  const int   iStartX     = rcSearchStruct.getBestPos().getX();
  const int   iStartY     = rcSearchStruct.getBestPos().getY();
  const UInt  uiDistance  = 2;
  const UInt  uiPoint     = 0;
  
  switch( rcSearchStruct.getBestPoint() )
  {
  case 1:
    {
      if( iStartX > rcSearchStruct.getMinPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX - 1, iStartY,      uiDistance, uiPoint );
      }
      if( iStartY > rcSearchStruct.getMinPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX,     iStartY - 1,  uiDistance, uiPoint );
      }
    }
    break;
  case 2:
    if( iStartY > rcSearchStruct.getMinPos().getY() )
    {
      if( iStartX > rcSearchStruct.getMinPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX - 1, iStartY - 1,  uiDistance, uiPoint );
      }
      if( iStartX < rcSearchStruct.getMaxPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX + 1, iStartY - 1, uiDistance, uiPoint );
      }
    }
    break;
  case 3:
    {
      if( iStartY > rcSearchStruct.getMinPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX,     iStartY - 1,  uiDistance, uiPoint );
      }
      if( iStartX < rcSearchStruct.getMaxPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX + 1, iStartY,      uiDistance, uiPoint );
      }
    }
    break;
  case 4:
    if( iStartX > rcSearchStruct.getMinPos().getX() )
    {
      if( iStartY < rcSearchStruct.getMaxPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX - 1, iStartY + 1,  uiDistance, uiPoint );
      }
      if( iStartY > rcSearchStruct.getMinPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX - 1, iStartY - 1,  uiDistance, uiPoint );
      }
    }
    break;
  case 5:
    if( iStartX < rcSearchStruct.getMaxPos().getX() )
    {
      if( iStartY > rcSearchStruct.getMinPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX + 1, iStartY - 1,  uiDistance, uiPoint );
      }
      if( iStartY < rcSearchStruct.getMaxPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX + 1, iStartY + 1,  uiDistance, uiPoint );
      }
    }
    break;
  case 6:
    {
      if( iStartX > rcSearchStruct.getMinPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX - 1, iStartY,      uiDistance, uiPoint );
      }
      if( iStartY < rcSearchStruct.getMaxPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX,     iStartY + 1,  uiDistance, uiPoint );
      }
    }
    break;
  case 7:
    if( iStartY < rcSearchStruct.getMaxPos().getY() )
    {
      if( iStartX > rcSearchStruct.getMinPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX - 1, iStartY + 1,  uiDistance, uiPoint );
      }
      if( iStartX < rcSearchStruct.getMaxPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX + 1, iStartY + 1,  uiDistance, uiPoint );
      }
    }
    break;
  case 8:
    {
      if( iStartX < rcSearchStruct.getMaxPos().getX() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX + 1, iStartY,      uiDistance, uiPoint );
      }
      if( iStartY < rcSearchStruct.getMaxPos().getY() )
      {
        xCheckInsidePos( rcSearchStruct, iStartX,     iStartY + 1,  uiDistance, uiPoint );
      }
    }
    break;
  default:
    AOT( 1 );
    break;
  }
}


template<class t_OrigPel,DistortionMeasure t_eDistMeasure,bool bTriangleUneq,class t_MvMeasure>
template<class t_RefPel>
ErrVal
FastLossySearch<t_OrigPel,t_eDistMeasure,bTriangleUneq,t_MvMeasure>::xFullPelSearch( SearchStruct<t_RefPel>&          rcSearchStruct, 
                                                                                     const std::vector<MotionVector>& rcCandMvList )
{
  ROT( rcCandMvList.empty() );
  ROT( rcSearchStruct.pointFound() );

  //===== get search range =====
  const PrimPixelPos  cDeltaMinMax  = rcSearchStruct.getMaxPos() - rcSearchStruct.getMinPos();
  const UInt          uiSearchRange = ( (UInt)gMin( cDeltaMinMax.getX(), cDeltaMinMax.getY() ) + 1 ) >> 1;

  //===== check predictors =====
  for( UInt uiCandIdx = 0; uiCandIdx < rcCandMvList.size(); uiCandIdx++ )
  {
    const MotionVector& rcCand  = rcCandMvList[ uiCandIdx ];
    ROF( rcCand.getPrecision() == 0 );
    xCheckPos( rcSearchStruct, rcCand.getX(), rcCand.getY(), 0, 0 );
  }
  ROF( rcSearchStruct.pointFound() );

  //===== start search =====
  const int iCenterX = rcSearchStruct.getBestPos().getX();
  const int iCenterY = rcSearchStruct.getBestPos().getY();
  for( UInt uiDistance = 1; uiDistance <= uiSearchRange; uiDistance <<= 1 )
  {
    if( m_bStartWithDiamondSearch )
    {
      xDiamondSearch( rcSearchStruct, iCenterX, iCenterY, uiDistance );
    }
    else
    {
      xSquareSearch ( rcSearchStruct, iCenterX, iCenterY, uiDistance );
    }
    if( rcSearchStruct.getBestRound() >= m_uiMaxFirstSearchRounds )
    {
      break;
    }
  }

  //===== check 2 points if best distance is equal to 1 =====
  if( rcSearchStruct.getBestDistance() == 1 )
  {
    rcSearchStruct.getBestDistance() = 0;
    xTwoPointSearch( rcSearchStruct );
  }

  //===== raster search if best distance is too large =====
  if( m_bAlwaysRasterSearch || ( m_bEnableRasterSearch && rcSearchStruct.getBestDistance() > m_uiRasterDistance ) )
  {
    rcSearchStruct.getBestDistance() = m_uiRasterDistance;
    for( int iTestY = rcSearchStruct.getMinPos().getY(); iTestY <= rcSearchStruct.getMaxPos().getY(); iTestY += (int)m_uiRasterDistance )
    for( int iTestX = rcSearchStruct.getMinPos().getX(); iTestX <= rcSearchStruct.getMaxPos().getX(); iTestX += (int)m_uiRasterDistance )
    {
      xCheckInsidePos( rcSearchStruct, iTestX, iTestY, m_uiRasterDistance, 0 );
    }
  }

  //===== raster refinement =====
  if( m_bEnableRasterRefinement && rcSearchStruct.getBestDistance() > 0 )
  {
    while( rcSearchStruct.getBestDistance() > 0 )
    {
      const int iCenterX = rcSearchStruct.getBestPos().getX();
      const int iCenterY = rcSearchStruct.getBestPos().getY();
      if( rcSearchStruct.getBestDistance() > 1 )
      {
        rcSearchStruct.getBestDistance() >>= 1;
        if( m_bRasterRefinementDiamond == 1 )
        {
          xDiamondSearch( rcSearchStruct, iCenterX, iCenterY, rcSearchStruct.getBestDistance() );
        }
        else
        {
          xSquareSearch ( rcSearchStruct, iCenterX, iCenterY, rcSearchStruct.getBestDistance() );
        }
      }
      if( rcSearchStruct.getBestDistance() == 1 )
      {
        rcSearchStruct.getBestDistance() = 0;
        if( rcSearchStruct.getBestPoint() != 0 )
        {
          xTwoPointSearch( rcSearchStruct );
        }
      }
    }
  }

  //===== star refinement =====
  if( m_bEnableStarRefinement && rcSearchStruct.getBestDistance() > 0 )
  {
    while( rcSearchStruct.getBestDistance() > 0 )
    {
      const int iCenterX                = rcSearchStruct.getBestPos().getX();
      const int iCenterY                = rcSearchStruct.getBestPos().getY();
      rcSearchStruct.getBestDistance()  = 0;
      rcSearchStruct.getBestPoint   ()  = 0;
      for( UInt uiDistance = 1; uiDistance <= uiSearchRange; uiDistance <<= 1 )
      {
        if( m_bStarRefinementDiamond )
        {
          xDiamondSearch( rcSearchStruct, iCenterX, iCenterY, uiDistance );
        }
        else
        {
          xSquareSearch ( rcSearchStruct, iCenterX, iCenterY, uiDistance );
        }
        if( rcSearchStruct.getBestRound() >= m_uiMaxStarRefinementRounds )
        {
          break;
        }
      }
      if( rcSearchStruct.getBestDistance() == 1 )
      {
        rcSearchStruct.getBestDistance() = 0;
        if( rcSearchStruct.getBestPoint() != 0 )
        {
          xTwoPointSearch( rcSearchStruct );
        }
      }
    }
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _FASTSEARCH_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
