/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_NALUNITENCODER_H__DA2EE2CC_46F5_4F11_B046_FA18CD441B65__INCLUDED_)
#define AFX_NALUNITENCODER_H__DA2EE2CC_46F5_4F11_B046_FA18CD441B65__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VideoCommonLib/BitWriterIf.h"

#include "VideoCommonLib/H264SliceHeader.h"
#include "VideoCommonLib/H264SequenceParameterSet.h"
#include "VideoCommonLib/H264PictureParameterSet.h"

// h264 namespace begin
VIDEO_NAMESPACE_BEGIN

class TraceCommon;
typedef BitWriterIf<UInt> MyBitWriterIf;
typedef std::vector<UChar> NalUnit;


//!\ingroup H264EncoderLib
//!\NalUnitEncoder
class NalUnitEncoder
{
public:
	NalUnitEncoder();
	virtual ~NalUnitEncoder();
  static ErrVal create( NalUnitEncoder*& rpcNalUnitEncoder );
  ErrVal destroy();
  ErrVal init( MyBitWriterIf* pcBitWriteBuffer, HeaderSymbolWriteIf* pcHeaderSymbolWriteIf, HeaderSymbolWriteIf* pcHeaderSymbolTestIf = NULL, bool bByteStream = false );
  ErrVal uninit();
  MyBitWriterIf* getBitBuffer() { return m_pcBitWriteBuffer; }
  ErrVal initNalUnit( );
  ErrVal closeNalUnit( std::list< std::vector<UChar> >& rcBitstreamInOut );
  ErrVal writeSPS( const SequenceParameterSet& rcSPS, std::vector<UChar>& rcNalUnit, TraceCommon* pcTrace );
  ErrVal writePPS( const PictureParameterSet& rcPPS, std::vector<UChar>& rcNalUnit, TraceCommon* pcTrace );
  ErrVal writeSliceHeader( const SliceHeader& rcSH );
  bool   isByteStream()   { return m_bByteStream; }

  ErrVal appendCabacZeroWords( NalUnit* pcNalUnit, int iMbInPic, int iBytesInAU, int iUsedCabacBins, UInt& ruiAddBytes );

private:
  ErrVal closeNalUnit( std::vector<UChar>& rcNalUnit, std::vector<UChar>& cInBuffer );

  ErrVal xConvertRBSPToPayload( std::vector<UChar>& rcNalUnit, std::list<std::vector<UChar> >& rcBitStreamIn );
  ErrVal xWriteTrailingBits( UInt uiFixedNumberOfBits = 0);

protected:
  MyBitWriterIf* m_pcBitWriteBuffer;
  HeaderSymbolWriteIf*  m_pcHeaderSymbolWriteIf;
  HeaderSymbolWriteIf*  m_pcHeaderSymbolTestIf;
  std::vector<UChar> m_cTempBuffer;
  NalUnitType m_eNalUnitType;
  NalRefIdc   m_eNalRefIdc;
  UInt m_uiBitsWritten;
  bool m_bIsUnitActive;
  bool m_bByteStream;
  bool m_bTraceEnable;
};

// h264 namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_NALUNITENCODER_H__DA2EE2CC_46F5_4F11_B046_FA18CD441B65__INCLUDED_)
