/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTERTRANSLATIONALESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTERTRANSLATIONALESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "FastFullSearch.h"
#include "FastLossySearch.h"
#include "PredictionEstimator.h"
#include "VideoCommonLib/InterTranslational.h"
#include "VideoCommonLib/PlanePackage.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif

class EstInfoInterTranslational
  : public EstInfo
{
public:
  EstInfoInterTranslational( UInt uiMvAccuracy, BlockSize cSearchRange, BlockSize cIterSearchRange, UInt uiSearchModeFullPel, UInt uiSearchModeSubPel, bool bSearchFirstPlaneOnly, const Lambda &rcMotionLambda, UInt uiMaxNumRefBlocksPerPred, UInt uiMaxNumIterations )
    : m_uiMvAccuracy( uiMvAccuracy )
    , m_cSearchRange( cSearchRange )
    , m_cIterSearchRange( cIterSearchRange )
    , m_uiSearchModeFullPel( uiSearchModeFullPel )
    , m_uiSearchModeSubPel( uiSearchModeSubPel )
    , m_bSearchFirstPlaneOnly( bSearchFirstPlaneOnly )
    , m_cMotionLambda( rcMotionLambda )
    , m_uiMaxNumRefBlocksPerPred( uiMaxNumRefBlocksPerPred )
    , m_uiMaxNumIterations( uiMaxNumIterations ) {}

  const EstInfoInterTranslational* getEstInfoInterTranslational() const { return this; }
  PredInfo::PredictionMode    getPredictionMode() const { return PredInfo::INTER_TRANSLATIONAL; }

  UInt getMvAccuracy() const { return m_uiMvAccuracy; }
  BlockSize getSearchRange() const { return m_cSearchRange; }
  BlockSize getIterSearchRange() const { return m_cIterSearchRange; }
  UInt getSearchModeFullPel() const { return m_uiSearchModeFullPel; }
  UInt getSearchModeSubPel() const { return m_uiSearchModeSubPel; }
  bool searchFirstPlaneOnly() const { return m_bSearchFirstPlaneOnly; }
  const Lambda& getMotionLambda() const { return m_cMotionLambda; }
  UInt getMaxNumRefBlocksPerPred() const { return m_uiMaxNumRefBlocksPerPred; }
  UInt getMaxNumIterations() const { return m_uiMaxNumIterations; }

private:
  UInt m_uiMvAccuracy;
  BlockSize m_cSearchRange;
  BlockSize m_cIterSearchRange;
  UInt m_uiSearchModeFullPel;
  UInt m_uiSearchModeSubPel;
  bool m_bSearchFirstPlaneOnly;
  Lambda m_cMotionLambda;
  UInt m_uiMaxNumRefBlocksPerPred;
  UInt m_uiMaxNumIterations;
};


class InterTranslationalEstimator
  : public PredictionEstimator
  , public InterTranslational
{
public:
  InterTranslationalEstimator( UInt uiSearchModeFullPel );

  typedef int OrigPel;

  UInt64 getMEDuration() const { return m_uiDuration; }
  double getTriangleInequalityEliminationRatio() const { return m_pcMotionSearchIf->getTriangleInequalityEliminationRatio(); }

  virtual ~InterTranslationalEstimator()
  {
    delete m_pcMotionSearchIf;
  }

  PredInfo::PredictionMode getPredictionMode() const { return PredInfo::INTER_TRANSLATIONAL; }

  ErrVal estimateAndCalculatePrediction( PredInfo *&rpcPredInfo,
                                         const QTNavi &rcNav,
                                         const EstInfo &rcEstInfo,
                                         const CodingState &rcCodingState,
                                         const AccessUnitParameters &rcAUPars,
                                         const PlanePackageParameters &rcPlaPaPars,
                                         const SliceData &rcCurrSlice,
                                         const PicBlock<const Pel> &rcOrigPicBlock,
                                         const PicBlock<const Pel> *pcEncPic,
                                         const PicBlock<DPel> &rcPredPicBlock,
                                         bool &rbPredictionValid );

  ErrVal reportRDCost( const QTNavi &rcNav,
                       RDCost cRDCost,
                       bool &rbEstimatePredictionAgain )
  {
    ROF( rcNav == m_cNav );
    if( !cRDCost.isMax() && cRDCost < m_cBestRDCost )
    {
      m_cBestPredInfo = m_cCurrPredInfo;
      m_cBestRDCost = cRDCost;
    }
//    const bool bNoPredictionFoundForCurrentNumberOfHypotheses = m_apcPredInfo.empty() && m_cBestRDCost.isMax();
//    rbEstimatePredictionAgain = !bNoPredictionFoundForCurrentNumberOfHypotheses && (!m_apcPredInfo.empty() || m_uiCurrNumPred < m_uiMaxNumRefBlocksPerPred);
    m_cCurrPredInfo.reset();
    rbEstimatePredictionAgain = !m_apcPredInfo.empty() || m_cBestPredInfo.getNumMv() < m_uiMaxNumRefBlocksPerPred;
    if( !rbEstimatePredictionAgain )
    {
      m_uiMaxNumRefBlocksPerPred = 0;
      m_cBestPredInfo.reset();
    }
    return Err::m_nOK;
  }


private:
  static ErrVal xGetMvCandVector( PicOrderCnt cRefPoc,
                                  const QTNavi &rcNav,
                                  const QuadtreeGroup<PredInfo*> &rcTree,
                                  const MotionVector& rcMvPred,
                                  std::vector<MotionVector>& rcMvCandList );

  static ErrVal xSearchOneAdditionalHypothesis( const QTNavi &rcNav,
                                                MotionSearchIf<OrigPel> &rcMotionSearchIf,
                                                InterTranslational &rcPredictionCalculator,
                                                const PredSyntaxInterTranslational &rcPredInfoToExtend,
                                                const EstInfoInterTranslational &rcEstInfo,
                                                const CodingState &rcCodingState,
                                                const PlanePackageParameters &rcPlaPaPars,
                                                const AccessUnitParameters &rcAUPars,
                                                const SliceData &rcCurrSlice,
                                                const PicBlock<const Pel> &rcOrigPicBlock,
                                                std::vector<PredSyntaxInterTranslational*> &rapcPredInfo,
                                                bool &rbPredictionValid );

  static ErrVal xRefineMVs( const QTNavi &rcNav,
                            MotionSearchIf<OrigPel> &rcMotionSearchIf,
                            InterTranslational &rcPredictionCalculator,
                            const PlanePackageParameters& rcPlaPaPars,
                            const EstInfoInterTranslational &rcEstInfo,
                            PicBlock<int> &rcResidualCurr,
                            const std::vector<const PicBlock<DPel>*> &rapcPreds,
                            const CodingState &rcCodingState,
                            const SliceData &rcCurrSlice,
                            PredInfoInterTranslational &rcPredInfo,
                            Lambda cLambda );

  QTNavi m_cNav;
  RDCost m_cBestRDCost;
  PredSyntaxInterTranslational m_cBestPredInfo;
  PredSyntaxInterTranslational m_cCurrPredInfo;
  std::vector<PredSyntaxInterTranslational*> m_apcPredInfo;
  UInt m_uiMaxNumRefBlocksPerPred;
  MotionSearchIf<OrigPel> *m_pcMotionSearchIf;
  UInt64 m_uiDuration;
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTERTRANSLATIONALESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
