/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _INTRADIRECTIONALESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _INTRADIRECTIONALESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "PredictionEstimator.h"
#include "VideoCommonLib/IntraDirectional.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif


class EstInfoIntraDirectional
  : public EstInfo
{
public:
  enum SmoothingEstimationMethod
  {
    SMOOTH_PREDICTION,
    DO_NOT_SMOOTH_PREDICTION,
    REESTIMATE_BEST_WITH_TOGGLED_SMOOTHING,
  };

  const EstInfoIntraDirectional* getEstInfoIntraDirectional() const { return this; }

  EstInfoIntraDirectional() : m_cPrSyntax( true ) {}
  EstInfoIntraDirectional( PredInfoIntraDirectional::IntraDirections eDir, SmoothingEstimationMethod eSmoothingMethod )
    : m_cPrSyntax        ( true, eDir )
    , m_eSmoothingMethod ( eSmoothingMethod )
  {}

  PredInfo::PredictionMode          getPredictionMode           ()  const { return PredInfo::INTRA_DIRECTIONAL; }
  const PredSyntaxIntraDirectional& getPredInfo                 ()  const { return m_cPrSyntax; }
  SmoothingEstimationMethod         getSmoothingEstimationMethod()  const { return m_eSmoothingMethod; }

private:
  PredSyntaxIntraDirectional m_cPrSyntax;
  SmoothingEstimationMethod  m_eSmoothingMethod;
};


class IntraDirectionalEstimator
  : public PredictionEstimator
  , public IntraDirectional
{
public:
  IntraDirectionalEstimator() : m_cNav( NULL ), m_cCurr( true ), m_cBest( true ) { m_cMinCost.setMax(); }

  PredInfo::PredictionMode getPredictionMode() const { return PredInfo::INTRA_DIRECTIONAL; }

  ErrVal estimateAndCalculatePrediction( PredInfo *&rpcPredInfo,
                                         const QTNavi &rcNav,
                                         const EstInfo &rcEstInfo,
                                         const CodingState &rcCodingState,
                                         const AccessUnitParameters &rcAUPars,
                                         const PlanePackageParameters &rcPlaPaPars,
                                         const SliceData &rcCurrSlice,
                                         const PicBlock<const Pel> &rcOrigPicBlock,
                                         const PicBlock<const Pel> *pcEncPic,
                                         const PicBlock<DPel> &rcPredPicBlock,
                                         bool &rbPredictionValid );
  ErrVal reportRDCost( const QTNavi &rcNav,
                       RDCost cRDCost,
                       bool &rbEstimatePredictionAgain )
  {
    if( m_cNav != rcNav )
    {
      // new block
      m_cMinCost.setMax();
      m_cNav = rcNav;
    }
    
    if( !cRDCost.isMax() && cRDCost < m_cMinCost )
    {
      m_cMinCost = cRDCost;
      m_cBest = m_cCurr;
    }
    rbEstimatePredictionAgain = false;
    return Err::m_nOK;
  }

private:
  RDCost m_cMinCost;
  QTNavi m_cNav;
  PredSyntaxIntraDirectional m_cCurr;
  PredSyntaxIntraDirectional m_cBest;
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _INTRADIRECTIONALESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
