/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PREDESTANDENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PREDESTANDENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "PredEstAndEncIf.h"
#include "ResEstInfoVecSrcIf.h"
#include "PredictionEstimator.h"
#include "SyntaxElementEncoderIf.h"
#include "PredictionEstimatorManager.h"
#include "AccessUnitEncoderParameters.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class PredInfoAndPredEstInfo
{
public:
  PredInfoAndPredEstInfo() : m_pcPredEstInfo( NULL ), m_pcPredInfo( NULL ), m_pcVecResEstInfo( NULL ) {}
  virtual ~PredInfoAndPredEstInfo() { delete m_pcPredInfo; }

  const EstInfo *m_pcPredEstInfo;
  PredInfo *m_pcPredInfo;
  const VecResEstInfo *m_pcVecResEstInfo;
};


class PredEstAndEnc
  : public PredEstAndEncIf<PredInfoAndPredEstInfo>
  , public ResEstInfoVecSrcIf
{
public:
  typedef PredInfoAndPredEstInfo PredIf;

  PredEstAndEnc( const PlanePackageParameters &rcPlaPaPars,
                 const AccessUnitParameters &rcAUPars,
                 const AccessUnitEncoderParameters& rcAUEstPars,
                 const VecPredEstInfo &rcEstPars,
                 SyntaxElementEncoderIf *pcSEEncoderIf,
                 const PicBlock<const Pel> &rcOrigPic,
                 const PicBlock<const Pel> *pcEncPic,
                 const CodingState &rcCodingState,
                 SliceData &rcSlice )
    : m_rcPlaPaPars( rcPlaPaPars )
    , m_rcAUPars( rcAUPars )
    , m_rcEstPars( rcEstPars )
    , m_pcSEEncoderIf( pcSEEncoderIf )
    , m_rcOrigPic( rcOrigPic )
    , m_pcEncPic( pcEncPic )
    , m_rcCodingState( rcCodingState )
    , m_rcSlice( rcSlice )
    , m_cPredMng( rcAUEstPars.getSearchModeFullPel() )
    , m_cActivePredNav( NULL )
  {}

  virtual ~PredEstAndEnc() {}

  ErrVal resetPredictionEstimationForBlock( const QTNavi &rcNav );
  ErrVal estimateAndEncodePrediction( const QTNavi &rcNav,
                                      PredInfoAndPredEstInfo *&rpResult,
                                      const PredInfo *&rpcEncodedPred,
                                      PicBlock<DPel> &rcPredPicBlock,
                                      bool &rbPredictionValidAndEncoded );
  ErrVal reportRDCost( const QTNavi &rcNav,
                       RDCost cCost,
                       bool &rbLastPredictionToTest );

  ErrVal proceedToNextPrediction( const QTNavi &rcNav );
  ErrVal setBestPrediction( const QTNavi &rcNav, PredInfoAndPredEstInfo *&rpResult );
  ErrVal releaseChildNodes( const QTNavi &rcNav );
  ErrVal releaseCurrentNode( const QTNavi &rcNav );

  UInt64 getTotalMEDuration() const { return m_cPredMng.getTotalMEDuration(); }
  double getTriangleInequalityEliminationRatio() const { return m_cPredMng.getTriangleInequalityEliminationRatio(); }

  ResEstInfoVecSrcIf& getResEstInfoVecSrcIf() { return *this; }

  const VecResEstInfo& getResEstInfoVec( const QTNavi &rcNav );

  bool currModeIsIntra() const
  {
    TOF( m_iActivePredEstInfo < int( m_rcEstPars.size() ) );
    const EstInfo *pcEstPars = m_rcEstPars[m_iActivePredEstInfo].getPtr();
    TOF( pcEstPars );
    bool bIntra = false;
    switch( pcEstPars->getPredictionMode() )
    {
    case PredInfo::INTRA_DC:
    case PredInfo::INTRA_DIRECTIONAL:
    case PredInfo::INTRA_PLANE:
      bIntra= true;
      break;
    default:
      break;
    };
    return bIntra;
  }

private:
  const PlanePackageParameters &m_rcPlaPaPars;
  const AccessUnitParameters &m_rcAUPars;
  const VecPredEstInfo &m_rcEstPars;
  SyntaxElementEncoderIf *m_pcSEEncoderIf;
  const PicBlock<const Pel> &m_rcOrigPic;
  const PicBlock<const Pel> *m_pcEncPic;
  const CodingState &m_rcCodingState;
  SliceData &m_rcSlice;
  PredictionEstimatorManager m_cPredMng;
  QTNavi m_cActivePredNav;
  bool m_bEstPredAgain;
  int m_iActivePredEstInfo;
};


// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _PREDESTANDENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
