/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PREDICTIONESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PREDICTIONESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/QuadtreeGroup.h"
#include "VideoCommonLib/PicBlock.h"
#include "VideoCommonLib/PredictionCalculator.h"
#include "ResidualEstimator.h"
#include "RDCost.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4251 )
#endif

class EstInfoInterTranslational;
class EstInfoIntraDirectional;
class EstInfoInterPlanePackage;

class EstInfo
{
public:
  virtual PredInfo::PredictionMode getPredictionMode() const = 0;

  virtual const EstInfoInterTranslational*  getEstInfoInterTranslational()  const { return NULL; }
  virtual const EstInfoIntraDirectional*  getEstInfoIntraDirectional()  const { return NULL; }
  virtual const EstInfoInterPlanePackage* getEstInfoInterPlanePackage() const { return NULL; }

  void setVecResEstInfo( const VecResEstInfo* pcVecResEstInfo ) { m_pcVecResEstInfo = pcVecResEstInfo; }
  const VecResEstInfo& getVecResEstInfo() const { return *m_pcVecResEstInfo; }
  virtual ~EstInfo() {}

private:
  const VecResEstInfo* m_pcVecResEstInfo;

protected:
  EstInfo() {}
};

//typedef std::vector<const EstInfo*> VecPredEstInfo;
typedef std::vector<Ptr<EstInfo> > VecPredEstInfo;

class PredictionEstimator
  : public virtual PredictionCalculator
{
public:
  virtual ~PredictionEstimator() {}

  virtual ErrVal estimateAndCalculatePrediction( PredInfo *&rpcPredInfo,
                                                 const QTNavi &rcNav,
                                                 const EstInfo &rcEstInfo,
                                                 const CodingState &rcCodingState,
                                                 const AccessUnitParameters &rcAUPars,
                                                 const PlanePackageParameters &rcPlaPaPars,
                                                 const SliceData &rcCurrSlice,
                                                 const PicBlock<const Pel> &rcOrigPicBlock,
                                                 const PicBlock<const Pel> *pcEncPic,
                                                 const PicBlock<DPel> &rcPredPicBlock,
                                                 bool &rbPredictionValid ) = 0;
  virtual ErrVal reportRDCost( const QTNavi &rcNav,
                               RDCost cRDCost,
                               bool &rbEstimatePredictionAgain ) { rbEstimatePredictionAgain = false; return Err::m_nOK; }

};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4251 )
#endif


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _PREDICTIONESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
