/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _PROBABILITYALLOCATORENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _PROBABILITYALLOCATORENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/ProbabilityAllocator.h"
#include "VideoCommonLib/CabacContextModel.h"
#include "VideoCommonLib/BinStreamIf.h"
#include "SyntaxElementEncoder.h"
#include "StateIf.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef BinStreamIf<ProbBin&, true> ProbBinStreamEncIf;

template<class t_ProbBinProcessor = BinStreamProcessor<ProbBin&, true>, class t_ContextModel = CabacContextModel>
class ProbabilityAllocatorEnc
  : public ProbabilityAllocator<t_ContextModel>
  , private CtxBinStreamEncIf
  , private StateIf
  , private t_ProbBinProcessor
{
public:
  typedef ProbabilityAllocatorEnc TypeCall;
  typedef ProbabilityAllocatorEnc BoolCall;

  struct CtorPars
  {
    UInt m_uiNumCtx;
    TraceCommon *m_pcTrace;
    typename t_ProbBinProcessor::CtorPars &m_rcProbBinProcessorPars;
  };

  ProbabilityAllocatorEnc( CtorPars &rcPars )
    : ProbabilityAllocator<t_ContextModel>( rcPars.m_uiNumCtx, rcPars.m_pcTrace )
    , t_ProbBinProcessor( rcPars.m_rcProbBinProcessorPars ) {}

  virtual ~ProbabilityAllocatorEnc() {}

  FORCEINLINE ErrVal pushBack( const CtxBin &rcT );
  FORCEINLINE ErrVal pushBack( const EPCtxBin &rcT );

  ErrVal getState( State *&rpcState );

  StateIf&           getStateIf()        { return *this; }
  CtxBinStreamEncIf& getInStreamCtxBin() { return *this; }

  t_ProbBinProcessor& getProbBinProcessor() { return *this; }

private:

  class ProbAllocatorState
    : public State
  {
  public:
    ProbAllocatorState( ProbabilityAllocatorEnc *pcStatesObject ) : m_pcStatesObject( pcStatesObject ) {}

    t_ContextModel *m_pacContextStored;

#if COLLECT_CABAC_BITS
    UInt m_uiBinIndexStored;
#endif

  private:
    ProbabilityAllocatorEnc *m_pcStatesObject;

    ErrVal setState();
    ErrVal deleteState();
    ErrVal setAndDeleteState();
  };

  ErrVal xGetNewProbAllocatorState( ProbAllocatorState *&rpcState )
  {
    rpcState = new ProbAllocatorState( this );
    rpcState->m_pacContextStored = new t_ContextModel[ProbabilityAllocator<t_ContextModel>::m_uiNumCtx];
    return Err::m_nOK;
  }

};


template<class t_ProbBinProcessor, class t_ContextModel>
FORCEINLINE ErrVal ProbabilityAllocatorEnc<t_ProbBinProcessor, t_ContextModel>::pushBack( const CtxBin &rcT )
{
  t_ContextModel &rcCCModel = ProbabilityAllocator<t_ContextModel>::m_pacContext[rcT.getCtx()];

  bool bIsLps = rcT.getBin() != ( 0 != rcCCModel.getMps());

  const ProbBin cProbBin( rcCCModel.getCabacState(), bIsLps );
  t_ProbBinProcessor::TypeCall::pushBack( cProbBin );

#if TRACE_PROB_BINS
  TRACE_NUMBER;
  TRACE_TO_STR( "Bin: ");
  TRACE_TO_STR( rcT.getBin() );
  TRACE_TO_STR( " Ctx: ");
  TRACE_TO_STR( rcT.getCtx() );
  TRACE_TO_STR( "Prob: " );
  TRACE_TO_STR( (UInt)rcCCModel.getCabacState() );
  TRACE_TO_STR( " isMps: " );
  TRACE_TO_STR( rcCCModel.getMps() ? 1 : 0 );
  TRACE_TO_STR( std::endl );
#endif

#if COLLECT_CABAC_BITS
  rcCCModel.addBits( cProbBin.getProbBinRate() );
  rcCCModel.addBin( rcT.getBin(), ProbabilityAllocator<t_ContextModel>::m_uiBinIndex++ );
#endif

  rcCCModel.updateCtx( bIsLps );
  return Err::m_nOK;
}


template<class t_ProbBinProcessor, class t_ContextModel>
FORCEINLINE ErrVal ProbabilityAllocatorEnc<t_ProbBinProcessor, t_ContextModel>::pushBack( const EPCtxBin &rcT )
{
#if COLLECT_CABAC_BITS
  t_ProbBinProcessor::BoolCall::pushBack( rcT.m_bBin );
  t_ContextModel &rcCCModel = ProbabilityAllocator<t_ContextModel>::m_pacContext[rcT.m_uiCtx];
  rcCCModel.addBits( 1.0 );
  rcCCModel.addBin( rcT.m_bBin, ProbabilityAllocator<t_ContextModel>::m_uiBinIndex++ );
#else
  t_ProbBinProcessor::BoolCall::pushBack( rcT );
#endif

#if TRACE_PROB_BINS
  TRACE_NUMBER;
  TRACE_TO_STR( "Bin: ");
  TRACE_TO_STR( rcT );
  TRACE_TO_STR( " EqProb");
  TRACE_TO_STR( std::endl );
#endif
  return Err::m_nOK;
}


template<class t_ProbBinProcessor, class t_ContextModel>
ErrVal ProbabilityAllocatorEnc<t_ProbBinProcessor, t_ContextModel>::getState( State *&rpcState )
{
  ROF(( TypeCompare<t_ContextModel,CabacContextModel>::m_bEqual )); // memcpy will not work properly due to std::vector<> in t_ContextModel

  ProbAllocatorState *pcPAState;
  RNOK( xGetNewProbAllocatorState( pcPAState ) );
  ::memcpy( pcPAState->m_pacContextStored, ProbabilityAllocator<t_ContextModel>::m_pacContext, ProbabilityAllocator<t_ContextModel>::m_uiNumCtx * sizeof( t_ContextModel ) );

#if COLLECT_CABAC_BITS
  pcPAState->m_uiBinIndexStored = ProbabilityAllocator<t_ContextModel>::m_uiBinIndex;
#endif

  rpcState = pcPAState;
  return Err::m_nOK;
}


template<class t_ProbBinProcessor, class t_ContextModel>
ErrVal ProbabilityAllocatorEnc<t_ProbBinProcessor, t_ContextModel>::ProbAllocatorState::setState()
{
  ROF(( TypeCompare<t_ContextModel,CabacContextModel>::m_bEqual )); // memcpy will not work properly due to std::vector<> in t_ContextModel
  ::memcpy( m_pcStatesObject->m_pacContext, m_pacContextStored, m_pcStatesObject->m_uiNumCtx * sizeof( t_ContextModel ) );

#if COLLECT_CABAC_BITS
  m_pcStatesObject->m_uiBinIndex = m_uiBinIndexStored;
#endif
  return Err::m_nOK;
}


template<class t_ProbBinProcessor, class t_ContextModel>
ErrVal ProbabilityAllocatorEnc<t_ProbBinProcessor, t_ContextModel>::ProbAllocatorState::deleteState()
{
  ROF(( TypeCompare<t_ContextModel,CabacContextModel>::m_bEqual )); // memcpy will not work properly due to std::vector<> in t_ContextModel

  delete[] m_pacContextStored;
  delete this;

  return Err::m_nOK;
}


template<class t_ProbBinProcessor, class t_ContextModel>
ErrVal ProbabilityAllocatorEnc<t_ProbBinProcessor, t_ContextModel>::ProbAllocatorState::setAndDeleteState()
{
  ROF(( TypeCompare<t_ContextModel,CabacContextModel>::m_bEqual )); // memcpy will not work properly due to std::vector<> in t_ContextModel

  delete[] m_pcStatesObject->m_pacContext;
  m_pcStatesObject->m_pacContext = m_pacContextStored;

#if COLLECT_CABAC_BITS
  m_pcStatesObject->m_uiBinIndex = m_uiBinIndexStored;
#endif

  delete this;

  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _PROBABILITYALLOCATORENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
