/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RDOQ_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RDOQ_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "ProbabilityAllocatorEnc.h"
#include "RDCost.h"
#include "SliceEstimator.h"
#include "AccessUnitEncoderParameters.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/SyntaxElementCoder.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class RDOQ
{
public:
  RDOQ( const ProbabilityAllocatorBase<CabacContextModel>& rcProbAlloc,
        const ContextArrayStartIndices&                   rcCtxArrayStartIndices,
        const CtxAccess&                                  rcCtxAccess,
        const AccessUnitEncoderParameters&                rcAUEstPars,
        const PlanePackageParameters&                     rcPlaPaPar,
        const SliceEstimator::Parameters&                 rcSliceEstPars );
  virtual ~RDOQ();

  bool          isActive  ()                const { return m_uiQuantizationMode > 0; }
  bool          skipIDCT  ()                const { return m_uiQuantizationMode > 1; }
  const Lambda& getLambda ( UInt uiPlane )  const { return m_acLambda[ uiPlane ]; }

  ErrVal  quantizePow2  ( int                         iQP,
                          UInt                        uiLog2PrimSize,
                          UInt                        uiPlane,
                          UInt                        uiCBFCtxInc,
                          const Block<const int>&     rcOrgCoeff,
                          const Block<int>&           rcLevel,
                          bool&                       rbCoded,
                          RDCost&                     rcRdCost );

private:
  Int64   xGetUncodedCost   ( const Block<const int>&         rcOrgCoeff,
                              const UShort                    ui16CtxNumCBF,
                              UInt&                           ruiMaxLineNum ) const;
  Int64   xQuantize         ( const Block<const int>&         rcOrgCoeff,
                              const Block<int>&               rcLevel,
                              bool&                           rbCoded       ) const;
#if CHECK_RDOQ_COST
  Int64   xGetDistortion    ( const Block<const int>&         rcOrgCoeff,
                              const Block<const int>&         rcLevel       ) const; 
#endif

private:
  //===== specific to integer implementation ======
  FORCEINLINE int     xGetIEPRate    ()                        const;
  FORCEINLINE int     xGetIRate      ( UShort  ui16CtxNumber, 
                                       UInt    uiBit         ) const;
  FORCEINLINE Int64   xGetICost      ( int     iRate         ) const; 
  FORCEINLINE Int64   xGetICost      ( UShort  ui16CtxNumber, 
                                       UInt    uiBit         ) const;
  FORCEINLINE Int64   xGetICRateCost ( UInt    uiAbsLevel,
                                       bool    bLastScanPos,
                                       UShort  ui16CtxNumSig,
                                       UShort  ui16CtxNumOne,
                                       UShort  ui16CtxNumAbs ) const;
  FORCEINLINE int     xGetCodedLevel ( Int64&  ri64UncodedCost,
                                       Int64&  ri64CodedCost,
                                       int     iOrgCoeff,
                                       bool    bLastScanPos,
                                       UShort  ui16CtxNumSig,
                                       UShort  ui16CtxNumOne,
                                       UShort  ui16CtxNumAbs,
                                       UInt    uiQuantIdx ) const;

private:
  const ProbabilityAllocatorBase<CabacContextModel>& m_rcProbAlloc;
  UInt                         m_uiQuantizationMode;
  UInt                         m_uiMinLog2PrimSize;
  UInt                         m_uiMaxLog2PrimSize;
  UInt                         m_uiNumPlanes;
  Lambda                       m_cPrimLambda;
  std::vector<Lambda>          m_acLambda;
  std::vector<std::vector<std::vector<UInt> > > m_aaauiScanPattern;
  const ContextArrayStartIndices&               m_rcCtxArrayStartIndices;
  const CtxAccess&                              m_rcCtxAccess;
  UInt                         m_uiLog2PrimSize;
  UInt                         m_uiLog2BlkSize;
  UInt                         m_uiPlane;
  UInt                         m_uiCBFCtxInc;
  //===== specific to integer implementation ======
  static const UInt            sm_uiQuantShift;
  static const Int64           sm_aai64QStep    [3][6];
  static const int             sm_aiCtxModelRate[2][64];
  Int64                        m_i64Lambda;
  Int64                        m_ai64QuantStepSize [3];
  Int64                        m_ai64QuantFactor   [3];
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _RDOQ_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
