/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _REFERENCEPICTUREENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _REFERENCEPICTUREENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/ReferencePicture.h"
#include "VideoCommonLib/PicBlock.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_PelIn>
class ReferencePictureEnc : public ReferencePicture<t_PelIn>
{
  enum { NORM_WINDOW_SIZE = 4 };
public:
  ReferencePictureEnc( const PicOrderCnt                             cPoc,
                       const PicBlockStats&                          rcAUStats,
                       UInt                                          uiUpsShift,
                       UInt                                          uiMargin,
                       UInt                                          uiSearchRange,
                       bool                                          bStoreDecPic,
                       bool                                          bStoreNorms,
                       typename ReferencePicture<t_PelIn>::UpsMethod cUpsMethod );

  virtual ~ReferencePictureEnc()
  {
    delete m_pc4x4BlockL1NormsPic;
    delete m_pcDecPic;
    delete m_pcOrigRefPic;
  }

  virtual ErrVal assignPlanePackage( UInt uiPrimPlaneId, const PicBlock<const t_PelIn> &rcPic, typename ReferencePicture<t_PelIn>::UpsMethod cUpsMethod, const UInt uiMargin = 0 );

  ErrVal setOrigRefPic( const ReferencePictureEnc<t_PelIn>* pcOrigRefPic )
  {
    ROT( m_pcOrigRefPic );
    ROF( pcOrigRefPic );
    ROF( pcOrigRefPic->getPicOrderCnt () == ReferencePicture<t_PelIn>::getPicOrderCnt() );
    ROF( pcOrigRefPic->getUpsShift    () == ReferencePicture<t_PelIn>::getUpsShift   () );
    ROF( pcOrigRefPic->getPrimOrigin  () == ReferencePicture<t_PelIn>::getPrimOrigin () );
    ROF( pcOrigRefPic->getStats       () == ReferencePicture<t_PelIn>::getStats      () );
    m_pcOrigRefPic = pcOrigRefPic;
    return Err::m_nOK;
  }

  virtual const PrimBlockSize&                                       get4x4BlockL1NormsPrimOrigin()  const { return m_c4x4BlockL1NormsPrimOrigin; }
  virtual const PicBlock<UInt>*                                      get4x4BlockL1NormsPic       ()  const { return m_pc4x4BlockL1NormsPic; }
  virtual const PicBlock<typename ReferencePicture<t_PelIn>::DType>* getDecPic                   ()  const { return m_pcDecPic; }
  virtual const ReferencePictureEnc<t_PelIn>*                        getOrigRefPicture           ()  const { return m_pcOrigRefPic; }

private:
  static ErrVal xGenerateL1NormPlane( const Block<const t_PelIn> &rcEncPlane, const Block<UInt> &rcNormPlane, const PixelPos& cOriginInNormPlane, const BlockSize& cNormWindowSize );

  const PrimBlockSize                                  m_c4x4BlockL1NormsPrimOrigin;
  PicBlock<UInt>*                                      m_pc4x4BlockL1NormsPic;
  PicBlock<typename ReferencePicture<t_PelIn>::DType>* m_pcDecPic;
  const ReferencePictureEnc<t_PelIn>*                  m_pcOrigRefPic;
};


template<class t_PelIn>
class ReferencePlanePackageEnc : public ReferencePlanePackage<t_PelIn>
{
public:
  ReferencePlanePackageEnc( const ReferencePicture<t_PelIn>& rcRefPic, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling& rcPrimPlaneSubsampling );
  
  virtual ~ReferencePlanePackageEnc()
  {
    delete m_pc4x4BlockL1NormsPlaPa;
    delete m_pcDecRefPlaPa;
    delete m_pcOrigRefPlaPa;
  }

  virtual ErrVal getL1Norm4x4Block( const PicBlock<UInt> &cDst, const MotionVector cPosVecToTopLeft ) const;

  virtual const PrimBlockSize&                                            get4x4BlockL1NormsOrigin  ()  const { return  m_c4x4BlockL1NormsOrigin; }
  virtual const PicBlock<UInt>&                                           get4x4BlockL1NormsPlaPa   ()  const { return *m_pc4x4BlockL1NormsPlaPa; }
  virtual const PicBlock<typename ReferencePlanePackage<t_PelIn>::DType>& getDecRefPlaPa            ()  const { return *m_pcDecRefPlaPa; }
  virtual const ReferencePlanePackage<t_PelIn>&                           getOrigRefPlaPa           ()  const { return *m_pcOrigRefPlaPa; }

private:
  const PrimBlockSize                                             m_c4x4BlockL1NormsOrigin;
  const PicBlock<UInt>*                                           m_pc4x4BlockL1NormsPlaPa;
  const PicBlock<typename ReferencePlanePackage<t_PelIn>::DType>* m_pcDecRefPlaPa;
  const ReferencePlanePackageEnc<t_PelIn>*                        m_pcOrigRefPlaPa;
};


template<class t_PelIn>
ReferencePictureEnc<t_PelIn>::ReferencePictureEnc( const PicOrderCnt                             cPoc, 
                                                   const PicBlockStats&                          rcAUStats,
                                                   UInt                                          uiUpsShift,
                                                   UInt                                          uiMargin,
                                                   UInt                                          uiSearchRange,
                                                   bool                                          bStoreDecPic,
                                                   bool                                          bStoreNorms,
                                                   typename ReferencePicture<t_PelIn>::UpsMethod cUpsMethod )
: ReferencePicture<t_PelIn>   ( cPoc, rcAUStats, uiUpsShift, uiMargin, cUpsMethod )
, m_c4x4BlockL1NormsPrimOrigin( PixelPos( NORM_WINDOW_SIZE - 1, NORM_WINDOW_SIZE - 1 ) / ReferencePicture<t_PelIn>::xGetMinSubsampling( rcAUStats ) )
, m_pc4x4BlockL1NormsPic      ( bStoreNorms ? new AllocPicBlock<UInt>( PicBlockStats( rcAUStats.getPrimBlockSize() + m_c4x4BlockL1NormsPrimOrigin, rcAUStats.getSubsampling() ) ) : 0 )
, m_pcDecPic                  ( bStoreDecPic ? new AllocPicBlock<typename ReferencePicture<t_PelIn>::DType>( rcAUStats ) : 0 )
, m_pcOrigRefPic              ( 0 )
{
}

template<class t_PelIn>
ErrVal ReferencePictureEnc<t_PelIn>::xGenerateL1NormPlane( const Block<const t_PelIn> &rcEncPlane, const Block<UInt> &rcNormPlane, const PixelPos& cOriginInNormPlane, const BlockSize& cNormWindowSize )
{
  // Initialize column norms for first line
  std::vector<UInt> auiColumnNorms( rcNormPlane.getSize().getX() + cNormWindowSize.getX(), 0 );
  for( int x = 0; x < int( auiColumnNorms.size() ); x++ )
  {
    for( int y = 0; y < cNormWindowSize.getY(); y++ )
    {
      PixelPos cPosInEncPlane = PixelPos( x, y ) - cOriginInNormPlane;
      cPosInEncPlane.getX()   = gMinMax( cPosInEncPlane.getX(), 0, rcEncPlane.getSize().getX() - 1 );
      cPosInEncPlane.getY()   = gMinMax( cPosInEncPlane.getY(), 0, rcEncPlane.getSize().getY() - 1 );
      auiColumnNorms[ x ]    += rcEncPlane.getEl( cPosInEncPlane );
    }
  }

  for( int y = 0; y < rcNormPlane.getSize().getY(); y++ )
  {
    UInt uiNormRun = 0;
    for( int x = 0; x < cNormWindowSize.getX(); x++ )
    {
      uiNormRun += auiColumnNorms[ x ];
    }

    for( int x = 0; x < rcNormPlane.getSize().getX(); x++ )
    {
      rcNormPlane.getEl( PixelPos( x, y ) ) = uiNormRun;
      uiNormRun -= auiColumnNorms[ x ];
      uiNormRun += auiColumnNorms[ x + cNormWindowSize.getX() ];
    }

    // update column norm vector
    for( int x = 0; x < int( auiColumnNorms.size() ); x++ )
    {
      PixelPos cPosInEncPlane = PixelPos( x, y ) - cOriginInNormPlane;
      cPosInEncPlane.getX()   = gMinMax( cPosInEncPlane.getX(), 0, rcEncPlane.getSize().getX() - 1 );
      cPosInEncPlane.getY()   = gMinMax( cPosInEncPlane.getY(), 0, rcEncPlane.getSize().getY() - 1 );
      auiColumnNorms[ x ]    -= rcEncPlane.getEl( cPosInEncPlane );
      cPosInEncPlane.getY()   = gMinMax( y - cOriginInNormPlane.getY() + cNormWindowSize.getY(), 0, rcEncPlane.getSize().getY() - 1 );
      auiColumnNorms[ x ]    += rcEncPlane.getEl( cPosInEncPlane );
    }
  }
  return Err::m_nOK;
}

template<class t_PelIn>
ErrVal ReferencePictureEnc<t_PelIn>::assignPlanePackage( UInt uiPrimPlaneId, const PicBlock<const t_PelIn> &rcPic, typename ReferencePicture<t_PelIn>::UpsMethod cUpsMethod, const UInt uiMargin )
{
  RNOK( ReferencePicture<t_PelIn>::assignPlanePackage( uiPrimPlaneId, rcPic, cUpsMethod, 0 ) );

  if( ReferencePicture<t_PelIn>::getInterpolationIf() )
  {
    if( cUpsMethod == ReferencePicture<t_PelIn>::CUBIC_MOMS_FIXED_MC || cUpsMethod == ReferencePicture<t_PelIn>::CUBIC_SPLINE_FIXED_MC )
    {
      for( UInt uiInPlane = 0; uiInPlane < rcPic.getStats().getNumPlanes(); uiInPlane++ )
      {
        typedef typename TypeLimits<t_PelIn>::DType DType;
        const Block< DType >&         rcUpsPlane  = ReferencePicture<t_PelIn>::getPicBlock().getPlane( uiInPlane + uiPrimPlaneId );
//        const Block<const t_PelIn>& rcInPlane   = rcPic      .getPlane( uiInPlane );

        // interpolate using filter coeffs but only for 0,0 filter position
        ReferencePicture<t_PelIn>::getInterpolationIf()->interpolate00(uiInPlane + uiPrimPlaneId, rcUpsPlane, 1<<ReferencePicture<t_PelIn>::getUpsShift());
      }
    }
  }


  if( m_pc4x4BlockL1NormsPic )
  {
    ROF( m_pc4x4BlockL1NormsPic->getStats().getNumPlanes() >= uiPrimPlaneId + rcPic.getStats().getNumPlanes() );
    for( UInt uiPlane = 0; uiPlane < rcPic.getStats().getNumPlanes(); uiPlane++ )
    {
      const Block<const t_PelIn>& rcInPlane   = rcPic                  .getPlane( uiPlane );
      const Block<UInt>&          rcNormPlane = m_pc4x4BlockL1NormsPic->getPlane( uiPlane + uiPrimPlaneId );
      const Subsampling&          rcSub       = m_pc4x4BlockL1NormsPic->getStats().getSubsampling()[ uiPlane + uiPrimPlaneId ];
      RNOK( xGenerateL1NormPlane( rcInPlane, rcNormPlane, m_c4x4BlockL1NormsPrimOrigin * rcSub, BlockSize( NORM_WINDOW_SIZE, NORM_WINDOW_SIZE ) ) );
      rcNormPlane *= ( 1 << EXTRA_PRECISION_BITS );
    }
  }
  if( m_pcDecPic )
  {
    ROF( m_pcDecPic->getStats().getNumPlanes() >= uiPrimPlaneId + rcPic.getStats().getNumPlanes() );
    for( UInt uiPlane = 0; uiPlane < rcPic.getStats().getNumPlanes(); uiPlane++ )
    {
      const Block<const t_PelIn>&                             rcInPlane   = rcPic      .getPlane( uiPlane );
      const Block<typename ReferencePicture<t_PelIn>::DType>& rcDecPlane  = m_pcDecPic->getPlane( uiPlane + uiPrimPlaneId );

      // this is the potential place to use an extra large picture for reconstruction

      RNOK( rcDecPlane.assign( rcInPlane ) );
      rcDecPlane *= ( 1 << EXTRA_PRECISION_BITS );
    }
  }
  return Err::m_nOK;
}

template<class t_PelIn>
ReferencePlanePackageEnc<t_PelIn>::ReferencePlanePackageEnc( const ReferencePicture<t_PelIn> &rcRefPic, UInt uiPrimPlaneId, UInt uiNumPlanes, const Subsampling &rcPrimPlaneSubsampling )
: ReferencePlanePackage<t_PelIn>( rcRefPic, uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling )
, m_c4x4BlockL1NormsOrigin      ( ( rcRefPic.get4x4BlockL1NormsPrimOrigin() * rcPrimPlaneSubsampling ).getX(), ( rcRefPic.get4x4BlockL1NormsPrimOrigin() * rcPrimPlaneSubsampling ).getY() )
, m_pc4x4BlockL1NormsPlaPa      ( rcRefPic.get4x4BlockL1NormsPic() ? new PicBlock<UInt>                                          ( *rcRefPic.get4x4BlockL1NormsPic (), uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling ) : NULL )
, m_pcDecRefPlaPa               ( rcRefPic.getDecPic            () ? new PicBlock<typename ReferencePlanePackage<t_PelIn>::DType>( *rcRefPic.getDecPic             (), uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling ) : NULL )
, m_pcOrigRefPlaPa              ( rcRefPic.getOrigRefPicture    () ? new ReferencePlanePackageEnc<t_PelIn>                       ( *rcRefPic.getOrigRefPicture     (), uiPrimPlaneId, uiNumPlanes, rcPrimPlaneSubsampling ) : NULL )
{
}

template<class t_PelIn>
ErrVal ReferencePlanePackageEnc<t_PelIn>::getL1Norm4x4Block( const PicBlock<UInt> &rcDst, const MotionVector cPosVecToTopLeft ) const
{
  ROF( cPosVecToTopLeft.getPrecision() == 0 );
  for( UInt uiPl = 0; uiPl < rcDst.getStats().getNumPlanes(); uiPl++ )
  {
    const Subsampling &rcSub = m_pc4x4BlockL1NormsPlaPa->getStats().getSubsampling()[uiPl];
    ROF( rcSub == rcDst.getStats().getSubsampling()[uiPl] );
    const Block<UInt> &rcDstPlane = rcDst.getPlane( uiPl );
    const Block<UInt> &rcNormPlane = m_pc4x4BlockL1NormsPlaPa->getPlane( uiPl );
    PixelPos cTopLeft = ( PrimPixelPos( cPosVecToTopLeft.getX(), cPosVecToTopLeft.getY() ) + m_c4x4BlockL1NormsOrigin ) * rcSub;
    if( cTopLeft >= PixelPos( 0, 0 ) && cTopLeft + rcDstPlane.getSize() < rcNormPlane.getSize() )
    {
      // fast copying without range check
      rcDstPlane.assign( Block<UInt>( cTopLeft, rcDstPlane.getSize(), rcNormPlane ) );
    }
    else
    {
      for( UInt y = 0, y2 = cTopLeft.getY(); int(y) < rcDstPlane.getSize().getY(); y++, y2++ )
      {
        for( UInt x = 0, x2 = cTopLeft.getX(); int(x) < rcDstPlane.getSize().getX(); x++, x2++ )
        {
          PixelPos cNormPos( x2, y2 );
          cNormPos.getX() = gMinMax( cNormPos.getX(), 0, rcNormPlane.getSize().getX() - 1 );
          cNormPos.getY() = gMinMax( cNormPos.getY(), 0, rcNormPlane.getSize().getY() - 1 );
          rcDstPlane.getEl( PixelPos( x, y ) ) = rcNormPlane.getEl( cNormPos );
        }
      }
    }
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _REFERENCEPICTUREENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
