/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RESESTANDENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RESESTANDENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "ResEstAndEncIf.h"
#include "ResEstInfoVecSrcIf.h"
#include "SyntaxElementEncoderIf.h"
#include "ResidualEstimatorManager.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class ResInfoAndResEstInfo
{
public:
  ResInfoAndResEstInfo() : m_pcResEstInfo( NULL ), m_pcResInfo( NULL ) {}
  virtual ~ResInfoAndResEstInfo() { delete m_pcResInfo; }

  const ResEstInfo *m_pcResEstInfo;
  ResInfo *m_pcResInfo;
};


class ResEstAndEnc
  : public ResEstAndEncIf<ResInfoAndResEstInfo>
{
public:
  typedef ResInfoAndResEstInfo ResIf;

  ResEstAndEnc( const PlanePackageParameters &rcPlaPaPars,
                const AccessUnitParameters &rcAUPars,
                SyntaxElementEncoderIf *pcSEEncoderIf,
                const CodingState &rcCodingState,
                SliceData &rcSlice,
                const ResidualEstimatorManager &rcResMng,
                ResEstInfoVecSrcIf &rcResEstInfoVecSrcIf )
    : m_rcPlaPaPars( rcPlaPaPars )
    , m_rcAUPars( rcAUPars )
    , m_pcSEEncoderIf( pcSEEncoderIf )
    , m_rcCodingState( rcCodingState )
    , m_rcSlice( rcSlice )
    , m_cResMng( rcResMng )
    , m_cActiveResNav( NULL )
    , m_rcResEstInfoVecSrcIf( rcResEstInfoVecSrcIf ) 
    , m_bEstResAgain( false ) {}

  virtual ~ResEstAndEnc() {}

  ErrVal resetResidualEstimationForBlock( const QTNavi &rcNav );
  ErrVal estimateAndEncodeResidual( const QTNavi &rcNav,
                                    bool& rbAllBelowFmdThreshold,
                                    ResInfoAndResEstInfo *&rpResult,
                                    const PredInfo *pcPredInfo,
                                    const PicBlock<const DPel> &rcResidualBlock,
                                    const PicBlock<DPel> &rcResidualReconstructionBlock,
                                    bool &rbResidualValidAndEncoded );
  ErrVal reportRDCost( const QTNavi &rcNav,
                       RDCost cCost,
                       bool &rbLastResidualToTest );
  ErrVal proceedToNextResidual( const QTNavi &rcNav );
  ErrVal setBestResidual( const QTNavi &rcNav, ResInfoAndResEstInfo *&rpResult );
  ErrVal releaseChildNodes( const QTNavi &rcNav );
  ErrVal releaseCurrentNode( const QTNavi &rcNav );

private:
  const PlanePackageParameters &m_rcPlaPaPars;
  const AccessUnitParameters &m_rcAUPars;
  SyntaxElementEncoderIf *m_pcSEEncoderIf;
  const CodingState &m_rcCodingState;
  SliceData &m_rcSlice;

  const ResidualEstimatorManager &m_cResMng;
  QTNavi m_cActiveResNav;
  int m_iActiveResEstInfo;

  ResEstInfoVecSrcIf &m_rcResEstInfoVecSrcIf;
  bool m_bEstResAgain;
};


// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _RESESTANDENC_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
