/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _RESTRAFOPOW2ESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _RESTRAFOPOW2ESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "ResidualEstimator.h"
#include "VideoCommonLib/ResTrafoPow2.h"
#include "SingleNodeEstimatorIf.h"
#include "SyntaxElementEncoderIf.h"
#include "BitRateEstimator.h"
#include "StateIf.h"
#include "RDOQ.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

#if defined( MSYS_WIN32 )
#pragma warning( disable: 4250 )
#endif

class ResEstInfoTrafoPow2
  : public ResEstInfo
{
public:

  ResEstInfoTrafoPow2( const std::vector<int> &raiQp, const Ratio& cDeadzone, UInt uiLog2MaxBlockSize, UInt uiLog2MinBlockSize, bool bSubdivide )
    : m_aiQp( raiQp )
    , m_cDeadzone( cDeadzone )
    , m_uiLog2MaxBlockSize( uiLog2MaxBlockSize )
    , m_uiLog2MinBlockSize( uiLog2MinBlockSize )
    , m_bSubdivide( bSubdivide ) {}

  ResInfo::ResidualMode       getResidualMode()         const { return ResInfo::RES_TRAFO_POW2; }
  const ResEstInfoTrafoPow2*  getResEstInfoTrafoPow2()  const { return this; }
  const std::vector<int>&     getQps()                  const { return m_aiQp; }
  const Ratio&                getDeadzone()             const { return m_cDeadzone; }
  UInt                        getLog2MaxBlockSize()     const { return m_uiLog2MaxBlockSize; }
  UInt                        getLog2MinBlockSize()     const { return m_uiLog2MinBlockSize; }
  bool                        getSubdivide()            const { return m_bSubdivide; }

  ErrVal setMaxLog2TrafoBlockSize( UInt uiMax )
  {
    ROT( uiMax < m_uiLog2MinBlockSize );
    m_uiLog2MaxBlockSize = uiMax;
    return Err::m_nOK;
  }

private:
  const std::vector<int> m_aiQp;
  const Ratio m_cDeadzone;
  UInt m_uiLog2MaxBlockSize;
  UInt m_uiLog2MinBlockSize;
  bool m_bSubdivide;
};

class ResTrafoPow2Estimator
  : public ResidualEstimator,
    public ResTrafoPow2
{
public:
  ResTrafoPow2Estimator( const PlanePackageParameters&      rcPlaPaPars,
                         const AccessUnitEncoderParameters& rcAUEstPars,
                         SyntaxElementEncoderIf&            rcSEEncoderIf, 
                         BitRateEstimatorIf&                rcBitRateEstIf,
                         StateIf&                           rcStateIf, 
                         RDOQ&                              rcRDOQ, 
                         Lambda                             cLambda );
  virtual ~ResTrafoPow2Estimator();

  // inherited from ResidualEstimator
  ErrVal estimateResInfo( ResInfo *&rpcResInfo,
                          bool& rbAllBelowFmdThreshold,
                          const PredInfo *pcPredInfo,
                          const QTNavi &rcNav,
                          const PlanePackageParameters &rcPlaPaPars,
                          const AccessUnitParameters &rcAUPars,
                          const SliceData &rcContextSlice,
                          const ResEstInfo &rcResEstInfo,
                          const PicBlock<const DPel> &rcOrigResidual,
                          bool &rbResidualConversionValid );
  void  resetResEst() { m_bResEstResetted = true; }

private:
  ErrVal xInit( const QTNavi &rcNav,
                const ResEstInfoTrafoPow2 *pcEstInfo,
                const PicBlockStats &rcOrigResidualStats );

  ErrVal xGrowAndPruneTree      ( const QTNavi &rcNav,
                                  const AccessUnitParameters &rcAUPars,
                                  RDCost &rcCost,
                                  bool& rbCoded,
                                  bool& rbAllBelowFMDThreshold,
                                  const PredInfo *pcPredInfo,
                                  UInt uiLog2TrafoMinSize,
                                  QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcResTree,
                                  const QTNavi &rcPredBlkNavi,
                                  const PlanePackageParameters &rcPlaPaPars,
                                  const SliceData &rcCurrSlice,
                                  const PicBlock<const DPel> &rcOriginalResidualPredBlk,
                                  const std::vector<int> &raiQPs,
                                  const Ratio& rcDeadzone );
  ErrVal xEncodeSingleNode      ( const QTNavi &rcNav,
                                  const AccessUnitParameters &rcAUPars,
                                  ResInfoTrafoPow2Leaf *&rpResult,
                                  RDCost *pcCost,
                                  const PredInfo *pcPredInfo,
                                  bool& rbCoded,
                                  bool& rbAllBelowFMDThreshold,
                                  QuadtreeGroup<ResInfoTrafoPow2Leaf*> &rcResTree,
                                  const QTNavi &rcPredBlkNavi,
                                  const PlanePackageParameters &rcPlaPaPars,
                                  const SliceData &rcCurrSlice,
                                  const PicBlock<const DPel> &rcOriginalResidualPredBlk,
                                  const std::vector<int> &raiQPs,
                                  const Ratio& rcDeadzone );
  ErrVal xEncodeSubDivisionInfo ( UInt uiLog2BlockSize, bool bSubdiv, RDCost &rcRDCost );

  ErrVal xReconstruct         ( const ResInfoTrafoPow2Leaf &rcQuantized, const PicBlock<DPel> &rcReconstruction, const std::vector<int> &raiQPs );
  ErrVal xQuantize            ( PrimPixelPos cPos, PrimBlockSize cSize, UInt uiLog2BlockSize, const PicBlock<const DPel> &rcOriginal, ResInfoTrafoPow2Leaf &rcQuantized, bool& rbAllBelowFMDThreshold, const std::vector<int> &raiQPs, const Ratio& rcDeadzone );
  ErrVal xQuantizeRDOQ        ( PrimPixelPos cPos, PrimBlockSize cSize, UInt uiLog2BlockSize, const PicBlock<const DPel> &rcOriginal, ResInfoTrafoPow2Leaf &rcQuantized, bool& rbAllBelowFMDThreshold, RDCost& rcRdCost, const std::vector<int> &raiQPs );

  void   xFast1dDCT           ( int* paiData, UInt uiInc, UInt uiLog2TSize, UInt uiInsideShift, UInt uiFinalShift );
  ErrVal xFast2dDCT           ( const Block<const DPel>& rcIn, const Block<int>& rcOut );
  ErrVal xFastQuant           ( int iQP, const Ratio &rcDeadzone, const Block<const int> &rcIn, const Block<int> &rcOut, bool &rbOutNonzero );
  ErrVal xCheckCoeffs         ( UInt& ruiAllBelowFMDThreshold, const Block<const int>    &rcIn, int    iThreshold )  const;

  // rate estimation tools
  SyntaxElementEncoderIf&         m_rcSEEncoderIf;
  BitRateEstimatorIf&             m_rcBitRateEstIf;
  StateIf&                        m_rcStateIf;
  RDOQ&                           m_rcRDOQ;
  Lambda                          m_cLambda;
  bool                            m_bResEstResetted;
  int                             m_iMaxQP0ToCheck;
  std::vector<PicBlock<int>*>     m_apcTICoeffs;
  PicBlock<DPel>*                 m_pcPaddedBlock;

  UInt                            m_uiMinLog2TrafoBlkSize;
  UInt                            m_uiMaxLog2PredBlkSize;
  bool                            m_bFastTSizeDecision;
  bool                            m_bCheckFmdThreshold;
  std::vector<int>                m_aiQThresholds;
  std::vector<std::vector<UInt> > m_aauiTrafoDone;
  std::vector<std::vector<UInt> > m_aauiAllBelowFmdThreshold;
};

#if defined( MSYS_WIN32 )
#pragma warning( default: 4250 )
#endif

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _RESTRAFOPOW2ESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
