/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SINGLENODEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SINGLENODEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "SingleNodeEstimatorIf.h"
#include "SyntaxElementEncoderIf.h"
#include "BitRateEstimator.h"
#include "PredEstAndEncIf.h"
#include "ResEstAndEncIf.h"
#include "RDCost.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef PicBlock<DPel> SingleNodeEstimator_IfType;

template<class t_PredEstResult, class t_ResEstResult>
class SingleNodeEstimator
  : public SingleNodeEstimatorIf<PicBlock<DPel> >
{
public:
  SingleNodeEstimator( const AccessUnitEncoderParameters& rcAUEstPars,
                       SyntaxElementEncoderIf *pcSEEncoderIf, 
                       BitRateEstimatorIf *pcBitRateEstIf,
                       PredEstAndEncIf<t_PredEstResult> *pcPredEstAndEncIf,
                       ResEstAndEncIf<t_ResEstResult> *pcResEstAndEncIf, 
                       Lambda cLambda, 
                       const PicBlock<Pel> &rcEncPic,
                       const PicBlock<Pel> *pcEncPrePic,
                       const PicBlock<Pel> *pcEncResPic, 
                       const PicBlock<const Pel> &rcOrigPic, 
                       StateIf *pcStateIf )
    : m_pcSEEncoderIf( pcSEEncoderIf )
    , m_pcBitRateEstIf( pcBitRateEstIf )
    , m_pcPredEstAndEncIf( pcPredEstAndEncIf )
    , m_pcResEstAndEncIf( pcResEstAndEncIf )
    , m_cLambda( cLambda )
    , m_rcEncPic( rcEncPic )
    , m_pcEncPrePic( pcEncPrePic )
    , m_pcEncResPic( pcEncResPic )
    , m_rcOrigPic( rcOrigPic )
    , m_pcStateIf( pcStateIf )
    , m_bFastModeDecision( rcAUEstPars.getFastModeDecision() )
    , m_bCheckCoefficients( rcAUEstPars.getFastModeDecision() )
    , m_bAllBelowFmdThreshold( true ) {}

  virtual ~SingleNodeEstimator() {}

  ErrVal processSubdivisionInfo( const QTNavi &rcPos, bool &rbSubdiv, RDCost *pcCost = NULL );
  ErrVal estimateSingleNode( const QTNavi &rcNav, PicBlock<DPel> *&rpResult, PicBlock<DPel>* &rpPreResult, PicBlock<DPel>* &rpResResult, RDCost *pcCost = NULL );
  ErrVal setSingleResult( const QTNavi &rcNav, PicBlock<DPel> *&rpResult, PicBlock<DPel> *&rpPreResult, PicBlock<DPel> *&rpResResult );
  ErrVal releaseChildNodes( const QTNavi &rcNav );
  ErrVal releaseCurrentNode( const QTNavi &rcNav );
  bool   getAllBelowFmdThreshold() const { TOF( m_bCheckCoefficients ); return m_bAllBelowFmdThreshold; }

private:
  SyntaxElementEncoderIf *m_pcSEEncoderIf;
  BitRateEstimatorIf *m_pcBitRateEstIf;
  PredEstAndEncIf<t_PredEstResult> *m_pcPredEstAndEncIf;
  ResEstAndEncIf<t_ResEstResult> *m_pcResEstAndEncIf;
  Lambda m_cLambda;
  const PicBlock<Pel> &m_rcEncPic;
  const PicBlock<Pel> *m_pcEncPrePic;
  const PicBlock<Pel> *m_pcEncResPic;
  const PicBlock<const Pel> &m_rcOrigPic;
  StateIf *m_pcStateIf;
  bool  m_bFastModeDecision;
  bool  m_bCheckCoefficients;
  bool  m_bAllBelowFmdThreshold;

  class EstimationResult
  {
  public:
    EstimationResult( Lambda cLambda )
      : m_cCost( cLambda )
      , m_pcEnc( NULL )
      , m_pcEncPre( NULL )
      , m_pcEncRes( NULL )
      , m_pcState( NULL ) {}

    ErrVal setEncBlock( PicBlock<DPel> *pcEncBlock )       { ROT( !m_pcEnc   == !pcEncBlock ); m_pcEnc = pcEncBlock;  return Err::m_nOK; }
    ErrVal setEncPreBlock( PicBlock<DPel> *pcEncPreBlock ) { ROT( !m_pcEncPre == !pcEncPreBlock  ); m_pcEncPre = pcEncPreBlock; return Err::m_nOK; }
    ErrVal setEncResBlock( PicBlock<DPel> *pcEncResBlock ) { ROT( !m_pcEncRes == !pcEncResBlock  ); m_pcEncRes = pcEncResBlock; return Err::m_nOK; }
    ErrVal setState( State *pcState )                { ROT( !m_pcState == !pcState );    m_pcState = pcState;   return Err::m_nOK; }

    RDCost& getRDCost()           { return m_cCost; }
    PicBlock<DPel>* getEncBlock() { return m_pcEnc; }
    PicBlock<DPel>* getEncPreBlock()  { return m_pcEncPre; }
    PicBlock<DPel>* getEncResBlock()  { return m_pcEncRes; }
    State* getState()             { return m_pcState; }

  private:
    RDCost m_cCost;
    PicBlock<DPel> *m_pcEnc;
    PicBlock<DPel> *m_pcEncPre;
    PicBlock<DPel> *m_pcEncRes;
    State *m_pcState;
  };

};


template<class t_PredEstResult, class t_ResEstResult>
ErrVal SingleNodeEstimator<t_PredEstResult,t_ResEstResult>::processSubdivisionInfo( const QTNavi &rcPos, bool &rbSubdiv, RDCost *pcCost )
{
  if( m_pcSEEncoderIf )
  {
    BitRate cRateBefore( 0 );
    if( pcCost )
    {
      cRateBefore = m_pcBitRateEstIf->getBitRateEstimate();
    }
    RNOK( m_pcSEEncoderIf->processTreeSubdivisionFlag( rcPos, rbSubdiv ) );

    if( pcCost )
    {
      pcCost->addRate( m_pcBitRateEstIf->getBitRateEstimate() - cRateBefore );
    }
  }
  return Err::m_nOK;
}


template<class t_PredEstResult, class t_ResEstResult>
ErrVal SingleNodeEstimator<t_PredEstResult,t_ResEstResult>::estimateSingleNode( const QTNavi &rcNav, PicBlock<DPel> *&rpResult, PicBlock<DPel>* &rpPreResult, PicBlock<DPel>* &rpResResult, RDCost *pcCost )
{
  ROT( pcCost && !m_pcBitRateEstIf );
  m_bAllBelowFmdThreshold   = false;
  bool  bAllBelowThresInter = false;

  EstimationResult* pcCurrentResult = new EstimationResult( m_cLambda );
  EstimationResult* pcBestResult    = new EstimationResult( m_cLambda );
  pcBestResult->getRDCost().setMax();

  t_PredEstResult* pcCurrentPred = NULL;
  t_PredEstResult* pcBestPred = NULL;
  t_ResEstResult* pcCurrentRes = NULL;
  t_ResEstResult* pcBestRes = NULL;

  BitRate cRootRate( 0 );
  if( m_pcBitRateEstIf )
  {
    cRootRate = m_pcBitRateEstIf->getBitRateEstimate();
  }

  RNOK( m_pcPredEstAndEncIf->resetPredictionEstimationForBlock( rcNav ) );

  // save initial bitstream state
  State *pcPredictionRootState = NULL;
  if( m_pcStateIf )
  {
    RNOK( m_pcStateIf->getState( pcPredictionRootState ) );
  }

  const PicBlock<const Pel> cOrigPicBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), m_rcOrigPic );
  AllocPicBlock<DPel> cPredictionBlock( cOrigPicBlock.getStats() );
  AllocPicBlock<DPel> cResidualBlock( cOrigPicBlock.getStats() );

  ROT( rpResult );
  ROT( rpPreResult );
  ROT( rpResResult );

  RNOK( pcCurrentResult->setEncBlock( new AllocPicBlock<DPel>( cOrigPicBlock.getStats() ) ) );
  RNOK( pcBestResult   ->setEncBlock( new AllocPicBlock<DPel>( cOrigPicBlock.getStats() ) ) );

  if ( m_pcEncPrePic )
  {
    RNOK( pcCurrentResult->setEncPreBlock( new AllocPicBlock<DPel>( cOrigPicBlock.getStats() ) ) );
    RNOK( pcBestResult   ->setEncPreBlock( new AllocPicBlock<DPel>( cOrigPicBlock.getStats() ) ) );
  }

  if ( m_pcEncResPic )
  {
    RNOK( pcCurrentResult->setEncResBlock( new AllocPicBlock<DPel>( cOrigPicBlock.getStats() ) ) );
    RNOK( pcBestResult   ->setEncResBlock( new AllocPicBlock<DPel>( cOrigPicBlock.getStats() ) ) );
  }

  bool bFirstRunPrediction = true;
  // cycle through all prediction calculations to test
  for( bool bLastPredictionToTest = false; !bLastPredictionToTest; m_pcPredEstAndEncIf->proceedToNextPrediction( rcNav ) )
  {
    bool bIsIntra = m_pcPredEstAndEncIf->currModeIsIntra();
    if( m_bFastModeDecision )
    { // assuming correct ordering of inter/intra modes
      if( bAllBelowThresInter && bIsIntra )
      {
        bLastPredictionToTest = true;
        continue;
      }
    }

    if( !bFirstRunPrediction )
    {
      RNOK( pcPredictionRootState->setState() );
    }
    bFirstRunPrediction = false;

    bool bPredictionValidAndEncoded = false;
    ROT( pcCurrentPred );
    const PredInfo *pcEncodedPred = NULL;
    RNOK( m_pcPredEstAndEncIf->estimateAndEncodePrediction( rcNav, pcCurrentPred, pcEncodedPred, cPredictionBlock, bPredictionValidAndEncoded ) );

    RDCost cBestCostOfCurrentPrediction( m_cLambda );
    cBestCostOfCurrentPrediction.setMax();

    if( bPredictionValidAndEncoded )
    {
      bool bCostWasReduced  = false;
      bool bAllBelowFmdPred = false;

      RNOK( m_pcResEstAndEncIf->resetResidualEstimationForBlock( rcNav ) );

      State *pcResidualRootState = NULL;
      if( m_pcStateIf )
      {
        RNOK( m_pcStateIf->getState( pcResidualRootState ) );
      }

      // calculate residual from original picture block and its prediction
      cResidualBlock.assign( cOrigPicBlock );
      cResidualBlock <<= EXTRA_PRECISION_BITS;
      cResidualBlock  -= cPredictionBlock;

      bool bFirstRunResidual = true;
      bool bAtLeastOneResidualProcessingSucceeded = false;
      for( bool bLastResidualToTest = false; !bLastResidualToTest; m_pcResEstAndEncIf->proceedToNextResidual( rcNav ) )
      {
        if( !bFirstRunResidual )
        {
          RNOK( pcResidualRootState->setState() );
        }
        bFirstRunResidual = false;

        // convert and encode, then reconstruct the residual
        PicBlock<DPel>& cReconstructionBlock = *pcCurrentResult->getEncBlock();
        PicBlock<DPel>& cReconstructedResBlock = *pcCurrentResult->getEncResBlock();
        PicBlock<DPel>& cReconstructedPreBlock = *pcCurrentResult->getEncPreBlock();

        ROT( pcCurrentRes );

        bool bResidualValidAndEncoded = false;
        bool bAllBelowFMDThresholdRes = false;
        RNOK( m_pcResEstAndEncIf->estimateAndEncodeResidual( rcNav, bAllBelowFMDThresholdRes, pcCurrentRes, pcEncodedPred, cResidualBlock, cReconstructionBlock, bResidualValidAndEncoded ) );

        if( bResidualValidAndEncoded )
        {
          if( bAllBelowFMDThresholdRes )
          {
            bAllBelowFmdPred = true;
          }

          bAtLeastOneResidualProcessingSucceeded = true;

          if (m_pcEncPrePic)
          {
            cReconstructedPreBlock.assign(0);
            cReconstructedPreBlock.addShiftAndClip<((1 << EXTRA_PRECISION_BITS ) >> 1), EXTRA_PRECISION_BITS , DPel>( cPredictionBlock, 0, 255 );
          }

          if (m_pcEncResPic)
          {
            cReconstructedResBlock.assign(0);
            cReconstructedResBlock.addShiftAndClip<((1 << EXTRA_PRECISION_BITS ) >> 1), EXTRA_PRECISION_BITS , DPel>( cReconstructionBlock, -255, 255 );
          }

          // calculate original-reconstruction from reconstructed residual and prediction
          cReconstructionBlock.addShiftAndClip<((1 << EXTRA_PRECISION_BITS ) >> 1), EXTRA_PRECISION_BITS , DPel>( cPredictionBlock, 0, 255 );

          bool bNewBest = true;
          // calculate rate and distortion measure
          if( m_pcBitRateEstIf )
          {
            pcCurrentResult->getRDCost().reset();
            pcCurrentResult->getRDCost().addRate( m_pcBitRateEstIf->getBitRateEstimate() - cRootRate );
            pcCurrentResult->getRDCost().addDistortion( cOrigPicBlock, cReconstructionBlock );
            bNewBest = pcCurrentResult->getRDCost() < pcBestResult->getRDCost();

            // remember the best cost for the current prediction
            if( pcCurrentResult->getRDCost() < cBestCostOfCurrentPrediction )
            {
              cBestCostOfCurrentPrediction = pcCurrentResult->getRDCost();
            }
          }

          // adopt current result if its better than the best found so far
          if( bNewBest )
          {
            bCostWasReduced = true;

            gSwapPtr( pcCurrentResult, pcBestResult );

            // delete former best state
            if( NULL != pcCurrentResult->getState() )
            {
              RNOK( pcCurrentResult->getState()->deleteState() );
              RNOK( pcCurrentResult->setState( NULL ) );
            }

            // save new best state
            ROT( pcBestResult->getState() );
            State *pcBestState = NULL;
            if( m_pcStateIf )
            {
              RNOK( m_pcStateIf->getState( pcBestState ) );
              RNOK( pcBestResult->setState( pcBestState ) );
            }

            // delete former best residual conversion and save new best residual conversion
            if( pcBestRes )
            {
              delete pcBestRes;
            }
            pcBestRes = pcCurrentRes;
          }
          else
          {
            delete pcCurrentRes;
          }
        }
        pcCurrentRes = NULL;
        RNOK( m_pcResEstAndEncIf->reportRDCost( rcNav, cBestCostOfCurrentPrediction, bLastResidualToTest ) );
      } // end of ResidualConversion loop

      if( bAllBelowFmdPred )
      {
        m_bAllBelowFmdThreshold = true;
        if( ! bIsIntra )
        {
          bAllBelowThresInter   = true;
        }
      }

      ROF( bAtLeastOneResidualProcessingSucceeded );

      if ( pcResidualRootState )
      {
        RNOK( pcResidualRootState->deleteState() );
        pcResidualRootState = NULL;
      }

      if (bCostWasReduced)
      {
        // delete former best prediction and save new best prediction
        delete pcBestPred;
        pcBestPred = pcCurrentPred;
      }
      else
      {
        delete pcCurrentPred;
      }
      pcCurrentPred = NULL;
    }

    RNOK( m_pcPredEstAndEncIf->reportRDCost( rcNav, cBestCostOfCurrentPrediction, bLastPredictionToTest ) );
  } // end of Prediction loop

  if ( pcPredictionRootState )
  {
    RNOK( pcPredictionRootState->deleteState() );
    pcPredictionRootState = NULL;
  }

  // return contents of pcBestResult
  if( pcCost )
  {
    if( pcBestResult->getRDCost().isMax() )
      pcCost->setMax();
    else
      *pcCost += pcBestResult->getRDCost();
  }

  rpResult = pcBestResult->getEncBlock();
  delete pcCurrentResult->getEncBlock();

  if( m_pcEncPrePic )
  {
    PicBlock<DPel>& cPredictionBlock = *pcBestResult->getEncPreBlock();
    cPredictionBlock.drawBorder( false, false, true, true );
    rpPreResult = &cPredictionBlock;
    delete pcCurrentResult->getEncPreBlock();
  }

  if( m_pcEncResPic )
  {
    PicBlock<DPel>& cResidualBlock = *pcBestResult->getEncResBlock();
    cResidualBlock.clip( -127, 128 );
    cResidualBlock += 127;
    cResidualBlock.drawBorder( false, false, true, true );
    rpResResult = &cResidualBlock;
    delete pcCurrentResult->getEncResBlock();
  }

  delete pcCurrentPred;
  delete pcCurrentRes;

  if( !pcBestResult->getRDCost().isMax() )
  {
    ROF( pcBestPred );
    ROF( pcBestRes );

    RNOK( m_pcPredEstAndEncIf->setBestPrediction( rcNav, pcBestPred ) );
    RNOK( m_pcResEstAndEncIf->setBestResidual( rcNav, pcBestRes ) );
  }

  // return to best bit stream state
  if( m_pcStateIf && pcBestResult->getState() )
  {
    RNOK( pcBestResult->getState()->setAndDeleteState() );
  }

  delete pcCurrentResult;
  delete pcBestResult;

  return Err::m_nOK;
}


template<class t_PredEstResult, class t_ResEstResult>
ErrVal SingleNodeEstimator<t_PredEstResult,t_ResEstResult>::setSingleResult( const QTNavi &rcNav, PicBlock<DPel> *&rpResult, PicBlock<DPel> *&rpPreResult, PicBlock<DPel> *&rpResResult )
{
  PicBlock<Pel> cEncBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), m_rcEncPic );
  PicBlock<Pel> *pcEncPreBlock = rpPreResult ? new PicBlock<Pel>( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), *m_pcEncPrePic ) : NULL;
  PicBlock<Pel> *pcEncResBlock = rpResResult ? new PicBlock<Pel>( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), *m_pcEncResPic ) : NULL;

  cEncBlock.assign( *rpResult );
  if ( pcEncPreBlock )
  {
    pcEncPreBlock->assign( *rpPreResult );
    delete pcEncPreBlock;
  }

  if ( pcEncResBlock )
  {
    pcEncResBlock->assign( *rpResResult );
    delete pcEncResBlock;
  }

  delete rpResult;
  delete rpPreResult;
  delete rpResResult;
  rpResult = NULL;
  rpPreResult = NULL;
  rpResResult = NULL;
  return Err::m_nOK;
}


template<class t_PredEstResult, class t_ResEstResult>
ErrVal SingleNodeEstimator<t_PredEstResult,t_ResEstResult>::releaseChildNodes( const QTNavi &rcNav )
{
  RNOK( m_pcPredEstAndEncIf->releaseChildNodes( rcNav ) );
  RNOK( m_pcResEstAndEncIf->releaseChildNodes( rcNav ) );
  return Err::m_nOK;
}


template<class t_PredEstResult, class t_ResEstResult>
ErrVal SingleNodeEstimator<t_PredEstResult,t_ResEstResult>::releaseCurrentNode( const QTNavi &rcNav )
{
  RNOK( m_pcPredEstAndEncIf->releaseCurrentNode( rcNav ) );
  RNOK( m_pcResEstAndEncIf->releaseCurrentNode( rcNav ) );
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _SINGLENODEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
