/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SINGLENODEESTIMATORWF_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SINGLENODEESTIMATORWF_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "SingleNodeEstimatorIf.h"
#include "SyntaxElementEncoderIf.h"
#include "BitRateEstimator.h"
#include "VideoCommonLib/WienerFilterParameter.h"


#include "RDCost.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef PicBlock<Pel> SingleNodeEstimatorWF_IfType;

template<class t_BlockPar>
class SingleNodeEstimatorWF
  : public SingleNodeEstimatorIf<PicBlock<Pel> >
{
public:
  SingleNodeEstimatorWF( SyntaxElementEncoderIf *pcSEEncoderIf, BitRateEstimatorIf *pcBitRateEstIf, QuadtreeGroup<t_BlockPar> &rcQTBlockPars, Lambda cLambda, const PicBlock<Pel> &rcEncPic, const PicBlock<Pel> &rcFilteredPic, const PicBlock<const Pel> &rcOrigPic, StateIf *pcStateIf )
    : m_pcSEEncoderIf( pcSEEncoderIf )
    , m_pcBitRateEstIf( pcBitRateEstIf )
    , m_rcQTBlockPars(rcQTBlockPars )
    , m_cLambda( cLambda )
    , m_rcEncPic( rcEncPic )
    , m_rcOrigPic( rcOrigPic )
    , m_rcFilteredPic( rcFilteredPic )
    , m_pcStateIf( pcStateIf ) {}

  virtual ~SingleNodeEstimatorWF() {}

  ErrVal processSubdivisionInfo( const QTNavi &rcPos, bool &rbSubdiv, RDCost *pcCost = NULL );
  ErrVal estimateSingleNode( const QTNavi &rcNav, PicBlock<Pel> *&rpResult, PicBlock<Pel> *&rpDummyResult, PicBlock<Pel> *&rpNextDummyResult,  RDCost *pcCost = NULL );
  ErrVal setSingleResult( const QTNavi &rcNav, PicBlock<Pel> *&rpResult, PicBlock<Pel> *&rpDummyResult, PicBlock<Pel> *&rpNextDummyResult );
  ErrVal releaseChildNodes( const QTNavi &rcNav );
  ErrVal releaseCurrentNode( const QTNavi &rcNav );

private:
  SyntaxElementEncoderIf *m_pcSEEncoderIf;
  BitRateEstimatorIf *m_pcBitRateEstIf;
  QuadtreeGroup<t_BlockPar>& m_rcQTBlockPars;
  Lambda m_cLambda;
  const PicBlock<Pel> &m_rcEncPic;
  const PicBlock<const Pel> &m_rcOrigPic;
  const PicBlock<Pel> &m_rcFilteredPic;
  StateIf *m_pcStateIf;


};

template<class t_BlockPar>
ErrVal SingleNodeEstimatorWF<t_BlockPar>::processSubdivisionInfo( const QTNavi &rcPos, bool &rbSubdiv, RDCost *pcCost )
{
  if( m_pcSEEncoderIf )
  {
    BitRate cRateBefore( 0 );
    if( pcCost )
    {
      cRateBefore = m_pcBitRateEstIf->getBitRateEstimate();
    }
    RNOK( m_pcSEEncoderIf->processWienerTreeSubdivisionFlag( rcPos, rbSubdiv ) );

    if( pcCost )
    {
      pcCost->addRate( m_pcBitRateEstIf->getBitRateEstimate() - cRateBefore );
    }
  }
  return Err::m_nOK;
}


template<class t_BlockPar>
ErrVal SingleNodeEstimatorWF<t_BlockPar>::estimateSingleNode( const QTNavi &rcNav, PicBlock<Pel> *&rpResult, PicBlock<Pel> *&rpDummyResult, PicBlock<Pel> *&rpNextDummyResult, RDCost *pcCost )
{
  ROT( pcCost && !m_pcBitRateEstIf );
  PicBlock<Pel> rcEncPicBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), m_rcEncPic );
  PicBlock<Pel> rcFiltPicBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), m_rcFilteredPic );
  const PicBlock<const Pel> cOrigPicBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), m_rcOrigPic );
  rpResult = new AllocPicBlock<Pel>( rcEncPicBlock.getStats() );

  BitRate cRootRate( 0 );
  if( m_pcBitRateEstIf )
  {
     cRootRate = m_pcBitRateEstIf->getBitRateEstimate();
  }

  // save initial bitstream state
//  State::State *pcRootState = NULL;
  State *pcRootState = NULL;
  if( m_pcStateIf )
  {
    RNOK( m_pcStateIf->getState( pcRootState ) );
  }

  // filter off
  RNOK( m_pcSEEncoderIf->processWienerFlag( false ) );
  RDCost cCostOff( pcCost->getLambda() );
  cCostOff.reset();
  cCostOff.addRate( m_pcBitRateEstIf->getBitRateEstimate() - cRootRate );
  cCostOff.addDistortion( cOrigPicBlock, rcEncPicBlock);

//  State::State *pcOffState = NULL;
  State *pcOffState = NULL;
  if( m_pcStateIf )
  {
    RNOK( m_pcStateIf->getState( pcOffState ) );
  }

//  RNOK( m_pcStateIf->setState( pcRootState ) );
  RNOK(  pcRootState->setState() );
  //filter on
  RNOK( m_pcSEEncoderIf->processWienerFlag( true ) );
  RDCost cCostOn( pcCost->getLambda() );;
  cCostOn.reset();
  cCostOn.addRate( m_pcBitRateEstIf->getBitRateEstimate() - cRootRate );
  cCostOn.addDistortion( cOrigPicBlock, rcFiltPicBlock);

  WienerBlockPar* pcBlockPar = m_rcQTBlockPars.getT( rcNav );
  if (pcBlockPar == NULL)
  {
    pcBlockPar = new WienerBlockPar();
    m_rcQTBlockPars.getT( rcNav ) = pcBlockPar;
  }

  if ( cCostOn < cCostOff)
  {
    *pcCost += cCostOn;
    rpResult->assign( rcFiltPicBlock);
    pcBlockPar->m_FilterBlock = true;
  }
  else
  {
    *pcCost += cCostOff;
    rpResult->assign( rcEncPicBlock );
    if ( m_pcStateIf )
    {
//      RNOK( m_pcStateIf->setState( pcOffState ) );
      RNOK(  pcOffState->setState() );
    }
    pcBlockPar->m_FilterBlock = false;
  }

  if (pcRootState)
     pcRootState->deleteState();

  if (pcOffState)
     pcOffState->deleteState();
  return Err::m_nOK;
}

template<class t_BlockPar>
ErrVal SingleNodeEstimatorWF<t_BlockPar>::setSingleResult( const QTNavi &rcNav, PicBlock<Pel> *&rpResult, PicBlock<Pel> *&rpDummyResult, PicBlock<Pel> *&rpNextDummyResult )
{
  PicBlock<Pel> cEncBlock( rcNav.getPrimPixelPos(), rcNav.getPrimBlockSize(), m_rcEncPic );
  cEncBlock.assign( *rpResult );
  delete rpResult;
  rpResult = NULL;
  return Err::m_nOK;
}

template<class t_BlockPar>
ErrVal SingleNodeEstimatorWF<t_BlockPar>::releaseChildNodes( const QTNavi &rcNav )
{
    RNOK( QuadtreeGroupPointerDeleter::deleteChildNodePointers(m_rcQTBlockPars, rcNav))
//	m_rcQTBlockPars.deleteChildNodePointers( rcNav );


  return Err::m_nOK;
}

template<class t_BlockPar>
ErrVal SingleNodeEstimatorWF<t_BlockPar>::releaseCurrentNode( const QTNavi &rcNav )
{

  delete(m_rcQTBlockPars.getT( rcNav ) );
  m_rcQTBlockPars.getT( rcNav ) = NULL;
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _SINGLENODEESTIMATORWF_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
