/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SLICEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SLICEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "RDCost.h"
#include "PredictionEstimator.h"
#include "ResidualEstimator.h"
#include "VideoCommonLib/AccessUnit.h"
#include "VideoCommonLib/CodedPictureStore.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN


class SliceEstimator
{
public:
  class Parameters
  {
  public:

    void setLambda( const AccessUnitParameters& rcAUPars,
                    PicOrderCnt                 iLastKeyPicPoc0,
                    PicOrderCnt                 iLastKeyPicPoc1,
                    PicOrderCnt                 iNextPoc,
                    UInt                        uiLambdaCalculationMode,
                    UInt                        uiQuantizationMode,
                    int                         iPrimQp )
    {
      double dLambda = 0.0;
      const double dQPX      = (double)( iPrimQp - QP_SHIFT1 );
      const double dQPFactor = pow( 2.0, dQPX / 3.0 );

      if( uiLambdaCalculationMode ) // JM-like method
      {
        const bool bIsKeyPic  = rcAUPars.getPicOrderCnt() >= iLastKeyPicPoc0;
        const bool bIsRefPic  = rcAUPars.getRefPicFlag();
        const bool bIntraRDOQ = rcAUPars.getMaxMvsPerBlk() == 0 && uiQuantizationMode;

        Int64 iNumNonKeyPic = 0;
        if( rcAUPars.getPicOrderCnt() > iLastKeyPicPoc0 )
        {
          iNumNonKeyPic = rcAUPars.getPicOrderCnt() - iLastKeyPicPoc0 - 1;
        }
        else if( bIsKeyPic ) // applies to first frame only
        {
          iNumNonKeyPic = iNextPoc - rcAUPars.getPicOrderCnt() - 1;
        }
        else
        {
          iNumNonKeyPic = iLastKeyPicPoc0 - iLastKeyPicPoc1 - 1;
        }

        double dBaseFactor = iNumNonKeyPic ? ( bIntraRDOQ ? 0.57 : 0.68 ) : ( bIntraRDOQ ? 0.71 : 0.85 );

        if( bIsKeyPic && bIsRefPic )
        { // key and reference picture
          dLambda = dBaseFactor * dQPFactor * gMax( 1.0 - 0.05 * (double)iNumNonKeyPic, 0.5 );
        }
        else if( bIsRefPic )
        { // non-key reference picture
          dLambda = dBaseFactor * dQPFactor * gMax( 1.0 - 0.05 * (double)iNumNonKeyPic, 0.5 ) * gMinMax( dQPX / 6.0, 2.0, 4.0 ) * 0.8;
        }
        else
        { // non-reference picture
          dLambda = dBaseFactor * dQPFactor                                                   * gMinMax( dQPX / 6.0, 2.0, 4.0 );
        }
      }
      else // old method
      {
        dLambda = 0.85 * dQPFactor;
      }
      m_cLambda = dLambda;
    }

    Lambda getLambda() const { return m_cLambda; }

  private:
    Lambda m_cLambda;
  };

  static ErrVal addIntraPars       ( VecPredEstInfo &racPars,
                                     const VecResEstInfo &racResEstInfo );
  static ErrVal addInterPars       ( VecPredEstInfo &racPars,
                                     UInt uiMvAccuracy,
                                     UInt uiSearchModeFullPel,
                                     UInt uiSearchModeSubPel,
                                     bool bPrimPlaneOnlySearch,
                                     BlockSize cSearchRange,
                                     BlockSize cIterSearchRange,
                                     Lambda cLambdaMotion,
                                     const VecResEstInfo &racResEstInfo,
                                     UInt uiMaxNumRefBlocksPerInterTranslationalPred,
                                     UInt uiMaxNumIterations );
  static ErrVal addInterPlaPaPars  ( VecPredEstInfo&      racPars,
                                     const VecResEstInfo& racResEstInfo,
                                     const PlanePackage*  pcRefPlanePackage );
  static ErrVal addResEstInfos     ( VecResEstInfo &racResEstInfo,
                                     const Ratio& cDeadzone,
                                     const std::vector<int> &raiQps,
                                     UInt uiMaxAbsQPDelta,
                                     UInt uiLog2MaxTransformBlockSize,
                                     UInt uiLog2MinTransformBlockSize,
                                     bool bTrafoPow2Subdivide,
                                     UInt uiQuantizers,
                                     bool bEnableResidualSkip );
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _SLICEESTIMATOR_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
