/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _STATEIF_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _STATEIF_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_


// video namespace begin
VIDEO_NAMESPACE_BEGIN


class State
{
protected:
  State() {}
  virtual ~State() {}

public:
  virtual ErrVal setState()    = 0;  // restores the state
  virtual ErrVal deleteState() = 0;  // deletes a state that was given by getState()

  virtual ErrVal setAndDeleteState()
  {
    RNOK( setState() );
    RNOK( deleteState() );
    return Err::m_nOK;
  }

};


class StateIf
{
protected:
  StateIf() {}
  virtual ~StateIf() {}

public:
  virtual ErrVal getState( State *&rpcState ) = 0;  // gives a pointer that represents the state of the implementing class
};


class StateMultiplier
  : public StateIf
{
public:
  virtual ~StateMultiplier() {}

  void addStateIf( StateIf* pcStateIf ) { m_apcStateIfs.push_back( pcStateIf ); }

  ErrVal getState( State *&rpcState )
  {
    StateMultiplierState* pcState = new StateMultiplierState( UInt( m_apcStateIfs.size() ) );
    for( UInt ui = 0; ui < UInt( m_apcStateIfs.size() ); ++ui )
    {
      m_apcStateIfs[ui]->getState( pcState->m_apcStates[ui] );
    }
    rpcState = pcState;
    return Err::m_nOK;
  }

private:
  std::vector<StateIf*> m_apcStateIfs;

  class StateMultiplierState : public State
  {
  public:
    StateMultiplierState( UInt uiNumStates ) : m_apcStates( uiNumStates ) {}

    ErrVal setState()
    {
      for( UInt ui = 0; ui < UInt( m_apcStates.size() ); ++ui )
      {
        m_apcStates[ui]->setState();
      }
      return Err::m_nOK;
    }

    ErrVal deleteState()
    {
      for( UInt ui = 0; ui < UInt( m_apcStates.size() ); ++ui )
      {
        m_apcStates[ui]->deleteState();
      }
      delete this;
      return Err::m_nOK;
    }

    ErrVal setAndDeleteState()
    {
      for( UInt ui = 0; ui < UInt( m_apcStates.size() ); ++ui )
      {
        m_apcStates[ui]->setAndDeleteState();
      }
      delete this;
      return Err::m_nOK;
    }

    std::vector<State*> m_apcStates;
  };
};


// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _STATEIF_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
