/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _SYNTAXELEMENTENCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _SYNTAXELEMENTENCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "VideoCommonLib/SyntaxElementCoder.h"
#include "VideoCommonLib/BinStreamIf.h"

#include "VideoCommonLib/TraceCommon.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN

typedef SyntaxElementCoderIf<true> SyntaxElementEncoderIf;

#if COLLECT_CABAC_BITS
typedef BinStreamIf<CtxBin&, true, EPCtxBin&> CtxBinStreamEncIf;
#else
typedef BinStreamIf<CtxBin&, true> CtxBinStreamEncIf;
#endif

template<class t_BinProcessor = BinStreamProcessor<CtxBin&, true, EPCtxBin&> >
class SyntaxElementEncoder
  : public SyntaxElementCoder<t_BinProcessor, true>
{
private:
  typedef SyntaxElementCoder<t_BinProcessor, true> BaseCall;
public:
  SyntaxElementEncoder( const CtxAccess &rcCtxAccess,
                        const ContextArrayStartIndices &rcCtxArrayStartIndices,
                        const PlanePackageParameters& rcPlaPaPars,
                        TraceCommon *pcTrace,
                        typename t_BinProcessor::CtorPars &rcBinProcessorPars )
    : SyntaxElementCoder<t_BinProcessor, true>( rcCtxAccess, rcCtxArrayStartIndices, rcPlaPaPars, pcTrace, rcBinProcessorPars ) {}
  ~SyntaxElementEncoder() {}

private:
  ErrVal xProcessUnarySymbol           ( const UInt &ruiSymbol,
                                         UInt uiCtxStart,
                                         UInt uiNumCtx,
                                         UInt uiSymbolMax );
  ErrVal xProcessDQp                   ( UInt uiDQpCtx,
                                         const CtxNames eContext,
                                         bool bNonZeroCoeffsPresent,
                                         const std::vector<int>& raiPredQp,
                                         const std::vector<int> & raiQp );
  ErrVal xProcessSigLastPow2           ( const Block<const int> &rcBlock,
                                         UInt uiPlane,
                                         int iLog2BlockSize,
                                         int iLog2PlaneBlockDim );
  ErrVal xProcessAbsGreaterOnePow2     ( const int &rcCoeff, UInt uiCtx );
  ErrVal xProcessCoeffLevelPow2        ( const int &rcCoeff,
                                         UInt uiCtx,
                                         UInt uiEpCtx );
  ErrVal xProcessSignPow2               ( const int &rcCoeff, UInt uiEpCtx );
  ErrVal xProcessPredMode               ( const bool bIsIntraOnlySlice,
                                          bool &rbIsIntra,
                                          const PredSyntax *const & rpcPredSyntax,
                                          const PredSyntax *pcLeftData,
                                          const PredSyntax *pcTopData );
  ErrVal xProcessResMode                ( const QTNavi &rcNav,
                                          bool bIsIntra,
                                          bool bLeftIsAvailable,
                                          bool bTopIsAvailable,
                                          ResInfo::ResidualMode eLeftMode,
                                          ResInfo::ResidualMode eTopMode,
                                          const SliceData &rcCurrSlice,
                                          const PlanePackageParameters &rcPlaPaPars,
                                          const ResInfo *const &rpcResInfo );
  ErrVal xProcessMvDComponent           ( const int &riMvDComponent,
                                          UInt uiLocalCtx,
                                          UInt uiCtxOffset );
  ErrVal xProcessInterPlaPaPredFlag     ( bool& rbIsInterPlaPaPred,
                                          const PredSyntax *const &rcPredSyntax,
                                          const QTNavi &rcPos,
                                          const PredSyntax* pcLeft,
                                          const PredSyntax* pcTop,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const AccessUnitParameters& rcAUPars );
  ErrVal xInferInterPlaPaPredFlag       ( bool& rbIsInterPlaPaPred,
                                          const PredSyntax *const &rcPredSyntax,
                                          const QTNavi &rcPos,
                                          const PredSyntax* pcLeft,
                                          const PredSyntax* pcTop,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const AccessUnitParameters& rcAUPars );
  ErrVal xProcessDirectModeFlag         ( bool& rbIsDirect,
                                          const PredSyntax *const &rcPredSyntax,
                                          const QTNavi &rcPos,
                                          const PredSyntax* pcLeft,
                                          const PredSyntax* pcTop,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const AccessUnitParameters& rcAUPars );
  ErrVal xProcessNumMvs                 ( const PredSyntaxInterTranslational& rcPredSyntax,
                                          UInt uiContextStart,
                                          UInt uiMaxNumContexts,
                                          UInt uiMaxNumMvs );
  ErrVal xProcessRefIdxs                ( const PredSyntaxInterTranslational& rcPredSyntax,
                                          UInt uiRefListSize );
  ErrVal xProcessMvPrecision            ( const PlanePackageParameters &rcPlaPaPars,
                                          const int &riPrecision );
  ErrVal xProcessMvPredictor            ( const QTNavi &rcNav,
                                          const PredSyntaxInterTranslational& rcPredSyntax,
                                          const PlanePackageParameters& rcPlaPaPars,
                                          const SliceData& rcSliceData,
                                          const UInt uiIdx );
  ErrVal xProcessResTrafoSkip           ( const QTNavi &rcPos,
                                          const SliceData &rcContextSlice,
                                          UInt uiNumPlanes,
                                          const ResInfoTrafoSkip &rcResInfo );
  ErrVal xWriteEpExGolomb               ( UInt uiSymbol, UInt uiCount, UInt uiEpCtx );
  ErrVal xProcessInterTranslational       ( const QTNavi &rcPos,
                                          const PlanePackageParameters &rcPlaPaPars,
                                          const AccessUnitParameters &rcAUPars,
                                          const PredSyntax *const &rpcPredSyntax,
                                          const SliceData& rcSliceData,
                                          const PredSyntax *pcLeft,
                                          const PredSyntax *pcTop );

  ErrVal xProcessWienerParameters       ( const WienerFilterParameters &rcWFParam , UInt uiNumOfPlanes );
  ErrVal xProcessWienerFilterLength     ( const UInt &ruiFilterLength ) ;

  ErrVal xProcessWienerFilterCoeffs     (const FilterCoefficients& rcCoeffs) ;
};


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessSigLastPow2( const Block<const int>&  rcBlock,
                                                                  UInt uiPlane,
                                                                  int iLog2PrimaryBlockSize,
                                                                  int iLog2PlaneBlockDim )
{
  ROT( iLog2PlaneBlockDim > iLog2PrimaryBlockSize );

  const int uiPlaneBlkSize = 1 << iLog2PlaneBlockDim;
  const int iStride = rcBlock.getStride();
  UInt uiNumSig = 0;
  {
    const int* pi = rcBlock.getPtr();
    const int* piEnd = pi + iStride * uiPlaneBlkSize;
    for( ; pi < piEnd; pi+= iStride  )
    {
      for( int x = 0; x < uiPlaneBlkSize; x++ )
      {
        if( pi[x] )
          uiNumSig++;
      }
    }
  }
  ROF( uiNumSig );
  const UInt  uiMaxNumCoeff     = uiPlaneBlkSize * uiPlaneBlkSize;
  const UInt  uiMaxNumCoeffM1   = uiMaxNumCoeff  - 1;
  UInt        uiDownLeft        = 1;
  UInt        uiNumSigTopRight  = 0;
  UInt        uiNumSigBotLeft   = 0;

  std::vector<std::vector<UInt> >&  raauiScanPattern = BaseCall::m_aaauiScanPattern[ iLog2PlaneBlockDim ];

  for( UInt uiScanPos = 0; uiScanPos < uiMaxNumCoeffM1; uiScanPos++ )
  {
    UInt  uiBlkPos  = raauiScanPattern[ uiDownLeft ][ uiScanPos ];
    UInt  uiPosY    = uiBlkPos >> iLog2PlaneBlockDim;
    UInt  uiPosX    = uiBlkPos - ( uiPosY << iLog2PlaneBlockDim );

    //===== code significance flag =====
    bool  bSig          = rcBlock.getEl( PixelPos( uiPosX, uiPosY ) ) != 0;
    UInt  uiCtxSigBase  = BaseCall::m_cCtxArrayStartIndices.m_acResCtx[ uiPlane ].m_uiStartTrafoPow2Sig + UInt( iLog2PrimaryBlockSize );
    UInt  uiCtxSigInc   = ResTrafoPow2::getSigCtxInc( rcBlock, uiPosX, uiPosY, UInt( iLog2PlaneBlockDim ), ( uiDownLeft > 0 ) );
    UInt  uiCtxSig      = BaseCall::m_cCtxAccess.getContextOffset( uiCtxSigBase, uiCtxSigInc );
    RNOK( BaseCall::xProcessBin( bSig, uiCtxSig ) );
    TRACE_NUMBER;
    TRACE_TO_STR( "Sig: " );
    TRACE_TO_STR( bSig );
    TRACE_TO_STR( std::endl );

    if( bSig )
    {
      uiNumSig--;
      if( uiPosX > uiPosY )
      {
        uiNumSigTopRight++;
      }
      else if( uiPosY > uiPosX )
      {
        uiNumSigBotLeft++;
      }

      //===== code last flag =====
      bool  bLast         = ( uiNumSig == 0 );
      UInt  uiCtxLastBase = BaseCall::m_cCtxArrayStartIndices.m_acResCtx[ uiPlane ].m_uiStartTrafoPow2Last + UInt( iLog2PrimaryBlockSize );
      UInt  uiCtxLastInc  = ResTrafoPow2::getLastCtxInc( rcBlock, uiPosX, uiPosY, UInt( iLog2PlaneBlockDim ) );
      UInt  uiCtxLast     = BaseCall::m_cCtxAccess.getContextOffset( uiCtxLastBase, uiCtxLastInc );
      RNOK( BaseCall::xProcessBin( bLast, uiCtxLast ) );
      TRACE_NUMBER;
      TRACE_TO_STR( "Last: " );
      TRACE_TO_STR( bLast );
      TRACE_TO_STR( std::endl );

      if( bLast )
      {
        break;
      }
    }

    //===== update scan direction =====
    if( ( uiDownLeft == 1 && ( uiPosX == 0 || uiPosY+1 == (UInt)uiPlaneBlkSize ) ) ||
        ( uiDownLeft == 0 && ( uiPosY == 0 || uiPosX+1 == (UInt)uiPlaneBlkSize ) )   )
    {
      uiDownLeft = ( uiNumSigTopRight >= uiNumSigBotLeft ? 1 : 0 );
    }
  }

  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessAbsGreaterOnePow2( const int &rcCoeff, UInt uiCtx )
{
  RNOK( BaseCall::xProcessBin( abs( rcCoeff ) > 1, uiCtx ) );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xWriteEpExGolomb( UInt uiSymbol, UInt uiCount, UInt uiEpCtx )
{
  while( uiSymbol >= (UInt)(1<<uiCount) )
  {
    RNOK( BaseCall::xProcessEqProbBin( true, uiEpCtx ) );
    uiSymbol -= 1<<uiCount;
    uiCount  ++;
  }
  RNOK( BaseCall::xProcessEqProbBin( false, uiEpCtx ) );
  while( uiCount-- )
  {
    RNOK( BaseCall::xProcessEqProbBin( (uiSymbol>>uiCount) & 1, uiEpCtx ) );
  }

  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessCoeffLevelPow2( const int &rcCoeff,
                                                                     UInt uiCtx,
                                                                     UInt uiEpCtx )
{
  UInt uiSymbol = abs( rcCoeff ) - 2;
  if( uiSymbol )
  {
    RNOK( BaseCall::xProcessBin( 1, uiCtx ) );
    UInt uiCount = 0;
    bool bNoExGo = (uiSymbol < 13);

    while( --uiSymbol && ++uiCount < 13 )
    {
      RNOK( BaseCall::xProcessBin( 1, uiCtx ) );
    }
    if( bNoExGo )
    {
      RNOK( BaseCall::xProcessBin( 0, uiCtx ) );
    }
    else
    {
      RNOK( xWriteEpExGolomb( uiSymbol, 0, uiEpCtx ) );
    }
  }
  else
  {
    RNOK( BaseCall::xProcessBin( 0, uiCtx ) );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessSignPow2( const int &rcCoeff, UInt uiEpCtx )
{
  RNOK( BaseCall::xProcessEqProbBin( rcCoeff < 0, uiEpCtx ) );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessUnarySymbol( const UInt &ruiSymbol,
                                                                  UInt uiCtxStart,
                                                                  UInt uiNumCtx,
                                                                  UInt uiSymbolMax )
{
  ROT( uiNumCtx == 0 );
  ROT( ruiSymbol > uiSymbolMax );

  TRACE_NUMBER;
  TRACE_TO_STR( "Unary: " );

  const UInt uiNumBins = gMin( ruiSymbol + 1, uiSymbolMax );
  for( UInt ui = 0; ui < uiNumBins; ui++ )
  {
    const UInt uiCtx = uiCtxStart + gMin( ui, uiNumCtx - 1 );
    RNOK( BaseCall::xProcessBin( ruiSymbol != ui ? 1 : 0, uiCtx ) );
    TRACE_TO_STR( ruiSymbol != ui ? 1 : 0 );
    TRACE_TO_STR( "," );
    TRACE_TO_STR( uiCtx );
    TRACE_TO_STR( " " );
  }

  TRACE_TO_STR( std::endl  );

  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessDQp( UInt uiDQpCtx,
                                                          const CtxNames eContext,
                                                          bool bNonZeroCoeffsPresent,
                                                          const std::vector<int>& raiPredQp,
                                                          const std::vector<int>& raiQp )
{
  ROF( raiPredQp.size() == raiQp.size() );
  int iDQp = 0;
  if( bNonZeroCoeffsPresent )
  {
    iDQp = raiQp[ 0 ] - raiPredQp[ 0 ];
    RNOK( BaseCall::xProcessBin( iDQp != 0, BaseCall::m_cCtxAccess.getContextOffset( eContext, uiDQpCtx ) ) );
    if( iDQp != 0 )
    {
      UInt uiDQp = (UInt)( iDQp > 0 ? ( 2 * iDQp - 2 ) : ( -2 * iDQp - 1 ) );
      RNOK( xProcessUnarySymbol( uiDQp, BaseCall::m_cCtxAccess.getContextOffset( eContext, 3 ), 2, TypeLimits<UInt>::m_iMax ) );
    }
  }

  for( UInt uiPlane = 0; uiPlane < raiQp.size(); uiPlane++ )
  {
    ROF( raiQp[ uiPlane ] == raiPredQp[ uiPlane ] + iDQp );
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessPredMode( const bool bIsIntraOnlySlice,
                                                               bool &rbIsIntra,
                                                               const PredSyntax *const &rpcPredSyntax,
                                                               const PredSyntax *pcLeftData,
                                                               const PredSyntax *pcTopData )
{
  UInt  uiLeftIntra = ( pcLeftData && pcLeftData->isIntraCoded() ? 1 : 0 );
  UInt  uiTopIntra  = ( pcTopData  && pcTopData ->isIntraCoded() ? 1 : 0 );
  rbIsIntra         = rpcPredSyntax->isIntraCoded();

  if( !bIsIntraOnlySlice )
  {
    UInt uiIntraCtx = 2 + uiLeftIntra + uiTopIntra;
    RNOK( BaseCall::xProcessBin( rbIsIntra, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, uiIntraCtx ) ) );
    TRACE_NUMBER;
    TRACE_TO_STR( "intra_flag: " );
    TRACE_TO_STR( rbIsIntra );
    TRACE_TO_STR( std::endl );
  }
  else
  {
    ROF( rbIsIntra );
  }

  if( rbIsIntra )
  {
    UInt  uiIntraIdx    = 0;
    UInt  uiLeftModeIdx = 0;
    UInt  uiTopModeIdx  = 0;
    if( uiLeftIntra )
    {
      RNOK( BaseCall::xGetIntraModeIdx( *pcLeftData, uiLeftModeIdx ) );
    }
    if( uiTopIntra )
    {
      RNOK( BaseCall::xGetIntraModeIdx( *pcTopData,  uiTopModeIdx  ) );
    }
    RNOK( BaseCall::xGetIntraModeIdx( *rpcPredSyntax, uiIntraIdx  ) );

    UInt uiMostProbable = 2; // assume INTRA_DC
    if( uiLeftIntra && uiTopIntra )
    {
      uiMostProbable = gMin( uiLeftModeIdx, uiTopModeIdx );
    }

    UInt uiMostProbableFlagCtx = 0;
    if( uiLeftIntra && uiTopIntra )
    {
      uiMostProbableFlagCtx = 5 + ( uiLeftModeIdx == uiTopModeIdx ? 0 : 1 );
    }

    RNOK( BaseCall::xProcessBin( uiMostProbable == uiIntraIdx, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, uiMostProbableFlagCtx ) ) );

    TRACE_NUMBER;
    TRACE_TO_STR( "most_probable_ipred_mode_flag: " );
    TRACE_TO_STR( uiMostProbable == uiIntraIdx );
    TRACE_TO_STR( std::endl );

    if( uiMostProbable != uiIntraIdx )
    {
      UInt uiIntraIdxBits = uiIntraIdx;
      if( uiIntraIdxBits > uiMostProbable )
      {
        uiIntraIdxBits--;
      }
      ROT( uiIntraIdxBits > 7 );
      RNOK( BaseCall::xProcessBin( (uiIntraIdxBits     ) & 1, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, 1 ) ) );
      RNOK( BaseCall::xProcessBin( (uiIntraIdxBits >> 1) & 1, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, 1 ) ) );
      RNOK( BaseCall::xProcessBin( (uiIntraIdxBits >> 2) & 1, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_PRED_MODE_INTRA, 1 ) ) );

      TRACE_NUMBER;
      TRACE_TO_STR( "intra_pred_mode: " );
      TRACE_TO_STR( uiIntraIdxBits );
      TRACE_TO_STR( std::endl );
    }

    if( rpcPredSyntax->getCodingMode() == PredInfo::INTRA_DIRECTIONAL )
    {
      TRACE_NUMBER;
      TRACE_TO_STR( "INTRA_DIRECTIONAL: " );
      TRACE_TO_STR( rpcPredSyntax->getPredSyntaxIntraDirectional()->getIntraDirection() );
      TRACE_TO_STR( std::endl );

      RNOK( BaseCall::xProcessBin( rpcPredSyntax->getPredSyntaxIntraDirectional()->getSmoothedRefPixelFlag(),
                                   BaseCall::m_cCtxAccess.getContextOffset( CTX_START_INTRA_SMOOTHED, rpcPredSyntax->getPredSyntaxIntraDirectional()->getIntraDirection() ) ) );
      TRACE_NUMBER;
      TRACE_TO_STR( "smooth_ref_samples_flag: " );
      TRACE_TO_STR( rpcPredSyntax->getPredSyntaxIntraDirectional()->getSmoothedRefPixelFlag() );
      TRACE_TO_STR( std::endl );
    }
    else if( rpcPredSyntax->getCodingMode() == PredInfo::INTRA_DC )
    {
      TRACE_NUMBER;
      TRACE_TO_STR( "INTRA_DC " );
      TRACE_TO_STR( std::endl );
    }
    else
    {
      TRACE_NUMBER;
      TRACE_TO_STR( "INTRA_PLANE " );
      TRACE_TO_STR( std::endl );
    }
  }
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessResMode( const QTNavi &rcNav,
                                                              bool bIsIntra,
                                                              bool bLeftIsAvailable,
                                                              bool bTopIsAvailable,
                                                              ResInfo::ResidualMode eLeftMode,
                                                              ResInfo::ResidualMode eTopMode,
                                                              const SliceData &rcCurrSlice,
                                                              const PlanePackageParameters &rcPlaPaPars,
                                                              const ResInfo *const &rpcResInfo )
{
  // possible transform modes are given in access unit header
  const ResInfo::ResidualMode eResMode = rpcResInfo->getResidualMode();
  UInt uiNumQuantizers = rcPlaPaPars.getNumQuantizers();
  UInt uiQuantizers    = rcPlaPaPars.getTransformMode();
  if( uiNumQuantizers > 1 && (uiQuantizers & ResInfo::RES_TRAFO_SKIP) )
  {
    if( !bIsIntra )
    {
      UInt uiSkipCtx = 0;
      RNOK( BaseCall::xGetSkipModeCtx( bIsIntra, bLeftIsAvailable, bTopIsAvailable, eLeftMode, eTopMode, uiSkipCtx  ) );
      RNOK( BaseCall::xProcessBin( eResMode == ResInfo::RES_TRAFO_SKIP, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_RES_SKIP_MODE, uiSkipCtx ) ) );

      TRACE_NUMBER;
      TRACE_TO_STR( "trafo_skip_flag: " );
      TRACE_TO_STR( eResMode == ResInfo::RES_TRAFO_SKIP );
      TRACE_TO_STR( std::endl );

      ROTRS( eResMode == ResInfo::RES_TRAFO_SKIP, Err::m_nOK );
    }
    else
    {
      ROT( eResMode == ResInfo::RES_TRAFO_SKIP );
    }
    uiNumQuantizers--;
    uiQuantizers &= ~UInt( ResInfo::RES_TRAFO_SKIP );
  }
  
  ROT( uiNumQuantizers > 1 );

  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessResTrafoSkip( const QTNavi &rcPos,
                                                                   const SliceData &rcContextSlice,
                                                                   UInt uiNumPlanes,
                                                                   const ResInfoTrafoSkip &rcResInfo )
{
  for( UInt ui = 0; ui < uiNumPlanes; ui++ )
  {
    int iPredQp = 0;
    RNOK( rcContextSlice.derivePredQP( rcPos, ui, iPredQp ) );
    ROF ( iPredQp == rcResInfo.getQps()[ui] );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessInterTranslational( const QTNavi &rcPos,
                                                                       const PlanePackageParameters &rcPlaPaPars,
                                                                       const AccessUnitParameters &rcAUPars,
                                                                       const PredSyntax *const &rpcPredSyntax,
                                                                       const SliceData &rcSliceData,
                                                                       const PredSyntax *pcLeft,
                                                                       const PredSyntax *pcTop )
{
  ROF( rpcPredSyntax->getCodingMode() == PredInfo::INTER_TRANSLATIONAL );
  const PredSyntaxInterTranslational* pcTransCurr = rpcPredSyntax->getPredSyntaxInterTranslational();
  const PredSyntaxInterTranslational* pcTransLeft = ( pcLeft ? pcLeft->getPredSyntaxInterTranslational() : 0 );
  const PredSyntaxInterTranslational* pcTransTop  = ( pcTop  ? pcTop ->getPredSyntaxInterTranslational() : 0 );
  RNOK(( SyntaxElementCoder<t_BinProcessor,true>::xProcessInterTranslational( rcPos, rcPlaPaPars, rcAUPars, *pcTransCurr, rcSliceData, pcTransLeft, pcTransTop ) ));
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessMvDComponent( const int &riMvDComponent,
                                                                   UInt uiLocalCtx,
                                                                   UInt uiCtxOffset )
{
  const UInt  uiAbsDiff = abs( riMvDComponent );
  UInt        uiValue   = 0;
  bool        bSymbol   = false;
  // unary part
  do
  {
    bSymbol = ( uiAbsDiff > uiValue );
    RNOK( BaseCall::xProcessBin( bSymbol, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MVD, uiCtxOffset + uiLocalCtx ) ) );
    if( uiValue < BaseCall::m_uiNumMvDUnaryCtxPerComponent )
    {
      uiLocalCtx = 3 + uiValue;
    }
    uiValue++;
  } while( bSymbol && uiValue < 9 );
  // exp-golomb part
  if( bSymbol )
  {
    RNOK( xWriteEpExGolomb( uiAbsDiff - uiValue, 3, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MVD ) ) );
  }
  // sign
  if( uiAbsDiff != 0)
  {
    RNOK( BaseCall::xProcessEqProbBin( riMvDComponent < 0, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MVD ) ) );
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessInterPlaPaPredFlag ( bool &rbIsInterPlaPaPred,
                                                                          const PredSyntax *const &rpcPredSyntax,
                                                                          const QTNavi &rcPos,
                                                                          const PredSyntax *pcLeft,
                                                                          const PredSyntax *pcTop,
                                                                          const PlanePackageParameters &rcPlaPaPars,
                                                                          const AccessUnitParameters &rcAUPars )
{
  rbIsInterPlaPaPred = false;
  ROFRS( rcPlaPaPars.getInterPlaPaPredFlag(),         Err::m_nOK );
  ROFRS( rcPlaPaPars.getAdaptiveInterPlaPaPredFlag(), Err::m_nOK );
  rbIsInterPlaPaPred  = (     rpcPredSyntax->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE );
  UInt  uiLeft        = ( !pcLeft || pcLeft->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE ? 1 : 0 );
  UInt  uiTop         = ( !pcTop  || pcTop ->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE ? 1 : 0 );
  UInt  uiCtx         = uiLeft + uiTop;
  RNOK( BaseCall::xProcessBin( rbIsInterPlaPaPred, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_INTER_PLAPA_PRED, uiCtx ) ) );
  TRACE_NUMBER;
  TRACE_TO_STR( "inter_pla_pa_pred_flag: " );
  TRACE_TO_STR( rbIsInterPlaPaPred );
  TRACE_TO_STR( std::endl );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xInferInterPlaPaPredFlag( bool &rbIsInterPlaPaPred,
                                                                       const PredSyntax *const &rpcPredSyntax,
                                                                       const QTNavi &rcPos,
                                                                       const PredSyntax *pcLeft,
                                                                       const PredSyntax *pcTop,
                                                                       const PlanePackageParameters &rcPlaPaPars,
                                                                       const AccessUnitParameters &rcAUPars )
{
  ROF( rcAUPars.getMaxMvsPerBlk() > 0 );
  rbIsInterPlaPaPred = ( rcPlaPaPars.getInterPlaPaPredFlag() && !rcPlaPaPars.getAdaptiveInterPlaPaPredFlag() );
  ROF( ( rpcPredSyntax->getCodingMode() == PredInfo::INTER_PLANE_PACKAGE ) == rbIsInterPlaPaPred );
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessDirectModeFlag( bool &rbIsDirect,
                                                                     const PredSyntax *const &rpcPredSyntax,
                                                                     const QTNavi &rcPos,
                                                                     const PredSyntax *pcLeft,
                                                                     const PredSyntax *pcTop,
                                                                     const PlanePackageParameters &rcPlaPaPars,
                                                                     const AccessUnitParameters &rcAUPars )
{
  ROT( ! rcAUPars.isDirectModeFlagPresent( rcPlaPaPars.getPlaPaId() ) && rpcPredSyntax->getPredSyntaxInterTranslational() && rpcPredSyntax->getPredSyntaxInterTranslational()->getDirectModeFlag() );
  ROFRS( rcAUPars.isDirectModeFlagPresent( rcPlaPaPars.getPlaPaId() ), Err::m_nOK );
  rbIsDirect    = ( rpcPredSyntax->getPredSyntaxInterTranslational() && rpcPredSyntax->getPredSyntaxInterTranslational()->getDirectModeFlag() );
  UInt  uiLeft  = ( !pcLeft ? 0 : ( !pcLeft->getPredSyntaxInterTranslational() ? 1 : ( pcLeft->getPredSyntaxInterTranslational()->getDirectModeFlag() ? 0 : 1 ) ) );
  UInt  uiTop   = ( !pcTop  ? 0 : ( !pcTop ->getPredSyntaxInterTranslational() ? 1 : ( pcTop ->getPredSyntaxInterTranslational()->getDirectModeFlag() ? 0 : 1 ) ) );
  UInt  uiCtx   = uiLeft + uiTop;
  RNOK( BaseCall::xProcessBin( rbIsDirect, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_DIRECT_MODE, uiCtx ) ) );
  TRACE_NUMBER;
  TRACE_TO_STR( "direct_mode_flag: " );
  TRACE_TO_STR( rbIsDirect );
  TRACE_TO_STR( std::endl );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessNumMvs( const PredSyntaxInterTranslational& rcPredSyntax,
                                                             UInt uiContextStart,
                                                             UInt uiMaxNumContexts,
                                                             UInt uiMaxNumMvs )
{
  RNOK( xProcessUnarySymbol( rcPredSyntax.getNumEntries() - 1, uiContextStart, uiMaxNumContexts, uiMaxNumMvs - 1 ) );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessRefIdxs( const PredSyntaxInterTranslational&  rcPredSyntax,
                                                              UInt uiRefListSize )
{
  RNOK( xProcessUnarySymbol( rcPredSyntax.getRefIdx( 0 ),
                             BaseCall::m_cCtxAccess.getContextOffset ( BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ),
                             BaseCall::m_cCtxAccess.getNumCtxElements( BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ),
                             uiRefListSize - 1 ) );
  if( uiRefListSize - 1 > 0 )
  {
    TRACE_NUMBER;
    TRACE_TO_STR( "ref_idx: " );
    TRACE_TO_STR( rcPredSyntax.getRefIdx( 0 ) );
    TRACE_TO_STR( std::endl );
  }

  for( UInt uiIdx = 1; uiIdx < rcPredSyntax.getNumEntries(); uiIdx++ )
  {
    UInt  uiLastRefIdx  = rcPredSyntax.getRefIdx( uiIdx - 1 );
    UInt  uiDeltaRefIdx = rcPredSyntax.getRefIdx( uiIdx     ) - uiLastRefIdx;
    UInt  uiRemListSize = uiRefListSize - uiLastRefIdx;
    UInt  uiSymbol      = ( uiDeltaRefIdx > 0 ? uiDeltaRefIdx : uiRemListSize ) - 1;
    RNOK( xProcessUnarySymbol( uiSymbol,
                               BaseCall::m_cCtxAccess.getContextOffset ( BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ) + uiLastRefIdx,
                               BaseCall::m_cCtxAccess.getNumCtxElements( BaseCall::m_cCtxArrayStartIndices.m_uiStartRefIdx ) - uiLastRefIdx,
                               uiRemListSize - 1 ) );
    if( uiRemListSize - 1 > 0 )
    {
      TRACE_NUMBER;
      TRACE_TO_STR( "delta_ref_idx_minus1: " );
      TRACE_TO_STR( uiSymbol );
      TRACE_TO_STR( std::endl );
    }
  }
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessMvPrecision( const PlanePackageParameters& rcPlaPaPars, const int& riPrecision )
{
  ROF( riPrecision == (int)rcPlaPaPars.getMvAccuracy() );
  return Err::m_nOK;
}


template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessMvPredictor( const QTNavi &rcNav,
                                                                  const PredSyntaxInterTranslational& rcPredSyntax,
                                                                  const PlanePackageParameters& rcPlaPaPars,
                                                                  const SliceData& rcSliceData,
                                                                  const UInt uiIdx )
{
  ROFRS( rcPlaPaPars.getMvCompetition(), Err::m_nOK );
  const QuadtreeGroup<PredInfo*> &rcTree( rcSliceData.getPredInfoTree() );
  UInt  uiNumMvPred = 0;
  UInt  auiCtx[2]   = { 0, 0 }; 
  RNOK ( InterTranslational::getNumMvPred( rcPredSyntax.getRefIdx( uiIdx ), rcNav, rcTree, uiNumMvPred, auiCtx ) );
  ROF  ( rcPredSyntax.getMvPredIdx( uiIdx ) < uiNumMvPred );
  if( uiNumMvPred > 1 )
  {
    bool  bFlag = ( rcPredSyntax.getMvPredIdx( uiIdx ) > 0 );
    RNOK( BaseCall::xProcessBin( bFlag, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MV_PRED_IDX, auiCtx[0] ) ) );
    if( bFlag && uiNumMvPred > 2 )
    {
      bFlag = ( rcPredSyntax.getMvPredIdx( uiIdx ) > 1 );
      RNOK( BaseCall::xProcessBin( bFlag, BaseCall::m_cCtxAccess.getContextOffset( CTX_START_MV_PRED_IDX, auiCtx[1] ) ) );
    }
  }
  TRACE_NUMBER;
  TRACE_TO_STR( "mv_predictor_idc: " );
  TRACE_TO_STR( rcPredSyntax.getMvPredIdx( uiIdx ) );
  TRACE_TO_STR( std::endl );
  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessWienerParameters( const WienerFilterParameters &rcWFParam , UInt uiNumOfPlanes )
{
  if ( rcWFParam.getUseFilterFlag() )
  {
    BaseCall::xProcessWienerSeparableFlag( true );
    TRACE_NUMBER;
    TRACE_TO_STR( "quadtree_based_inloop_filter_separable_flag: " );
    TRACE_TO_STR( true );
    TRACE_TO_STR( std::endl );

    BaseCall::xProcessWienerFlagForQuadtreeUse( rcWFParam.getUseBlocksFlag() );
    TRACE_NUMBER;
    TRACE_TO_STR( "quadtree_based_inloop_filter_quadtree_used_flag: " );
    TRACE_TO_STR( rcWFParam.getUseBlocksFlag() );
    TRACE_TO_STR( std::endl   );

    if ( rcWFParam.getUseBlocksFlag() )
    {
      BaseCall::xProcessWienerBlockSize(rcWFParam.getPower2MaxBlock(),rcWFParam.getPower2MinBlock());
      TRACE_NUMBER;
      TRACE_TO_STR( "quadtree_based_inloop_filter_log2_blocksize_max: " );
      TRACE_TO_STR( rcWFParam.getPower2MaxBlock() );
      TRACE_TO_STR( std::endl );
      TRACE_NUMBER;
      TRACE_TO_STR( "quadtree_based_inloop_filter_log2_blocksize_min: " );
      TRACE_TO_STR( rcWFParam.getPower2MinBlock() );
      TRACE_TO_STR( std::endl );
    }
    for( UInt uiPlane=0; uiPlane < uiNumOfPlanes ; uiPlane++)
    {
      UInt uiSet = 0;
      if ( rcWFParam.getFilterPlaneFlag( uiPlane ) )
      {
        uiSet = 1;
        UInt uiSetIdx = 0;
        while( uiSetIdx < rcWFParam.getSetNumOfPlane( uiPlane) )
        {
          if( rcWFParam.getNumberOfPlanesWhichUsesSet( uiSetIdx ) > 0 )
          {
            uiSet++;
          }
          uiSetIdx++;
        }
      }
      BaseCall::xProcessWienerPlaneSet( uiSet );
      TRACE_TO_STR( "quadtree_based_inloop_filter_allowed_modes_idx[ " );
      TRACE_TO_STR( uiPlane ) ;
      TRACE_TO_STR( " ]: " ) ;
      TRACE_TO_STR( uiSet );
      TRACE_TO_STR( std::endl );
    }

    for ( UInt uiSetNum=0; uiSetNum < rcWFParam.getNewSetNumberSep(); uiSetNum++ )
    {
      if( rcWFParam.getNumberOfPlanesWhichUsesSet( uiSetNum ) > 0 )
      {
        xProcessWienerFilterLength( rcWFParam.getHorizontalFilterCoeff( uiSetNum ).getFilterLength() );
        TRACE_TO_STR( "filter_length_minus1_div2[ " );
        TRACE_TO_STR( uiSetNum ) ;
        TRACE_TO_STR( " ][ 0 ]: " ) ;
        TRACE_TO_STR( ( rcWFParam.getHorizontalFilterCoeff( uiSetNum ).getFilterLength() - 1 ) / 2 );
        TRACE_TO_STR( std::endl );

        xProcessWienerFilterLength( rcWFParam.getVerticalFilterCoeff( uiSetNum ).getFilterLength() );
        TRACE_TO_STR( "filter_length_minus1_div2[ " );
        TRACE_TO_STR( uiSetNum ) ;
        TRACE_TO_STR( " ][ 1 ]: " ) ;
        TRACE_TO_STR( ( rcWFParam.getVerticalFilterCoeff( uiSetNum ).getFilterLength() - 1 ) / 2 );
        TRACE_TO_STR( std::endl );

        BaseCall::xProcessWienerFilterType( rcWFParam.getHorizontalFilterCoeff( uiSetNum ).getFilterType() );
        TRACE_TO_STR( "quadtree_based_inloop_filter_type[ " );
        TRACE_TO_STR( uiSetNum ) ;
        TRACE_TO_STR( " ]: " ) ;
        TRACE_TO_STR( rcWFParam.getHorizontalFilterCoeff( uiSetNum ).getFilterType() );
        TRACE_TO_STR( std::endl );

        xProcessWienerFilterCoeffs( rcWFParam.getHorizontalFilterCoeff( uiSetNum ) );
        xProcessWienerFilterCoeffs( rcWFParam.getVerticalFilterCoeff( uiSetNum ) );
      }
    }
  }

  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessWienerFilterLength(const UInt& ruiFilterLength)
{
  UInt iFilterLengthIdx = (ruiFilterLength-1)/2 ;

  UInt uiMaxIdxOfFilterLength = 7;
  xProcessUnarySymbol( iFilterLengthIdx, BaseCall::m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_LENGTH), 1, uiMaxIdxOfFilterLength );

  return Err::m_nOK;
}

template<class t_BinProcessor>
ErrVal SyntaxElementEncoder<t_BinProcessor>::xProcessWienerFilterCoeffs(const FilterCoefficients& rcCoeffs)
{
  //UInt uiMaxCoeff = 8192 ;
  for (UInt i = 0; i < rcCoeffs.getNumberOfCoeffs() ; i++)
  {
    int     thisCoeff   = rcCoeffs.getPredQuantCoeffTransmit(i);
    bool    thisSign    = (thisCoeff  < 0 ) ? 1:0 ;
    UInt    thisAbsCoeff           = abs(thisCoeff);
    RNOK( xProcessBin( thisSign, BaseCall::m_cCtxAccess.getContextOffset( CTX_WIENER_FILTER_COEFFS_SIGN ) ) );
    TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_sign[ i ][ hv ][ " );
    TRACE_TO_STR( i ) ;
    TRACE_TO_STR( " ]: " ) ;
    TRACE_TO_STR( thisSign ? 1 : 0 );
    TRACE_TO_STR( std::endl );
    xWriteEpExGolomb( thisAbsCoeff, 14, BaseCall::m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_ABS_COEFFS) );
    TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_abs[ i ][ hv ][ " );
    TRACE_TO_STR( i ) ;
    TRACE_TO_STR( " ]: " ) ;
    TRACE_TO_STR( thisAbsCoeff );
    TRACE_TO_STR( std::endl );
  }

  int iDC = rcCoeffs.getPredQuantDcTransmit();
  bool bDcSign = ( iDC < 0 ) ? 1 : 0;
  UInt uiDcAbs = abs( iDC );
  RNOK( xProcessBin( bDcSign, BaseCall::m_cCtxAccess.getContextOffset( CTX_WIENER_FILTER_COEFFS_SIGN ) ) );
  TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_dc_sign[ i ][ hv ]: " );
  TRACE_TO_STR( bDcSign ? 1 : 0 );
  TRACE_TO_STR( std::endl );
  xWriteEpExGolomb( uiDcAbs, 14, BaseCall::m_cCtxAccess.getContextOffset(CTX_WIENER_FILTER_ABS_COEFFS) );
  TRACE_TO_STR( "quadtree_based_inloop_filter_coeff_dc_abs[ i ][ hv ]: " );
  TRACE_TO_STR( uiDcAbs );
  TRACE_TO_STR( std::endl );

  return Err::m_nOK ;
}

// video namespace end
VIDEO_NAMESPACE_END


#endif // #ifndef _SYNTAXELEMENTENCODER_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
