/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VLCENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_VLCENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_


#include "VideoCommonLib/InStreamIf.h"
#include "VideoCommonLib/BitWriterIf.h"
#include "VideoCommonLib/VLCodingTables.h"
#include "VLCodewordBuffer.h"
#include "VLCStateEncoder.h"
#include "ProbabilityAllocatorEnc.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_ValType>
class VLCEncoder
  : public  ProbBinStreamEncIf
  , public  CabaCoderStartFinishIf
  , private VLCodewordBuffer<t_ValType>
{
  DISABLE_DEFAULT_CPYCTOR_AND_ASOP( VLCEncoder );

public:
  VLCEncoder( BitWriterIf<UInt>& rcBitWriterIf, const UInt uiMaxDelayInBits );
  virtual ~VLCEncoder();

  ErrVal  startCabac ();
  ErrVal  finishCabac();
  ErrVal  pushBack   ( const ProbBin& rcProbBin );
  ErrVal  pushBack   ( const bool&    rbBin     );

private:
  const UInt                    m_uiNumBaseEncoders;
  const UInt*                   m_pauiStateToIdxMap;
  const UInt                    m_uiStateZeroIdx;
  VLCStateEncoder<t_ValType>**  m_papcVLCStateEncoders;
};


template<class t_ValType>
FORCEINLINE
VLCEncoder<t_ValType>::VLCEncoder( BitWriterIf<UInt>& rcBitWriterIf, const UInt uiMaxDelayInBits )
: VLCodewordBuffer<t_ValType> ( rcBitWriterIf, VLCodingTables<t_ValType,true>::getNumVLCTables() )
, m_uiNumBaseEncoders         ( VLCodingTables<t_ValType,true>::getNumVLCTables() )
, m_pauiStateToIdxMap         ( VLCodingTables<t_ValType,true>::getState2IdxMap() )
, m_uiStateZeroIdx            ( m_pauiStateToIdxMap[ 0 ] )
, m_papcVLCStateEncoders      ( new VLCStateEncoder<t_ValType>* [ m_uiNumBaseEncoders ] )
{
  for( UInt uiEncoderId = 0; uiEncoderId < m_uiNumBaseEncoders; uiEncoderId++ )
  {
    m_papcVLCStateEncoders[ uiEncoderId ] = new VLCStateEncoder<t_ValType>( uiEncoderId, *this );
  }
  TNOK( VLCodewordBuffer<t_ValType>::xInitMaxDelay( uiMaxDelayInBits, m_papcVLCStateEncoders ) );
}


template<class t_ValType>
FORCEINLINE
VLCEncoder<t_ValType>::~VLCEncoder()
{
  for( UInt uiEncoderId = 0; uiEncoderId < m_uiNumBaseEncoders; uiEncoderId++ )
  {
    delete  m_papcVLCStateEncoders[ uiEncoderId ];
  }
  delete [] m_papcVLCStateEncoders;
}


template<class t_ValType>
FORCEINLINE
ErrVal
VLCEncoder<t_ValType>::startCabac()
{
  return Err::m_nOK;
}


template<class t_ValType>
FORCEINLINE
ErrVal
VLCEncoder<t_ValType>::finishCabac()
{
  for( UInt uiEncoderId = 0; uiEncoderId < m_uiNumBaseEncoders; uiEncoderId++ )
  {
    RNOK( m_papcVLCStateEncoders[ uiEncoderId ]->flush() );
  }
  RNOK( VLCodewordBuffer<VLCodeType>::finish() );
  return Err::m_nOK;
}


template<class t_ValType>
FORCEINLINE
ErrVal
VLCEncoder<t_ValType>::pushBack( const ProbBin& rcProbBin )
{
  m_papcVLCStateEncoders[ m_pauiStateToIdxMap[ rcProbBin.getStateIdx() ] ]->pushBack( rcProbBin.isLps() );
  return Err::m_nOK;
}


template<class t_ValType>
FORCEINLINE
ErrVal
VLCEncoder<t_ValType>::pushBack( const bool& rbBin )
{
  m_papcVLCStateEncoders[ m_uiStateZeroIdx ]->pushBack( rbBin );
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_VLCENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
