/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VLC_STATE_ENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_VLC_STATE_ENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_


#include "VideoCommonLib/VLCodingTables.h"
#include "VLCodewordBuffer.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_ValType>
class VLCStateEncoder : private VLCodingTable<t_ValType>
{
public:
  VLCStateEncoder( const UInt uiEncoderId, VLCodewordBuffer<t_ValType>& rcVLCodewordBuffer );
  virtual ~VLCStateEncoder()  {}

  void    pushBack( const bool& rbBin );
  ErrVal  flush   ();
                   
private:
  const UInt                    m_uiEncoderId;
  VLCodewordBuffer<t_ValType>&  m_rcCodewordBuffer;
  UInt                          m_uiState;
};


template<class t_ValType>
VLCStateEncoder<t_ValType>::VLCStateEncoder( const UInt uiEncoderId, VLCodewordBuffer<t_ValType>& rcVLCodewordBuffer )
: VLCodingTable<t_ValType>( VLCodingTables<t_ValType,true>::getVLCodingTable( uiEncoderId ) )
, m_uiEncoderId           ( uiEncoderId )
, m_rcCodewordBuffer      ( rcVLCodewordBuffer )
, m_uiState               ( 0 )
{
}


template<class t_ValType>
void
VLCStateEncoder<t_ValType>::pushBack( const bool& rbBin )
{
  if( m_uiState == 0 )
  {
    m_rcCodewordBuffer.reserve( m_uiEncoderId );
  }
  const StateTransition         eTransition = ( rbBin ? STATE_TRANSITION_BIT1 : STATE_TRANSITION_BIT0 );
  const VLCodeword<t_ValType>&  rcCodeword  = VLCodingTable<t_ValType>::updateState( m_uiState, eTransition );
  if( rcCodeword.getLength() )
  {
    m_rcCodewordBuffer.write( m_uiEncoderId, rcCodeword );
    m_uiState = 0;
  }
}


template<class t_ValType>
ErrVal
VLCStateEncoder<t_ValType>::flush()
{
  if( m_uiState )
  {
    const VLCodeword<t_ValType>& rcCodeword = VLCodingTable<t_ValType>::updateState( m_uiState, STATE_TRANSITION_END );
    ROF( rcCodeword.getLength() );
    m_rcCodewordBuffer.write( m_uiEncoderId, rcCodeword );
    m_uiState = 0;
  }
  return Err::m_nOK;
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_VLC_STATE_ENCODER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
