/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#if !defined(AFX_VLCODEWORD_BUFFER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
#define AFX_VLCODEWORD_BUFFER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_


#include "VideoCommonLib/BitWriterIf.h"
#include "VideoCommonLib/VLCodeword.h"


// video namespace begin
VIDEO_NAMESPACE_BEGIN


template<class t_ValType>
class VLCStateEncoder;


template<class t_ValType>
class VLCodewordBuffer
{
  enum { STD_CODEWORD_BUFFER_ALLOC_SIZE = 256 };

private:
  class BufferEntry
  {
  public:
    BufferEntry() : m_uiEncoderId ( TypeLimits<UInt>::getMax() ), m_cCodeword( VLCodeword<t_ValType>() ) {}
    void  clear() { m_uiEncoderId = TypeLimits<UInt>::getMax()  ; m_cCodeword.clear(); }
  public:
    UInt                  m_uiEncoderId;
    VLCodeword<t_ValType> m_cCodeword;
  };

public:
  VLCodewordBuffer( BitWriterIf<UInt>& rcBitWriterIf, const UInt uiNumVLCStateEncoders );
  virtual ~VLCodewordBuffer();

public:
  void    reserve ( const UInt uiEncoderId );
  void    write   ( const UInt uiEncoderId, const VLCodeword<t_ValType>& rcVLCodeword );
  ErrVal  finish  ();

protected:
  ErrVal  xInitMaxDelay         ( const UInt                    uiMaxDelayInBits,
                                  VLCStateEncoder<t_ValType>**  papcVLCStateEncoders );

private:
  void    xIncreaseBufferSize   ();
  void    xWrite                ();

private:
  BitWriterIf<UInt>&            m_rcBitWriterIf;                  
  const UInt                    m_uiNumVLCStateEncoders;
  UInt                          m_uiNextWriteIdx;
  UInt                          m_uiNextReserveIdx;
  UInt                          m_uiCodewordBufferSize;
  UInt                          m_uiNumFreeBufferEntries;
  UInt                          m_uiMaxCodewordLength;
  UInt                          m_uiMaxDelayBitCounter;
  UInt                          m_uiMaxDelayInBits;
  VLCStateEncoder<t_ValType>**  m_papcVLCStateEncoders;
  UInt*                         m_pauiMaxCodewordLength;
  UInt*                         m_pauiReservedIdBuffer;
  BufferEntry*                  m_pacCodewordBuffer;
};


template<class t_ValType>
VLCodewordBuffer<t_ValType>::VLCodewordBuffer( BitWriterIf<UInt>& rcBitWriterIf, const UInt uiNumVLCStateEncoders )
: m_rcBitWriterIf         ( rcBitWriterIf )
, m_uiNumVLCStateEncoders ( uiNumVLCStateEncoders )
, m_uiNextWriteIdx        ( 0 )
, m_uiNextReserveIdx      ( 0 )
, m_uiCodewordBufferSize  ( STD_CODEWORD_BUFFER_ALLOC_SIZE )
, m_uiNumFreeBufferEntries( m_uiCodewordBufferSize         )
, m_uiMaxCodewordLength   ( 0 )
, m_uiMaxDelayBitCounter  ( 0 )
, m_uiMaxDelayInBits      ( TypeLimits<UInt>::getMax() )
, m_papcVLCStateEncoders  ( 0 )
, m_pauiMaxCodewordLength ( new UInt                  [ m_uiNumVLCStateEncoders ] )
, m_pauiReservedIdBuffer  ( new UInt                  [ m_uiNumVLCStateEncoders ] )
, m_pacCodewordBuffer     ( new BufferEntry           [ m_uiCodewordBufferSize  ] )
{
  for( UInt uiIdx = 0; uiIdx < m_uiNumVLCStateEncoders; uiIdx++ )
  {
    m_pauiReservedIdBuffer  [ uiIdx ] = TypeLimits<UInt>::getMax();
    m_pauiMaxCodewordLength [ uiIdx ] = VLCodingTables<t_ValType,true>::getVLCodingTable( uiIdx ).getMaxCodewordLength();
    m_uiMaxCodewordLength             = gMax<UInt>( m_uiMaxCodewordLength, m_pauiMaxCodewordLength[ uiIdx ] );
  }
}


template<class t_ValType>
VLCodewordBuffer<t_ValType>::~VLCodewordBuffer()
{
  delete [] m_pauiMaxCodewordLength;
  delete [] m_pauiReservedIdBuffer;
  delete [] m_pacCodewordBuffer;
}


template<class t_ValType>
ErrVal VLCodewordBuffer<t_ValType>::xInitMaxDelay( const UInt uiMaxDelayInBits, VLCStateEncoder<t_ValType>** papcVLCStateEncoders )
{
  ROF( papcVLCStateEncoders );
  ROF( m_uiCodewordBufferSize == m_uiNumFreeBufferEntries );
  ROF( uiMaxDelayInBits       >= m_uiMaxCodewordLength    || uiMaxDelayInBits == 0 );
  m_uiMaxDelayInBits      = ( uiMaxDelayInBits ? uiMaxDelayInBits : TypeLimits<UInt>::getMax() );
  m_papcVLCStateEncoders  = papcVLCStateEncoders;
  return Err::m_nOK;
}


template<class t_ValType>
void VLCodewordBuffer<t_ValType>::reserve( const UInt uiEncoderId )
{
  TOF_DBG( uiEncoderId < m_uiNumVLCStateEncoders );
  TOF_DBG( m_pauiReservedIdBuffer[ uiEncoderId ] == TypeLimits<UInt>::getMax() );
  UInt uiMaxBitCounter = m_uiMaxDelayInBits - m_pauiMaxCodewordLength[ uiEncoderId ];
  while( m_uiMaxDelayBitCounter > uiMaxBitCounter )
  {
    m_papcVLCStateEncoders[ m_pacCodewordBuffer[ m_uiNextWriteIdx ].m_uiEncoderId ]->flush();
  }
  if( m_uiNumFreeBufferEntries == 0 )
  {
    xIncreaseBufferSize();
  }
  m_pacCodewordBuffer[ m_uiNextReserveIdx ].m_uiEncoderId = uiEncoderId;
  m_pauiReservedIdBuffer[ uiEncoderId ] = m_uiNextReserveIdx++;
  if( m_uiNextReserveIdx == m_uiCodewordBufferSize )
  {
    m_uiNextReserveIdx = 0;
  }
  m_uiNumFreeBufferEntries--;
  m_uiMaxDelayBitCounter += m_pauiMaxCodewordLength[ uiEncoderId ];
}


template<class t_ValType>
void VLCodewordBuffer<t_ValType>::write( const UInt uiEncoderId, const VLCodeword<t_ValType>& rcVLCodeword )
{
  TOF_DBG( uiEncoderId < m_uiNumVLCStateEncoders );
  TOF_DBG( m_pauiReservedIdBuffer[ uiEncoderId ] != TypeLimits<UInt>::getMax() );
  const UInt uiCurrIdx = m_pauiReservedIdBuffer[ uiEncoderId ];
  m_pauiReservedIdBuffer[ uiEncoderId ]         = TypeLimits<UInt>::getMax();
  m_pacCodewordBuffer[ uiCurrIdx ].m_cCodeword  = rcVLCodeword;
  if( uiCurrIdx == m_uiNextWriteIdx )
  {
    xWrite();
  }
}


template<class t_ValType>
ErrVal VLCodewordBuffer<t_ValType>::finish()
{
  ROF ( m_uiNumFreeBufferEntries == m_uiCodewordBufferSize );
  RNOK( m_rcBitWriterIf.write( 1 ) );
  return Err::m_nOK;
}


template<class t_ValType>
void VLCodewordBuffer<t_ValType>::xIncreaseBufferSize()
{
  UInt                    uiNewBufferSize       = m_uiCodewordBufferSize    + STD_CODEWORD_BUFFER_ALLOC_SIZE;
  UInt                    uiNewFreeEntries      = m_uiNumFreeBufferEntries  + STD_CODEWORD_BUFFER_ALLOC_SIZE;
  UInt                    uiNewNextWriteIdx     = 0;
  UInt                    uiNewNextReservedIdx  = m_uiCodewordBufferSize    - m_uiNumFreeBufferEntries;
  BufferEntry*            pacNewVLCBuffer       = new BufferEntry           [ uiNewBufferSize ];
  for( UInt uiSrcIdx = m_uiNextWriteIdx, uiDesIdx = uiNewNextWriteIdx; uiDesIdx < uiNewNextReservedIdx; uiDesIdx++, uiSrcIdx = ( uiSrcIdx + 1 ) % m_uiCodewordBufferSize )
  {
    pacNewVLCBuffer[ uiDesIdx ] = m_pacCodewordBuffer[ uiSrcIdx ];
    if( pacNewVLCBuffer[ uiDesIdx ].m_cCodeword.getLength() == 0 )
    {
      TOF_DBG( m_pauiReservedIdBuffer[ pacNewVLCBuffer[ uiDesIdx ].m_uiEncoderId ] == uiSrcIdx );
      m_pauiReservedIdBuffer[ pacNewVLCBuffer[ uiDesIdx ].m_uiEncoderId ] = uiDesIdx;
    }
  }
  delete [] m_pacCodewordBuffer;
  m_uiNextWriteIdx          = uiNewNextWriteIdx;
  m_uiNextReserveIdx        = uiNewNextReservedIdx;
  m_uiCodewordBufferSize    = uiNewBufferSize;
  m_uiNumFreeBufferEntries  = uiNewFreeEntries;
  m_pacCodewordBuffer       = pacNewVLCBuffer;
}


template<class t_ValType>
void
VLCodewordBuffer<t_ValType>::xWrite()
{
  while( m_pacCodewordBuffer[ m_uiNextWriteIdx ].m_cCodeword.getLength() )
  {
    BufferEntry&            rcBufferEntry = m_pacCodewordBuffer[ m_uiNextWriteIdx ];
    VLCodeword<t_ValType>&  rcCodeword    = rcBufferEntry.m_cCodeword;
    t_ValType               tValue        = rcCodeword.getValue  ();
    UInt                    uiNumBits     = rcCodeword.getLength ();
    //>>>>> the case inside should never happen -> can be removed later >>>>>
    if( sizeof( t_ValType ) > sizeof( UInt ) )
    {
      const UInt uiMaxBits = ( sizeof( UInt ) << 3 );
      while( uiNumBits > uiMaxBits )
      {
        uiNumBits  -= uiMaxBits;
        UInt  uiVal = UInt( tValue >> uiNumBits );
        tValue     &= ( t_ValType( 1 ) << uiNumBits ) - t_ValType( 1 );
        TNOK( m_rcBitWriterIf.write( uiVal, uiMaxBits ) );
      }
    }
    //<<<<< the case inside should never happen -> can be removed later <<<<<
    TNOK( m_rcBitWriterIf.write( UInt( tValue ), uiNumBits ) );
    m_uiMaxDelayBitCounter -= m_pauiMaxCodewordLength[ rcBufferEntry.m_uiEncoderId ];
    rcBufferEntry.clear();
    m_uiNextWriteIdx++;
    if( m_uiNextWriteIdx == m_uiCodewordBufferSize )
    {
      m_uiNextWriteIdx = 0;
    }
    m_uiNumFreeBufferEntries++;
  }
}


// video namespace end
VIDEO_NAMESPACE_END

#endif // !defined(AFX_VLCODEWORD_BUFFER_H__19223BF1_DA9C_475F_9AEE_D1A55237EB1E__INCLUDED_)
