/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef _VIDEOENCODERIMPL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
#define _VIDEOENCODERIMPL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_

#include "AUCodingStarter.h"
#include "VideoCommonLib/DecodingToDisplayOrderBuffer.h"
#include "VideoCommonLib/PocAndFrameNum.h"
#include "VideoEncoderLib/VideoEncoderIf.h"
#include "VideoEncoderLib/VideoEncoderInstance.h"
#include "AccessUnitEncoder.h"
#include "VideoCommonLib/CodedPictureStore.h"
#include "SequenceStructure.h"
#include "ReferencePictureEnc.h"
#include "VideoEncoderLib/VideoEncoderParameter.h"

#include "RelRefPOCGenerator.h"

// video namespace begin
VIDEO_NAMESPACE_BEGIN

class SequenceParameterSet;
class PictureParameterSet;


template<class T, class t_Container>
class PushBackSink
  : public VideoCodecSinkIf<T>
{
public:
  PushBackSink( t_Container &rcCont ) : m_rcCont( rcCont ) {}

  ErrVal receive( T pT ) { m_rcCont.push_back( pT ); return Err::m_nOK; }

private:
  t_Container &m_rcCont;
};


class VideoEncoderImpl
  : private VideoEncoderIf, VideoEncoderInstance
{
public:
  static ErrVal runAlternative( int argc, char **argv, bool &rbExit );

  static ErrVal create( VideoEncoderInstance *&rpcVideoEncoderInstance );
  ErrVal destroy();

  VideoEncoderIf* getVideoEncoderIf() { return this; }
  
  UInt64 getFrameEncodingDurationSumMSec() const { return m_uiFrameEncodingDurationSumMSec; }
  UInt64 getFrameMEDurationSumMSec()       const { return m_uiFrameMEDurationSumMSec; }

  ErrVal encodePicture( const VideoCodecPicture<const VideoCodecPel> *pcOriginalPic,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedPic,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedPrePic,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedResPic,
                        VideoCodecSinkIf<const VideoCodecPicture<const VideoCodecPel>*> &rcOriginalPicSinkIf,
                        VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> &rcReconstructedPicSinkIf,
                        VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                        VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                        VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
  {
    try
    {
      RNOK( xEncodePicture( pcOriginalPic, pcReconstructedPic, pcReconstructedPrePic, pcReconstructedResPic, rcOriginalPicSinkIf, rcReconstructedPicSinkIf, pcReconstructedPrePicSinkIf, pcReconstructedResPicSinkIf, rcBitstreamPacketSinkIf ) );
    }
    RETURN_ERR_ON_EXCEPTION;
    return Err::m_nOK;
  }

  ErrVal insertPicture( const VideoCodecPicture<const VideoCodecPel> *pcOriginalPic,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedPic,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedPrePic,
                        const VideoCodecPicture<VideoCodecPel> *pcReconstructedResPic )
  {
    try
    {
      RNOK( xInsertPicture( pcOriginalPic, pcReconstructedPic, pcReconstructedPrePic, pcReconstructedResPic ) );
    }
    RETURN_ERR_ON_EXCEPTION;
    return Err::m_nOK;
  }

  ErrVal encodeNextPicture( VideoCodecSinkIf<const VideoCodecPicture<const VideoCodecPel>*> &rcOriginalPicSinkIf,
                            VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> &rcReconstructedPicSinkIf,
                            VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                            VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                            VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
  {
    try
    {
      RNOK( xEncodeNextPicture( rcOriginalPicSinkIf, rcReconstructedPicSinkIf, pcReconstructedPrePicSinkIf, pcReconstructedResPicSinkIf, rcBitstreamPacketSinkIf ) );
    }
    RETURN_ERR_ON_EXCEPTION;
    return Err::m_nOK;
  }

  const VideoCodecPicture<const VideoCodecPel> *getNextPicture()
  {
      InputPicMap::iterator cIter;
      if( ( cIter = m_acInputPicMap.find( m_cSeqIter.getPoc() ) ) == m_acInputPicMap.end() )
      {
        return NULL;
      }
      return cIter->second.m_pcOriginalPic;
  }

  ErrVal setEncoderParameter( const VideoEncoderParameter* pcEncoderParameter );

  ErrVal getPSNR( std::vector<double> &radPSNR ) const;

  bool isUncodedPictureWaiting() const;
  bool hasUncodedPictures() const;

protected:
  VideoEncoderImpl();
  virtual ~VideoEncoderImpl();

private:
  typedef std::map<PicOrderCnt, OriginalAndReconstructedPic> InputPicMap;

  template <class t_Pel>
  ErrVal xEncodePicture( const VideoCodecPicture<const t_Pel> *pcOriginalPic,
                         const VideoCodecPicture<t_Pel> *pcReconstructedPic,
                         const VideoCodecPicture<t_Pel> *pcReconstructedPrePic,
                         const VideoCodecPicture<t_Pel> *pcReconstructedResPic,
                         VideoCodecSinkIf<const VideoCodecPicture<const t_Pel>*> &rcOriginalPicSinkIf,
                         VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                         VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                         VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                         VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf );

  ErrVal xCreateNextCodingItem( AccessUnitEncodingItem *&rpcAUCodingItem, const OriginalAndReconstructedPic &rcInputPics, VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf );

  template <class t_Pel>
  ErrVal xReleaseCodingItem( AccessUnitEncodingItem *pcAUCodingItem,
                             VideoCodecSinkIf<const VideoCodecPicture<const t_Pel>*> &rcOriginalPicSinkIf,
                             VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                             VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf );

  ErrVal xReleaseAllCodedItems( VideoCodecSinkIf<const VideoCodecPicture<const Pel>*> &rcOriginalPicSinkIf,
                                VideoCodecSinkIf<const VideoCodecPicture<Pel>*> &rcReconstructedPicSinkIf,
                                VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                                VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                                VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf,
                                bool bWaitForOne );

  ErrVal xSetAccessUnitEncoderParameters( AccessUnitParameters &rcAUPars,
                                          AccessUnitEncoderParameters &rcAUEstPars,
                                          std::ostream &rcOut );

  template <class t_Pel>
  ErrVal xInsertPicture( const VideoCodecPicture<const t_Pel> *pcOriginalPic,
                         const VideoCodecPicture<t_Pel> *pcReconstructedPic,
                         const VideoCodecPicture<t_Pel> *pcReconstructedPrePic,
                         const VideoCodecPicture<t_Pel> *pcReconstructedResPic );

  template <class t_Pel>
  ErrVal xEncodeNextPicture( VideoCodecSinkIf<const VideoCodecPicture<const t_Pel>*> &rcOriginalPicSinkIf,
                             VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                             VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf );

  ErrVal xWriteParameterSets( VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf );

  const VideoEncoderParameter *m_pcEncPar;
  CodedPictureStoreMT m_cCodedPictureStore;
  PacketParser<LargeFile> *m_pcPrecodedBitstreamParser;
  LargeFile m_cPrecodedBitstreamFile;
  UInt m_uiNumAUsToParse;
  PicOrderCnt m_cNextInputPoc;

  InputPicMap m_acInputPicMap;
  DecodingToDisplayOrderBuffer m_cDecToDispBuffer;
  DecodingToDisplayOrderBuffer m_cDecPreToDispBuffer;
  DecodingToDisplayOrderBuffer m_cDecResToDispBuffer;
  SequenceStructure m_cSequenceStructure;
  SequenceStructure::Iterator m_cSeqIter;
  SequenceStructure::Iterator m_cNextIter;

  std::vector<std::map<Int64,double> > m_acPSNR;

  RelRefPOCGenerator m_cRelRefPocGenerator;
  Int64 m_iCodOrder;
  Sequencer<Int64, std::list<std::vector<UChar> >*> m_cBitstreamSequencer;
  Sequencer<Int64, std::string> m_cConsoleOutputSequencer;
  AUCodingStarter m_cAUCodingStarter;
  UInt64 m_uiFrameEncodingDurationSumMSec;
  UInt64 m_uiFrameMEDurationSumMSec;
  SequenceParameterSet* m_pcSequenceParameterSet;
  PictureParameterSet*  m_pcPictureParameterSet;
  SequenceParameterSet* m_pcSequenceParameterSetParse;
  PictureParameterSet*  m_pcPictureParameterSetParse;
  PocAndFrameNum m_cPocAndFrameNum;
  UInt m_uiIdrPicId;
};


template <class t_Pel>
ErrVal VideoEncoderImpl::xReleaseCodingItem( AccessUnitEncodingItem *pcAUCodingItem,
                                             VideoCodecSinkIf<const VideoCodecPicture<const t_Pel>*> &rcOriginalPicSinkIf,
                                             VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                                             VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
{
  PicOrderCnt cPoc = pcAUCodingItem->getAUParameters().getPicOrderCnt();
  RNOK( m_cCodedPictureStore.markAUsAsUnused( 0, pcAUCodingItem->getPocsToRemoveFromStoreAfterCurrentPoc() ) );
  RNOK( m_cCodedPictureStore.markAUsAsUnused( cPoc, pcAUCodingItem->getAUParameters().getRefDeltaPocList() ) );

  m_uiFrameEncodingDurationSumMSec += pcAUCodingItem->getEncodingDurationMSec();
  m_uiFrameMEDurationSumMSec       += pcAUCodingItem->getMEDurationMSec();

  if( pcAUCodingItem->getAUParameters().getRefPicFlag() )
  {
    RNOK( m_cCodedPictureStore.addAU( pcAUCodingItem->getAccessUnit() ) );
  }
  else
  {
    delete pcAUCodingItem->getAccessUnit();
  }

  {
    std::vector<std::string> acOutput;
    PushBackSink<std::string&,std::vector<std::string> > cSink( acOutput );
    RNOK( m_cConsoleOutputSequencer.insertItem( pcAUCodingItem->getCodingOrder(), pcAUCodingItem->getOutput() ) );
    RNOK( m_cConsoleOutputSequencer.outputItemsIfPossible( cSink ) );
    for( UInt ui = 0; ui < UInt( acOutput.size() ); ui++ )
    {
      std::cout << acOutput[ui];
    }
    std::cout << std::flush;
  }

  RNOK( m_cBitstreamSequencer.insertItem( pcAUCodingItem->getCodingOrder(), pcAUCodingItem->getBitstreamData() ) );
  RNOK( m_cBitstreamSequencer.outputItemsIfPossible( rcBitstreamPacketSinkIf ) );

  ROF( m_acPSNR.size() == pcAUCodingItem->getAUParameters().getNumPlanes() );
  for( UInt uiPlane = 0; uiPlane < pcAUCodingItem->getAUParameters().getNumPlanes(); uiPlane++ )
  {
    m_acPSNR[uiPlane][pcAUCodingItem->getCodingOrder()] = pcAUCodingItem->getAUPSNR()[uiPlane];
  }

  if ( pcReconstructedPrePicSinkIf )
  {
    RNOK( m_cDecPreToDispBuffer.insertItem( cPoc, pcAUCodingItem->getInputPics().m_pcReconstructedPrePic ) );
    RNOK( m_cDecPreToDispBuffer.outputItemsIfPossible( *pcReconstructedPrePicSinkIf ) );
  }

  if ( pcReconstructedResPicSinkIf )
  {
    RNOK( m_cDecResToDispBuffer.insertItem( cPoc, pcAUCodingItem->getInputPics().m_pcReconstructedResPic ) );
    RNOK( m_cDecResToDispBuffer.outputItemsIfPossible( *pcReconstructedResPicSinkIf ) );
  }

  RNOK( m_cDecToDispBuffer.insertItem( cPoc, pcAUCodingItem->getInputPics().m_pcReconstructedPic ) );
  RNOK( rcOriginalPicSinkIf.receive( pcAUCodingItem->getInputPics().m_pcOriginalPic ) );

  RNOK( m_cDecToDispBuffer.outputItemsIfPossible( rcReconstructedPicSinkIf ) );

  delete pcAUCodingItem;

  return Err::m_nOK;
}


template <class t_Pel>
ErrVal VideoEncoderImpl::xEncodePicture( const VideoCodecPicture<const t_Pel> *pcOriginalPic,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedPic,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedPrePic,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedResPic,
                                         VideoCodecSinkIf<const VideoCodecPicture<const t_Pel>*> &rcOriginalPicSinkIf,
                                         VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                                         VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                                         VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                                         VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
{
  CAOF( (TypeCompare<t_Pel, Pel>::m_bEqual) );

  if( NULL != pcOriginalPic )
  {
    RNOK( xInsertPicture( pcOriginalPic, pcReconstructedPic, pcReconstructedPrePic, pcReconstructedResPic ) );
  }

  const bool bHasNoNewOriginalPictures = NULL == pcOriginalPic;
  while( isUncodedPictureWaiting() || ( bHasNoNewOriginalPictures && hasUncodedPictures() ) )
  {
    RNOK( xEncodeNextPicture( rcOriginalPicSinkIf, rcReconstructedPicSinkIf, pcReconstructedPrePicSinkIf, pcReconstructedResPicSinkIf, rcBitstreamPacketSinkIf ) );
  }

  if( bHasNoNewOriginalPictures )
  {
    while( m_cAUCodingStarter.getNumCodingItems() > 0 )
    {
      const UInt uiNumCodedItemsBefore = m_cAUCodingStarter.getNumCodingItems();

      RNOK( m_cAUCodingStarter.startCodableItems( m_cCodedPictureStore ) );
      RNOK( xReleaseAllCodedItems( rcOriginalPicSinkIf, rcReconstructedPicSinkIf, pcReconstructedPrePicSinkIf, pcReconstructedResPicSinkIf, rcBitstreamPacketSinkIf, true ) );
      ROF( m_cAUCodingStarter.getNumCodingItems() < uiNumCodedItemsBefore ); // avoid infinite loop
    }
  }

  return Err::m_nOK;
}


template <class t_Pel>
ErrVal VideoEncoderImpl::xInsertPicture( const VideoCodecPicture<const t_Pel> *pcOriginalPic,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedPic,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedPrePic,
                                         const VideoCodecPicture<t_Pel> *pcReconstructedResPic )
{
  CAOF( (TypeCompare<t_Pel, Pel>::m_bEqual) );

  ROT( NULL == pcOriginalPic );
  ROT( NULL == pcReconstructedPic );

  m_acInputPicMap.insert( std::make_pair( m_cNextInputPoc++, OriginalAndReconstructedPic( pcOriginalPic, pcReconstructedPic, pcReconstructedPrePic, pcReconstructedResPic ) ) );

  return Err::m_nOK;
}


template <class t_Pel>
ErrVal VideoEncoderImpl::xEncodeNextPicture( VideoCodecSinkIf<const VideoCodecPicture<const t_Pel>*> &rcOriginalPicSinkIf,
                                             VideoCodecSinkIf<const VideoCodecPicture<t_Pel>*> &rcReconstructedPicSinkIf,
                                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedPrePicSinkIf,
                                             VideoCodecSinkIf<const VideoCodecPicture<VideoCodecPel>*> *pcReconstructedResPicSinkIf,
                                             VideoCodecSinkIf<std::list<std::vector<UChar> >*> &rcBitstreamPacketSinkIf )
{
  CAOF( (TypeCompare<t_Pel, Pel>::m_bEqual) );
  ROT( m_acInputPicMap.empty() );

  InputPicMap::iterator cIter = m_acInputPicMap.find( m_cSeqIter.getPoc() );
  const bool bPictureAvailable = cIter != m_acInputPicMap.end();
  if( bPictureAvailable )
  {
    AccessUnitEncodingItem *pcAUCodingItem;
    RNOK( xCreateNextCodingItem( pcAUCodingItem, cIter->second, rcBitstreamPacketSinkIf ) );
    m_acInputPicMap.erase( cIter );
    RNOK( m_cAUCodingStarter.pushBackAUCodingItem( pcAUCodingItem ) );
    RNOK( m_cAUCodingStarter.startCodableItems( m_cCodedPictureStore ) );

    while( m_cAUCodingStarter.getNumCodingItems() > m_pcEncPar->getMultithreadingControlParams().getNumPicsInMultithreadingQueue() )
    {
      const UInt uiNumCodedItemsBefore = m_cAUCodingStarter.getNumCodingItems();
      RNOK( m_cAUCodingStarter.startCodableItems( m_cCodedPictureStore ) );
      RNOK( xReleaseAllCodedItems( rcOriginalPicSinkIf, rcReconstructedPicSinkIf, pcReconstructedPrePicSinkIf, pcReconstructedResPicSinkIf, rcBitstreamPacketSinkIf, true ) );
      ROF( m_cAUCodingStarter.getNumCodingItems() < uiNumCodedItemsBefore ); // avoid infinite loop
    }
  }
  else
  {
    std::cout << std::setw(8) << m_cSeqIter.getPoc() << " skipped due to sequence end" << std::endl;
  }

  RNOK( m_cRelRefPocGenerator.goToNextPoc( bPictureAvailable && m_cSeqIter.getFrameDescriptor().getStoreForRef() ) );

  // prepare for the next frame
  if( m_cSeqIter.getFrameDescriptor().getStoreForRef() )
  {
    m_cPocAndFrameNum.incFrameNum();
  }

  ++m_cSeqIter;
  ++m_cNextIter;

  return Err::m_nOK;
}

// video namespace end
VIDEO_NAMESPACE_END

#endif // #ifndef _VIDEOENCODERIMPL_H__E43399A8_CF59_4184_984E_FAB9FC0DF52B__INCLUDED_
