/* ====================================================================================================================

  The copyright in this software is being made available under the license included below. 
  This software may be subject to other third party and contributor rights, including patent rights, and no such 
  rights are granted under this license.
 
  Copyright (c) 2010, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for 
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and 
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V. nor the
      names of its contributors may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#ifndef WIENERCORRESTIMATOR_H_
#define WIENERCORRESTIMATOR_H_

#include "VideoCommonLib/WienerFilterParameter.h"

VIDEO_NAMESPACE_BEGIN


template<class t_Pel>
class WienerCorrEstimatorBase
{
public:
  WienerCorrEstimatorBase( const PicBlock<Pel> &rcDecPic , const PicBlock<const Pel> &rcOrigPic ,std::vector<UInt> auiSets ,  std::vector<UInt> auiFilterType, std::vector<UInt>auiFilterLength )
  : m_rcDecPic( rcDecPic )
  , m_rcOrigPic( rcOrigPic )
  , m_auiSets( auiSets )
  {
    m_aadCorr.clear();
    for( UInt uiSetNum=0; uiSetNum < auiSets.size(); uiSetNum++ )
    {
        FilterCoefficients cTemp;
        cTemp.setFilterTypeSep( auiFilterType[uiSetNum] , auiFilterLength[uiSetNum] );
        m_auiNumOfCoeffs.push_back( cTemp.getNumberOfCoeffs() );
        m_auiEstLength.push_back( auiFilterLength[ uiSetNum ] );
        m_auiFilterType.push_back( auiFilterType[ uiSetNum ] );
        m_aauiMapping.push_back( cTemp.getMappingRule() );

        m_aadCorr.push_back(std::vector< std::vector<double> >() );

        for( UInt uiRow = 0; uiRow < m_auiNumOfCoeffs[ uiSetNum ] + 1; uiRow ++ )
        {
            m_aadCorr[uiSetNum].push_back(std::vector<double>());
            m_aadCorr[uiSetNum][uiRow].assign( m_auiNumOfCoeffs[ uiSetNum ] + 2, 0);
        }

    }
  }

  //virtual ~WienerCorrEstimator();

  ErrVal addCorrHorizontal();
  ErrVal addCorrVertical();
  ErrVal addCorrHorizontal(UInt uiPlane, const Block<t_Pel> &rcDecBlock, const Block<const t_Pel> &rcOrigBlock);
  ErrVal addCorrVertical(UInt uiPlane, const Block<t_Pel> &rcDecBlock, const Block<const t_Pel> &rcOrigBlock);
  std::vector< std::vector<double> > &getCorr( UInt uiPlane ) { xPopulateSymElements(); return m_aadCorr[ m_auiSets[ uiPlane ] - 1]; }
  std::vector< std::vector<double> > getCorrSeparable( UInt uiPlane , UInt uiFilterLength);

protected:
    ErrVal xPopulateSymElements()
    {
         for(UInt uiSet=0; uiSet < m_auiNumOfCoeffs.size(); uiSet++ )
            for(UInt j=0; j < m_auiNumOfCoeffs[ uiSet ] - 1; j++)
                for(UInt i=j+1; i < m_auiNumOfCoeffs[ uiSet ]; i++)
                         m_aadCorr[uiSet][i][j] = m_aadCorr[uiSet][j][i];
         return Err::m_nOK;
    }

    std::vector<UInt> m_auiNumOfCoeffs;
    std::vector<UInt> m_auiEstLength;
    std::vector<UInt> m_auiFilterType;
    const PicBlock<Pel> &m_rcDecPic;
    const PicBlock<const Pel> &m_rcOrigPic;
    std::vector<UInt> m_auiSets;
    std::vector< std::vector<UInt> > m_aauiMapping;

    std::vector< std::vector< std::vector<double> > > m_aadCorr;


};


template<class t_Pel>
std::vector< std::vector<double> > WienerCorrEstimatorBase<t_Pel>::getCorrSeparable( UInt uiPlane , UInt uiFilterLength)
{

    if( m_auiEstLength[ m_auiSets[uiPlane] -1 ] == uiFilterLength )
      return getCorr( uiPlane );

    AOF( m_auiEstLength[ m_auiSets[uiPlane] -1 ] > uiFilterLength );
    UInt uiLengthDiff = m_auiEstLength[ m_auiSets[uiPlane] -1 ] - uiFilterLength;
    UInt uiMargin = uiLengthDiff >> 1;

    std::vector< std::vector<double> > cCorrNew;
    FilterCoefficients cTemp;
    cTemp.setFilterTypeSep( m_auiFilterType[ m_auiSets[ uiPlane ] -1] , uiFilterLength );

    for( UInt uiRow = 0; uiRow < UInt(cTemp.getNumberOfCoeffs() + 1); uiRow ++ )
    {
       cCorrNew.push_back(std::vector<double>());
       cCorrNew[uiRow].assign( cTemp.getNumberOfCoeffs() + 2, 0);
    }


    std::vector< std::vector<double> > &rcBaseCorr = getCorr( uiPlane );
    std::vector<UInt> auiOldCoeffs;

    for( UInt uiCoeff = 0; uiCoeff < UInt(cTemp.getNumberOfCoeffs()); uiCoeff ++ )
    {
       UInt uiPos = 0;
       while (cTemp.getMappingRule()[ uiPos ] != uiCoeff )
           uiPos++;


       UInt uiOldPos = uiMargin + uiPos;
       auiOldCoeffs.push_back(m_aauiMapping[m_auiSets[uiPlane] -1][uiOldPos]);
    }


    for ( UInt uiRow = 0; uiRow < UInt(cTemp.getNumberOfCoeffs()); uiRow++ )
    {
        for ( UInt uiCol = 0; uiCol < UInt(cTemp.getNumberOfCoeffs()); uiCol++ )
        {
            cCorrNew[uiRow][uiCol] = rcBaseCorr[auiOldCoeffs[uiRow]][auiOldCoeffs[uiCol]];
        }

        cCorrNew[cTemp.getNumberOfCoeffs()][uiRow] = rcBaseCorr[ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1 ] ][auiOldCoeffs[uiRow]];
        cCorrNew[uiRow][cTemp.getNumberOfCoeffs()] = rcBaseCorr[auiOldCoeffs[uiRow]][ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1] ];
        cCorrNew[uiRow][cTemp.getNumberOfCoeffs() + 1] = rcBaseCorr[auiOldCoeffs[uiRow]][ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1 ] + 1];
    }


    cCorrNew[cTemp.getNumberOfCoeffs()][cTemp.getNumberOfCoeffs()] = rcBaseCorr[ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1 ] ][ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1 ] ];
    cCorrNew[cTemp.getNumberOfCoeffs()][cTemp.getNumberOfCoeffs() + 1] = rcBaseCorr[ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1 ] ][ m_auiNumOfCoeffs[ m_auiSets[uiPlane] -1 ] + 1];


      //correlation matrix is symmetric, fill out empty places
      for(UInt j=0; j < cTemp.getNumberOfCoeffs() - 1; j++)
        for(UInt i=j+1; i < cTemp.getNumberOfCoeffs(); i++)
          cCorrNew[i][j] = cCorrNew[j][i];


    return cCorrNew;
}


template<class t_Pel>
ErrVal WienerCorrEstimatorBase<t_Pel>::addCorrHorizontal()
{

    for(UInt uiPlane = 0; uiPlane < m_rcDecPic.getStats().getNumPlanes() ; uiPlane++)
    {
        if ( m_auiSets[ uiPlane ] > 0 )
           addCorrHorizontal( m_auiSets[ uiPlane ] -1 ,m_rcDecPic.getPlane(uiPlane),m_rcOrigPic.getPlane(uiPlane));
    }

    return Err::m_nOK;
}

template<class t_Pel>
ErrVal WienerCorrEstimatorBase<t_Pel>::addCorrVertical()
{

    for(UInt uiPlane = 0; uiPlane < m_rcDecPic.getStats().getNumPlanes() ; uiPlane++)
    {
        if ( m_auiSets[ uiPlane ] > 0 )
           addCorrVertical( m_auiSets[ uiPlane ] -1 ,m_rcDecPic.getPlane(uiPlane),m_rcOrigPic.getPlane(uiPlane));
    }

    return Err::m_nOK;
}


template<class t_Pel>
ErrVal WienerCorrEstimatorBase<t_Pel>::addCorrHorizontal(UInt uiSet, const Block<t_Pel> &rcDecBlock, const Block<const t_Pel> &rcOrigBlock)
{

      int uiPosX = 0;
      UInt uiHorizontalFilterLength = m_auiEstLength[ uiSet ];
      int iHorizontalOverlap = uiHorizontalFilterLength >> 1;
      UInt uiHorizontalNumberOfCoeffs = m_auiNumOfCoeffs[ uiSet ];
      std::vector< UInt > auiHorizontalMapping = m_aauiMapping[ uiSet ];

      std::vector< std::vector<double> > &rcCorrHor = m_aadCorr[ uiSet ];

      // th subject of optimization
      rcCorrHor.clear() ;
      for (UInt i = 0; i < uiHorizontalNumberOfCoeffs + 1; i++)
      {
        rcCorrHor.push_back(std::vector<double>());
        rcCorrHor[i].assign(uiHorizontalNumberOfCoeffs + 2, 0);
      }


      UInt uiValueJ ;
      UInt uiHorizontalValueI ;


      // th subject of optimization
      // horizontal correlation
      for (int y = 0; y < rcDecBlock.getSize().getY() ; y++)
      {
        for (int x = 0; x < rcDecBlock.getSize().getX() ;x++)  // x-position of mask
        {
          for (UInt i = 0; i<uiHorizontalFilterLength; i++)
          {
            uiPosX              =  std::max(0 , int(x - iHorizontalOverlap + i) );
            uiPosX              =  std::min(rcDecBlock.getSize().getX() - 1 , uiPosX);
            uiHorizontalValueI  = rcDecBlock.getEl(PixelPos(uiPosX, y));

            rcCorrHor[auiHorizontalMapping[i]][uiHorizontalNumberOfCoeffs] += uiHorizontalValueI ;
            rcCorrHor[uiHorizontalNumberOfCoeffs][auiHorizontalMapping[i]] += uiHorizontalValueI ;
            rcCorrHor[auiHorizontalMapping[i]][uiHorizontalNumberOfCoeffs+1] += uiHorizontalValueI *rcOrigBlock.getEl(PixelPos(x, y)) ;

            for (UInt j = 0; j<uiHorizontalFilterLength; j++)
            {
              uiPosX        =  std::max(0 , int(x - iHorizontalOverlap + j));
              uiPosX        =  std::min(rcDecBlock.getSize().getX() - 1 , uiPosX);
              uiValueJ      = rcDecBlock.getEl(PixelPos(uiPosX, y));
              rcCorrHor[auiHorizontalMapping[i]][auiHorizontalMapping[j]] += uiHorizontalValueI * uiValueJ;
            }
          }
          rcCorrHor[uiHorizontalNumberOfCoeffs][uiHorizontalNumberOfCoeffs + 1] += rcOrigBlock.getEl(PixelPos(x,y));
          rcCorrHor[uiHorizontalNumberOfCoeffs][uiHorizontalNumberOfCoeffs ]    += rcDecBlock.getEl(PixelPos(x,y));
        }
      }

  return Err::m_nOK;
}


template<class t_Pel>
ErrVal WienerCorrEstimatorBase<t_Pel>::addCorrVertical(UInt uiSet, const Block<t_Pel> &rcDecBlock, const Block<const t_Pel> &rcOrigBlock)
{

      int uiPosY = 0;
      UInt uiVerticalFilterLength = m_auiEstLength[ uiSet ];
      int iVerticalOverlap = uiVerticalFilterLength >> 1;
      UInt uiVerticalNumberOfCoeffs = m_auiNumOfCoeffs[ uiSet ];
      std::vector< UInt > auiVerticalMapping = m_aauiMapping[ uiSet ];

      std::vector< std::vector<double> > &rcCorrVer = m_aadCorr[ uiSet ];

      // th subject of optimization
      rcCorrVer.clear() ;
      for (UInt i = 0; i < uiVerticalNumberOfCoeffs + 1; i++)
      {
        rcCorrVer.push_back(std::vector<double>());
        rcCorrVer[i].assign(uiVerticalNumberOfCoeffs + 2, 0);
      }


      UInt uiValueJ ;
      UInt uiVerticalValueI ;


      // th subject of optimization
      // Vertical correlation
      for (int y = 0; y < rcDecBlock.getSize().getY(); y++)
        {
          for (int x = 0; x < rcDecBlock.getSize().getX();x++)  // x-position of mask
          {
            for (UInt i = 0; i<uiVerticalFilterLength; i++)
            {

              uiPosY                =  std::max(0 , int(y - iVerticalOverlap + i));
              uiPosY                =  std::min(rcDecBlock.getSize().getY() - 1 , uiPosY);

              uiVerticalValueI      = rcDecBlock.getEl(PixelPos(x, uiPosY));
              rcCorrVer[auiVerticalMapping[i]][uiVerticalNumberOfCoeffs] += uiVerticalValueI ;
              rcCorrVer[uiVerticalNumberOfCoeffs][auiVerticalMapping[i]] += uiVerticalValueI ;
              rcCorrVer[auiVerticalMapping[i]][uiVerticalNumberOfCoeffs+1] += uiVerticalValueI *rcOrigBlock.getEl(PixelPos(x, y)) ;
              for (UInt j = 0; j<uiVerticalFilterLength; j++)
              {

                uiPosY        =  std::max(0 , int(y - iVerticalOverlap + j));
                uiPosY        =  std::min(rcDecBlock.getSize().getY() - 1 , uiPosY);
                uiValueJ      = rcDecBlock.getEl(PixelPos(x, uiPosY));
                rcCorrVer[auiVerticalMapping[i]][auiVerticalMapping[j]] += uiVerticalValueI * uiValueJ ;
              }
            }
            rcCorrVer[uiVerticalNumberOfCoeffs][uiVerticalNumberOfCoeffs + 1] += rcOrigBlock.getEl(PixelPos(x,y));
            rcCorrVer[uiVerticalNumberOfCoeffs][uiVerticalNumberOfCoeffs ]    += rcDecBlock.getEl(PixelPos(x,y));
          }
        }


  return Err::m_nOK;
}


VIDEO_NAMESPACE_END

#endif /* WIENERCORRESTIMATOR_H_ */
